#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

#include "algorithm.h"
#include "GTXLWriter.h"
#include "hypergraph.h"
#include "gts.h"
#include <string.h>


using namespace data_structures;

namespace algorithms {



namespace writers {

using namespace data_structures;

GTXLWRITER::GTXLWRITER()
{
    /*int n = strlen(pfilename);
    
    filename = new char[n+1];
    strcpy(filename,pfilename);
    doc = xmlNewDoc((const xmlChar *)"1.0");
    xmlNodePtr root = xmlNewNode(NULL,(const xmlChar *)"gxl");
    xmlDocSetRootElement(doc,root); */
}


GTXLWRITER::~GTXLWRITER()
{
    xmlFreeDoc(doc);
    delete filename;
}
    
void GTXLWRITER::initialize(char *pfilename)
{
    int n = strlen(pfilename);
    
    filename = new char[n+1];
    strcpy(filename,pfilename);
    doc = xmlNewDoc((const xmlChar *)"1.0");
    
    xmlNodePtr info = xmlNewNode(NULL, (const xmlChar*)"! DOCTYPE gtxl-document");
    xmlNodePtr root = xmlNewNode(NULL,(const xmlChar *)"GTS");
    
    xmlDocSetRootElement(doc,root);
}
    
void GTXLWRITER::run(GTS &GraphGrammar)
{
	xmlNodePtr xml_gts = xmlDocGetRootElement(doc);
	//xmlNodePtr xml_gts = xmlNewChild(root, NULL, (const xmlChar*)"GTS", NULL);
	xmlSetProp(xml_gts,(const xmlChar *)"id",(const xmlChar *)"id");
	
	xmlNodePtr initial = xmlNewChild(xml_gts, NULL,(const xmlChar*)"Initial", NULL);
	
	HYPERGRAPH *hg;
	hg = GraphGrammar.getStartgraph();
	WriteHypergraph(initial, *hg);
//HIER!	
	// variables for initializing ids for Rulegraph and Mapping
	IDTYPE id_initialize = 1;
	string str0 = "initialized id";
	string id_str;
	char id_char[15];
	
	map<IDTYPE, RULE>* rules_tmp = GraphGrammar.getRules();
	for(map<IDTYPE, RULE>::iterator it = rules_tmp->begin(); it != rules_tmp->end() && !rules_tmp->empty(); it++)
	{
		xmlNodePtr rule_xml = xmlNewChild(xml_gts, NULL, (const xmlChar*)"Rule", NULL);

		//timur
                list<EXPRESSION> le=(it->second).getGuard();
                string s_guard = "";
		for(list<EXPRESSION>::iterator itl = le.begin(); itl != le.end() && !le.empty(); itl++)
		{
		   if(itl != le.begin())
			s_guard += ",";
		   if(itl->getName() != "")
			s_guard += itl->getExpr();
		}
		WriteRuleGuard(rule_xml, s_guard);
		//end 

		xmlSetProp(rule_xml,(const xmlChar *)"id", (const xmlChar *)((it->second).getName()).c_str());
		xmlNodePtr lhs = xmlNewChild(rule_xml, NULL, (const xmlChar *)"LHS", NULL);
		xmlNodePtr RuleGraph = xmlNewChild(lhs, NULL, (const xmlChar *)"RuleGraph", NULL);
		
		sprintf(id_char,"%d\0",id_initialize);
		id_str = str0 + id_char;
		id_initialize++;
		xmlSetProp(RuleGraph, (const xmlChar*)"id", (const xmlChar*)id_str.c_str());
		
		HYPERGRAPH *hg_left = (it->second).getLeft();
		WriteHypergraph(RuleGraph, *hg_left);
		//WriteHypergraph(RuleGraph, it->second)
		
		
		xmlNodePtr rhs = xmlNewChild(rule_xml, NULL, (const xmlChar *)"RHS", NULL);
		RuleGraph = xmlNewChild(rhs, NULL, (const xmlChar *)"RuleGraph", NULL);
		
		sprintf(id_char,"%d\0",id_initialize);
		id_str = str0 + id_char;
		id_initialize++;
		xmlSetProp(RuleGraph, (const xmlChar*)"id", (const xmlChar*)id_str.c_str());
		
		HYPERGRAPH *hg_right = (it->second).getRight();
		WriteHypergraph(RuleGraph, *hg_right);
		
		//write mappings
		xmlNodePtr mapping = xmlNewChild(rule_xml, NULL, (const xmlChar*)"Mapping", NULL);
		
		sprintf(id_char,"%d\0",id_initialize);
		id_str = str0 + id_char;
		id_initialize++;
		xmlSetProp(mapping, (const xmlChar*)"id", (const xmlChar*)id_str.c_str());
		
		map<IDTYPE, IDTYPE> match_tmp = (it->second).getInterface().getMatch();
		for(map<IDTYPE, IDTYPE>::iterator itm = match_tmp.begin();
			itm != match_tmp.end() && !match_tmp.empty(); itm++)
		{
			xmlNodePtr map_elem = xmlNewChild(mapping, NULL, (const xmlChar*)"MapElem", NULL);
			if((it->second).getLeft()->hasVertex(itm->first))
				xmlSetProp(map_elem, (const xmlChar*)"from", (const xmlChar*)(it->second).getLeft()->getVertex(itm->first).getIdLabel().c_str());
			if((it->second).getLeft()->hasEdge(itm->first))
				xmlSetProp(map_elem, (const xmlChar*)"from", (const xmlChar*)(it->second).getLeft()->getEdge(itm->first).getIdLabel().c_str());
			if((it->second).getRight()->hasVertex(itm->second))
				xmlSetProp(map_elem, (const xmlChar*)"to", (const xmlChar*)(it->second).getRight()->getVertex(itm->second).getIdLabel().c_str());
			if((it->second).getRight()->hasEdge(itm->second))
				xmlSetProp(map_elem, (const xmlChar*)"to", (const xmlChar*)(it->second).getRight()->getEdge(itm->second).getIdLabel().c_str());
			
		}
	}
	
	
    if(-1 == xmlSaveFormatFile(filename,doc,true))
    {
      cout << "Error writing XML file\n";
      exit(1);
    }	
}
    

     
void GTXLWRITER::WriteHypergraph(xmlNodePtr node, HYPERGRAPH &hg)
{
	xmlNodePtr graph = xmlNewChild(node, NULL, (const xmlChar *)"Graph", NULL);
	xmlSetProp(graph, (const xmlChar *)"id",(const xmlChar *)(hg.getName()).c_str());
	xmlSetProp(graph, (const xmlChar*)"edgeids", (const xmlChar*)"true");
	xmlSetProp(graph, (const xmlChar*)"hypergraph", (const xmlChar*)"true");
	xmlSetProp(graph, (const xmlChar*)"edgemode", (const xmlChar*)"undirected");

	//start to write nodes
	
	map<IDTYPE, VERTEX> map_tmp = hg.getVertices();
	for(map<IDTYPE, VERTEX>::iterator it = map_tmp.begin(); it != map_tmp.end() && !map_tmp.empty(); it++)
	{
		xmlNodePtr node = xmlNewChild(graph, NULL, (const xmlChar*)"node", NULL);
		
		xmlSetProp(node,(const xmlChar*)"id",(const xmlChar*)(it->second).getIdLabel().c_str());
	}
	
	//start to write edges (describe all edges as hyperedges)
	map<IDTYPE, EDGE> edges_tmp = hg.getEdges();
	for(map<IDTYPE, EDGE>::iterator it = edges_tmp.begin(); it != edges_tmp.end() && !edges_tmp.empty(); it++)
	{
		xmlNodePtr edge = xmlNewChild(graph, NULL, (const xmlChar*)"rel", NULL);
		xmlSetProp(edge, (const xmlChar*)"id", (const xmlChar*)(it->second).getIdLabel().c_str());
		xmlNodePtr attr = xmlNewChild(edge, NULL, (const xmlChar*)"attr", NULL);
		xmlSetProp(attr, (const xmlChar*)"name", (const xmlChar*)"label");
		xmlNewChild(attr, NULL, (const xmlChar*)"string",(const xmlChar*)(it->second).getLabel().c_str());
		
		list<IDTYPE> relends = hg.getConnectedVertices((it->second).getID());

		//timur
                //insert attribute
            list <EXPRESSION> expression=(it->second).getAttributes();
			if(!expression.empty()){
                for (list<EXPRESSION>::iterator exp_it=expression.begin(); exp_it != expression.end(); exp_it++)
                {
                   TYPES types=exp_it->getTYPES();
                   string attrname=exp_it->getName();
                   string attrvalue=exp_it->getExpr();

                   int size=5*types.size()+5;
                   char attrtype[size];
                   strcpy(attrtype, "attr");
                   for (vector<T>::iterator type_it=types.begin(); type_it != types.end(); type_it++)
                   {
                      switch(*type_it) {
                         case INTEGER: {
                            strcat(attrtype, "-int");
                            break;
                         }
                         case STRING: {
                            strcat(attrtype, "-str");
                            break;
                         }
                         case BOOL: {
                            strcat(attrtype, "-bool");
                            break;
                         }
                         default:assert(false);
                      }
                   }
                   xmlNodePtr attr = xmlNewChild(edge, NULL, (const xmlChar*)"attr", NULL);
                   xmlSetProp(attr, (const xmlChar*)"kind", (const xmlChar*)attrtype);
                   xmlSetProp(attr, (const xmlChar*)"name", (const xmlChar*)attrname.c_str());
                   xmlNewChild(attr, NULL, (const xmlChar*)"string",(const xmlChar*)attrvalue.c_str());
                }
			}
		//end insert attribute

		VERTEX v;
		int order = 0;
		string s;
		char buffer[25];
		for(list<IDTYPE>::iterator itr = relends.begin(); itr != relends.end() && !relends.empty(); itr++)
		{
			xmlNodePtr relend = xmlNewChild(edge, NULL, (const
			xmlChar*)"relend", NULL);
			v = hg.getVertex(*itr);
			xmlSetProp(relend, (const xmlChar*)"target", (const xmlChar*)v.getIdLabel().c_str());
			sprintf(buffer, "%i", order);
  			xmlSetProp(relend, (const xmlChar*)"startorder", (const xmlChar*)buffer);
			order++;
		}
		
	}
	
}

//timur
void GTXLWRITER::WriteRuleGuard(xmlNodePtr &node, string exp)
{
	if(exp != ""){
		xmlNodePtr p1 = xmlNewChild(node, NULL, (const xmlChar*)"precondition", NULL);
		xmlNodePtr p2= xmlNewChild(p1, NULL, (const xmlChar*)"condition",NULL);
		xmlNodePtr p3= xmlNewChild(p2, NULL, (const xmlChar*)"attrCondition",NULL);
        xmlNewChild(p3, NULL, (const xmlChar*)"string",(const xmlChar*)exp.c_str());
	}
}

};
};
