#include "aggreader.h"
#include "GTXLWriter.h"
#include "id_checker.h"

using namespace algorithms;

using namespace readers;

// using namespace algorithms::readers::new_gtxl_reader;

AGGREADER::AGGREADER()
{

}

AGGREADER::~AGGREADER()
{

}

TYPES AGGREADER::types(string input){
	TYPES result;
	string type;
	for(int i = 0, k = 0; i < input.size();i++){
		if(input[i] == '-'){
			type = substr(input,k,i-1);
			k = i+1;
		}
		else if(i == (input.size()-1)){
			type = substr(input,k,i);
		}
		if(input[i] == '-' || i == (input.size()-1)){
			if(type == "int")
				result.push_back(INTEGER);
			else if(type == "str" || type == "string")
				result.push_back(STRING);
			else if(type == "bool")
				result.push_back(BOOL);
			else{
				cerr << "The type: " << type << " is not supported yet" << endl;
				exit(1);
			}
		}
	}
	return result;
}

string AGGREADER::getProp(xmlNodePtr ptr, char* name){
	return (string)(char*)xmlGetProp(ptr, (const xmlChar *)name);
}

xmlNodePtr AGGREADER::GetNextChild(xmlNodePtr child1, char* name_child)
{
	 xmlNodePtr child2 = child1->next;
	while((child2 != NULL) && (xmlStrcasecmp(child2->name, (const xmlChar*)name_child)))
		child2 = child2->next;
	return child2;
}
xmlNodePtr AGGREADER::GetChildrenNode(xmlNodePtr parent, char* name_child)
{
	xmlNodePtr child = parent->xmlChildrenNode;
	while((child != NULL) && (xmlStrcasecmp(child->name, (const xmlChar*)name_child)))
		child = child->next;
	return child;	
}

void AGGREADER::initialize(char* filename){
	doc = xmlParseFile(filename);
	if(doc == 0){
		cerr << "Parse error. File could not be read" << endl;
		xmlFreeDoc(doc);
		exit(1);
	}
}

string AGGREADER::between(string search, char first, char second){
	int indexFirst = -1;
	int indexSecond = -1;
	for(int i = 0; i < search.size() && indexFirst == -1; i++)
		if(search[i] == first)
			indexFirst = i;
	for(int j = indexFirst+1; j < search.size() && indexSecond == -1; j++)
		if(search[j] == second)
			indexSecond = j;
	if(indexFirst == -1 || indexSecond == -1){
		cerr << "No occurrence of the sequence: " << first << " or " << second << " int the word: " << search << endl;
		exit(1);
	}
	char* temp = new char[indexSecond-indexFirst];
	for(int k = indexFirst+1, l = 0; k < indexSecond; k++, l++)
		temp[l] = search[k];
	temp[indexSecond-indexFirst-1] = 0;
	string result(temp);
	delete temp;
	return result;
}

string AGGREADER::untill(string search, char till){
	int index = -1;
	for(int i = 0; i < search.size() && index == -1; i++)
		if(search[i] == till)
			index = i;
	if(index == -1){
		cerr << "There is no occurrence of " << till << " in the string: " << search << endl;
		exit(1);
	}
	char* temp = new char[index+1];
	for(int j = 0; j < index; j++)
		temp[j] = search[j];
	temp[index] = 0;
	string result(temp);
	delete temp;
	return result;
}

IDTYPE AGGREADER::getID(string xmlID){
	char* temp = new char[xmlID.size()-1];
	for(int i = 1; i < xmlID.size(); i++)
		temp[i-1] = xmlID[i];
	int ret = atoi(temp);
	delete temp;
	return ret;
}

void AGGREADER::readTypes(xmlNodePtr types){
	string prop, type, name, xmlID;
	int attrCounter;
	for(xmlNodePtr typ = GetChildrenNode(types, "NodeType"); typ != 0;
			typ = GetNextChild(typ, "NodeType")){
		attrCounter = 0;
		for(xmlNodePtr attr = GetChildrenNode(typ, "AttrType"); 
			attr != 0; attr = GetNextChild(attr, "AttrType")){
			EXPRESSION ex;
			ex.addName(getProp(attr, "attrname"));
			exprs.insert(make_pair(getProp(attr, "ID"), ex));
			attrCounter++;
		}
		prop = getProp(typ, "name");
		xmlID = getProp(typ, "ID");
		numOfExprs.insert(make_pair(xmlID, attrCounter));
		name = untill(prop, '%');
		type = between(prop, ':',':');
		if(type == "RECT"){//EDGE
			EDGE edge;
			edge.setLabel(name);
			edges.insert(make_pair(xmlID,edge));
		}
		else if(type == "CIRCLE"){//NODE
			VERTEX v;
			v.setLabel(name);
			vertexes.insert(make_pair(xmlID,v));
		}
	}
	for(xmlNodePtr edgeTyp = GetChildrenNode(types, "EdgeType"); edgeTyp != 0;
			edgeTyp = GetNextChild(edgeTyp, "EdgeType")){
		prop = getProp(edgeTyp, "name");
		xmlID = getProp(edgeTyp, "ID");
		name = untill(prop, '%');
		lines.insert(make_pair(xmlID, atoi(name.c_str())));
	}
}

bool AGGREADER::isVertex(string ID){
	for(map<string, VERTEX>::iterator iter = vertexes.begin();
		iter != vertexes.end(); iter++)
		if(iter->first == ID)
			return true;
	return false;
}

HYPERGRAPH AGGREADER::readHypergraph(xmlNodePtr graph){
	HYPERGRAPH hg;
	list<EXPRESSION> listOfExprs;
	TYPES typen;
	hg.setID(getID(getProp(graph, "ID")));
	hg.setName(getProp(graph, "name"));
	for(xmlNodePtr node = GetChildrenNode(graph, "Node"); node != 0; node = GetNextChild(node, "Node")){
		listOfExprs.clear();
		for(xmlNodePtr attr = GetChildrenNode(node, "Attribute"); attr != 0; attr = GetNextChild(attr, "Attribute")){
			EXPRESSION ex = get_2(exprs, getProp(attr, "type"));
			xmlNodePtr value = GetChildrenNode(GetChildrenNode(GetChildrenNode(attr, "Value"), "java"), "string");
			value = value->xmlChildrenNode;
			string val = (string)(char*)value->content;
			int index = val.find(":");
			string type = substr(val, 0, index-1);
			val = substr(val, index+1, val.size()-1);
			typen = types(type);
			ex.addType(typen);
			ex.addExpr(val);
			listOfExprs.push_back(ex);
		}
		string xmlID = getProp(node, "type");
		if(listOfExprs.size() != get_2(numOfExprs, xmlID)){
			cerr << "Some error occurd, based on missing or to many attributes" << endl;
			cerr << xmlID << " should have " << get_2(numOfExprs, xmlID) << " attributes. In line: " << xmlGetLineNo(node);
			cerr << " are: " << listOfExprs.size() << " expressions." << endl;
			exit(1);
		}
		if(isVertex(xmlID)){//Knoten
			VERTEX v = get_2(vertexes, xmlID);
			v.setIdLabel(getProp(node, "ID"));
			v.setID(getID(v.getIdLabel()));
			v.setAttributes(listOfExprs);
			hg.addVertex(v,v.getID());
		}
	}
	list<IDTYPE> listOfVertices;
	for(xmlNodePtr node = GetChildrenNode(graph, "Node"); node != 0; node = GetNextChild(node, "Node")){
		listOfExprs.clear();
		for(xmlNodePtr attr = GetChildrenNode(node, "Attribute"); attr != 0; attr = GetNextChild(attr, "Attribute")){
			EXPRESSION ex = get_2(exprs, getProp(attr, "type"));
			xmlNodePtr value = GetChildrenNode(GetChildrenNode(GetChildrenNode(attr, "Value"), "java"), "string");
			value = value->xmlChildrenNode;
			string val = (string)(char*)value->content;
			int index = val.find(":");
			string type = substr(val, 0, index-1);
			val = substr(val, index+1, val.size()-1);
			typen = types(type);
			ex.addType(typen);
			ex.addExpr(val);
			listOfExprs.push_back(ex);
		}
		string xmlID = getProp(node, "type");
		if(!isVertex(xmlID)){//Kante
			EDGE edge = get_2(edges, xmlID);
			edge.setIdLabel(getProp(node, "ID"));
			edge.setID(getID(edge.getIdLabel()));
			edge.setAttributes(listOfExprs);
			int i = 0;
			for(xmlNodePtr xmlEdge = GetChildrenNode(graph, "Edge"); xmlEdge != 0; xmlEdge = GetNextChild(xmlEdge, "Edge")){//Relends
				string source = getProp(xmlEdge, "source");
				string target = getProp(xmlEdge, "target");
				string nodeID;
				if(source == edge.getIdLabel() || target == edge.getIdLabel()){
					int startorder = get_2(lines, getProp(xmlEdge, "type"));
					if(i == startorder){
						if(source == edge.getIdLabel())
							nodeID = target;
						else
							nodeID = source;
						listOfVertices.push_back(getID(nodeID));
						i++;
						xmlEdge = GetChildrenNode(graph, "Edge");
					}
				} 
			}
			hg.addEdge(edge,edge.getID(),listOfVertices);
			listOfVertices.clear();
		}
	}
	return hg;
}

void AGGREADER::readRules(xmlNodePtr firstRule, GTS& GraphGrammar){
//	IDTYPE id = 0;
	while(firstRule != 0){
		RULE rule;
		rule.setID(getID(getProp(firstRule, "ID")));
		HYPERGRAPH rhs, lhs;
		IDTYPE image, orig;
		xmlNodePtr graph = GetChildrenNode(firstRule, "Graph");
		lhs = readHypergraph(graph);
		graph = GetNextChild(graph, "Graph");
		rhs = readHypergraph(graph);
	//Create Morphism (Matching)
		MATCH result;
		map<IDTYPE, IDTYPE> match;
		xmlNodePtr morph = GetChildrenNode(firstRule, "Morphism");
		rule.setName(getProp(morph, "name"));
		string strImage, strOrig;
		for(xmlNodePtr mapping = GetChildrenNode(morph, "Mapping");
			mapping != 0; mapping = GetNextChild(mapping, "Mapping")){
			strImage = getProp(mapping, "image");
			strOrig = getProp(mapping, "orig");
			image = getID(strImage);
			orig = getID(strOrig);
			match.insert(make_pair(orig, image));
		}
		result.setMatch(match);
	//Read Attrribut conditions || GUARDS
		xmlNodePtr acon = GetChildrenNode(firstRule, "ApplCondition");
		if(acon != 0){
			xmlNodePtr attrcon = GetChildrenNode(acon, "AttrCondition");
			for(xmlNodePtr con = GetChildrenNode(attrcon, "Condition");
				con != 0; con = GetNextChild(con, "Condition")){
				xmlNodePtr value = GetChildrenNode(con, "Value");
				xmlNodePtr str = GetChildrenNode(GetChildrenNode(value, "java"), "string");
				str = str->xmlChildrenNode;
				string content = (string)(char*)str->content;
				EXPRESSION ex;
				TYPES types;
				ex.addName("guard");
				types.push_back(BOOL);
				ex.addType(types);
				ex.addExpr(content);
				rule.addGuard(ex);
			}
		}
	//Add everything to rule
		rule.setInterface(result);
		rule.setLeft(&lhs);
		rule.setRight(&rhs);
		GraphGrammar.addRule(&rule);
		firstRule = GetNextChild(firstRule,"Rule");
	}
}

void AGGREADER::run(GTS &GraphGrammar){
	xmlNodePtr root = xmlDocGetRootElement(doc);
	xmlNodePtr gts = GetChildrenNode(root,"GraphTransformationSystem");
	readTypes(GetChildrenNode(gts,"Types"));
	initial = readHypergraph(GetChildrenNode(gts, "Graph"));
	GraphGrammar.setStartgraph(&initial);
	readRules(GetChildrenNode(gts, "Rule"), GraphGrammar);

	/*
	ID_CHECKER ic;
	ic.initialize(GraphGrammar);
	ic.run();
	*/
}

void AGGREADER::start(){
	GTS gts;
	initialize("firewall2.ggx");
	run(gts);
	algorithms::writers::GTXLWRITER testWriter;
	testWriter.initialize("GTXLfile.xml");
	testWriter.run(gts);
}
