/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef ALGORITHMSALGORITHM_H
#define ALGORITHMSALGORITHM_H

#include "globals.h"
#include "match.h"
#include "petrigraph.h"
#include "validator.h"
#include "marking.h"
#include "binding.h"
#include "expression.h"
#include "table.h"

namespace algorithm
{
   /*
    * define values used for algorithm IDs
    */
   #define ALGORITHM_NONE 0
   #define ALGORITHM_GENERIC_SCENARIO 2
   #define ALGORITHM_AUNFOLD 3
   #define ALGORITHM_TRIVIAL_MATCHHANDLER 17
   #define ALGORITHM_FIRST_MATCHHANDLER 18
   #define ALGORITHM_SECOND_MATCHHANDLER 19
   #define ALGORITHM_FINDMATCH 20
   #define ALGORITHM_CAUSALITY_RELATION 21
   #define ALGORITHM_FIND_ALL_MATCHES 22
   #define ALGORITHM_MERGE_PETRIGRAPH 23
   #define ALGORITHM_MERGE_PETRIGRAPH_UF 923
   #define ALGORITHM_FOLD 24
   #define ALGORITHM_UNFOLD 25
   #define ALGORITHM_REDUCE_TRANSITIONS 26
   #define ALGORITHM_APPROXIMATED_UNFOLDING 27
   #define ALGORITHM_WRITE_LLNET 28
   #define ALGORITHM_GTXL_READER 29
   #define ALGORITHM_COVERABILITY 30
   #define ALGORITHM_REFINEMENT 31
   #define ALGORITHM_GXLWRITER 32
   #define ALGORITHM_SECOND_MATCHHANDLER_REFINEMENT 33
   #define ALGORITHM_GXLREADER 34
   #define ALGORITHM_COVER 35
   #define ALGORITHM_SPONGE 36
   #define ALGORITHM_REFINEMENT_LOOP 37
   #define ALGORITHM_GTS_EMULATOR 38
   #define ALGORITHM_FIND_ALL_MATCHES_BY_SP 39
   #define ALGORITHM_FINDMATCH_BY_SP 40
   #define ALGORITHM_SPONGE_CALL 41
   #define ALGORITHM_BWRA_SCENARIO 42
   #define ALGORITHM_METRICFF_CALL 43
   #define ALGORITHM_REGEXPR2MARKING 44
   #define ALGORITHM_QUICK_AUNFOLD 45

   class ALGORITHM: public VALIDATOR, public VALIDATED
   {
      public:
         ALGORITHM();
	 virtual ~ALGORITHM();

         list<LABEL> getIdentifiers();
	 string getName();
	 
         virtual void initialize();
         virtual void initialize( PETRIGRAPH* i_petrigraph, IDTYPE i_prule_id,
	                          RULE* i_prule);
         virtual void initialize( PETRIGRAPH* i_petrigraph, IDTYPE i_prule_id,
	                          RULE* i_prule, bool* i_canFold,
				  bool* i_canUnfold);
         virtual void initialize( PETRIGRAPH* i_petrigraph, IDTYPE i_prule_id,
	                          RULE* i_prule, bool* i_canFold,
				  bool* i_canUnfold,
				  MATCH* i_phicpy, MATCH* i_phi2cpy );
	 virtual void initialize( PETRIGRAPH* i_petrigraph, IDTYPE i_prule_id,
	                          RULE* i_prule, MARKING* i_phi_E_L,
                                  MATCH* i_match_first);
 	 virtual void initialize(PETRIGRAPH * i_petrigraph,
				 IDTYPE i_prule_id, RULE * i_prule,
				 bool * i_canFold, bool * i_canUnfold,
				 MATCH* m1, MATCH* m2, MARKING* i_phi_E_L );
         virtual void initialize(HYPERGRAPH* i_t);
         virtual void initialize(HYPERGRAPH* i_t, MATCH* phi, MATCH* phi2, bool* icf, bool* icu, PETRIGRAPH* pg);
         virtual void initialize(PTNET* i_ptnet);
	 virtual void initialize(PETRIGRAPH* pg);
	 virtual void initialize(GTS* i_grammar, ALGORITHM* i_gxlwriter);
	 virtual void initialize( PTNET* i_ptnet,
	                          map<IDTYPE, unsigned>* pp2l, map<unsigned, IDTYPE>* pl2p,
				  map<IDTYPE, unsigned>* tp2l, map<unsigned, IDTYPE>* tl2p  );
	 virtual void initialize(char *filename);
	 virtual bool initialize(list<string> files);
	 virtual void initialize(BINDING pb);
	 
	 virtual void run();
         virtual bool run(MATCH* match);
         virtual bool run(MATCH* phi_in, MATCH* phi2_out);
	 virtual MATCH* run(HYPERGRAPH* hgraph, bool* found, IDTYPE ruleid, MARKING* phi_E_L);
	 virtual MATCH* run(HYPERGRAPH* hgraph, bool* found, IDTYPE ruleid);
	 virtual MATCH* run(HYPERGRAPH* hgraph, IDTYPE ruleid);
	 virtual list<MATCH> run( HYPERGRAPH* hgraph, ALGORITHM* i_matchhandler,
	                          bool surjective );
         virtual bool run(IDTYPE elem1, IDTYPE elem2); 
	 virtual void run( set< pair<IDTYPE, IDTYPE> > vunion,
	                   set< pair<IDTYPE, IDTYPE> > eunion,
			   set< pair<IDTYPE, IDTYPE> > tunion );
	 virtual void run( map<IDTYPE, IDTYPE> vunion,
	                   map<IDTYPE, IDTYPE> eunion, map<IDTYPE, IDTYPE> tunion );
	 virtual void run(RULE* r, MATCH phi2, MATCH phi);
	 virtual void run(RULE* r, MATCH phi);
	 virtual void run(PETRIGRAPH* pg);
	 virtual void run(PETRIGRAPH &PtGraph, char const* id_name);
	 virtual void run(char* filename);
	 virtual void run(GTS& GraphGrammar);
	 virtual bool run(MARKING m, long * psize);
	 virtual bool run(MATCH m, long * psize);
	 virtual void run(PETRIGRAPH &PtGraph, GTS &GraphGrammar);
	 virtual list<MARKING> run(int type);
	 virtual EXPRESSION run(EXPRESSION & e);
	 virtual string run(EXPRESSION & e, PRINT_TYPE type);
	 virtual bool run(list<EXPRESSION> guards);
	 virtual bool run(list<MARKING> m, char * tracefile);
	 virtual bool run(IDTYPE trans_id, TABLE * * table);

	 virtual bool refine();
	 virtual void evaluate(ExpressionElement & e);

	 friend ostream& operator<<(ostream& ost, ALGORITHM& th);
	 
	 void setReusable(bool val);
	 bool getReusable();
	 
	 virtual void setAlgorithm(ALGORITHM * p_algorithm);

	 virtual BINDING abstractBinding(BINDING b);

	 // <= : Compare EVALUATED expressions
	 virtual bool smallereq(list<EXPRESSION> l1, list<EXPRESSION> l2);

	 virtual bool smallereq(EXPRESSION e1, EXPRESSION e2);

	 virtual bool smallereq(ExpressionElement ee1, ExpressionElement ee2);

      protected:
         PATH_DB path;
         IDTYPE ID;
         string name;
         list<LABEL> lablist;
         /*
	  * The history stores successive calls. For example,
	  * suppose A calls B at position a, and later B calls C at position b.
	  * Then, C should have the history (A.history, (A, a), (B, b)).
	  * By evaluating their own history, algorithms can adapt to their context.
	  * Here, algorithms are identified by their names.
	  */
         list< pair<string, string> > history;
	 
	 void setPath(PATH_DB &p);
	 ALGORITHM* load(LABEL lb);
	 
	 bool reusable;
   };
   
   ostream& operator<<(ostream& ost, PATH_DB& th);
};
using namespace algorithm;

#endif
