/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "algorithm_storage.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "aunfold.h"
#include "causality_relation.h"
#include "findmatch.h"
#include "fold.h"
#include "matchhandler.h"
#include "merge_petrigraph.h"
#include "merge_petrigraph_uf.h"
#include "reduce_transitions.h"
#include "unfold.h"
#include "aggreader.h"
#include "coverability.h"
#include "abstraction_refinement.h"
#include "second_matchhandler_refinement.h"
#include "GTXLReader.h"
#include "gtxl_nac_reader.h"
#include "gxlreader.h"
#include "gxlwriter.h"
#include "persistable.h"
#include "refinement.h"
#include "spl_scenario.h"
#include "rules_converter.h"
#include "hg_converter.h"
#include "pn_converter.h"
#include "cover_scenario.h"
#include "sponge_call.h"
#include "refinement_loop.h"
#include "testolga.h"
#include "testscenario.h"
#include "gts_emulator.h"
#include "expression_engine.h"
#include "expression_engine_mp.h"
#include "expression_engine_mod.h"
#include "logic_converter.h"
#include "reachability_analyse.h"
#include "bwra.h"
#include "findmatch_by_sp.h"
#include "expression_engine_pn.h"
#include "expression_engine_mn.h"
#include "expression_engine_mp0.h"
#include "find_all_matches.h"
#include "find_all_matches_by_sp.h"
#include "newGTXLReader.h"
#include "pointer_lang_reader.h"
#include "sponge_scenario.h"
#include "metrikff_call_scenario.h"
#include "bwra_scenario.h"
#include "regexpr2marking.h"
#include "quickunfoldscenario.h"

#define GETALG(p) \
	if(reusable[name])\
		{	if(reusableStorage.find(name) == reusableStorage.end())\
			{	p *g = new p();\
				g->setReusable(true);\
				reusableStorage[name] = (ALGORITHM*)g;\
			}\
			return reusableStorage[name];\
		}\
		else\
		{\
			p *g = new p();\
			return (ALGORITHM*)g;\
		}
	
using namespace algorithms;
using namespace readers;
using namespace special;
using namespace special::param;

using namespace testspace;

namespace algorithm {

ALGORITHM* ALGORITHM_STORAGE::getAlgorithm(PATH_DB path)
{
	string name = getFunction(path);
	
	// scenarios
	if(name == "aunfold") GETALG(AUNFOLD);
	if(name == "refinement") GETALG(REFINEMENT);
	if(name == "coverability_scenario") GETALG(COVER_SCENARIO);
	if(name == "sponge_call") GETALG(SPONGE_CALL_SCENARIO);
	if(name == "sponge") GETALG(SPONGE_SCENARIO);
	if(name == "refinement_loop") GETALG(REFINEMENT_LOOP);
	if(name == "gts_emulator") GETALG(GTS_EMULATOR);
	if(name == "metricff_scenario") GETALG(METRIKFF_CALL_SCENARIO);
	if(name == "bwra_scenario") GETALG(BWRA_SCENARIO);
	if(name == "reg2marking") GETALG(REGEXPR_2_MARKING);
	if(name == "qunfold") GETALG(QUICK_UNFOLD_SCENARIO);

	// convertors
	if (name == "spl_converter") GETALG(SPL_CONVERTER);
	if (name == "rules_converter") GETALG(RULES_CONVERTER);
	if (name == "hg_converter") GETALG(HG_CONVERTER);
	if (name == "pn_converter") GETALG(PN_CONVERTER);
	if(name == "logic_converter") GETALG(LOGIC_CONVERTER);
	
	// algorithms
	if(name == "causality_relation") GETALG(CAUSALITY_RELATION);
	if(name == "findmatch") GETALG(FINDMATCH);
	if(name == "findmatch_by_sp") GETALG(FINDMATCH_BY_SP);
	if(name == "first_matchhandler") GETALG(FIRST_MATCHHANDLER);
	if(name == "find_all_matches") GETALG(FIND_ALL_MATCHES);
	if(name == "find_all_matches_by_sp") GETALG(FIND_ALL_MATCHES_BY_SP);
	if(name == "fold") GETALG(FOLD);
	if(name == "merge_petrigraph") GETALG(MERGE_PETRIGRAPH);
	if(name == "merge_petrigraph_uf") GETALG(MERGE_PETRIGRAPH_UF);
	if(name == "reduce_transitions") GETALG(REDUCE_TRANSITIONS);
	if(name == "second_matchhandler") GETALG(SECOND_MATCHHANDLER);
	if(name == "unfold") GETALG(UNFOLD);
	if(name == "agg_reader") GETALG(AGGREADER);
	if(name == "coverability") GETALG(COVERABILITY);
	if(name == "abstraction_refinement") GETALG(ABSTRACTION_REFINEMENT);
	if(name == "second_matchhandler_refinement") GETALG(SECOND_MATCHHANDLER_REFINEMENT);
	if(name == "gtxl_reader") GETALG(GTXLREADER);
	if(name == "gtxl_nac_reader") GETALG(GTXL_NAC_READER);
	if(name == "new_gtxl_reader") GETALG(newGTXLREADER);
	if(name == "spl_reader") GETALG(POINTER_LANG_READER);
	if(name == "gxl_reader") GETALG(GXLREADER);
	if(name == "gxl_writer") GETALG(GXLWRITER);
	if(name == "approximated_unfolding") GETALG(APPROXIMATED_UNFOLDING);
	if(name == "coverability_test") GETALG(COVERABILITY);
	if(name == "expression_engine") GETALG(EXPRESSION_ENGINE);
	if(name == "expression_engine_mp") GETALG(EXPRESSION_ENGINE_MP);
	if(name == "expression_engine_mod") GETALG(EXPRESSION_ENGINE_MOD);
	if(name == "expression_engine_mp0") GETALG(EXPRESSION_ENGINE_MP0);
	if(name == "expression_engine_pn") GETALG(EXPRESSION_ENGINE_PN);
	if(name == "expression_engine_mn") GETALG(EXPRESSION_ENGINE_MN);
	if(name == "bwra") GETALG(BWRA);
	if(name == "reachability_analyse") GETALG(REACHABILITY_ANALYSE);

	// test scenarios
	if(name == "testolga") GETALG(TestOlga);
	if(name == "test_scenario") GETALG(TestScenario);	
	
	cerr << "\nerror: algorithm \"" << name << "\" not found in database\n";
	exit(1);
}

ALGORITHM_STORAGE::ALGORITHM_STORAGE()
{
	name = "storage";
	filename = NULL;
}


ALGORITHM_STORAGE::~ALGORITHM_STORAGE()
{
}

bool ALGORITHM_STORAGE::alike(pair<string, LABEL> p1, pair<string, LABEL> p2)
{
	if (((p1.first == p2.first)&&(p1.second == p2.second))
		||((p1.first == p2.first)&&(p2.second == "#"))
		||((p2.first == "#")&&(p1.second == p2.second))
		||((p2.first == "#")&&(p2.second == "#")))
			return true;
	else
			return false;
}

bool ALGORITHM_STORAGE::comparePATH_DB2(PATH_DB p, PATH_DB pDB)
{
	PATH_DB::iterator i = p.begin();
	PATH_DB::iterator j = pDB.begin();
	
	while((i != p.end())&&(j != pDB.end()))
	{
		if(alike((*i), (*j))) 
		{
		 i++; 
		 j++;
		}
		else
		  i++;
	}
	
	if(j == pDB.end())
	 return true;
	else
	 return false; 
}

bool ALGORITHM_STORAGE::comparePATH_DB(PATH_DB p, PATH_DB pDB, bool count)
{
	bool result = false;
	bool f = true;

	bool countj = count;
		
	PATH_DB pDB_alternative = pDB;
	PATH_DB p_alternative = p;
	bool alternative_full = false;

	PATH_DB::iterator i = p.begin();
	PATH_DB::iterator j = pDB.begin();
	
	while((i != p.end())&&(j != pDB.end())&&(f))
	{
		f = false;
		if((j->first == "+")&&(j->second == "+"))
		{	if(countj)
			{
				j++;
				if(j != pDB.end())
				{
					if(alike((*i),(*j)))
					{	
						if(!alternative_full)
							alternative_full = true;
						p_alternative.pop_front();
						j++;
						countj = false;
					}
					else	j--;
					if(!alternative_full)
						p_alternative.pop_front();
					
				}
				else j--;
			}
			else	countj = true;
			
			f = true;
			i++;
			
		}
				
		else if((j->first == "*")&&(j->second == "*"))
		{
			j++;
			if(j != (pDB.end()))
			{
				if(alike((*i),(*j)))
				{
					if(!alternative_full)
						alternative_full = true;
					
					j++;
					countj = false;
				}
				else	j--;
				
				if(!alternative_full)
					p_alternative.pop_front();
			}
			else j--;	
			i++;
			f = true;
		}	
				
		else if(alike((*i), (*j))) 
		{
			i++; 
			j++;
			if(!alternative_full)
			{
				p_alternative.pop_front();
				pDB_alternative.pop_front();
			}
			countj = false;
			f = true;
		}

	}
	
	if((i == p.end())&&(j == pDB.end()))
		result = true;

	PATH_DB::iterator last = pDB.end();
	last--;
	if((j == last)&&((((*j).first == "*")&&((*j).second == "*"))||(((*j).first == "+")&&((*j).second == "+")&&(countj))))
		result = true;
	if((!result)&&(alternative_full)&&(p_alternative.size() != 0))
		result = comparePATH_DB(p_alternative, pDB_alternative, true);
		
	return result;
		

}

string ALGORITHM_STORAGE::getFunction(PATH_DB path)
{
try{	
	pair<string, LABEL> func_lab(path.back());
	CALL call;
	map<pair<string, LABEL>, CALL>::iterator i = datenBank.begin();
	
	bool f = false;
	while(!f)
	{
		if(i == datenBank.end())
		{
		  cout << "\nalgorithm <" << path.back().first << ", " << path.back().second;
		  cout << "> not found! severe database error! aborting ...\n";
		  exit(1);
		}
		  
		if(alike(i->first, func_lab))
		{
			call = (i->second);
			f = true;
		}
		
		i++;
	}

	if(call.status)
	{
		path.pop_back();
		list<pair<PATH_DB, string> >::iterator cF = call.callingFunctions.begin();
	
		while(cF != call.callingFunctions.end())
		{	
			// bool comp = comparePATH_DB(path, (*cF).first, false);
			bool comp = comparePATH_DB2(path, (*cF).first);
			if(comp)
				return (*cF).second;
			cF++;
		}
	}
	
	return call.defaultFunction;
}
catch(exception e)
{
 cout << "error in ALGORITHM_STORAGE::getFunction()\n";
 exit(1);
} 
	
}

xmlNodePtr ALGORITHM_STORAGE:: GetChildrenNode(xmlNodePtr parent, char* name_child)
{
	xmlNodePtr child = parent->xmlChildrenNode;
	while((child != NULL) && (xmlStrcasecmp(child->name, (const xmlChar*)name_child)))
	{
		child = child->next;
	}
	
	return child;
	
		
}
 
xmlNodePtr ALGORITHM_STORAGE::GetNextChild(xmlNodePtr child1, char* name_child)
{
	
	xmlNodePtr child2 = child1->next;
	while((child2 != NULL) && (xmlStrcasecmp(child2->name, (const xmlChar*)name_child)))
	{
		child2 = child2->next;
	}
	
	return child2;
	
	
}

bool ALGORITHM_STORAGE::setParam(xmlNodePtr cur, char* name)
{
		
	xmlNodePtr par = GetChildrenNode(cur, name);
	string str_par = (string)(char*)xmlGetProp(par, (const xmlChar*)"val");
	if(str_par == "false")
		return false;
	else
		return true;
}

unsigned long ALGORITHM_STORAGE::setLongParam(xmlNodePtr cur, char* name)
{
	char* pEnd;
	xmlNodePtr par = GetChildrenNode(cur, name);
	string str_par = (string)(char*)xmlGetProp(par, (const xmlChar*)"val");
//	unsigned long l = atol(str_par.c_str());
	unsigned long l = strtoul(str_par.c_str(), &pEnd, 0);
	return l;
}

void ALGORITHM_STORAGE::OutParam(xmlNodePtr cur, char* name, bool val)
{
	xmlNodePtr par = xmlNewChild(cur, NULL, (const xmlChar*)name, NULL);
    	if(val)
    		xmlSetProp(par, (const xmlChar*)"val", (const xmlChar*)"true");
    	else
    		xmlSetProp(par, (const xmlChar*)"val", (const xmlChar*)"false");	
}

void ALGORITHM_STORAGE::OutParam(xmlNodePtr cur, char* name, unsigned long val)
{
	xmlNodePtr par = xmlNewChild(cur, NULL, (const xmlChar*)name, NULL);
	char* ch;
	sprintf(ch, "%d", val);
    	xmlSetProp(par, (const xmlChar*)"val", (const xmlChar*)ch);
    	
}

void ALGORITHM_STORAGE::initialize (const char* filename)
{
    doc = xmlParseFile(filename);

    this->filename = new char[strlen(filename)+1];
    strcpy(this->filename, filename);
	
    if (doc == NULL ) 
    {
	fprintf(stderr,"Document not parsed successfully. \n");
	xmlFreeDoc(doc);
	exit(1);
    }
	
    if (xmlStrcmp(xmlDocGetRootElement(doc)->name, (const xmlChar *) "Datenbank")) 
      {
	fprintf(stderr,"document of the wrong type, root node != Datenbank");
	xmlFreeDoc(doc);
	exit(1);
      }
     
     xmlNodePtr root = xmlDocGetRootElement(doc); 
      
     // read the global parameters      
      
	xmlNodePtr globals = GetChildrenNode(root, "Globals");
	
	debug = setParam(globals, "debug");
	quiet = setParam(globals, "quiet");
	ncov = setParam(globals, "ncov");
	coloured = setParam(globals, "coloured");
	nfold = setParam(globals, "nfold");
	red = setParam(globals, "red");
	mapping = setParam(globals, "mapping");
	trace = setParam(globals, "trace");
	depth = setLongParam(globals, "depth");
	unfdepth = setLongParam(globals, "unfdepth");
	maxstep = setLongParam(globals, "maxstep");
	coloured = setParam(globals, "coloured");
	full_coverability_graph=setParam(globals, "full_coverability_graph");
	clear_cov_count=setLongParam(globals, "clear_cov_count");
	mod_base=setLongParam(globals, "mod_base");
	interval_from=setLongParam(globals, "interval_from");
	interval_to=setLongParam(globals, "interval_to");
	ref_loop_limit=setLongParam(globals, "ref_loop_limit");
	attr_ref_count=setLongParam(globals, "attr_ref_count");

	// read DB
	
	xmlNodePtr algorithms = GetChildrenNode(root, "Algorithms_DB");
	xmlNodePtr func = GetChildrenNode(algorithms, "algorithm");
	while(func != NULL)
	{
		string name = (string)(char*)xmlGetProp(func, (const xmlChar*)"name");
		string s_reuse = (string)(char*)xmlGetProp(func, (const xmlChar*)"reusable");
		bool reuse = false;
		if(!strcmp(s_reuse.c_str(), "true"))
			reuse = true;
		this->reusable[name] = reuse;
			
		xmlNodePtr label = GetChildrenNode(func, "label");
		set<LABEL> labels;
		while(label != NULL)
		{
			string label_name = (string)(char*)xmlGetProp(label, (const xmlChar*)"name");
			pair<string, LABEL> func_lab(name, label_name);
			labels.insert(label_name);
			
			CALL call;
			xmlNodePtr defFunc = GetChildrenNode(label, "default");
			if(defFunc != NULL)
			{
				string f_name = (string)(char*)xmlGetProp(defFunc, (const xmlChar*)"algorithm");
				call.defaultFunction = f_name;
			}
			
			xmlNodePtr info = GetChildrenNode(label, "info");
			if((info != NULL)&&(info->xmlChildrenNode != NULL))
			{
				string info_str = (string)(char*)(info->xmlChildrenNode->content);
				call.info = info_str;
			}
			
			xmlNodePtr expert = GetChildrenNode(label, "expert");
			if(expert != NULL)
			{
				string status = (string)(char*)xmlGetProp(expert, (const xmlChar*)"status");
				if(!strcmp(status.c_str(), "true"))
					call.status = true;
				else
					call.status = false;
					
				if(call.status)
				{
					xmlNodePtr history = GetChildrenNode(expert, "history");
					
					while(history != NULL)
					{
						xmlNodePtr happen = GetChildrenNode(history, "happen");
						PATH_DB path;
						typedef pair<string, LABEL> path_type;
						while(happen != NULL)
						{
							string alg_name = (string)(char*)xmlGetProp(happen, (const xmlChar*)"algorithm");
							string label_name = (string)(char*)xmlGetProp(happen, (const xmlChar*)"label");
							path.push_back(path_type(alg_name, label_name));
							happen = GetNextChild(happen, "happen");
						}
						
						xmlNodePtr calling = GetChildrenNode(history, "call");
						if(calling != NULL)
						{
							string call_name = (string)(char*)xmlGetProp(calling, (const xmlChar*)"algorithm");
							typedef pair<PATH_DB, string> call_type;
							call.callingFunctions.push_back(call_type(path, call_name));
						}
						history = GetNextChild(history, "history");
					
					}
				}
			}
			
			this->datenBank[func_lab] = call;
			label = GetNextChild(label, "label");
		}
		this->functions[name] = labels;
		func = GetNextChild(func, "algorithm");
		
	}
	
	// special case coverability
	// reusable must be set only in the case of !ncov
	// and if the working algorithm is coverability
	if(reusable.find(string("coverability")) != reusable.end())
	{
		if(reusable[string("coverability")] && ncov)
		{
			cout << "Warning: no coverability check - reusable is set to false \n";
			reusable[string("coverability")] = false;
		}

		pair<string, LABEL> p(string("first_matchhandler"), string("call_coverability"));

		if(datenBank.find(p) != datenBank.end())
		{
			if(reusable[string("coverability")] && datenBank[p].defaultFunction != string("coverability"))
			{
				cout << "Warning: algorithm coverability is not used - reusable is set to false \n";
				reusable[string("coverability")] = false;
			}
		}
	}
	
}

void ALGORITHM_STORAGE::writeDB()
{
	if(filename != NULL)
	 writeDB(filename);
}

void ALGORITHM_STORAGE::writeDB(const char* filename)
{
    doc = xmlNewDoc((const xmlChar *)"1.0");
    xmlNodePtr root = xmlNewNode(NULL,(const xmlChar *)"Datenbank");
    xmlDocSetRootElement(doc,root);
    
    xmlNodePtr globals = xmlNewChild(root, NULL, (const xmlChar*)"Globals", NULL);
  
	OutParam(globals, "debug", debug);
	OutParam(globals, "quiet",quiet);
	OutParam(globals, "ncov",ncov);
	OutParam(globals, "coloured",coloured);
	OutParam(globals, "nfold",nfold);
	OutParam(globals, "red",red);
	OutParam(globals, "mapping",mapping);
	OutParam(globals, "trace",trace);
	OutParam(globals, "depth",depth);
	OutParam(globals, "unfdepth",unfdepth);
	OutParam(globals, "maxstep",maxstep);
	OutParam(globals, "full_coverability_graph",full_coverability_graph);
	OutParam(globals, "clear_cov_count",clear_cov_count);
	OutParam(globals, "mod_base",mod_base);
	OutParam(globals, "interval_from",interval_from);
	OutParam(globals, "interval_to",interval_to);
	OutParam(globals, "ref_loop_limit",ref_loop_limit);
	OutParam(globals, "attr_ref_count",attr_ref_count);

    xmlNodePtr algorithm = xmlNewChild(root, NULL, (const xmlChar*)"Algorithms_DB", NULL);
    for(map<string, set<LABEL> >::iterator it = this->functions.begin(); it != this->functions.end(); it++)
    	{
    		xmlNodePtr func = xmlNewChild(algorithm, NULL, (const xmlChar*)"algorithm", NULL);
		xmlSetProp(func, (const xmlChar*)"name", (const xmlChar*)(it->first).c_str());
		if(this->reusable[it->first])
			xmlSetProp(func, (const xmlChar*)"reusable", (const xmlChar*)"true");
		else
			xmlSetProp(func, (const xmlChar*)"reusable", (const xmlChar*)"false");
			
		for(set<LABEL>::iterator lb = it->second.begin(); lb != it->second.end(); lb++)
		{
			xmlNodePtr label = xmlNewChild(func, NULL, (const xmlChar*)"label", NULL);
			xmlSetProp(label, (const xmlChar*)"name", (const xmlChar*)(*lb).c_str());
			
			pair<string, LABEL> func_lab(it->first, *lb);
			CALL call = this->datenBank[func_lab];
			xmlNodePtr def = xmlNewChild(label, NULL, (const xmlChar*)"default", NULL);
			xmlSetProp(def, (const xmlChar*)"algorithm", (const xmlChar*)call.defaultFunction.c_str());
			
			xmlNewChild(label, NULL, (const xmlChar*)"info", (const xmlChar*)call.info.c_str());
			
			xmlNodePtr expert = xmlNewChild(label, NULL, (const xmlChar*)"expert", NULL);
			if(call.status)
			{
				xmlSetProp(expert, (const xmlChar*)"status", (const xmlChar*)"true");
				for(list<pair<PATH_DB, string> >::iterator pth = call.callingFunctions.begin();
					pth != call.callingFunctions.end(); pth++)
				{
					xmlNodePtr history = xmlNewChild(expert, NULL, (const xmlChar*)"history", NULL);
					for(PATH_DB::iterator p = (*pth).first.begin(); p != (*pth).first.end(); p++)
					{
						xmlNodePtr happen = xmlNewChild(history, NULL, (const xmlChar*)"happen", NULL);
						xmlSetProp(happen, (const xmlChar*)"algorithm", (const xmlChar*)((*p).first).c_str());
						xmlSetProp(happen, (const xmlChar*)"label", (const xmlChar*)((*p).second).c_str());
						
					}
					xmlNodePtr calling = xmlNewChild(history, NULL, (const xmlChar*)"call", NULL);
					xmlSetProp(calling, (const xmlChar*)"algorithm", (const xmlChar*)((*pth).second).c_str());				
				}
			}
			else
				xmlSetProp(expert, (const xmlChar*)"status", (const xmlChar*)"false");
		}
		
	}
        
    if(-1 == xmlSaveFormatFile(filename,doc,true))
    {
      cout << "Error writing XML file (Datenbank)\n";
      exit(1);
    }
}

void ALGORITHM_STORAGE::setData(string to_alg, DATA* d)
{
	if(d==NULL)
	 exchanged_data.erase(to_alg);
	else 
	 exchanged_data[to_alg]=d;
}

DATA * ALGORITHM_STORAGE::getData(string alg)
{
	map<string, DATA*>::iterator it = exchanged_data.find(alg);
	if(it!=exchanged_data.end())
	 return it->second;
	else
	 return NULL; 
}

ALGORITHM * ALGORITHM_STORAGE::getStdEngine()
{
	return this->load("attribute_engine_std");
}

ALGORITHM * ALGORITHM_STORAGE::getEngine()
{
	return this->load("attribute_engine");
}

};

namespace special
{
   ALGORITHM_STORAGE storage;
};
using namespace special;
