/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef ALGORITHMALGORITHM_STORAGE_H
#define ALGORITHMALGORITHM_STORAGE_H

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <stdexcept>

#include "globals.h"
#include "algorithm.h"
#include "GTXLReader.h"
#include "gxlreader.h"
#include "persistable.h"

#include "testolga.h"


using namespace std;

namespace algorithm {


struct CALL
{
	string defaultFunction;
	string info;
	bool status;					/* status defines having Expert-Table (callingFunctions)*/
	list<pair<PATH_DB, string> > callingFunctions;	/* corresponding function to call, 
								depends on PATH_DB */
	
};

class ALGORITHM_STORAGE: public ALGORITHM
{
public:
    ALGORITHM_STORAGE();

    ~ALGORITHM_STORAGE();
    
    void initialize (const char* filename);
    
    void writeDB (const char* filename);
    void writeDB ();
    
    ALGORITHM* getAlgorithm(PATH_DB path);
    
    // methods for changing the data between two algorithms
    // which can not communicate directly 
    void setData(string to_alg, DATA * d);
    DATA * getData(string alg);

    ALGORITHM * getStdEngine();
    ALGORITHM * getEngine();
    
private:
	xmlNodePtr GetChildrenNode(xmlNodePtr parent, char* child_name);
        xmlNodePtr GetNextChild(xmlNodePtr child1, char* child_name);
	string getFunction(PATH_DB path);    	
	
	bool comparePATH_DB(PATH_DB p, PATH_DB pDB, bool count);
	bool comparePATH_DB2(PATH_DB p, PATH_DB pDB);
	
	bool alike(pair<string, LABEL> p1, pair<string, LABEL> p2);
	
	bool setParam(xmlNodePtr cur, char* name);
	unsigned long setLongParam(xmlNodePtr cur, char* name);
	void OutParam(xmlNodePtr cur, char* name, bool val);
	void OutParam(xmlNodePtr cur, char* name, unsigned long val);
	
	map<pair<string, LABEL>, CALL> datenBank;	/* function with the name string calls CALL if 
								it get corresponding LABEL*/
	
	map<string, set<LABEL> > functions;		/* all Labels a corresponding function has*/
	
	map<string, bool> reusable;
	
	map<string, ALGORITHM*> reusableStorage;
	
	list<ALGORITHM> algos;
	
	map<string, DATA*> exchanged_data;
	
	//set<string> functions;
	
	xmlDocPtr doc;

	// filename for database
	char * filename;
};

};

namespace special
{
   extern ALGORITHM_STORAGE storage;
}
using namespace special;

#endif
