/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "apply_rule.h"

namespace algorithms
{

namespace hypergraph
{

APPLY_RULE::APPLY_RULE()
{
 p_engine=NULL;
}


APPLY_RULE::~APPLY_RULE()
{
}
 
void APPLY_RULE::initialize(HYPERGRAPH * hg)
{
 this->hg=hg;
}

void APPLY_RULE::initialize(HYPERGRAPH * hg, ALGORITHM * p_engine)
{
 this->hg = hg;
 this->p_engine = p_engine;	
}

HYPERGRAPH APPLY_RULE::run(RULE * r, MATCH & m)
{
 HYPERGRAPH result = *hg;
   
 map<IDTYPE, VERTEX> vertices_right = r->getRight()->getVertices();

 // map to the inserted nodes
 map<IDTYPE, IDTYPE> inserted;
 
 // map to remained nodes
 map<IDTYPE, IDTYPE> remained;
 
 // interface
 map<IDTYPE, IDTYPE> interface = r->getInterface().getMatch();
 
 // iota match
 map<IDTYPE, IDTYPE> iota_match = m.getMatch();
 
 // insert nodes
 for(map<IDTYPE, VERTEX>::iterator v = vertices_right.begin(); v != vertices_right.end(); v++)
 {
 	bool found = false;
	for(map<IDTYPE, IDTYPE>::iterator it = interface.begin(); it != interface.end(); it++)
	{
		if(it->second == v->first)
		{
			found=true;
			remained.insert(pair<IDTYPE, IDTYPE>(v->first, iota_match[it->first]));
			break;
		}
	}
	
	if(!found)
	{
		VERTEX vertex = v->second;
		IDTYPE id = result.getLastUsedID()+1;
		result.addVertex(vertex, id);
		inserted.insert(pair<IDTYPE, IDTYPE>(v->first, id));
	}
 }

 //edges
 map<IDTYPE, EDGE> edges_left = r->getLeft()->getEdges();
 map<IDTYPE, EDGE> edges_right = r->getRight()->getEdges();

 // delete L edges
 for(map<IDTYPE, EDGE>::iterator e=edges_left.begin(); e!=edges_left.end(); e++)
	result.deleteEdge(m.getMatch(e->first));

 // insert R edges
 for(map<IDTYPE, EDGE>::iterator ri=edges_right.begin(); ri!=edges_right.end(); ri++)
 {
	list<IDTYPE> cV_ = r->getRight()->getConnectedVertices(ri->first);
	list<IDTYPE> cV;
	for(list<IDTYPE>::iterator icV = cV_.begin(); icV != cV_.end(); icV++)
	{
	 if(inserted.find(*icV)!=inserted.end())
	  cV.push_back(inserted[*icV]);
	 else
	  cV.push_back(remained[*icV]); 
	} 
	EDGE edge = ri->second;
	result.addEdge(edge, result.getLastUsedID()+1, cV);
 }
 
 return result;

}

HYPERGRAPH APPLY_RULE::run(RULE * r, MATCH & m, BINDING & b)
{
 if(p_engine==NULL)
 {
	cout << "\nerror: APPLY_RULE - no engine\n";
	exit(1);
 }

 // set binding
 p_engine->initialize(b);

 HYPERGRAPH result = *hg;
   
 map<IDTYPE, VERTEX> vertices_right = r->getRight()->getVertices();

 // map to the inserted nodes
 map<IDTYPE, IDTYPE> inserted;
 
 // map to remained nodes
 map<IDTYPE, IDTYPE> remained;
 
 // interface
 map<IDTYPE, IDTYPE> interface = r->getInterface().getMatch();
 
 // iota match
 map<IDTYPE, IDTYPE> iota_match = m.getMatch();
 
 // insert nodes
 for(map<IDTYPE, VERTEX>::iterator v = vertices_right.begin(); v != vertices_right.end(); v++)
 {
 	bool found = false;
	for(map<IDTYPE, IDTYPE>::iterator it = interface.begin(); it != interface.end(); it++)
	{
		if(it->second == v->first)
		{
			found=true;
			remained.insert(pair<IDTYPE, IDTYPE>(v->first, iota_match[it->first]));
			break;
		}
	}
	
	if(!found)
	{
		VERTEX vertex = v->second;
		IDTYPE id = result.getLastUsedID()+1;
		result.addVertex(vertex, id);
		inserted.insert(pair<IDTYPE, IDTYPE>(v->first, id));
	}
 }

 //edges
 map<IDTYPE, EDGE> edges_left = r->getLeft()->getEdges();
 map<IDTYPE, EDGE> edges_right = r->getRight()->getEdges();

 // delete L edges
 for(map<IDTYPE, EDGE>::iterator e=edges_left.begin(); e!=edges_left.end(); e++)
	result.deleteEdge(m.getMatch(e->first));

 // insert R edges
 for(map<IDTYPE, EDGE>::iterator ri=edges_right.begin(); ri!=edges_right.end(); ri++)
 {
	list<IDTYPE> cV_ = r->getRight()->getConnectedVertices(ri->first);
	list<IDTYPE> cV;
	for(list<IDTYPE>::iterator icV = cV_.begin(); icV != cV_.end(); icV++)
	{
	 if(inserted.find(*icV)!=inserted.end())
	  cV.push_back(inserted[*icV]);
	 else
	  cV.push_back(remained[*icV]); 
	} 
	
	EDGE edge = ri->second;

	// attributes
	list<EXPRESSION> attributes = edge.getAttributes();

	list<EXPRESSION> le;
	for(list<EXPRESSION>::iterator ita = attributes.begin(); ita != attributes.end(); ita++)
	{		
		// evaluate
		EXPRESSION e = p_engine->run(*ita);

		le.push_back(e);
	}

	edge.setAttributes(le);

	result.addEdge(edge, result.getLastUsedID()+1, cV);
 }
 
 return result;

}

bool APPLY_RULE::IsAppliable(RULE * r,  BINDING * b)
{
	p_engine->initialize(*b);
	return p_engine->run(r->getGuard());
}

///////////////// namespaces ////////////////////

}

}


