/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#ifndef ALGORITHM_APPROXIMATED_UNFOLDING_CPP
#define ALGORITHM_APPROXIMATED_UNFOLDING_CPP

#include "algorithm_storage.h"
#include "approximated_unfolding.h"
#include "fold.h"
#include "matchhandler.h"
#include "reduce_transitions.h"
#include "find_all_matches.h"
#include "unfold.h"
#include "globals.h"

// for tests
#include "coverability.h"

namespace algorithms
{
   namespace petrigraph
   {
      APPROXIMATED_UNFOLDING::APPROXIMATED_UNFOLDING()
      {
         this->ID = ALGORITHM_APPROXIMATED_UNFOLDING;
         this->name = "approximated_unfolding";
	 this->grammar = NULL;
	 this->gxlwriter = NULL;
      }   
      APPROXIMATED_UNFOLDING::~APPROXIMATED_UNFOLDING() {}
      
      void APPROXIMATED_UNFOLDING::initialize(GTS* i_grammar, ALGORITHM* i_gxlwriter)
      {
         this->grammar = i_grammar;
	 this->gxlwriter = i_gxlwriter;
      }
      
      /*
       * parameter pg contains the return value ...
       */
      void APPROXIMATED_UNFOLDING::run(PETRIGRAPH* pg)
      {
         ALGORITHM* rt = this->load("call_reduce_transitions");
         rt->initialize(pg);

	 /*
	  * step 0
	  *
	  * construct a petri graph from the start graph
	  * without any transitions and with the identity as its iota mapping
	  * set all depth values to 0
	  */
	 *pg = PETRIGRAPH();
	 IDTYPE pgid = pg->getLastUsedID() + 1;
	 pg->setID(pgid);
	 pg->setLastUsedID(debug? this->grammar->getStartgraph()->getLastUsedID(): pgid);
	 HYPERGRAPH* startgraph = this->grammar->getStartgraph();
	 map<IDTYPE, VERTEX> V = startgraph->getVertices();
	 map<IDTYPE, VERTEX>::iterator vit;
	 map<IDTYPE, EDGE> E = startgraph->getEdges();
	 map<IDTYPE, EDGE>::iterator eit;
	 MARKING initial;
	 MATCH iota;
	 for(vit = V.begin(); vit != V.end(); vit++)
	 {
	    pg->addPlace(vit->first, vit->second);
	    pg->setDepth(vit->first, 0);
	    
	    if(nodes_as_places)
	    {
		TOKEN t;
		t.setCount(1);

	        if(coloured)
		{
		 ALGORITHM * engine = storage.getEngine();
		 list<EXPRESSION> le = vit->second.getAttributes();
		 for(list<EXPRESSION>::iterator ita = le.begin(); ita != le.end(); ita++)
			t.addAttributes(engine->run(*ita));
		}

		initial.putToken(vit->first, t);
	    }
	
	    iota.setMatch(vit->first, vit->first);
	 }
	 for(eit = E.begin(); eit != E.end(); eit++)
	 {
	    pg->addPlace(eit->first, eit->second, startgraph->getConnectedVertices(eit->first));
	    pg->setDepth(eit->first, 0);

	    TOKEN t;
	    t.setCount(1);
		
	    if(coloured)
	    {
		ALGORITHM * engine = storage.getEngine();
		list<EXPRESSION> le = eit->second.getAttributes();
		for(list<EXPRESSION>::iterator ita = le.begin(); ita != le.end(); ita++)
			t.addAttributes(engine->run(*ita));
	    }

	    initial.putToken(eit->first, t);

	    iota.setMatch(eit->first, eit->first);
	 }

	 pg->PTNET::setInitialMarking(initial);
	 pg->PTNET::setCurrentMarking(initial);
	 pg->setIota(iota);
	 pg->setGrammar(this->grammar);

	 ALGORITHM* fold = this->load("call_fold");
	 fold->initialize(pg);
	 ALGORITHM* unfold = this->load("call_unfold");
	 unfold->initialize(pg);

	 map<IDTYPE, RULE>* rules = this->grammar->getRules();
	 map<IDTYPE, RULE>::iterator ruleit;

	 set<IDTYPE> sori;  // set of rule IDs corresponding to *rules
	 set<IDTYPE>::iterator sorit;
	 set<IDTYPE> current_sori;  // goal: avoid unproductive loops
	 unsigned storednobj = pg->getVertices().size() + pg->getEdges().size() + pg->getTransitions().size();
	 bool wholesori = true;
	 /*
	  * initialize sori := *rules
	  */
	 for(ruleit = rules->begin(); ruleit != rules->end(); ruleit++)
	 {
	    sori.insert(ruleit->first);
	 }
	 current_sori = sori;

	 bool terminated = false;
	 unsigned step = 0;
	 iteration = 0;
	 while(!terminated)
	 {
	    if(maxstep==0)
	     break;
	    
	    terminated = true;
	    if(trace)
	    {
	       char graphname[256];
	       sprintf(graphname, "step_%i", step);
	       /* writer.addGraph(pg, graphname); */
	    }

            if(!quiet) cout << "step " << step+1 << ":\n" << flush;
	    
	    // iteration
	    iteration++;
	  
	    // find matching algorithm
	    bool canFold = false;
	    bool canUnfold = false;
	    MATCH* phi = new MATCH();
            MATCH* phi2 = new MATCH();
	    ALGORITHM* findmatch = this->load("call_findmatch");
	    findmatch->initialize((HYPERGRAPH*) pg, phi, phi2, &canFold, &canUnfold, pg);
	      
	    /*
	     * find a rule satisfying the conditions of Def. (11)
	     */
	    if(!quiet) cout << "searching matches" << flush;
	    for(sorit = current_sori.begin(); sorit != current_sori.end(); sorit++)
	    {
	       ruleit = rules->find(*sorit);
	       if(!quiet) cout << '.' << flush;
	       /*
	        * results of invoking the FINDMATCH algorithm (with the first match handler):
		*    (*) fold, unfold or neither of these?
		*    (*) where to fold/unfold
	        */
	       bool foundMatch = false;

	       MATCH * m_temp = findmatch->run(ruleit->second.getLeft(), &foundMatch, ruleit->first);
		// :)
		if(m_temp)
		 delete(m_temp);
		
	       /*
	        * apply the results
		*/
	       if(debug) cout << "canFold: " << ((unsigned)canFold) << " (" << canFold << ")\n";
	       if(debug) cout << "canUnfold: " << ((unsigned)canUnfold) << " (" << canUnfold << ")\n";
	       if(canFold)
	       {
	          step++;		  
	          if(!quiet) cout << "folding rule " << ruleit->first << " ("
		                  << ruleit->second.getName() << ") ..." << flush;
		  if(debug)
		  {
		     cout << "\nPHI2 promulgation: reading phi2 from " << phi2
		          << " (phi: " << phi << ")\n";
		     cout << "\n-~-~-~-\nmatch phi: " << *phi
		          << "\nmatch phi2: " << *phi2 << "\n-~-~-~-\n";
		  }
		  fold->run(&(ruleit->second), *phi2, *phi);
		  if(!quiet) cout << "done.\n" << flush;
		  if(debug) cout << "\n" << *pg << "\n\n";
		  terminated = false;
		  if(nfold && force_alternation)
		  {
		     wholesori = false;
		     current_sori.erase(ruleit->first);
		  }
                  if(red) rt->run();
		  break;  // and continue with "while" loop
	       }
	       else if(canUnfold)
	       {
	          step++;
		  if(!quiet) cout << "unfolding rule " << ruleit->first << " ("
		                  << ruleit->second.getName() << ") ..." << flush;
		  unfold->run(&(ruleit->second), *phi);
		  if(!quiet) cout << "done.\n" << flush;
		  if(debug) cout << "\n" << *pg << "\n\n";
		  terminated = false;
		  if(nfold && force_alternation)
		  {
		     wholesori = false;
		     current_sori.erase(ruleit->first);
		  }
		  break;  // and continue with "while" loop
	       }
	       else if(debug) cout << "rule " << ruleit->first << " ("
		                   << ruleit->second.getName() << ") not applicable\n" << flush;
	       // else try the next rule
	    }   
	    
	    if(findmatch)
		 delete findmatch;

	    if(debug)
	    {
	       cout << "\ncurrent size of unfolding:"
	            << "\ntransitions:\t" << pg->getTransitions().size()
		    << "\nedges:      \t" << pg->getEdges().size()
		    << "\nvertices:   \t" << pg->getVertices().size() << "\n\n";
	    }

	    if(force_alternation && (terminated == true) && !wholesori)
	    {
	       unsigned nobj = pg->getVertices().size() + pg->getEdges().size() + pg->getTransitions().size();
	       if(nobj > storednobj)
	       {
	          wholesori = true;
	          terminated = false;
	          current_sori = sori;
		  storednobj = nobj;
	       }
	       else if((storednobj >= nobj) && (!quiet))
	       {
	          cout << "\nobservation: size of Petri graph is stagnating.\n";
		  cout << "the unfolding process will be stopped due to this observation.\n\n" << flush;
	       }
	    }
	    
	    if(step >= maxstep) terminated = true;

	    if(phi)
	     delete phi;
	    if(phi2)
             delete phi2;

	   // timeout
	   if(timeout != 0)
	   {
		if(get_time() > timeout)
		{
			terminated = true;
			cout << "----- Timeout Termination -----" << endl;
		}
	   }
	 }
	 
	 if(red)
	 {
	    /*
	     * reduce transitions with identical labels and presets
	     */
	    rt->run();
	 }
	 if(!quiet) cout << "calculation finished.\n" << flush;
      }
   };
};
using namespace algorithms;
using namespace algorithms::petrigraph;

#endif
