/***************************************************************************
 *   Copyright (C) 2006 by root   *
 *   root@Vitali   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "attr_example_checker.h"
#include "algorithm_storage.h"
#include "find_all_matches.h"
#include "matchhandler.h"
#include "binder.h"
#include "fire_transition.h"

namespace attributes {

ATTR_EXAMPLE_CHECKER::ATTR_EXAMPLE_CHECKER()
{
}


ATTR_EXAMPLE_CHECKER::~ATTR_EXAMPLE_CHECKER()
{
}

void ATTR_EXAMPLE_CHECKER::initialize(PETRIGRAPH * pg, COUNTER_EXAMPLE & ce)
{
	this->pn = (PTNET*)pg;
	this->ce = ce;
}

bool ATTR_EXAMPLE_CHECKER::run(bool real_engine)
{
	if(!coloured)
	{
		cout << "Possible conflict between attributed and not attributed parts\n";
		exit(0);
	}

	// attribute engine
	engine = NULL;
	
	if(real_engine)
	 engine = storage.getStdEngine();
	else
	 engine = storage.getEngine();

	FIRE_TRANSITION ft;
	ft.initialize(pn, engine);

	list<MARKING> current_markings;
	current_markings.push_back(pn->getInitialMarking());

	list<MARKING> obtained_markings;
	
	for(unsigned int index = 0; index < ce.trans_ids.size(); index++)
	{
		obtained_markings.clear();

		for(list<MARKING>::iterator itm = current_markings.begin(); itm != current_markings.end(); itm++)
		{
			list<BINDING> lb = ft.getEnabledBindings(*itm, ce.trans_ids[index]);
			list<BINDING> selected_bindings;
			for(list<BINDING>::iterator itb = lb.begin(); itb != lb.end(); itb++)
			{	
				// both real and abstract attributes
				// call abstractBinding() because ft checkes the binding with abstraction,
				// but gives back the original binding
				if(BINDING::compareBindingsAbs(ce.bindings[index], engine->abstractBinding(*itb)))
				 selected_bindings.push_back(*itb);
			}

			// can not fire
			if(selected_bindings.size()==0)
			 continue;

			// fire
			for(list<BINDING>::iterator its = selected_bindings.begin(); its != selected_bindings.end(); its++)
			 obtained_markings.push_back(ft.run(pn->getTransition(ce.trans_ids[index]), *itm, *its));
		}

		// empty set of obtained markings
		if(obtained_markings.size()==0)
		 return false;

		current_markings = obtained_markings;
	}

	return true;
}

}
