/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <string.h>
#include "scenario.h"
#include "algorithm_storage.h"

using namespace std;

int main(int argc, char** argv)
{
   const char* usage = "Usage: augur -sc=<sc> [-q] [-d] [-t] [-nc] [-nr] [-cut] [-kin] [-fa] \
[-depth=k | -nfold=k] [-step=n] [-nmap] (infile)* (outfile)*\n\nOptions:\n\n\
-sc=<sc>\tcall scenario <sc>, for instance -sc=aunfold\n\n\
-db=<db>\tuse database <db>, for instance -db=db/mydb.xml\n\n\
-q quiet\tquiet mode, disable verbose output\n\
-d debug\tdebug mode, promulgate technical details of the implementation\n\
-t trace\tinclude each step of calculation into output file\n\
-nc     \tdo not check covering information\n\
-nr     \twithout irredundancy condition\n\
-cut    \tchecking for cut-off\n\
-kin    \tkeep isolated nodes\n\
-fa     \tforce rule alternation\n\
-depth=k\tbuild over-approximation C(k)\n\
-nfold=k\tno folding steps, maximal depth k\n\
-step=n \tmaximal number of unfolding steps\n\
-nmap   \twithout mapping functions\n\
-timeout \ttimeout in seconds\n\
infile  \tinput depending on the chosen scenario\n\
outfile \toutput destination depending on the chosen scenario\n\n";

   /*
    * which database?
    */
   unsigned i;
   string database = "../db/default.xml";
   for(i=1; i < argc; i++)
   {
      if(*argv[i] != '-') break;
      else if(strstr(argv[i], "-db=") == argv[i])
      {
         database = (string)&(argv[i][4]);	 
      }
   }

   /*
    * load default arguments and algorithm DB
    */
   if(!quiet) cout << "loading database\n";
   storage.initialize(database.c_str());

    /*
    * parse arguments
    */
   string scenario = "";
   for(i=1; i < argc; i++)
   {
//      if(!strcmp(argv[i], "-test")) test = true;
      if(!strcmp(argv[i], "-q")) quiet = true;
      else if(!strcmp(argv[i], "-d")) debug = true;
//      else if(!strcmp(argv[i], "-cut")) cutoff = true;
      else if(!strcmp(argv[i], "-nc")) ncov = true;
      else if(!strcmp(argv[i], "-nr")) red = false;
      else if(!strcmp(argv[i], "-nmap")) mapping = false;
//      else if(!strcmp(argv[i], "-kin")) remove_isolated_nodes = false;
      else if(!strcmp(argv[i], "-fa")) force_alternation = true;
      else if(!strcmp(argv[i], "-cov_test")) cov_test = true;
      else if(!strcmp(argv[i], "-coloured")) coloured = true;
      else if(!strcmp(argv[i], "-pid")) pid = true;
      else if(strstr(argv[i], "-depth=") == argv[i])
      {
         char buffer[10];
	 strcpy(buffer, argv[i] + 7);
	 if(atoi(buffer) < 0)
	 {
	    cout << "depth must be some positive number or 0.\n";
	    return 1;
	 }
	 else depth = atoi(buffer);
      }
      else if(strstr(argv[i], "-nfold=") == argv[i])
      {
         if(depth > 0)
	 {
	    cout << "depth and nfold cannot be defined together.\n";
	    return 1;
	 }
	 else
	 {
	    nfold = true;
	    char buffer[10];
	    strcpy(buffer, argv[i] + 7);
	    if(atoi(buffer) < 0)
	    {
   	       cout << "unfolding depth must be some positive number or 0.\n";
	       return 1;
	    }
	    else unfdepth = atoi(buffer);
	 }
      }
      else if(strstr(argv[i], "-step=") == argv[i])
      {
         char buffer[10];
	 strcpy(buffer, argv[i] + 6);
	 if(0 > atoi(buffer))
	 {
	    cout << "invalid (negative) number of steps.\n";
	    return 1;
	 }
	 else maxstep = atoi(buffer);
      }
      else if(strstr(argv[i], "-db=") == argv[i]) continue;
      else if(strstr(argv[i], "-sc=") == argv[i])
      {
         scenario = (string)&(argv[i][4]);
      }
      else if(strstr(argv[i], "-ref_loop_limit=") == argv[i])
      {
         char buffer[10];
	 strcpy(buffer, argv[i] + 12);
	 if(0 > atoi(buffer))
	 {
	    cout << "invalid (negative) loop limit.\n";
	    return 1;
	 }
	 else ref_loop_limit = atoi(buffer);
      }
      else if(strstr(argv[i], "-timeout=") == argv[i])
      {
         char buffer[10];
	 strcpy(buffer, argv[i] + 9);
	 if(0 > atoi(buffer))
	 {
	    cout << "invalid (negative) timeout.\n";
	    return 1;
	 }
	 else timeout = atoi(buffer);
      }
      else if(strstr(argv[i], "-size=") == argv[i])
      {
         char buffer[10];
	 strcpy(buffer, argv[i] + 6);
	 if(atof(buffer) <= 0)
	 {
	    cout << "size must be some positive number.\n";
	    return 1;
	 }
	 else rules_size = atof(buffer);
      }
      else if(*argv[i] == '-')
      {
         if (*(argv[i] + 1) != ' ')
	 {
	    cout << "unknown option: " << argv[i] << "\n\n" << usage;
            return -1;
	 }
	 else
	 {
	    i++;
	    break;
	 }
      }
      else break;
   }
//   if(debug && trace) test = true;

   /*
    * construct argument (file) list for the scenario
    */
   list<string> argl;
   while(i < argc)
   {
      argl.push_back((string)(argv[i]));
      i++;
   }

   if(scenario == "")
   {
      cout << "no scenario provided!\n\n" << usage;
      return -1;
   }
   
   PATH_DB np = PATH_DB();
   np.push_back(pair<string, LABEL>("main", ("scenario_"+scenario)));
   SCENARIO* scn;
   try
   {
      scn = (SCENARIO*)(storage.getAlgorithm(np));
   }
   catch(exception e)
   {
      cout << "scenario \"" << scenario << "\" not found.\n\n" << usage;
   }
   scn->initialize(argl);

   cout << "Running scenario ..." << endl;
   scn->run();
   
   return 0;
}

