/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#ifndef SCENARIOS_AUNFOLD_CPP
#define SCENARIOS_AUNFOLD_CPP

#include "aunfold.h"

#include "approximated_unfolding.h"
#include "gts.h"
#include "GTXLReader.h"
#include "GTXLWriter.h"
#include "petrigraph.h"
#include "gxlreader.h"
#include "gxlwriter.h"
#include "gtxl_nac_reader.h"

namespace scenarios
{
   AUNFOLD::AUNFOLD()
   {
      this->ID = ALGORITHM_AUNFOLD;
      this->name = "aunfold";
      
      this->inspec.push_back(pair<string, DATA_FORMAT>("grammar", GTXL));
      this->outspec.push_back(pair<string, DATA_FORMAT>("petrigraph", GXL_PETRIGRAPH));
      
      this->call_descriptor = "Usage: augur -sc=aunfold [options] grammar petrigraph\n\n\
grammar\tgraph grammar in GTXL format\n\
petrigraph\toutput destination for the unfolding, a Petri graph in GXL format\n\n";
   }
   AUNFOLD::~AUNFOLD() {}
   
   void AUNFOLD::run()
   {
      GTS grammar;
      ALGORITHM* reader = this->load("call_gtxl_reader");
      reader->initialize((char*)(this->filemap["grammar"]).c_str());
      reader->run(grammar);

      ALGORITHM* writer = this->load("call_writer");
      writer->initialize((char*)(this->filemap["petrigraph"]).c_str());

      PETRIGRAPH pg;
      init_time();
      ALGORITHM* unfolder = new APPROXIMATED_UNFOLDING();
      unfolder->initialize(&grammar, writer);
      unfolder->run(&pg);
      double timeused = get_time();
      cout << "time used: " << timeused << " sec.\n";

      if(debug) cout << "\n" << pg << "\n";

      /*
       * promulgate Petri graph size
       */
      cout << "edges: " << pg.getEdges().size()
           << ", vertices: " << pg.getVertices().size()
           << ", transitions: " << pg.getTransitions().size() << "\n";
   
      // try
      // {
         writer->run(pg, "augur2.out");
      // }
      // catch(exception e)
      // {
         // cerr << "error while writing output file.\n";
      // }
   }
};
using namespace scenarios;

#endif
