/***************************************************************************
 *   Copyright (C) 2006 by root   *
 *   root@Vitali   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "binder.h"

namespace attributes {

BINDER::BINDER()
{
}


BINDER::~BINDER()
{
}

void BINDER::initialize(HYPERGRAPH * hg)
{
	this->hg = hg;
}

list<BINDING> BINDER::run(HYPERGRAPH * left, MATCH * m)
{
	// nodes as places (attributes in nodes) is not yet supported
	if(nodes_as_places)
	{
		cout << "\nnodes as places (attributes in nodes) is not yet supported\n";
		exit(1);
	}

	list<BINDING> result;

	map<IDTYPE, EDGE> edges_hg = hg->getEdges();
	map<IDTYPE, EDGE> edges_left = left->getEdges();

	// check if there are some attributes at all
	bool is_any = false;
	for(map<IDTYPE, EDGE>::iterator ite = edges_left.begin(); ite != edges_left.end(); ite++)
	{
		if(ite->second.getAttributes().size() != 0)
		{
			is_any = true;
			break;
		}
	}
	
	// only not attributed
	if(!is_any)
	{
		BINDING b;
		result.push_back(b);
		return result;
	}

	for(map<IDTYPE, EDGE>::iterator ite = edges_left.begin(); ite != edges_left.end(); ite++)
	{
		list<EXPRESSION> le1 = ite->second.getAttributes();

		// no attributes
		if(le1.size() == 0)
		 continue;

		// find image
		if(! m->hasMatch(ite->first))
		{
			cout << "\nBINDER :: internal error (no matching found)\n";
			exit(1);
		}

		IDTYPE id = m->getMatch(ite->first);

		// no such edge
		if(edges_hg.find(id) == edges_hg.end())
		{
			cout << "\nBINDER :: internal error (no edge found)\n";
			exit(1);
		}

		list<EXPRESSION> le2 = edges_hg[id].getAttributes();

		// empty edge 
		// binding can not be constructed
		if(le2.size()==0)
		{
			result.clear();
			cout << "\nwarning: BINDER :: empty edge\n";
			break;
		}

		for(list<EXPRESSION>::iterator ite = le1.begin(); ite != le1.end(); ite++)
		{
			string name1 = ite->getName();
				
			bool found = false;
			
			// find expression by name
			list<EXPRESSION>::iterator ite_1;
			list<BINDING> lb;
			for(ite_1 = le2.begin(); ite_1 != le2.end(); ite_1++)
			{
				string name2 = ite_1->getName();
				if(name1 == name2)
				{
					// we use here BINDING in order to have all checks,
					// it will be later (mergeList) transformed in another data structure (variables)
					BINDING b;
					b.addBinding(*ite, *ite_1);
					lb.push_back(b);
					found = true;
				}
			}

			// not found
			if(!found)
			{
				cout << "\nBINDER :: internal error (no expression found)\n";
				exit(1);
			}

			result = mergeLists(result, lb);
		}
	}

	return result;
}

list<BINDING> BINDER::mergeLists(list<BINDING> & l1, list<BINDING> & l2)
{
	list<BINDING> result;

	// l2 will be transformed in this map
	map<string, list<ExpressionElement> > variables;

	for(list<BINDING>::iterator it = l2.begin(); it != l2.end(); it++)
	{
		
		list<string> ls = it->getVariableNames();

		for(list<string>::iterator itstr = ls.begin(); itstr != ls.end(); itstr++)
		{
			list<ExpressionElement> lee;

			// find in map
			if(variables.find(*itstr) == variables.end())
			{
				lee.push_back(it->getVariable(*itstr));
			}
			else
			{
				lee = variables[*itstr];
				lee.push_back(it->getVariable(*itstr));
			}

			variables[*itstr] = lee;
		}
	}

	if(variables.size() == 0)
	{
		cout << "\nerror in BINDER: (empty variable list)\n";
		exit(1);
	}

	if(l1.size() == 0)
	{
		// take the first variable
		map<string, list<ExpressionElement> >::iterator itv = variables.begin();

		for(list<ExpressionElement>::iterator iteel = itv->second.begin(); iteel != itv->second.end(); iteel++)
		{
			BINDING b;
			b.setVariable(itv->first, *iteel);
			result.push_back(b);
		}

		// delete first variable (we already have it in the list)
		variables.erase(itv);
	}
	else
	{
		result = l1;
	}

	while(variables.size() != 0)
	{
		list<BINDING> result_;

		// take the first variable
		map<string, list<ExpressionElement> >::iterator itv = variables.begin();

		for(list<BINDING>::iterator it1 = result.begin(); it1 != result.end(); it1++)
		{
			for(list<ExpressionElement>::iterator iteel = itv->second.begin(); iteel != itv->second.end(); iteel++)
			{
				BINDING b = *it1;

				b.setVariable(itv->first, *iteel);

				result_.push_back(b);
			}
		}

		result = result_;

		// delete
		variables.erase(itv);
	}

	return result;
}

void BINDER::initialize(PTNET * pn)
{
	this->pn = pn;
}

BINDING BINDER::unifyList(list<BINDING> l)
{
	BINDING result;

	for(list<BINDING>::iterator it = l.begin(); it != l.end(); it++)
	{
		
		list<string> ls = it->getVariableNames();

		for(list<string>::iterator itstr = ls.begin(); itstr != ls.end(); itstr++)
		{
			result.setVariable(*itstr, it->getVariable(*itstr));
		}
	}

	return result;
}

list<BINDING> BINDER::run(MARKING * marking, IDTYPE idtrans)
{
	// nodes as places (attributes in nodes) is not yet supported
	if(nodes_as_places)
	{
		cout << "\nnodes as places (attributes in nodes) is not yet supported\n";
		exit(1);
	}

	list<BINDING> result;

	map< IDTYPE, vector<TOKEN> > marking_in = marking->getMarkings();
	vector<TCONNECTION> connections = pn->getPreset(idtrans);

	// check if there are some attributes at all
	bool is_any = false;
	for(vector<TCONNECTION>::iterator ite = connections.begin(); ite != connections.end(); ite++)
	{
		if(ite->attributes.size() != 0)
		{
			is_any = true;
			break;
		}
	}
	
	// only not attributed
	if(!is_any)
	{
		BINDING b;
		result.push_back(b);
		return result;
	}

	// List of bindings for each place
	// Each place has a list of bindings for tokens
	// It is merged each time with all possible variants
	list<BINDING> places_list;

	for(vector<TCONNECTION>::iterator ite = connections.begin(); ite != connections.end(); ite++)
	{
		list<EXPRESSION> le1 = ite->attributes;

		// no attributes
		if(le1.size() == 0)
		 continue;

		// find image
		vector<TOKEN> image = marking_in[ite->ID];

		list<BINDING> lb_tokens;
	
		for(vector<TOKEN>::iterator itv = image.begin(); itv!=image.end(); itv++)
		{
			// check count
			if(itv->getCount()<ite->weight)
				continue;

			list<EXPRESSION> le2 = itv->getAttributes();

			// token without attributes  
			// binding can not be constructed
			if(le2.size()==0)
			{
				// result.clear();
				// return result;
				
				cout << "BINDER Error: Token without attributes is found on attributed place\n";
				exit(1);
			}

			// binding
			BINDING b;
	
			for(list<EXPRESSION>::iterator ite_e = le1.begin(); ite_e != le1.end(); ite_e++)
			{
				string name1 = ite_e->getName();
				// list<BINDING> lb;	
				bool found = false;
				
				// find expression by name
				list<EXPRESSION>::iterator ite_1;
				for(ite_1 = le2.begin(); ite_1 != le2.end(); ite_1++)
				{
					string name2 = ite_1->getName();
					if(name1 == name2)
					{
						b.addBinding(*ite_e, *ite_1);
						found = true;
						break;
					}
				}
	
				// not found
				if(!found)
				{
					cout << "\nBINDER Error:: no expression with name " << name1 << " is found on place " << ite->ID <<"\n";
					exit(1);
				}
			}

			if(b.getSize()!=0)
			 lb_tokens.push_back(b);
		}

		if(places_list.size()==0)
		{
		 places_list = lb_tokens;
		}
		else
		{
			list<BINDING> places_list_;

			for(list<BINDING>::iterator it_places = places_list.begin(); it_places != places_list.end(); it_places++)
			{
				// make lists (compatibility with othe run method in the signature og mergeLists)
				list<BINDING> lb1;
				lb1.push_back(*it_places);
				for(list<BINDING>::iterator it_tokens = lb_tokens.begin(); it_tokens != lb_tokens.end(); it_tokens++)
				{
					list<BINDING> lb2;
					lb2.push_back(*it_tokens);

					places_list_.push_back(unifyList(mergeLists(lb1,lb2)));
				}
			}

			places_list = places_list_;
		}
	}	

	// write binding
	/*
	cout << "\nBinding:\n";
	for(list<BINDING>::iterator it = places_lists.begin(); it != places_lists.end(); it++)
	{
		list<string> ls = it->getVariableNames();
		for(list<string>::iterator its = ls.begin(); its != ls.end(); its++)
			cout << *its << " : " << it->getVariable(*its).int_evaluated << " , " << it->getVariable(*its).str_evaluated << "\n";
	}
	cout << "------------\n";
	*/
	return places_list;
}

}

