/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "binding.h"

namespace attributes {

BINDING::BINDING()
{
}


BINDING::~BINDING()
{
}

// clear binding
void BINDING::clear()
{
	variables.clear();
}

// add binding
void BINDING::addBinding(EXPRESSION e1, EXPRESSION e2)
{
	vector<ExpressionElement> vee1 = e1.getExprParsed();
	vector<ExpressionElement> vee2 = e2.getExprParsed();

	// check sizes
	if(vee1.size() != vee2.size())
	{
		cout << "\nerror while BINDING\n";
		exit(1);
	} 
	
	vector<ExpressionElement>::iterator it1 = vee1.begin();
	for(vector<ExpressionElement>::iterator it2 = vee2.begin();
		it2 != vee2.end(); it2++)
	{
		// check, that e1 is variable
		if(it1->type != VAR_UNKNOWN)
		{
			cout << "\nerror: only variables should be used in BINDING\n";
			exit(1);
		}
	
		// check if such variable already exists
		if(variables.find(it1->variable_name)!=variables.end())
		{
			cout << "\nerror: repeated variable name in BINDING\n Variable: " << it1->variable_name << "\n";
			exit(1);
		}
	
		variables[it1->variable_name] = *it2;

		it1++;
	}
}

ExpressionElement BINDING::getVariable(string name)
{
	return variables[name];
}

void BINDING::setVariable(string name, ExpressionElement value)
{
	variables[name] = value;
}

list<string> BINDING::getVariableNames()
{
	list<string> result;

	for(map<string, ExpressionElement>::iterator it = variables.begin(); it != variables.end(); it++)
	{
		result.push_back(it->first);
	}

	return result; 
}

bool BINDING::hasVariable(string name)
{
	if(variables.find(name) != variables.end())
		return true;
	else
		return false;
}

int BINDING::getSize()
{
	return variables.size();
}

bool BINDING::compareBindingsAbs(BINDING b1, BINDING b2)
{
	list<string> lv = b1.getVariableNames();

	bool result = true;

	for(list<string>::iterator it = lv.begin(); it != lv.end(); it++)
	{
		ExpressionElement ee1 = b1.getVariable(*it);
		ExpressionElement ee2 = b2.getVariable(*it);

		if(ee1.bool_val != ee2.bool_val || ee1.str_evaluated != ee2.str_evaluated || ee1.int_evaluated != ee2.int_evaluated)
		{
			result = false;
			break;
		}
	}

	return result;
}

bool BINDING::compareBindings(BINDING b1, BINDING b2)
{
	list<string> lv = b1.getVariableNames();

	bool result = true;

	for(list<string>::iterator it = lv.begin(); it != lv.end(); it++)
	{
		ExpressionElement ee1 = b1.getVariable(*it);
		ExpressionElement ee2 = b2.getVariable(*it);

		if(ee1.bool_val != ee2.bool_val || ee1.str_val != ee2.str_val || ee1.int_val != ee2.int_val)
		{
			result = false;
			break;
		}
	}

	return result;
}

ostream &operator<<(ostream &ost, BINDING &b)
{
	int counter = b.variables.size();
	for(map<string, ExpressionElement>::iterator it = b.variables.begin(); it != b.variables.end(); it++)
	{
		switch(it->second.type)
		{
			case NONE_:
				counter--;
				ost << it->first << ": not initialized";
				break;

			case OP_NUM_PLUS:
			case OP_NUM_MINUS:
			case OP_NUM_MULT:
			case OP_NUM_DIV:
			case OP_NUM_MOD:
			case NUMBER:
			case VAR_NUM:
			case OP_STR_LEN:
			case OP_NUM_UMINUS:
				counter--;
				ost << it->first << "(int): " << it->second.int_evaluated;
				break;
			case OP_BOOL_AND:
			case OP_BOOL_OR:
			case OP_BOOL_NOT:
			case OP_NUM_BOOL_EQ:
			case OP_NUM_BOOL_LE:
			case OP_NUM_BOOL_GE:
			case OP_NUM_BOOL_NE:
			case OP_NUM_BOOL_L:
			case OP_NUM_BOOL_G:
			case BOOLEAN:
			case OP_STR_BOOL_EQ:
				counter--;
				ost << it->first << "(bool): " << it->second.bool_val;
				break;
			case _STR_:
			case VAR_STR:
			case OP_STR_CONC:
				counter--;
				ost << it->first << "(str): " << it->second.str_evaluated;
				break;
			case VAR_UNKNOWN:
				counter--;
				ost << it->first << ": str=" << it->second.str_evaluated << ", int=" << it->second.int_evaluated;
				break;
			default:
				break;
		}

		if(counter != 0)
		 ost << ", ";
	}

	ost << endl;
	
	return ost;
}

};
