#include "ConstraintSet.h"


ConstraintSet::ConstraintSet() { count = 0; } // constructor, initialization

int ConstraintSet::size() { return cs.size(); }// return number of Constraints
bool ConstraintSet::empty() { return cs.empty(); } // is this CS empty?

// delete and return any of the constraints
Constraint* ConstraintSet::pop() {
  csit = cs.begin();
  int id = csit->first;
  Constraint* res = csit->second;
  this->del(id);
  return res;
}

// delete constraint given its ID
void ConstraintSet::del(int id) { cs.erase(id); }

// add a given constraint and assign it a unique ID
int ConstraintSet::add(Constraint* in) {
  count++;
  cs[count] = in;
  return count;
}

/* move constraints from given ConstraintSet to this set, 
   assigning a unique ID to all constraints,
*/
void ConstraintSet::move(ConstraintSet* in) {
  while(!in->empty()) {
    this->add(in->pop());
  }
  
  delete in;
}

// iterator functions: place iterator at begining/end of this CS
ConstraintSet::iterator ConstraintSet::begin() { return cs.begin(); }
ConstraintSet::iterator ConstraintSet::end() { return cs.end(); }

// return ID of constraint, where iterator is pointing to
int ConstraintSet::getID(ConstraintSet::iterator iter) { return iter->first; }

// return constraint, where iterator is pointing to
Constraint* ConstraintSet::getConstraint(ConstraintSet::iterator iter) { 
  return iter->second; 
}

// does the input-constraint cover this CS, i.e.
// do we have any constraint in CS which is covered by the input-constraint?
bool ConstraintSet::covers(Constraint* in) { 
  for (csit=cs.begin(); csit!=cs.end(); ++csit) {
    if ((*in) <= csit->second) { 
      return true; 
    }
  }
  return false; 
}

// does this CS cover the input-constraint, i.e.
// do we have any constraint in CS covering the input-constraint?
bool ConstraintSet::isCoveredBy(Constraint* in) { 
  for (csit=cs.begin(); csit!=cs.end(); ++csit) {
    if ((*(csit->second)) <= in) { 
      return true; 
    }
  }
  return false; 
}

// return a string-representation of this CS
string ConstraintSet::asString(stringMap* placeNames) { 
  ostringstream os;
  for (csit=cs.begin(); csit!=cs.end(); ++csit) { 
    os << "Constraint with ID " << csit->first << "\t: ";
    os << (csit->second)->asString(placeNames); 
  } 
  return os.str();
}

// return a new copy of this CS
ConstraintSet* ConstraintSet::copy() {
  ConstraintSet* res = new ConstraintSet;
  for (csit=cs.begin(); csit!=cs.end(); ++csit) {
    res->add(csit->second->copy());
  }
  return res;
}

// are all of the constraints in this CS smaller
// as the input-constraint at place, specified by given index?
bool ConstraintSet::smallerAs(Constraint* in, int index) { 
  
  // ??  if (this->isEmpty()) { return false; }
  for (csit=cs.begin(); csit!=cs.end(); ++csit) {
    if (!(csit->second->smallerAs(in,index))) { 
      return false;
    }
  }
  return true;
}

