#include <sstream>
#include <map>
#include "stringMap.h"
#include "Constraint.h"

using namespace std;

#ifndef CONSTRAINTSET_H
#define CONSTRAINTSET_H

/* this class provides a container for constraints,
   (set of constraints CS)
   and iterator-functions for access to constraints
   within the set.
*/
class ConstraintSet {
  map<int,Constraint*> cs; // the internal container for constraints
  map<int,Constraint*>::iterator csit; // internal used iterator
  int count; // used for assigning unique ID to constraints
 public:
  ConstraintSet(); // constructor

  int size(); // return number of Constraints
  bool empty(); // check if CS is empty
  
  // update/access methods 
  Constraint* pop(); // delete an element of the CS, and return it
  void del(int id); // delete Constraint with given ID
  int add(Constraint* in); // add a Constraint, and return its id

  /* move constraints from given ConstraintSet to this set, 
     assigning a new ID to each contraint */
  void move(ConstraintSet* in);
  
  // iterator: type and methods
  typedef map<int,Constraint*>::iterator iterator;
  iterator begin();
  iterator end();
  
  // access to CS via iterator
  
  // get the ID of the constraint, which iterator is pointing to
  int getID(ConstraintSet::iterator iter);
  // get the constraint, which iterator is pointing to
  Constraint* getConstraint(ConstraintSet::iterator iter);
  
  // does this set cover the input-constraint?
  bool covers(Constraint* in); 
  
  // does the input-constraint cover any of the constraints in this CS?
  bool isCoveredBy(Constraint* in);
  
  // return a new copy of this CS
  ConstraintSet* copy();
  
  // provide string-representation of this CS
  string asString(stringMap* placeNames);
  
  // are all of the constraints in this CS smaller as input-constraint,
  // at given input-place/index?
  bool smallerAs(Constraint* in, int index);
};

#endif

