#include <sstream>
#include <time.h>

#include "stringMap.h"
#include "intSet.h"
#include "Constraint.h"
#include "ConstraintSet.h"
#include "PredicateTransformer.h"
#include "PredicateTransformerSet.h"

#ifndef CVS_H
#define CVS_H

// extern long timeout;

/* Class Cvs, which is implementing a Constraint Verification
   System, in this case the Backward Reachability Algorithm.
 */
class Cvs {
  ConstraintSet* V; // List of visited Constraints
  ConstraintSet* W; // Working List of Constraints
  ConstraintSet* initial; // set of initial Constraints
  PredicateTransformerSet* T; // set of Transitions / Predicate Transformers
  bool verbose; // used to show, how much messages to print
  bool quiet;// used to show, that a minimum of messages shall be print
  stringMap* placeNames; // available place names, and their Integer-IDs
  stringMap* transformerNames; // available transition names, and their Integer-IDs
 public:
  Cvs(PredicateTransformerSet* T,ConstraintSet* initial_in,ConstraintSet* final_in,
      stringMap* inPlaceNames,stringMap* inTransNames); // Constructor
  void setVerbose(); // print verbose messages
  void setQuiet(); // print just the result, "true" or "false", nothing else
  intSet* select(PredicateTransformerSet* ps,Constraint* c);

  /* run the Backward Rechability Algorithm, and return the Result,
     "true", if final marking(s) is coverable,
     "false", if final marking(s) is not coverable,
  */
  bool run(char * tracefile);
};

#endif
