#include "InputCvs.h"
#include <fstream>
#include <sstream>
#include <stdio.h>

InputCvs::InputCvs() { // constructor
  final = new ConstraintSet;
  initial = new ConstraintSet;
  ptSet = new PredicateTransformerSet;
  placeNames = new stringMap;
  transformerNames = new stringMap;
  verbose = false;
}

// methods, for obtaining infos
PredicateTransformerSet* InputCvs::getPredicateTransformerSet() { return ptSet; }
ConstraintSet* InputCvs::getFinalConstraintSet() { return final; }  
ConstraintSet* InputCvs::getInitialConstraintSet() { return initial; }  
stringMap* InputCvs::getTransformerNames() { return transformerNames; }  
stringMap* InputCvs::getPlaceNames() { return placeNames; }  

bool InputCvs::run(PTNET * pn, list<MARKING> m) { 
 
  // places
  set<IDTYPE> places = pn->getPlaces();
  for(set<IDTYPE>::iterator its = places.begin(); its != places.end(); its++)
  {
    char buffer[16];
    sprintf(buffer, "%u", *its);

    placeNames->add(string(buffer));
  }

  // initial marking
  map<IDTYPE, unsigned> init_map;
  if(!coloured)
   init_map = pn->getInitialMarking().getUncoloured();
  else
  {
   cout << "\nProgramming Error: non-coloured bwra is called with coloured parameters\n";
   exit(1);
  }

  Constraint * newCI = new Constraint;

  for(map<IDTYPE, unsigned>::iterator it = init_map.begin(); it != init_map.end(); it++)
  {
   // is the place name known?
   char buffer[16];
   sprintf(buffer, "%u", it->first);
   
   int place_int = placeNames->has(string(buffer));
   if (place_int == 0)
   {
    cout << "\nBWRA: no such PLACE: " << buffer << "\n";
    exit(1);
   }
   newCI->set(place_int,it->second);
  }

  initial->add(newCI);

  // end marking
  for(list<MARKING>::iterator itlm = m.begin(); itlm != m.end(); itlm++)
  {	
	map<IDTYPE, unsigned> end_map;
	if(!coloured)
	end_map = itlm->getUncoloured();
	else
	{
	cout << "\nbwra coloured, not yet implemented\n";
	exit(1);
	}
	
	Constraint * newCF = new Constraint;
	
	for(map<IDTYPE, unsigned>::iterator it = end_map.begin(); it != end_map.end(); it++)
	{
	// is the place name known?
	char buffer[16];
	sprintf(buffer, "%u", it->first);
	
	int place_int = placeNames->has(string(buffer));
	if (place_int == 0)
	{
	cout << "\nBWRA: No such PLACE: " << buffer << "\n";
	exit(1);
	}
	newCF->set(place_int,it->second);
	}
	
	final->add(newCF);
  }
  // transitions

  map<IDTYPE, TRANSITION> transitions = pn->getTransitions();
  for(map<IDTYPE, TRANSITION>::iterator it = transitions.begin(); it != transitions.end(); it++)
  {
   PredicateTransformer* newPT = new PredicateTransformer;

   newPT->setName(it->second.getLabel());
   newPT->setID(it->first);

   // preset
   vector<TCONNECTION> preset = pn->getPreset(it->first);
   for(vector<TCONNECTION>::iterator it_pre = preset.begin(); it_pre != preset.end(); it_pre++)
   {
	char buffer[16];
	sprintf(buffer, "%u", it_pre->ID);
	int place_int = placeNames->has(string(buffer));
	if (place_int == 0)
   	{
    		cout << "\nBWRA: No such PLACE: " << buffer << "\n";
    		exit(1);
   	}
	newPT->setIn(place_int,it_pre->weight);
   }

   // postset
   vector<TCONNECTION> postset = pn->getPostset(it->first);
   for(vector<TCONNECTION>::iterator it_post = postset.begin(); it_post != postset.end(); it_post++)
   {
	char buffer[16];
	sprintf(buffer, "%u", it_post->ID);

	int place_int = placeNames->has(string(buffer));
	if (place_int == 0)
   	{
    		cout << "\nBWRA: No such PLACE: " << buffer << "\n";
    		exit(1);
   	}
	newPT->setOut(place_int,it_post->weight);
   }

   // set
  transformerNames->set(it->second.getLabel(),ptSet->add(newPT));

  }
  return true; 
}

// return parsed informations as string
string InputCvs::asString() {
  ostringstream os;
  os << "Initial Markings:\n"
     << this->getInitialConstraintSet()->asString(placeNames)
     << "Test Markings:\n"
     << this->getFinalConstraintSet()->asString(placeNames)
     << "Net:\n"
     << this->getPredicateTransformerSet()->asString(placeNames,transformerNames);
  return os.str();
}
