#include "Constraint.h"
#include "ConstraintSet.h"
#include "PredicateTransformer.h"
#include "PredicateTransformerSet.h"
#include "stringMap.h"

#include "../ptnet.h"
#include "../marking.h"

#ifndef InputCvs_H
#define InputCvs_H

/* class InputCvs: read Petri-Net description 
   and provide Final-Set of Constraints,
   Initial-Set of Constraints and PredicateTransformerSet.
*/
class InputCvs {
  // variables, to store information:
  // places, markings, transitions, place and transition names.
  ConstraintSet* final;
  ConstraintSet* initial;
  PredicateTransformerSet* ptSet;
  stringMap* placeNames;
  stringMap* transformerNames;

  // incr. info messages
  bool verbose;

 public:
  InputCvs();
  void setVerbose() { verbose = true; }  
  
  // initiate parsing on files
  bool run(PTNET * pn, list<MARKING> m);
  
  // provide markings and transitions
  PredicateTransformerSet* getPredicateTransformerSet();
  ConstraintSet* getFinalConstraintSet();
  ConstraintSet* getInitialConstraintSet();
  stringMap* getTransformerNames();
  stringMap* getPlaceNames();
  
  // provide markings and transitions as string
  string asString();
};
#endif
