#include "PredicateTransformer.h"


PredicateTransformer::PredicateTransformer() {} // constructor

// set place-value-pairs for in-/out-function
void PredicateTransformer::setIn(int key,int value) { 
  if (value > 0) {
    in[key] = value;
  } else {
    this->delIn(key);
  }
}

void PredicateTransformer::setOut(int key,int value) { 
  if (value > 0) {
    out[key] = value;
  } else {
    this->delOut(key);
  }
}

// del value for in-/out-function at given place
void PredicateTransformer::delIn(int key) { in.erase(key); }
void PredicateTransformer::delOut(int key) { out.erase(key); }

// get value for in-/out-function at given place
int PredicateTransformer::getIn(int key) { return in[key]; }
int PredicateTransformer::getOut(int key) { return out[key]; }

// is this PT empty?
bool PredicateTransformer::empty() { return (in.empty() && out.empty()); }

// return a new copy of this PT
PredicateTransformer* PredicateTransformer::copy() {
  PredicateTransformer* pt = new PredicateTransformer;
  for (i=in.begin(); i!=in.end(); ++i) {
    pt->setIn(i->first,i->second);
  }
  for (i=out.begin(); i!=out.end(); ++i) {
    pt->setOut(i->first,i->second);
  }
  pt->setName(name);
  pt->setID(trans_id);
  return pt;
}

/* generate new constraint by applying this Transition to input-constraint,
   for all places p of out-function: input(p) = max(input(p) - out(p),0)
   for all places p of in-function: input(p) = input(p) + in(p) */
Constraint* PredicateTransformer::apply(Constraint* input){
  int temp = 0;
  int input_val = 0; 
  int my_val = 0;
  int index = 0; // storing place-index
  
  Constraint* res = input->copy(); // working on a new copy
  
  // set the name and id
  TRANSITION_INFO ti;
  ti.name = name;
  ti.ID = trans_id;
  res->addTransition(ti);
  
  // apply out-function
  for (i=out.begin(); i!=out.end(); ++i) {
    index = i->first;
    my_val = i->second;
    input_val = res->get(index);
    temp = input_val - my_val;
    if (temp <= 0) {
      res->del(index); // 0-values not needed
    } else {
      res->set(index,temp); // save value > 0
    }
  }
  
  // apply in-function
  for (i=in.begin(); i!=in.end(); ++i) {
    res->set(i->first,res->get(i->first)+(i->second));
  }
    
  return res;
}

/* check, if this PT is deficient for place p at given constraint c,
   (if c(p) < out(PT)(p)),
   and in(input-PT)(p) > 0 
*/
bool PredicateTransformer::deficient(PredicateTransformer* p, Constraint* c) {
  for (i=out.begin(); i!=out.end(); ++i) {
    if ((i->second > c->get(i->first)) && (p->getIn(i->first) > 0)) {
      return true;
    }
  }
  return false;
}

/*this PT and input-PT are in conflict, 
  if for some place p we have (out(PT)(p) > 0) 
  and out(input-PT)(p) > 0)
 */
bool PredicateTransformer::conflicts(PredicateTransformer* input) {
  for (i=out.begin(); i!=out.end(); ++i) {
    if ((i->second > 0) && (input->getOut(i->first) > 0)) {
      return true;
    }
  }
  return false;
}

/* check, if this PT separates constraint c from constraint set s, 
   i.e. is there a place p with out(PT)(p) > 0 such that
   c(p) > sc(p) for each constraint sc in s.
*/
bool PredicateTransformer::separates(ConstraintSet* s,Constraint* c){
  for (i=out.begin(); i!=out.end(); ++i) {
    if ((i->second > 0) && (s->smallerAs(c,i->first))) {
      return true;
    }
  }
  return false;
}

// return string-representation of in-/out-function
string PredicateTransformer::asString(stringMap* placeNames) {
  ostringstream os;
  os << "\tIn Function: ";
  for (i=in.begin(); i!=in.end(); ++i) {
    os << '(' << placeNames->getString(i->first) << ',' << i->second << ") ";
  }
  os << "\t\tOut Function: ";
  for (i=out.begin(); i!=out.end(); ++i) {
    os << '(' << placeNames->getString(i->first) << ',' << i->second << ") ";
  }
  os << endl;
  return os.str();
}

void PredicateTransformer::setName(string s)
{
    name = s;
}

void PredicateTransformer::setID(int id)
{
    trans_id = id;
}
