#include <map>
#include <iostream>
#include <sstream>

#include "Constraint.h"
#include "ConstraintSet.h"
#include "stringMap.h"

using namespace std;

#ifndef PREDICATETRANSFORMER_H
#define PREDICATETRANSFORMER_H

/* A Predicate Transformer (PT) is similar to a Transition,
   having an IN- and an OUT-function, but in case of 
   backward reachibility, applying a transition to a constraint
   marking(some_place) = max(marking(some_place) - out(some_place),0)
*/
class PredicateTransformer {
 protected:
  int trans_id;   // transition's id
  string name;     // name of transition
  map<int,int> in; // internal container for in-function
  map<int,int> out; // internal container for out-function
  map<int,int>::iterator i; // used iterating over in- and out-function
 public:
  PredicateTransformer(); // constructor

  // insert/update place-value-pair for in-function
  void setIn(int key,int value); 

  // insert/update place-value-pair for out-function
  void setOut(int key,int value);

  // delete place-value-pair for in-function
  void delIn(int key); 

  // delete place-value-pair for in-function
  void delOut(int key);
  
  // get in-function-value for spec. place
  int getIn(int key);
  // get out-function-value for spec. place
  int getOut(int key);
  
  bool empty();  // is this PT empty?

  PredicateTransformer* copy(); // return new copy of this PT

  // apply this PT to input-Constraint
  Constraint* apply(Constraint* input); 
  
  // methods specifically needed for the backward reachability algorithm

  // does this PT separate spec. constraint from constraint set
  bool separates(ConstraintSet* s,Constraint* c);
  
  // does this PT conflict with input-PT
  bool conflicts(PredicateTransformer* input);

  // is this PT deficient for input-constraint regarding input-PT
  bool deficient(PredicateTransformer* p, Constraint* c);
  
  // return String-Representation of this PT
  string asString(stringMap* placeNames);
  
  //name
  void setName(string s);
  
  // id
  void setID(int id);
  
};
#endif
