#include "PredicateTransformerSet.h"


// constructor
PredicateTransformerSet::PredicateTransformerSet() { count = 0;};

// is this set empty?
bool PredicateTransformerSet::isEmpty() { return pts.empty(); }

// get specified PT
PredicateTransformer* PredicateTransformerSet::get(int id) { return pts[id]; }

// add a PT to this set, and return its assigned ID
int PredicateTransformerSet::add(PredicateTransformer* in) {
  count++;
  pts[count] = in;
  return count;
} 

// return all PTs as with place and transition Names
string PredicateTransformerSet::asString(stringMap* placeNames,stringMap* transNames) {
  ostringstream os;
  for (i=pts.begin(); i!=pts.end(); ++i) { 
    os << "Transition " << transNames->getString(i->first) 
       << ": " << i->second->asString(placeNames);
  } 
  return os.str();
}


// apply all PTs to the input Constraint, and
// return resulting Constraint Set
ConstraintSet* PredicateTransformerSet::apply(Constraint* in) {
  ConstraintSet* res = new ConstraintSet;
  for (i=pts.begin(); i!=pts.end(); ++i) { 
    res->add(i->second->apply(in));
  }
  return res;
}

// apply all PTs with their ID specified in input-ID-set to the input Constraint, and
// return resulting Constraint Set
ConstraintSet* PredicateTransformerSet::apply(intSet* ids, Constraint* in) {
  ConstraintSet* res = new ConstraintSet;
  for (intSet::iterator it=ids->begin(); it!=ids->end(); ++it) { 
    res->add(this->get(ids->get(it))->apply(in));
  }
  return res;
}


// return a new copy of this PT Set
PredicateTransformerSet* PredicateTransformerSet::copy() {
  PredicateTransformerSet* pts_new = new PredicateTransformerSet;
  for (i=pts.begin(); i!=pts.end(); ++i) {
    pts_new->add(i->second->copy());
  }
  return pts_new;
}

// place iterator to first element
PredicateTransformerSet::iterator PredicateTransformerSet::begin(){
  return pts.begin();
}

// place iterator to last element
PredicateTransformerSet::iterator PredicateTransformerSet::end(){
  return pts.end();
}

// get ID of PT which iterator is pointing to
int PredicateTransformerSet::getID(PredicateTransformerSet::iterator iter) {
  return iter->first;
}

// get PT which the iterator is pointing to
PredicateTransformer* PredicateTransformerSet::getPredicateTransformer(PredicateTransformerSet::iterator iter) {
  return iter->second;
}
