#include <sstream>
#include <map>

#include "intSet.h"
#include "stringMap.h"
#include "Constraint.h"
#include "ConstraintSet.h"
#include "PredicateTransformer.h"

using namespace std;

#ifndef PREDICATETRANSFORMERSET_H
#define PREDICATETRANSFORMERSET_H

/* class provides containter for 
   Constraint-Modifiers/Predicate Transformers (PT)/Transitions
*/
class PredicateTransformerSet {
protected:
  int count; // used for assigning unique ID to transitions
  map<int,PredicateTransformer*> pts; // map: id -> Predicate Transformer

  // iterator used to iterate over this set of transitions
  map<int,PredicateTransformer*>::iterator i; 
 public:
  PredicateTransformerSet(); // constructor
  bool isEmpty(); // is this set empty?

  // methods for updating/accessing elements of this set
  int add(PredicateTransformer* in); // add new PT and return its ID
  PredicateTransformer* get(int id); // return specified PT
  
  PredicateTransformerSet* copy(); // create a new copy of this set
  
  // applying this set of PT to input-Constraint
  ConstraintSet* apply(Constraint* in);
  
  // applying PTs which are specified in input-set of IDs to input-Constraint
  ConstraintSet* apply(intSet* ids, Constraint* in);

  // iterator: type and methods,
  // for easily iterating over this set
  typedef map<int,PredicateTransformer*>::iterator iterator;
  iterator begin();
  iterator end();

  // access to PT or its ID via iterator
  int getID(PredicateTransformerSet::iterator iter);
  PredicateTransformer* getPredicateTransformer(PredicateTransformerSet::iterator iter);
  
  // return all PTs as string
  string asString(stringMap* placeNames,stringMap* transNames);
};

#endif
