#include "ConstraintSet_Col.h"


ConstraintSet_Col::ConstraintSet_Col() { count = 0; } // constructor, initialization

int ConstraintSet_Col::size() { return cs.size(); }// return number of Constraints
bool ConstraintSet_Col::empty() { return cs.empty(); } // is this CS empty?

// delete and return any of the constraints
Constraint_Col* ConstraintSet_Col::pop() {
  csit = cs.begin();
  int id = csit->first;
  Constraint_Col* res = csit->second;
  this->del(id);
  return res;
}

// delete constraint given its ID
void ConstraintSet_Col::del(int id) { cs.erase(id); }

// add a given constraint and assign it a unique ID
int ConstraintSet_Col::add(Constraint_Col* in) {
  count++;
  cs[count] = in;
  return count;
}

// add given constraints and assign it a unique ID
int ConstraintSet_Col::add(ConstraintSet_Col* in)
{
  for(iterator it = in->begin(); it != in->end(); it++)
  {
   count++;
   cs[count] = it->second;
  }

  return count;
}

/* move constraints from given ConstraintSet to this set, 
   assigning a unique ID to all constraints,
*/
void ConstraintSet_Col::move(ConstraintSet_Col* in) {
  while(!in->empty()) {
    this->add(in->pop());
  }
  
  delete in;
}

// iterator functions: place iterator at begining/end of this CS
ConstraintSet_Col::iterator ConstraintSet_Col::begin() { return cs.begin(); }
ConstraintSet_Col::iterator ConstraintSet_Col::end() { return cs.end(); }

// return ID of constraint, where iterator is pointing to
int ConstraintSet_Col::getID(ConstraintSet_Col::iterator iter) { return iter->first; }

// return constraint, where iterator is pointing to
Constraint_Col* ConstraintSet_Col::getConstraint(ConstraintSet_Col::iterator iter) { 
  return iter->second; 
}

// does the input-constraint cover this CS, i.e.
// do we have any constraint in CS which is covered by the input-constraint?
bool ConstraintSet_Col::covers(Constraint_Col* in) { 
  for (csit=cs.begin(); csit!=cs.end(); ++csit) {
    if ((*in) <= csit->second) { 
      return true; 
    }
  }
  return false; 
}

// does this CS cover the input-constraint, i.e.
// do we have any constraint in CS covering the input-constraint?
bool ConstraintSet_Col::isCoveredBy(Constraint_Col* in) {
  for (csit=cs.begin(); csit!=cs.end(); ++csit) {
    if (*(csit->second) <= in) { 
      return true; 
    }
  }
  return false; 
}

// return a string-representation of this CS
string ConstraintSet_Col::asString(stringMap* placeNames) { 
  ostringstream os;
  for (csit=cs.begin(); csit!=cs.end(); ++csit) { 
    os << "Constraint with ID " << csit->first << "\t: ";
    os << (csit->second)->asString(placeNames); 
  } 
  return os.str();
}

// return a new copy of this CS
ConstraintSet_Col* ConstraintSet_Col::copy() {
  ConstraintSet_Col* res = new ConstraintSet_Col;
  for (csit=cs.begin(); csit!=cs.end(); ++csit) {
    res->add(csit->second->copy());
  }
  return res;
}

// are all of the constraints in this CS smaller
// as the input-constraint at place, specified by given index?
bool ConstraintSet_Col::smallerAs(Constraint_Col* in, int index)
{
  for (csit=cs.begin(); csit!=cs.end(); ++csit) {
    if (!(csit->second->smallerAs(in,index))) { 
      return false;
    }
  }
  return true;
}

// are all of the constraints in this CS have a shortage of some token from the input-constraint
// at given input-place/index
bool ConstraintSet_Col::Shortage(Constraint_Col* in, int index)
{
  for (csit=cs.begin(); csit!=cs.end(); ++csit) {
    if (!(csit->second->Shortage(in,index))) { 
      return false;
    }
  }
  return true;
}
