#include <sstream>
#include <map>
// #include "../bwra/stringMap.h"
#include "Constraint_Col.h"

using namespace std;

#ifndef CONSTRAINTSET_COL_H
#define CONSTRAINTSET_COL_H

/* this class provides a container for constraints,
   (set of constraints CS)
   and iterator-functions for access to constraints
   within the set.
*/
class ConstraintSet_Col {
  map<int,Constraint_Col*> cs; // the internal container for constraints
  map<int,Constraint_Col*>::iterator csit; // internal used iterator
  int count; // used for assigning unique ID to constraints
 public:
  ConstraintSet_Col(); // constructor

  int size(); // return number of Constraints
  bool empty(); // check if CS is empty
  
  // update/access methods 
  Constraint_Col* pop(); // delete an element of the CS, and return it
  void del(int id); // delete Constraint with given ID
  int add(Constraint_Col* in); // add a Constraint, and return its id

  int add(ConstraintSet_Col* in); // add a Constraint, and return its id

  /* move constraints from given ConstraintSet to this set, 
     assigning a new ID to each contraint */
  void move(ConstraintSet_Col* in);
  
  // iterator: type and methods
  typedef map<int,Constraint_Col*>::iterator iterator;
  iterator begin();
  iterator end();
  
  // access to CS via iterator
  
  // get the ID of the constraint, which iterator is pointing to
  int getID(ConstraintSet_Col::iterator iter);
  // get the constraint, which iterator is pointing to
  Constraint_Col* getConstraint(ConstraintSet_Col::iterator iter);
  
  // does this set cover the input-constraint?
  bool covers(Constraint_Col* in); 
  
  // does the input-constraint cover any of the constraints in this CS?
  bool isCoveredBy(Constraint_Col* in);
  
  // return a new copy of this CS
  ConstraintSet_Col* copy();
  
  // provide string-representation of this CS
  string asString(stringMap* placeNames);
  
  // are all of the constraints in this CS smaller as input-constraint,
  // at given input-place/index?
  bool smallerAs(Constraint_Col* in, int index);

  // are all of the constraints in this CS have a shortage of some token from input-constraint
  // at given input-place/index
  bool Shortage(Constraint_Col* in, int index);
};

#endif

