#include "Constraint_Col.h"

// constructor
Constraint_Col::Constraint_Col()
{
} 

// is this constraint empty?
bool Constraint_Col::empty()
{ 
 return (ctm.getMarkings().size() == 0);
}

// set the marks at place specified by key to given value
void Constraint_Col::set(int key, TOKEN t) { 
  if (t.getCount() > 0)
  {
	ctm.putToken(key, t);
  }
  else
  {
	ctm.removeTokens(key); // setting to 0 is same as deleting key
  }
}

void Constraint_Col::set(int key, vector<TOKEN> vt) { 
  if (vt.size() > 0)
  {
	ctm.setMarking(key, vt);
  }
  else
  {
	ctm.removeTokens(key); // setting to 0 is same as deleting key
  }
}

void Constraint_Col::set(MARKING m)
{
	this->ctm = m;
}

// is this constraint smaller as given constraint at place given by index?
bool Constraint_Col::smallerAs(Constraint_Col* in, int index)
{ 
  MARKING m = in->getMarking();

  if(MARKING::isStrictlyLowerPlace(ctm, m, (IDTYPE)index)) {
    return true; 
  } else { 
    return false;
  }
}

bool Constraint_Col::Shortage(Constraint_Col* in, int index)
{ 
  MARKING m = in->getMarking();

  if(MARKING::ShortagePlace(ctm, m, (IDTYPE)index)) {
    return true; 
  } else { 
    return false;
  }
}

bool Constraint_Col::smallerEqualAs(Constraint_Col* in)
{
  MARKING m = in->getMarking();

  if(MARKING::isLower(ctm, m)) {
    return true; 
  } else { 
    return false;
  }
}

// return a new copy of this constraint
Constraint_Col* Constraint_Col::copy() {
  Constraint_Col* c = new Constraint_Col;
  c->set(ctm);
  for(int j=0; j<ltr.size();j++)
    c->addTransition(ltr[j]);
  return c;
}

// return string-representation of this constraint, given the place names
string Constraint_Col::asString(stringMap* placeNames) {
  ostringstream os;
  if (this->empty()) {
    os << "()\n";
  } else {
    os << ctm;
  }
  return os.str();
}

void Constraint_Col::addTransition(TRANSITION_INFO trans)
{
    ltr.push_back(trans);
}

int Constraint_Col::getTransCount()
{
    return ltr.size();
}

TRANSITION_INFO Constraint_Col::getTransition(int number)
{
    TRANSITION_INFO ti;
    if(ltr.size()<number+1)
	return ti;
	
    return ltr[number];
}

MARKING Constraint_Col::getMarking()
{
	return ctm;
}

// marking simulation

unsigned int Constraint_Col::getTokenNumber(IDTYPE id)
{
	return ctm.getTokenNumber(id);
}

void Constraint_Col::removeTokens(IDTYPE id)
{
	ctm.removeTokens(id);
}

void Constraint_Col::setMarking(IDTYPE id, vector<TOKEN> m)
{
	ctm.setMarking(id, m);
}

void Constraint_Col::putTokens(IDTYPE id, TOKEN tkn)
{
	ctm.putToken(id, tkn);
}

void Constraint_Col::removeTokens(IDTYPE p, TOKEN tkn)
{
	ctm.removeTokens(p, tkn);
}
