#include <map>
#include <iostream>
#include <vector>
#include <sstream>
// #include "../bwra/stringMap.h"
#include "../marking.h"
#include "../bwra/Constraint.h"

using namespace std;

#ifndef CONSTRAINT_COL_H
#define CONSTRAINT_COL_H

/* A constraint is similar to a marking, 
   a mapping from Places P to Integers,
   but a constraint represents a set of markings:
   constraint_m represents all markings m' which
   are greater equal m.
*/
class Constraint_Col
{

 private:

  MARKING ctm;
  vector<TRANSITION_INFO> ltr; // list of fired transitions

 public:

  Constraint_Col(); // constructor

  bool empty();  // is this constraint empty?

  /* update and access methods: */
  
  // set the marks at place specified by key to given value
  void set(int key, TOKEN t); 
  void set(int key, vector<TOKEN> vt); 
  void set(MARKING m);

  // return string-representation of this constraint
  string asString(stringMap* placeNames);
  
  // return a new copy of this constraint
  Constraint_Col* copy();

  /* comparision methods */
  
  // is this constraint smaller as given constraint at place given by index?
  bool smallerAs(Constraint_Col* in, int index);

  // has this constraint a shortage of some token from the given constraint at place given by index?
  bool Shortage(Constraint_Col* in, int index);

  // is this constraint smaller eqaual as given constraint?
  bool smallerEqualAs(Constraint_Col* in);
  
  // shortcuts for method from above
  bool operator<=(Constraint_Col* in) {return this->smallerEqualAs(in); }

  // transitions
  void addTransition(TRANSITION_INFO trans);
  
  int getTransCount();
  
  TRANSITION_INFO getTransition(int number);

  // get marking
  MARKING getMarking();

  // for uncoloured transition
  // get token number at place
  unsigned int getTokenNumber(IDTYPE id);
  // remove all tokens
  void removeTokens(IDTYPE id);
  // set some nunber of tokes
  void setMarking(IDTYPE id, vector<TOKEN> m);
  // put tokens
  void putTokens(IDTYPE id, TOKEN tkn);
  // remove tokens
  void removeTokens(IDTYPE p, TOKEN tkn);
};

#endif
