#include "Cvs_Col.h"
#include<fstream>

#include<cstdio>

extern bool result;

// Constructor
Cvs_Col::Cvs_Col(PredicateTransformerSet_Col* T_in,
	 ConstraintSet_Col* initial_in,
	 ConstraintSet_Col* final_in)
{ 
  // initialize intern variables
  V = new ConstraintSet_Col;
  // initialize intern variables, using input parameters
  initial = initial_in->copy();
  W = final_in->copy();
  T = T_in->copy();
  // placeNames = inPlaceNames;
  // transformerNames = inTransNames;
}

intSet* Cvs_Col::selectAll(PredicateTransformerSet_Col* ps,Constraint_Col* c)
{
	intSet* des = new intSet;
	// store all IDs of input ps in des
  	for(PredicateTransformerSet_Col::iterator it = ps->begin(); it != ps->end(); ++it) {
    		des->add(ps->getID(it));
	}
  
	return des;
}

/* select-function, which is used to improve  the Algorithm, 
   by applying only a subset of available transitions
   to a given constraint.
   Return the IDs of the Transitions, which should
   be applied to given input Constraint c.
*/
intSet* Cvs_Col::select(PredicateTransformerSet_Col* ps,Constraint_Col* c) {
  int id = 0; // store IDs temporarily
  intSet* src = new intSet; // save IDs of ps
  intSet* des = new intSet; // save resulting IDs

  bool found = false;
  bool change = true;

  // store all IDs of input ps in src
  for(PredicateTransformerSet_Col::iterator it = ps->begin(); it != ps->end(); ++it) {
    src->add(ps->getID(it));
  }

  /* search for a separating Transition.
     If found some appropriate Transition T, add T to result-set,
     and delete from src-set.
   */
  for(intSet::iterator it = src->begin(); it != src->end(); ++it) {
    id = src->get(it);
    if (ps->get(id)->separates(initial,c)) {
      found = true;
      src->del(id);
      des->add(id);
      break;
    }
  }
  
  // no separating transition found, therefore return IDs of whole ps
  if (!found) { return src; }
  
  // found separating transition, so go on
  int des_id = 0;
 SCHLEIFE:  while(change) { // as long as there is a transfer from src- to des-set.
   change = false; // reset the flag
   for(intSet::iterator srcit = src->begin(); srcit != src->end(); ++srcit) {
     id = src->get(srcit);
     for(intSet::iterator desit = des->begin(); desit != des->end(); ++desit) {
       des_id = des->get(desit);
       if (ps->get(des_id)->conflicts(ps->get(id)) || 
	   (ps->get(des_id)->deficient(ps->get(id),c))) {
	 des->add(id); // making transfer from src- ...
	 src->del(id); // ... to des-set
	 change = true; // transfer is made,
	 goto SCHLEIFE; // so restart loop after transfer
       }
     }
   }
   return des;
 }
}

/* run the algorithm, i.e. check if some given final
   marking is reachable/coverable from some given
   inital marking, using backward search.
 */
bool Cvs_Col::run(char * tracefile) {
  // private variables
  Constraint_Col* temp = new Constraint_Col;
  intSet* ids = new intSet;
    
  /* at least 1 initial marking is needed,
     if not available, exit with error
   */
  if (initial->empty()) {
    cout << "Error: Initial Constraint Set is empty.\n";
    std::exit(1);
  }

  ofstream of(tracefile);
  
  /* visit constraints in working list one-by-one,
     until we have some appropriate Constraint,
     which (set-)includes a initial marking.
  */
 SCHLEIFE: while (!W->empty()) {

   // timeout
   if(timeout != 0)
   {
	if(get_time() > timeout)
	{
		cout << "Exited by Timeout" << endl;
		break;
	}
   }

   temp = W->pop(); // remove constraint to visit from working list

   // if an equal or better constraint is already visited,
   // skip current constraint, and take the next one
   if (V->isCoveredBy(temp)) { goto SCHLEIFE; } 
   
   if (initial->covers(temp)) { // an appropriate constraint is found
     V->add(temp);
     if (quiet) { // print only result
       cout << "true\n";
     } else { // print all visited constraints and elapsed time
       // cout << "Visited Constraints:\n" << V->asString(placeNames); 
       cout << "The Final Marking(s) are coverable.\n";
     }

     // write transitions in the file
     int i;
     for(i=temp->getTransCount()-1; i>=0; i--)
     {
	BINDING b = temp->getTransition(i).binding;
	of << temp->getTransition(i).name.c_str() << endl << temp->getTransition(i).ID << endl  << "% " << b << endl;
     }

		
    of.close();	
     
     return true;
   } 
   /* use the select-funktion to determine
      the appropriate transitions to apply to
      the current constraint.
      Then apply these transitions to the
      current constraint, and add the 
      resulting constraints to the working list W.
   */
   //   W->add(T->apply(this->select(T,temp),temp)); 
   W->move(T->apply(this->select(T,temp),temp)); 
   
   V->add(temp); // add current constraint to the list of visited.
 }
  
  /* W is empty, hence return "False",
     as no appropriate Constraint was generated.
  */
  
  if (quiet) { // print only result
    cout << "false\n";
  } else {// print all visited constraints and elapsed time
    // cout << "Visited Constraints:\n" << V->asString(placeNames); 
    cout << "The Final Marking(s) are not coverable.\n";
  }

/*  if(result)
  {
    ofstream of;
    of.open("result", ofstream::out | ofstream::app);
    of << "false\n";
    of << "time used " << this->getTime() << " sec." << endl;
  }
*/
  
  return false;
}
