#include "PredicateTransformerSet_Col.h"


// constructor
PredicateTransformerSet_Col::PredicateTransformerSet_Col() { count = 0;};

// is this set empty?
bool PredicateTransformerSet_Col::isEmpty() { return pts.empty(); }

// get specified PT
PredicateTransformer_Col* PredicateTransformerSet_Col::get(int id) { return pts[id]; }

// add a PT to this set, and return its assigned ID
int PredicateTransformerSet_Col::add(PredicateTransformer_Col* in) {
  count++;
  pts[count] = in;
  return count;
} 

// return all PTs as with place and transition Names
string PredicateTransformerSet_Col::asString(stringMap* placeNames,stringMap* transNames) {
  ostringstream os;
  for (i=pts.begin(); i!=pts.end(); ++i) { 
    os << "Transition " << transNames->getString(i->first) 
       << ": " << i->second->asString(placeNames);
  }
  return os.str();
}


// apply all PTs to the input Constraint, and
// return resulting Constraint Set
ConstraintSet_Col* PredicateTransformerSet_Col::apply(Constraint_Col* in) {
  ConstraintSet_Col* res = new ConstraintSet_Col;
  for (i=pts.begin(); i!=pts.end(); ++i) { 
    res->add(i->second->apply(in));
  } 
  return res;
}

// apply all PTs with their ID specified in input-ID-set to the input Constraint, and
// return resulting Constraint Set
ConstraintSet_Col* PredicateTransformerSet_Col::apply(intSet* ids, Constraint_Col* in) {
  ConstraintSet_Col* res = new ConstraintSet_Col;
  for (intSet::iterator it=ids->begin(); it!=ids->end(); ++it) { 
    res->add(this->get(ids->get(it))->apply(in));
  } 
  return res;
}


// return a new copy of this PT Set
PredicateTransformerSet_Col* PredicateTransformerSet_Col::copy() {
  PredicateTransformerSet_Col* pts_new = new PredicateTransformerSet_Col;
  for (i=pts.begin(); i!=pts.end(); ++i) {
    pts_new->add(i->second->copy());
  }
  return pts_new;
}

// place iterator to first element
PredicateTransformerSet_Col::iterator PredicateTransformerSet_Col::begin(){
  return pts.begin();
}

// place iterator to last element
PredicateTransformerSet_Col::iterator PredicateTransformerSet_Col::end(){
  return pts.end();
}

// get ID of PT which iterator is pointing to
int PredicateTransformerSet_Col::getID(PredicateTransformerSet_Col::iterator iter) {
  return iter->first;
}

// get PT which the iterator is pointing to
PredicateTransformer_Col* PredicateTransformerSet_Col::getPredicateTransformer(PredicateTransformerSet_Col::iterator iter) {
  return iter->second;
}
