#include "PredicateTransformer_Col.h"


PredicateTransformer_Col::PredicateTransformer_Col() {} // constructor

void PredicateTransformer_Col::setTransition(IDTYPE idt)
{
	trans_id = idt;
}

void PredicateTransformer_Col::setPetriNet(PTNET * pn)
{
	this->pn = pn;
}

void PredicateTransformer_Col::setEngine(ALGORITHM * engine)
{
	this->engine = engine;
}

// return a new copy of this PT
PredicateTransformer_Col* PredicateTransformer_Col::copy() {
  PredicateTransformer_Col* pt = new PredicateTransformer_Col;

  pt->setTransition(trans_id);
  pt->setPetriNet(pn);
  pt->setEngine(engine);

  return pt;
}

/* generate new constraint by applying this Transition to input-constraint,
   for all places p of out-function: input(p) = max(input(p) - out(p),0)
   for all places p of in-function: input(p) = input(p) + in(p) */
ConstraintSet_Col* PredicateTransformer_Col::apply(Constraint_Col* input){

  ConstraintSet_Col* result = new ConstraintSet_Col();

  // get table
  TABLE * table;
  bool col = engine->run(trans_id, &table);

  // coloured transition
  if(col)
  {
	for(TABLE::iterator it = table->begin(); it != table->end(); it++)
	{
		// check guard
		if(!(it->guard_result))
		 continue;

		Constraint_Col* res = input->copy(); // working on a new copy
	
		// set the name
		TRANSITION_INFO ti;
		ti.name = pn->getTransition(trans_id).getLabel();
		ti.ID = pn->getTransition(trans_id).getID();
		ti.binding = it->binding;

		res->addTransition(ti);

		// apply out-function
		vector<TCONNECTION> postset = pn->getPostset(trans_id);
		// table iterator
		vector<list<EXPRESSION> >::iterator it_postset_table = it->postset.begin();
		// iterate through postset
		for (vector<TCONNECTION>::iterator i=postset.begin(); i!=postset.end(); i++)
		{
			TOKEN t;
			t.setCount(i->weight);
			t.setAttributes(*it_postset_table);

			res->removeTokens(i->ID, t);

			// next table value
			it_postset_table++;
		}

		// apply in-function
		vector<TCONNECTION> preset = pn->getPreset(trans_id);
		// table iterator
		vector<list<EXPRESSION> >::iterator it_preset_table = it->preset.begin();
		// iterate through preset
		for (vector<TCONNECTION>::iterator i=preset.begin(); i!=preset.end(); i++)
		{
			TOKEN t;
			t.setCount(i->weight);
			t.setAttributes(*it_preset_table);

			res->putTokens(i->ID,t);

			// next table value
			it_preset_table++;
		}

		result->add(res);
	}
  }
  // not coloured transition
  else
  {
	Constraint_Col* res = input->copy(); // working on a new copy
	
	// set the name
	TRANSITION_INFO ti;
	ti.name = pn->getTransition(trans_id).getLabel();
	ti.ID = pn->getTransition(trans_id).getID();

	res->addTransition(ti);
	
	// apply out-function
	vector<TCONNECTION> postset = pn->getPostset(trans_id);
	for (vector<TCONNECTION>::iterator i=postset.begin(); i!=postset.end(); i++)
	{
		int temp = 0;
		temp = res->getTokenNumber(i->ID) - i->weight;
		if (temp <= 0)
		{
		 res->removeTokens(i->ID); // 0-values not needed
		}
		else
		{
		 TOKEN t;
		 t.setCount(temp);
		 vector<TOKEN> vt;
		 vt.push_back(t);
		 res->setMarking(i->ID,vt); // save value > 0
		}
	}
	
	// apply in-function
	vector<TCONNECTION> preset = pn->getPreset(trans_id);
	for (vector<TCONNECTION>::iterator i=preset.begin(); i!=preset.end(); i++)
	{
	 TOKEN t;
	 t.setCount(i->weight);
	 res->putTokens(i->ID,t);
	}

	result->add(res);
  }

  return result;
}

/* check, if this PT is deficient for place p at given constraint c
   
	In Uncoloured case it was: (if c(p) < out(PT)(p)) and in(input-PT)(p) > 0 

	In Coloured case: the condition c(p) < out(PT)(p) should be calculated for all colours
	which is not so probably to obtain.
	We just ignore this condition.
*/
bool PredicateTransformer_Col::deficient(PredicateTransformer_Col* p, Constraint_Col* c)
{
	// this postset 
	vector<TCONNECTION> vc_postset = pn->getPostset(trans_id);

	// preset of p
	vector<TCONNECTION> vc_preset = pn->getPostset(p->getTransition());

	for(vector<TCONNECTION>::iterator it1 = vc_postset.begin(); it1 != vc_postset.end(); it1++)
	{
		// is deficient
		// if(it1->weight > c->getTokenNumber(it1->ID))
		// {
			// check transition
			for(vector<TCONNECTION>::iterator it2 = vc_preset.begin(); it2 != vc_preset.end(); it2++)
			{
				if(it1->ID == it2->ID)
					return true;
			}
		// }
	}

	return false;
}

/*this PT and input-PT are in conflict, 
  if for some place p we have (out(PT)(p) > 0) 
  and out(input-PT)(p) > 0)
 */
bool PredicateTransformer_Col::conflicts(PredicateTransformer_Col* input)
{
	// this postset 
	vector<TCONNECTION> vc1 = pn->getPostset(trans_id);

	// input postset
	vector<TCONNECTION> vc2 = pn->getPostset(input->getTransition());

	for(vector<TCONNECTION>::iterator it1 = vc1.begin(); it1 != vc1.end(); it1++)
	{
		for(vector<TCONNECTION>::iterator it2 = vc2.begin(); it2 != vc2.end(); it2++)
		{
			if(it1->ID == it2->ID)
				return true;
		}
	}

	return false;
}

/* check, if this PT separates constraint c from constraint set s, 
   i.e. is there a place p with out(PT)(p) > 0 such that
   c(p) > sc(p) for each constraint sc in s.
*/
bool PredicateTransformer_Col::separates(ConstraintSet_Col* s,Constraint_Col* c)
{
	// this postset 
	vector<TCONNECTION> vc1 = pn->getPostset(trans_id);

	for(vector<TCONNECTION>::iterator it1 = vc1.begin(); it1 != vc1.end(); it1++)
	{
		if ((it1->weight > 0) && (s->Shortage(c,it1->ID)))
			return true;
	}
	
	return false;
}

// return string-representation of in-/out-function
string PredicateTransformer_Col::asString(stringMap* placeNames) {
  ostringstream os;
  os << "ID: " << trans_id << ", Label: " << pn->getTransition(trans_id).getLabel() << endl;
  os << endl;
  return os.str();
}

IDTYPE PredicateTransformer_Col::getTransition()
{
	return trans_id;
}
