#include <map>
#include <iostream>
#include <sstream>

#include "Constraint_Col.h"
#include "ConstraintSet_Col.h"
// #include "../bwra/stringMap.h"

#include "algorithm.h"

#include "ptnet.h"
#include "binding.h"

using namespace std;

#ifndef PREDICATETRANSFORMER_COL_H
#define PREDICATETRANSFORMER_COL_H

/* A Predicate Transformer (PT) is similar to a Transition,
   having an IN- and an OUT-function, but in case of 
   backward reachibility, applying a transition to a constraint
   marking(some_place) = max(marking(some_place) - out(some_place),0)
*/
class PredicateTransformer_Col
{

 protected:

  IDTYPE trans_id; // transition
  PTNET * pn; // Petri net
  ALGORITHM * engine;

 public:

  PredicateTransformer_Col(); // constructor

  PredicateTransformer_Col* copy(); // return new copy of this PT

  // apply this PT to input-Constraint
  ConstraintSet_Col* apply(Constraint_Col* input); 

  // return String-Representation of this PT
  string asString(stringMap* placeNames);

  void setTransition(IDTYPE idt);

  void setPetriNet(PTNET * pn);

  void setEngine(ALGORITHM * engine);

  IDTYPE getTransition();

  // select help methods

  // does this PT separate spec. constraint from constraint set
  bool separates(ConstraintSet_Col* s,Constraint_Col* c);

  // does this PT conflict with input-PT
  bool conflicts(PredicateTransformer_Col* input);

  // is this PT deficient for input-constraint regarding input-PT
  bool deficient(PredicateTransformer_Col* p, Constraint_Col* c);

};
#endif
