/***************************************************************************
 *   Copyright (C) 2006 by root   *
 *   root@Vitali   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "costgraph.h"

namespace algorithms {

namespace matchings
{
namespace data_str_searchplan
{

COSTGRAPH::COSTGRAPH()
 : GRAPH()
{
}


COSTGRAPH::~COSTGRAPH()
{
}

void COSTGRAPH::addVertex( VERTEX v ){
	this->vertices[v.getID()] = v;
// 	this->cost[v.getID()] = cost;
      	if(v.getID() > lastUsedID) lastUsedID = v.getID();
}

void COSTGRAPH::addEdge( EDGE e, IDTYPE v1, IDTYPE v2, float cost ){

	if(this->vertices.count(v1) == 0 || this->vertices.count(v2) == 0){
		cout << "Can not add Edge" << endl;
		return;
	}
	this->edges[e.getID()]=e;
	if(e.getID() > lastUsedID) lastUsedID = e.getID();

	this->cost[e.getID()] = cost;
	this->incomingEdges[v2].push_back(e.getID());
	this->outgoingEdges[v1].push_back(e.getID());
	this->targetVertex[e.getID()] = v2;
	this->sourceVertex[e.getID()] = v1;
	
}

void COSTGRAPH::clean(){
	
	GRAPH::clean();
	incomingEdges.clear();
    	outgoingEdges.clear();
    	sourceVertex.clear();
    	targetVertex.clear();
    	cost.clear();
}

IDTYPE COSTGRAPH::getSource(IDTYPE i){

	return this->sourceVertex[i];
}

IDTYPE COSTGRAPH::getTarget(IDTYPE i){

	return this->targetVertex[i];
}


list<IDTYPE> COSTGRAPH::getIncoming(IDTYPE i){

	return this->incomingEdges[i];
}

list<IDTYPE> COSTGRAPH::getOutgoing(IDTYPE i){

	return this->outgoingEdges[i];
}

list<IDTYPE> COSTGRAPH::getOrderedOutgoing(IDTYPE i){

	list<IDTYPE> list1 = this->outgoingEdges[i];
	if(list1.empty()) return list1;
	multimap<float, IDTYPE> mapList;
	list<IDTYPE> orderedList;
	for(list<IDTYPE>::iterator list1_iter = list1.begin(); list1_iter != list1.end(); list1_iter++){

// 		mapList[cost[*list1_iter]] = *list1_iter;
		mapList.insert(pair<float, IDTYPE>(cost[*list1_iter], *list1_iter));
	}
// 	cout << "getOutgoing ";
	for(multimap<float, IDTYPE>::iterator map_iter = mapList.begin(); map_iter!=mapList.end(); map_iter++){

		orderedList.push_back(map_iter->second);
// 		cout << map_iter->first<< "("<<map_iter->second<<"), ";
	}
// 	cout << endl;
	return orderedList;
}

float COSTGRAPH::getCost(IDTYPE i){

	return this->cost[i];
}
}
}
}
