/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "counter_example_reader.h"
#include <string.h>

namespace algorithms {

namespace readers {

COUNTER_EXAMPLE_READER::COUNTER_EXAMPLE_READER()
{
}


COUNTER_EXAMPLE_READER::~COUNTER_EXAMPLE_READER()
{
}

void COUNTER_EXAMPLE_READER::initialize(char *filename)
{
	fname = (string) filename;
}
    
COUNTER_EXAMPLE COUNTER_EXAMPLE_READER::run()
{
 fstream trans;
 trans.open(fname.c_str());

 if(!trans.is_open())
 {
  cout << "\nError while reading counter-example file\n";
  exit(1);
 }
 
 COUNTER_EXAMPLE ce;
 
 bool odd = true;

 while(!trans.eof())
 {
    char ch[256];

    trans.getline(ch, 256);

    if(strlen(ch)==0)
	continue;

    // try to read the binding
    if(ch[0]=='%')
    {
	BINDING b;
	char * ch_ = new char[256];
	strcpy(ch_, string(ch).substr(2, 256).c_str());
	if(readBinding(&b, ch_))
	{
	 int ltr_size = ce.bindings.size();
	 ce.bindings.resize(ltr_size+1);
	 ce.bindings[ltr_size] = b;
	}
	continue;
    }

    if(odd)
    {
	int ltr_size = ce.trans_strings.size();
    
	ce.trans_strings.resize(ltr_size+1);
	string s(ch);
	ce.trans_strings[ltr_size]= s;

	odd = false;
    }
    else
    {
	int ltr_size = ce.trans_ids.size();
			
	ce.trans_ids.resize(ltr_size+1);
	ce.trans_ids[ltr_size]=atoi(ch);

	odd = true;
    }

 } 
 
 return ce;
 
}

bool COUNTER_EXAMPLE_READER::readBinding(BINDING * b, char * ch)
{
	int index = 0;
	
	string current_name = "";
	string current_value = "";
	T * current_type = NULL;

	bool reading_var = true;

	while(index<strlen(ch))
	{
		// variable
		if(reading_var)
		{
			if(ch[index] != '(')
			{
				current_name += ch[index];
				index++;
				continue;
			}
			else
			{
				index++;
				if(ch[index] == 'i')
				 current_type = new T(INTEGER);
				if(ch[index] == 'b')
				 current_type = new T(BOOL);
				if(ch[index] == 's')
				 current_type = new T(STRING);

				// ):_
				for(int i=0; i<6; i++)
				{
					if(ch[index] != '\0')	
					 index++;
					else
					 return false;
				}

				reading_var = false;
				continue;
			}
		}
		// value
		else
		{
			if(ch[index] != ',' && index<strlen(ch))
			{
				current_value += ch[index];
				index++;
				continue;
			}
			else
			{
				if(current_name!="" && current_value!="" && current_type!=NULL)
				{
					ExpressionElement e;
					e.is_abstract = true;

					if(*current_type == INTEGER)
					{
						e.type = NUMBER;
						e.int_evaluated = atoi(current_value.c_str());
					}
					else if(*current_type == STRING)
					{
						e.type = _STR_;
						e.str_evaluated = current_value;
					}
					else
					{
						cout << "can not read BOOLEAN binding (not yet implemented)" << endl;
					}

					b->setVariable(current_name, e);					
				}

				current_name = "";
				current_value = "";
				current_type = NULL;

				reading_var = true;
				index += 2;
			}
		}
	}

	// insert last element
	if(current_name!="" && current_value!="" && current_type!=NULL)
	{
		ExpressionElement e;
		e.is_abstract = true;

		if(*current_type == INTEGER)
		{
			e.type = NUMBER;
			e.int_evaluated = atoi(current_value.c_str());
		}
		else if(*current_type == STRING)
		{
			e.type = _STR_;
			e.str_evaluated = current_value;
		}
		else
		{
			cout << "can not read BOOLEAN binding (not yet implemented)" << endl;
		}

		b->setVariable(current_name, e);					
	}

	return true;
}

//////// namespaces ////////

}

}


