/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "cov_graph.h"

namespace algorithms {

namespace ptnet {

namespace coverability {

namespace data_str_coverability {

COV_GRAPH::COV_GRAPH()
{
  node_counter=0;
  status = READY;
  unfold_trans_id=0;
}

COV_GRAPH::~COV_GRAPH()
{
}

void COV_GRAPH::clear()
{
     m_nodeMapping.clear();
     m_markingMapping.clear();
     m_succ.clear();
     m_pred.clear();
     status=READY;
} 

long COV_GRAPH::getSize()
{
    return m_markingMapping.size();
}

IDTYPE COV_GRAPH::getID(Marking & m)
{
  return m_nodeMapping[m]; 
}
   
IDTYPE COV_GRAPH::addNode(Marking & m)
{
    // optimization
    // assert(!containsNode(m));    

    node_counter++;
    IDTYPE newID = node_counter;
    
    m_nodeMapping[m] = newID;
    m_markingMapping[newID] = m;

    return newID;
}
  

void COV_GRAPH::addEdge(Marking & m1, Marking & m2, TRANSITION & transition)
{
    // optimization
    // assert(containsNode(m1));    
    // assert(containsNode(m2));     

    IDTYPE id1 = getID(m1);
    IDTYPE id2 = getID(m2);
    
    m_succ[id1][((GRAPH_ELEMENT)transition).getID()] = id2; 
    m_pred[id2][((GRAPH_ELEMENT)transition).getID()] = id1; 
}

void COV_GRAPH::addEdge(IDTYPE id1, IDTYPE id2, IDTYPE transition_id)
{    
    m_succ[id1][transition_id] = id2;
    m_pred[id2][transition_id] = id1;
} 
 
bool COV_GRAPH::containsNode(Marking& marking, unsigned * node)
{
  *node=0;
  map<Marking,IDTYPE>::iterator it = m_nodeMapping.find(marking);
  if(it != m_nodeMapping.end())
  {
   *node = it->second;
   return true;
  }
  else
   return false;  
}
  
bool COV_GRAPH::getNodeWithSmallerMarking(Marking & node, Marking & marking, Marking * smallerNode)
{
//    if (m_nodeMapping.find(node) == m_nodeMapping.end())
//      throw invalid_argument("CovGraph::containsNodeWithSmallerMarking() - No such node");
      
    map<IDTYPE,bool> visited;
    for(map<IDTYPE,Marking>::iterator i = m_markingMapping.begin(); i != m_markingMapping.end(); i++)
      visited[i->first] = false;
      
    IDTYPE actualNode = m_nodeMapping[node];
    stack<IDTYPE> work;
    
    bool finished = false;
    
    while (! finished)
    {
      if (isStrictlyLower(m_markingMapping[actualNode], marking))
      {
        *smallerNode = m_markingMapping[actualNode];
        return true;
      }
      for(map<IDTYPE,IDTYPE>::iterator i = m_pred[actualNode].begin(); i != m_pred[actualNode].end(); i++)
      {
        if (! visited[i->second])
        {
          work.push(i->second);
          visited[i->second] = true;
        }
      }
      if (! work.empty())
      {
        actualNode = work.top();
        work.pop();
      }
      else
        finished = true;       
    }
    
    return false;
}


void COV_GRAPH::changeIDs(map<IDTYPE, IDTYPE> eqrel, map<IDTYPE, IDTYPE> transmap)
{
   // nodes
   map<Marking,IDTYPE> m_nodeMapping_;
   map<IDTYPE,Marking> m_markingMapping_;
   
   map<IDTYPE,IDTYPE> nodes;
      
   for(map<IDTYPE,Marking>::iterator it = m_markingMapping.begin();
   	it!=m_markingMapping.end(); it++)
   {
   	map<IDTYPE,unsigned> Marking_;
	
	for(map<IDTYPE,unsigned>::iterator im = it->second.begin();
		im!=it->second.end(); im++)
	{
		IDTYPE id = eqrel[im->first];
		if(id!=0)
		{
			unsigned a = 0;
			Marking::iterator itm = Marking_.find(id);
			if(itm!=Marking_.end())
			 a = itm->second;
			Marking_[id]=plus(a, im->second);
		}	
		else
		{
			unsigned a = 0;
			Marking::iterator itm = Marking_.find(im->first);
			if(itm!=Marking_.end())
			 a = itm->second;
			Marking_[im->first]=plus(a, im->second);
		}	
	}
	
	map<Marking, IDTYPE>::iterator itnM = m_nodeMapping_.find(Marking_);
	if(itnM==m_nodeMapping_.end())
	{
		m_nodeMapping_[Marking_]=it->first;
		nodes[it->first]=it->first;
		m_markingMapping_[it->first] = Marking_;
	}
	else
		nodes[it->first]=itnM->second;
   }
   
   m_markingMapping=m_markingMapping_;
   m_nodeMapping=m_nodeMapping_;
   
   // edges
   map<IDTYPE,map<IDTYPE,IDTYPE> > m_succ_;
   map<IDTYPE,map<IDTYPE,IDTYPE> > m_pred_;
   
   // succ
   for(map<IDTYPE,map<IDTYPE,IDTYPE> >::iterator itS = m_succ.begin();
   	itS!=m_succ.end(); itS++)
   {
   	IDTYPE place1;
	IDTYPE node1 = nodes[itS->first];
   	if(node1!=0)
		place1 = node1;
	else
		place1 = itS->first;
		
   	map<IDTYPE, IDTYPE> m;
	for(map<IDTYPE,IDTYPE>::iterator itS1 = itS->second.begin();
		itS1!=itS->second.end(); itS1++)
	{
		IDTYPE place2, trans;
		IDTYPE node2 = nodes[itS1->second];
		if(node2!=0)
			place2 = node2;
		else
			place2 = itS1->second;
		
		IDTYPE trans_id = transmap[itS1->first];
		if(trans_id!=0)
			trans = trans_id;
		else
			trans = itS1->first;	
		
		m[trans]=place2;
	}
	
	m_succ_[place1]=m;
   }
   
   // pred
   for(map<IDTYPE,map<IDTYPE,IDTYPE> >::iterator itS = m_pred.begin();
   	itS!=m_pred.end(); itS++)
   {
   	IDTYPE place1;
	IDTYPE node1 = nodes[itS->first];
   	if(node1!=0)
		place1 = node1;
	else
		place1 = itS->first;
		
   	map<IDTYPE, IDTYPE> m;
	for(map<IDTYPE,IDTYPE>::iterator itS1 = itS->second.begin();
		itS1!=itS->second.end(); itS1++)
	{
		IDTYPE place2, trans;
		IDTYPE node2 = nodes[itS1->second];
		if(node2!=0)
			place2 = node2;
		else
			place2 = itS1->second;
		
		IDTYPE trans_id = transmap[itS1->first];
		if(trans_id!=0)
			trans = trans_id;
		else
			trans = itS1->first;
		
		m[trans]=place2;
	}
	
	m_pred_[place1]=m;
   }
  
   m_pred = m_pred_;
   m_succ = m_succ_;

   // trick
   // if the graph was ready and we have made only unfolding step,
   // then we add the transition and the graph remains ready
   DATA * d = storage.getData("coverability");
   storage.setData("coverability", NULL);
   // unfolding
   if(d!=NULL)
   { 
     if(status==READY)
     {
      TRANSITION_INFO * ti = (TRANSITION_INFO*)d;
      unfold_trans_id = transmap[ti->transition];    
     } 
     else
     {
      unfold_trans_id=0;
     }
   }
   // folding
   else
   {
    unfold_trans_id=0;
    status=SMTH;
   } 
}
  
  unsigned COV_GRAPH::plus(unsigned a, unsigned b)
  {  
      if((a==OMEGA) || (b==OMEGA))
       return OMEGA;
      else
       return a+b;
  }
  
  bool COV_GRAPH::coverNode(Marking & m, IDTYPE * id)
  {
    for(map<IDTYPE,Marking>::iterator i = m_markingMapping.begin();
        i != m_markingMapping.end();
        i++)
    {
      if (isLower(m,i->second) )
      {
       *id = i->first;	
       return true;
      }
    }    
    
    return false;
  }

  bool COV_GRAPH::coverNodes(list<Marking> lmu, IDTYPE * id)
  {
    for(list<Marking>::iterator it = lmu.begin(); it != lmu.end(); it++)
    {
      if(coverNode(*it, id))
       return true;
    }    
    
    return false;
  }

  list<IDTYPE> COV_GRAPH::getPath(IDTYPE search_id)
  {
	list<IDTYPE> result;

	// work list of node ids and path to this node
	list<pair<IDTYPE, list<IDTYPE> > > workList;

	// visited nodes
	list<IDTYPE> visited_nodes;

	// initial marking is number one
	list<IDTYPE> emptyList;
	workList.push_back(pair<IDTYPE, list<IDTYPE> >(1, emptyList));

	// bread search
	while(workList.size() != 0)
	{
		// found
		if(workList.front().first == search_id)
		 return workList.front().second;

		// call possible transitions
		map<IDTYPE,IDTYPE> mt = m_succ[workList.front().first];
		for(map<IDTYPE,IDTYPE>::iterator its = mt.begin(); its!= mt.end(); its++)
		{
			// check the loop
			bool found = false;
			for(list<IDTYPE>::iterator it = visited_nodes.begin(); it != visited_nodes.end(); it++)
			{
				if(*it == its->second)
				{
					found = true;
					break;
				}
			}

			// not found
			if(!found)
			{
				list<IDTYPE> l = workList.front().second;
				l.push_back(its->first);
				workList.push_back(pair<IDTYPE, list<IDTYPE> >(its->second, l));
			}
		}

		// delete current node
		visited_nodes.push_back(workList.front().first);
		workList.pop_front();
	}

	return result;
  }

  ostream& operator<<(ostream& ost, COV_GRAPH& th)
  {
	ost << endl << "coverability graph" << endl;

	for(map<IDTYPE,Marking>::iterator i = th.m_markingMapping.begin();
        i != th.m_markingMapping.end();
        i++)
	{
		ost << "marking" << endl;
		for(Marking::iterator it = i->second.begin(); it != i->second.end(); it++)
			ost << it->first << " " << it->second << endl;

	}

	ost << "------------------" << endl;

	return ost;
  }

};
};
};
};





