/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef COV_GRAPH_H
#define COV_GRAPH_H

#include "globals.h"

#include "marking.h"
#include "token.h"
#include "transition.h"
#include "validator.h"
#include "persistable.h"

#include "algorithm_storage.h"

#include <map>
#include <list>
#include <stdexcept>
#include <vector>

namespace algorithms {
namespace ptnet {
namespace coverability {
namespace data_str_coverability {

// SMTH is not empty and schould be checked for the possible extensions
enum STATUS {READY, SMTH};

class COV_GRAPH {

public:

    COV_GRAPH();

    ~COV_GRAPH();

    IDTYPE addNode(Marking & m);
    void addEdge(Marking & m1, Marking & m2, TRANSITION & t);
    void addEdge(IDTYPE id1, IDTYPE id2, IDTYPE transition_id);
    
    void clear();
    
    IDTYPE getID(Marking & m);
    
    bool containsNode(Marking & marking, unsigned * node);  
    bool getNodeWithSmallerMarking(Marking & node, Marking & marking, Marking * smallerNode);
    
    void changeIDs(map<IDTYPE, IDTYPE> eqrel, map<IDTYPE, IDTYPE> transmap);

    long getSize();  
    bool coverNode(Marking & m, IDTYPE * id);
    bool coverNodes(list<Marking> lmu, IDTYPE * id);
        
    // status of the graph
    STATUS status;
    
    // trick, to optimize the calculation after unfolding step
    IDTYPE unfold_trans_id;
    
    IDTYPE node_counter;

    map<Marking,IDTYPE> m_nodeMapping;
    map<IDTYPE,Marking> m_markingMapping;

    /**
     * 1. IDTYPE: Node of the coverability graph
     * 2. IDTYPE: ID of the corresponding transition in the PTNet
     * 3. IDTYPE: Successor node in coverability graph
     */
    map<IDTYPE,map<IDTYPE,IDTYPE> > m_succ;
    map<IDTYPE,map<IDTYPE,IDTYPE> > m_pred;
      
    unsigned plus(unsigned a, unsigned b);

    list<IDTYPE> getPath(IDTYPE search_id);

    friend ostream& operator<<(ostream& ost, COV_GRAPH& th);
};

class TRANSITION_INFO: public DATA
{
 public: 
  IDTYPE transition;
};

};

};

};

};

using namespace algorithms;
using namespace ptnet;
using namespace coverability;
using namespace data_str_coverability;

#endif
