/***************************************************************************
 *   Copyright (C) 2006 by root   *
 *   root@Vitali   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "cov_graph_coloured.h"

namespace algorithms {
namespace ptnet {
namespace coverability {
namespace data_str_coverability {

COV_GRAPH_COLOURED::COV_GRAPH_COLOURED()
{
	node_counter = 0;
	status = READY;
	unfold_trans_id=0;
}


COV_GRAPH_COLOURED::~COV_GRAPH_COLOURED()
{
}

void COV_GRAPH_COLOURED::clear()
{
//     m_nodeMapping.clear();
     m_markingMapping.clear();
     m_succ.clear();
     m_pred.clear();
     m_transitions.clear();
     status=READY;
     node_counter = 0;
     trans_counter = 0;
} 

long COV_GRAPH_COLOURED::getSize()
{
    return m_markingMapping.size();
}

IDTYPE COV_GRAPH_COLOURED::getID(MARKING & m)
{
  for(map<IDTYPE, MARKING>::iterator it = m_markingMapping.begin(); it != m_markingMapping.end(); it++)
  {
	if(it->second == m)
	 return it->first;
  }

  cout << "\nerror: COV_GRAPH_COLOURED - no such marking\n";
  exit(1);
}
   
IDTYPE COV_GRAPH_COLOURED::addNode(MARKING & m)
{
    // optimization
    // assert(!containsNode(m));    

    node_counter++;
    IDTYPE newID = node_counter;

//    m_nodeMapping[m] = newID;
    m_markingMapping[newID] = m;

    return newID;
}


void COV_GRAPH_COLOURED::addEdge(MARKING & m1, MARKING & m2, TRANSITION & transition, BINDING & binding)
{
    IDTYPE id1 = getID(m1);
    IDTYPE id2 = getID(m2);

	// optimization
	if(m_succ.find(id1) != m_succ.end())
	{
		map<IDTYPE,IDTYPE> temp_map = m_succ[id1];
		for(map<IDTYPE,IDTYPE>::iterator it = temp_map.begin(); it != temp_map.end(); it++)
		{
			if(it->second == id2)
			 return;
		}
	}

    trans_counter++;
    m_transitions[trans_counter] = pair<IDTYPE, BINDING>(((GRAPH_ELEMENT)transition).getID(), binding);

    m_succ[id1][trans_counter] = id2; 
    m_pred[id2][trans_counter] = id1; 
}

void COV_GRAPH_COLOURED::addEdge(IDTYPE id1, IDTYPE id2, TRANSITION & transition, BINDING & binding)
{
	// optimization
	if(m_succ.find(id1) != m_succ.end())
	{
		map<IDTYPE,IDTYPE> temp_map = m_succ[id1];
		for(map<IDTYPE,IDTYPE>::iterator it = temp_map.begin(); it != temp_map.end(); it++)
		{
			if(it->second == id2)
			 return;
		}
	}

    trans_counter++;
    m_transitions[trans_counter] = pair<IDTYPE, BINDING>(((GRAPH_ELEMENT)transition).getID(), binding);

    m_succ[id1][trans_counter] = id2; 
    m_pred[id2][trans_counter] = id1; 
} 
 
bool COV_GRAPH_COLOURED::containsNode(MARKING& marking, unsigned * node)
{
  *node=0;

  for(map<IDTYPE, MARKING>::iterator it = m_markingMapping.begin(); it != m_markingMapping.end(); it++)
  {
	if(MARKING::isEqual(marking, it->second))
	{
    		*node = it->first;
    		return true;
	}
  }

   return false;
}

MARKING COV_GRAPH_COLOURED::getNode(IDTYPE id)
{
	if(m_markingMapping.find(id) != m_markingMapping.end())
		return m_markingMapping[id];
	else
		return *(new MARKING());
}
  
bool COV_GRAPH_COLOURED::getNodeWithSmallerMarking(MARKING & node, MARKING & marking, MARKING * smallerNode)
{
//    if (m_nodeMapping.find(node) == m_nodeMapping.end())
//      throw invalid_argument("CovGraph::containsNodeWithSmallerMarking() - No such node");

    map<IDTYPE,bool> visited;
    for(map<IDTYPE,MARKING>::iterator i = m_markingMapping.begin(); i != m_markingMapping.end(); i++)
      visited[i->first] = false;

    IDTYPE actualNode = getID(node);
    stack<IDTYPE> work;

    bool finished = false;

    while (! finished)
    {
      if (MARKING::isStrictlyLower(m_markingMapping[actualNode], marking))
      {
        *smallerNode = m_markingMapping[actualNode];
        return true;
      }
      for(map<IDTYPE,IDTYPE>::iterator i = m_pred[actualNode].begin(); i != m_pred[actualNode].end(); i++)
      {
        if (! visited[i->second])
        {
          work.push(i->second);
          visited[i->second] = true;
        }
      }
      if (! work.empty())
      {
        actualNode = work.top();
        work.pop();
      }
      else
        finished = true;
    }

    return false;
}


void COV_GRAPH_COLOURED::changeIDs(map<IDTYPE, IDTYPE> eqrel, map<IDTYPE, IDTYPE> transmap)
{
   // nodes
   // map<Marking,IDTYPE> m_nodeMapping_;
   map<IDTYPE,MARKING> m_markingMapping_;

   map<IDTYPE,IDTYPE> nodes;

   for(map<IDTYPE,MARKING>::iterator it = m_markingMapping.begin();
   	it!=m_markingMapping.end(); it++)
   {
   	map< IDTYPE, vector<TOKEN> > M_;

	map< IDTYPE, vector<TOKEN> > M = it->second.getMarkings();
	
	for(map<IDTYPE, vector<TOKEN> >::iterator im = M.begin(); im!=M.end(); im++)
	{
		IDTYPE id = eqrel[im->first];
		if(id!=0)
		{
			map< IDTYPE, vector<TOKEN> >::iterator itm = M_.find(id);
			if(itm!=M_.end())
			 M_[id]=plus(itm->second, im->second);
			else
			 M_[id] = im->second;
		}	
		else
		{
			map< IDTYPE, vector<TOKEN> >::iterator itm = M_.find(id);
			if(itm!=M_.end())
			 M_[im->first]=plus(itm->second, im->second);
			else
			 M_[im->first]=itm->second;
		}	
	}

	MARKING Marking_;
	Marking_.setMarkings(M_);
	
	map<IDTYPE, MARKING>::iterator itnM;
	for(itnM = m_markingMapping_.begin(); itnM != m_markingMapping_.end(); itnM++)
	{
		if(MARKING::isEqual(Marking_, itnM->second))
			break;
	}

	if(itnM==m_markingMapping_.end())
	{
		// m_nodeMapping_[Marking_]=it->first;
		nodes[it->first]=it->first;
		m_markingMapping_[it->first] = Marking_;
	}
	else
		nodes[it->first]=itnM->first;	
   }
   
   m_markingMapping=m_markingMapping_;
   // m_nodeMapping=m_nodeMapping_;
   
   // edges
   map<IDTYPE,map<IDTYPE,IDTYPE> > m_succ_;
   map<IDTYPE,map<IDTYPE,IDTYPE> > m_pred_;
   
   // succ
   for(map<IDTYPE,map<IDTYPE,IDTYPE> >::iterator itS = m_succ.begin();
   	itS!=m_succ.end(); itS++)
   {
   	IDTYPE place1;
	IDTYPE node1 = nodes[itS->first];
   	if(node1!=0)
		place1 = node1;
	else
		place1 = itS->first;
		
   	map<IDTYPE, IDTYPE> m;
	for(map<IDTYPE,IDTYPE>::iterator itS1 = itS->second.begin();
		itS1!=itS->second.end(); itS1++)
	{
		IDTYPE place2;
		IDTYPE node2 = nodes[itS1->second];
		if(node2!=0)
			place2 = node2;
		else
			place2 = itS1->second;

		m[itS1->first]=place2;
	}
	
	m_succ_[place1]=m;
   }
   
   // pred
   for(map<IDTYPE,map<IDTYPE,IDTYPE> >::iterator itS = m_pred.begin();
   	itS!=m_pred.end(); itS++)
   {
   	IDTYPE place1;
	IDTYPE node1 = nodes[itS->first];
   	if(node1!=0)
		place1 = node1;
	else
		place1 = itS->first;
		
   	map<IDTYPE, IDTYPE> m;
	for(map<IDTYPE,IDTYPE>::iterator itS1 = itS->second.begin();
		itS1!=itS->second.end(); itS1++)
	{
		IDTYPE place2;
		IDTYPE node2 = nodes[itS1->second];
		if(node2!=0)
			place2 = node2;
		else
			place2 = itS1->second;

		m[itS1->first]=place2;
	}
	
	m_pred_[place1]=m;
   }
  
   m_pred = m_pred_;
   m_succ = m_succ_;

   map<IDTYPE, pair<IDTYPE, BINDING> > m_transitions_;

   // bindings
   for(map<IDTYPE, pair<IDTYPE, BINDING> >::iterator itB = m_transitions.begin(); itB != m_transitions.end(); itB++)
   {
	IDTYPE trans = 0;
	IDTYPE trans_id = transmap[itB->second.first];
	if(trans_id!=0)
		trans = trans_id;
	else
		trans = itB->second.first;	

	m_transitions_[itB->first] = pair<IDTYPE, BINDING>(trans, itB->second.second);
   }

   // trick
   // if the graph was ready and we have made only unfolding step,
   // then we add the transition and the graph remains ready
   DATA * d = storage.getData("coverability");
   storage.setData("coverability", NULL);
   // unfolding
   if(d!=NULL)
   { 
     if(status==READY)
     {
      TRANSITION_INFO * ti = (TRANSITION_INFO*)d;
      unfold_trans_id = transmap[ti->transition];    
     } 
     else
     {
      unfold_trans_id=0;
     }
   }
   // folding
   else
   {
    unfold_trans_id=0;
    status=SMTH;
   } 
}

  vector<TOKEN> COV_GRAPH_COLOURED::plus(vector<TOKEN> v1, vector<TOKEN> v2)
  {
	vector<TOKEN> result;

	// to delete
	vector<TOKEN> copy_v2 = v2;

	for(vector<TOKEN>::iterator it1 = v1.begin(); it1 != v1.end(); it1++)
	{
		bool found = false;

		for(vector<TOKEN>::iterator it2 = copy_v2.begin(); it2 != copy_v2.end(); it2++)
		{
			if(EXPRESSION::equal(it1->getAttributes(), it2->getAttributes()))
			{
				TOKEN t = (*it1);
				t.setCount(plus(it1->getCount(), it2->getCount()));
				result.push_back(t);

				found = true;

				copy_v2.erase(it2);
				break;
			}
		}

		// not found
		if(!found)
		 result.push_back(*it1);		
	}

	// copy_v2
	for(vector<TOKEN>::iterator it_2 = copy_v2.begin(); it_2 != copy_v2.end(); it_2++)
		result.push_back(*it_2);
  }

  unsigned COV_GRAPH_COLOURED::plus(unsigned a, unsigned b)
  {  
      if((a==OMEGA) || (b==OMEGA))
       return OMEGA;
      else
       return a+b;
  }
  
  bool COV_GRAPH_COLOURED::coverNode(MARKING & m, IDTYPE * id)
  {
    // here are two cases: 
    // 1. The real coloured marking
    // 2. The number of tokens which can be covered by any attributed tokens

    // check the case
    bool is_coloured = m.isColoured();

    *id = 0;

    for(map<IDTYPE,MARKING>::iterator i = m_markingMapping.begin();
		i != m_markingMapping.end();
		i++)
    {
	if(is_coloured)
	{
		if (MARKING::isLower(m,i->second) )
		{
			*id = i->first;
			return true;
		}
	}
	else
	{
		if (MARKING::isLower_UC(m,i->second) )
		{
			*id = i->first;
			return true;
		}
	}
    }
	
    return false;

  }

  bool COV_GRAPH_COLOURED::coverNodes(list<MARKING> lm, IDTYPE * id)
  {	
	for(list<MARKING>::iterator it = lm.begin(); it != lm.end(); it++)
	{
		if(coverNode(*it, id))
		 return true;
	}

	return false;
  }

  ostream& operator<<(ostream& ost, COV_GRAPH_COLOURED& th)
  {
	ost << endl << "COLOURED COVERABILITY GRAPH" << endl;


	for(map<IDTYPE,MARKING>::iterator i = th.m_markingMapping.begin(); i != th.m_markingMapping.end(); i++)
	{
		ost << "NODE: " << i->first << endl << i->second << endl;

	}

	ost << "----------------\n";

	for(map<IDTYPE,map<IDTYPE,IDTYPE> >::iterator it = th.m_succ.begin(); it != th.m_succ.end(); it++)
	{
		for(map<IDTYPE,IDTYPE>::iterator itm = it->second.begin(); itm != it->second.end(); itm++)
		{
			ost << it->first << " " << itm->first << " " << itm->second << "\n";
		}
	}

	ost << "--- END COVERABILITY GRAPH ---" << endl;

	return ost;
  }

  list<IDTYPE> COV_GRAPH_COLOURED::getPath(IDTYPE search_id)
  {
	list<IDTYPE> result;

	// work list of node ids and path to this node
	list<pair<IDTYPE, list<IDTYPE> > > workList;

	// visited nodes
	list<IDTYPE> visited_nodes;

	// initial marking is number one
	list<IDTYPE> emptyList;
	workList.push_back(pair<IDTYPE, list<IDTYPE> >(1, emptyList));

	// bread search
	while(workList.size() != 0)
	{
		// found
		if(workList.front().first == search_id)
		 return workList.front().second;

		// call possible transitions
		map<IDTYPE,IDTYPE> mt = m_succ[workList.front().first];
		for(map<IDTYPE,IDTYPE>::iterator its = mt.begin(); its!= mt.end(); its++)
		{
			// check the loop
			bool found = false;
			for(list<IDTYPE>::iterator it = visited_nodes.begin(); it != visited_nodes.end(); it++)
			{
				if(*it == its->second)
				{
					found = true;
					break;
				}
			}

			// not found
			if(!found)
			{
				list<IDTYPE> l = workList.front().second;
				l.push_back(its->first);
				workList.push_back(pair<IDTYPE, list<IDTYPE> >(its->second, l));
			}
		}

		// delete current node
		visited_nodes.push_back(workList.front().first);
		workList.pop_front();
	}

	return result;
  }

}
}
}
}
