/***************************************************************************
 *   Copyright (C) 2006 by root   *
 *   root@Vitali   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef COV_GRAPH_COLOURED_H
#define COV_GRAPH_COLOURED_H

#include "globals.h"

#include "marking.h"
#include "token.h"
#include "transition.h"
#include "validator.h"
#include "persistable.h"

#include "algorithm_storage.h"

#include "cov_graph.h"

#include <map>
#include <list>
#include <stdexcept>
#include <vector>

namespace algorithms {
namespace ptnet {
namespace coverability {
namespace data_str_coverability {

class COV_GRAPH_COLOURED
{

public:

    COV_GRAPH_COLOURED();

    ~COV_GRAPH_COLOURED();

    IDTYPE addNode(MARKING & m);
    void addEdge(MARKING & m1, MARKING & m2, TRANSITION & transition, BINDING & binding);
    void addEdge(IDTYPE id1, IDTYPE id2,  TRANSITION & transition, BINDING & binding);

    void clear();

    IDTYPE getID(MARKING & m);
    MARKING getNode(IDTYPE id);

    bool containsNode(MARKING & marking, unsigned * node);  
    bool getNodeWithSmallerMarking(MARKING & node, MARKING & marking, MARKING * smallerNode);

    void changeIDs(map<IDTYPE, IDTYPE> eqrel, map<IDTYPE, IDTYPE> transmap);

    long getSize();
    bool coverNode(MARKING & m, IDTYPE * id);
    bool coverNodes(list<MARKING> lm, IDTYPE * id);

    // status of the graph
    STATUS status;

    // trick, to optimize the calculation after unfolding step
    IDTYPE unfold_trans_id;

    IDTYPE node_counter;

    // map<MARKING,IDTYPE> m_nodeMapping;
    map<IDTYPE,MARKING> m_markingMapping;

    /**
     * 1. IDTYPE: Node of the coverability graph
     * 2. IDTYPE: ID of the corresponding transition in the PTNet
     * 3. IDTYPE: Successor node in coverability graph
     */
    map<IDTYPE,map<IDTYPE,IDTYPE> > m_succ;
    map<IDTYPE,map<IDTYPE,IDTYPE> > m_pred;

    // bindings of transitions
    // number -> trans_id, BINDING
    map<IDTYPE, pair<IDTYPE, BINDING> > m_transitions; 
    // counter for number
    long trans_counter;

    vector<TOKEN> plus(vector<TOKEN> v1, vector<TOKEN> v2);
    unsigned plus(unsigned a, unsigned b);

    friend ostream& operator<<(ostream& ost, COV_GRAPH_COLOURED& th);

    list<IDTYPE> getPath(IDTYPE search_id);
};

};
};
};
};

#endif
