/***************************************************************************
 *   Copyright (C) 2006 by root   *
 *   root@Vitali   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "cover_scenario.h"
#include "gxlreader.h"
#include "marking_reader.h"
#include "bwra.h"

namespace scenarios
{

COVER_SCENARIO::COVER_SCENARIO()
{
      this->ID = ALGORITHM_COVER;
      this->name = "coverability_scenario";
      
      this->inspec.push_back(pair<string, DATA_FORMAT>("petrinet", GXL));
      this->inspec.push_back(pair<string, DATA_FORMAT>("marking", PLAIN_TEXT));	
      this->outspec.push_back(pair<string, DATA_FORMAT>("trace", PLAIN_TEXT));
      
      this->call_descriptor = "Usage: augur -sc=coverability petrinet marking trace\n\n\
petrinet\tpetri net in GXL format\n\
marking\ttested marking\n\
trace\ttrace to coverable marking\n\n";
}


COVER_SCENARIO::~COVER_SCENARIO()
{
}

void COVER_SCENARIO::run()
{
 // read Petri net
 GXLREADER gxlreader;
 gxlreader.initialize((char*)(this->filemap["petrinet"]).c_str());
 PETRIGRAPH pg;
 GTS gts;
 gxlreader.run(pg, gts);
 PTNET pn = (PTNET)pg;

 // read markings
 MARKING_READER mreader;
 mreader.initialize((char*)(this->filemap["marking"]).c_str());
 list<MARKING> lm = mreader.run(0);

 writePPID();

 // run bwra
 ALGORITHM* cover = this->load("cover_algorithm");
 cover->initialize(&pn);

 init_time();

 cover->run(lm, (char*)(this->filemap["trace"]).c_str()); 

 cout << "time used: " << get_time() << " sec." << endl;
}

}


