/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "coverability.h"
#include "algorithm_storage.h"

namespace algorithms
{

namespace ptnet
{

namespace coverability
{

COVERABILITY::COVERABILITY()
{
 this->ID = ALGORITHM_COVERABILITY;
 this->name = "coverability";
 tracefile = NULL;
}

COVERABILITY::~COVERABILITY()
{
}

void COVERABILITY::initialize(PTNET * i_ptnet)
{
    this->pn = i_ptnet;
    
    // firing algorithm
    if(coloured)
     ft.initialize(i_ptnet, storage.getEngine()); 
    else
     ft.initialize(i_ptnet); 
}

bool COVERABILITY::run(MARKING m, long * psize)
{ 
    if(psize!=NULL)
     *psize = 0;
    else
     cout << "internal pointer error" << endl;

    if(coloured)
    {
	list<MARKING> lm;
    	lm.push_back(m);
	return internal_run(lm, psize);
    }
    else
    {
	list<Marking> lmu;
    	lmu.push_back(m.getUncoloured());
	return internal_run(lmu, psize);
    }
}

bool COVERABILITY::internal_run(list<Marking> lmu, long * psize)
{
    bool result = false;

    if (m_covGraph.getSize()==0)
    {
	//first call at all or since last CovGraph::clear()
	Marking initial = pn->getInitialMarking().getUncoloured();
	m_covGraph.addNode(initial);
	IDTYPE firstNode = m_covGraph.getID(initial);
	m_work.push_back(pair<IDTYPE,list<TRANSITION> >(firstNode, 
	ft.getEnabledTransitions(initial)));
	if(m_work.size()!=0)
	 m_covGraph.status = SMTH;
	// cout << "work size: " << m_work.front().second.size() << endl;
    }
    else if(m_covGraph.status!=READY)
    {	
	// set work    
    	for(map<IDTYPE, Marking>::iterator itn=m_covGraph.m_markingMapping.begin();
			itn!=m_covGraph.m_markingMapping.end(); itn++)
	{
	 m_work.push_back(pair<IDTYPE,list<TRANSITION> >(itn->first,
	 	ft.getEnabledTransitions(itn->second, m_covGraph.m_succ[itn->first])));
	}
    }
    else if(m_covGraph.status==READY)
    {
      // unfolding, only one transition
      if(this->reusable && m_covGraph.unfold_trans_id!=0)
      {
        addTransition(m_covGraph.unfold_trans_id);
	m_covGraph.unfold_trans_id=0;
      }
    }

    IDTYPE id = 0;

    // check coverability in the existing part    
    if (m_covGraph.coverNodes(lmu, &id))
    {
	if(full_coverability_graph)
	{
		result=true;
	}
	else
	{ 
		*psize = m_covGraph.getSize();
		m_work.clear();
		writeCovGraph(id);
		return true;
	}	
    }
    else
    { 
	 // if status==READY then not coverable
	 if(m_covGraph.status==READY)
	  return false;
    }

    int counter = 0;

    while (! m_work.empty())
    {
	// timeout
	if(timeout != 0)
	{
		if(get_time() > timeout)
		{
			cout << "Exited by Timeout" << endl;
			break;
		}
	}

      counter++;

      IDTYPE actualNode = m_work.begin()->first;
      Marking actualMarking = m_covGraph.m_markingMapping[actualNode];

      if ((m_work.begin()->second).empty())
      {
        m_work.erase(m_work.begin());
      }
      else
      {
        TRANSITION actualTransition = *(m_work.begin()->second).begin();
        (m_work.begin()->second).erase((m_work.begin()->second).begin());      

        Marking newMarking = ft.run(actualTransition, actualMarking);

        Marking smallerMarking; 

        bool smaller = m_covGraph.getNodeWithSmallerMarking(actualMarking, newMarking, &smallerMarking);

        if (smaller)
        {
	  map<IDTYPE, EDGE> * edges = pn->getEdgesPointer();  
	  for(map<IDTYPE, EDGE>::iterator i = edges->begin(); i!= edges->end(); i++)
          {
	    Marking::iterator smallerMarkingIt = smallerMarking.find(i->first);
            Marking::iterator newMarkingIt = newMarking.find(i->first);
	    unsigned smallerMarkingValue=0;
            unsigned newMarkingValue=0;
	    if(smallerMarkingIt!=smallerMarking.end())
	     smallerMarkingValue = smallerMarking[i->first];
	    if(newMarkingIt!=newMarking.end())
	     newMarkingValue = newMarking[i->first]; 
            if(smallerMarkingValue < newMarkingValue)
              newMarking[i->first] = OMEGA;
          }
        }

	unsigned node_id;
        if(! m_covGraph.containsNode(newMarking, &node_id))
        {
	  IDTYPE id_ = m_covGraph.addNode(newMarking);

	  for(list<Marking>::iterator it_lmu = lmu.begin(); it_lmu != lmu.end(); it_lmu++)
	  {
		if(isLower(*it_lmu, newMarking))
		{
			if(full_coverability_graph)
			{
				if(!result)
				{
				 id = id_;
				 result=true;
				}
			}
			else
			{
				m_work.clear();
				*psize=m_covGraph.getSize();
				m_covGraph.addEdge(actualMarking, newMarking, actualTransition);
				writeCovGraph(id_);
				return true;
			} 
		}
	  }

	  list<TRANSITION> trans = ft.getEnabledTransitions(newMarking);
          if (trans.size()!=0)
            m_work.push_back(pair<IDTYPE,list<TRANSITION> >
	         (m_covGraph.getID(newMarking), trans));
	} 

        m_covGraph.addEdge(actualMarking, newMarking, actualTransition);
      }
    } 

    *psize=m_covGraph.getSize();
    m_covGraph.status = READY;

    if(full_coverability_graph && result==true)
    {
	if(cov_test)
		cout << endl << "coverability size: " << *psize << ", ";
	else
		cout << "coverability size: " << *psize << " ";

	writeCovGraph(id);

     	return true;
    } 
    else 
     return false;
}

bool COVERABILITY::internal_run(list<MARKING> lm, long * psize)
{
    bool result = false;

    if (m_covGraphCol.getSize()==0)
    {
	//first call at all or since last CovGraph::clear()
	MARKING initial = pn->getInitialMarking();

//	cout << "\ninitial marking\n" << initial;
//	cout << "\nmarking to test\n" << marking;

	// evaluate expressions
	ALGORITHM * engine = storage.getEngine();
	map< IDTYPE, vector<TOKEN> > initial_marking = initial.getMarkings();
	map< IDTYPE, vector<TOKEN> > initial_marking_evaluated;
	for(map< IDTYPE, vector<TOKEN> >::iterator it=initial_marking.begin(); it!=initial_marking.end(); it++)
	{
		vector<TOKEN> vt;
		for(vector<TOKEN>::iterator itt  = it->second.begin(); itt != it->second.end(); itt++)	
		{
			TOKEN t;
			t.setCount(itt->getCount());
			list<EXPRESSION> le = itt->getAttributes();
			list<EXPRESSION> le_evaluated;
			for(list<EXPRESSION>::iterator ite = le.begin(); ite != le.end(); ite++)
			{
				EXPRESSION e = engine->run(*ite);
				le_evaluated.push_back(e);
			}
		
			t.setAttributes(le_evaluated);
			vt.push_back(t);
		}
		initial_marking_evaluated[it->first] = vt;
	}

	initial.setMarkings(initial_marking_evaluated);

	IDTYPE firstNode =  m_covGraphCol.addNode(initial);
	m_workCol.push_back(pair<IDTYPE,list<pair<TRANSITION, BINDING> > >(firstNode, 
	ft.getEnabledTransitions(initial)));
	if(m_workCol.size()!=0)
	 m_covGraphCol.status = SMTH;
	// cout << "work size: " << m_work.front().second.size() << endl;
    }
    else if(m_covGraphCol.status!=READY)
    {	
	// set work    
    	for(map<IDTYPE, MARKING>::iterator itn=m_covGraphCol.m_markingMapping.begin();
			itn!=m_covGraphCol.m_markingMapping.end(); itn++)
	{
	 m_workCol.push_back(pair<IDTYPE,list<pair<TRANSITION, BINDING> > >(itn->first,
	 	ft.getEnabledTransitions(itn->second, m_covGraphCol.m_succ[itn->first])));
	}
    }
    else if(m_covGraphCol.status==READY)
    {
      // unfolding, only one transition
      if(this->reusable && m_covGraphCol.unfold_trans_id!=0)
      {
        addTransitionCol(m_covGraphCol.unfold_trans_id);
	m_covGraphCol.unfold_trans_id=0;
      }
    }

    IDTYPE id = 0;

    // check coverability in the existing part    
    if (m_covGraphCol.coverNodes(lm, &id))
    {
	if(full_coverability_graph)
	{
		result=true;
	}
	else
	{ 
		*psize = m_covGraphCol.getSize();
		m_workCol.clear();
		writeCovGraph(id);
		return true;
	}	
    }
    else
    { 
	 // if status==READY then not coverable
	 if(m_covGraphCol.status==READY)
	 {
	  // writeCovGraph(id);
	  return false;
	 }
    }

    while (! m_workCol.empty())
    {

	// timeout
	if(timeout != 0)
	{
		if(get_time() > timeout)
		{
			cout << "Exited by Timeout" << endl;
			break;
		}
	}

      IDTYPE actualNode = m_workCol.begin()->first;
      MARKING actualMarking = m_covGraphCol.m_markingMapping[actualNode];

      if ((m_workCol.begin()->second).empty())
      {
        m_workCol.erase(m_workCol.begin());
	// tmp
	// cout << "erase" << endl;
      }
      else
      {
        TRANSITION actualTransition = (*(m_workCol.begin()->second).begin()).first;
	BINDING actualBinding = (*(m_workCol.begin()->second).begin()).second;

        (m_workCol.begin()->second).erase((m_workCol.begin()->second).begin());

        MARKING newMarking = ft.run(actualTransition, actualMarking, actualBinding);

        MARKING smallerMarking; 

        bool smaller = m_covGraphCol.getNodeWithSmallerMarking(actualMarking, newMarking, &smallerMarking);

        if (smaller)
        {
	  map<IDTYPE, vector<TOKEN> > smallerMarking_ = smallerMarking.getMarkings();
	  map<IDTYPE, vector<TOKEN> > newMarking_ = newMarking.getMarkings();

	  map<IDTYPE, EDGE> * edges = pn->getEdgesPointer();

	  for(map<IDTYPE, EDGE>::iterator i = edges->begin(); i!= edges->end(); i++)
          {
	    map<IDTYPE, vector<TOKEN> >::iterator smallerMarkingIt = smallerMarking_.find(i->first);
            map<IDTYPE, vector<TOKEN> >::iterator newMarkingIt = newMarking_.find(i->first);


	    if(newMarkingIt == newMarking_.end())
		continue;

	    // tokens to insert in newMarking_
	    vector<TOKEN> tokens;

	    for(vector<TOKEN>::iterator itn = newMarkingIt->second.begin(); itn != newMarkingIt->second.end(); itn++)
	    {
		// copy
		TOKEN t = *itn;

		if(smallerMarkingIt != smallerMarking_.end())
		{
			vector<TOKEN>::iterator its;
			for(its = smallerMarkingIt->second.begin(); its != smallerMarkingIt->second.end(); its++)
			{
				// found
				if(EXPRESSION::equal(itn->getAttributes(), its->getAttributes()))
				{
					if(its->getCount()<itn->getCount())
						t.setCount(OMEGA);
					break;
				}
			}

			// not found
			if(its == smallerMarkingIt->second.end())
		 		t.setCount(OMEGA);
		}
		else
		{
			// empty place, all values will be set in OMEGA
			t.setCount(OMEGA);
		}

		tokens.push_back(t);
	    }

	    newMarking_[newMarkingIt->first] = tokens;
          }

	  newMarking.setMarkings(newMarking_);
        }

	unsigned node_id = 0;
        if(! m_covGraphCol.containsNode(newMarking, &node_id))
        {
	  // add the node
	  IDTYPE id_ = m_covGraphCol.addNode(newMarking);

	  // tmp
	  // cout << "coverability size: " << m_covGraphCol.getSize() << endl;
	  // cout << newMarking;

	  list<pair<TRANSITION, BINDING> > trans = ft.getEnabledTransitions(newMarking);
          if (trans.size()!=0)
            m_workCol.push_back(pair<IDTYPE,list<pair<TRANSITION, BINDING> > >
	         (m_covGraphCol.getID(newMarking), trans));

	  // tmp	  
	  // cout << "work: " << m_workCol.size() << endl;

	  // here are two cases: 
	  // 1. The real coloured marking
    	  // 2. The number of tokens which can be covered by any attributed tokens

	  bool lower = false;
	  for(list<MARKING>::iterator it_lm = lm.begin(); it_lm!=lm.end(); it_lm++)
	  {
		// check the case
		bool is_coloured = it_lm->isColoured();		
	
		if(is_coloured)
			lower = MARKING::isLower(*it_lm, newMarking);
		else
			lower = MARKING::isLower_UC(*it_lm, newMarking);

		if(lower)
		 break;
	  }

          if(lower)
          {
	       	if(full_coverability_graph)
	        {
		 if(!result)
		 {
		  id = id_;
	          result=true;
		 }
	        }
		else
		{
	  	 m_workCol.clear();
		 *psize=m_covGraphCol.getSize(); 
		 m_covGraphCol.addEdge(actualMarking, newMarking, actualTransition, actualBinding);
		 writeCovGraph(id_);
		 return true;
		} 
          }
	}
	else
	{
		// cout << "contained" << endl;
	} 

        m_covGraphCol.addEdge(actualMarking, newMarking, actualTransition, actualBinding);
      }
    } 

    *psize=m_covGraphCol.getSize();
    m_covGraphCol.status = READY;

    if(full_coverability_graph)
    {
	if(cov_test)
		cout << endl << "coverability size: " << *psize << ", ";
	else
		cout << "coverability size: " << *psize << "\n";

	if(result==true)
	 writeCovGraph(id);
    }

    return result;
}

void COVERABILITY::update(vector<EVENT> events)
{
  if(events.size()==0)
   return;
     
  if(events.size()>1)
  {   
   cout << "COVERABILITY::validate unknown event\n";
   exit(1);   
  } 
  
  // trick
  // if the coverability graph will be cleaned next time, then
  // it makes no sense to update it now
  if(clear_cov_count!=0)
  {
     if((iteration+1)%clear_cov_count==0)
     {
        storage.setData("coverability", NULL);
	if(coloured)
	 m_covGraphCol.clear();
	else
	 m_covGraph.clear();
        return;
     }
  }  
    
  // places and transitions
  map<IDTYPE, IDTYPE> p_rel;
  map<IDTYPE, IDTYPE> t_rel;
     
  for(map<IDTYPE, IDTYPE>::iterator it = events[0].ids.begin(); it!=events[0].ids.end(); it++)
  {
   // nodes_as_places is already in hasPlace() implemented
   if(pn->hasPlace(it->second)) 
     p_rel.insert(pair<IDTYPE, IDTYPE>(*it));
   
   if(pn->hasTransition(it->second))
     t_rel.insert(pair<IDTYPE, IDTYPE>(*it));    
  } 

  if(cov_test)
  {
   // number of equivalence classes
   // map is because of find
   map<IDTYPE, IDTYPE> eq_classes;
   long counter = 0;
   for(vector<EVENT>::iterator it = events.begin();
	it != events.end(); it++)
   {
	for(map<IDTYPE, IDTYPE>::iterator itm = it->ids.begin();
		itm !=it->ids.end(); itm++)
	{
		if(eq_classes.find(itm->second)==eq_classes.end())
			eq_classes[itm->second]=1;
		counter++;
	}
   }
   
   cout << "\nmerged elements (%): " << ((double)(counter-eq_classes.size()))/((double)counter) << endl;
  }

  // change IDs in coverability graph
  if(cov_test)
   init_time();

  if(coloured)
   m_covGraphCol.changeIDs(p_rel, t_rel);
  else
   m_covGraph.changeIDs(p_rel, t_rel);

  if(cov_test)
   cout << "time used for reducing: " << get_time() << " sec." << endl;
   
}

void COVERABILITY::addTransition(IDTYPE t)
{
    if(coloured)
    {
	addTransitionCol(t);
	return;
    }

    // transition
    TRANSITION transition = pn->getTransition(t);
    transition.setID(t);
    	
    // find nodes covering the preset
    list<pair<IDTYPE, Marking> > nodes;
    for(map<IDTYPE, Marking>::iterator it = m_covGraph.m_markingMapping.begin();
    		it!=m_covGraph.m_markingMapping.end(); it++)
    {   
     if(ft.isEnabledBy(it->second, t))
     	nodes.push_back(*it);
    }	
	
    for(list<pair<IDTYPE, Marking> >::iterator itn = nodes.begin(); itn !=nodes.end(); itn++)
    {     
	// fire transition
	list<Marking> lm = ft.run_n_times(transition, itn->second);
	
	for(list<Marking>::iterator itl = lm.begin();
		itl != lm.end(); itl++)
	{
		// node 2
		IDTYPE id;
		if(!m_covGraph.containsNode(*itl, &id))
		{
			m_covGraph.addNode(*itl);
			id=m_covGraph.node_counter;			
		}
			
		// insert edge
		m_covGraph.addEdge(itn->first, id, t);
	}
    } 
}

void COVERABILITY::addTransitionCol(IDTYPE t)
{
    // transition
    TRANSITION transition = pn->getTransition(t);
    transition.setID(t);
	
    // find nodes covering the preset
    list<pair<IDTYPE, pair<MARKING, BINDING> > > nodes;
    for(map<IDTYPE, MARKING>::iterator it = m_covGraphCol.m_markingMapping.begin();
    		it!=m_covGraphCol.m_markingMapping.end(); it++)
    {
	list<BINDING> lb = ft.getEnabledBindings(it->second, t);
	for(list<BINDING>::iterator itb = lb.begin(); itb != lb.end(); itb++)
		nodes.push_back(pair<IDTYPE, pair<MARKING, BINDING> >(it->first, pair<MARKING, BINDING>(it->second, *itb)));
    }
	
    for(list<pair<IDTYPE, pair<MARKING, BINDING> > >::iterator itn = nodes.begin(); itn !=nodes.end(); itn++)
    {
	// fire transition
	list<MARKING> lm = ft.run_n_times(transition, itn->second.first, itn->second.second);
	
	for(list<MARKING>::iterator itl = lm.begin(); itl != lm.end(); itl++)
	{
		// node 2
		IDTYPE id;
		if(!m_covGraphCol.containsNode(*itl, &id))
		{
			m_covGraphCol.addNode(*itl);
			id=m_covGraphCol.node_counter;
		}

		// insert edge
		m_covGraphCol.addEdge(itn->first, id, transition, itn->second.second);
	}
    }
}

COV_GRAPH * COVERABILITY::getCovGraph()
{
	return &m_covGraph;
}

COV_GRAPH_COLOURED * COVERABILITY::getCovGraphCol()
{
	return &m_covGraphCol;
}

bool COVERABILITY::run(list<MARKING> lm, char * tracefile)
{
	this->tracefile = tracefile;

	// size of graph
	long psize;

	if(coloured)
	{
		if(internal_run(lm, &psize))
		{			
			cout << endl << "The Final Marking(s) are coverable\n";
			return true;
		}
		else
		{
			cout << endl << "The Final Marking(s) are not coverable.\n";
			return false;
		}
	}
	else
	{
		list<Marking> lmu;
		for(list<MARKING>::iterator it = lm.begin(); it != lm.end(); it++)
			lmu.push_back(it->getUncoloured());

		if(internal_run(lmu, &psize))
		{			
			cout << endl << "The Final Marking(s) are coverable\n";
			return true;
		}
		else
		{
			cout << endl << "The Final Marking(s) are not coverable.\n";
			return false;
		}
	}
}

void COVERABILITY::writeCovGraph(IDTYPE id)
{
	// write somthing only in case of coverability scenario
	// this is checked by the initialization of tracefile
	if(tracefile == NULL)
	 return;

	if(coloured)
	{
		cout << "Searching the trace to id=" << id << endl;
		list<IDTYPE> ltb = m_covGraphCol.getPath(id);

		// output in file
		ofstream of(tracefile);

		cout << "write counter-example to " << tracefile << endl;
		
		for(list<IDTYPE>::iterator it = ltb.begin(); it != ltb.end(); it++)
		{
			of << (pn->getTransition(m_covGraphCol.m_transitions[*it].first)).getLabel() << endl;
			of << m_covGraphCol.m_transitions[*it].first << endl;
			of << "% " << m_covGraphCol.m_transitions[*it].second << endl;
		}
		
		of.close();
	}
	
	if(!coloured)
	{
		cout << "Searching the trace to id=" << id << endl;
		list<IDTYPE> ltb = m_covGraph.getPath(id);

		// output in file
		ofstream of(tracefile);

		cout << "write counter-example to " << tracefile << endl;
		
		for(list<IDTYPE>::iterator it = ltb.begin(); it != ltb.end(); it++)
		{
			of << pn->getTransition(*it).getLabel() << endl;
			of << *it << endl;
			of << endl;
		}
		
		of.close();
	}
}

};

};

};

