/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef COVERABILITY_H
#define COVERABILITY_H

#include "algorithm.h"
#include "ptnet.h"
#include "globals.h"
#include "cov_graph.h"
#include "fire_transition.h"
#include "cov_graph_coloured.h"

namespace algorithms
{

namespace ptnet 
{

namespace coverability
{

class COVERABILITY: public ALGORITHM {
public:
    COVERABILITY();

    ~COVERABILITY();

    virtual void initialize(PTNET * i_ptnet);
    virtual bool run(MARKING m, long * psize);

    // validate in case of incremental coverability graph
    virtual void update(vector<EVENT> events);

    COV_GRAPH * getCovGraph();
    COV_GRAPH_COLOURED * getCovGraphCol();

    // coverability with counter-example
    virtual bool run(list<MARKING> lm, char * tracefile);

private:

    PTNET * pn;

    COV_GRAPH m_covGraph;

    // coloured graph
    COV_GRAPH_COLOURED m_covGraphCol;

    FIRE_TRANSITION ft;

    list<pair<IDTYPE,list<TRANSITION> > > m_work;
    list<pair<IDTYPE, list<pair<TRANSITION, BINDING> > > > m_workCol;

    void addTransition(IDTYPE t);
    void addTransitionCol(IDTYPE t);

    // internal runs for coloured and not coloured cases
    bool internal_run(list<MARKING> lm, long * psize);
    bool internal_run(list<Marking> lmu, long * psize);

    void writeCovGraph(IDTYPE id);

    char * tracefile;
};

};

};

};

using namespace algorithms;
using namespace ptnet;
using namespace coverability;

#endif
