/***************************************************************************
 *   Copyright (C) 2007 by Salil Joshi   *
 *   hellfeuer@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef ALGORITHMS_FINITE_AUTOMATONCROSS_PRODUCT_H
#define ALGORITHMS_FINITE_AUTOMATONCROSS_PRODUCT_H

#include "hypergraph.h"
#include "finite_automaton.h"
#include "globals.h"

#include <stdlib.h>

namespace algorithms {

namespace hypergraph {

/**
	@author hellfeuer <hellfeuer@gmail.com>
	Computes cross product of finite automaton and hypergraph, as given in "Graph transformation systems, Petri nets and Semilinear Sets: Checking for the Absence of Forbidden Paths in Graphs" Barbara Koenig, page 7.

*/

class CROSS_PRODUCT
{

private:

class PRODUCTMAP
{
	list<pair<pair<IDTYPE, IDTYPE>, STATE*> > prmap;

public:
	void insert(pair<IDTYPE, IDTYPE> p, STATE * s);
	pair<pair<IDTYPE, IDTYPE>, STATE*>  * find(pair<IDTYPE, IDTYPE> p);
};

	FINITE_AUTOMATON* fa;
 	HYPERGRAPH* hg;
public:
    CROSS_PRODUCT();
    
    void initialize(FINITE_AUTOMATON* f, HYPERGRAPH* h);
    void run(FINITE_AUTOMATON* fb);

    ~CROSS_PRODUCT();

    void createTransition(IDTYPE fromId, IDTYPE toId, STATE* fromState, 
	set<STATE*> &toStates, string label, FINITE_AUTOMATON* fb, PRODUCTMAP * productmap);

};

}

}

using namespace algorithms;
using namespace hypergraph;

#endif
