/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "equivalence_constructor.h"

namespace algorithms {

namespace refinement {

EQUIVALENCE_CONSTRUCTOR::EQUIVALENCE_CONSTRUCTOR()
{
}


EQUIVALENCE_CONSTRUCTOR::~EQUIVALENCE_CONSTRUCTOR()
{
}

void EQUIVALENCE_CONSTRUCTOR::initialize(PETRIGRAPH * pg, vector<IDTYPE> trans, SAMPLE * s)
{
	this->pg = pg;
	
	this->trans = trans;

	this->m_list = &(s->tree);

	// sample itself is needed to cut the transitions after optimization
	this->sample = s;
}

void EQUIVALENCE_CONSTRUCTOR::run()
{

try
{
  // from empty level in m_list
  const int level = (*m_list).size() - 2;

  HYPERGRAPH * L  =pg->getAssociatedRule(trans[level])->getLeft();

  MATCH mu = pg->getMu(trans[level]);

  map<IDTYPE, VERTEX> nodesL = L->getVertices();

  assert((*m_list)[level].size()!=0);

  // for(int linklevel=level; linklevel>=0; linklevel--)
  int linklevel=level;
  {
	// hypergraphs
	for(list<HYPERGRAPH_INFO>::iterator it = (*m_list)[linklevel].begin(); it!=(*m_list)[linklevel].end(); it++)
	{
		// consider only if not linked
		if(isLinked(it->id, linklevel))
			continue;
		
		// assert that the matchings are initialized
		assert(it->matchesGgM.begin()!=it->matchesGgM.end());

		// matches G -> M
		for(list<MATCH>::iterator itmGM = it->matchesGgM.begin(); itmGM!=it->matchesGgM.end(); itmGM++)
		{
			// nodes
			for(map<IDTYPE, VERTEX>::iterator it_nodes=nodesL.begin(); it_nodes!=nodesL.end(); it_nodes++)
			{
				// class
				set<NODE_INFO> eq_class;
				multiset<IDTYPE> edges = L->getConnectedEdges(it_nodes->first);

				// edges
				for(multiset<IDTYPE>::iterator it_edges=edges.begin(); it_edges!=edges.end(); it_edges++)
				{
               				IDTYPE id_edge_m = mu.getMatch(*it_edges);

               				// find pre-image in matching G -> M
               				map<IDTYPE, IDTYPE>::iterator itr;
					
					map<IDTYPE, IDTYPE> mapGM = itmGM->getMatch();
               
               				for(itr = mapGM.begin(); itr!=mapGM.end(); itr++)
               				{
						 if(itr->second == id_edge_m)
							 break;
               				} 

               				if(itr==mapGM.end())
					{
						cout << "Error in EQUIVALENCE_CONSTRUCTOR: no GM matches\n";
						exit(1);
					}

               				IDTYPE id_edge_g = itr->first;

               				// order in L
					list<IDTYPE> verticesL = L->getConnectedVertices(*it_edges);

					// order in G
					list<IDTYPE> verticesG = it->hg.getConnectedVertices(id_edge_g);

					// find the corresponding node in G
					list<IDTYPE>::iterator it_nG = verticesG.begin();
					for(list<IDTYPE>::iterator it_nL = verticesL.begin(); it_nL!=verticesL.end(); it_nL++)
					{
						 if(*it_nL == it_nodes->first)
						 {
		 					 NODE_INFO ni;
							 ni.id = *it_nG;
							 ni.graph_id = it->id;
							 ni.level = linklevel;

							 if(eq_class.find(ni)==eq_class.end())
								 eq_class.insert(ni);
						 } 

						 it_nG++;
					 }

				} // edges

            			if(eq_class.size()>1)
            			{
					// is the equivalence already in list
					bool is_in_list = false;
					for(list<set<NODE_INFO> >::iterator itE = equivalences.begin(); itE!=equivalences.end(); itE++)
					{
                				bool not_found = false;
						for(set<NODE_INFO>::iterator itN = eq_class.begin(); itN!=eq_class.end(); itN++)
						{
							if(itE->find(*itN)==itE->end())
							{
								not_found = true;
								break;
							}	
						}

						if(!not_found)
						{
							is_in_list = true;
							break;
						}
					}

						if(!is_in_list)
						{
							equivalences.push_back(eq_class);
							cout << "equivalence class: nodes ( ";
							for(set<NODE_INFO>::iterator it=eq_class.begin(); it!=eq_class.end(); it++)
							 cout << it->id << " ";
							cout << ")\n"; 
							 
						}	
				}	
	
			} // nodes

		} // GM

	  } // hypergraphs
	  
  } // linklevel

  // find appearing
  findAppearing();
	  
 }
 catch(exception e)
 {
	cout << "Error in EQUIVALENCE_CONSTRUCTOR\n";
	exit(1);
 }
}

/** find first appearing of each node in equivalence classes in [M_0] ... [M_n] */
void EQUIVALENCE_CONSTRUCTOR::findAppearing()
{

try
{
	// number of equivalence class
	unsigned int eq_number = 1;
	  
	for(list<set<NODE_INFO> >::iterator it_list = equivalences.begin();
				it_list!=equivalences.end(); it_list++)
	{
		for(set<NODE_INFO>::iterator it = (*it_list).begin();
							it!=(*it_list).end(); it++)
		{
			int linklevel = it->level;
			HYPERGRAPH_INFO * hgi = findHypergraph(it->graph_id, linklevel);
			HYPERGRAPH_INFO * hgi_last = NULL;
			unsigned int graph_id = it->graph_id;
			unsigned int graph_id_last;
			
			while(linklevel>=0)
			{
				// G0
				if(linklevel==0)
				{
					if(hgi_last != NULL)
					{
						if(hgi_last->eq_labels.count(it->id)==0)
						{
							list<unsigned> lu;
							lu.push_back(eq_number);
							hgi->eq_labels[it->id] = lu;
						}
						else
						{
							hgi->eq_labels[it->id].push_back(eq_number);
						}
					}
					else
					{
						hgi->eq_labels[it->id].push_back(eq_number);
					}
					break; 
				}

        			// copy last info
        			hgi_last = hgi;
        			graph_id_last = graph_id;

        			graph_id = findLink(graph_id_last, linklevel);

        			// find linked graph
				hgi = findHypergraph(graph_id, linklevel-1);
			
				// check if the node is still there
				map<IDTYPE, VERTEX> vertices = hgi->hg.getVertices();
			
				if(vertices.count(it->id)==0)
				{
					if(hgi_last->eq_labels.count(it->id)==0)
					{
						list<unsigned> lu;
						lu.push_back(eq_number);
						hgi_last->eq_labels[it->id] = lu;
					} 
					else
					{
						hgi_last->eq_labels[it->id].push_back(eq_number); 
					} 
					break;
				}

        			linklevel--;
			}
		}

		eq_number++;
	}

	optimize();
	
}
catch(exception e)
{
	cout << "Error in EQUIVALENCE_CONSTRUCTOR\n";
	exit(1);
}

}

/** find in the level link of the graph with id */
unsigned int EQUIVALENCE_CONSTRUCTOR::findLink(unsigned int id, unsigned int level)
{
  for(list<HYPERGRAPH_INFO>::iterator it = (*m_list)[level].begin(); it!=(*m_list)[level].end(); it++)
  {
	if(it->id==id)
	 return it->link;
  }

  assert(false);
}

/** find hypergraph with id at the level */
HYPERGRAPH_INFO * EQUIVALENCE_CONSTRUCTOR::findHypergraph(unsigned int id, unsigned int level)
{
  for(list<HYPERGRAPH_INFO>::iterator it = (*m_list)[level].begin(); it!=(*m_list)[level].end(); it++)
  {
	if(it->id==id)
	 return &(*it);
  }

  assert(false);
}

/** optimize sample list by deleting the hypergraphs
		where no labels are marked
		and no labels will be marked at the next levels */
void EQUIVALENCE_CONSTRUCTOR::optimize()
{
	 for(int level = (*m_list).size() - 2; level >=0; level--)
	 {
		  list<int> to_delete;
		  
		  for(list<HYPERGRAPH_INFO>::iterator it = (*m_list)[level].begin(); it!=(*m_list)[level].end(); it++)
		  {
				if(isLinked(it->id, level))
					continue;

				if(it->eq_labels.size()==0)
					to_delete.push_back(it->id);
		  }

		  // to get no problems with deleted iterators let us make it so
		  for(list<int>::iterator itd = to_delete.begin(); itd!=to_delete.end(); itd++)
		  {
			for(list<HYPERGRAPH_INFO>::iterator it = (*m_list)[level].begin(); it!=(*m_list)[level].end(); it++)
			{
				if(it->id == *itd)
				{
					(*m_list)[level].erase(it);
					break;
				}
			}
		  }
				
	 }

	// delete empty sets
	int counter = 0;
	int count = (*m_list).size();
	for(vector<list<HYPERGRAPH_INFO> >::iterator it = (*m_list).begin(); it!=(*m_list).end(); it++)
	{
		if(it->size()==0)
		{
		        for(int i = 0; i<(count-counter); i++)
			 (*m_list).pop_back();

			break;
		}

		counter++;
	 }

	// now propagate labels
   	for(vector<list<HYPERGRAPH_INFO> >::iterator itVec = (*m_list).begin(); itVec!=(*m_list).end(); itVec++)
	{
		 vector<list<HYPERGRAPH_INFO> >::iterator itTmp = itVec;

		 itTmp++;

  		 if(itTmp==(*m_list).end())
			break;
	 
		 for(list<HYPERGRAPH_INFO>::iterator it1 = (*itVec).begin(); it1!=(*itVec).end(); it1++)
		 {
			  for(list<HYPERGRAPH_INFO>::iterator it2 = (*itTmp).begin(); it2!=(*itTmp).end(); it2++)
			  {
					if(it1->id!=it2->link)
					 continue;

					map<IDTYPE, VERTEX> nodes1=it1->hg.getVertices();
					map<IDTYPE, VERTEX> nodes2=it2->hg.getVertices();

					for(map<IDTYPE, VERTEX>::iterator itnode1 = nodes1.begin(); itnode1!=nodes1.end(); itnode1++)
					{
						if(nodes2.find(itnode1->first)!=nodes2.end())
							if(it1->eq_labels[itnode1->first].size()!=0)
								it2->eq_labels[itnode1->first]=it1->eq_labels[itnode1->first];
							
					}
        		  }
		 }
	 }
	 
	 cout << "optimized sample size: " << m_list->size() << endl;

	// only m_list->size()-1 transition is needed
	vector<string> trans_optimized;
	for(unsigned int i=0; i< m_list->size()-1; i++)
	 trans_optimized.push_back(sample->trans[i]);
	sample->trans = trans_optimized;
}

/** is the graph with id at the level linked from some graph at the level + 1 */
bool EQUIVALENCE_CONSTRUCTOR::isLinked(unsigned int id, unsigned int level)
{
	if(level>=(*m_list).size()-1)
	 return false;

	// last not empty level in m_list
	for(list<HYPERGRAPH_INFO>::iterator it = (*m_list)[level+1].begin(); it!= (*m_list)[level+1].end(); it++)
	{
		if(it->link==id)
		 return true;
	}

	return false;
}


/////////// namespaces ///////////////

}

}
