/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "example_checker.h"

#include "globals.h"
#include "find_all_matches.h"
#include "apply_rule.h"

#include "write_hypergraph.h"

namespace algorithms {

namespace refinement {

EXAMPLE_CHECKER::EXAMPLE_CHECKER()
{
	this->name = "example_checker";
}

EXAMPLE_CHECKER::~EXAMPLE_CHECKER()
{
}

void EXAMPLE_CHECKER::initialize(PETRIGRAPH * pg, vector<IDTYPE> & trans)
{
	  this->pg = pg;

	  this->trans = trans;
	  
	  counter = 0;
	  
	  ft.initialize((PTNET*)pg);
}

SAMPLE EXAMPLE_CHECKER::run(int * trans_num, bool * result)
{
  try
  {
  
  // [M_0] ... [M_n]
  SAMPLE m_list;
  
  *trans_num=0;

  //  map<IDTYPE, VERTEX> v = pg->getVertices();

  m_list.tree.resize(1);
  // start graph
  m_list.tree[0].push_back(HYPERGRAPH_INFO(*(pg->getGrammar()->getStartgraph()), counter, 0));

  counter++;

  unsigned int count = 0;
  
  Marking current_marking = pg->getInitialMarking().getUncoloured();
  
  while(1)
  {
    if(trans.size()==0)
    {
	*result = true;
	return m_list;
    }	

    // resize
    m_list.tree.resize(count+2);
    		
    checkLevel(count, current_marking, &(m_list.tree));

    TRANSITION t = pg->getTransition(trans[count]);
    
    if(!ft.isEnabledBy(current_marking, t.getID()))
    {
        cout << "internal error\n";
	cout << "counter-example could not be reproduced in the EXAMPLE_CHECKER\n";
	exit(1);
    }
    else
    {
    	current_marking = ft.run(t, current_marking);	
    }

    m_list.trans.push_back(t.getLabel());

    count++;

    if( (count==trans.size()) && (m_list.tree[count].size()!=0) )
    {
	*result = true;
	return m_list;
    }

    if(m_list.tree[count].size()==0)
    {
	*trans_num=count;
	*result = false;
	return m_list;
    }
    
   }
  
   return m_list;
 }
 catch(exception e)
 {
	cout << "Error in EXAMPLE_CHECKER.\n";
	exit(1);
 }
 
}

/** check [M_i] set */
void EXAMPLE_CHECKER::checkLevel(int level, Marking m, vector<list<HYPERGRAPH_INFO> > * m_list)
{
  try
  {
	GTS * ggr = pg->getGrammar();
	
	map<IDTYPE, IDTYPE> mergedEdges;

	// graph(M)
	HYPERGRAPH hmarking = pg->getMarkedHypergraph(m, &mergedEdges);

	// test
	// WRITE_HYPERGRAPH whm;
	// whm.initialize("testm.xml");
	// whm.run(hmarking);
	
	TRIVIAL_MATCHHANDLER mhandler;
	
	RULE * r = pg->getAssociatedRule(trans[level]);
	
//	IDTYPE id_rule = r->getID();

	// L
	HYPERGRAPH * hleft = r->getLeft();

	list<HYPERGRAPH_INFO>::iterator it = (*m_list)[level].begin();

	// must have at least one element
	assert(it!=(*m_list)[level].end());

	int size_m = (*m_list)[level].size();
		
	for(int i = 0; i<size_m; i++)
	{
		ALGORITHM * famLG = this->load("call_find_all");
		ALGORITHM * famGm = this->load("call_find_all");

		// matches from L to G
		famLG->initialize(&(it->hg));
		list<MATCH> matchesLG = famLG->run(hleft, &mhandler, false);

		// matches from G to graph(M)
		famGm->initialize(&hmarking);
		list<MATCH> matchesGgM = famGm->run(&(it->hg), &mhandler, true);
		
		// choosen matches L -> G
        	list<MATCH> matches;
		
		// check commuting
		for(list<MATCH>::iterator lg = matchesLG.begin(); lg!=matchesLG.end(); lg++)
		{
			bool commutes = false;
			for(list<MATCH>::iterator gm = matchesGgM.begin(); gm!=matchesGgM.end(); gm++)
			{
				MATCH mu_mapping = pg->getMu(trans[level]);
		
				if(checkCommuting(lg->getMatch(), gm->getMatch(), mu_mapping.getMatch(), mergedEdges))
				{
					commutes=true;
					break;
				}
			}

			if(commutes)
			{
				matches.push_back(*lg);
				commutes=false;
			}
		}

		// apply all matchings
		for(list<MATCH>::iterator it_match = matches.begin(); it_match!=matches.end(); it_match++)
		{
			APPLY_RULE ar;
			ar.initialize(&(it->hg));
			HYPERGRAPH h = ar.run(r, (*it_match));

			// test
			// WRITE_HYPERGRAPH wh;
			// wh.initialize("test.xml");
			// wh.run(h);
	
			HYPERGRAPH_INFO hi = HYPERGRAPH_INFO(h, counter, it->id);
			(*m_list)[level+1].push_back(hi);
			it->matchesLG = matches;
			counter++;
		}
	
		// if it is the last not empty element of m_list,
		// then initialize matchesGgM 
	//	if(m_list[level+1].size()==0)
	//	{
		for(list<MATCH>::iterator itGgMList = matchesGgM.begin();
					itGgMList!=matchesGgM.end(); itGgMList++)
		{
			map<IDTYPE, IDTYPE> matchGgM = itGgMList->getMatch();
			for(map<IDTYPE, IDTYPE>::iterator itGgM = matchGgM.begin();
							itGgM!=matchGgM.end(); itGgM++)
			{
				if(mergedEdges.find(itGgM->second)!=mergedEdges.end())
					itGgM->second=mergedEdges[itGgM->second];
			}
			
			itGgMList->setMatch(matchGgM);
		}
		
		it->matchesGgM = matchesGgM;
	//	}
	
	
		it++;
	}
}
catch(exception e)
{
	cout << "error while checking the level" << endl;
	throw e;
}

}

/** check if the diagram commutes */
bool EXAMPLE_CHECKER::checkCommuting(map<IDTYPE, IDTYPE> ab, map<IDTYPE, IDTYPE> bc,
					map<IDTYPE, IDTYPE> ac, map<IDTYPE, IDTYPE> mergedEdges)
{
	for(map<IDTYPE, IDTYPE>::iterator it = ab.begin(); it!=ab.end(); it++)
	{
		 IDTYPE vertedge;
		 map<IDTYPE, IDTYPE>::iterator itFind = mergedEdges.find(bc[it->second]);

		 if(itFind==mergedEdges.end())
			 vertedge = bc[it->second];
		 else
		   vertedge = itFind->second;
		 
		 if(vertedge!=ac[it->first])
		 	 return false;
	}

	 return true;
}

///////// namespaces //////////

}

}
