/***************************************************************************
 *   Copyright (C) 2006 by root   *
 *   root@Vitali   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "example_checker_col.h"

#include "globals.h"
#include "find_all_matches.h"
#include "apply_rule.h"

#include "write_hypergraph.h"
#include "algorithm_storage.h"
#include "binder.h"

namespace algorithms {

namespace refinement {

EXAMPLE_CHECKER_COL::EXAMPLE_CHECKER_COL()
{
	this->name = "example_checker_col";
	std_engine = NULL;
	abs_engine = NULL;
}


EXAMPLE_CHECKER_COL::~EXAMPLE_CHECKER_COL()
{
}

void EXAMPLE_CHECKER_COL::initialize(PETRIGRAPH * pg, vector<IDTYPE> & trans, vector<BINDING> & bindings)
{
	this->pg = (*pg); // copy 

	this->trans = trans;

	this->bindings = bindings;
	
	counter = 0;	
}

SAMPLE EXAMPLE_CHECKER_COL::run(int * trans_num, bool * result)
{
  try
  {
  
  // [M_0] ... [M_n]
  SAMPLE m_list;
  
  *trans_num=0;

  //  map<IDTYPE, VERTEX> v = pg->getVertices();

  m_list.tree.resize(1);
  // start graph
  m_list.tree[0].push_back(HYPERGRAPH_INFO(*(pg.getGrammar()->getStartgraph()), counter, 0));

  counter++;

  unsigned int count = 0;

  // engines
  std_engine = storage.getStdEngine();
  abs_engine = storage.getEngine();

  // list of markings
  list<MARKING> l_markings;
  l_markings.push_back(pg.getInitialMarking());

  // set initial marking to the petrigraph again from G_0 (without abstraction)
  MATCH iota = pg.getIota();

  map<IDTYPE, EDGE> hedges = pg.getGrammar()->getStartgraph()->getEdges();

  MARKING initial;

  for(map<IDTYPE, EDGE>::iterator ith = hedges.begin(); ith != hedges.end(); ith++)
  {
  	TOKEN t;
	t.setCount(1);
		
	list<EXPRESSION> le = ith->second.getAttributes();
	for(list<EXPRESSION>::iterator ita = le.begin(); ita != le.end(); ita++)
	 t.addAttributes(std_engine->run(*ita));	

	initial.putToken(ith->first, t);
  }

  pg.setInitialMarking(initial);

  // initialize FIRE_TRANSITION
  ft.initialize((PTNET*)&pg, storage.getStdEngine());

  while(1)
  {
    if(trans.size()==0)
    {
	*result = true;
	return m_list;
    }

    // resize
    m_list.tree.resize(count+2);

    checkLevel(count, l_markings, &(m_list.tree));

    TRANSITION t = pg.getTransition(trans[count]);

    list<MARKING> nl_markings;
    for(list<MARKING>::iterator it_ = l_markings.begin(); it_ != l_markings.end(); it_++)
    {
	list<BINDING> lb = ft.getEnabledBindings(*it_, t.getID());
	for(list<BINDING>::iterator itb = lb.begin(); itb != lb.end(); itb++)
	{
		BINDING b = abs_engine->abstractBinding(*itb);
		if(BINDING::compareBindingsAbs(b, bindings[count]))
		 nl_markings.push_back(ft.run(t, *it_, *itb));
	}
    }

    // this is a case of attributes, is denoted by the -transition
    if(nl_markings.size() == 0)
    {
	// return empty sample
	SAMPLE m_list_empty;
	count++;
        *trans_num=-count;
	*result = false;
	return m_list_empty;
    }

    l_markings = nl_markings;

    m_list.trans.push_back(t.getLabel());

    count++;

    if( (count==trans.size()) && (m_list.tree[count].size()!=0) )
    {
	*result = true;
	return m_list;
    }

    if(m_list.tree[count].size()==0)
    {
	*trans_num=count;
	*result = false;
	return m_list;
    }
    
   }
  
   return m_list;
 }
 catch(exception e)
 {
	cout << "Error in EXAMPLE_CHECKER.\n";
	exit(1);
 }
 
}

/** check [M_i] set */
void EXAMPLE_CHECKER_COL::checkLevel(int level, list<MARKING> lm, vector<list<HYPERGRAPH_INFO> > * m_list)
{
  try
  {
	GTS * ggr = pg.getGrammar();
	
	map<IDTYPE, IDTYPE> mergedEdges;

	// graph(M)
	vector<HYPERGRAPH> vhmarking;
	for(list<MARKING>::iterator itl = lm.begin(); itl != lm.end(); itl++)
	{
		HYPERGRAPH h = pg.getMarkedHypergraph(*itl, &mergedEdges);
		vhmarking.push_back(h);
	}

	assert(vhmarking.size() != 0);

	TRIVIAL_MATCHHANDLER mhandler;
	
	RULE * r = pg.getAssociatedRule(trans[level]);
	
	// L
	HYPERGRAPH * hleft = r->getLeft();

	list<HYPERGRAPH_INFO>::iterator it = (*m_list)[level].begin();

	// must have at least one element
	assert(it!=(*m_list)[level].end());

	int size_m = (*m_list)[level].size();
		
	for(int i = 0; i<size_m; i++)
	{
		ALGORITHM * famLG = this->load("call_find_all");
		ALGORITHM * famGm = this->load("call_find_all");

		// matches from L to G
		famLG->initialize(&(it->hg));
		list<MATCH> matchesLG = famLG->run(hleft, &mhandler, false);
		
		// matches from G to graph(M)
		famGm->initialize(&(vhmarking[0]));
		list<MATCH> matchesGgM_ = famGm->run(&(it->hg), &mhandler, true);

		// choosen matches from G to graph(M)
		list<MATCH> matchesGgM;

		// check attributes in G -> graph(M)
		for(list<MATCH>::iterator lGM = matchesGgM_.begin(); lGM!=matchesGgM_.end(); lGM++)
		{
			// graph(m) ,graph(m'), ...
			for(vector<HYPERGRAPH>::iterator it_gm = vhmarking.begin(); it_gm != vhmarking.end(); it_gm++)
			{
				bool corresponds = true;

				// match
				map<IDTYPE, IDTYPE> match = lGM->getMatch();
				for(map<IDTYPE, IDTYPE>::iterator itm = match.begin(); itm != match.end(); itm++)
				{
					if((it->hg).hasEdge(itm->first))
					{
						list<EXPRESSION> attr_G = (it->hg).getEdge(itm->first).getAttributes();
						list<EXPRESSION> attr_gM = it_gm->getEdge(itm->second).getAttributes();
	
						// if(!(EXPRESSION::equal(attr_G, attr_gM)))
						if(!(abs_engine->smallereq(attr_G, attr_gM)))
						{
							corresponds = false;
							break;
						}
					}
				}

				if(corresponds)
				{
					matchesGgM.push_back(*lGM);
					break;
				}
			}
		}
		
		// commutating matches L -> G
        	list<MATCH> matches_;
		
		// check commuting
		for(list<MATCH>::iterator lg = matchesLG.begin(); lg!=matchesLG.end(); lg++)
		{
			bool commutes = false;
			for(list<MATCH>::iterator gm = matchesGgM.begin(); gm!=matchesGgM.end(); gm++)
			{
				MATCH mu_mapping = pg.getMu(trans[level]);	
		
				if(checkCommuting(lg->getMatch(), gm->getMatch(), mu_mapping.getMatch(), mergedEdges))
				{
					commutes=true;
					break;
				}
			}

			if(commutes)
			{
				matches_.push_back(*lg);
				commutes=false;
			}
		}

		// matches to apply
		list<MATCH> matches;
		list<BINDING> m_bindings;

		// check attributes in L -> G
		for(list<MATCH>::iterator lm = matches_.begin(); lm!=matches_.end(); lm++)
		{
			BINDER binder;
			binder.initialize(&(it->hg));
			list<BINDING> lb = binder.run(r->getLeft(), &(*lm));

			assert(lb.size() == 1);

			BINDING b = abs_engine->abstractBinding(lb.front());
			if(BINDING::compareBindingsAbs(b, bindings[level]))
			{
				matches.push_back(*lm);
				m_bindings.push_back(lb.front());
			}
		}

		// apply all matchings
		list<BINDING>::iterator itb = m_bindings.begin();
		for(list<MATCH>::iterator it_match = matches.begin(); it_match!=matches.end(); it_match++)
		{
			APPLY_RULE ar;
			ar.initialize(&(it->hg), std_engine);

			if(ar.IsAppliable(r, &(*itb)))
			{
				HYPERGRAPH h = ar.run(r, (*it_match), *itb);
	
				HYPERGRAPH_INFO hi = HYPERGRAPH_INFO(h, counter, it->id);
				(*m_list)[level+1].push_back(hi);
				it->matchesLG.push_back(*it_match);
				counter++;
			}

			itb++;
		}
	
		// initialize matchesGgM 
		for(list<MATCH>::iterator itGgMList = matchesGgM.begin();
					itGgMList!=matchesGgM.end(); itGgMList++)
		{
			map<IDTYPE, IDTYPE> matchGgM = itGgMList->getMatch();
			for(map<IDTYPE, IDTYPE>::iterator itGgM = matchGgM.begin();
							itGgM!=matchGgM.end(); itGgM++)
			{
				if(mergedEdges.find(itGgM->second)!=mergedEdges.end())
					itGgM->second=mergedEdges[itGgM->second];
			}
			
			itGgMList->setMatch(matchGgM);
		}
		
		it->matchesGgM = matchesGgM;
	
		it++;
	}
}
catch(exception e)
{
	cout << "error while checking the level" << endl;
	throw e;
}

}


/** check if the diagram commutes */
bool EXAMPLE_CHECKER_COL::checkCommuting(map<IDTYPE, IDTYPE> ab, map<IDTYPE, IDTYPE> bc,
					map<IDTYPE, IDTYPE> ac, map<IDTYPE, IDTYPE> mergedEdges)
{
	for(map<IDTYPE, IDTYPE>::iterator it = ab.begin(); it!=ab.end(); it++)
	{
		 IDTYPE vertedge;
		 map<IDTYPE, IDTYPE>::iterator itFind = mergedEdges.find(bc[it->second]);

		 if(itFind==mergedEdges.end())
			 vertedge = bc[it->second];
		 else
		   vertedge = itFind->second;
		 
		 if(vertedge!=ac[it->first])
		 	 return false;
	}

	 return true;
}

// namespaces

}

}


