%{
#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <vector>

#include "expr.h"

extern int ttlex();
extern int ttdebug;

extern "C"
{
   int ttwrap(){
       return 1; 
   }
   extern int ttparse();

   extern void tterror(const char* s)
   {
      fprintf(stderr, "%s\n", s);
   }

   using namespace std;
}

vector<ExpressionElement> exprtree;

%}

%token EQ NE GE LE LT GT
%token PLUS MINUS MULT DIV MOD
%token TRUE FALSE
%token NOT AND OR
%token COMMA 
%token VARIABLE
%token INT_VALUE
%token STR_VALUE
%token LEFT RIGHT
%token LEN
%token CONC
%token ECKIG_L ECKIG_R

%left AND OR 
%nonassoc NOT
%left EQ NE GE LE LT GT EQUALS
%left CONC
%left MINUS PLUS      /* left associative and lowest precedence */
%left MULT DIV MOD    /* higher level than + and - */
%nonassoc UMINUS   /* no associativity for unariy - sign but hightest level */

%start expression

%union {
int integer;
string * pstr;
ExpressionElement* element;
}

%type <integer> INT_VALUE
%type <pstr> VARIABLE
%type <pstr> STR_VALUE
%type <element> expression expr_num expr_bool expr_str expr_num_ expr_str_

%%

expression : ECKIG_L expression ECKIG_R {
	}
	|  expression COMMA expr_num_ {		
		exprtree.push_back(*$3);
	}	
	| expr_num_ {
		exprtree.push_back(*$1);
	}
	| expression COMMA expr_bool {		
		exprtree.push_back(*$3);
	}	
	| expr_bool {
		exprtree.push_back(*$1);
	}
	| expression COMMA expr_str_ {		
		exprtree.push_back(*$3);
	}	
	| expr_str_ {
		exprtree.push_back(*$1);
	}
	| expression COMMA VARIABLE {
		ExpressionElement* a = new ExpressionElement;
		a->type = VAR_UNKNOWN;
		a->variable_name = *$3;
		exprtree.push_back(*a);
	}
	| VARIABLE {
		ExpressionElement* a = new ExpressionElement;
		a->type = VAR_UNKNOWN;
		a->variable_name = *$1;
		exprtree.push_back(*a);
	}
	| { tterror("empty expression"); }
;

expr_num_  : INT_VALUE {
		ExpressionElement* a = new ExpressionElement;
		a->type = NUMBER;
		a->int_val = $1;
		$$ = a;
	}
	| expr_num PLUS expr_num {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_NUM_PLUS;
		a->elements.push_back(*$1);
		a->elements.push_back(*$3);
		$$ = a;
		}
	| expr_num MINUS expr_num {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_NUM_MINUS;
		a->elements.push_back(*$1);
		a->elements.push_back(*$3);
		$$ = a;
		}	
	| expr_num MULT expr_num {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_NUM_MULT;
		a->elements.push_back(*$1);
		a->elements.push_back(*$3);
		$$ = a;
		}
	| expr_num DIV expr_num {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_NUM_DIV;
		a->elements.push_back(*$1);
		a->elements.push_back(*$3);
		$$ = a;
		}
	| expr_num MOD expr_num {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_NUM_MOD;
		a->elements.push_back(*$1);
		a->elements.push_back(*$3);
		$$ = a;
		}
	| LEFT expr_num RIGHT {
		$$ = $2;
		}
	| LEN LEFT expr_str RIGHT {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_STR_LEN;
		a->elements.push_back(*$3);		
		$$ = a;
		}
	| MINUS expr_num %prec UMINUS {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_NUM_UMINUS;
		a->elements.push_back(*$2);		
		$$ = a;
		}	
;

expr_num  : INT_VALUE {
		ExpressionElement* a = new ExpressionElement;
		a->type = NUMBER;
		a->int_val = $1;
		$$ = a;
	}
	| VARIABLE {
		ExpressionElement* a = new ExpressionElement;
		a->type = VAR_NUM;
		a->variable_name = *$1;
		$$ = a;
	}
	| expr_num PLUS expr_num {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_NUM_PLUS;
		a->elements.push_back(*$1);
		a->elements.push_back(*$3);
		$$ = a;
		}
	| expr_num MINUS expr_num {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_NUM_MINUS;
		a->elements.push_back(*$1);
		a->elements.push_back(*$3);
		$$ = a;
		}	
	| expr_num MULT expr_num {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_NUM_MULT;
		a->elements.push_back(*$1);
		a->elements.push_back(*$3);
		$$ = a;
		}
	| expr_num DIV expr_num {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_NUM_DIV;
		a->elements.push_back(*$1);
		a->elements.push_back(*$3);
		$$ = a;
		}
	| expr_num MOD expr_num {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_NUM_MOD;
		a->elements.push_back(*$1);
		a->elements.push_back(*$3);
		$$ = a;
		}
	| LEFT expr_num RIGHT {
		$$ = $2;
		}
	| LEN LEFT expr_str RIGHT {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_STR_LEN;
		a->elements.push_back(*$3);		
		$$ = a;
		}
	| MINUS expr_num %prec UMINUS {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_NUM_UMINUS;
		a->elements.push_back(*$2);		
		$$ = a;
		}	
;

expr_bool  : TRUE {
		ExpressionElement* a = new ExpressionElement;
		a->type = BOOLEAN;
		a->bool_val = TRUE_VAL;
		$$ = a;
	}
	| FALSE {
		ExpressionElement* a = new ExpressionElement;
		a->type = BOOLEAN;
		a->bool_val = FALSE_VAL;
		$$ = a;
	}
	| expr_bool AND expr_bool {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_BOOL_AND;
		a->elements.push_back(*$1);
		a->elements.push_back(*$3);
		$$ = a;
		}
	| expr_bool OR expr_bool {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_BOOL_OR;
		a->elements.push_back(*$1);
		a->elements.push_back(*$3);
		$$ = a;
		}
	| NOT expr_bool {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_BOOL_NOT;
		a->elements.push_back(*$2);		
		$$ = a;
		}
	| expr_num EQ expr_num {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_NUM_BOOL_EQ;
		a->elements.push_back(*$1);
		a->elements.push_back(*$3);
		$$ = a;
		}
	| expr_num LE expr_num {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_NUM_BOOL_LE;
		a->elements.push_back(*$1);
		a->elements.push_back(*$3);
		$$ = a;
		}	
	| expr_num GE expr_num {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_NUM_BOOL_GE;
		a->elements.push_back(*$1);
		a->elements.push_back(*$3);
		$$ = a;
		}	
	| expr_num NE expr_num {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_NUM_BOOL_NE;
		a->elements.push_back(*$1);
		a->elements.push_back(*$3);
		$$ = a;
		}
	| expr_num LT expr_num {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_NUM_BOOL_L;
		a->elements.push_back(*$1);
		a->elements.push_back(*$3);
		$$ = a;
		}	
	| expr_num GT expr_num {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_NUM_BOOL_G;
		a->elements.push_back(*$1);
		a->elements.push_back(*$3);
		$$ = a;
		}
	| expr_str EQUALS expr_str {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_STR_BOOL_EQ;
		a->elements.push_back(*$1);
		a->elements.push_back(*$3);
		$$ = a;
		}	
	| LEFT expr_bool RIGHT {
		$$ = $2;
		}
;

expr_str_  : STR_VALUE {
		ExpressionElement* a = new ExpressionElement;
		a->type = _STR_;
		a->str_val = *$1;
		$$ = a;
	}
	| expr_str CONC expr_str {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_STR_CONC;
		a->elements.push_back(*$1);
		a->elements.push_back(*$3);
		$$ = a;
	}
;

expr_str  : STR_VALUE {
		ExpressionElement* a = new ExpressionElement;
		a->type = _STR_;
		a->str_val = *$1;
		$$ = a;
	}
	| expr_str CONC expr_str {
		ExpressionElement* a = new ExpressionElement;
		a->type = OP_STR_CONC;
		a->elements.push_back(*$1);
		a->elements.push_back(*$3);
		$$ = a;
	}
	| VARIABLE {
		ExpressionElement* a = new ExpressionElement;
		a->type = VAR_STR;
		a->variable_name = *$1;
		$$ = a;
	}	
;

%%


