/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "expression.h"
#include <string.h>

extern "C"
{
  extern int ttparse();
  extern char *ttstrin;
}

extern vector<ExpressionElement> exprtree;

namespace attributes
{

   EXPRESSION::EXPRESSION()
   {
      this->name="";
   }

   EXPRESSION::EXPRESSION(TYPES type, string expr, string name)
   {
      this->type=type;
      this->expr=expr;
      this->name=name;
   }

   EXPRESSION EXPRESSION::clone()
   {
      EXPRESSION e;
      e.expr=this->expr;
      e.type=this->type;
      e.name=this->name;
      return e;
   }

   void EXPRESSION::addName(string name)
   {
      this->name=name;
   }

   void EXPRESSION::addExpr(string expr)
   {
      this->expr=expr;

      ttstrin = new char[256];
      strcpy(ttstrin, (char*)expr.c_str());

      // clear global tree
      exprtree.clear();

      ttparse();

      expr_parsed = exprtree;
   }

   void EXPRESSION::addType(TYPES type)
   {
      this->type=type;
   }

   TYPES EXPRESSION::getTYPES()
   {
      return this->type;
   }

   string EXPRESSION::getExpr()
   {
      return this->expr;
   }

   string EXPRESSION::getName()
   {
      return this->name;
   }

   ostream &operator<<(ostream &ost, EXPRESSION &expr)
   {
      ost << expr.name << ": ";
	 
/*    vector<T>::iterator it=expr.type.begin();
      if (it !=  expr.type.end()) {
         switch (*it) {
            case INTEGER: {
               ost <<"INTEGER";
	       break;
            }
            case STRING: {
               ost <<"STRING";
               break;
            }
            case BOOL: {
               ost <<"BOOL";
               break;
            }
            default: assert(false);
         }
         it++;
      }
      while(it != expr.type.end())
      {
         ost <<'-';
         switch (*it) {
            case INTEGER: {
               ost <<"INTEGER";
	       break;
            }
            case STRING: {
               ost <<"STRING";
               break;
            }
            case BOOL: {
               ost <<"BOOL";
               break;
            }
            default: assert(false);
         }
         it++;
      }
      ost <<" (" <<expr.expr <<')';

*/

	vector<ExpressionElement> expr_parsed = expr.getExprParsed();
	for(vector<ExpressionElement>::iterator ite = expr_parsed.begin(); ite != expr_parsed.end(); ite++)
		ost << "string_eval=" << ite->str_evaluated << " int_eval=" << ite->int_evaluated;	
	
      return ost;
   }

   EXPRESSION::~EXPRESSION()
   {
   }

   vector<ExpressionElement> & EXPRESSION::getExprParsed()
   {
	return this->expr_parsed;
   }

   void EXPRESSION::addExprParsed(vector<ExpressionElement> vee)
   {
	this->expr_parsed = vee;
   }

   void EXPRESSION::write(ExpressionElement * el)
   {
	cout << "Type: " << el->type << endl;
	cout << "Int_Value: " << el->int_val << endl;
	cout << "Str_Value: " << el->str_val << endl;
	cout << "---------------" << endl;

	for(vector<ExpressionElement>::iterator it = el->elements.begin();
		it != el->elements.end(); it++)
	{
		write((ExpressionElement*)(&(*it)));
	}
   }

   void EXPRESSION::writeParsed()
   {
	for(vector<ExpressionElement>::iterator it = expr_parsed.begin(); it != expr_parsed.end(); it++)
	{
		write(&(*it));
		cout << "..............\n";
	}
   }

   bool EXPRESSION::equal(list<EXPRESSION> l1, list<EXPRESSION> l2)
   {
	if(l1.size() != l2.size())
		return false;

	list<EXPRESSION>::iterator it1 = l1.begin();
	list<EXPRESSION>::iterator it2 = l2.begin();

	while(it1 != l1.end())
	{
		if(!EXPRESSION::equal(*it1, *it2))
			return false;

		it1++;
		it2++;
	}

	return true;
   }
 
   bool EXPRESSION::equal(EXPRESSION e1, EXPRESSION e2)
   {
	vector<ExpressionElement> vee1 = e1.getExprParsed();
	vector<ExpressionElement> vee2 = e2.getExprParsed();

	if(vee1.size() != vee2.size())
		return false;

	vector<ExpressionElement>::iterator it1 = vee1.begin();
	vector<ExpressionElement>::iterator it2 = vee2.begin();

	while(it1 != vee1.end())
	{
		if( (it1->type != it2->type) ||
			(it1->bool_val != it2->bool_val) ||
			(it1->str_evaluated != it2->str_evaluated) ||
			(it1->int_evaluated != it2->int_evaluated)	
		  )
			return false;

		it1++;
		it2++;
	}

	return true;
   }

   bool EXPRESSION::equal(ExpressionElement ee1, ExpressionElement ee2)
   {
	if( (ee1.type != ee2.type) ||
		(ee1.bool_val != ee2.bool_val) ||
		(ee1.str_evaluated != ee2.str_evaluated) ||
		(ee1.int_evaluated != ee2.int_evaluated)	
	  )
		return false;

	return true;
   }

   string EXPRESSION::getTypesString()
   {
	if(type.size()==0)
	 return "";

	string result = "attr";

	for(TYPES::iterator it = type.begin(); it != type.end(); it++)
	{
		switch(*it)
		{
			case INTEGER:
			{
			 result += "-int";
			 break;
			}
			case STRING:
			{
			 result += "-str";
			 break;
			}
			case BOOL:
			{
			 result += "-bool";
			 break;
			}
		}
	}

	return result;
   }

};

