/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "expression_engine.h"

namespace attributes {

EXPRESSION_ENGINE::EXPRESSION_ENGINE()
{
 pn = NULL;
}


EXPRESSION_ENGINE::~EXPRESSION_ENGINE()
{
}

void EXPRESSION_ENGINE::initialize(BINDING pb)
{
	this->pb = pb;

	// evaluate all values
	list<string> names = this->pb.getVariableNames();

	for(list<string>::iterator it = names.begin(); it != names.end(); it++)
	{
		ExpressionElement ee = this->pb.getVariable(*it);
		evaluate(ee);
		this->pb.setVariable(*it, ee);
	}
}

EXPRESSION EXPRESSION_ENGINE::run(EXPRESSION & e)
{
/*	if(pb==NULL)
	{
		cout << "\nError while evaluating expression (No binding set)\n";
		exit(1);
	}
*/
	// result expression
	EXPRESSION re = e;

	// result expression parsed
	vector<ExpressionElement> rv;

	// expression parsed
	vector<ExpressionElement> vee = e.getExprParsed();

	for(vector<ExpressionElement>::iterator it = vee.begin();
		it != vee.end(); it++)
	{
		evaluate(*it);

		// copy the result
		ExpressionElement ee;
		ee.str_val = it->str_evaluated;
		ee.int_val = it->int_evaluated;
		ee.bool_val = it->bool_val;
		ee.str_evaluated = it->str_evaluated;
		ee.int_evaluated = it->int_evaluated;

		// is abstract also for the standard engine, because of problems with variables
		// other solution would be to make abstract only in case of variables
		ee.is_abstract = true;
		
		switch(it->type)
		{
			case OP_NUM_PLUS:
			case OP_NUM_MINUS:
			case OP_NUM_MULT:
			case OP_NUM_DIV:
			case OP_NUM_MOD:
			case NUMBER:
			case OP_STR_LEN:
			case OP_NUM_UMINUS:
			case VAR_NUM:
			{
				ee.type = NUMBER;
				break;
			}


			case OP_BOOL_AND:
			case OP_BOOL_OR:
			case OP_NUM_BOOL_EQ:
			case OP_NUM_BOOL_LE:
			case OP_NUM_BOOL_GE:
			case OP_NUM_BOOL_NE:
			case OP_NUM_BOOL_L:
			case OP_NUM_BOOL_G:
			case OP_STR_BOOL_EQ:
			case BOOLEAN:
			{
				ee.type = BOOLEAN;
				break;

			}
		
			case _STR_:
			case OP_STR_CONC:
			case VAR_STR:
			{
				ee.type = _STR_;
				break;

			}

			case VAR_UNKNOWN:
			{
				ee.type = pb.getVariable(it->variable_name).type;
				break;
			}
			default:
			{
				cout << "\ninternal error in EXPRESSION_ENGINE: Unknown EXPRESSION\n";
				exit(1);
			}
		}

		rv.push_back(ee);
	}

	re.addExprParsed(rv);

	return re;
}

void EXPRESSION_ENGINE::evaluate(ExpressionElement & e)
{
	if(e.is_abstract==true)
	 return;

	switch(e.type)
	{
		case OP_NUM_PLUS:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			e.int_evaluated = e.elements[0].int_evaluated + e.elements[1].int_evaluated;
			e.int_val = e.elements[0].int_evaluated + e.elements[1].int_evaluated;
			break;
		}
		case OP_NUM_MINUS:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			e.int_evaluated = e.elements[0].int_evaluated - e.elements[1].int_evaluated;
			e.int_val = e.elements[0].int_evaluated - e.elements[1].int_evaluated;
			break;
		}
		case OP_NUM_MULT:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			e.int_evaluated = e.elements[0].int_evaluated * e.elements[1].int_evaluated;
			e.int_val = e.elements[0].int_evaluated * e.elements[1].int_evaluated;
			break;
		}
		case OP_NUM_DIV:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			e.int_evaluated = e.elements[0].int_evaluated / e.elements[1].int_evaluated;
			e.int_val = e.elements[0].int_evaluated / e.elements[1].int_evaluated;
			break;
		}
		case OP_NUM_MOD:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			e.int_evaluated = e.elements[0].int_evaluated % e.elements[1].int_evaluated;
			e.int_val = e.elements[0].int_evaluated % e.elements[1].int_evaluated;
			break;
		}
		case NUMBER:
		{
			checkArgumentNumber(e.elements, 0);
			e.int_evaluated = e.int_val;
			break;
		}
		case VAR_NUM:
		{
			checkArgumentNumber(e.elements, 0);
			ExpressionElement ee = pb.getVariable(e.variable_name);
			e.int_evaluated = ee.int_evaluated;
			e.int_val = ee.int_evaluated;
			break;
		}
		case OP_STR_LEN:
		{
			checkArgumentNumber(e.elements, 1);
			e.int_evaluated = e.elements[0].str_val.size();
			e.int_val = e.elements[0].str_val.size();
			break;
		}
		case OP_NUM_UMINUS:
		{
			checkArgumentNumber(e.elements, 1);
			evaluate(e.elements[0]);
			e.int_evaluated = -e.elements[0].int_evaluated;
			e.int_val = -e.elements[0].int_val;
			break;
		}

		case OP_BOOL_AND:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			e.bool_val = ExpressionElement::logic_3(e.elements[0].bool_val, e.elements[1].bool_val, OP_BOOL_AND);
			break;
		}
		case OP_BOOL_OR:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			e.bool_val = ExpressionElement::logic_3(e.elements[0].bool_val, e.elements[1].bool_val, OP_BOOL_OR);
			break;
		}
		case OP_BOOL_NOT:
		{
			checkArgumentNumber(e.elements, 1);
			evaluate(e.elements[0]);
			e.bool_val = ExpressionElement::logic_3(e.elements[0].bool_val, e.elements[0].bool_val, OP_BOOL_NOT);
			break;
		}
		case OP_NUM_BOOL_EQ:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			if(e.elements[0].int_evaluated == e.elements[1].int_evaluated)
				e.bool_val = TRUE_VAL;
			else
				e.bool_val = FALSE_VAL;
			break;
		}
		case OP_NUM_BOOL_LE:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			if(e.elements[0].int_evaluated <= e.elements[1].int_evaluated)
				e.bool_val = TRUE_VAL;
			else
				e.bool_val = FALSE_VAL;
			break;
		}
		case OP_NUM_BOOL_GE:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			if(e.elements[0].int_evaluated >= e.elements[1].int_evaluated)
				e.bool_val = TRUE_VAL;
			else
				e.bool_val = FALSE_VAL;
			break;
		}
		case OP_NUM_BOOL_NE:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			if(e.elements[0].int_evaluated != e.elements[1].int_evaluated)
				e.bool_val = TRUE_VAL;
			else
				e.bool_val = FALSE_VAL;
			break;
		}
		case OP_NUM_BOOL_L:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			if(e.elements[0].int_evaluated < e.elements[1].int_evaluated)
				e.bool_val = TRUE_VAL;
			else
				e.bool_val = FALSE_VAL;
			break;
		}
		case OP_NUM_BOOL_G:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			if(e.elements[0].int_evaluated > e.elements[1].int_evaluated)
				e.bool_val = TRUE_VAL;
			else
				e.bool_val = FALSE_VAL;
			break;
		}
		case BOOLEAN:
		{
			checkArgumentNumber(e.elements, 0);
			break;
		}
		
		case _STR_:
		{
			checkArgumentNumber(e.elements, 0);
			e.str_evaluated = e.str_val;
			break;
		}
		case VAR_STR:
		{
			checkArgumentNumber(e.elements, 0);
			ExpressionElement ee = pb.getVariable(e.variable_name);
			e.str_val = ee.str_val;
			e.str_evaluated = ee.str_evaluated;
			break;
		}
		case OP_STR_CONC:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			e.str_evaluated = e.elements[0].str_evaluated + e.elements[1].str_evaluated;
			e.str_val = e.elements[0].str_evaluated + e.elements[1].str_evaluated;
			break;
		}
		case OP_STR_BOOL_EQ:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			if(e.elements[0].str_evaluated == e.elements[1].str_evaluated)
				e.bool_val = TRUE_VAL;
			else
				e.bool_val = FALSE_VAL;
			break;
		}
		case VAR_UNKNOWN:
		{
			checkArgumentNumber(e.elements, 0);
			ExpressionElement ee = pb.getVariable(e.variable_name);
			e.str_val = ee.str_val;
			e.int_val = ee.int_val;
			e.str_evaluated = ee.str_evaluated;
			e.int_evaluated = ee.int_evaluated;
			e.bool_val = ee.bool_val;
			break;
		}
		default:
		{
			cout << "\ninternal error in EXPRESSION_ENGINE: unknown EXPRESSION\n";
			exit(1);
		}
	}
}

string EXPRESSION_ENGINE::run(EXPRESSION & e, PRINT_TYPE type)
{
	vector<ExpressionElement> vee = e.getExprParsed();

	string s = string("");

	if(vee.size()>1)
	 s += "[";

	for(vector<ExpressionElement>::iterator it = vee.begin();
		it != vee.end(); it++)
	{
		if(s.size()>1)
		{
			s += string(",");
			s += run((ExpressionElement&)(*it), type);
		}
		else
			s += run((ExpressionElement&)(*it), type);
	}

	if(vee.size()>1)
	 s += "]";

	return s;
}

string EXPRESSION_ENGINE::run(ExpressionElement &e, PRINT_TYPE type)
{
	switch(e.type)
	{
		case OP_NUM_PLUS:
		case OP_NUM_MINUS:
		case OP_NUM_MULT:
		case OP_NUM_DIV:
		case OP_NUM_MOD:
		case NUMBER:
		case VAR_NUM:
		case OP_STR_LEN:
		case OP_NUM_UMINUS:
		{
			char ch[256];

			if(type==EVALUATED)
			{
				sprintf(ch, "%i", e.int_evaluated);
				return string(ch);
			}
			else
			{				
				sprintf(ch, "%i", e.int_val);
				return string(ch);
			}
		}

		case OP_BOOL_AND:
		case OP_BOOL_OR:
		case OP_NUM_BOOL_EQ:
		case OP_NUM_BOOL_LE:
		case OP_NUM_BOOL_GE:
		case OP_NUM_BOOL_NE:
		case OP_NUM_BOOL_L:
		case OP_NUM_BOOL_G:
		case OP_STR_BOOL_EQ:
		case BOOLEAN:
		{
			string bresult;

			if(e.bool_val)
				bresult = "TRUE";
			else
				bresult = "FALSE";

			return bresult;
		}
		
		case _STR_:
		case VAR_STR:
		case OP_STR_CONC:
		{
			if(type==EVALUATED)
				return string("'") + e.str_evaluated + string("'");
			else
				return string("'") + e.str_val + string("'");
		}

		case VAR_UNKNOWN:
		{
			checkArgumentNumber(e.elements, 0);
			ExpressionElement ee = pb.getVariable(e.variable_name);
			return run(ee, type);
		}
		default:
		{
			cout << "\ninternal error in EXPRESSION_ENGINE: unknown EXPRESSION\n";
			exit(1);
		}
	}

	return "";
}

// check guard
bool EXPRESSION_ENGINE::run(list<EXPRESSION> guards)
{
	// guard should be boolean
	for(list<EXPRESSION>::iterator ite = guards.begin();
		ite != guards.end(); ite++)
	{
		vector<ExpressionElement> vee = ite->getExprParsed();

		// check, that the values are boolean
		for(vector<ExpressionElement>::iterator it = vee.begin();
			it != vee.end(); it++)
		{
			if( it->type != OP_BOOL_AND &&
			it->type != OP_BOOL_OR &&
			it->type != OP_BOOL_NOT &&
			it->type != OP_NUM_BOOL_EQ &&
			it->type != OP_NUM_BOOL_LE &&
			it->type != OP_NUM_BOOL_GE &&
			it->type != OP_NUM_BOOL_NE &&
			it->type != OP_NUM_BOOL_L &&
			it->type != OP_NUM_BOOL_G &&
			it->type != OP_STR_BOOL_EQ &&
			it->type != BOOLEAN)
			{
				cout << "\nerror: wrong type of guard function\n";
				exit(1);
			}

			evaluate(*it);
			
			if(it->bool_val == FALSE_VAL)
				return false;
		}
	}
	
	return true;
}

void EXPRESSION_ENGINE::checkArgumentNumber(vector<ExpressionElement> & ve, unsigned int number)
{
	if(ve.size()!=number)
	{
		cout << "\ninternal error due to parsed expression\n";
		exit(1);
	}
}

bool EXPRESSION_ENGINE::refine()
{
	return false;
}

bool EXPRESSION_ENGINE::refine(COUNTER_EXAMPLE * pe)
{
	return refine();
}

void EXPRESSION_ENGINE::initialize(PTNET * i_ptnet)
{
	this->pn = i_ptnet;
}

bool EXPRESSION_ENGINE::run(IDTYPE trans_id, TABLE * * table)
{	
	if(expr_tables.count(trans_id)==0)
	{
	 if(IsColoured(trans_id))
	  CalculateTable(trans_id);
	 else
	  return false;
	}

	*table = &(expr_tables[trans_id]);

	return true;
}

void EXPRESSION_ENGINE::CalculateTable(IDTYPE trans_id)
{	
	if(pn==NULL)
	{
	 cout << "Programming Error: Petri net is not set in the ENGINE\n";
	 exit(1);
	}

	// preset possible values
	vector<list<vector<vector<ExpressionElement> > > > possible_values;

	vector<TCONNECTION> vc = pn->getPreset(trans_id);

	for(vector<TCONNECTION>::iterator it1 = vc.begin(); it1 != vc.end(); it1++)
	{
		list<vector<vector<ExpressionElement> > > list_expr;

		for(list<EXPRESSION>::iterator it2 = it1->attributes.begin(); it2 != it1->attributes.end(); it2++)
		{
			vector<vector<ExpressionElement> > v_eel;

			TYPES types = it2->getTYPES();

			for(vector<T>::iterator it3 = types.begin(); it3 != types.end(); it3++)
				v_eel.push_back(GetPossibleValues(*it3));
			

			list_expr.push_back(v_eel);
		}

		possible_values.push_back(list_expr);
	}

	// current configuration
	vector<list<vector<pair<long, vector<ExpressionElement> > > > > current_configuration;

	// fill current configuration with start and with possible values
	for(vector<list<vector<vector<ExpressionElement> > > >::iterator it_1 = possible_values.begin();
		it_1 != possible_values.end(); it_1++)
	{
		list<vector<pair<long, vector<ExpressionElement> > > > l_current;
		for(list<vector<vector<ExpressionElement> > >::iterator it_2 = it_1->begin(); it_2 != it_1->end(); it_2++)
		{
			vector<pair<long, vector<ExpressionElement> > > v_current;
			for(vector<vector<ExpressionElement> >::iterator it_3 = it_2->begin(); it_3 != it_2->end(); it_3++)
			{
				pair<long, vector<ExpressionElement> > p;
				p.first = 0;
				p.second = *it_3;

				// add value
				v_current.push_back(p);
			}

			// add value
			l_current.push_back(v_current);
		}

		// add to the current configuration
		current_configuration.push_back(l_current);
	}
	
	bool res = true;
	
	// fill table
	TABLE table;
	while(res)
	{
		TABLE_CELL table_cell;
		CalculateCell(trans_id, &current_configuration, &table_cell);
		table.push_back(table_cell);
		res = GetNextValue(&current_configuration);
	}

	// save table
	expr_tables[trans_id] = table;

}

vector<ExpressionElement> EXPRESSION_ENGINE::GetPossibleValues(T type)
{
	vector<ExpressionElement> le;

	cout << "Engine Error: Standard engine can not enumerate its values\n";
	exit(1);

	return le;
}

bool EXPRESSION_ENGINE::GetNextElement(pair<long, vector<ExpressionElement> > * p)
{
	if(p->first >= p->second.size()-1)
	{
		p->first = 0;
		return false;
	}
	else
	{
		(p->first)++;
		return true;
	}
}

bool EXPRESSION_ENGINE::GetNextVector(vector<pair<long, vector<ExpressionElement> > > * v)
{
	for(vector<pair<long, vector<ExpressionElement> > >::iterator it = v->begin(); it != v->end(); it++)
	{
		if(GetNextElement(&(*it)))
		 return true;
	}

	return false;
}

bool EXPRESSION_ENGINE::GetNextList(list<vector<pair<long, vector<ExpressionElement> > > > * l)
{
	for(list<vector<pair<long, vector<ExpressionElement> > > >::iterator it = l->begin(); it != l->end(); it++)
	{
		if(GetNextVector(&(*it)))
		 return true;
	}

	return false;
}

bool EXPRESSION_ENGINE::GetNextValue(vector<list<vector<pair<long, vector<ExpressionElement> > > > > * value)
{
	for(vector<list<vector<pair<long, vector<ExpressionElement> > > > >::iterator it = value->begin(); it != value->end(); it++)
	{
		if(GetNextList(&(*it)))
		 return true;
	}

	return false;
}

void EXPRESSION_ENGINE::CalculateCell(IDTYPE trans_id, 
			vector<list<vector<pair<long, vector<ExpressionElement> > > > > * current_configuration,
			TABLE_CELL * table_cell)
{
	// binding
	BINDING binding;

	vector<list<vector<pair<long, vector<ExpressionElement> > > > >::iterator itc = current_configuration->begin();

	vector<TCONNECTION> preset = pn->getPreset(trans_id);	
	for(vector<TCONNECTION>::iterator it1 = preset.begin(); it1 != preset.end(); it1++)
	{
		list<vector<pair<long, vector<ExpressionElement> > > >::iterator it_list = itc->begin();

		// list of expressions for evaluated preset expressions
		list<EXPRESSION> le;

		for(list<EXPRESSION>::iterator it2 = it1->attributes.begin(); it2 != it1->attributes.end(); it2++)
		{
			// expression for evaluated preset expression
			EXPRESSION e = *it2;
			vector<ExpressionElement> ev;
			for(vector<pair<long, vector<ExpressionElement> > >::iterator it_expr = it_list->begin(); it_expr != it_list->end(); it_expr++)
			 ev.push_back(it_expr->second[it_expr->first]);
			e.addExprParsed(ev);

			le.push_back(e);

			// binding
			vector<pair<long, vector<ExpressionElement> > >::iterator itv = it_list->begin();

			vector<ExpressionElement> expr_parsed = it2->getExprParsed();
			for(vector<ExpressionElement>::iterator it3 = expr_parsed.begin(); it3 != expr_parsed.end(); it3++)
			{
				// check, that e1 is variable
				if(it3->type != VAR_UNKNOWN)
				{
					cout << "\nerror: only variables should be used in BINDING\n";
					exit(1);
				}

				// set variable
				binding.setVariable(it3->variable_name, itv->second[itv->first]);

				itv++;
			}
			
			it_list++;
		}

		table_cell->preset.push_back(le);

		itc++;
	}

	// set binding
	table_cell->binding = binding;
	initialize(binding);

	// guard
	bool guard_res = run(pn->getTransition(trans_id).getGuard());
	
	table_cell->guard_result = guard_res;

	// calculate postsets
	if(guard_res)
	{
		vector<TCONNECTION> vpost = pn->getPostset(trans_id);

		for(vector<TCONNECTION>::iterator it_post = vpost.begin(); it_post != vpost.end(); it_post++)
		{
			list<EXPRESSION> l_expr_postset;
			for(list<EXPRESSION>::iterator it_expr_postset = it_post->attributes.begin(); 
					it_expr_postset != it_post->attributes.end(); it_expr_postset++)
			{
				l_expr_postset.push_back(run(*it_expr_postset));
			}

			table_cell->postset.push_back(l_expr_postset);
		}
	}	
}

bool EXPRESSION_ENGINE::IsColoured(IDTYPE trans_id)
{
	bool res = false;

	vector<TCONNECTION> vc = pn->getPreset(trans_id);

	for(vector<TCONNECTION>::iterator it = vc.begin(); it != vc.end(); it++)
	{
		if(it->attributes.size() != 0)
		{
			res = true;
			break;
		}
	}

	return res;
}

BINDING EXPRESSION_ENGINE::abstractBinding(BINDING b)
{
	BINDING result;

	list<string> variables = b.getVariableNames();

	for(list<string>::iterator it = variables.begin(); it != variables.end(); it++)
	{
		ExpressionElement ee = b.getVariable(*it);

		// make back the abstract flag
		// the hope is, that this method will be applied only for bindings evaluated in the standard engine
		if(ee.is_abstract==true)
		{
		 ee.is_abstract = false;
		 evaluate(ee);
		 ee.is_abstract = true;
		}
		else
		{
		 evaluate(ee);
		}
		result.setVariable(*it, ee);
	}

	return result;
}

bool EXPRESSION_ENGINE::smallereq(list<EXPRESSION> l1, list<EXPRESSION> l2)
{
	if(l1.size() != l2.size())
		return false;

	list<EXPRESSION>::iterator it1 = l1.begin();
	list<EXPRESSION>::iterator it2 = l2.begin();

	while(it1 != l1.end())
	{
		if(!smallereq(*it1, *it2))
		 return false;

		it1++;
		it2++;
	}

	return true;
}

bool EXPRESSION_ENGINE::smallereq(EXPRESSION e1, EXPRESSION e2)
{
	vector<ExpressionElement> vee1 = e1.getExprParsed();
	vector<ExpressionElement> vee2 = e2.getExprParsed();

	if(vee1.size() != vee2.size())
		return false;

	vector<ExpressionElement>::iterator it1 = vee1.begin();
	vector<ExpressionElement>::iterator it2 = vee2.begin();

	while(it1 != vee1.end())
	{
		if(!smallereq(*it1, *it2))
		 return false;

		it1++;
		it2++;
	}

	return true;
}

bool EXPRESSION_ENGINE::smallereq(ExpressionElement ee1, ExpressionElement ee2)
{
	return EXPRESSION::equal(ee1, ee2);
}

};
