/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef ATTRIBUTESEXPRESSION_ENGINE_H
#define ATTRIBUTESEXPRESSION_ENGINE_H

#include "algorithm.h"
#include "expression.h"
#include "binding.h"

#include "counter_example.h"

#include "table.h"

namespace attributes {

class EXPRESSION_ENGINE: public ALGORITHM
{
public:

	EXPRESSION_ENGINE();
	
	~EXPRESSION_ENGINE();
	
	// initialize with binding
	virtual void initialize(BINDING pb);
	
	// evaluate
	virtual EXPRESSION run(EXPRESSION & e);
	
	// to string
	virtual string run(EXPRESSION & e, PRINT_TYPE type);
	
	// check guards
	virtual bool run(list<EXPRESSION> guards);

	// refine
	virtual bool refine();

	// refine with given counter example
	virtual bool refine(COUNTER_EXAMPLE * pe);

	// set petri net
	virtual void initialize(PTNET * i_ptnet);

	// table with all possible values of attributes
	virtual bool run(IDTYPE trans_id, TABLE * * table);

	// evaluate
	virtual void evaluate(ExpressionElement & e);

	// evaluate all values in b 
	virtual BINDING abstractBinding(BINDING b);

	// <= : Compare EVALUATED expressions
	virtual bool smallereq(list<EXPRESSION> l1, list<EXPRESSION> l2);

	virtual bool smallereq(EXPRESSION e1, EXPRESSION e2);

	virtual bool smallereq(ExpressionElement ee1, ExpressionElement ee2);

protected:

	virtual void checkArgumentNumber(vector<ExpressionElement> & ve, unsigned int number);

	virtual string run(ExpressionElement &e, PRINT_TYPE type);

	virtual void CalculateTable(IDTYPE trans_id);

	// must be overwritten
	virtual vector<ExpressionElement> GetPossibleValues(T type);

	virtual bool GetNextElement(pair<long, vector<ExpressionElement> > * p);
	
	virtual bool GetNextVector(vector<pair<long, vector<ExpressionElement> > > * v);
	
	virtual bool GetNextList(list<vector<pair<long, vector<ExpressionElement> > > > * l);
	
	virtual bool GetNextValue(vector<list<vector<pair<long, vector<ExpressionElement> > > > > * value);

	virtual void CalculateCell(IDTYPE trans_id,
		vector<list<vector<pair<long, vector<ExpressionElement> > > > > * current_configuration,
		TABLE_CELL * table_cell);

	virtual bool IsColoured(IDTYPE trans_id);

	BINDING pb;

	PTNET * pn;

	map<IDTYPE, TABLE> expr_tables;

	map<IDTYPE, bool> coloured;
};

};

#endif
