/***************************************************************************
 *   Copyright (C) 2006 by root   *
 *   root@Vitali   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "expression_engine_mn.h"
#include "algorithm_storage.h"

namespace attributes {

EXPRESSION_ENGINE_MN::EXPRESSION_ENGINE_MN()
{
	numberM = interval_from;
	numberN = interval_to;
}


EXPRESSION_ENGINE_MN::~EXPRESSION_ENGINE_MN()
{
}

void EXPRESSION_ENGINE_MN::evaluate(ExpressionElement & e)
{
	if(e.is_abstract==true)
	 return;

	switch(e.type)
	{
		case OP_NUM_PLUS:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			e.int_evaluated = e.elements[0].int_evaluated + e.elements[1].int_evaluated;
			if(e.int_evaluated > numberN)
				e.int_evaluated = numberN+1;
			e.is_abstract=true;
			break;
		}
		case OP_NUM_MINUS:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);

			if(e.elements[0].int_evaluated > numberN || e.elements[1].int_evaluated > numberN)
				e.int_evaluated = numberN+1;
			else
			{
				e.int_evaluated = e.elements[0].int_evaluated - e.elements[1].int_evaluated;
				if(e.int_evaluated < numberM)
					e.int_evaluated = numberN+1;
			}

			e.is_abstract=true;

			break;
		}
		case OP_NUM_MULT:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			e.int_evaluated = e.elements[0].int_evaluated * e.elements[1].int_evaluated;
			if(e.int_evaluated > numberN)
				e.int_evaluated = numberN+1;
			e.is_abstract=true;
			break;
		}
		case OP_NUM_DIV:
		{
			cout << "\nEXPRESSION_ENGINE_MN does not support division\n";
			exit(1);
		}
		case OP_NUM_MOD:
		{
			cout << "\nEXPRESSION_ENGINE_MN does not support modulo operation\n";
			exit(1);
		}
		case NUMBER:
		{
			checkArgumentNumber(e.elements, 0);
			
			if(e.int_val>=numberM && e.int_val<=numberN)
				e.int_evaluated = e.int_val;
			else
				e.int_evaluated = numberN+1;			
			break;
		}
		case VAR_NUM:
		{
			checkArgumentNumber(e.elements, 0);
			ExpressionElement ee = pb.getVariable(e.variable_name);
			e.int_evaluated = ee.int_evaluated;
			e.is_abstract=true;
			break;
		}
		case OP_STR_LEN:
		{
			checkArgumentNumber(e.elements, 1);
			e.int_evaluated = 0;
			e.is_abstract=true;
			break;
		}
		case OP_NUM_UMINUS:
		{
			checkArgumentNumber(e.elements, 1);
			evaluate(e.elements[0]);

			e.int_evaluated = - e.elements[0].int_evaluated;
			if(e.int_evaluated > numberN < numberM || e.int_evaluated > numberN)
				e.int_evaluated = numberN+1;

			e.is_abstract=true;

			break;
		}

		case OP_BOOL_AND:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			e.bool_val = ExpressionElement::logic_3(e.elements[0].bool_val, e.elements[1].bool_val, OP_BOOL_AND);
			e.is_abstract=true;
			break;
		}
		case OP_BOOL_OR:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			e.bool_val = ExpressionElement::logic_3(e.elements[0].bool_val, e.elements[1].bool_val, OP_BOOL_OR);
			e.is_abstract=true;
			break;
		}
		case OP_BOOL_NOT:
		{
			checkArgumentNumber(e.elements, 1);
			evaluate(e.elements[0]);
			e.bool_val = ExpressionElement::logic_3(e.elements[0].bool_val, e.elements[0].bool_val, OP_BOOL_NOT);
			e.is_abstract=true;
			break;
		}
		case OP_NUM_BOOL_EQ:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			if(e.elements[0].int_evaluated<=numberN && e.elements[1].int_evaluated<=numberN)
			{
				if(e.elements[0].int_evaluated == e.elements[1].int_evaluated)
					e.bool_val = TRUE_VAL;
				else
					e.bool_val = FALSE_VAL;
			}
			else
				e.bool_val = DN_VAL;
			e.is_abstract=true;
			break;
		}
		case OP_NUM_BOOL_LE:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			if(e.elements[0].int_evaluated<=numberN && e.elements[1].int_evaluated<=numberN)
			{
				if(e.elements[0].int_evaluated <= e.elements[1].int_evaluated)
					e.bool_val = TRUE_VAL;
				else
					e.bool_val = FALSE_VAL;
			}
			else
			{
				e.bool_val = DN_VAL;
			}
			e.is_abstract=true;
			break;
		}
		case OP_NUM_BOOL_GE:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			if(e.elements[0].int_evaluated<=numberN && e.elements[1].int_evaluated<=numberN)
			{
				if(e.elements[0].int_evaluated >= e.elements[1].int_evaluated)
					e.bool_val = TRUE_VAL;
				else
					e.bool_val = FALSE_VAL;
			}
			else
			{
				e.bool_val = DN_VAL;
			}
			e.is_abstract=true;
			break;
		}
		case OP_NUM_BOOL_NE:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			if(e.elements[0].int_evaluated<=numberN || e.elements[1].int_evaluated<=numberN)
			{
				if(e.elements[0].int_evaluated != e.elements[1].int_evaluated)
					e.bool_val = TRUE_VAL;
				else
					e.bool_val = FALSE_VAL;
			}
			else
				e.bool_val = DN_VAL;
			e.is_abstract=true;
			break;
		}
		case OP_NUM_BOOL_L:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			if(e.elements[0].int_evaluated<=numberN && e.elements[1].int_evaluated<=numberN)
			{
				if(e.elements[0].int_evaluated < e.elements[1].int_evaluated)
					e.bool_val = TRUE_VAL;
				else
					e.bool_val = FALSE_VAL;
			}
			else
			{
				e.bool_val = DN_VAL;
			}
			e.is_abstract=true;
			break;
		}
		case OP_NUM_BOOL_G:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			if(e.elements[0].int_evaluated<=numberN && e.elements[1].int_evaluated<=numberN)
			{
				if(e.elements[0].int_evaluated > e.elements[1].int_evaluated)
					e.bool_val = TRUE_VAL;
				else
					e.bool_val = FALSE_VAL;
			}
			else
			{
				e.bool_val = DN_VAL;
			}
			e.is_abstract=true;
			break;
		}
		case BOOLEAN:
		{
			checkArgumentNumber(e.elements, 0);
			break;
		}
		
		case _STR_:
		{
			checkArgumentNumber(e.elements, 0);
			e.str_evaluated = "";
			e.is_abstract=true;
			break;
		}
		case VAR_STR:
		{
			checkArgumentNumber(e.elements, 0);
			// ExpressionElement ee = pb->getVariable(e.variable_name);
			e.str_evaluated = "";
			e.is_abstract=true;
			break;
		}
		case OP_STR_CONC:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			e.str_evaluated = "";
			e.is_abstract=true;
			break;
		}
		case OP_STR_BOOL_EQ:
		{
			checkArgumentNumber(e.elements, 2);
			evaluate(e.elements[0]);
			evaluate(e.elements[1]);
			e.bool_val = DN_VAL;
			e.is_abstract=true;
			break;
		}
		case VAR_UNKNOWN:
		{
			checkArgumentNumber(e.elements, 0);
			ExpressionElement ee = pb.getVariable(e.variable_name);
			e.str_evaluated = ee.str_evaluated;
			e.int_evaluated = ee.int_evaluated;
			e.bool_val = ee.bool_val;
			e.is_abstract=true;
			break;
		}
		default:
		{
			cout << "\ninternal error in EXPRESSION_ENGINE: Unknown EXPRESSION\n";
			exit(1);
		}
	}
}

string EXPRESSION_ENGINE_MN::run(ExpressionElement &e, PRINT_TYPE type)
{
	switch(e.type)
	{
		case _STR_:
		case VAR_STR:
		case OP_STR_CONC:
		{
			return string("STRING");
		}
		default:
		{
			return EXPRESSION_ENGINE::run(e,type);
		}
	}
}

bool EXPRESSION_ENGINE_MN::refine()
{
	interval_from++;
	numberM = interval_from;
	interval_to++;
	numberN = interval_to;	

	storage.writeDB();

	return true;
}

vector<ExpressionElement> EXPRESSION_ENGINE_MN::GetPossibleValues(T type)
{
	vector<ExpressionElement> result;

	// integer
	if(type == INTEGER)
	{
		for(int i=-numberM; i<=numberN+1; i++)
		{	
			ExpressionElement ee_plus;
			ee_plus.type = NUMBER;
			ee_plus.int_evaluated = i;
			ee_plus.is_abstract = true;
			result.push_back(ee_plus);
		}
	}
	// string
	else if(type == STRING)
	{
		// string
		ExpressionElement ee_string;
		ee_string.type = _STR_;
		ee_string.is_abstract = true;
		result.push_back(ee_string);
	}
	// boolean
	else
	{
		// plus
		ExpressionElement ee_true;
		ee_true.type = BOOLEAN;
		ee_true.bool_val = TRUE_VAL;
		ee_true.is_abstract = true;
		result.push_back(ee_true);
	
		// minus
		ExpressionElement ee_false;
		ee_false.type = BOOLEAN;
		ee_false.bool_val = FALSE_VAL;
		ee_false.is_abstract = true;
		result.push_back(ee_false);

		// all
		ExpressionElement ee_dn;
		ee_dn.type = BOOLEAN;
		ee_dn.bool_val = DN_VAL;
		ee_dn.is_abstract = true;
		result.push_back(ee_dn);
	}

	return result;
}

bool EXPRESSION_ENGINE_MN::smallereq(ExpressionElement ee1, ExpressionElement ee2)
{
	if( (ee1.type != ee2.type) ||
		(ee1.bool_val != ee2.bool_val) ||
		(ee1.str_evaluated != ee2.str_evaluated)
	  )

	if(ee1.int_evaluated != ee2.int_evaluated && ee2.int_evaluated != numberN+1)	
		return false;

	return true;
}

}


