%{
/*****************************************************************************
 * reglex.ll: lex file for regular expressions
 *****************************************************************************
 * Authors: Salil Joshi <hellfeuer@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/


#include <stdio.h>
#include <string>
#include "../semilin/cfexpression.h"
#include "../finite_automaton.h"
#include "expparse.tab.hh"
using std::string;
#undef YY_INPUT
char *ex_strin;		/* current pointer on input string */

#define YY_INPUT(buf, result, max_size) \
	strncpy(buf, ex_strin, max_size-1);\
	ex_strin += result = strlen(buf);

%}

delim	[ \t\n]
white	{delim}+
smallchar	[a-z_]
bigchar [A-Z]
digit	[2-9]
alphanum	{smallchar}|{digit}|{bigchar}
star	\*
plus	\+
lb	\(
rb	\)
equal \=

%%
{white}	;/* eat up whitespace */
{smallchar}+ {
		ex_lval.str = new string(ex_text);
		return STR;
		}
{bigchar} {
		ex_lval.str = new string(&ex_text[0]);
		return VAR;
		}
{digit}+ {
		ex_lval.str = new string(ex_text);
		return POWER;
	}
{star}	{return STAR;}
{plus}	{return PLUS;}
{lb}	{return LB;}
{rb}	{return RB;}
{equal} {return EQUAL;}
%%

