%{
/*****************************************************************************
 * regparse.yy: yacc grammar for regular expressions
 *****************************************************************************
 * Authors: Salil Joshi <hellfeuer@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/
#include "../semilin/cfexpression.h"
#include <string>
#include <cstdlib>
#include <iostream>

extern "C"
{
	int ex_parse(void);
	int ex_wrap()
	{
		return 1;
	}
	void ex_error(const char* str)
	{
		fprintf(stderr,"error: %s\n",str);
	}

	using namespace std;
}
int ex_lex(void);
CFEXPRESSION * finalExp;
%}

%token STR VAR STAR PLUS POWER CONCAT LB RB EQUAL

%union {
CFEXPRESSION * exp;
std::string* str;
}

%type <exp> expr;
%type <exp> term;
%type <exp> factor;
%type <str> VAR;
%type <str> STR;
%type <str> POWER;

%%
final	: expr
	{
		finalExp = $1;
	}
	;
expr	: expr EQUAL expr
	{
		$$ = new CFEXPRESSION(semilin::OP_EQUAL, $1, $3);
	}
	| expr PLUS term
	{
		$$ = new CFEXPRESSION(semilin::OP_PLUS, $1, $3);
	}
	| term
	{
	}
	;
term	: term factor
	{
		$$ = new CFEXPRESSION(semilin::OP_CONCAT, $1, $2);
	}
	| factor
	{
	}
	;
factor	: LB expr RB
	{
		$$ = $2;
	}
	| factor STAR
	{
		$$ = new CFEXPRESSION(semilin::OP_STAR, $1);
	}
	| factor POWER
	{
		$$ = new CFEXPRESSION(semilin::OP_POWER, $1, atoi($2->c_str()));
	}
	| STR
	{
		$$ = new CFEXPRESSION(semilin::STRING, $1);
		delete $1;
	}
	| VAR
	{
		$$ = new CFEXPRESSION(semilin::VARIABLE, $1);
		delete $1;
	}
	;

%%

