/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include "find_all_matches.h"

namespace algorithms
{
   namespace matchings
   {
      FIND_ALL_MATCHES::FIND_ALL_MATCHES()
      {
         this->ID = ALGORITHM_FIND_ALL_MATCHES;
         this->name = "find_all_matches";
	 this->ggraph = new HYPERGRAPH;
      }
      FIND_ALL_MATCHES::~FIND_ALL_MATCHES() {}
   
      void FIND_ALL_MATCHES::initialize(HYPERGRAPH* i_t)
      {
         assert(i_t != NULL);
         this->ggraph = i_t;
      }
      
      list<MATCH> FIND_ALL_MATCHES::run( HYPERGRAPH* hgraph, ALGORITHM* i_matchhandler, 
	                                 bool surjective )
      {
         assert(hgraph != NULL);
         //assert(i_matchhandler != NULL);
	 
	 stack< map<IDTYPE, EDGE>::iterator > mappings;
	 stack< pair<IDTYPE, EDGE> > unmapped_edges;
	 /*
	  * initialize unmapped_edges := hgraph->edges
	  */
	 map<IDTYPE, EDGE> hedges = hgraph->getEdges();
         for(map<IDTYPE, EDGE>::iterator i = hedges.begin(); i != hedges.end(); i++)
	 {
	    unmapped_edges.push(*i);
	 }
	 
	 MATCH phi_v, phi_e;
	 list<MATCH> found_matches;
	 this->next( hgraph, &mappings, &unmapped_edges, &phi_v, &phi_e, &found_matches,
	             i_matchhandler, surjective );
		     
         /*
	  * next(...) should have cleared phi and mappings
	  */
	 assert(phi_v.count() == 0);
	 assert(phi_e.count() == 0);
	 assert(mappings.size() == 0);
	 assert(unmapped_edges.size() == hedges.size());
	 
	 return found_matches;
      }
      
      void FIND_ALL_MATCHES::next( HYPERGRAPH* hgraph,
                                   stack< map<IDTYPE, EDGE>::iterator >* mappings,
	                           stack< pair<IDTYPE, EDGE> >* unmapped_edges,
				   MATCH* phi_v, MATCH* phi_e,
				   list<MATCH>* found_matches,
				   ALGORITHM* i_matchhandler, bool surjective )
      {
         map<IDTYPE, IDTYPE> m_phi_e = phi_e->getMatch();
         map<IDTYPE, IDTYPE> m_phi_v = phi_v->getMatch();
	 map<IDTYPE, IDTYPE>::iterator mapit;
	    
	 /*
	  * Did we find a match? (or are there still any unmapped edges?)
	  */
	 if(unmapped_edges->size() == 0)
	 {
	    /*
	     * if the corresponding flag is set, check whether the match is surjective
	     */
	    bool surjective_checked = false;
	    if(surjective)
	    {
	       set<IDTYPE> ev;
	       for(mapit = m_phi_e.begin(); mapit != m_phi_e.end(); mapit++)
	          ev.insert(mapit->second);
	       for(mapit = m_phi_v.begin(); mapit != m_phi_v.end(); mapit++)
	          ev.insert(mapit->second);
		  
	       if( ev.size() == ( this->ggraph->getEdges().size() +
	                          this->ggraph->getVertices().size() ) )
	          surjective_checked = true;
	    }
	    else surjective_checked = true;
	    
	    if(surjective_checked)
	    {
  	       MATCH thematch = phi_v->clone();
	       for(mapit = m_phi_e.begin(); mapit != m_phi_e.end(); mapit++)
	          thematch.setMatch(mapit->first, mapit->second);
	       
		if(i_matchhandler==NULL)
		{
			found_matches->push_back(thematch);
		}
		else
		{
			if(i_matchhandler->run(&thematch)) 
				found_matches->push_back(thematch);
		}
	    }
	 }
	 else
	 {
	    /*
	     * the match is not yet complete. get the next edge to be mapped.
	     */
	    pair<IDTYPE, EDGE> hedge = unmapped_edges->top();
	    LABEL hedgelabel = hedge.second.getLabel();
	    list<IDTYPE> hcV = hgraph->getConnectedVertices(hedge.first);
	    unmapped_edges->pop();
	    
	    /* 
	     * the iterator on the mappings stack iterates through all the edges of g
	     * in order to find an appropriate mapping
	     */
	    map<IDTYPE, EDGE> gedges = this->ggraph->getEdges();
	    mappings->push(gedges.begin());
	    
	    while(mappings->top() != gedges.end())
	    {
	       pair<IDTYPE, EDGE> gedge = *(mappings->top());
	       LABEL gedgelabel = gedge.second.getLabel();
	       list<IDTYPE> gcV = this->ggraph->getConnectedVertices(gedge.first);
	       if((gedgelabel != hedgelabel) || (gcV.size() != hcV.size()))
	       {
	          (mappings->top()) ++;
		  continue;
	       }
	       
	       /*
	        * store all new vertex mappings in order to delete them later
		*/
	       list<IDTYPE> newMappedVertices;
	       bool vertexMatchingFailed = false;
	       list<IDTYPE>::iterator gv = gcV.begin();
	       list<IDTYPE>::iterator hv = hcV.begin();
	       map<IDTYPE, IDTYPE> phi_v_content = phi_v->getMatch();
	       map<IDTYPE, IDTYPE> phi_e_content = phi_e->getMatch();
	       map<IDTYPE, IDTYPE>::iterator mapit;
	       
	       while(!vertexMatchingFailed && (hv != hcV.end()))
	       {
	          mapit = phi_v_content.find(*hv);
		  if(mapit == phi_v_content.end())
		  {
		     newMappedVertices.push_back(*hv);
		     phi_v->setMatch(*hv, *gv);
		     // Lösung Von Maxim Zaks
		     phi_v_content[*hv] = *gv;
		     ////////////////////////
		  }
		  else if(mapit->second != *gv) vertexMatchingFailed = true;
		  gv ++;
		  hv ++;
	       }
	       
	       /*
	        * Now, either all vertices are mapped, or vertexMatchingFailed == true.
		*/
	       if(!vertexMatchingFailed)
	       {
	          phi_e->setMatch(hedge.first, gedge.first);
		  this->next( hgraph, mappings, unmapped_edges, phi_v, phi_e,
		              found_matches, i_matchhandler, surjective );
		  phi_e->deleteKey(hedge.first);
	       }
	       
	       /*
	        * undo all changes in phi_v
		*/
	       for( list<IDTYPE>::iterator v = newMappedVertices.begin();
	            v != newMappedVertices.end();
		    v++ )
	       {
	          phi_v->deleteKey(*v);
	       }
	       
	       /*
	        * try next edge mapping
		*/
	       (mappings->top()) ++;
	    }
	    
	    unmapped_edges->push(hedge);
	    mappings->pop();
	    
	    return;
	 }
      }

////////// namespaces ///////////      
   }
}
