/***************************************************************************
 *   Copyright (C) 2006 by Maxim Zaks   *
 *   mzaks@neckar   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "find_all_matches_by_sp.h"

namespace algorithms {

namespace matchings {

FIND_ALL_MATCHES_BY_SP::FIND_ALL_MATCHES_BY_SP()
{
	this->ID = ALGORITHM_FIND_ALL_MATCHES_BY_SP;
        this->name = "find_all_matches_by_sp";
	this->ggraph = new HYPERGRAPH;
}


FIND_ALL_MATCHES_BY_SP::~FIND_ALL_MATCHES_BY_SP()
{
}

void FIND_ALL_MATCHES_BY_SP::initialize(HYPERGRAPH* i_t)
      {
         assert(i_t != NULL);
         this->ggraph = i_t;
	 costMatrix.intitialize( i_t );
// 	 costMatrix.run();
      }
      
list<MATCH> FIND_ALL_MATCHES_BY_SP::run( HYPERGRAPH* hgraph, ALGORITHM* i_matchhandler, 
	                                 bool surjective )
      {
        assert(hgraph != NULL);
//          assert(i_matchhandler != NULL);
	
	list<MATCH> matches;
	if(!searchPlan.initialize( &costMatrix, hgraph, NULL, false )) return matches;
	
	MATCH m;
	searchPlan.runOperationForAll( &m, 0 );

 
// 	list<MATCH> found_matches = searchPlan.allMatches;
	
	if(searchPlan.allMatches.empty()) return matches;
	 
	for(list<MATCH>::iterator match_iter = searchPlan.allMatches.begin(); match_iter!=searchPlan.allMatches.end(); match_iter++){
		
		map<IDTYPE,IDTYPE> thisMatch = (*match_iter).getMatch();
// 		bool surjective_checked = false;
		if(surjective)
	    	{
			set<IDTYPE> ev;
	       		for(map<IDTYPE,IDTYPE>::iterator mapit = thisMatch.begin(); mapit != thisMatch.end(); mapit++)
	          		ev.insert(mapit->second);
	       				  
	       		if( ev.size() != ( this->ggraph->getEdges().size() + this->ggraph->getVertices().size() ) )
				continue;

	    	}
	    	 	       		
		if(i_matchhandler!=NULL && !i_matchhandler->run(&(*match_iter)))
		{ 
			continue;
		}
		matches.push_back(*match_iter);
	}
	
	return matches;
      }

}

}
