/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "findmatch.h"

namespace algorithms
{
   namespace matchings
   {
      FINDMATCH::FINDMATCH()
      {
         this->ID = ALGORITHM_FINDMATCH;
         this->name = "findmatch";
//          this->ggraph = new HYPERGRAPH;
	 matchhandler = NULL;
      }

      FINDMATCH::~FINDMATCH()
      {
      }
       
      void FINDMATCH::initialize(HYPERGRAPH* i_t, MATCH* im1, MATCH* im2, bool* icf, bool* icu, PETRIGRAPH* pg)
      {
         assert(i_t != NULL);
	 assert(im1 != NULL);
	 this->ggraph = i_t;
	 this->m1 = im1;
	 this->m2 = im2;
	 this->canFold = icf;
	 this->canUnfold = icu;
	 this->petrigraph = pg;

// 	 map<IDTYPE, EDGE> edges = this->ggraph->getEdges();
//          for(map<IDTYPE, EDGE>::iterator i = edges.begin(); i != edges.end(); i++)
// 	 {
// 	     
// 	 /////////////////////////
// 	 //directMatch berechnen
// 	 /////////////////////////
// 	    		
// 		edgeLookUp[i->second.getLabel()].push_back( i->first );
// 
// 			
// 	 }
      }
      void FINDMATCH::initialize(HYPERGRAPH* i_t)
      {
         this->initialize(i_t, new MATCH(), NULL, new bool, new bool, NULL);
      }
      
      MATCH* FINDMATCH::run(HYPERGRAPH* hgraph, bool* found, IDTYPE ruleid, MARKING* phi_E_L)
      {
         assert(hgraph != NULL);
	 
	 MATCH* result = new MATCH();
	 
	 if(matchhandler==NULL)
	 	matchhandler = this->load("call_matchhandler");
		
	 if(debug)
	 {
	  cout << "PHI2 promulgation: matchhandler recvs phi2 = " << m2 << "\n";
	 }

	 if(matchhandler && petrigraph)
	  matchhandler->initialize( this->petrigraph, ruleid,
	 			this->petrigraph->getGrammar()->getRule(ruleid),
				this->canFold,this->canUnfold, m1, m2, phi_E_L );

	 stack< map<IDTYPE, EDGE>::iterator > mappings;
	 stack< pair<IDTYPE, EDGE> > unmapped_edges;
	 /*
	  * initialize unmapped_edges := hgraph->edges
	  */
	 map<IDTYPE, EDGE> hedges = hgraph->getEdges();
         for(map<IDTYPE, EDGE>::iterator i = hedges.begin(); i != hedges.end(); i++)
	 {
	    unmapped_edges.push(*i);
	 }
	 
	 MATCH* phi_v = new MATCH;
	 MATCH* phi_e = new MATCH;
	 if( this->findNextMatching( hgraph, &mappings,
	                             &unmapped_edges, phi_v, phi_e, matchhandler ) )
         {
	    /*
	     * set *matching to (*phi_v U *phi_e)
	     */
            *result = phi_v->clone();
	    map<IDTYPE, IDTYPE> m_phi_e = phi_e->getMatch();
	    for(map<IDTYPE, IDTYPE>::iterator i = m_phi_e.begin(); i != m_phi_e.end(); i++)
	    {
	       result->setMatch(i->first, i->second);
	    }
	    if(found != NULL) *found = true;
	 }
	 else
	 {
	    /*
	     * reset *result
	     */
            *result = MATCH();
	    if(found != NULL) *found = false;
	 }
	 return result;
      }
      MATCH* FINDMATCH::run(HYPERGRAPH* hgraph, bool* found, IDTYPE ruleid)
      {
         this->run(hgraph, found, ruleid, NULL);
      }
      MATCH* FINDMATCH::run(HYPERGRAPH* hgraph, IDTYPE ruleid)
      {
         return this->run(hgraph, NULL, ruleid, NULL);
      }  
      
      /*
       * protected method called by FINDMATCH::run
       * corresponds to the Hypergraph::findNextMatching method of AUGUR I
       */
      bool FINDMATCH::findNextMatching( HYPERGRAPH* hgraph,
                                        stack<map<IDTYPE, EDGE>::iterator>* mappings,
	                                stack< pair<IDTYPE, EDGE> >* unmapped_edges, MATCH* phi_v,
					MATCH* phi_e, ALGORITHM* matchhandler )
      {
         assert(hgraph != NULL);
         assert(mappings != NULL);
	 assert(unmapped_edges != NULL);
	 assert(phi_v != NULL);
	 assert(phi_e != NULL);
	 assert(matchhandler != NULL);
      
         if(unmapped_edges->size() == 0)
	 {
	    /*
	     * a complete match (*phi_v U *phi_e) was found
	     * call the matchhandler to "use this match" and determine whether it is applicable
	     */
	    MATCH* matching = new MATCH;
	    *matching = phi_v->clone();
	    map<IDTYPE, IDTYPE> m_phi_e = phi_e->getMatch();
	    for(map<IDTYPE, IDTYPE>::iterator i = m_phi_e.begin(); i != m_phi_e.end(); i++)
	    {
	       matching->setMatch(i->first, i->second);
	    }
	    if(matchhandler && petrigraph)
	     return matchhandler->run(matching);
	    else
	     return true;
	 }
	 else
	 {
	    /*
	     * The matching (*phi_v U *phi_e) is incomplete.
	     *
	     * Edges and vertices in this->g are named with prefix g,
	     * edges and vertices in h with prefix h.
	     *
	     * Now, get the next edge to be mapped.
	     */
	    pair<IDTYPE, EDGE> hedge = unmapped_edges->top();
	    unmapped_edges->pop();
	    LABEL hedgelabel = hedge.second.getLabel();
	    /*
	     * all vertices connected to the edge to be mapped
	     */
	    list<IDTYPE> hcV = hgraph->getConnectedVertices(hedge.first);
	    
	    /*
	     * The stack *mappings stores iterators to edges of *t
	     * that are scanned through in order to find an appropriate mapping
	     */
	    map<IDTYPE, EDGE> gedges = this->ggraph->getEdges();
	    
	    /*
	     * iterate through all edges of g and try to map hedge to them
	     */
	    for( mappings->push(gedges.begin());
	         mappings->top() != gedges.end();
		 (mappings->top())++ )
            {
	       pair<IDTYPE, EDGE> gedge = *(mappings -> top());
	       LABEL gedgelabel = gedge.second.getLabel();
	       list<IDTYPE> gcV = this->ggraph->getConnectedVertices(gedge.first);
	       if ((gedgelabel == hedgelabel) && (gcV.size() == hcV.size()))
	       {
		  map<IDTYPE, IDTYPE> phi_v_content = phi_v->getMatch();
		  map<IDTYPE, IDTYPE>::iterator phi_v_it;
	          /*
		   * remember which vertices are added to phi_v
		   * this is necessary to allow undoing these changes
		   */
		  list<IDTYPE> newMappedVertices;
		  bool vertexMatchingFailed = false;
		  list<IDTYPE>::iterator gv, hv;
		  
		  for( gv = gcV.begin(), hv = hcV.begin();
		       !vertexMatchingFailed && (gv != gcV.end());
		       gv++, hv++ )
		  {
		     phi_v_it = phi_v_content.find(*hv);
		     if(phi_v_it == phi_v_content.end())  // *hv is not yet mapped
		     {
		        newMappedVertices.push_back(*hv);
			phi_v->setMatch(*hv, *gv);
			// Lösung Von Maxim Zaks
			phi_v_content[*hv] = *gv;
			////////////////////////
		     }
		     else  // *hv was already mapped
		     {
		        /*
			 * compare phi_v(*hv) to *gv
			 */
			if(phi_v_it->second != *gv) vertexMatchingFailed = true;
		     }
		  }
		  
		  /*
		   * now, all vertices connected to hedge are mapped,
		   * or otherwise vertexMatchingFailed is true
		   */
		  if(!vertexMatchingFailed)
		  {
		     phi_e->setMatch(hedge.first, gedge.first);
		     if(this->findNextMatching( hgraph, mappings, unmapped_edges,
		                                phi_v, phi_e, matchhandler ))
                     {
		        return true;
		     }
		     else phi_e->deleteKey(hedge.first);
		  }
		  
		  /*
		   * remove all additions to phi_v
		   */
		  for( list<IDTYPE>::iterator v = newMappedVertices.begin();
		       v != newMappedVertices.end();
		       v ++ )
		  {
		     phi_v->deleteKey(*v);
		  }
	       }
	    }
	    
	    unmapped_edges->push(hedge);
	    mappings->pop();
	    
	    return false;
	 }
      }
   };
};
