/***************************************************************************
 *   Copyright (C) 2006 by Maxim Zaks   *
 *   mzaks@neckar   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "findmatch_by_sp.h"
#include "algorithm_storage.h"

namespace algorithms {

namespace matchings {

FINDMATCH_BY_SP::FINDMATCH_BY_SP()
{
	this->ID = ALGORITHM_FINDMATCH_BY_SP;
        this->name = "findmatch_by_sp";
//         this->ggraph = new HYPERGRAPH;
	matchhandler = NULL;
	loaded = false;
}


FINDMATCH_BY_SP::~FINDMATCH_BY_SP()
{
}

void FINDMATCH_BY_SP::initialize(HYPERGRAPH* i_t, MATCH* im1, MATCH* im2, bool* icf, bool* icu, PETRIGRAPH* pg)
      {
         assert(i_t != NULL);
	 assert(im1 != NULL);
	 this->ggraph = i_t;
	 this->m1 = im1;
	 this->m2 = im2;
	 this->canFold = icf;
	 this->canUnfold = icu;
	 this->petrigraph = pg;

	 // try to load DATA from Algorithm Storage
	 DATA * d = storage.getData(this->name);
	 if(d != NULL)
	 {
   	 	storage.setData(this->name, NULL);
		loaded = true;

		SEARCHPLAN * sp = static_cast<SEARCHPLAN *>(d);

		// make copy
		searchPlan = *sp;
		searchPlan.loaded = true;
	 }
	 else
	 {
		costMatrix.intitialize( i_t );
// 		costMatrix.run();
	 }
      }

void FINDMATCH_BY_SP::initialize(HYPERGRAPH* i_t)
      {
	 MATCH m;
	 bool b1,b2;
         this->initialize(i_t, &m, NULL, &b1, &b2, NULL);
      }
      
MATCH* FINDMATCH_BY_SP::run(HYPERGRAPH* hgraph, bool* found, IDTYPE ruleid, MARKING* phi_E_L)
      {
         assert(hgraph != NULL);

	 
// 	 MATCH * found_match1 = new MATCH();
// 	 MATCH * found_match = new MATCH(); 
// 
// 	 if(!searchPlan.initialize( &costMatrix, hgraph ))return found_match;
// 	 *found_match = searchPlan.runOperation( found_match1, 0 );
// 	 
	 if(matchhandler==NULL)
	 	matchhandler = this->load("call_matchhandler");
		
	 if(debug)
	 {
	  cout << "PHI2 promulgation: matchhandler recvs phi2 = " << m2 << "\n";
	 }
	 bool correct = true;
 	 if(matchhandler && petrigraph){
	  	matchhandler->initialize( this->petrigraph, ruleid,
	 			this->petrigraph->getGrammar()->getRule(ruleid),
				this->canFold, this->canUnfold, m1, m2, phi_E_L );
 	 	
// 		correct = matchhandler->run( found_match );	 	 
 	 }
	 MATCH * found_match = new MATCH(); 
	 MATCH m;
	 
	 if(!loaded)
	 {
		if(!searchPlan.initialize( &costMatrix, hgraph, matchhandler, (petrigraph!=NULL) )){
			cout<< "no!";
			return found_match;
		}
// 		if(searchPlan.startMatching().count()==0){ 
// 			cout << "durty"; return found_match; //Durty trick!!!
// 		}
	 }
	 else
	  searchPlan.setMatchHandler(matchhandler);
	
	 *found_match = searchPlan.runOperation( &m, 0 );
	 
	 if(found_match->count() == 0) *found = false;
	 else *found = true;
	 
	 return found_match;



//  	 if(correct) return found_match;
// 	 
//  	 if(found_match == NULL) {
// 		
//  		MATCH* result = new MATCH();
//  		return result;
//  	 }
//  	 else{ 
// 		list<MATCH> resultList = searchPlan.startMatchingForAll();
// 		list<MATCH>::iterator iter = resultList.begin();
// // 		iter++;
// 		int i = 0;
// 		for( iter; iter!=resultList.end(); iter++){
// 			
// 			correct = matchhandler->run( &(*iter) );
// 			if(correct){ 
// 
// 				cout << "#" << i << "/" << resultList.size();
// 				return &(*iter);
// 			}	
// 			i++;
// 		}
// 		cout << "All";
//  	 }
// 	 MATCH* result = new MATCH();
// 	 return result;
      }

MATCH* FINDMATCH_BY_SP::run(HYPERGRAPH* hgraph, bool* found, IDTYPE ruleid)
      {
         this->run(hgraph, found, ruleid, NULL);
      }

MATCH* FINDMATCH_BY_SP::run(HYPERGRAPH* hgraph, IDTYPE ruleid)
      {
	 bool found;
         return this->run(hgraph, &found, ruleid, NULL);
      }

}

}
