/***************************************************************************
 *   Copyright (C) 2007 by Salil Joshi   *
 *   hellfeuer@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "finite_automaton.h"
#include <iostream>
#include <string>
using std::cout;
using std::endl;

namespace data_structures {

IDTYPE STATE::currentId = 0;
STATE::STATE()
{
	id = STATE::currentId++;
}

STATE::STATE(const STATE& s)
{
	id = s.id;
}

STATE::STATE(IDTYPE i)
{
	id = i;
}
STATE::~STATE()
{
}

STATE* FINITE_AUTOMATON::addState()
{
	pair<set<STATE*>::iterator, bool> result = allStates.insert(new STATE());
	return (*(result.first));
}

STATE* FINITE_AUTOMATON::addState(IDTYPE id)
{
	pair<set<STATE*>::iterator, bool> result = allStates.insert(new STATE(id));
	return (STATE *)&(*(result.first));
}

void FINITE_AUTOMATON::addTransition(STATE* from, string a, STATE* to)
{
	pair<STATE*, string> key = pair<STATE*, string>(from, a);
	set<STATE*> data;
	data.insert(to);
	pair< pair<STATE*, string>, set<STATE*> > value= pair< pair<STATE*, string>, set<STATE*> >(key, data);
	transitions.insert(value);
}

void FINITE_AUTOMATON::addTransition(STATE* from, string a, set<STATE*> to)
{
	pair<STATE*, string> key = pair<STATE*, string>(from, a);
	pair< pair<STATE*, string>, set<STATE*> > value= pair< pair<STATE*, string>, set<STATE*> >(key, to);
	transitions.insert(value);
}


//Finds a transition from the 'from' state and containing the 'to' state in its destination set
set<string> *FINITE_AUTOMATON::findTransitions(STATE* from, STATE* to)
{
	set<string> * result = new set<string>();
	for (map<pair<STATE*, string>, set<STATE*> >::iterator it=transitions.begin(); it != transitions.end(); it++)
	{
		if (it->first.first == from) {
			if ((it->second.find(to)) != it->second.end()) {
				result->insert((*it).first.second);
			}
		}	
	}
	return result;
}



void FINITE_AUTOMATON::write(char * filename) {

	ofstream of(filename);
	
	map<pair<STATE*, string>, set<STATE*> >::iterator i;
	set<STATE *>::iterator j;
	
	of<<"All States:"<<endl;
	for (j = allStates.begin(); j != allStates.end(); j++)
	{
		of<<"\t"<<(*j)->id<<endl;
	}
	of<<"Initial States:"<<endl;
	for (j = initialStates.begin(); j != initialStates.end(); j++)
	{
		of<<"\t"<<(*j)->id<<endl;
	}
	of<<"Final States:"<<endl;
	for (j = finalStates.begin(); j != finalStates.end(); j++)
	{
		of<<"\t"<<(*j)->id<<endl;
	}
	for (i = transitions.begin(); i != transitions.end(); i++)
	{
		of <<(*i).first.first->id<<" - "<<(*i).first.second<<" - ";
		for (j = (*i).second.begin(); j != (*i).second.end(); j++)
		{
			of<<(*j)->id<<", ";
		}
		
		of<<endl;
	}

	of.close();
}

FINITE_AUTOMATON::FINITE_AUTOMATON()
{
}


FINITE_AUTOMATON::~FINITE_AUTOMATON()
{
}


}
