/***************************************************************************
 *   Copyright (C) 2007 by Salil Joshi   *
 *   hellfeuer@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef DATA_STRUCTURESFINITE_AUTOMATON_H
#define DATA_STRUCTURESFINITE_AUTOMATON_H

#include <set>
#include <map>
#include <utility>
#include <string>

using namespace std;

#include "globals.h"

namespace data_structures {
/**
	@author hellfeuer,,, <hellfeuer@hellfeuer-ubuntu>
*/

/*
Represents a single state of the FA
*/
class STATE {
public:
	static IDTYPE currentId;
	IDTYPE id;
	STATE();
	STATE(IDTYPE i);
	STATE(const STATE& s);
	bool operator< (STATE s);
	STATE& operator= (const STATE &s);
	~STATE();
};

/*
Used to represent a non-deterministic finite automaton
*/
class FINITE_AUTOMATON{
public:
	set<int> s;
	set<STATE*> allStates;
	set<STATE*> initialStates;
	set<STATE*> finalStates;
 	map<pair<STATE*, string>, set<STATE*> > transitions; //transition from state to set of states, labelled by a string
	
	STATE* addState();
	STATE* addState(IDTYPE id);
	void addTransition(STATE* from, string a, STATE * to);
	void addTransition(STATE* from, string a, set<STATE*> to);
	set<string> *findTransitions(STATE* from, STATE* to);
	
	void write(char * filename);

    FINITE_AUTOMATON();
    FINITE_AUTOMATON(const FINITE_AUTOMATON &fa);
    
    void clear();
    ~FINITE_AUTOMATON();
};

};
using namespace data_structures;

#endif
