/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "fire_transition.h"
#include "algorithm_storage.h"
#include "binder.h"

namespace algorithms
{
namespace ptnet
{

FIRE_TRANSITION::FIRE_TRANSITION()
{
	pn = NULL;
}


FIRE_TRANSITION::~FIRE_TRANSITION()
{
}

void FIRE_TRANSITION::initialize(PTNET * pn)
{
 this->pn = pn;
}

Marking FIRE_TRANSITION::run(TRANSITION t, Marking m)
{
    if(nodes_as_places)
    {
	 // the same for-loop for nodes
	 cout << "\nPTNET::fireTransition() with nodes as places is not implemented\n";
	 exit(1);
    }

    Marking result;

	map<IDTYPE, EDGE> * edges = pn->getEdgesPointer();
	
	for(map<IDTYPE, EDGE>::iterator i = edges->begin();
		i!=edges->end();
		i++)
	{
	// number of tokens on the place before fireing
	// precondition ensures tokens>=0
	unsigned tokens=m[i->first];

	// ...plus added tokens in the postset of *this
	unsigned w = pn->WeightInPostSet(t.getID(), i->first);
	if(w>0)
	{
		if(OMEGA==tokens) // handle "infinity" case
		tokens = OMEGA;
		else
		tokens += w;
	}
	
	// ... minus removed tokens in the preset of *this
	w = pn->WeightInPreSet(t.getID(), i->first); 
	if(w>0)
	{
		if(OMEGA==tokens)
		tokens = OMEGA;
		else
		tokens -= w;
	}
	
	if(tokens>0)
	 result[i->first] = tokens;  
	
	}
 
    return result;

}

list<Marking> FIRE_TRANSITION::run_n_times(TRANSITION t, Marking m)
{
    // assert(isEnabledBy(m, t));

    list<Marking> result;

    if(coloured)
    {
     cout << "\nerror: FIRE_TRANSITION::run_n_times - wrong method\n";
     exit(1);
    }
    else
    {
        if(nodes_as_places)
	{
	 // the same for-loop for nodes
	 cout << "PTNET::fireTransition() with nodes as places is not implemented\n";
	 exit(1);
	}
	
	// check omega case
	vector<TCONNECTION> vtc = pn->getPreset(t.getID());
	bool all_omegas = true;
	for(vector<TCONNECTION>::iterator it = vtc.begin();
		it != vtc.end(); it++)
	{
		if(it->weight != 0 && m[it->ID] != OMEGA)
		{
			all_omegas = false;
			break;
		}
	}

	if(all_omegas)
	{
		Marking m1 = run(t,m);
		vtc.clear();
		vtc = pn->getPostset(t.getID());
		for(vector<TCONNECTION>::iterator it = vtc.begin();
			it != vtc.end(); it++)
		{
			if(it->weight != 0)
				m1[it->ID]=OMEGA;
		}

		result.push_back(m1);
		return result;
    	}

	// standard case
	Marking m1 = run(t, m);
	result.push_back(m1);
	while(isEnabledBy(m1, t.getID()))
	{
		m1 = run(t, m1);
		result.push_back(m1);
	}
	
	}
 
    return result;
}

list<TRANSITION> FIRE_TRANSITION::getEnabledTransitions(Marking m)
{
    list<TRANSITION> lt;

    map<IDTYPE, TRANSITION> transitions = pn->getTransitions();

    for(map<IDTYPE, TRANSITION>::iterator i = transitions.begin();
        i != transitions.end(); i++)
    {
      if (isEnabledBy(m, i->second.getID()))
      {
        lt.push_back(i->second);
      }
    }

    return lt;
}

list<TRANSITION> FIRE_TRANSITION::getEnabledTransitions(Marking m, map<IDTYPE, IDTYPE> without_this)
{
    list<TRANSITION> lt;

    map<IDTYPE, TRANSITION> transitions = pn->getTransitions();

    for(map<IDTYPE, TRANSITION>::iterator i = transitions.begin();
        i != transitions.end(); i++)
    {
      // check without_this 
      if(without_this.find(i->first)!=without_this.end())
       continue;

      if (isEnabledBy(m, i->second.getID()))
      {
        lt.push_back(i->second);
      }
    }

    return lt;
}

bool FIRE_TRANSITION::isEnabledBy(Marking marking, IDTYPE transition_id) 
{
	vector<TCONNECTION> preset = pn->getPreset(transition_id);
	map<IDTYPE, EDGE> * edges = pn->getEdgesPointer();
	
	for(vector<TCONNECTION>::iterator i = preset.begin(); i!=preset.end(); i++)
	{
		if(!nodes_as_places)
		 if(!edges->count(i->ID))
		  continue;
			
		if(i->weight > marking[i->ID])
		 return false;
	}
	
	return true;
}

//////////////// coloured ///////////////////////

void FIRE_TRANSITION::initialize(PTNET * pn, ALGORITHM * p_engine)
{
 this->pn = pn;
 this->p_engine = p_engine;
}

MARKING FIRE_TRANSITION::run(TRANSITION t, MARKING m, BINDING b)
{
	if(pn == NULL)
	{
		cout << "\nerror: FIRE_TRANSITION - not initialized\n";
		exit(1);
	}

	MARKING result;

	map<IDTYPE, vector<TOKEN> > result_;	

	if(nodes_as_places)
	{
		// the same for-loop for nodes
		cout << "\nFIRE_TRANSITION with nodes as places is not implemented\n";
		exit(1);
	}

	// engine
	p_engine->initialize(b);
	
	// edges
	map<IDTYPE, EDGE> * edges = pn->getEdgesPointer();

	// preset
	vector<TCONNECTION> preset = pn->getPreset(t.getID());

	// postset
	vector<TCONNECTION> postset = pn->getPostset(t.getID());

	
	for(map<IDTYPE, EDGE>::iterator i = edges->begin(); i!=edges->end(); i++)
	{
		// marking on edge
		vector<TOKEN> tokens = m.getMarking(i->first);

		// preset
		list<EXPRESSION> preset_attributes;
		
		// find in preset
		for(vector<TCONNECTION>::iterator it = preset.begin(); it != preset.end(); it++)
		{
			unsigned preset_w = 0;

			if(it->ID == i->first)
			{
				preset_attributes = it->attributes;
				preset_w = it->weight;
			}
		
			if(preset_w > 0)
			{
				// is there any tokens ?
				if(tokens.size() == 0)
				{
					cout << "\nerror: FIRE_TRANSITION - not enough tokens\n";
					exit(1);
				}
	
				// evaluate preset attributes
				list<EXPRESSION> preset_attributes_evaluated;
				for(list<EXPRESSION>::iterator it_pre = preset_attributes.begin(); it_pre != preset_attributes.end(); it_pre++)
				{
					preset_attributes_evaluated.push_back(p_engine->run(*it_pre));
				}
		
				// find token with this attributes
				unsigned int count_pre = 0;
				for(count_pre = 0; count_pre < tokens.size(); count_pre++)
				{
					if(EXPRESSION::equal(preset_attributes_evaluated, tokens[count_pre].getAttributes()))
						break;
				}
		
				// not found (not attributed preset)
				if(count_pre == tokens.size())
				count_pre=0;
	
				// remove tokens
				if(tokens[count_pre].getCount() != OMEGA)
				{
					int count = tokens[count_pre].getCount() - preset_w;
	
					// not enough tokens
					if(count < 0)
					{
						cout << "\nerror: FIRE_TRANSITION - not enough tokens\n";
						exit(1);
					}
	
					// delete if count == 0
					if(count == 0)
					{
						vector<TOKEN>::iterator itd = tokens.begin();
						for(int c=0; c<count_pre; c++)
							itd++;
						
						tokens.erase(itd);
					}
					else
					{
						TOKEN t = tokens[count_pre];
						t.setCount(count);
						tokens[count_pre] = t;
					}
				}
			}
		}

		// postset
		list<EXPRESSION> postset_attributes;
		
		// find in postset
		for(vector<TCONNECTION>::iterator it = postset.begin(); it != postset.end(); it++)
		{
			unsigned postset_w = 0;

			if(it->ID == i->first)
			{
				postset_attributes = it->attributes;
				postset_w = it->weight;
			}
				
			if(postset_w > 0)
			{
				// evaluate postset attributes
				list<EXPRESSION> postset_attributes_evaluated;
				for(list<EXPRESSION>::iterator it_post = postset_attributes.begin(); it_post != postset_attributes.end(); it_post++)
				{
					postset_attributes_evaluated.push_back(p_engine->run(*it_post));
				}
		
				// find token with this attributes
				unsigned int count_post = 0;
				for(count_post = 0; count_post < tokens.size(); count_post++)
				{
					if(EXPRESSION::equal(postset_attributes_evaluated, tokens[count_post].getAttributes()))
						break;
				}
		
				// not found
				if(count_post == tokens.size())
				{
					TOKEN t;
					t.setAttributes(postset_attributes_evaluated);
					t.setCount(postset_w);
					tokens.push_back(t);
				}
				// add tokens
				else if(tokens[count_post].getCount() != OMEGA)
				{
					TOKEN t = tokens[count_post];
					t.setCount(t.getCount() + postset_w);
					tokens[count_post] = t;
				}
			}
		}
		
		if(tokens.size()>0)
		 result_[i->first] = tokens;
		else
		 result_.erase(i->first);
	}

	result.setMarkings(result_);
	return result;
}

bool FIRE_TRANSITION::isEnabledBy(MARKING marking, IDTYPE transition_id, BINDING b)
{
	if(pn == NULL)
	{
		cout << "\nerror: FIRE_TRANSITION - not initialized\n";
		exit(1);
	}

	if(nodes_as_places)
	{
		// the same for-loop for nodes
		cout << "\nFIRE_TRANSITION with nodes as places is not implemented\n";
		exit(1);
	}
	
	// engine
	// ALGORITHM * p_engine = storage.getEngine();
	p_engine->initialize(b);

	// check guards
	if(! p_engine->run(pn->getTransition(transition_id).getGuard()))
	 return false;

	// preset
	vector<TCONNECTION> preset = pn->getPreset(transition_id);

	// find in preset
	for(vector<TCONNECTION>::iterator it = preset.begin(); it != preset.end(); it++)
	{
		// marking on edge
		vector<TOKEN> tokens = marking.getMarking(it->ID);

		if(it->weight > 0)
		{
			// is there any tokens ?
			if(tokens.size()==0)
			 return false;

			// evaluate preset attributes
			list<EXPRESSION> preset_attributes_evaluated;
			for(list<EXPRESSION>::iterator it_pre = it->attributes.begin(); it_pre != it->attributes.end(); it_pre++)
			{
				preset_attributes_evaluated.push_back(p_engine->run(*it_pre));
			}

			// find token with this attributes
			unsigned int count_pre = 0;
			for(count_pre = 0; count_pre < tokens.size(); count_pre++)
			{
				if(EXPRESSION::equal(preset_attributes_evaluated, tokens[count_pre].getAttributes()))
					break;
			}

			// not found (not attributed preset)
			if(count_pre == tokens.size())
			 count_pre=0;

			// count
			if(tokens[count_pre].getCount() >= it->weight)
			 continue;
			else
			 return false;
		}
	}

	return true;
}

list<BINDING> FIRE_TRANSITION::getEnabledBindings(MARKING marking, IDTYPE transition_id)
{
	if(pn == NULL)
	{
		cout << "\nerror: FIRE_TRANSITION - not initialized\n";
		exit(1);
	}

	list<BINDING> result;

	BINDER binder;
	binder.initialize(pn);

	list<BINDING> bindings = binder.run(&marking, transition_id);

	for(list<BINDING>::iterator it = bindings.begin(); it != bindings.end(); it++)
	{
		if(isEnabledBy(marking, transition_id, *it))
			result.push_back(*it);
	}

	return result;
}

list<pair<TRANSITION, BINDING> > FIRE_TRANSITION::getEnabledTransitions(MARKING m)
{
	list<pair<TRANSITION, BINDING> > lt;

	map<IDTYPE, TRANSITION> transitions = pn->getTransitions();
	
	for(map<IDTYPE, TRANSITION>::iterator i = transitions.begin();
		i != transitions.end(); i++)
	{
		list<BINDING> lb = getEnabledBindings(m, i->second.getID());
		
		for(list<BINDING>::iterator it = lb.begin(); it != lb.end(); it++)
			lt.push_back(pair<TRANSITION, BINDING>(i->second, *it));
	}

	return lt;
}

list<pair<TRANSITION, BINDING> > FIRE_TRANSITION::getEnabledTransitions(MARKING m, map<IDTYPE, IDTYPE> without_this)
{
	list<pair<TRANSITION, BINDING> > lt;

	map<IDTYPE, TRANSITION> transitions = pn->getTransitions();
	
	for(map<IDTYPE, TRANSITION>::iterator i = transitions.begin();
		i != transitions.end(); i++)
	{
		// check without_this 
		if(without_this.find(i->first)!=without_this.end())
			continue;
		
		list<BINDING> lb = getEnabledBindings(m, i->second.getID());
		
		for(list<BINDING>::iterator it = lb.begin(); it != lb.end(); it++)
			lt.push_back(pair<TRANSITION, BINDING>(i->second, *it));
	}

	return lt;
}


list<MARKING> FIRE_TRANSITION::run_n_times(TRANSITION t, MARKING m, BINDING b)
{
	list<MARKING> result;
	
	if(nodes_as_places)
	{
		// the same for-loop for nodes
		cout << "PTNET::fireTransition() with nodes as places is not implemented\n";
		exit(1);
	}

	map<IDTYPE, vector<TOKEN> > marking = m.getMarkings();
	
	// check omega case
	vector<TCONNECTION> vtc = pn->getPreset(t.getID());

	bool all_omegas = true;
	bool break_ = false;

	for(vector<TCONNECTION>::iterator it = vtc.begin(); it != vtc.end(); it++)
	{
		if(break_)
		 break;

		if(it->weight != 0)
		{
			vector<TOKEN> vt = marking[it->ID];
			for(vector<TOKEN>::iterator itt = vt.begin(); itt != vt.end(); itt++)
			{
				if(itt->getCount() != OMEGA)
				{
					all_omegas = false;
					break_ = true;
					break;
				}
			}
		}
	}

	if(all_omegas)
	{
		MARKING m1 = run(t,m,b);
		map<IDTYPE, vector<TOKEN> > marking1 = m1.getMarkings();
		vtc.clear();
		vtc = pn->getPostset(t.getID());
		for(vector<TCONNECTION>::iterator it = vtc.begin(); it != vtc.end(); it++)
		{
			if(it->weight != 0)
			{
				vector<TOKEN> vt = marking1[it->ID];
				for(vector<TOKEN>::iterator itt = vt.begin(); itt != vt.end(); itt++)
					itt->setCount(OMEGA);
				
				marking1[it->ID] = vt;
			}
		}

		m1.setMarkings(marking1);

		result.push_back(m1);
		return result;
    	}

	// standard case
	MARKING m1 = run(t, m, b);
	result.push_back(m1);
	while(isEnabledBy(m1, t.getID(), b))
	{
		m1 = run(t, m1, b);
		result.push_back(m1);
	}
	
    return result;
}

////////////// namespaces ////////////// 

}

}


