/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
  
#ifndef ALGORITHM_FOLD_CPP
#define ALGORITHM_FOLD_CPP

#include "fold.h"
#include "algorithm_storage.h"

namespace algorithms
{
   namespace petrigraph
   {
      FOLD::FOLD()
      {
         this->ID = ALGORITHM_FOLD;
         this->name = "fold";
	 this->petrigraph = NULL;
      }
      FOLD::~FOLD() {}
      
      void FOLD::initialize(PETRIGRAPH* pg)
      {
         this->petrigraph = pg;
      }
      
      void FOLD::run(RULE* r, MATCH phi2, MATCH phi)
      {
         HYPERGRAPH* left = r->getLeft();
	 map<IDTYPE, EDGE> edges_left = left->getEdges();
	 set<pair<IDTYPE, IDTYPE> > emptymap;
	 
	 set<pair<IDTYPE, IDTYPE> > eunion;
	 map<IDTYPE, EDGE>::iterator eit;
	 for(eit = edges_left.begin(); eit != edges_left.end(); eit++)
	 {
	    if(phi.hasMatch(eit->first) && phi2.hasMatch(eit->first))
	    {
	       IDTYPE phi2_e = phi2.getMatch(eit->first);
	       IDTYPE phi_e = phi.getMatch(eit->first);
	       if(phi_e != phi2_e) eunion.insert(pair<IDTYPE, IDTYPE>(phi2_e, phi_e));
	    }
	 }
	 
	 // set data for coverability graph (for incremental coverability)
	 storage.setData("coverability", NULL);
	 
	 ALGORITHM* merge = this->load("call_merge");
	 merge->initialize((PETRIGRAPH*)(this->petrigraph));
	 merge->run(emptymap, eunion, emptymap);
      }

   };
};
using namespace algorithms;
using namespace algorithms::petrigraph;

#endif
