/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#ifndef TYPEDEF_GLOBALS_CPP
#define TYPEDEF_GLOBALS_CPP

#include <string.h>
#include "globals.h"

namespace special
{
   namespace param
   {
      bool debug = false;
      bool quiet = false;
//      bool test = false;
      bool ncov = false;
      bool nfold = false;
//      bool alternative = false;
//      bool fullConflict = false;
//      bool concurrency = false;
//      bool directConflict = false;
//      bool context = false;
//      bool cutoff = false;
      bool pid = false;
      bool red = true;
      bool mapping = true;
      bool trace = false;
      bool remove_isolated_nodes = true;
      bool force_alternation = false;
      unsigned long depth = 0;
      unsigned long unfdepth = MINUS_ONE;
      unsigned long maxstep = MINUS_ONE;
      bool coloured = false;
      bool nodes_as_places = false;
      bool full_coverability_graph = false;
      unsigned long iteration = 0;
      unsigned long clear_cov_count=0;
      bool cov_test = false;
      double rules_size = 0.5;
      unsigned long ref_loop_limit = 4;
      unsigned long mod_base = 2;
      unsigned long interval_from = 0;
      unsigned long interval_to = 2;
      unsigned long attr_ref_count = 3;
      unsigned long timeout = 0;
      bool fullConflict = false;
      bool concurrency = false;
      bool directConflict = false;
      bool context = false;
      bool cutoff = false;
      bool exCerts = false;

   };

ostream& operator<<(ostream& os, std::map<IDTYPE,IDTYPE>& m)
{
   for(map<IDTYPE,IDTYPE>::iterator i = m.begin(); i!=m.end(); i++)
      os << (*i).first << "->" << (*i).second << "  ";
   return os;
}

double global_timer = 0.0;

time_t global_timer_sec = 0;

void init_time()
{
	global_timer = (double)((double)(clock()) / (double)CLOCKS_PER_SEC);
	global_timer_sec = time(NULL);
}

double get_time()
{
	double result = (double)((double)(clock()) / (double)CLOCKS_PER_SEC) - global_timer;
	double global_timer_sec_ = time(NULL) - global_timer_sec;

	if(global_timer_sec_ > 1000)
		return (double)global_timer_sec_;

	if(result<0.00001)
		result=0.0;

	return (double)result;
}

// neato
void callNeato(char * dot_File, char * ps_File, NEATO_CALL_TYPE t)
{
	switch(t)
	{
		case STANDARD:
			string s = string("neato -Tps -Gspline=true -Goverlap=false ") +
					string(dot_File) + string(" >") +
					string(ps_File);
			system(s.c_str());
			break;
	}
}
// dot
void callDot(char * dot_File, char * ps_File, NEATO_CALL_TYPE t)
{
	switch(t)
	{
		case STANDARD:
			string s = string("dot -Tps -Goverlap=false ") +
					string(dot_File) + string(" >") +
					string(ps_File);
			system(s.c_str());
			break;
	}
}

// latex
void callLatex(char * tex_File, char * ps_File)
{
	string s = string("latex ") + string(tex_File);
	system(s.c_str());

	string dviFile;
	char ch[strlen(tex_File)+1];
	strcpy(ch, tex_File);
	char * pch;
	pch = strtok(ch,".");
	
	while (pch!=NULL && strcmp(pch,"tex")!=0)
  	{
		if(dviFile.size()!=0)
			dviFile += (string)(".");
		dviFile += (string)pch;
    		pch = strtok (NULL, ".");
  	}

	dviFile += (string)(".dvi");

	s=string("dvips -o ") + string(ps_File) + string(" ") + dviFile;
	system(s.c_str());
}

string substr(string input, int begin, int end){
	int size = end-begin;
	char buffer[size+2];
	for(int i = begin, k = 0; i < input.size() && i <= end && k <= size; i++, k++)
		buffer[k] = input[i];
	buffer[size+1] = 0;
	return string(buffer);
}

};

using namespace special;
using namespace special::param;

namespace typedefs
{

// methods for the typedef Marking
// ideally should not be in global, 
// but let us make some exception      
bool isEqual(const Marking& m1, const Marking& m2)
{
    if(m1.size()!=m2.size())
     return false;    

    for(Marking::const_iterator i = m1.begin();
        i!= m1.end();
        i++)
    {      
      unsigned found_marking = i->second;
      Marking::const_iterator loop_marking_it = m2.find(i->first);
      unsigned loop_marking = 0;
      if(loop_marking_it != m2.end())
       loop_marking_it->second;
      if(found_marking != loop_marking) // mismatch
       return false;
    }
    return true;
}

bool isLower(const Marking& m1, const Marking& m2)
{
    if(m1.size()>m2.size())
     return false;
      
    for(Marking::const_iterator i=m1.begin();
        i != m1.end();
        i++)
    {
      unsigned m1_marking = i->second;
      Marking::const_iterator m2_marking_it = m2.find(i->first);
      unsigned m2_marking = 0;
      if(m2_marking_it != m2.end())
       m2_marking = m2_marking_it->second;
      if(m1_marking>m2_marking)
        return false;
    }
    return true;
}

bool isStrictlyLower(const Marking& m1, const Marking& m2)
{
    bool different = false; // set to true if any difference is found
    
    if(m1.size()>m2.size())
     return false; 
    
    for(Marking::const_iterator i=m1.begin();
        i != m1.end();
        i++)
      {
        unsigned m1_marking = i->second;
        Marking::const_iterator m2_marking_it = m2.find(i->first);
        unsigned m2_marking = 0;
	if(m2_marking_it != m2.end())
	  m2_marking = m2_marking_it->second;
        if(m1_marking>m2_marking)
          return false;
        if(m1_marking < m2_marking)
          different = true;
      }
      
    return (different || (m1.size()<m2.size()));
     
}

ostream& operator<<(ostream& ost, Marking& th)
{
 ost << "marking" << endl;

 for(Marking::iterator it = th.begin(); it != th.end(); it++)
	ost << it->first << " " << it->second << endl;

 return ost;
}

void writePPID()
{
	if(!pid)
	 return;
	string s = string("echo %% $PPID");	
	system(s.c_str());
}

}

#endif
