/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef TYPEDEF_GLOBALS_H
#define TYPEDEF_GLOBALS_H 

#define MINUS_ONE (unsigned long)(-1)
 
  // common includes
  #include <iostream>
  #include <cstdlib>
  #include <assert.h>
 
  // it is a common include of STL for all classes, which use it
  #include <vector>
  #include <string>
  #include <list>
  #include <map>
  #include <set>
  #include <stack>
  #include <stdexcept>  
  #include <fstream>
  #include <string>

  using namespace std;

  // global typedefs are collected here
  namespace typedefs {
  
    // possible types by expression
    typedef enum{INTEGER, STRING, BOOL} T;
    typedef vector<T> TYPES;
    
    /*
     * graph element types
     */
    typedef enum{IS_VERTEX, IS_EDGE, IS_TRANSITION} GRAPH_ELEMENT_TYPE;
    
    const TYPES NONE;
    
    // type for ID numbers
    typedef unsigned int IDTYPE;
    
    // type of labels in the run() of Algorithms
    typedef string LABEL;

    typedef list<pair<string, LABEL> > PATH_DB;

    typedef enum{ACCESS, CHANGE, DELETE, MERGE, CHANGE_ATTRIBUTE} ATYPES;

    typedef enum{DOT, GTXL, GXL, GXL_PETRIGRAPH, LLNET, PLAIN_TEXT, PS, XML_GENERIC} DATA_FORMAT;

    struct EVENT
    {
       ATYPES type;
       map<IDTYPE, IDTYPE> ids;
    };

    typedef enum{STANDARD} NEATO_CALL_TYPE; 

    // not coloured marking, have only number of tokens
    typedef map<IDTYPE, unsigned> Marking;

    // expression evaluated by some engine
    struct EXPR_VALUE
    {
	EXPR_VALUE()
	{
		int_val = 0;
		str_val = "";
		bool_val = true;
	}

	int int_val;
	string str_val;
	bool bool_val;
    };

    typedef enum {FULL, NORMAL, EVALUATED} PRINT_TYPE;

    bool isEqual(const Marking& m1, const Marking& m2);
    bool isLower(const Marking& m1, const Marking& m2);
    bool isStrictlyLower(const Marking& m1, const Marking& m2);
    ostream& operator<<(ostream& ost, Marking& th);

    void writePPID();

  };
  using namespace typedefs;

  namespace special
  {
     namespace param
     {
        extern bool debug;
        extern bool quiet;
//        extern bool test;
        extern bool ncov;
        extern bool nfold;
//        extern bool alternative;
//        extern bool fullConflict;
//        extern bool concurrency;
//        extern bool directConflict;
//        extern bool context;
//        extern bool cutoff;
	extern bool pid;
        extern bool red;
        extern bool mapping;
	extern bool trace;
	extern bool remove_isolated_nodes;
	extern bool force_alternation;
        extern unsigned long depth;
        extern unsigned long unfdepth;
        extern unsigned long maxstep;
	extern bool coloured;
	extern bool nodes_as_places;
	extern bool full_coverability_graph;
	// global counter, not a parameter
	extern unsigned long iteration;
	extern unsigned long clear_cov_count;
	extern bool cov_test;
	extern double rules_size;
	extern unsigned long ref_loop_limit;
	extern unsigned long mod_base;
	extern unsigned long interval_from;
	extern unsigned long interval_to;
	extern unsigned long attr_ref_count;
	extern unsigned long timeout;
	extern bool fullConflict;
        extern bool concurrency;
        extern bool directConflict;
        extern bool context;
        extern bool cutoff;
	extern bool exCerts;
     };

  // cout any map
  ostream& operator<<(ostream& os, std::map<IDTYPE,IDTYPE>& m); 

  extern double global_timer;

  extern time_t global_timer_sec;
 
  void init_time();
  double get_time();

  // neato
  void callNeato(char * dot_File, char * ps_File, NEATO_CALL_TYPE t=STANDARD);
  // dot
  void callDot(char * dot_File, char * ps_File, NEATO_CALL_TYPE t=STANDARD);

  // latex
  void callLatex(char * tex_File, char * ps_File);

class NoSuchElementException : public exception{
public:
	NoSuchElementException(){}
};

 template<class T>
	T get(string id, map<string, T>& m){
		for(typename map<string,T>::iterator iter = m.begin();
			iter != m.end(); iter++){
			if(iter->first == id)
				return iter->second;
		}
		throw NoSuchElementException();
	}

 template<class K, class V> V get_2(map<K,V> m, K key){
		for(typename map<K,V>::iterator iter = m.begin();
			iter != m.end(); iter++){
			if(iter->first == key)
				return iter->second;
		}
	}

 string substr(string input, int begin, int end);

  };

  using namespace special;
  using namespace special::param;

#endif
