/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>

#include "graph.h"

namespace data_structures
{
   GRAPH::GRAPH()
   {
   	this->lastUsedID = 0;
	this->id = 0;
	this->name = "";
   }

   GRAPH::GRAPH(map<IDTYPE, VERTEX> vertices,
                map<IDTYPE, EDGE> edges, 
	        string name,
	        IDTYPE id,
	        IDTYPE lastUsedID)
   {
   	this->vertices=vertices;
	this->edges=edges;
	this->lastUsedID = lastUsedID;
	this->id = id;
	this->name = name;
   }

   GRAPH::~GRAPH()
   {
   }

   map<IDTYPE, VERTEX> GRAPH::getVertices() { return this->vertices; }
   map<IDTYPE, EDGE> GRAPH::getEdges() { return this->edges; }

   map<IDTYPE, VERTEX> * GRAPH::getVerticesPointer() { return &(this->vertices); }
   map<IDTYPE, EDGE> * GRAPH::getEdgesPointer() { return &(this->edges); }

   void GRAPH::setName(string s) 
   {
   	this->name = s;
   }

   void GRAPH::setID(IDTYPE ID) 
   {
   	this->id = ID;
   }

   void GRAPH::setID()
   {
   	id = ++lastUsedID;
   }

   IDTYPE GRAPH::getLastUsedID()
   {
   	return lastUsedID;
   }

   void GRAPH::setLastUsedID(IDTYPE id)
   {
   	lastUsedID = id;
   }
   string GRAPH::getName() 
   {
   	return this->name;
   }

   IDTYPE GRAPH::getID() 
   {
   	return this->id;
   }

   EDGE GRAPH::getEdge(IDTYPE id)
   { 
      assert(edges.count(id) > 0);     
      return this->edges[id];
   }

   bool GRAPH::hasEdge(IDTYPE id)
   {
      return (this->edges.count(id) > 0);
   }

   bool GRAPH::hasEdge(string s)
   {
      bool ch = false;	
      for(map<IDTYPE, EDGE>::iterator it = this->edges.begin(); it != this->edges.end(); it++)
      {
      	if (!strcmp(((it->second).getIdLabel()).c_str(), s.c_str()))
		{ ch = true;}
      }
      return ch;	
   }

   EDGE GRAPH::getEdge(string s)
   {
      map<IDTYPE, EDGE> es = this->edges;
      for(map<IDTYPE, EDGE>::iterator it = es.begin(); it != es.end(); it++)
      {
      	if (!strcmp(((it->second).getIdLabel()).c_str(), s.c_str()))
		return it->second;
      }
//      cerr << "warning: edge with label '" << s << "' not found.\n";
      return EDGE();
   }

   VERTEX GRAPH::getVertex(IDTYPE id)
   {
      assert(this->vertices.count(id) > 0);
      return this->vertices[id];
   }

   bool GRAPH::hasVertex(IDTYPE id)
   { 
      return (this->vertices.count(id) > 0);
   }
 
   bool GRAPH::hasVertex(string s)
   {
      bool ch = false;	
      for(map<IDTYPE, VERTEX>::iterator it = this->vertices.begin(); it != this->vertices.end(); it++)
      {
      	if (!strcmp(((it->second).getIdLabel()).c_str(), s.c_str()))
		{ ch = true;}
      }
      return ch;	
   }

   VERTEX GRAPH::getVertex(string s)
   {
	map<IDTYPE, VERTEX> vs = this->vertices;
      for(map<IDTYPE, VERTEX>::iterator it = vs.begin(); it != vs.end(); it++)
      {
      	if (!strcmp(((it->second).getIdLabel()).c_str(), s.c_str()))
		return it->second;
      }
//      cerr << "warning: vertex with label '" << s << "' not found.\n";
      return VERTEX();
   }
	
   void GRAPH::clean(){
	
	vertices.clear();
        edges.clear(); 
	name="";
	id=0;
	lastUsedID=0;
   }
};
