/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>

#include "gts.h"

namespace data_structures
{
   GTS::GTS()
   {
   }

   GTS::~GTS()
   {
   }

   GTS GTS::clone()
   {
      GTS gg = GTS();
      gg.rules = this->rules;
      gg.startgraph = this->startgraph;
      gg.typegraph = this->typegraph;
      return gg;
   }

   map<IDTYPE, RULE>* GTS::getRules()
   {
      return &(this->rules);
   }
   
   RULE* GTS::getRule(IDTYPE id)
   {
      assert(this->rules.count(id) > 0);
      return &(this->rules[id]);
   }
   
   RULE* GTS::getRule(string s)
   {
   	map<IDTYPE, RULE> *r = this->getRules();
   	for(map<IDTYPE, RULE>::iterator it = r->begin(); it != r->end(); it++)
	{
		string s1 = (it->second).getName();
		if((!strcmp(s1.c_str(), s.c_str())))
			return &(it->second);
	}
	
//	cerr << "warning: rule with label '" << s << "' not found.\n";

        return new RULE;
   }
   
   bool GTS::hasRule(IDTYPE id)
   {
      return (this->rules.count(id) > 0);
   }
   
   void GTS::setRules(map<IDTYPE, RULE>* rulemap)
   {
      this->rules = *rulemap;
   }
   
   void GTS::addRule(RULE* r)
   {
      assert(this->rules.count(r->getID()) == 0);
      this->rules[r->getID()] = *r;
   }
   
   void GTS::deleteRule(IDTYPE id)
   {
      this->rules.erase(id);
   }

   HYPERGRAPH* GTS::getStartgraph()
   {
      return &(this->startgraph);
   }
   
   void GTS::setStartgraph(HYPERGRAPH* g)
   {
      this->startgraph = *g;
   }
   
   HYPERGRAPH* GTS::getTypegraph()
   {
      return &(this->typegraph);
   }
   
   void GTS::setTypegraph(HYPERGRAPH* g)
   {
      this->typegraph = *g;
   }
   
   ostream& operator<<(ostream& ost, GTS& th)
   {
      ost << "graph grammar with the following elements\n---\nstart graph\n" << th.startgraph;
      map<IDTYPE, RULE>::iterator ruleit;
      for(ruleit = th.rules.begin(); ruleit != th.rules.end(); ruleit++)
      {
         ost << "\n---graph rewriting rule with ID " << ruleit->first << "\n" << ruleit->second;
      }
      return ost;
   }
};
