/***************************************************************************
 *   Copyright (C) 2006 by root   *
 *   root@Vitali   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "gts_emulator.h"
#include "find_all_matches.h"
#include "matchhandler.h"
#include "apply_rule.h"
#include "vis_hypergraph.h"
#include "binder.h"
#include "algorithm_storage.h"
#include "attr_example_checker.h"

GTS_EMULATOR::GTS_EMULATOR()
{
      this->ID = ALGORITHM_GTS_EMULATOR;
      this->name = "gts_emulator";
      
      this->inspec.push_back(pair<string, DATA_FORMAT>("grammar", GTXL));
      
      this->call_descriptor = "Usage: augur -sc=gts_emulator [options] grammar \n\
grammar\tgraph grammar in GTXL format\n\n";
}


GTS_EMULATOR::~GTS_EMULATOR()
{
}

void GTS_EMULATOR::run()
{
	// nodes as places (attributes in nodes) is not yet supported
	if(nodes_as_places)
	{
		cout << "\nnodes as places (attributes in nodes) is not yet supported\n";
		exit(1);
	}

	// attribute engine
	engine = NULL;
	if(coloured)
	{
		engine = storage.getEngine();
	}

	// read grammatik
	GTS grammar;
	ALGORITHM* reader = this->load("call_gtxl_reader");
	reader->initialize((char*)(this->filemap["grammar"]).c_str());
	reader->run(grammar);

	HYPERGRAPH hg = *(grammar.getStartgraph());

	// visualize initial graph with attributes
	VIS_HYPERGRAPH vh;
	if(coloured)
	{
		vh.initialize("hg0.ps", engine);
		vh.run(hg, EVALUATED);
	}
	else
	{
		vh.initialize("hg0.ps");
		vh.run(hg, FULL);

	}	

	map<IDTYPE, RULE>* rules = grammar.getRules();

	HYPERGRAPH hgresult;

	bool exit_=false;

	int iteration = 1;

	while(!exit_)
	{
		vector<BINDING_MATCH> found_matches;
		vector<BINDING_MATCH> binded_matches;
		
		// check all rules
		for(map<IDTYPE, RULE>::iterator it =(*rules).begin();
				it != (*rules).end(); it++)
		{
			FIND_ALL_MATCHES fam;
			TRIVIAL_MATCHHANDLER mh;
	
			fam.initialize(&hg);
			list<MATCH> lm = fam.run(it->second.getLeft(), (ALGORITHM*)(&mh), false);

			for(list<MATCH>::iterator itm = lm.begin(); itm != lm.end(); itm++)
			  found_matches.push_back(BINDING_MATCH((*it).second, *itm, BINDING()));
		}
		
		if(coloured)
		{		
			// find binding
			for(vector<BINDING_MATCH>::iterator itfm = found_matches.begin(); itfm != found_matches.end(); itfm++)
			{
				BINDER binder;
				binder.initialize(&hg);
				list<BINDING> lb = binder.run(itfm->rule.getLeft(), &(itfm->match));
				
				// empty binding
				if(lb.size()==0)
				{
					BINDING_MATCH bm = *itfm;
					binded_matches.push_back(bm);
				}
				
				for(list<BINDING>::iterator itl = lb.begin(); itl != lb.end(); itl++)
				{
					if(checkBinding(*itl, itfm->rule.getGuard()))
					{
						BINDING_MATCH bm = *itfm;
						bm.binding = *itl;
						binded_matches.push_back(bm);
					}
				}
			}
		}
		else
		{
			binded_matches = found_matches;
		}

		// communication with user
		int counter = 0;
		if(binded_matches.size() == 0)
		{
			if(found_matches.size() == 0)
				cout << "\nNo rule can be applied (No match found)\n";
			else
				cout << "\nNo rule can be applied (Guards are violated)\n";
			exit(1);
		}
		else
		{
			cout << endl << "Step " << iteration << ":\n" << "Please choose rule to be applied "<< endl;
			int count = 1;
			for(vector<BINDING_MATCH>::iterator itb = binded_matches.begin(); itb != binded_matches.end(); itb++)
			{
				// evaluate binding
				list<string> names = itb->binding.getVariableNames();
				BINDING b;
				for(list<string>::iterator itn = names.begin(); itn != names.end(); itn++)
				{					
					ExpressionElement ee = itb->binding.getVariable(*itn);
					engine->evaluate(ee);
					b.setVariable(*itn, ee);
				}
				cout << count << ": " << itb->rule.getName() << " % " << b;
				count++;
			}
		
			cout << "-1: Exit\n";

			// read input	
			cin >> counter;

			if(counter==-1)
			 exit(0);

			if(counter > 0)
			 counter--;
			else
			 counter = 0;
			
			if(counter >= binded_matches.size())
			 counter = 0;			
		}

		// apply rule
		APPLY_RULE ar;
		
		if(coloured)
		{
			ar.initialize(&hg, engine);
			hgresult = ar.run(&(binded_matches[counter].rule), binded_matches[counter].match, binded_matches[counter].binding);
		}
		else
		{
			ar.initialize(&hg, engine);
			hgresult = ar.run(&(binded_matches[counter].rule), binded_matches[counter].match);
		}

		hg = hgresult;

		char buffer[20];
		sprintf(buffer, "%i", iteration);
		string sf = string("hg") + string(buffer) + string(".ps");

		if(coloured)
		{
			vh.initialize((char*)sf.c_str(), engine);
			vh.run(hg, EVALUATED);
		}
		else
		{
			vh.initialize((char*)sf.c_str());
			vh.run(hg, FULL);
		}

		iteration++;
	}
}

bool GTS_EMULATOR::checkBinding(BINDING b, list<EXPRESSION> guards)
{
	if(engine == NULL)
	{
		cout << "\nGTS_EMULATOR :: No Attribute Engine is set - internal error\n";
		exit(1);
	}

	engine->initialize(b);
	return engine->run(guards);
}
