#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <stdexcept>
#include <string.h>

#include "rule.h"
#include "gxlreader.h"

#include "match.h"

namespace algorithms{
namespace readers {

using namespace data_structures;

GXLREADER::GXLREADER()
{

}


GXLREADER::~GXLREADER()
{
	xmlFreeDoc(doc);
}

void GXLREADER::initialize(char *filename)
{

    doc = xmlParseFile(filename);
	
    if (doc == NULL ) 
      {
	fprintf(stderr,"Document not parsed successfully. \n");
	xmlFreeDoc(doc);
	exit(1);
      }
	
	
    if (xmlStrcmp(xmlDocGetRootElement(doc)->name, (const xmlChar *) "gxl")) 
      {
	fprintf(stderr,"document of the wrong type, root node != gxl");
	xmlFreeDoc(doc);
	exit(1);
      }
}

xmlNodePtr GXLREADER::GetAttributeNode(xmlNodePtr node, char* name)
{
	xmlNodePtr cur = GetChildrenNode(node, "attr");
	while(cur != NULL)
		{
			if (!(xmlStrcmp(xmlGetProp(cur, (const xmlChar*)"name"), (const xmlChar*)name)))
				return cur;
			else
				cur = GetNextChild(cur, "attr");
		}
	return cur;
}


xmlNodePtr GXLREADER::GetChildrenNode(xmlNodePtr parent, char* name_child)
{
	xmlNodePtr child = parent->xmlChildrenNode;
	while((child != NULL) && (xmlStrcasecmp(child->name, (const xmlChar*)name_child)))
	{
		child = child->next;
	}
	
	return child;	
}

xmlNodePtr GXLREADER::GetRelend(xmlNodePtr rel, int order)
{
	string s;
	xmlNodePtr relend = GetChildrenNode(rel, "relend");
	while(relend != NULL)
	{
		string s_role = (string)(char*)xmlGetProp(relend,(const xmlChar *)"role");
		if(!strcmp(s_role.c_str(), "vertex"))
		{
			s = (string)(char*)xmlGetProp(relend,(const xmlChar *)"startorder");
			if(atoi(s.c_str()) == order)
			return relend;
		}
		relend = GetNextChild(relend, "relend");
	}
	
	return NULL;
}


xmlNodePtr GXLREADER::GetNextChild(xmlNodePtr child1, char* name_child)
{
	
	xmlNodePtr child2 = child1->next;
	while((child2 != NULL) && (xmlStrcasecmp(child2->name, (const xmlChar*)name_child)))
	{
		child2 = child2->next;
	}
	
	return child2;
	
	
}

void GXLREADER::run(PETRIGRAPH &PtGraph, GTS &GraphGrammar)
{

try
{
	PtGraph.setGrammar(&GraphGrammar);
	xmlNodePtr root = xmlDocGetRootElement(doc);
	xmlNodePtr graph = GetChildrenNode(root, "graph");
	if(graph == NULL)
	{
		cout << "GXLREADER: Node 'graph' is expected.\n ";
		exit(1);
	}
	
	xmlNodePtr node = GetChildrenNode(graph, "node");
	
	string s;
	char* ch;
	char* ch_map;
	MATCH iota_match;
	MATCH mu_map;
	//VERTEX v;
	map<IDTYPE, MATCH> mu;
	//for iota_mapping (vertex)
	map<IDTYPE, IDTYPE> matching;
	map<IDTYPE, IDTYPE> mapping;
	list<IDTYPE> connected_vertices;
	typedef pair<IDTYPE, MATCH> mu_type;
	typedef pair<IDTYPE, IDTYPE> matching_type;
	IDTYPE id, id1, id2, id_match;
	
	while(node != NULL)
	{
		xmlNodePtr attr = GetAttributeNode(node, "vertex");
		if(attr != NULL)
			{
				VERTEX v;
				
				// set vertex id
				string s = (string)(char*)xmlGetProp(node,(const xmlChar *)"id");
				s.erase(0, 1);
				id = atoi(s.c_str());
				PtGraph.addVertex(v, id);
				
				xmlNodePtr iotaMap = GetAttributeNode(node, "iota_mapping");
				if(iotaMap != NULL)
				{
					xmlNodePtr str_node = GetChildrenNode(iotaMap, "string");
					if((str_node != NULL)&&(str_node->xmlChildrenNode != NULL)) 
						{
							s = (string)(char*)str_node->xmlChildrenNode->content;
							if (s != "")
    							{
								ch = strtok((char*)s.c_str(), ";");

								while(ch!=NULL)
								{
									v = (GraphGrammar.getStartgraph())->getVertex(ch);
									id_match = v.getID();
									matching.insert (matching_type(id_match, id));
									ch = strtok(NULL, ";");
								}
							}
							
						}
				}
			}
		
		attr = GetAttributeNode(node, "transition");
		if(attr != NULL)
		{
			TRANSITION tr;
			s = (string)(char*)xmlGetProp(node,(const xmlChar *)"id");
			s.erase(0,1);
			id = atoi(s.c_str());
				
			RULE *rule;
			xmlNodePtr tr_name = GetAttributeNode(node, "rule");
			if (tr_name != NULL)
			{
				tr_name = GetChildrenNode(tr_name, "string");
				if((tr_name != NULL)&&(tr_name->xmlChildrenNode != NULL))
				{
					string str = (string)(char*)tr_name->xmlChildrenNode->content;
					tr.setLabel(str);
					rule = GraphGrammar.getRule(str);
				}
			}
	
			xmlNodePtr mu_mapping = GetAttributeNode(node, "mu_mapping");
			if(mu_mapping != NULL)
			{
				xmlNodePtr str_node = GetChildrenNode(mu_mapping, "string");
				if((str_node != NULL)&&(str_node->xmlChildrenNode != NULL)) 
				{
					s = (string)(char*)str_node->xmlChildrenNode->content;
					if (s != "")
    					{
						mapping.clear();
						char * ch1;
						ch1 = strtok((char*)s.c_str(), ",;");
    						char * ch2;
    						ch2 = strtok(NULL, ",;");

						while((ch1!=NULL) && (ch2!=NULL))
						{			
							VERTEX v;
							EDGE e;
							bool check1 = false;
							bool check2 = false;
							//VERTEX v = (GraphGrammar.getStartgraph())->getVertex(ch);
							if(rule->getLeft()->hasVertex(ch1))
							{
								v = rule->getLeft()->getVertex(ch1);
								check1 = true;
							}
										
							if(rule->getRight()->hasVertex(ch1))
							{
								v = rule->getRight()->getVertex(ch1);
								check1 = true;
							}
					
							if(rule->getLeft()->hasEdge(ch1))
							{
								e = rule->getLeft()->getEdge(ch1);
								check2 = true;
							}
					
							if(rule->getRight()->hasEdge(ch1))
							{
								e = rule->getRight()->getEdge(ch1);
								check2 = true;
							}
			
							// if((!check1)&&(!check2))
							//cout<<"connections in mu_mapping are not found!!! ch = "<<ch1<<endl;
										
							if(check1)
							 id1 = v.getID();
							if(check2)
							 id1 = e.getID();
									
							id2 = atoi(ch2);
							mapping.insert(matching_type(id1, id2));
							ch1 = strtok(NULL, ",;");
							ch2 = strtok(NULL, ",;");
						}
					}
							
				}
			}
			mu_map.setMatch(mapping);
			mu.insert(mu_type(id, mu_map));

			// guards
			xmlNodePtr attribut = GetChildrenNode(node, "attr");
			while(attribut != NULL)
			{
				xmlChar* sattr  = xmlGetProp(attribut,(const xmlChar *)"kind");
				if(sattr==NULL)
				{
					attribut = GetNextChild(attribut, "attr");
					continue;
				}

				if(xmlStrcmp(sattr,(const xmlChar *)"attr-bool")) 
					continue;
		
				EXPRESSION e;
				TYPES t;
				t.push_back(BOOL);
				e.addType(t);
		
				xmlNodePtr attribut_attr;
				attribut_attr = GetChildrenNode(attribut, "string");
				if(attribut_attr != NULL) 
					attribut_attr = attribut_attr->xmlChildrenNode;
					if (attribut_attr != NULL)
					{
						e.addExpr((string)(char*)attribut_attr->content);
						tr.addGuard(e);
					}
				
				attribut = GetNextChild(attribut, "attr");	
			}

			PtGraph.addTransition(id, tr);
			if(rule!=NULL)
			 PtGraph.setAssociatedRule(id, rule->getID());
		}
		node = GetNextChild(node, "node");
	}
	
	PtGraph.setMu(&mu);
	
	
	
	
	vector<TOKEN> tokens;
	list<IDTYPE> list_of_vertices;
	MARKING marking;
	xmlNodePtr rel = GetChildrenNode(graph, "rel");
	
	
	while(rel != NULL)
	{
		EDGE edge;
		
		
		s = (string)(char*)xmlGetProp(rel,(const xmlChar *)"id");
		s.erase(0,1);
		id = atoi(s.c_str());
		
		edge.setID(id);
		
		xmlNodePtr rel_label = GetAttributeNode(rel, "label");
		if(rel_label != NULL)
		{
			rel_label = GetChildrenNode(rel_label, "string");
			if((rel_label != NULL)&&(rel_label->xmlChildrenNode != NULL))
				{
					s = (string)(char*)rel_label->xmlChildrenNode->content;
					edge.setLabel(s);
				}
		}
		tokens.clear();
		xmlNodePtr initialNode = GetAttributeNode(rel, "initial_marking");
		if(initialNode != NULL)
		{
			xmlNodePtr tokenNode = GetChildrenNode(initialNode, "token");
			while(tokenNode != NULL)
			{
				TOKEN token;
				xmlNodePtr numberNode = GetChildrenNode(tokenNode, "int");
				if((numberNode != NULL)&&(numberNode->xmlChildrenNode != NULL))
				{
					s = (string)(char*)numberNode->xmlChildrenNode->content;
					int n = atoi(s.c_str());
					token.setCount(n);
				}

				xmlNodePtr attribut = GetChildrenNode(tokenNode, "attr");
				while(attribut != NULL)
				{
					xmlChar* sattr  = xmlGetProp(attribut,(const xmlChar *)"kind");
					if(sattr==NULL)
						continue;
		
					EXPRESSION e;
					TYPES t;
		
					char * ch;
					ch = strtok((char*)sattr, "-");
		
					while(ch!=NULL)
					{
						if(strcmp(ch, "int")==0)
							t.push_back(INTEGER);
						else if(strcmp(ch, "str")==0)
							t.push_back(STRING);
						else if(strcmp(ch, "bool")==0)
							t.push_back(BOOL);
		
						ch = strtok(NULL, "-");
					}
					
					e.addType(t);
					xmlChar* sname = xmlGetProp(attribut,(const xmlChar *)"name");
					e.addName((string)(char*)sname);
					xmlNodePtr attribut_attr;
					attribut_attr = GetChildrenNode(attribut, "string");
					if(attribut_attr != NULL) 
						attribut_attr = attribut_attr->xmlChildrenNode;
						if (attribut_attr != NULL)
							e.addExpr((string)(char*)attribut_attr->content);
				
					edge.setType(t);
					token.addAttributes(e);
					
					attribut = GetNextChild(attribut, "attr");	
				}
				tokens.push_back(token);
				tokenNode = GetNextChild(tokenNode, "token");
			}

			marking.setMarking(id, tokens);
		}
		
		PtGraph.setInitialMarking(marking);
		
		
		xmlNodePtr iota = GetAttributeNode(rel, "iota_mapping");
		if(iota != NULL)
		{
			xmlNodePtr str_node = GetChildrenNode(iota, "string");
				if((str_node != NULL)&&(str_node->xmlChildrenNode != NULL)) 
					{
						s = (string)(char*)str_node->xmlChildrenNode->content;
						if (s != "")
    						{
							ch = strtok((char*)s.c_str(), ";");
							
							while(ch!=NULL)
							{
								EDGE ed = (GraphGrammar.getStartgraph())->getEdge(ch);
								id_match = ed.getID();
								matching.insert(matching_type(id_match, id));
								ch = strtok(NULL, ";");
							}
						}
						
					}	
		}
		
		
		int order = 0;
		xmlNodePtr relend = GetRelend(rel, order);
		list_of_vertices.clear();
		
		while(relend != NULL)
		{
			s = (string)(char*)xmlGetProp(relend,(const xmlChar *)"target");
			s.erase(0,1);
			list_of_vertices.push_back(atoi(s.c_str()));
			order++;
			relend = GetRelend(rel, order);
			PtGraph.addEdgeToCE(atoi(s.c_str()), id);
			
		}
		int n = ((HYPERGRAPH)PtGraph).getVertices().size();
		PtGraph.addEdge(edge, id, list_of_vertices);
		
	/*	xmlNodePtr relend = GetAttributeNode(rel, "relend");
		while(relend != NULL)
		{
			
			
			s = (char*)xmlGetProp(relend, (const xmlChar*)"role");
			if(!strcmp(s.c_str(), "vertex")) 
				{
					s = (char*)xmlGetProp(relend, (const xmlChar*)"target");
					id = atoi(s.c_str());	
				}
			relend = GetNextChild(relend, "relend");
		}*/
		
		// vector<TCONNECTION> preset;
		// vector<TCONNECTION> postset;
		
		relend = GetChildrenNode(rel, "relend");
		while(relend != NULL)
		{
			s = (string)(char*)xmlGetProp(relend, (const xmlChar*)"target");
			s.erase(0,1);
			id1 = atoi(s.c_str());
			
			s = (string)(char*)xmlGetProp(relend, (const xmlChar*)"role");
			
			if(!strcmp(s.c_str(), "preset")) 
				{
					TCONNECTION tc;
					tc.ID = id1;
					xmlNodePtr weight = GetAttributeNode(relend, "weight");
					if(weight != NULL)
					{
						weight = GetChildrenNode(weight, "int");
						if((weight != NULL)&&(weight->xmlChildrenNode != NULL))
						{
							s = (string)(char*)weight->xmlChildrenNode->content;
							tc.weight = atoi(s.c_str());
						}
					}
					// preset.push_back(tc);

					xmlNodePtr attribut = GetChildrenNode(relend, "attr");
					while(attribut != NULL)
					{
						xmlChar* sattr  = xmlGetProp(attribut,(const xmlChar *)"kind");
						if(sattr==NULL)
						{
							attribut = GetNextChild(attribut, "attr");
							continue;
						}
			
						EXPRESSION e;
						TYPES t;
			
						char * ch;
						ch = strtok((char*)sattr, "-");
			
						while(ch!=NULL)
						{
							if(strcmp(ch, "int")==0)
								t.push_back(INTEGER);
							else if(strcmp(ch, "str")==0)
								t.push_back(STRING);
							else if(strcmp(ch, "bool")==0)
								t.push_back(BOOL);
			
							ch = strtok(NULL, "-");
						}

						e.addType(t);
						xmlChar* sname = xmlGetProp(attribut,(const xmlChar *)"name");
						e.addName((string)(char*)sname);
						xmlNodePtr attribut_attr;
						attribut_attr = GetChildrenNode(attribut, "string");
						if(attribut_attr != NULL) 
							attribut_attr = attribut_attr->xmlChildrenNode;
							if (attribut_attr != NULL)
								e.addExpr((string)(char*)attribut_attr->content);
					
						tc.attributes.push_back(e);
						
						attribut = GetNextChild(attribut, "attr");	
					}	
	
					// add preset to the transition
					PtGraph.addTConnection(id, tc);
				}
				
			if(!strcmp(s.c_str(), "postset")) 
				{
					TCONNECTION tc;
					tc.ID = id;
					xmlNodePtr weight = GetAttributeNode(relend, "weight");
					if(weight != NULL)
					{
						weight = GetChildrenNode(weight, "int");
						if((weight != NULL)&&(weight->xmlChildrenNode != NULL))
						{
							s = (string)(char*)weight->xmlChildrenNode->content;
							tc.weight = atoi(s.c_str());
						}
					}
					// postset.push_back(tc);

					xmlNodePtr attribut = GetChildrenNode(relend, "attr");
					while(attribut != NULL)
					{
						xmlChar* sattr  = xmlGetProp(attribut,(const xmlChar *)"kind");
						if(sattr==NULL)
						{
							attribut = GetNextChild(attribut, "attr");
							continue;
						}
			
						EXPRESSION e;
						TYPES t;
			
						char * ch;
						ch = strtok((char*)sattr, "-");
			
						while(ch!=NULL)
						{
							if(strcmp(ch, "int")==0)
								t.push_back(INTEGER);
							else if(strcmp(ch, "str")==0)
								t.push_back(STRING);
							else if(strcmp(ch, "bool")==0)
								t.push_back(BOOL);
			
							ch = strtok(NULL, "-");
						}
						
						e.addType(t);
						xmlChar* sname = xmlGetProp(attribut,(const xmlChar *)"name");
						e.addName((string)(char*)sname);
						xmlNodePtr attribut_attr;
						attribut_attr = GetChildrenNode(attribut, "string");
						if(attribut_attr != NULL) 
							attribut_attr = attribut_attr->xmlChildrenNode;
							if (attribut_attr != NULL)
								e.addExpr((string)(char*)attribut_attr->content);
					
						tc.attributes.push_back(e);
						
						attribut = GetNextChild(attribut, "attr");	
					}	
	
					// add postset to the transition
					PtGraph.addTConnection(id1, tc);
				}
				
			relend = GetNextChild(relend, "relend");
		}
		
		// for nodes and edges preset is postset and vise versa
		// PtGraph.setPostset(id, preset);
		// PtGraph.setPreset(id, postset);
		rel = GetNextChild(rel, "rel");
	}
	
	
	
	iota_match.setMatch(matching);
	PtGraph.setIota(iota_match);
	PtGraph.setInitialMarking(marking);


}
catch(exception e)
{
	cout << "Error in GXLReader.\n";
	exit(1);
}
}



};
};
