/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdexcept>
#include <string.h>

#include "gxlwriter.h"

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/tree.h>


namespace algorithms {



namespace writers {

using namespace data_structures;

GXLWRITER::GXLWRITER()
{
}


GXLWRITER::~GXLWRITER()
{
//    xmlFreeDoc(doc);
//    delete filename;
}
    
void GXLWRITER::initialize(char *pfilename)
{
    //filename = pfilename;
    int n = strlen(pfilename);
    
    filename = new char[n+1];
    
    strcpy(filename,pfilename);
    doc = xmlNewDoc((const xmlChar *)"1.0");
    xmlNodePtr root = xmlNewNode(NULL,(const xmlChar *)"gxl");
    xmlDocSetRootElement(doc,root);
}


    
void GXLWRITER::run(PETRIGRAPH &PtGraph, char const* id_name)
{
	xmlNodePtr root = xmlDocGetRootElement(doc);
	
	xmlNodePtr gxl_graph = xmlNewChild(root, NULL, (const xmlChar*)"graph", NULL);
	//xmlSetProp(gxl_graph,(const xmlChar *)"id",(const xmlChar *)PtGraph.getGrammar()->getStartgraph()->getName());
	xmlSetProp(gxl_graph,(const xmlChar *)"id",(const xmlChar*)id_name);
	xmlSetProp(gxl_graph,(const xmlChar *)"hypergraph",(const xmlChar *)"true");
	xmlSetProp(gxl_graph,(const xmlChar*)"edgemode", (const xmlChar*)"undirected");
	
	
	map<IDTYPE, VERTEX> vertices = ((HYPERGRAPH)PtGraph).getVertices();
    	map<IDTYPE, EDGE> edges = ((HYPERGRAPH)PtGraph).getEdges();
    	map<IDTYPE, TRANSITION> transitions = PtGraph.getTransitions();
	
	MATCH iota_mapping = PtGraph.getIota();
	GTS *GraphGrammar = PtGraph.getGrammar();
	HYPERGRAPH *hg = GraphGrammar->getStartgraph();
	//hg = GraphGrammar.getStartgraph();
		
// store vertices:
    for(map<IDTYPE, VERTEX>::iterator it = vertices.begin(); it != vertices.end(); it++)
    {
      char id_char[50];
      
      IDTYPE id = it->first;
      sprintf(id_char,"%i",id);
      string s = id_char;
      s.insert(0, "_");
     // id_char = s.c_str();
      xmlNodePtr node = xmlNewChild(gxl_graph, NULL, (const xmlChar *)"node", NULL);
      xmlSetProp(node, (const xmlChar *)"id", (const xmlChar *)s.c_str());
      xmlNodePtr attr = xmlNewChild(node,NULL,(const xmlChar *)"attr",NULL);
      xmlSetProp(attr,(const xmlChar *)"name",(const xmlChar *)"vertex");

     
      // iota_mapping
      
      //if(iota_mapping.hasMatch(id))
     
      xmlNodePtr iotaAttr = xmlNewChild(node, NULL, (const xmlChar *)"attr", NULL);
      xmlSetProp(iotaAttr, (const xmlChar *)"name", (const xmlChar *)"iota_mapping");
     
      	//string s;
	s = "";
      	map<IDTYPE, IDTYPE> iota_tmp = iota_mapping.getMatch();
      
      	for(map<IDTYPE, IDTYPE>::iterator it_map = iota_tmp.begin(); it_map != iota_tmp.end(); it_map++)
        {
          if((it_map->second) == id)
	  	{
			//VERTEX v = PtGraph.getGrammar()->getStartgraph()->getVertex(it_map->first);
			string s1 = PtGraph.getGrammar()->getStartgraph()->getVertex(it_map->first).getIdLabel();
			s = s + s1 + ";";
		}
	}
	int l = s.length();
	if(l > 0) s.erase(l-1, 1);
        xmlNewChild(iotaAttr,NULL,(const xmlChar *)"string",(const xmlChar *)s.c_str());
      
      

    }
    
        
// store edges:

    // to do
    assert(!nodes_as_places);

    for(map<IDTYPE, EDGE>::iterator it = edges.begin(); it != edges.end(); it++)
    {
    
      char id_char[50];
      // IDTYPE id = (it->second).getID();
      IDTYPE id = it->first;
      sprintf(id_char,"%i",id);	
      string s = id_char;
      s.insert(0, "_");
      //id_char = s.c_str();
      xmlNodePtr rel = xmlNewChild(gxl_graph, NULL, (const xmlChar *)"rel", NULL);
      xmlSetProp(rel,(const xmlChar *)"id",(const xmlChar *)s.c_str());

      //name
      xmlNodePtr attr = xmlNewChild(rel, NULL, (const xmlChar *)"attr", NULL);
      xmlSetProp(attr, (const xmlChar *)"name", (const xmlChar *)"label");
      xmlNewChild(attr, NULL, (const xmlChar *)"string", (const xmlChar *)(it->second).getLabel().c_str());

      //initial marking
      xmlNodePtr markingAttr = xmlNewChild(rel, NULL, (const xmlChar*) "attr", NULL);
      xmlSetProp(markingAttr, (const xmlChar *)"name", (const xmlChar *)"initial_marking");

      vector<TOKEN> vt = PtGraph.getInitialMarking().getMarking(id);

      for(vector<TOKEN>::iterator itv = vt.begin(); itv != vt.end(); itv++)
      {
	xmlNodePtr tokenAttr = xmlNewChild(markingAttr, NULL, (const xmlChar*) "token", NULL);

	char buf[50];
	sprintf(buf, "%d", itv->getCount());
	xmlNewChild(tokenAttr, NULL, (const xmlChar *)"int", (const xmlChar*) buf);
	list<EXPRESSION> le = itv->getAttributes();
	for(list<EXPRESSION>::iterator ite = le.begin(); ite != le.end(); ite++)
	{
	  xmlNodePtr attrNode = xmlNewChild(tokenAttr, NULL, (const xmlChar *)"attr", NULL);
	  xmlSetProp(attrNode, (const xmlChar *)"kind", (const xmlChar *) ite->getTypesString().c_str());
	  xmlSetProp(attrNode, (const xmlChar *)"name", (const xmlChar *) ite->getName().c_str());
	  xmlNewChild(attrNode, NULL, (const xmlChar *)"string", (const xmlChar*) ite->getExpr().c_str());
	}

      }

      // iota_mapping
      //if(iota_mapping.hasMatch(id))
		xmlNodePtr iotaAttr = xmlNewChild(rel, NULL, (const xmlChar *) "attr", NULL);
      		xmlSetProp(iotaAttr, (const xmlChar *)"name", (const xmlChar *)"iota_mapping");


		//string s;
		s = "";
		map<IDTYPE, IDTYPE> iota_tmp = iota_mapping.getMatch();
      		for(map<IDTYPE, IDTYPE>::iterator it_map = iota_tmp.begin(); it_map != iota_tmp.end(); it_map++)
      		{
          		if((it_map->second) == id)
	  		{
				string s1 = PtGraph.getGrammar()->getStartgraph()->getEdge(it_map->first).getIdLabel();
				s = s + s1 + ";";
			}
		}
		if(s != "")
		{
			int l = s.length();
			s.erase(l-1, 1);
		}
      		xmlNewChild(iotaAttr, NULL, (const xmlChar *)"string", (const xmlChar *)s.c_str());

      // write relends

      // connected vertices
      list<IDTYPE> relends = PtGraph.getConnectedVertices(id);

      int order = 0;

      for(list<IDTYPE>::iterator it_rel = relends.begin(); it_rel != relends.end(); it_rel++)
      {
        char id_char[50];
        char ch_order[50];
        sprintf(id_char,"%d",(*it_rel));
        sprintf(ch_order,"%d",order);
	string s = id_char;
	s.insert(0, "_");
        xmlNodePtr relend = xmlNewChild(rel, NULL, (const xmlChar *)"relend", NULL);
        xmlSetProp(relend, (const xmlChar *)"target", (const xmlChar*)s.c_str());
        xmlSetProp(relend,(const xmlChar *)"role",(const xmlChar*)"vertex");
        xmlSetProp(relend,(const xmlChar *)"startorder",(const xmlChar *)ch_order);
        order++;
      }

      //post- and presets
      vector<TCONNECTION> preset = PtGraph.getPreset(id);
      vector<TCONNECTION> postset = PtGraph.getPostset(id);

      for(vector<TCONNECTION>::iterator it_pre = preset.begin(); it_pre != preset.end(); it_pre++)
      {
		xmlNodePtr relend = xmlNewChild(rel, NULL, (const xmlChar *)"relend", NULL);
		char ch[50];
		sprintf(ch, "%d", (*it_pre).ID);
		string s = ch;
		s.insert(0, "_");
		xmlSetProp(relend, (const xmlChar *)"target", (const xmlChar*)s.c_str());
		// xmlSetProp(relend,(const xmlChar *)"role",(const xmlChar*)"preset");	
		xmlSetProp(relend,(const xmlChar *)"role",(const xmlChar*)"postset");	
		
		xmlNodePtr weightAttr = xmlNewChild(relend, NULL, (const xmlChar*)"attr", NULL);
		xmlSetProp(weightAttr, (const xmlChar *)"name", (const xmlChar *)"weight");
		sprintf(ch, "%d", (*it_pre).weight);
		xmlNewChild(weightAttr, NULL, (const xmlChar*)"int", (const xmlChar*)ch);

		list<EXPRESSION> le = it_pre->attributes;
		for(list<EXPRESSION>::iterator ite = le.begin(); ite != le.end(); ite++)
		{
		xmlNodePtr attrNode = xmlNewChild(relend, NULL, (const xmlChar *)"attr", NULL);
		xmlSetProp(attrNode, (const xmlChar *)"kind", (const xmlChar *) ite->getTypesString().c_str());
		xmlSetProp(attrNode, (const xmlChar *)"name", (const xmlChar *) ite->getName().c_str());
		xmlNewChild(attrNode, NULL, (const xmlChar *)"string", (const xmlChar*) ite->getExpr().c_str());
		}
      }


      for(vector<TCONNECTION>::iterator it_post = postset.begin(); it_post != postset.end(); it_post++)
      {
      		xmlNodePtr relend = xmlNewChild(rel, NULL, (const xmlChar *)"relend", NULL);
		char ch[50];
		sprintf(ch, "%d", (*it_post).ID);
		string s = ch;
		s.insert(0, "_");
		xmlSetProp(relend, (const xmlChar *)"target", (const xmlChar*)s.c_str());
		// xmlSetProp(relend,(const xmlChar *)"role",(const xmlChar*)"postset");
		xmlSetProp(relend,(const xmlChar *)"role",(const xmlChar*)"preset");
		
		xmlNodePtr weightAttr = xmlNewChild(relend, NULL, (const xmlChar*)"attr", NULL);
		xmlSetProp(weightAttr, (const xmlChar *)"name", (const xmlChar *)"weight");
		sprintf(ch, "%d", (*it_post).weight);
		xmlNewChild(weightAttr, NULL, (const xmlChar*)"int", (const xmlChar*)ch);

		list<EXPRESSION> le = it_post->attributes;
		for(list<EXPRESSION>::iterator ite = le.begin(); ite != le.end(); ite++)
		{
		xmlNodePtr attrNode = xmlNewChild(relend, NULL, (const xmlChar *)"attr", NULL);
		xmlSetProp(attrNode, (const xmlChar *)"kind", (const xmlChar *) ite->getTypesString().c_str());
		xmlSetProp(attrNode, (const xmlChar *)"name", (const xmlChar *) ite->getName().c_str());
		xmlNewChild(attrNode, NULL, (const xmlChar *)"string", (const xmlChar*) ite->getExpr().c_str());
		}
      }
    }

    //store transitions

    for(map<IDTYPE, TRANSITION>::iterator it = transitions.begin(); it != transitions.end(); it++)
    {
	char id_char[50];
	IDTYPE id = it->first;
	//IDTYPE id = (it->second).getID();
	sprintf(id_char,"%d",id);
	string s = id_char;
      	s.insert(0, "_");
      	//id_char = s.c_str();
	xmlNodePtr node = xmlNewChild(gxl_graph, NULL, (const xmlChar *)"node", NULL);
	xmlSetProp(node, (const xmlChar *)"id", (const xmlChar *)s.c_str());
	xmlNodePtr attr = xmlNewChild(node, NULL,(const xmlChar *)"attr", NULL);
	xmlSetProp(attr, (const xmlChar *)"name", (const xmlChar *)"transition");
	
	//write name of transition
	attr = xmlNewChild(node, NULL,(const xmlChar *)"attr", NULL);
	xmlSetProp(attr, (const xmlChar *)"name", (const xmlChar *)"rule");
	// string rule_name = (it->second).getLabel();
	string rule_name = PtGraph.getAssociatedRule(it->first)->getName();
	xmlNewChild(attr, NULL, (const xmlChar*)"string", (const xmlChar*)rule_name.c_str());

	// guards
	if(it->second.getGuard().size() != 0)
	{
		list<EXPRESSION> guards = it->second.getGuard();
		for(list<EXPRESSION>::iterator itg = guards.begin(); itg != guards.end(); itg++)
		{
			attr = xmlNewChild(node, NULL,(const xmlChar *)"attr", NULL);
			xmlSetProp(attr, (const xmlChar *)"kind", (const xmlChar *)"attr-bool");
			xmlNewChild(attr, NULL, (const xmlChar*)"string", (const xmlChar*)itg->getExpr().c_str());
		}
	}

	//write mu_mapping
	attr = xmlNewChild(node, NULL,(const xmlChar *)"attr", NULL);
	xmlSetProp(attr, (const xmlChar *)"name", (const xmlChar *)"mu_mapping");
		
	RULE* rule = PtGraph.getGrammar()->getRule(rule_name);
	map<IDTYPE, IDTYPE> mu_mapping = PtGraph.getMu(id).getMatch();
	s = "";
	
	for(map<IDTYPE, IDTYPE>::iterator it_mu = mu_mapping.begin(); it_mu != mu_mapping.end(); it_mu++)
	{
		char ch[50];
		string s1 = "";
		if(rule->getLeft()->hasVertex(it_mu->first))
			s1 = rule->getLeft()->getVertex(it_mu->first).getIdLabel();
		if(rule->getLeft()->hasEdge(it_mu->first))
			s1 = rule->getLeft()->getEdge(it_mu->first).getIdLabel();
		if(rule->getRight()->hasVertex(it_mu->first))
			s1 = rule->getRight()->getVertex(it_mu->first).getIdLabel();
		if(rule->getRight()->hasEdge(it_mu->first))
			s1 = rule->getRight()->getEdge(it_mu->first).getIdLabel();
		s = s + s1 + ",";
		sprintf(ch, "%d", (it_mu->second));
		s = s + ch + ";";
	}
	
	xmlNewChild(attr, NULL, (const xmlChar*)"string", (const xmlChar*)s.c_str());
	
	
    }

	
	
    if(-1 == xmlSaveFormatFile(filename,doc,true))
    {
      cout << "GXLWriter: Error while writing XML file\n";
      exit(1);
    }
}


};
};


