/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "hypergraph.h"

namespace data_structures
{
   HYPERGRAPH::HYPERGRAPH()
   {
   }

   HYPERGRAPH::~HYPERGRAPH()
   {
   }
   
   HYPERGRAPH HYPERGRAPH::clone()
   {
      HYPERGRAPH h;
      h.vertices = this->vertices;
      h.edges = this->edges;
      h.connectedVertices = this->connectedVertices;
      h.connectedEdges = this->connectedEdges;
      return h;
   }
   
   void HYPERGRAPH::addVertex(VERTEX v, IDTYPE id)
   {
      if(vertices.count(id) != 0)
	 return;

      this->vertices[id] = v;
      this->vertices[id].setID(id);
      this->connectedEdges[id] = multiset<IDTYPE>();
      if(id > lastUsedID) lastUsedID = id;
   }
   
   void HYPERGRAPH::deleteVertex(IDTYPE id)
   {
      if(this->vertices.count(id) > 0)
      {
         /*
          * delete connected edges
          */
         multiset<IDTYPE> cE = this->connectedEdges[id];
         multiset<IDTYPE>::iterator connectit;
         for(connectit = cE.begin(); connectit != cE.end(); connectit++)
         {
            this->deleteEdge(*connectit);
         }
     
         /*
          * delete the vertex
          */
         this->vertices.erase(id);
         this->connectedEdges.erase(id);
      }
   }

   void HYPERGRAPH::addEdge(EDGE e, IDTYPE id, list<IDTYPE> cV)
   {
      if(this->edges.count(id) == 0)
      {
         /*
          * add edge
          */  
         this->edges[id] = e;
 	 this->edges[id].setID(id);
         this->connectedVertices[id] = cV;

         /*
          * add this edge to connection lists of its connected vertices
          */
         list<IDTYPE>::iterator connectit;
	 
         for(connectit = cV.begin(); connectit != cV.end(); connectit++)
         {
            assert(this->vertices.count(*connectit) > 0);
            this->connectedEdges[*connectit].insert(id);
         }

	 if(id > lastUsedID) lastUsedID = id;
      }
   }

   IDTYPE HYPERGRAPH::addEdge2(EDGE e, IDTYPE id, list<IDTYPE> cV)
   {
      IDTYPE edge_id;

      if(this->edges.count(id) == 0)
      {
        edge_id=id;
      }
      else
      {
        lastUsedID++;
	edge_id = lastUsedID;
      }

      /*
       * add edge
       */  
      this->edges[edge_id] = e;
      this->edges[edge_id].setID(edge_id);
      this->connectedVertices[edge_id] = cV;

      /*
       * add this edge to connection lists of its connected vertices
       */
      list<IDTYPE>::iterator connectit;

      for(connectit = cV.begin(); connectit != cV.end(); connectit++)
      {
         assert(this->vertices.count(*connectit) > 0);
         this->connectedEdges[*connectit].insert(edge_id);
      }

      if(edge_id > lastUsedID) lastUsedID = edge_id;

      return edge_id;
   }

   void HYPERGRAPH::addEdgeToCE(IDTYPE ver_id, IDTYPE edge_id)
   {
   	multiset<IDTYPE> edges = this->getConnectedEdges(ver_id);
	edges.insert(edge_id);
   	this->connectedEdges[ver_id] = edges;
	this->connectedVertices[edge_id].push_back(ver_id);
   }

   void HYPERGRAPH::deleteEdge(IDTYPE id)
   {
      if(this->edges.count(id) > 0)
      {
         /*
          * remove this edge from all vertex connection lists where it occurs
          */
         list<IDTYPE>::iterator connectit;
         for( connectit = this->connectedVertices[id].begin();
              connectit != this->connectedVertices[id].end();
              connectit++ )
         {
            this->connectedEdges[*connectit].erase(id); 
         }

         /*
          * delete edge
          */  
         this->edges.erase(id); 
         this->connectedVertices.erase(id);
      }
   }   

   list<IDTYPE> HYPERGRAPH::getConnectedVertices(IDTYPE eid)
   {
	if(connectedVertices.count(eid)==0)
	 return list<IDTYPE>();

	return this->connectedVertices[eid];
   }
   
   multiset<IDTYPE> HYPERGRAPH::getConnectedEdges(IDTYPE vid)
   {
	if(connectedEdges.count(vid)==0)
	 return multiset<IDTYPE>();

	return this->connectedEdges[vid];
   }

   map< IDTYPE, list<IDTYPE> > & HYPERGRAPH::getConnectedVertices()
   {
	return this->connectedVertices;
   }

   map< IDTYPE, multiset<IDTYPE> > & HYPERGRAPH::getConnectedEdges()
   {
	return this->connectedEdges;
   }

   void HYPERGRAPH::addConnectedVertex(IDTYPE eid, VERTEX v)
   {
	addVertex(v, v.getID());

	list<IDTYPE> lv = getConnectedVertices(eid);
	lv.push_back(v.getID());

	connectedVertices[eid] = lv;
   }

   void HYPERGRAPH::initDepth()
   {
   for(std::map<IDTYPE, EDGE>::iterator it = edges.begin(); it != edges.end(); it++)
			depthmapEdges[(it->first)] = 0;
			
		for(std::map<IDTYPE, VERTEX>::iterator it = vertices.begin(); it != vertices.end(); it++)
			depthmapVertices[(it->first)] = 0;
    }


   ostream& operator<<(ostream& ost, HYPERGRAPH& th)
   {
      ost << "hypergraph \"" << th.name << "\" (ID " << th.id << ") with\nvertices";
      map<IDTYPE, VERTEX>::iterator vit;
      for(vit = th.vertices.begin(); vit != th.vertices.end(); vit++)
      {
         ost << " " << vit->second << "/" << vit->first;
      }
      ost << "\nedges";
      map<IDTYPE, EDGE>::iterator eit;
      for(eit = th.edges.begin(); eit != th.edges.end(); eit++)
      {
         ost << " " << eit->second << "/" << eit->first << ": ";
	 list<IDTYPE> cV = th.connectedVertices[eit->first];
	 list<IDTYPE>::iterator cVit;
	 if(cV.size() == 2)
	 {
	    cVit = cV.begin();
	    ost << *cVit << "->";
	    cVit++;
	    ost << *cVit << " ";
	 }
	 else
	 {
            ost << "|-";
	    for(cVit = cV.begin(); cVit != cV.end(); cVit++)
	    {
	       ost << '-' << *cVit;
	    }
	    ost << "--> ";
	 }
      }
      return ost;
   }
};
