/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef DATA_STRUCTURESHYPERGRAPH_H
#define DATA_STRUCTURESHYPERGRAPH_H

#include "graph.h"

namespace data_structures
{
   class HYPERGRAPH: public virtual GRAPH
   {
      public:
         HYPERGRAPH();
         virtual ~HYPERGRAPH();

         HYPERGRAPH clone();
  
         virtual void addVertex(VERTEX v, IDTYPE id);
         // delete a vertex and all edges connected to it
         virtual void deleteVertex(IDTYPE id);
         
         // add an edge with given connected vertices cV; id must be unique
         virtual void addEdge(EDGE e, IDTYPE id, list<IDTYPE> cV);
	 virtual IDTYPE addEdge2(EDGE e, IDTYPE id, list<IDTYPE> cV);
	 void addEdgeToCE(IDTYPE ver_id, IDTYPE edge_id);
         virtual void deleteEdge(IDTYPE id);         

         list<IDTYPE> getConnectedVertices(IDTYPE eid);
         multiset<IDTYPE> getConnectedEdges(IDTYPE vid);
	 map< IDTYPE, list<IDTYPE> > & getConnectedVertices();
         map< IDTYPE, multiset<IDTYPE> > & getConnectedEdges();
	 void addConnectedVertex(IDTYPE eid, VERTEX v);

         friend ostream& operator<<(ostream& ost, HYPERGRAPH& th);
	void initDepth();
	void setConnectedVertices(map<IDTYPE,list<IDTYPE> > cVerts){connectedVertices=cVerts;}
	void setConnectedEdges(map<IDTYPE, multiset<IDTYPE> > cEdges){connectedEdges=cEdges;}
	void setDepthmapEdges(map<IDTYPE, long> depthEdges){depthmapEdges=depthEdges;}
	void setDepthmapVertices(map<IDTYPE, long> depthVerts){depthmapVertices=depthVerts;}
	map<IDTYPE, long> getDepthmapEdges(){return depthmapEdges;}
	map<IDTYPE, long> getDepthmapVertices(){return depthmapVertices;}
	 
      protected:
         map< IDTYPE, list<IDTYPE> > connectedVertices;
         map< IDTYPE, multiset<IDTYPE> > connectedEdges;
	 std::map<IDTYPE, long> depthmapEdges;
         std::map<IDTYPE, long> depthmapVertices; 
   };
};
using namespace data_structures;

#endif
