/***************************************************************************
 *   Copyright (C) 2006 by root   *
 *   root@Vitali   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "id_checker.h"

namespace special
{

ID_CHECKER::ID_CHECKER()
{
	gts=NULL;
}


ID_CHECKER::~ID_CHECKER()
{
}

void ID_CHECKER::initialize(GTS & gts)
{
	this->gts = &gts;
}

void ID_CHECKER::run(RULE & r)
{
	HYPERGRAPH right_changed;
	
	// changed nodes
	map<IDTYPE, IDTYPE> changed_nodes;

	IDTYPE id = 0;

	// vertices
	map<IDTYPE, VERTEX> vl = r.getLeft()->getVertices();
	map<IDTYPE, VERTEX> vr = r.getRight()->getVertices();

	// edges
	map<IDTYPE, EDGE> el = r.getLeft()->getEdges();
	map<IDTYPE, EDGE> er = r.getRight()->getEdges();	

	// check hypergraphs
	checkHypergraph(vl,el);
	checkHypergraph(vr,er);

	// id
	id = findMaximum(id, vl, el);
	id = findMaximum(id, vr, er);
	id++;

	// nodes
	for(map<IDTYPE, VERTEX>::iterator itv = vr.begin(); itv != vr.end(); itv++)
	{
		if(findVID(itv->first, vl) || findEID(itv->first, el))
		{
			VERTEX v = itv->second;
			changed_nodes[itv->first] = id;
			v.setID(id);
			right_changed.addVertex(v, id);
			id++;
		}
		else
		{
			right_changed.addVertex(itv->second, itv->first);
		}
	}

	// edges
	for(map<IDTYPE, EDGE>::iterator ite = er.begin(); ite != er.end(); ite++)
	{
		// reconnect nodes
		list<IDTYPE> connv_changed;
		list<IDTYPE> connv = r.getRight()->getConnectedVertices(ite->first);

		for(list<IDTYPE>::iterator it = connv.begin(); it != connv.end(); it++)
		{
			if(changed_nodes.find(*it) != changed_nodes.end())
			 connv_changed.push_back(changed_nodes[*it]);
			else
			 connv_changed.push_back(*it);
		}

		if(findVID(ite->first, vl) || findEID(ite->first, el))
		{
			EDGE e = ite->second;
			e.setID(id);
			right_changed.addEdge(e, id, connv_changed);
			id++;
		}
		else
		{
			right_changed.addEdge(ite->second, ite->first, connv_changed);
		}
	}

	// match
	map<IDTYPE, IDTYPE> m = r.getInterface().getMatch();

	MATCH match;

	for(map<IDTYPE, IDTYPE>::iterator itm = m.begin(); itm != m.end(); itm++)
	{
		if(changed_nodes.find(itm->second) != changed_nodes.end())
			match.setMatch(itm->first, changed_nodes[itm->second]);
		else
			match.setMatch(itm->first, itm->second);
	}
	
	r.setRight(&right_changed);
	r.setInterface(match);
}

void ID_CHECKER::run()
{
	assert(gts != NULL);

	map<IDTYPE, RULE>* prules = gts->getRules();

	map<IDTYPE, RULE> * prules_changed = new map<IDTYPE, RULE>();

	for(map<IDTYPE, RULE>::iterator it = (*prules).begin(); it != (*prules).end(); it++)
	{
		RULE r = it->second;

		run(r);

		(*prules_changed)[it->first] = r;
	}

	gts->setRules(prules_changed);
}

IDTYPE ID_CHECKER::findMaximum(IDTYPE id, map<IDTYPE, VERTEX> v, map<IDTYPE, EDGE> e)
{
	IDTYPE res = id;

	for(map<IDTYPE, VERTEX>::iterator itv = v.begin(); itv != v.end(); itv++)
	{
		if(itv->first > res)
		 res = itv->first;
	}

	for(map<IDTYPE, EDGE>::iterator ite = e.begin(); ite != e.end(); ite++)
	{
		if(ite->first > res)
		 res = ite->first;
	}

	return res;
}

bool ID_CHECKER::findVID(IDTYPE id, map<IDTYPE, VERTEX> v)
{
	for(map<IDTYPE, VERTEX>::iterator itv = v.begin(); itv != v.end(); itv++)
	{
		if(itv->first == id)
		 return true;
	}

	return false;
}

bool ID_CHECKER::findEID(IDTYPE id, map<IDTYPE, EDGE> e)
{
	for(map<IDTYPE, EDGE>::iterator ite = e.begin(); ite != e.end(); ite++)
	{
		if(ite->first == id)
		 return true;
	}

	return false;
}

void ID_CHECKER::checkHypergraph(map<IDTYPE, VERTEX> v, map<IDTYPE, EDGE> e)
{
	map<IDTYPE, int> visited;

	for(map<IDTYPE, VERTEX>::iterator itv = v.begin(); itv != v.end(); itv++)
	{
		if(visited.find(itv->first) != visited.end())
		{
			cout << endl << "Error: non-unique IDs in the hypergraph, node with ID=" << itv->first << endl;
			exit(1);
		}
		else
		{
			visited[itv->first] = 1;
		}
	}

	for(map<IDTYPE, EDGE>::iterator ite = e.begin(); ite != e.end(); ite++)
	{
		if(visited.find(ite->first) != visited.end())
		{
			cout << endl << "Error: non-unique IDs in the hypergraph, edge with ID=" << ite->first << endl;
			exit(1);
		}
		else
		{
			visited[ite->first] = 1;
		}
	}
}

}
