/***************************************************************************
 *   Copyright (C) 2007 by Salil Joshi   *
 *   hellfeuer@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <iostream>
#include <set>
using std::pair;
using std::cout;
using std::set;

namespace data_structures {

template <typename T>LINEARSET<T>::LINEARSET(T& bp)
{
	base_point = T(bp);
	zero = base_point.isZero();
}

template <typename T>LINEARSET<T>::LINEARSET(T& bp, set<T*, Compare>& v)
{
	base_point = T(bp);
	vectors = set<T*,Compare>(v);
	zero = base_point.isZero();
	typename set<T*,Compare>::iterator i;
	for (i = vectors.begin(); i != vectors.end(); i++)
	{
		if ((*i)->isZero() == false) {
			zero = false;
			break;
		}
	}
}

template <typename T>LINEARSET<T>::LINEARSET()
{
}

/*Set contains only the zero element*/
template <typename T> bool LINEARSET<T>::isZero()
{
	return zero;
}

/* For taking the star: the base point is inserted into the set of vectors*/
template <typename T> void LINEARSET<T>::closure()
{
	vectors.insert(new T(base_point));
}


template <typename T> LINEARSET<T>* LINEARSET<T>::operator+(LINEARSET<T> &b)
{
	LINEARSET<T>* l = new LINEARSET<T>(base_point, vectors);
	*l += b;
	return l;
}

/*
Addition
1) Adds the base point
2) Inserts the vectors
*/
template <typename T> void LINEARSET<T>::operator+=(LINEARSET<T> &b)
{
	base_point += b.base_point;
	
	typename set<T*,Compare>::iterator i;
	for (i = b.vectors.begin(); i != b.vectors.end(); i++)
	{
		vectors.insert(*i);
	}
}

template <typename T> void LINEARSET<T>::print()
{
	base_point.print();
	typename set<T *,Compare>::iterator i;
	for (i = vectors.begin(); i != vectors.end(); i++)
	{
		cout<<", ";
		(*i)->print();
		cout<<"N";
	}
}


/* This is not a true less than.
This is a dirty hack and exists only to provide an ordering, which will enable LINEARSET to be used correctly in the set STL data structure (i.e., to prevent duplicates in a set). It provides a strict weak ordering, but no extra meaning should be attached to it.
*/
template <typename T> bool LINEARSET<T>::operator<(LINEARSET<T>& b)
{
		if (vectors.size() >b.vectors.size()) {
			return false;
		} else if (vectors.size()  < b.vectors.size()) {
			return true;
		} else {
			typename set<T *,Compare>::iterator i;
			typename set<T *,Compare>::iterator j = b.vectors.begin();
			for (i = vectors.begin(); i != vectors.end(); i++) {
				if (!(*(*i) < *(*j))) {
					return false;
				}
				j++;
			}
			return (base_point<b.base_point);
		}
}

/*
Function not defined, because it would require proving the existence of a solution in the naturals, of a certain set of linear equations
This is not necessary for Parikh image generation
*/
template <typename T> bool LINEARSET<T>::hasElement(T* element)
{
}

template <typename T> LINEARSET<T>::~LINEARSET()
{
}
}
