/***************************************************************************
 *   Copyright (C) 2007 by Salil Joshi   *
 *   hellfeuer@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef DATA_STRUCTURESLS_MARKINGS_H
#define DATA_STRUCTURESLS_MARKINGS_H
#include <iostream>
#include<set>
#include <utility>
using std::set;
using std::cout;


namespace data_structures {

/**
	@author hellfeuer <hellfeuer@gmail.com>
	Represents a linear set. The template parameter class T must support the following operations:
	1) +=
	2) <
	3) isZero() method
	4) copy constructor
	5) constructor that takes one unsigned int argument representing size of alphabet, and creates a "zero element"
	6) print() method

*/

template <typename T>
class LINEARSET
{
private:
	int n;
	bool zero;
	class Compare
	{
		public:
			bool operator()(T* a, T* b)
			{
				return (*a < *b);
			}
	};
public:
	T base_point;
	set<T*, Compare> vectors;

	LINEARSET();
	LINEARSET(T &bp);
	LINEARSET(T& bp, set<T*, Compare>& v);
    void setBasePoint(T* element);
    void addVector(T* element);
    bool hasElement(T* element);
    bool isZero();
    void closure();
    void print();
    bool operator<(LINEARSET<T>& b);	//This is not "true" less than. See comment in method definition
    LINEARSET<T>* operator+(LINEARSET<T> &b);
    void operator+=(LINEARSET<T> & b);
    ~LINEARSET();

};
}
#include "linearset.cpp"
#endif
