/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "logic_converter.h"
#include "gxlreader.h"
#include "graph_element.h"
#include "logicreader.h"

#include <fstream>
#include <map>
#include <string.h>
#include <typeinfo>

#include <strstream>

using namespace std;

namespace converters {

namespace logic_converter {

LOGIC_CONVERTER::LOGIC_CONVERTER()
 : SCENARIO()
{
  this->name = "logic_converter";

  this->inspec.push_back(pair<string, DATA_FORMAT>("petrigraph", GXL_PETRIGRAPH));
  this->inspec.push_back(pair<string, DATA_FORMAT>("logic", PLAIN_TEXT));
  this->outspec.push_back(pair<string, DATA_FORMAT>("formula", PLAIN_TEXT));

  this->call_descriptor = "Usage: augur -sc=logic_converter [options] logic petrigraph formula\n\n\
petrigraph\tinput petrigraph\n\
logic\t\tformula\n\
formula\t\toutput file\n";
}


LOGIC_CONVERTER::~LOGIC_CONVERTER()
{
}

void LOGIC_CONVERTER::run()
{ 
 algorithms::readers::LOGICREADER lr;
 lr.initialize(this->filemap["logic"].c_str());
 Node *formula = NULL;
 lr.run (&formula);
 
 cout << formula << endl;
 
 algorithms::readers::GXLREADER reader;
 reader.initialize((char*)(this->filemap["petrigraph"]).c_str());
 
 cout << "running reader" << endl;
 PETRIGRAPH pg;
 GTS gts;
 reader.run(pg, gts);
 // cout is gone???
 // FIXME: outputting causes a segfault
 //cerr << pg << endl;
 
 map<string, IDTYPE> mapping;
 PARTITION part;
 Node *result;
 
 /*
 mapping["edge1"] = 9336;
 mapping["edge2"] = 9338;
 mapping["edge3"] = 9341;
 mapping["edge4"] = 9349;
 */
 cout << "starting encoding..." << endl;
 result = encode (*formula, mapping, part, pg);

 cout << "done encoding:" << endl << endl;

 strstream oss;
 oss << result;

 if(oss.pcount()<2)
	cout << "FORMULA FALSE" << endl;
 else
 	cout << result << endl;

 ofstream output;
 output.open(this->filemap["formula"].c_str());
 output << result << endl;
 output.close();
 delete result;
}

Node* LOGIC_CONVERTER::encode_equals (const Node &F, map<string, IDTYPE> mapping, PARTITION P, const PETRIGRAPH &pg)
{
  const Operator *oper = dynamic_cast<const Operator *>(&F);
  if ((typeid (*oper->operands[0]) == typeid (Identifier)) &&
      (typeid (*oper->operands[1]) == typeid (Identifier))) {
    Identifier *id0 = dynamic_cast<Identifier *>(oper->operands[0]);
    Identifier *id1 = dynamic_cast<Identifier *>(oper->operands[1]);
    return new Boolean (P.isSameClass (*id0, *id1));
  } else if ((typeid (*oper->operands[0]) == typeid (Label)) &&
             (typeid (*oper->operands[1]) == typeid (Identifier))) {
    // Function = Identifier only allowed for function = label
    Label *func = dynamic_cast<Label *>(oper->operands[0]);
    Identifier *id = dynamic_cast<Identifier *>(oper->operands[1]);
    Identifier *edge = func->variable;
    // get label of edge, compare with id
    if (mapping.count (edge->name) == 0) {
      // TODO: error, maybe already disallow this in the parser ...
      cerr << "error, unmapped edge" << endl;
      exit(1);	
    } else {
      //cerr << "got labelfunc for edge with label: " << ((GRAPH)pg).getEdges()[mapping[edge->name]].getLabel() << endl;
      string label = ((GRAPH)pg).getEdges()[mapping[edge->name]].getLabel();
      if (label == id->name) {
        //cerr << "label(" << mapping[edge->name] << ") = " << id->name << endl;
        return new Boolean (true);
      } else {
        //cerr << "label(" << mapping[edge->name] << ") != " << id->name << endl;
        return new Boolean (false);
      }
    }
  } else if ((typeid (*oper->operands[0]) == typeid (Connection)) &&
             (typeid (*oper->operands[1]) == typeid (Connection))) {
    Connection *func0 = dynamic_cast<Connection *>(oper->operands[0]);
    Connection *func1 = dynamic_cast<Connection *>(oper->operands[1]);
    // compare IDTYPEs of the results
    EDGE edge0 = ((GRAPH)pg).getEdges()[mapping[func0->variable->name]];
    EDGE edge1 = ((GRAPH)pg).getEdges()[mapping[func1->variable->name]];
    //cerr << "edge0: " << edge0.getID() << " edge1: " << edge1.getID() << endl;
    list<IDTYPE> nodes0 = ((HYPERGRAPH)pg).getConnectedVertices (
        edge0.getID());
    IDTYPE connectedNode0 = 0;
    IDTYPE connectedNode1 = 0;
    list<IDTYPE>::iterator it = nodes0.begin();
    for (unsigned int i = 0; i <= func0->number; i++)
    {
      if (it == nodes0.end()) {
        //cerr << "no such connection: " << func0->number << " on edge: " <<
          //mapping[func0->variable->name] << endl;
        return new Boolean (false);
      }
      connectedNode0 = *it;
      ++it;
    }
    list<IDTYPE> nodes1 = ((HYPERGRAPH)pg).getConnectedVertices (
        edge1.getID());
    it = nodes1.begin();
    for (unsigned int i = 0; i <= func1->number; i++)
    {
      if (it == nodes1.end()) {
        //cerr << "no such connection: " << func1->number << " on edge: " <<
          //mapping[func1->variable->name] << endl;
        return new Boolean (false);
      }
      connectedNode1 = *it;
      ++it;
    }
    //cerr << "first node: " << connectedNode0 << " second node: " << connectedNode1 << endl;
    return new Boolean (connectedNode0 == connectedNode1);
  }
  cerr << "should not be reached" << endl;
  return NULL;
}

list<Node *> LOGIC_CONVERTER::mapToExistingClass (string variable, const Node &F, map<string, IDTYPE> mapping, PARTITION P, const PETRIGRAPH &pg)
{
  list<Node *> operands;
  for (std::set<IdSet, ltIdSet>::iterator it=P.subsets.begin ();
       it != P.subsets.end (); ++it)
  {
    IdSet set (*it);
    string rep = *(set.begin ());
    map<string, IDTYPE> newMapping (mapping);
    newMapping[variable] = newMapping[rep];
    PARTITION newPart (P);
    newPart.removeSet (set);
    set.insert (variable);
    newPart.addSet (set);
    Node *resultFragment = encode (F, newMapping, newPart, pg);
    operands.push_back (resultFragment);
  }
  return operands;
}

// removes all Boolean Nodes with value 'filter' from the list
void LOGIC_CONVERTER::filterBooleans (list<Node *> &nodes, bool filter)
{
  list<Node *>::iterator it = nodes.begin ();
  while (it != nodes.end ())
  {
    Node *current_node = *it;
    if (typeid (*current_node) == typeid (Boolean)) {
      Boolean *res = dynamic_cast<Boolean *>(current_node);
      if (res->value == filter) {
        delete res;
        it = nodes.erase (it);
        continue;
      }
    }
    ++it;
  }
}


// check if nodes contains a Boolean with value 'value'
bool LOGIC_CONVERTER::hasBoolean (const list<Node *> &nodes, bool value)
{
  for (list<Node *>::const_iterator it = nodes.begin();
       it != nodes.end();
       ++it)
  {
    if (typeid (**it) == typeid (Boolean)) {
      const Boolean *current = dynamic_cast<const Boolean *>(*it);
      if (current->value == value) {
        return true;
      }
    }
  }
  return false;
}

Node* LOGIC_CONVERTER::encode_sub_formula (const string &variable, const IDTYPE edge, const Node &F, const map<string, IDTYPE> &mapping, const PARTITION &P, const PETRIGRAPH &pg)
{
  map<string, IDTYPE> newMapping (mapping);
  newMapping[variable] = edge;
  PARTITION newPart (P);
  IdSet set;
  set.insert (variable);
  newPart.addSet (set);
  return encode (F, newMapping, newPart, pg);
}

Node* LOGIC_CONVERTER::encode_quantor (const Node &F, map<string, IDTYPE> mapping, PARTITION P, const PETRIGRAPH &pg)
{
  const Quantor *quantor = dynamic_cast<const Quantor *>(&F);
  if (quantor->name == "all1") {
    // convert ALL1
    // (1)
    list<Node *> operands = mapToExistingClass (quantor->variable->name, *quantor->formula, mapping, P, pg);
    if (hasBoolean (operands, false))
        return new Boolean (false);
    // (2)
    map<IDTYPE, EDGE> edges = ((GRAPH)pg).getEdges ();
    for (map<IDTYPE, EDGE>::iterator it = edges.begin();
         it != edges.end();
         ++it)
    {
      IDTYPE eid = (*it).first;
      Node *variable = new Variable (eid);
      Node *value = new Constant (P.nPp (eid, mapping) + 1);
      Node *part1 = new Operator (">=", variable, value);
      Node *part2 = encode_sub_formula (
          quantor->variable->name, eid, *quantor->formula,
          mapping, P, pg);
      if (typeid (*part2) == typeid (Boolean)) {
        Boolean *res = dynamic_cast<Boolean *>(part2);
        if (res->value) {
          operands.push_back (res);
          delete part1;
        } else {
          operands.push_back (new Operator ("~", part1, NULL));
          delete part2;
        }
      } else {
        // TODO:
        // what happens if part2 isn't a boolean?
        // shouldn't there be a implies operator then?
        //
        // for now just delete whatever we got
        delete part1;
        delete part2;
      }
    }
    filterBooleans (operands, true);
    if (operands.empty ())
      return new Boolean (false);
    if (operands.size () == 1)
      return operands.front ();
    Operator *retval = new Operator ("&", operands);
    //cerr << "returning: " << retval << endl;
    return retval;
  } else if (quantor->name == "ex1") {
    // convert EX1
    // (1)
    list<Node *> operands = mapToExistingClass (quantor->variable->name, *quantor->formula, mapping, P, pg);
    if (hasBoolean (operands, true))
      return new Boolean (true);

    // (2)
    map<IDTYPE, EDGE> edges = ((GRAPH)pg).getEdges ();
    for (map<IDTYPE, EDGE>::iterator it = edges.begin();
         it != edges.end();
         ++it)
    {
      IDTYPE eid = (*it).first;
      Node *part1 = encode_sub_formula (
          quantor->variable->name, eid, *quantor->formula,
          mapping, P, pg);
      Node *variable = new Variable (eid);
      Node *value = new Constant (P.nPp (eid, mapping) + 1);
      Node *part2 = new Operator (">=", variable, value);
      if (typeid (*part1) != typeid (Boolean)) {
        //cerr << "creating new &: " << part1 << " && " << part2 << endl;
        Node *oper = new Operator ("&", part1, part2);
        operands.push_back (oper);
      } else {
        const Boolean *res = dynamic_cast<const Boolean *>(part1);
        if (res->value) {
          //cerr << "[ex1] adding operand: " << part2 << endl;
          operands.push_back (part2);
          delete part1;
        } else {
          //cerr << "[ex1] adding operand: " << part1 << endl;
          operands.push_back (part1);
          delete part2;
        }
      }
    }
    filterBooleans (operands, false);
    if (operands.empty ())
      return new Boolean (false);
    if (operands.size () == 1)
      return operands.front ();
    Operator *retval = new Operator ("|", operands);
    //cerr << "returning: " << retval << endl;
    return retval;
  }
  cerr << "should not be reached" << endl;
  return NULL;
}

Node* LOGIC_CONVERTER::encode (const Node &F, map<string, IDTYPE> mapping, PARTITION P, const PETRIGRAPH &pg)
{
  if (typeid (F) == typeid (Operator)) {
    const Operator *oper = dynamic_cast<const Operator *>(&F);
    if (oper->name == "~") {
      Node *result = encode (*oper->operands[0], mapping, P, pg);
      //cerr << "[not] returning: " << result << endl;
      if (typeid (*result) == typeid (Boolean)) {
        ((Boolean *)result)->value = !((Boolean *)result)->value;
        return result;
      }
      return new Operator ("~", result, NULL);
    } else if (oper->name == "|") {
      Node *res0 = encode (*oper->operands[0], mapping, P, pg);
      Node *res1 = encode (*oper->operands[1], mapping, P, pg);
      //cerr << "[or] results: " << res0 << " || " << res1 << endl;
      if (typeid (*res0) == typeid (Boolean)) {
        if (((Boolean *)res0)->value) {
          delete res1;
          return res0;
        } else {
          delete res0;
          return res1;
        }
      }
      if (typeid (*res1) == typeid (Boolean)) {
        if (((Boolean *)res1)->value) {
          delete res0;
          return res1;
        } else {
          delete res1;
          return res0;
        }
      }
      return new Operator ("|", res0, res1);
    } else if (oper->name == "&") {
      Node *res0 = encode (*oper->operands[0], mapping, P, pg);
      Node *res1 = encode (*oper->operands[1], mapping, P, pg);
      //cerr << "[and] results: " << res0 << " || " << res1 << endl;
      if (typeid (*res0) == typeid (Boolean)) {
        if (!((Boolean *)res0)->value) {
          delete res1;
          return res0;
        } else {
          delete res0;
          return res1;
        }
      }
      if (typeid (*res1) == typeid (Boolean)) {
        if (!((Boolean *)res1)->value) {
          delete res0;
          return res1;
        } else {
          delete res1;
          return res0;
        }
      }
      return new Operator ("&", res0, res1);
    } else if (oper->name == "=") {
      return encode_equals (F, mapping, P, pg);
    }
  } else if (typeid (F) == typeid (Quantor)) {
    return encode_quantor (F, mapping, P, pg);
  }
  cerr << "should not be reached" << endl;
  return NULL;
}

};

};
