%{
  #include <stdlib.h>
  #include <stdio.h>

  #include "types.h"

  #include <iostream>

  extern "C" void lp_error (char *s);
  extern int lp_lineno;
  int lp_lex();
  extern "C" int lp_parse();
  Node *lp_startNode = NULL;
%}
%token LP_NOT LP_AND LP_OR LP_IMPLIES LP_IN LP_ALL1 LP_EX1 LP_NAME LP_LABEL
%token LP_CONNECTION LP_EQUAL LP_STRING LP_LPAREN LP_RPAREN LP_PAREN
%left LP_AND LP_OR LP_IMPLIES
%nonassoc LP_NOT
%union {
  Node* node;
  Operator *oper;
  Identifier *id;
  Function *func;
  Quantor *quant;
  char *string;
  unsigned int integer;
};
%type <oper> LP_EQUAL LP_NOT LP_AND LP_OR LP_IMPLIES LP_IN;
%type <string> LP_STRING LP_NAME;
%type <integer> LP_CONNECTION;
%type <id> name string;
%type <func> label connection;
%type <quant> LP_EX1 LP_ALL1;
%type <node> F start;
%%
start:
     F { lp_startNode = $1; }

F:
 name LP_EQUAL name { $$ = new Operator ("=", $1, $3); }
 | connection LP_EQUAL connection { $$ = new Operator ("=", $1, $3); }
 | label LP_EQUAL string { $$ = new Operator ("=", $1, $3); }
 | name LP_IN name { $$ = new Operator ("in", $1, $3); }
 | F LP_OR F { $$ = new Operator ("|", $1, $3); }
 | F LP_AND F { $$ = new Operator ("&", $1, $3); }
 | F LP_IMPLIES F { $$ = new Operator ("=>", $1, $3); }
 | LP_NOT F { $$ = new Operator ("~", $2, NULL); }
 | LP_ALL1 name F { $$ = new Quantor ("all1", $2, $3); }
 | LP_EX1 name F { $$ = new Quantor ("ex1", $2, $3); }
 | LP_LPAREN F LP_RPAREN { $$ = $2; }
 ;

string:
      LP_STRING { $$ = new Identifier ($1); free (lp_lval.string); };

name:
    LP_NAME { $$ = new Identifier ($1); free (lp_lval.string); };

connection:
          LP_CONNECTION LP_LPAREN name LP_RPAREN { $$ = new Connection ($1, $3); }
          ;

label:
     LP_LABEL LP_LPAREN name LP_RPAREN { $$ = new Label ($3); }
     ;

%%
void lp_error (char *s)
{
  fprintf (stderr, "line %d: %s\n", lp_lineno, s);
  exit (1);
}

extern int lp__flex_debug;

#ifdef TEST_PARSE
int main (void)
{
  lp__flex_debug = 0;
  lp_debug = 0;
  printf ("starting parse\n");
  lp_parse ();
  printf ("stopping parse\n");
  if (lp_startNode)
    std::cout << lp_startNode << std::endl;
  return 0;
}
#endif
