#ifndef TYPES_H
#define TYPES_H
#include <string>
#include <vector>
#include <list>
#include <iostream>

#ifdef TEST_PARSE
#define IDTYPE int
#else
#include <globals.h>
#endif

using namespace std;

class Node {
  friend ostream & operator<<(ostream &ostream, const Node *node);
  
protected:
  virtual string to_string() const = 0;

public:
  virtual Node* copy() const = 0;
  virtual ~Node();
};

class Identifier: public Node {
public:
  string name;

  Identifier (const char *name);
  virtual Identifier* copy() const;

protected:
  virtual string to_string() const;
};

class Function: public Node {
public:
  string name;
  Identifier *variable;
  
  Function (const char *name, Identifier *id);
  virtual Function* copy() const = 0;

protected:
  virtual string to_string() const = 0 ;
};

class Label: public Function {
public:
  Label (Identifier *id);
  virtual Label* copy() const;
  
protected:
  virtual string to_string() const;
};

class Connection: public Function {
public:
  unsigned int number;
  Connection (const unsigned int number, Identifier *id);
  virtual Connection* copy() const;
  
protected:
  virtual string to_string() const;
};

class Operator: public Node {
public:
  string name;
  vector<Node*> operands;

  Operator (const char *name, Node* op1, Node* op2);
  Operator (const char *name, const vector<Node*> &operands);
  Operator (const char *name, const list<Node*> &operands);
  virtual Operator* copy() const;
  
protected:
  virtual string to_string() const;
};

class Quantor: public Node {
public:
  string name;
  Identifier *variable;
  Node *formula;

  Quantor (const char *name, Identifier *id, Node *formula);
  virtual Quantor* copy() const;
  
protected:
  virtual string to_string() const;
};

class Variable: public Node {
public:
  IDTYPE edge;
  
  Variable (const IDTYPE edge);
  virtual Variable* copy() const;
  
protected:
  virtual string to_string() const;
};

class Constant: public Node {
public:
  int value;
  
  Constant (const int value);
  virtual Constant* copy() const;
  
protected:
  virtual string to_string() const;
};

class Boolean: public Node {
public:
  bool value;
  
  Boolean (const bool value);
  virtual Boolean* copy() const;
  
protected:
  virtual string to_string() const;
};

#endif
