/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "marking.h"

namespace data_structures
{
   MARKING::MARKING()
   {
   }

   MARKING::~MARKING()
   {
   }

   // void MARKING::addPlace(IDTYPE p)
   // {
      // this->marking[p] = vector<TOKEN>();
   // }

   void MARKING::deletePlace(IDTYPE p)
   {
      marking.erase(p);
   }

   vector<TOKEN> MARKING::getMarking(IDTYPE p)
   {
      return (marking.count(p) > 0)? marking[p]: vector<TOKEN>();
   }

   void MARKING::setMarking(IDTYPE p, vector<TOKEN> m)
   {
      marking[p] = m;
   }

   void MARKING::putToken(IDTYPE p, TOKEN tkn)
   {
     if(marking.count(p) == 0)
         marking[p] = vector<TOKEN>();

     if(marking[p].size()==0)
     {
      marking[p].push_back(tkn);
      return;
     }

     if(coloured)
     {
	// warning: MARKING is needed for second match handler
	
	TOKEN t = tkn;
	for(int i=0; i<marking[p].size(); i++)
	{
		if(EXPRESSION::equal(marking[p][i].getAttributes(), tkn.getAttributes()))
		{
			t.setCount(marking[p][i].getCount()+tkn.getCount());
			marking[p][i]=t;
			return;
		}
	}

	marking[p].push_back(tkn);
     }
     else
     {
 	 TOKEN t;
	 t.setCount(marking[p][0].getCount()+tkn.getCount());
	 marking[p][0]=t;
     }
   }

   void MARKING::removeTokens(IDTYPE p)
   {
	map< IDTYPE, vector<TOKEN> >::iterator it = marking.find(p);
	if(it != marking.end())
	 marking.erase(it);
   }

   ostream& operator<<(ostream& ost, MARKING& th)
   {
     ost << "MARKING:" << endl;
     if(coloured)
     {
	map< IDTYPE, vector<TOKEN> >::iterator markit;
        for(markit = th.marking.begin(); markit != th.marking.end(); markit++)
	{
		ost << "PLACE: " << markit->first << endl;
		for(vector<TOKEN>::iterator itt = markit->second.begin(); itt != markit->second.end(); itt++)
		{
		 ost << "NUMBER: " << markit->second[0].getCount() << *itt << endl;
		}
	}
     }
     else
     {
      map< IDTYPE, vector<TOKEN> >::iterator markit;
      for(markit = th.marking.begin(); markit != th.marking.end(); markit++)
         ost << "PLACE: " << markit->first << ", NUMBER: " << markit->second[0].getCount() << endl;
     } 
     return ost;
   } 

   bool operator==(const MARKING & m1, const MARKING & m2)
   {
     map<IDTYPE, vector<TOKEN> > map1 = (const_cast<MARKING*>(&m1))->getMarkings();
     map<IDTYPE, vector<TOKEN> > map2 = (const_cast<MARKING*>(&m2))->getMarkings();

     if(map1.size()!=map2.size())
      return false;

     for(map<IDTYPE, vector<TOKEN> >::iterator it1 = map1.begin();
     		it1!= map1.end(); it1++)
     {
      map<IDTYPE, vector<TOKEN> >::iterator it2 = map2.find(it1->first);
      if(it2==map2.end())
       return false;

      // number of tokens
      if(it1->second.size() != it2->second.size())
	return false;

      if(coloured)
      {
	for(vector<TOKEN>::iterator itt_1 = it1->second.begin(); itt_1 != it1->second.end(); itt_1++)
	{
		bool found = false;

		for(vector<TOKEN>::iterator itt_2 = it2->second.begin(); itt_2 != it2->second.end(); itt_2++)
		{
			if(EXPRESSION::equal(itt_1->getAttributes(), itt_2->getAttributes())) 
			{
				if(itt_1->getCount() == itt_2->getCount())
					found = true;
				else
					found = false;

				break;
			}
		}

		if(!found)
		 return false;
	}
      }
      else
      {
      	if(it1->second[0].getCount() != it2->second[0].getCount())
       		return false; 
      }
     }

     return true;
   }

   map< IDTYPE, vector<TOKEN> > & MARKING::getMarkings()
   {
    return marking;
   }

   void MARKING::setMarkings(map< IDTYPE, vector<TOKEN> > marking)
   {
    this->marking = marking;
   }

   map<IDTYPE, unsigned> MARKING::getUncoloured()
   {
    map<IDTYPE, unsigned> result;

    map< IDTYPE, vector<TOKEN> >::iterator markit;
    for(markit = marking.begin(); markit != marking.end(); markit++)
    {
	for(vector<TOKEN>::iterator it = markit->second.begin(); it != markit->second.end(); it++)
	 result[markit->first] += it->getCount();
    }

    return result;
   }

   unsigned MARKING::getTokenNumber(unsigned ID)
   {
    if(coloured)
    {
	unsigned count = 0;
	for(vector<TOKEN>::iterator it = marking[ID].begin(); it != marking[ID].end(); it++)
		count += it->getCount();
	return count;
    }
    else
    {
     if(marking.count(ID)>0)
     {
      // optimization
      // assert(marking[ID].size()>0);
      return marking[ID][0].getCount();
     } 
     else
      return 0;
    }
   }
   
   bool MARKING::IsMarked(unsigned ID)
   {
      if(marking.count(ID)>0)
      {
       // optimization
       // assert(marking[ID].size()>0);
       return true;
      }
      else
       return false;
   }

   bool MARKING::isColoured()
   {
	map< IDTYPE, vector<TOKEN> >::iterator it = marking.begin();

	if(it == marking.end())
	 return true;

	if(it->second[0].getAttributes().size() == 0)
	 return false;
	else
	 return true;
   }

//////////////////////////////////////////////////////////////
// static methods

bool MARKING::isEqual(MARKING & m1, MARKING & m2)
{
    map<IDTYPE, vector<TOKEN> > marking1 = m1.getMarkings();
    map<IDTYPE, vector<TOKEN> > marking2 = m2.getMarkings();

    if(marking1.size()!=marking2.size())
     return false;

    for(map<IDTYPE, vector<TOKEN> >::iterator it1 = marking1.begin(); it1 != marking1.end(); it1++)
    {
      map<IDTYPE, vector<TOKEN> >::iterator it2 = marking2.find(it1->first);

      // not found
      if(it2 == marking2.end())
	return false;

      // number of tokens
      if(it1->second.size() != it2->second.size())
	return false;

      // compare tokens
      if(coloured)
      {
	for(vector<TOKEN>::iterator itt_1 = it1->second.begin(); itt_1 != it1->second.end(); itt_1++)
	{
		bool found = false;

		for(vector<TOKEN>::iterator itt_2 = it2->second.begin(); itt_2 != it2->second.end(); itt_2++)
		{
			if(EXPRESSION::equal(itt_1->getAttributes(), itt_2->getAttributes())) 
			{
				if(itt_1->getCount() == itt_2->getCount())
					found = true;
				else
					found = false;

				break;
			}
		}

		if(!found)
		 return false;
	}
      }
      else
      {
      	if(it1->second[0].getCount() != it2->second[0].getCount())
       		return false; 
      }
    }

    return true;
}

bool MARKING::isLower(MARKING & m1, MARKING & m2)
{
    map<IDTYPE, vector<TOKEN> > marking1 = m1.getMarkings();
    map<IDTYPE, vector<TOKEN> > marking2 = m2.getMarkings();

    if(marking1.size()>marking2.size())
     return false;

    for(map<IDTYPE, vector<TOKEN> >::iterator it1 = marking1.begin(); it1 != marking1.end(); it1++)
    {
      map<IDTYPE, vector<TOKEN> >::iterator it2 = marking2.find(it1->first);

      // not found
      if(it2 == marking2.end())
	return false;

      // number of tokens
      if(it1->second.size() > it2->second.size())
	return false;

      // compare tokens
      for(vector<TOKEN>::iterator it3 = it1->second.begin(); it3 != it1->second.end(); it3++)
      {
	vector<TOKEN>::iterator it4;
	for(it4 = it2->second.begin(); it4 != it2->second.end(); it4++)
	{
		if(EXPRESSION::equal(it3->getAttributes(), it4->getAttributes()))
		{
			// token numbers
			if(it3->getCount() > it4->getCount())
				return false;

			break;
		}
	}

	// not found
	if(it4 == it2->second.end())
	 return false;
      }
    }

    return true;
}

bool MARKING::isLower_UC(MARKING & m1, MARKING & m2)
{
    map<IDTYPE, vector<TOKEN> > marking1 = m1.getMarkings();
    map<IDTYPE, vector<TOKEN> > marking2 = m2.getMarkings();

    if(marking1.size()>marking2.size())
     return false;

    for(map<IDTYPE, vector<TOKEN> >::iterator it1 = marking1.begin(); it1 != marking1.end(); it1++)
    {
      map<IDTYPE, vector<TOKEN> >::iterator it2 = marking2.find(it1->first);

      // not found
      if(it2 == marking2.end())
	return false;

      // numbers of tokens in m2
      int t_number = 0;
      for(vector<TOKEN>::iterator it = it2->second.begin(); it != it2->second.end(); it++)
	t_number += it->getCount();

      // number of tokens
      if(it1->second[0].getCount() > t_number)
	return false;
    }

    return true;
}


bool MARKING::isStrictlyLower(MARKING & m1, MARKING & m2)
{
    bool different = false; // set to true if any difference is found

    map<IDTYPE, vector<TOKEN> > marking1 = m1.getMarkings();
    map<IDTYPE, vector<TOKEN> > marking2 = m2.getMarkings();

    if(marking1.size()>marking2.size())
     return false;

    if(marking1.size()<marking2.size())
     different = true;

    for(map<IDTYPE, vector<TOKEN> >::iterator it1 = marking1.begin(); it1 != marking1.end(); it1++)
    {
      map<IDTYPE, vector<TOKEN> >::iterator it2 = marking2.find(it1->first);

      // not found
      if(it2 == marking2.end())
	return false;

      // number of tokens
      if(it1->second.size() > it2->second.size())
	return false;

      if(it1->second.size() < it2->second.size())
	different = true;

      // compare tokens
      for(vector<TOKEN>::iterator it3 = it1->second.begin(); it3 != it1->second.end(); it3++)
      {
	vector<TOKEN>::iterator it4;
	for(it4 = it2->second.begin(); it4 != it2->second.end(); it4++)
	{
		if(EXPRESSION::equal(it3->getAttributes(), it4->getAttributes()))
		{
			// token numbers
			if(it3->getCount() > it4->getCount())
				return false;

			if(it3->getCount() < it4->getCount())
				different = true;

			break;
		}
	}

	// not found
	if(it4 == it2->second.end())
	 return false;
      }
    }

    return different;
}

bool MARKING::isLowerPlace(MARKING & m1, MARKING & m2, IDTYPE place)
{
    map<IDTYPE, vector<TOKEN> > marking1 = m1.getMarkings();
    map<IDTYPE, vector<TOKEN> > marking2 = m2.getMarkings();

    map<IDTYPE, vector<TOKEN> >::iterator it1 = marking1.find(place);
    map<IDTYPE, vector<TOKEN> >::iterator it2 = marking2.find(place);

    // not found second
    if(it2 == marking2.end())
    {
	if(it1 == marking1.end())
	 return true;
	else
	 return false;
    }

    // not found first (but found second)
    if(it1 == marking1.end())
	return true;

    // number of tokens
    if(it1->second.size() > it2->second.size())
	return false;

    // compare tokens
    for(vector<TOKEN>::iterator it3 = it1->second.begin(); it3 != it1->second.end(); it3++)
    {
	vector<TOKEN>::iterator it4;
	for(it4 = it2->second.begin(); it4 != it2->second.end(); it4++)
	{
		if(EXPRESSION::equal(it3->getAttributes(), it4->getAttributes()))
		{
			// token numbers
			if(it3->getCount() > it4->getCount())
				return false;

			break;
		}
	}

	// not found
	if(it4 == it2->second.end())
	 return false;
     }

     return true;
}

bool MARKING::isStrictlyLowerPlace(MARKING & m1, MARKING & m2, IDTYPE place)
{
    bool different = false; // set to true if any difference is found

    map<IDTYPE, vector<TOKEN> > marking1 = m1.getMarkings();
    map<IDTYPE, vector<TOKEN> > marking2 = m2.getMarkings();

    map<IDTYPE, vector<TOKEN> >::iterator it1 = marking1.find(place);
    map<IDTYPE, vector<TOKEN> >::iterator it2 = marking2.find(place);

    // not found second
    if(it2 == marking2.end())
	return false;

    // not found first (but found second)
    if(it1 == marking1.end())
	return true;

    // number of tokens
    if(it1->second.size() > it2->second.size())
	return false;

    if(it1->second.size() < it2->second.size())
	different = true;

    // compare tokens
    for(vector<TOKEN>::iterator it3 = it1->second.begin(); it3 != it1->second.end(); it3++)
    {
	vector<TOKEN>::iterator it4;
	for(it4 = it2->second.begin(); it4 != it2->second.end(); it4++)
	{
		if(EXPRESSION::equal(it3->getAttributes(), it4->getAttributes()))
		{
			// token numbers
			if(it3->getCount() > it4->getCount())
				return false;

			if(it3->getCount() < it4->getCount())
				different = true;

			break;
		}
	}

	// not found
	if(it4 == it2->second.end())
	 return false;
    }

    return different;
}

void MARKING::removeTokens(IDTYPE p, TOKEN tkn)
{
	if(!coloured)
	{
	 cout << "Programming Error: Please use MARKING::removeTokens only in coloured case\n";
	 exit(1);
	}
	
	if(marking.count(p)==0)
	 return;

	vector<TOKEN>::iterator i=marking[p].begin();
	while(i!=marking[p].end())	
	{
		if(EXPRESSION::equal(i->getAttributes(), tkn.getAttributes()))
		{
			unsigned int count = i->getCount() - tkn.getCount();
			if(count <= 0)
			{
				marking[p].erase(i);
				break;
			}
			else
			{
				i->setCount(count);
			}
		}

		i++;
	}

	// check empty marking
	if(marking[p].size()==0)
	 marking.erase(p);
}

bool MARKING::ShortagePlace(MARKING & m1, MARKING & m2, IDTYPE place)
{
    map<IDTYPE, vector<TOKEN> > marking1 = m1.getMarkings();
    map<IDTYPE, vector<TOKEN> > marking2 = m2.getMarkings();

    map<IDTYPE, vector<TOKEN> >::iterator it1 = marking1.find(place);
    map<IDTYPE, vector<TOKEN> >::iterator it2 = marking2.find(place);

    // not found second
    if(it2 == marking2.end())
    	return false;

    // not found first (but found second)
    if(it1 == marking1.end())
	return true;

    // number of tokens
    if(it1->second.size() < it2->second.size())
	return true;

    // compare tokens
    for(vector<TOKEN>::iterator it3 = it2->second.begin(); it3 != it2->second.end(); it3++)
    {
	vector<TOKEN>::iterator it4;
	for(it4 = it1->second.begin(); it4 != it1->second.end(); it4++)
	{
		if(EXPRESSION::equal(it3->getAttributes(), it4->getAttributes()))
		{
			// token numbers
			if(it3->getCount() > it4->getCount())
				return true;

			break;
		}
	}

	// not found
	if(it4 == it1->second.end())
	 return true;
     }

     return false;
}

}
