%{
  #include <stdlib.h>
  #include <stdio.h>

  #include "../logic_lex_yacc/types.h"

  #include <iostream>

  extern "C" void mp_error (char *s);
  extern int mp_lineno;
  int mp_lex();
  extern "C" int mp_parse();
  Node *mp_startNode = NULL;
%}
%token MP_VARIABLE MP_AND MP_OR MP_GE MP_LE MP_GT MP_LT MP_EQUAL MP_NUMBER
%token MP_LPAREN MP_RPAREN MP_NOT
%left MP_AND MP_OR
%nonassoc MP_NOT
%union {
  Node* node;
  Operator *oper;
  Variable *var;
  Constant *cons;
  unsigned int integer;
};
%type <oper> MP_EQUAL MP_AND MP_OR MP_GE MP_LE MP_GT MP_LT MP_NOT;
%type <integer> MP_NUMBER;
%type <var> variable;
%type <cons> constant;
%type <node> F start;
%%
start:
     F { mp_startNode = $1; }

F:
 MP_LPAREN F MP_RPAREN { $$ = $2; }
 | F MP_AND F { $$ = new Operator ("&", $1, $3); }
 | F MP_OR F { $$ = new Operator ("|", $1, $3); }
 | MP_NOT F { $$ = new Operator ("~", $2, NULL); }
 | variable MP_EQUAL constant { $$ = new Operator ("=", $1, $3); }
 | variable MP_GE constant { $$ = new Operator (">=", $1, $3); }
 | variable MP_GT constant { $$ = new Operator (">", $1, $3); }
 | variable MP_LE constant { $$ = new Operator ("<=", $1, $3); }
 | variable MP_LT constant { $$ = new Operator ("<", $1, $3); }
 ;

variable:
       MP_VARIABLE { $$ = new Variable (mp_lval.integer); }

constant:
        MP_NUMBER { $$ = new Constant (mp_lval.integer); }

%%
void mp_error (char *s)
{
  fprintf (stderr, "line %d: %s\n", mp_lineno, s);
  exit (1);
}

extern int mp__flex_debug;

#ifdef TEST_PARSE
int main (void)
{
  mp__flex_debug = 0;
  mp_debug = 0;
  printf ("starting parse\n");
  mp_parse ();
  printf ("stopping parse\n");
  if (mp_startNode)
    std::cout << mp_startNode << std::endl;
  return 0;
}
#endif
