/***************************************************************************
 *   Copyright (C) 2006 by root   *
 *   root@Vitali   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>

#include "marking_reader.h"

namespace algorithms {

namespace readers {

MARKING_READER::MARKING_READER()
{
}


MARKING_READER::~MARKING_READER()
{
}

void MARKING_READER::initialize(char *filename)
{
	this->filename = filename;
}

list<MARKING> MARKING_READER::run(int type)
{
  list<MARKING> lm;

  string s = readFile(filename);
  string::size_type a,b;
  
  for (a = s.find("MARKING"); a < s.size(); a = s.find("MARKING")) {
    b = s.find(";");
    if (b < s.size()) {
      MARKING m = getMarking(string(s,a,(b-a)+1));
      lm.push_back(m);	
      s.erase(0,b+1);
    } else {
      this->error_exit("No delimiting \";\" found");
    }
  }

  return lm;
}

string MARKING_READER::readFile(const char* file)
{ 
  ostringstream os;
  char ch;
  std::ifstream input(file);
  if (!input) { 
    this->error_exit("Failed to open File: \"" +string(file)+"\"");
  }
  
  while (input.get(ch)) { 
    if ((ch != ' ') && (ch != '\n')) {
      os.put(ch);
    }
  }
  return os.str();
}

MARKING MARKING_READER::getMarking(string s)
{
  MARKING m;
  string token = "MARKING";
  string in = s;
  in.erase(0,token.size());
  string::size_type b,c;

  c = in.find_first_of(",;");

  while((!(in.empty())) && c < in.size()) 
  {
    b = in.find(":"); // find delimiter between place and value

    if (b < c)
    {
      string place = string(in,0,b);
      string val = string(in,b+1,c-b);

      // check if place-name is not empty
      if (!place.empty())
      {
	char * str = new char(place.size()+1);
	strcpy(str, place.c_str());
	char * pch = strtok (str,"_");

	string sid;
		
	while (pch!=NULL)
  	{
		sid = (string)pch;
    		pch = strtok (NULL, "_");
  	}

	TOKEN tkn;
	tkn.setCount(atoi(val.c_str()));
	m.putToken(atoi(sid.c_str()), tkn);
      }
      else
      {
	this->error_exit("Marking includes empty place name.\n");
      }
    }
    in.erase(0,c+1); // erase sub-string until one of [,;]
    c = in.find_first_of(",;"); // find next delimiter
  }

  return m;
}

// exit with error message
void MARKING_READER::error_exit(const string in) {
  cout << in << endl; 
  std::exit(1);
}

// namespace

}

}


