
/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "matchhandler.h"
#include <sstream>

#include "check_coverability.h"
#include "globals.h"

namespace algorithms
{
   namespace matchhandler
   {
      MATCHHANDLER::MATCHHANDLER()
      {
       alg_coverability=NULL;
      }
      MATCHHANDLER::~MATCHHANDLER()
      {
      }

      TRIVIAL_MATCHHANDLER::TRIVIAL_MATCHHANDLER()
      {
	 this->ID = ALGORITHM_TRIVIAL_MATCHHANDLER;
	 this->name = "trivial_matchhandler";
      }
      TRIVIAL_MATCHHANDLER::~TRIVIAL_MATCHHANDLER()
      {
      }

      void TRIVIAL_MATCHHANDLER::initialize()
      {
      }
      void TRIVIAL_MATCHHANDLER::initialize( PETRIGRAPH * i_petrigraph,
				             IDTYPE i_prule_id, RULE * i_prule,
				             bool * i_canFold, bool * i_canUnfold,
				             MATCH* m1, MATCH* m2, MARKING* i_phi_E_L )
      {
      }

      bool TRIVIAL_MATCHHANDLER::run(MATCH * match)
      {
	 assert(match != NULL);
	 if(debug) cout << "running trivial matchhandler\n";

	 return true;
      }

      FIRST_MATCHHANDLER::FIRST_MATCHHANDLER()
      {
	 this->ID = ALGORITHM_FIRST_MATCHHANDLER;
	 this->name = "first_matchhandler";
      }
      FIRST_MATCHHANDLER::~FIRST_MATCHHANDLER()
      {
      }

      void FIRST_MATCHHANDLER::initialize(PETRIGRAPH * i_petrigraph, IDTYPE i_prule_id,
					  RULE * i_prule, bool * i_canFold, bool * i_canUnfold,
					  MATCH* i_phicpy, MATCH* i_phi2cpy)
      {
	 assert(i_petrigraph != NULL);
	 assert(i_prule != NULL);

	 this->petrigraph = i_petrigraph;
	 this->prule_id = i_prule_id;
	 this->prule = i_prule;
	 this->canFold = (i_canFold != NULL) ? i_canFold : (new bool);
	 this->canUnfold = (i_canUnfold != NULL) ? i_canUnfold : (new bool);
	 *(this->canFold) = false;
	 *(this->canUnfold) = false;
 
	 this->phicpy = (i_phicpy != NULL)? i_phicpy: (new MATCH());
	 this->phi2cpy = (i_phi2cpy != NULL)? i_phi2cpy: (new MATCH());
      }
      void FIRST_MATCHHANDLER::initialize(PETRIGRAPH * i_petrigraph, IDTYPE i_prule_id,
					  RULE * i_prule,
					  bool * i_canFold, bool * i_canUnfold)
      {
         this->initialize( i_petrigraph, i_prule_id, i_prule,
	                   i_canFold, i_canUnfold, NULL, NULL);
      }
      void FIRST_MATCHHANDLER::initialize(PETRIGRAPH * i_petrigraph,
					  IDTYPE i_prule_id, RULE * i_prule)
      {
	 this->initialize(i_petrigraph, i_prule_id, i_prule,
			  NULL, NULL);
      }
      void FIRST_MATCHHANDLER::initialize( PETRIGRAPH * i_petrigraph,
				           IDTYPE i_prule_id, RULE * i_prule,
				           bool * i_canFold,
				           bool * i_canUnfold,
				           MATCH* m1, MATCH* m2, MARKING* i_phi_E_L )
      {
         this->initialize( i_petrigraph, i_prule_id, i_prule,
	                   i_canFold, i_canUnfold, m1, m2);
      }


      /*
       * *phi2_out can be used as a a second return value
       */
      bool FIRST_MATCHHANDLER::run(MATCH * phi_in, MATCH * phi2_out)
      {
	 assert(phi_in != NULL);
	 bool use_this_match = false;
	 // if(debug) cout << "running first matchhandler\n";
	 /*
	  * additional output phi2_out is optional
	  */
	 if(phi2_out == NULL) phi2_out = this->phi2cpy;
	 MATCH* phi2 = (phi2_out != NULL) ? phi2_out : (new MATCH);
	 *phi2 = MATCH();

	 /*
	  * calculate phi_E_L
	  */
	 MARKING phi_E_L = MARKING();
	 set < IDTYPE > places = this->petrigraph->getPlaces();
	 set < IDTYPE >::iterator placeit;
	 // for(placeit = places.begin(); placeit != places.end(); placeit++)
	 // {
	 //   phi_E_L.addPlace(*placeit);
	 // }
	 map < IDTYPE, EDGE > E_L = this->prule->getLeft()->getEdges();
	 map < IDTYPE, EDGE >::iterator elit;

	 for(elit = E_L.begin(); elit != E_L.end(); elit++)
	 {
	    TOKEN t;
	    t.setCount(1);
	    phi_E_L.putToken(phi_in->getMatch(elit->first), t);
	 }

	 /*
	  * AUGUR I says: "now [calculate??] the depth of vertices"
	  */
	 map < IDTYPE, VERTEX > V_L = this->prule->getLeft()->getVertices();
	 map < IDTYPE, VERTEX >::iterator vlit;
	 bool use_depth = true;
	 bool use_maxdepth = true;

	 /*
	  * does a match phi2 exist satisfying condition 6?
	  */
	 // ALGORITHM* handler = new SECOND_MATCHHANDLER();
	 // handler->initialize(this->petrigraph, this->prule_id, this->prule, &phi_E_L,
	 //                     phi_in, this->depth);
	 ALGORITHM* findmatch = this->load("call_findmatch");
	 findmatch->initialize((HYPERGRAPH *) (this->petrigraph), phi_in, phi2, canFold, canUnfold, this->petrigraph);
	 bool condition6holds = false;
	 // *phi2 = MATCH();

// Here it exists some problem probably concerned with the marking phi_E_L
// If coloured=true, then it does not work correct.
// Just put here temporarily coloured = false 
// It does not influence attributes in the result, because here we only search for a match
// To do: Find the problem

bool trick=false;
if(coloured)
{
	trick = true;
	coloured = false;
}
	 *phi2 = *(findmatch->run(this->prule->getLeft(), &condition6holds, prule_id, &phi_E_L));

if(trick)
	coloured = true;

	 delete findmatch;

	 /*
	  * checking depth values once more is important because of vertices and of unfdepth
	  */
	 for(vlit = V_L.begin(); (vlit != V_L.end()) && (use_depth && use_maxdepth); vlit++)
	 {
	    // do not make this for isolated nodes
	    // they are not in the MATCH -> by getMatch() the severe error will be obtained
	    if(this->prule->getLeft()->getConnectedEdges(vlit->first).size()==0)
		continue;

	    IDTYPE gvertex = phi_in->getMatch(vlit->first);
	    unsigned dep = this->petrigraph->getDepth(gvertex);

	    if(condition6holds)
	    {
	       /*
	        * AUGUR I: "if phi != phi2 and not depth then it doesn't pass for folding"
	        */
	       if((phi2->getMatch(vlit->first) != gvertex) && (dep < depth))
	       {
		  use_depth = false;
	       }
	    }

	    /*
	     * AUGUR I: "if more than unfdepth then it doesn't pass for unfolding"
	     */
	    if((unfdepth >= 0) && (dep >= (unsigned)(unfdepth)))
	       use_maxdepth = false;
	 }


	 /*
	  * repeat this for the edges
	  */
	 for(elit = E_L.begin(); (elit != E_L.end()) && (use_depth && use_maxdepth); elit++)
	 {
	    IDTYPE gedge = phi_in->getMatch(elit->first);
	    unsigned dep = this->petrigraph->getDepth(gedge);
	    if((condition6holds) && (dep < depth) &&
	       (phi2->getMatch(elit->first) != gedge))
	       use_depth = false;
	    if((unfdepth >= 0) && (dep >= (unsigned)(unfdepth)))
	       use_maxdepth = false;
	 }

	 if(debug) cout << "\n|phi2|: " << phi2->count() << "\tcondition 6: " << condition6holds;
	 if(debug) cout << "\tuse depth: " << use_depth << "\tnfold: " << nfold << "\n";
	 *(this->canFold) = false;
	 *(this->canUnfold) = false;

	 if(!condition6holds || !use_depth || nfold)  // do not fold
	 {
	    if(debug) cout << "confirming " << *phi_in << "\n";

	    /*
	     * unfolding possible? first, let us check maxdepth
	     */
	    if(!use_maxdepth)
	    {
	       if(phi2_out == NULL)
		  delete phi2;
	       return false;
	    }

	    /*
	     * is there a "this->prule_id"-labelled transition 
	     * tau in T_N such that preset(tau) = phi_E_L?
	     */
	    map < IDTYPE, TRANSITION > T_N = this->petrigraph->getTransitions();
	    map < IDTYPE, TRANSITION >::iterator tau;
	    for(tau = T_N.begin(); tau != T_N.end(); tau++)
	    {
	       /*
	        * p_N_i(tau) == this->prule_id?
	        */
	       unsigned taurule_id;
	       /*
	       istringstream ist;
	       ist.str(tau->second.getLabel());
	       ist >> taurule_id;
	       */
	       taurule_id = this->petrigraph->getAssociatedRule(tau->first)->getID();
	       if(this->prule_id == taurule_id)
	       {
	          if(debug)
		     cout << "comparing instance of rule ID " << this->prule_id
		          << " and t" << tau->first << "\n";
		  vector < TCONNECTION > tau_preset =
		     this->petrigraph->getPreset(tau->first);
		  vector < TCONNECTION >::iterator tpit;
		  
		  if(debug) cout << "is phi_E_L (" << phi_E_L << ") == *t" <<  tau->first << "?\n";
		  /*
		   * preset(tau) == phi_E_L?
		   */
		  bool equal = true;
		  for(placeit = places.begin(); placeit != places.end(); placeit++)
		  {
		     unsigned long weight = 0;
		     for(tpit = tau_preset.begin(); tpit != tau_preset.end(); tpit++)
			if(tpit->ID == *placeit) weight += tpit->weight;
		     if(phi_E_L.getTokenNumber(*placeit) != weight)
		     {
		        if(debug) cout << "marking at p" << *placeit << " = " << phi_E_L.getTokenNumber(*placeit) << " != " << weight << ". proceeding.\n";
			equal = false;
			break;
		     }
		  }

		  if(equal)
		  {
		     if(debug) cout << "this match was unfolded before. it will be discarded.\n";
		     if(phi2_out == NULL)
			delete phi2;
		     return false;
		  }
	       }
	    }
            if(debug) cout << "no previous unfolding of this match detected. proceeding.\n";
	    
	    // coverability check
	    bool coverable = true;
	    if(!ncov)
	    {
	     if(alg_coverability==NULL)
	     {
		alg_coverability = this->load("call_coverability");
		alg_coverability->initialize( (PTNET*)this->petrigraph);
	     }	

	     MARKING m = petrigraph->match2marking(*phi_in);

	     long psize;
	     if(cov_test)
	      init_time();

	     coverable=alg_coverability->run(m, &psize);

	     if(cov_test)
	        cout << "time used: " << get_time() << " sec." << endl; 
	    }
	    if(coverable)
	    {
	       *(this->canUnfold) = true;
	       use_this_match = true;

		// test coverability
		// call coverability one more time not incremental
		if(cov_test)
		{
			if(coloured)
			{
				cout << "\nerror: TEST COVERABILITY with attributed GTS is not implemented\n";
				exit(1);
			}
			ALGORITHM * alg_cover = this->load("call_coverability_test");
			assert(alg_cover!=NULL);
			alg_cover->initialize( (PTNET*)this->petrigraph);
	                long psize2 = 0;
			MARKING m = petrigraph->match2marking(*phi_in);
			init_time();
	     		if(alg_cover->run(m, &psize2))
	     	         cout << "time used: " << get_time() << " sec." << endl;

			// compare two coverability graphs
			CHECK_COVERABILITY cc;
			cc.initialize(alg_coverability, alg_cover);
			if(!cc.run())
			{
				cout << "problems with coverability\n";
				cout << *(((COVERABILITY*)alg_coverability)->getCovGraph());
				cout << *(((COVERABILITY*)alg_cover)->getCovGraph());
				exit(1);
			}
	     	}
	    }
	    else use_this_match = false;
	 }
	 else  // a matching satisfying condition 6 was found
	 {
	    // coverability check
	    bool coverable = true;
	    if(!ncov)
	    {
	     if(alg_coverability==NULL)
	     {
		alg_coverability = this->load("call_coverability");
		alg_coverability->initialize( (PTNET*)this->petrigraph);
	     }

	     MARKING m = petrigraph->match2marking(*phi_in);

	     long psize = 0;
	     if(cov_test)
	      init_time();

	     coverable=alg_coverability->run(m, &psize);

	     if(cov_test)
	      cout << "time used: " << get_time() << " sec." << endl;
	    } 

	    if(coverable)
	    {
	       *(this->canFold) = true;
	       use_this_match = true;

		// test coverability
		// call coverability one more time not incremental
		if(cov_test)
		{
			ALGORITHM * alg_cover = this->load("call_coverability_test");
			assert(alg_cover!=NULL);
			alg_cover->initialize( (PTNET*)this->petrigraph);
	                long psize2 = 0;
			MARKING m = petrigraph->match2marking(*phi_in);
			init_time();
	     		if(alg_cover->run(m, &psize2))
	     	         cout << "time used: " << get_time() << " sec." << endl;

			// compare two coverability graphs
			CHECK_COVERABILITY cc;
			cc.initialize(alg_coverability, alg_cover);
			if(!cc.run())
			{
				cout << "problems with coverability\n";
				cout << *(((COVERABILITY*)alg_coverability)->getCovGraph());
				cout << *(((COVERABILITY*)alg_cover)->getCovGraph());
				exit(1);
			}
	     	}
	    }
	    else use_this_match = false;
	 }

	 *(this->phicpy) = *phi_in;
	 if(phi2_out != NULL) *(this->phi2cpy) = *phi2_out;
	 
	 if(phi2_out == NULL)
	    delete phi2;
	 return use_this_match;
      }
      bool FIRST_MATCHHANDLER::run(MATCH * match)
      {
	 return this->run(match, NULL);
      }

      SECOND_MATCHHANDLER::SECOND_MATCHHANDLER()
      {
	 this->ID = ALGORITHM_SECOND_MATCHHANDLER;
	 this->name = "second_matchhandler";
      }
      SECOND_MATCHHANDLER::~SECOND_MATCHHANDLER()
      {
      }

      void SECOND_MATCHHANDLER::initialize(PETRIGRAPH * i_petrigraph, IDTYPE i_prule_id,
					   RULE * i_prule, MARKING * i_phi_E_L,
					   MATCH * i_match_first)
      {
	 assert(i_petrigraph != NULL);
	 assert(i_prule != NULL);
	 assert(i_phi_E_L != NULL);
	 assert(i_match_first != NULL);

	 this->petrigraph = i_petrigraph;
	 this->prule_id = i_prule_id;
	 this->prule = i_prule;
	 this->phi_E_L = i_phi_E_L;
	 this->match_first = i_match_first;
      }
      void SECOND_MATCHHANDLER::initialize( PETRIGRAPH * i_petrigraph,
				            IDTYPE i_prule_id, RULE * i_prule,
				            bool * i_canFold,
				            bool * i_canUnfold,
				            MATCH* m1, MATCH* m2, MARKING* i_phi_E_L )
      {
         this->initialize(i_petrigraph, i_prule_id, i_prule, i_phi_E_L, m1);
	 if(debug) cout << "PHI2 promulgation: match_second <- " << m2 << "\n";
	 this->match_second = m2;
      }

      bool SECOND_MATCHHANDLER::run(MATCH * match)
      {  
	 assert(match != NULL);
	 if(debug)
	 {
	    cout << "\nrunning second matchhandler:\n";
	    cout << "match: " << *match << "\n" << "phi_E_L: " << *phi_E_L << "\n";
	 }

	 /*
	  * phi(E_L) is passed on initialization. Calculate phi2(E_L):
	  */
	 MARKING phi2_E_L;
	 set < IDTYPE > places = this->petrigraph->getPlaces();
	 set < IDTYPE >::iterator placeit;
	 // for(placeit = places.begin(); placeit != places.end(); placeit++)
	 // {
	 //   phi2_E_L.addPlace(*placeit);
	 // }
	 map < IDTYPE, EDGE > E_L = this->prule->getLeft()->getEdges();
	 map < IDTYPE, EDGE >::iterator elit;
	 for(elit = E_L.begin(); elit != E_L.end(); elit++)
	 {
	    TOKEN t;
	    t.setCount(1);
	    phi2_E_L.putToken(match->getMatch(elit->first), t);
	 }
	 if(debug) cout << "phi2_E_L: " << phi2_E_L << "\n";

	 /*
	  * AUGUR I says: "we need two different matchings. phi == phi = phi2?"
	  * Actually, two markings are compared. We consider all tokens as equal.
	  */
	 bool phi_equals_phi2 = true;
	 for(placeit = places.begin(); placeit != places.end(); placeit++)
	 {
	    if(!(phi2_E_L==*phi_E_L))
	    {
	       phi_equals_phi2 = false;
	    }

	 }
	 if(phi_equals_phi2)
	 {
	    if(debug) cout << "discarded because phi equals phi2.\n";
	    return false;
	 }

	 /*
	  * let us check the depth of edges
	  */
	 for(elit = E_L.begin(); elit != E_L.end(); elit++)
	 {
	    /*
	     * depth of gedge
	     */
	    IDTYPE gedge = match->getMatch(elit->first);
	    unsigned long dep = this->petrigraph->getDepth(gedge);

	    /*
	     * AUGUR I says: "if phi != phi2 and not depth then it doesn't pass"
	     * "not depth" means that the considered edge has lower than maximal depth
	     */
	    if((this->match_first->getMatch(elit->first) != gedge) && (depth > dep))
	    {
	       if(debug)
	       {
	          cout << "this->depth: " << depth << "\n";
	          cout << "edge in phi1: e" << this->match_first->getMatch(elit->first) << "\n";
	          cout << "edge in phi2: e" << gedge << ":d" << dep << "\n";
		  cout << "\"if phi != phi2 and not depth then it doesn't pass\" (AUGUR I)\n";
	       }
	       return false;
	    }
	 }
	 /*
	  * now consider the depth values of vertices
	  */
	 map < IDTYPE, VERTEX > V_L = this->prule->getLeft()->getVertices();
	 map < IDTYPE, VERTEX >::iterator vlit;
	 for(vlit = V_L.begin(); vlit != V_L.end(); vlit++)
	 {
	    // do not make this for isolated nodes
	    // they are not in the MATCH -> by getMatch() the severe error will be obtained
	    if(this->prule->getLeft()->getConnectedEdges(vlit->first).size()==0)
		 continue;

	    IDTYPE gvertex = match->getMatch(vlit->first);
	    unsigned long dep = this->petrigraph->getDepth(gvertex);
	    if((this->match_first->getMatch(vlit->first) != gvertex) && (depth > dep))
	    {
	       if(debug)
	       {
	          cout << "this->depth: " << depth << "\n";
	          cout << "vertex in phi1: v" << this->match_first->getMatch(vlit->first) << "\n";
	          cout << "vertex in phi2: v" << gvertex << ":d" << dep << "\n";
		  cout << "\"if phi != phi2 and not depth then it doesn't pass\" (AUGUR I)\n";
	       }
	       return false;
	    }
	 }

	 /*
	  * Does a transition t exist such that ...
	  */
	 map < IDTYPE, TRANSITION > transitions = this->petrigraph->getTransitions();
	 map < IDTYPE, TRANSITION >::iterator transit;
	 ALGORITHM* causality_relation = this->load("call_causality");
	 causality_relation->initialize((PTNET *) (this->petrigraph));
	 for(transit = transitions.begin(); transit != transitions.end(); transit++)
	 {
	    /*
	     * ... p_N_i(t) == r and ...
	     */
	    unsigned transition_id;
	    /* istringstream ist;
	    ist.str(transit->second.getLabel());
	    ist >> transition_id; */
	    transition_id = this->petrigraph->getAssociatedRule(transit->first)->getID();
	    if(debug) cout << " {t" << transit->first << "/r" << transition_id << "?}";
	    if(this->prule_id == transition_id)
	    {
	       /*
	        * ... preset of t == phi2_E_L and ...
	        */
	       vector < TCONNECTION > tpreset = this->petrigraph->getPreset(transit->first);
	       vector < TCONNECTION >::iterator tpit;
	       bool equal = true;
	       for(placeit = places.begin(); placeit != places.end(); placeit++)
	       {		       
		  unsigned long weight = 0;
		  for(tpit = tpreset.begin(); tpit != tpreset.end(); tpit++)
		  {		     
		     if(tpit->ID == *placeit)
			weight += tpit->weight;
		  }
		  if(phi2_E_L.getTokenNumber(*placeit) != weight)
		     equal = false;
	       }
	       if(debug) cout << "[[*t" << transit->first << " == phi2_E_L?] " << equal << "]"; 

	       if(equal)
	       {
		  /* 
		   * for all e in phi_E_L: e in preset(t) or t <N e
		   * note that we just checked that *t == phi2_E_L, so the condition 
		   * of "e in *t" transforms to "e is marked in phi2_E_L".
		   */
		  bool e_condition_satisfied = true;
		  for(placeit = places.begin();
		      e_condition_satisfied && (placeit != places.end()); placeit++)
		  {
		     if(this->phi_E_L->IsMarked(*placeit) &&  // for all e in phi_E_L exclude:
			!phi2_E_L.IsMarked(*placeit) )  // e is not marked in phi2_E_L and ...
		     { 
		        /*
			 * ... does not depend on t.
			 */
			bool depends = causality_relation->run(transit->first, *placeit);
			if(debug) cout << "[[t" << transit->first << " <N p" << *placeit << "?] " << depends << "]";
			
			/*
			 * e_condition is satisfied only if *placeit is in tpreset
			 * as we checked before, this is the case if and only if
			 * the place *placeit is marked in phi2_E_L.
			 */
			if(!depends) e_condition_satisfied = false;
		     }
		  }
                  if(debug) cout << "[[\"e condition\" satisfied?] " << e_condition_satisfied << "] ";
		  if(e_condition_satisfied)
		  {
		     *(this->match_second) = *match;
		     if(debug) cout << "\nPHI2 promulgation: *(" << match_second << ") <- " << *match << "\n";
		     return true;
		  }
	       }
	       if(debug) cout << ' ';
	    }
	    else if (debug) cout << "[[wrong rule]] ";
	 }

	 if(debug) cout << "no corresponding instance of rule " << prule_id << " was found. discarding.\n";
	 return false;
      }

   };
};
