
/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef ALGORITHMSMATCHHANDLER_H
#define ALGORITHMSMATCHHANDLER_H

#include "algorithm.h"
#include "causality_relation.h"
#include "match.h"
#include "petrigraph.h"
#include "findmatch.h"

namespace algorithms
{
   namespace matchhandler
   {
      class MATCHHANDLER:public ALGORITHM
      {
       public:
	 MATCHHANDLER();
	 virtual ~ MATCHHANDLER();

	 virtual void initialize(PETRIGRAPH * i_petrigraph,
				 IDTYPE i_prule_id, RULE * i_prule,
				 bool * i_canFold, bool * i_canUnfold,
				 MATCH* m1, MATCH* m2, MARKING* i_phi_E_L) = 0;
	 virtual bool run(MATCH * match) = 0;
	 
       protected:

        // coverability algorithm
        ALGORITHM * alg_coverability;
      };

      class TRIVIAL_MATCHHANDLER:public MATCHHANDLER
      {
       public:
	 TRIVIAL_MATCHHANDLER();
	 virtual ~ TRIVIAL_MATCHHANDLER();

	 virtual void initialize();
	 virtual void initialize(PETRIGRAPH * i_petrigraph,
				 IDTYPE i_prule_id, RULE * i_prule,
				 bool * i_canFold, bool * i_canUnfold,
				 MATCH* m1, MATCH* m2, MARKING* i_phi_E_L);
	 virtual bool run(MATCH * match);
      };

      class FIRST_MATCHHANDLER:public MATCHHANDLER
      {
       public:
	 FIRST_MATCHHANDLER();
	 virtual ~ FIRST_MATCHHANDLER();

	 virtual void initialize(PETRIGRAPH * i_petrigraph,
				 IDTYPE i_prule_id, RULE * i_prule);
	 virtual void initialize(PETRIGRAPH * i_petrigraph,
				 IDTYPE i_prule_id, RULE * i_prule,
				 bool * i_canFold, bool * i_canUnfold);
	 virtual void initialize(PETRIGRAPH * i_petrigraph,
				 IDTYPE i_prule_id, RULE * i_prule,
				 bool * i_canFold, bool * i_canUnfold,
				 MATCH* i_phicpy, MATCH* i_phi2cpy);
	 virtual void initialize(PETRIGRAPH * i_petrigraph,
				 IDTYPE i_prule_id, RULE * i_prule,
				 bool * i_canFold, bool * i_canUnfold,
				 MATCH* m1, MATCH* m2, MARKING* i_phi_E_L);
	 virtual bool run(MATCH * phi_in, MATCH * phi2_out);
	 virtual bool run(MATCH * match);

	 bool *canFold;
	 bool *canUnfold;

       protected:
	 PETRIGRAPH * petrigraph;
	 IDTYPE prule_id;
	 RULE *prule;
	 MATCH* phicpy;
	 MATCH* phi2cpy;
      };

      class SECOND_MATCHHANDLER:public MATCHHANDLER
      {
       public:
	 SECOND_MATCHHANDLER();
	 virtual ~ SECOND_MATCHHANDLER();

	 virtual void initialize(PETRIGRAPH * i_petrigraph, IDTYPE i_prule_id,
				 RULE * i_prule, MARKING * i_phi_E_L,
				 MATCH * i_match_first);
	 virtual void initialize(PETRIGRAPH * i_petrigraph,
				 IDTYPE i_prule_id, RULE * i_prule,
				 bool * i_canFold, bool * i_canUnfold, 
				 MATCH* m1, MATCH* m2, MARKING* i_phi_E_L);
	 virtual bool run(MATCH * match);

       protected:
	 PETRIGRAPH * petrigraph;
	 IDTYPE prule_id;
	 RULE *prule;
	 MARKING *phi_E_L;
	 MATCH *match_first;
	 MATCH *match_second;
      };
   };
};
using namespace algorithms;
using namespace algorithms::matchhandler;

#endif
