/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef ALGORITHM_MERGE_PETRIGRAPH_CPP
#define ALGORITHM_MERGE_PETRIGRAPH_CPP

#include "merge_petrigraph.h"
#include "algorithm_storage.h"

namespace data_structures
{
   MERGE_RELATION::MERGE_RELATION(PETRIGRAPH* pg)
   {
      this->petrigraph = pg;
      this->rel[IS_VERTEX] = map< IDTYPE, set<IDTYPE> >();
      this->rel[IS_EDGE] = map< IDTYPE, set<IDTYPE> >();
      this->rel[IS_TRANSITION] = map< IDTYPE, set<IDTYPE> >();
   }
   MERGE_RELATION::~MERGE_RELATION() {}
   
   void MERGE_RELATION::addRel(IDTYPE x, IDTYPE y, GRAPH_ELEMENT_TYPE which)
   {
      // if(debug) cout << "   addRel(" << x << ", " << y << ", [[" << (unsigned)(which) << "]])\n";
      map< IDTYPE, set<IDTYPE> >::iterator xrel;
      xrel = this->rel[which].find(x);
      if(xrel == this->rel[which].end())
      {
         this->rel[which][x] = set<IDTYPE>();
	 xrel = this->rel[which].find(x);
      }
      xrel->second.insert(y);
   }
   
   bool MERGE_RELATION::isInRel(IDTYPE x, IDTYPE y, GRAPH_ELEMENT_TYPE which)
   {
      map< IDTYPE, set<IDTYPE> >::iterator xrel;
      xrel = this->rel[which].find(x);
      if(xrel == this->rel[which].end())
      {
         this->rel[which][x] = set<IDTYPE>();
      }
      return (this->rel[which][x].count(y) > 0);
   }
   
   map<IDTYPE, IDTYPE> MERGE_RELATION::getExtendingEquivRelation()
   {
      /*
       * make sure that every element is in relation with itself
       */
      map<IDTYPE, VERTEX> V = this->petrigraph->getVertices();
      map<IDTYPE, EDGE> E = this->petrigraph->getEdges();
      map<IDTYPE, TRANSITION> T = this->petrigraph->getTransitions();
      for(map<IDTYPE, VERTEX>::iterator vit = V.begin(); vit != V.end(); vit++)
         this->addRel(vit->first, vit->first, IS_VERTEX);
      for(map<IDTYPE, EDGE>::iterator eit = E.begin(); eit != E.end(); eit++)
         this->addRel(eit->first, eit->first, IS_EDGE);
      for(map<IDTYPE, TRANSITION>::iterator tit = T.begin(); tit != T.end(); tit++)
         this->addRel(tit->first, tit->first, IS_TRANSITION);
	 
      /*
       * extend to a closed equivalence relation
       * iterate until no extension could be performed
       */
      bool isClosedEquivRelation = false;
      if(debug) cout << "\n";
      while(!isClosedEquivRelation)
      {
         if(debug) cout << "e";
         isClosedEquivRelation = true;
	 
	 /*
	  * for each pair of transitions:
	  */
	 for(map<IDTYPE, TRANSITION>::iterator t1 = T.begin(); t1 != T.end(); t1++)
	 {
	    for(map<IDTYPE, TRANSITION>::iterator t2 = T.begin(); t2 != T.end(); t2++)
	    {
	       IDTYPE t1ID = t1->first;
	       IDTYPE t2ID = t2->first;
	       if(isInRel(t1ID, t2ID, IS_TRANSITION))   // is t1 ~ t2?
	       {
	          /*
		   * Def. 7, (4): t1~t2
		   * => for all e in (E_l U E_r): mu(t1)(e) ~ mu(t2)(e)
		   */
	          HYPERGRAPH* lefthandside;
		  lefthandside = petrigraph->getAssociatedRule(t1ID)->getLeft();
		  map<IDTYPE, EDGE> e_lefthandside = lefthandside->getEdges();
		  for( map<IDTYPE, EDGE>::iterator lhsit = e_lefthandside.begin();
		       lhsit != e_lefthandside.end();
		       lhsit++ )
		  {
		     IDTYPE id1 = petrigraph->getMu(t1ID).getMatch(lhsit->first);
		     IDTYPE id2 = petrigraph->getMu(t2ID).getMatch(lhsit->first);
		     if(!this->isInRel(id1, id2, IS_EDGE))
		     {
		        isClosedEquivRelation = false;
			this->addRel(id1, id2, IS_EDGE);
		     }
		  }
	          HYPERGRAPH* righthandside;
		  righthandside = petrigraph->getAssociatedRule(t1ID)->getRight();
		  map<IDTYPE, EDGE> e_righthandside = righthandside->getEdges();
		  for( map<IDTYPE, EDGE>::iterator rhsit = e_righthandside.begin();
		       rhsit != e_righthandside.end();
		       rhsit++ )
		  {
		     IDTYPE id1 = petrigraph->getMu(t1ID).getMatch(rhsit->first);
		     IDTYPE id2 = petrigraph->getMu(t2ID).getMatch(rhsit->first);
		     if(!this->isInRel(id1, id2, IS_EDGE))
		     {
		        isClosedEquivRelation = false;
			this->addRel(id1, id2, IS_EDGE);
		     }
		  }
	       }
	       else if(false)
	       {
	          /*
		   * Def. 7, (3):
		   * (p_N(t1) == pN(t2)) && for all e in E_L: mu(t1)(e) ~ mu(t2)(e)  =>  t1 ~ t2
		   */
		  if( this->petrigraph->getAssociatedRule(t1ID)->getID()
		      == this->petrigraph->getAssociatedRule(t2ID)->getID() )
		  {
		     /*
		      * E_L
		      */
		     map<IDTYPE, EDGE> edges = this->petrigraph->getAssociatedRule(t1ID)->getLeft()->getEdges();
		     bool ok = true;
		     for(map<IDTYPE, EDGE>::iterator i = edges.begin(); i != edges.end(); i++)
		     {
		        if( petrigraph->getMu(t1ID).getMatch(i->first)
			    != petrigraph->getMu(t2ID).getMatch(i->first) )
			{
			   ok = false;
			   break;
			}
		     }
		     if(ok)
		     {
		        this->addRel(t1ID, t2ID, IS_TRANSITION);
			isClosedEquivRelation = false;
		     }
		  }  // if p_N(t1) == p_N(t2)
	       }  // else (t1 !~ t2)
	    }  // for each relation between t1 and some t2
	 }  // for each transition t1
	 
	 /*
	  * Def. 7, (5):
	  * ((e1 ~ e2) && (c_g(e1) == v1 ... vm) && (c_g(e2) == w1 ... wm))  =>  vi ~ wi
	  */
	 for( map< IDTYPE, set<IDTYPE> >::iterator erel_it = this->rel[IS_EDGE].begin();
	      erel_it != this->rel[IS_EDGE].end();
	      erel_it++ )
	 {
	    IDTYPE e1 = erel_it -> first;
	    list<IDTYPE> v = this->petrigraph->getConnectedVertices(e1);
	    
	    for( set<IDTYPE>::iterator e2_it = erel_it->second.begin();
	         e2_it != erel_it->second.end();
		 e2_it++ )
            {
	       IDTYPE e2 = *e2_it;
	       /*
	        * now it holds that e1 ~ e2
		*/
	       list<IDTYPE> w = this->petrigraph->getConnectedVertices(e2);
	       if(v.size() == w.size())  // same arity?
	       {
	          list<IDTYPE>::iterator v_it = v.begin();
		  list<IDTYPE>::iterator w_it = w.begin();
		  while(v_it != v.end())
		  {
		     if(!this->isInRel(*v_it, *w_it, IS_VERTEX))
		     {
		        isClosedEquivRelation = false;
			this->addRel(*v_it, *w_it, IS_VERTEX);
		     }
		     v_it++;
		     w_it++;
		  }
	       }
	    }  // for each edge relation
	 }  // for each edge
	 
	 /*
	  * now ensure the relation is an equivalence relation
	  */
	 GRAPH_ELEMENT_TYPE which[3];
	 which[0] = IS_VERTEX; which[1] = IS_EDGE; which[2] = IS_TRANSITION;
	 for(unsigned char digamma = 0; digamma < 3; digamma++)
	 {
	    for( map< IDTYPE, set<IDTYPE> >::iterator i = this->rel[which[digamma]].begin();
	         i != this->rel[which[digamma]].end();
	         i++ )
	    {
	       /*
	        * v1 ~ v1
	        */
	       if(!this->isInRel(i->first, i->first, which[digamma]))
	       {
	          isClosedEquivRelation = false;
		  this->addRel(i->first, i->first, which[digamma]);
	       }
	    
	       for(set<IDTYPE>::iterator j = i->second.begin(); j != i->second.end(); j++)
	       {
	          /*
	           * v1 ~ v2  =>  v2 ~ v1
		   */
	          map< IDTYPE, set<IDTYPE> > :: iterator j2 = rel[which[digamma]].find(*j);
	          if(j2 == this->rel[which[digamma]].end())
	          { 
		     rel[which[digamma]][*j] = set<IDTYPE>();
	          }
	          if(!this->isInRel(*j, i->first, which[digamma]))
	          {
	             isClosedEquivRelation = false;
		     this->addRel(*j, i->first, which[digamma]);
		     j2 = this->rel[which[digamma]].find(*j);
	          }
		  
		  /*
		   * (v1 ~ v2) && (v2 ~ v3)  =>  (v1 ~ v3)
		   */
		  for(set<IDTYPE>::iterator k = j2->second.begin(); k != j2->second.end(); k++)
		  {
		     if(!this->isInRel(i->first, *k, which[digamma]))
		     {
		        isClosedEquivRelation = false;
			this->addRel(i->first, *k, which[digamma]);
		     }
		  }
  	       }
	    }
	 }
      }  // while(!isClosedEquivRelation)
      if(debug) cout << "\n";
      
      /*
       * create map from old IDs to equivalence class IDs
       */
      map<IDTYPE, IDTYPE> result;
      IDTYPE lastUsedID = this->petrigraph->getLastUsedID();
      IDTYPE currentID;
      for(map< IDTYPE, set<IDTYPE> > :: iterator i = rel[IS_VERTEX].begin(); i != rel[IS_VERTEX].end(); i++)
      {
         if(result.count(i->first) == 0)  // no ID for this vertex yet
	 {
	    lastUsedID++;
	    
	    /*
	     * calculate also min depth
	     */
	    unsigned min_depth = (unsigned)(-1);
	    unsigned current_depth;
	    for(set<IDTYPE>::iterator j = i->second.begin(); j != i->second.end(); j++)
	    {
	       result[*j] = lastUsedID;
	       current_depth = this->petrigraph->getDepth(*j);
	       if(current_depth < min_depth) min_depth = current_depth;
	    }
	    this->petrigraph->setDepth(lastUsedID, min_depth);
	 }
      }
      for(map< IDTYPE, set<IDTYPE> > :: iterator i = rel[IS_EDGE].begin(); i != rel[IS_EDGE].end(); i++)
      {
         if(result.count(i->first) == 0)  // no ID for this edge yet
	 {
	    lastUsedID++;
	    
	    /*
	     * calculate also min depth
	     */
	    unsigned min_depth = (unsigned)(-1);
	    unsigned current_depth;
	    for(set<IDTYPE>::iterator j = i->second.begin(); j != i->second.end(); j++)
	    {
	       result[*j] = lastUsedID;
	       current_depth = this->petrigraph->getDepth(*j);
	       if(current_depth < min_depth) min_depth = current_depth;
	    }
	    this->petrigraph->setDepth(lastUsedID, min_depth);
	 }
      }
      if(debug)  // imitate AUGUR1 behaviour
      {
         lastUsedID += rel[IS_TRANSITION].size();
      }
      for( map< IDTYPE, set<IDTYPE> > :: iterator i = rel[IS_TRANSITION].begin();
           i != rel[IS_TRANSITION].end();
	   i++ )
      {
         if(result.count(i->first) == 0)  // no ID for this transition yet
	 {
	    lastUsedID++;
    	    for(set<IDTYPE>::iterator j = i->second.begin(); j != i->second.end(); j++)
	    {
               result[*j] = lastUsedID;
	    }
         }
      }
      this->petrigraph->setLastUsedID(lastUsedID);
      
      return result;
   }
};
using namespace data_structures;

namespace algorithms
{
   namespace petrigraph
   {     
      MERGE_PETRIGRAPH::MERGE_PETRIGRAPH()
      {
         this->ID = ALGORITHM_MERGE_PETRIGRAPH_UF;
         this->name = "merge_petrigraph";	 
      }
      
      MERGE_PETRIGRAPH::~MERGE_PETRIGRAPH() {}
      
      void MERGE_PETRIGRAPH::initialize(PETRIGRAPH* pg)
      {
         this->petrigraph = pg;	 
	  
	 // trick
	 // if the coverability algorithm is reusable,
	 // then we want to use it incrementally
	 // it means update the coverability algorithm after merging
	 ALGORITHM* cov = this->load("ask_coverability");
	 bool incremental_cov = cov->getReusable();
	 if(incremental_cov)
	 {
	  list<ATYPES> atypes;
	  atypes.push_back(MERGE);
	  VALIDATOR::registerValidated((VALIDATED*)cov, atypes);
	 } 
	 
	 // the same trick for abstraction refinement
	 ALGORITHM* smh = this->load("ask_smh_abstref");
	 bool abstref = smh->getReusable();
	 if(abstref)
	 {
	  list<ATYPES> atypes;
	  atypes.push_back(MERGE);
	  VALIDATOR::registerValidated((VALIDATED*)smh, atypes);
	 } 
      }
      
      void MERGE_PETRIGRAPH::run(MERGE_RELATION* r)
      {
         map<IDTYPE, VERTEX> V = this->petrigraph->getVertices();
	 map<IDTYPE, VERTEX>::iterator vit;
         map<IDTYPE, EDGE> E = this->petrigraph->getEdges();
	 map<IDTYPE, EDGE>::iterator eit;
         map<IDTYPE, TRANSITION> T = this->petrigraph->getTransitions();
	 
         map<IDTYPE, VERTEX> nV;  // new vertex set
	 map<IDTYPE, EDGE> nE;  // new edge set
	 map<IDTYPE, TRANSITION> nT;  // new transition set
	 map< IDTYPE, list<IDTYPE> > ncV;   // new E -> V* connection function
	 map< IDTYPE, multiset<IDTYPE> > ncE;  // new V -> N^E connection function
	 MARKING nInitialMarking;
	 MARKING nCurrentMarking;
	 map<IDTYPE, MATCH> nMu;  // new mu mapping
	 MATCH nIota;  // new iota mapping
	 map<IDTYPE, unsigned> nDepth;  // new depth mapping
	 map< IDTYPE, vector<TCONNECTION> > nPresets;
	 map< IDTYPE, vector<TCONNECTION> > nPostsets;
	 map<IDTYPE, IDTYPE> n_associated_rules;
	 
	 /*
	  * get a mapping from the current petrigraph
	  * to the new equivalence class IDs
	  */
	 map<IDTYPE, IDTYPE> equivrel = r->getExtendingEquivRelation();

	 IDTYPE nLUI = this->petrigraph->getLastUsedID();
	 for(map<IDTYPE, IDTYPE>::iterator i = equivrel.begin(); i != equivrel.end(); i++)
	 {
	    /*
	     * initialize new vertices
	     */
	    if((nV.count(i->second) == 0) && (V.count(i->first) == 1))
	    {
		nDepth[i->second] = this->petrigraph->getDepth(i->second);
		nV[i->second] = V[i->first];
		nV[i->second].setID(i->second);
		if(ncE.count(i->second) == 0) ncE[i->second] = multiset<IDTYPE>();
		nPresets[i->second] = vector<TCONNECTION>();
		nPostsets[i->second] = vector<TCONNECTION>();
		if(i->second > nLUI) nLUI = i->second;
	    }

	    /*
	     * initialize new edges
	     */
	    if((nE.count(i->second) == 0) && (E.count(i->first) == 1))
	    {
	       nDepth[i->second] = this->petrigraph->getDepth(i->second);
	       nE[i->second] = E[i->first];
	       nE[i->second].setID(i->second);
	       
	       list<IDTYPE> connect = this->petrigraph->getConnectedVertices(i->first);
	       list<IDTYPE>::iterator connectit;
	       list<IDTYPE> nConnect;
	       for(connectit = connect.begin(); connectit != connect.end(); connectit++)
	       {
	          IDTYPE tV = equivrel[*connectit];
	          nConnect.push_back(tV);
		  if(ncE.count(tV) == 0) ncE[tV] = multiset<IDTYPE>();
	          ncE[tV].insert(i->second);
	       }
	       ncV[i->second] = nConnect;
	       
	       nPresets[i->second] = vector<TCONNECTION>();
	       nPostsets[i->second] = vector<TCONNECTION>();
	       if(i->second > nLUI) nLUI = i->second;
	    }
 
	    /*
	     * initialize new transitions
	     */
	    if((nT.count(i->second) == 0) && (T.count(i->first) == 1))
	    {
	       nDepth[i->second] = this->petrigraph->getDepth(i->second);
	       nT[i->second] = T[i->first];
	       nT[i->second].setID(i->second);
	       n_associated_rules[i->second] = this->petrigraph->getAssociatedRuleID(i->first);
	    
	       IDTYPE repr = i->first;
	       /*
	        * calculate preset, postset and weights of the new transition.
	        * this is a multiset join where the "weight" refers to the multiplicity
	        * of the set items in the multiset.
	        */
	       vector<TCONNECTION>::iterator connectit;
	       vector<TCONNECTION> tPreset = this->petrigraph->getPreset(repr);
	       vector<TCONNECTION> tPostset = this->petrigraph->getPostset(repr);
	       map<IDTYPE, vector<list<EXPRESSION> > >::iterator expressit;
	       map<IDTYPE, vector<list<EXPRESSION> > > preConnect;
	       map<IDTYPE, vector<list<EXPRESSION> > > postConnect;
	       map<IDTYPE, unsigned> preWeight;
	       map<IDTYPE, unsigned> postWeight;
	       map<IDTYPE, vector<unsigned> > preWeight_Col;
	       map<IDTYPE, vector<unsigned> > postWeight_Col;
	       vector<TCONNECTION> nPreset;
	       vector<TCONNECTION> nPostset;
	       for(connectit = tPreset.begin(); connectit != tPreset.end(); connectit++)
	       {
	          GRAPH_ELEMENT_TYPE which = (E.count(connectit->ID) > 0)? IS_EDGE: IS_VERTEX;
	          IDTYPE target = equivrel[connectit->ID];
	          if(preConnect.count(target) == 0)
	          {
		     vector<list<EXPRESSION> > vle;
		     preConnect[target] = vle;

		     if(coloured)
		     {
			vector<unsigned> lu;
			preWeight_Col[target] = lu;
			preWeight_Col[target].push_back(connectit->weight);
		     }
		     else
			preWeight[target] = connectit->weight;
	          }
	          else
		  { 
			if(coloured)
			 preWeight_Col[target].push_back(connectit->weight);
			else
			 preWeight[target] += connectit->weight;
		  }

		  preConnect[target].push_back(connectit->attributes);
	       }
	       for(connectit = tPostset.begin(); connectit != tPostset.end(); connectit++)
	       {
	          GRAPH_ELEMENT_TYPE which = (E.count(connectit->ID) > 0)? IS_EDGE: IS_VERTEX;
	          IDTYPE target = equivrel[connectit->ID];
	          if(postConnect.count(target) == 0)
	          {
		     vector<list<EXPRESSION> > vle;
		     postConnect[target] = vle;

		     if(coloured)
		     {
			vector<unsigned> lu;
			postWeight_Col[target] = lu;
			postWeight_Col[target].push_back(connectit->weight);
		     }
		     else
			postWeight[target] = connectit->weight;
	          }
	          else
		  {
		     if(coloured)
			postWeight_Col[target].push_back(connectit->weight);
		     else
		     	postWeight[target] += connectit->weight;
		  }

		  postConnect[target].push_back(connectit->attributes);
	       }
	       for(expressit = preConnect.begin(); expressit != preConnect.end(); expressit++)
	       {
		  if(coloured)
		  {
			for(int j=0; j<expressit->second.size(); j++)
			{
				TCONNECTION tc;
				tc.ID = expressit->first;
				tc.weight = preWeight_Col[tc.ID][j];
				tc.attributes = expressit->second[j];
				nPreset.push_back(tc);
		
				TCONNECTION dual;
				dual.ID = i->second;
				dual.weight = tc.weight;
				dual.attributes = expressit->second[j];
				nPostsets[expressit->first].push_back(dual);
			}
		  }
		  else
		  {
			TCONNECTION tc;
			tc.ID = expressit->first;
			tc.weight = preWeight[tc.ID];
			tc.attributes = *(new list<EXPRESSION>());
			nPreset.push_back(tc);
	
			TCONNECTION dual;
			dual.ID = i->second;
			dual.weight = tc.weight;
			dual.attributes = *(new list<EXPRESSION>());
			nPostsets[expressit->first].push_back(dual);
		  }
	       }
	       for(expressit = postConnect.begin(); expressit != postConnect.end(); expressit++)
	       {
		  if(coloured)
		  {
			for(int j=0; j<expressit->second.size(); j++)
			{
				TCONNECTION tc;
				tc.ID = expressit->first;
				tc.weight = postWeight_Col[tc.ID][j];
				tc.attributes = expressit->second[j];
				nPostset.push_back(tc);
			
				TCONNECTION dual;
				dual.ID = i->second;
				dual.weight = tc.weight;
				dual.attributes = expressit->second[j];
				nPresets[expressit->first].push_back(dual);
			}
		  }
		  else
		  {
			TCONNECTION tc;
			tc.ID = expressit->first;
			tc.weight = postWeight[tc.ID];
			tc.attributes = *(new list<EXPRESSION>());
			nPostset.push_back(tc);
		
			TCONNECTION dual;
			dual.ID = i->second;
			dual.weight = tc.weight;
			dual.attributes = *(new list<EXPRESSION>());
			nPresets[expressit->first].push_back(dual);
		  }
	       }
	       nPresets[i->second] = nPreset;
	       nPostsets[i->second] = nPostset;

	       /*
	        * calculate new mu(i->second)(x):
		*/
	       map<IDTYPE, IDTYPE> muT = this->petrigraph->getMu(repr).getMatch();
	       map<IDTYPE, IDTYPE>::iterator mutit;
	       MATCH munT;
	       for(mutit = muT.begin(); mutit != muT.end(); mutit++)
	       {
	          munT.setMatch(mutit->first, equivrel[mutit->second]);
	       }
	       nMu[i->second] = munT;
	    }
	 }
	  
	 /*
	  * calculate new iota mapping and initial marking
	  */
	 map<IDTYPE, IDTYPE> ioto = this->petrigraph->getIota().getMatch();
	 map<IDTYPE, IDTYPE>::iterator iotit;
	 list<IDTYPE> setInitial;
	 for(iotit = ioto.begin(); iotit != ioto.end(); iotit++)
	 {
	    nIota.setMatch(iotit->first, equivrel[iotit->second]);

	    if(!nodes_as_places)
	     if(nE.count(equivrel[iotit->second])<1)
	       continue;

	    TOKEN t;
	    t.setCount(1);
		
	    if(coloured)
	    {
		ALGORITHM * engine = storage.getEngine();
		list<EXPRESSION> le = this->petrigraph->getGrammar()->getStartgraph()->getEdge(iotit->first).getAttributes();
		for(list<EXPRESSION>::iterator ita = le.begin(); ita != le.end(); ita++)
			t.addAttributes(engine->run(*ita));
	    }

	    nInitialMarking.putToken(equivrel[iotit->second], t);
	 }
	 
	 /*
	  * set current marking to initial marking
	  */ 
	 nCurrentMarking = nInitialMarking;
	 
	 /*
	  * now replace the old Petri graph by the new one
	  */
	 this->petrigraph->setAllValues(nIota, nMu,
	  n_associated_rules, ncV, ncE, nT, nPresets, nPostsets,
	  nInitialMarking, nCurrentMarking, nDepth, nV, nE, nLUI, this->petrigraph->getGrammar());
	 
	 // update IDs only if the Petrigraph is real
	 if(this->petrigraph->IsReal)
	 {
		vector<EVENT> to_send;
		EVENT e;
		e.type = MERGE;
		e.ids = equivrel;
		to_send.push_back(e);
		VALIDATOR::validate(to_send);		
	 } 
      }
      
      void MERGE_PETRIGRAPH::run( map<IDTYPE, IDTYPE> vunion,
                                  map<IDTYPE, IDTYPE> eunion, map<IDTYPE, IDTYPE> tunion )
      {
         map<IDTYPE, IDTYPE>::iterator unionit;
         MERGE_RELATION* r = new MERGE_RELATION(this->petrigraph);
	 for(unionit = vunion.begin(); unionit != vunion.end(); unionit++)
	 {
	    r->addRel(unionit->first, unionit->second, IS_VERTEX);
	 }
	 for(unionit = eunion.begin(); unionit != eunion.end(); unionit++)
	 {
	    r->addRel(unionit->first, unionit->second, IS_EDGE);
	 }
	 for(unionit = tunion.begin(); unionit != tunion.end(); unionit++)
	 {
	    r->addRel(unionit->first, unionit->second, IS_TRANSITION);
	 }
	 this->run(r);
      }
      void MERGE_PETRIGRAPH::run( set< pair<IDTYPE, IDTYPE> > vunion,
                                  set< pair<IDTYPE, IDTYPE> > eunion,
			          set< pair<IDTYPE, IDTYPE> > tunion )
      {
         set< pair<IDTYPE, IDTYPE> >::iterator unionit;
         MERGE_RELATION* r = new MERGE_RELATION(this->petrigraph);
	 for(unionit = vunion.begin(); unionit != vunion.end(); unionit++)
	 {
	    r->addRel(unionit->first, unionit->second, IS_VERTEX);
	 }
	 for(unionit = eunion.begin(); unionit != eunion.end(); unionit++)
	 {
	    r->addRel(unionit->first, unionit->second, IS_EDGE);
	 }
	 for(unionit = tunion.begin(); unionit != tunion.end(); unionit++)
	 {
	    r->addRel(unionit->first, unionit->second, IS_TRANSITION);
	 }
	 this->run(r);      
      }      
   };
};
using namespace algorithms;
using namespace algorithms::petrigraph;

#endif
