/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#ifndef ALGORITHM_MERGE_PETRIGRAPH_H
#define ALGORITHM_MERGE_PETRIGRAPH_H

#include "algorithm.h"
#include "globals.h"
#include "petrigraph.h"
#include "validator.h"

namespace data_structures
{
   class MERGE_RELATION
   {
    public:
      MERGE_RELATION(PETRIGRAPH* pg);
      ~MERGE_RELATION();
      
      void addRel(IDTYPE x, IDTYPE y, GRAPH_ELEMENT_TYPE which);
      bool isInRel(IDTYPE x, IDTYPE y, GRAPH_ELEMENT_TYPE which);
      
      map<IDTYPE, IDTYPE> getExtendingEquivRelation();
    
    protected:
      PETRIGRAPH* petrigraph;
      map< GRAPH_ELEMENT_TYPE, map< IDTYPE, set<IDTYPE> > > rel;
   };
};
using namespace data_structures;

namespace algorithms
{
   namespace petrigraph
   {
      class MERGE_PETRIGRAPH: public ALGORITHM 
      {
       public:
              
         MERGE_PETRIGRAPH();
	 virtual ~MERGE_PETRIGRAPH();
       
         virtual void initialize(PETRIGRAPH* pg);
	 virtual void run(MERGE_RELATION* r);
	 virtual void run( map<IDTYPE, IDTYPE> vunion,
	                   map<IDTYPE, IDTYPE> eunion, map<IDTYPE, IDTYPE> tunion );
	 virtual void run( set< pair<IDTYPE, IDTYPE> > vunion,
	                   set< pair<IDTYPE, IDTYPE> > eunion,
			   set< pair<IDTYPE, IDTYPE> > tunion );
			   
       protected:
         PETRIGRAPH* petrigraph;
	 
	 // incremental coverability
	 bool incremental_cov;
      };
   };
};
using namespace algorithms;
using namespace algorithms::petrigraph;

#endif
