/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef ALGORITHM_MERGE_PETRIGRAPH_UF_CPP
#define ALGORITHM_MERGE_PETRIGRAPH_UF_CPP

#include "merge_petrigraph_uf.h"
#include "algorithm_storage.h"

namespace algorithms { namespace petrigraph {

   MERGE_PETRIGRAPH_UF::PARTITION::PARTITION()
   {
      this->unused_name = 1;
   }
   MERGE_PETRIGRAPH_UF::PARTITION::~PARTITION() {}

   IDTYPE MERGE_PETRIGRAPH_UF::PARTITION::insert(IDTYPE elem, IDTYPE tname)
   {
      IDTYPE assigned_name;
      /*
       * condition: name does not yet exist || there are no elements with this name
       */
      if((this->root.count(tname) == 0) || (this->count[root[tname]] == 0))
      {
         assigned_name = tname;
      }
      else
      {
         assigned_name = this->unused_name;
         this->unused_name++;
         while( (this->root.count(unused_name) > 0)
                 && (this->unused_name != 0) ) this->unused_name++;
         if(this->unused_name == 0)
         {
            cerr << "warning: PARTITION running out of equivalence class IDs.\n";
	    this->unused_name--;
         }
      }
      
      this->el.insert(elem);
      this->count[elem] = 1;
      this->name[elem] = assigned_name;
      this->root[assigned_name] = elem;
      this->succ[elem] = set<IDTYPE>();
      
      return assigned_name;
   }
   IDTYPE MERGE_PETRIGRAPH_UF::PARTITION::add(IDTYPE elem)
   {
      if(this->el.count(elem) > 0) return this->find(elem);
      else return this->insert(elem, elem);
   }
   void MERGE_PETRIGRAPH_UF::PARTITION::add(IDTYPE elem, IDTYPE tname)
   {
      if(this->el.count(elem) == 0)
      {
         IDTYPE assigned_name = this->insert(elem, tname);
         if(assigned_name != tname) this->mk_union(tname, assigned_name, tname);
      }
      else cerr << "warning: " << elem << " is already an element of this PARTITION.\n";
   }

   IDTYPE MERGE_PETRIGRAPH_UF::PARTITION::find(IDTYPE elem)
   {
      IDTYPE u = elem;
      while(this->pred.count(u) > 0) u = this->pred[u];
      IDTYPE root_element = u;
      
      IDTYPE v;
      u = elem;
      while(u != root_element)
      {
         v = this->pred[u];
	 this->succ[v].erase(u);
         this->pred[u] = root_element;
	 this->succ[root_element].insert(u);
         u = v;
      }
   
      return this->name[root_element];
   }

   /*
    * merge equivalence classes i and j to a new class called k
    */
   void MERGE_PETRIGRAPH_UF::PARTITION::mk_union(IDTYPE i, IDTYPE j, IDTYPE k)
   {
      if(debug) cout << "mk_union([" << i << ", " << j << "] -> " << k << ")\n";
      if(this->count[root[i]] > this->count[root[j]])
      {
         this->mk_union(j, i, k);
      }
      else
      {
         IDTYPE large = this->root[j];
         if(i != j)
         {
            IDTYPE small = this->root[i];
	    if(false)
	    {
	       cout << "merging partition subtrees:\n";
	       cout << "large is " << large << " and has " << this->count[large] << " elements.\n";
	       cout << "small is " << small << " and has " << this->count[small] << " elements.\n";
	    }
            this->pred[small] = large;
	    this->succ[large].insert(small);
            this->count[large] += this->count[small];
            this->count.erase(small);
            this->name.erase(small);
         }
         this->name[large] = k;
         this->root[k] = large;
	 if(!quiet)
	 {
	    if(this->count[large] != this->getPartition(k).size())
	    {
	       cout << "an __INCONSISTENCY__ was detected by void MERGE_PETRIGRAPH_UF::PARTITION::mk_union(...) - the results will almost certainly be wrong.\n";
	       if(debug)
	       {
	          cout << "new common name is " << large << ", it combines " << this->count[large] << " elements.\n";
	          set<IDTYPE> tpt = this->getPartition(k);
	          cout << "the corresponding section has " << tpt.size() << " elements:";
	          for(set<IDTYPE>::iterator tptit = tpt.begin(); tptit != tpt.end(); tptit++) cout << " " << *tptit;
	          cout << "\n";
	          exit(-1);
	       }
	       this->count[large] = this->getPartition(k).size();
	    }
         }
         
         if(i != k)
         {
            this->root.erase(i);
            if(this->unused_name > i) this->unused_name = i;
         }
         if(j != k)
         {
            this->root.erase(j);
            if(this->unused_name > j) this->unused_name = j;
         }
      }
   }
   void MERGE_PETRIGRAPH_UF::PARTITION::mk_union(IDTYPE i, IDTYPE j)
   {
      if(i != j) this->mk_union(i, j, i);
   }
   void MERGE_PETRIGRAPH_UF::PARTITION::mk_union(IDTYPE i, IDTYPE j, IDTYPE* k)
   {
      if(k != NULL) *k = i;
      if(i != j) this->mk_union(i, j, i);
   }
   
   IDTYPE MERGE_PETRIGRAPH_UF::PARTITION::getAnyElement(IDTYPE tname)
   {
      return this->root[tname];
   }
   
   set<IDTYPE> MERGE_PETRIGRAPH_UF::PARTITION::getNames()
   {
      set<IDTYPE> retv;
      for(map<IDTYPE, IDTYPE>::iterator rtit = root.begin(); rtit != root.end(); rtit++)
      {
         retv.insert(rtit->first);
      }
      return retv;
   }
   
   void MERGE_PETRIGRAPH_UF::PARTITION::exportSubtree(IDTYPE nd, IDTYPE rt, set<IDTYPE>* tpt)
   {
      set<IDTYPE>::iterator levit;
      for(levit = this->succ[nd].begin(); levit != this->succ[nd].end(); levit++)
      {
         this->exportSubtree(*levit, rt, tpt);
      }
      if(nd != rt)
      {
         if(pred[nd] != rt)
	 {
            this->pred[nd] = rt;
            this->succ[rt].insert(nd);
	 } 
         if(tpt != NULL) tpt->insert(nd);
	 this->succ[nd] = set<IDTYPE>();
      }
   }
   void MERGE_PETRIGRAPH_UF::PARTITION::exportSubtree(IDTYPE nd, IDTYPE rt)
   {
      this->exportSubtree(nd, rt, NULL);
   }
   set<IDTYPE> MERGE_PETRIGRAPH_UF::PARTITION::getPartition(IDTYPE name)
   {
      set<IDTYPE>* tpt = new set<IDTYPE>();
      tpt->insert(this->root[name]);
      this->exportSubtree(this->root[name], this->root[name], tpt);
      return *tpt;
   }
}; };
  
namespace algorithms
{
   namespace petrigraph
   {     
      MERGE_PETRIGRAPH_UF::MERGE_PETRIGRAPH_UF()
      {
         this->ID = ALGORITHM_MERGE_PETRIGRAPH_UF;
         this->name = "merge_petrigraph_uf";
	 this->petrigraph = new PETRIGRAPH;
      }   
      MERGE_PETRIGRAPH_UF::~MERGE_PETRIGRAPH_UF() {}
      
      void MERGE_PETRIGRAPH_UF::initialize(PETRIGRAPH* pg)
      {
         this->petrigraph = pg;
	 
	 // trick
	 // if the coverability algorithm is reusable,
	 // then we want to use it incrementally
	 // it means update the coverability algorithm after merging
	 ALGORITHM* cov = this->load("ask_coverability");
	 incremental_cov = cov->getReusable();
	 if(incremental_cov)
	 {
	  list<ATYPES> atypes;
	  // atypes.push_back(MERGE_PLACES);
	  // atypes.push_back(MERGE_TRANSITIONS);
	  VALIDATOR::registerValidated((VALIDATED *)cov, atypes);
	 } 
      }
      
      void MERGE_PETRIGRAPH_UF::run(RELATION* r)
      {
         map<IDTYPE, VERTEX> V = this->petrigraph->getVertices();
	 map<IDTYPE, VERTEX>::iterator vit;
         map<IDTYPE, EDGE> E = this->petrigraph->getEdges();
	 map<IDTYPE, EDGE>::iterator eit;
         map<IDTYPE, TRANSITION> T = this->petrigraph->getTransitions();
	 
         map<IDTYPE, VERTEX> nV;  // new vertex set
	 map<IDTYPE, EDGE> nE;  // new edge set
	 map<IDTYPE, TRANSITION> nT;  // new transition set
	 map< IDTYPE, list<IDTYPE> > ncV;   // new E -> V* connection function
	 map< IDTYPE, multiset<IDTYPE> > ncE;  // new V -> N^E connection function
	 MARKING nInitialMarking;
	 MARKING nCurrentMarking;
	 map<IDTYPE, MATCH> nMu;  // new mu mapping
	 MATCH nIota;  // new iota mapping
	 map<IDTYPE, unsigned> nDepth;  // new depth mapping
	 map< IDTYPE, vector<TCONNECTION> > nPresets;
	 map< IDTYPE, vector<TCONNECTION> > nPostsets;
	 map<IDTYPE, IDTYPE> n_associated_rules;
	 
	 IDTYPE nLUI = 0;
	 
	 /*
	  * construct the least closed equivalence relation that includes *r
	  */
	 if(false)
	 {
	    cout << "relation before extending\n" << *r;
	 }
	 r->extend();
	 if(debug)
	 {
	    cout << "relation �tir extending\n" << *r;
	 }
	 
	 /*
	  * calculate the merged current marking
	  * note that this section deviates from AUGUR I in assigning tokens to vertices
	  */
	 for(vit = V.begin(); vit != V.end(); vit++)
	 {
	    IDTYPE nvid = r->find(vit->first, IS_VERTEX);
	    vector<TOKEN> tokens = this->petrigraph->getCurrentMarking().getMarking(vit->first);
	    vector<TOKEN>::iterator tokit;
	    for(tokit = tokens.begin(); tokit != tokens.end(); tokit++)
	    {
	       nCurrentMarking.putToken(nvid, *tokit);
	    }
	 }
	 for(eit = E.begin(); eit != E.end(); eit++)
	 {
	    IDTYPE neid = r->find(eit->first, IS_EDGE);
	    vector<TOKEN> tokens = this->petrigraph->getCurrentMarking().getMarking(eit->first);
	    vector<TOKEN>::iterator tokit;
	    for(tokit = tokens.begin(); tokit != tokens.end(); tokit++)
	    {
	       nCurrentMarking.putToken(neid, *tokit);
	    }
	 }
	  
	 /*
	  * initialize new vertices
	  */
	 set<IDTYPE> nVset = r->getNames(IS_VERTEX);
	 for(set<IDTYPE>::iterator nVit = nVset.begin(); nVit != nVset.end(); nVit++)
	 {
	    nV[*nVit] = V[r->getAnyElement(*nVit, IS_VERTEX)];
	    nV[*nVit].setID(*nVit);
	    nDepth[*nVit] = r->getDepth(*nVit);
	    ncE[*nVit] = multiset<IDTYPE>();
	    if(*nVit > nLUI) nLUI = *nVit;
	    nPresets[*nVit] = vector<TCONNECTION>();
	    nPostsets[*nVit] = vector<TCONNECTION>();
	 }
	 
	 /*
	  * intialize new edges
	  */
	 set<IDTYPE> nEset = r->getNames(IS_EDGE);
	 for(set<IDTYPE>::iterator nEit = nEset.begin(); nEit != nEset.end(); nEit++)
	 {
	    IDTYPE repr = r->getAnyElement(*nEit, IS_EDGE);
	    nE[*nEit] = E[repr];
	    nE[*nEit].setID(*nEit);
	    list<IDTYPE> connect = this->petrigraph->getConnectedVertices(repr);
	    list<IDTYPE>::iterator connectit;
	    list<IDTYPE> nConnect;
	    for(connectit = connect.begin(); connectit != connect.end(); connectit++)
	    {
	       IDTYPE tV = r->find(*connectit, IS_VERTEX);
	       nConnect.push_back(tV);
	       ncE[tV].insert(*nEit);
	    }
	    ncV[*nEit] = nConnect;
	    /*
	     * markings will be updated later
	     */
	    nDepth[*nEit] = r->getDepth(*nEit);
	    
	    if(*nEit > nLUI) nLUI = *nEit;
	    nPresets[*nEit] = vector<TCONNECTION>();
	    nPostsets[*nEit] = vector<TCONNECTION>();
	 }
	 
	 /*
	  * remove isolated nodes
	  */
	 if(remove_isolated_nodes)
	 {
	    for(set<IDTYPE>::iterator nVit = nVset.begin(); nVit != nVset.end(); nVit++)
	    {
   	       if(ncE[*nVit].size() == 0)
	       {
	          nV.erase(*nVit);
	          ncE.erase(*nVit);
	          nDepth.erase(*nVit);
	       }
	    }
	 }
	 
	 /*
	  * initialize new transitions
	  */
	 set<IDTYPE> nTset = r->getNames(IS_TRANSITION);
	 for(set<IDTYPE>::iterator nTit = nTset.begin(); nTit != nTset.end(); nTit++)
	 {
	    IDTYPE repr = r->getAnyElement(*nTit, IS_TRANSITION);
	    TRANSITION newt;
	    newt.setID(*nTit);
	    newt.setIdLabel(T[repr].getIdLabel());
	    newt.setLabel(T[repr].getLabel());
	    newt.setGuard(T[repr].getGuard());
	    nT[*nTit] = newt;
	    
	    /*
	     * calculate preset, postset and weights of the new transition.
	     * this is a multiset join where the "weight" refers to the multiplicity
	     * of the set items in the multiset.
	     */
	    vector<TCONNECTION>::iterator connectit;
	    vector<TCONNECTION> tPreset = this->petrigraph->getPreset(repr);
	    vector<TCONNECTION> tPostset = this->petrigraph->getPostset(repr);
	    map<IDTYPE, vector<list<EXPRESSION> > >::iterator expressit;
	    map<IDTYPE, vector<list<EXPRESSION> > > preConnect;
	    map<IDTYPE, vector<list<EXPRESSION> > > postConnect;
	    map<IDTYPE, unsigned> preWeight;
	    map<IDTYPE, unsigned> postWeight;
	    map<IDTYPE, vector<unsigned> > preWeight_Col;
	    map<IDTYPE, vector<unsigned> > postWeight_Col;
	    vector<TCONNECTION> nPreset;
	    vector<TCONNECTION> nPostset;
	    for(connectit = tPreset.begin(); connectit != tPreset.end(); connectit++)
	    {
	       GRAPH_ELEMENT_TYPE which = (E.count(connectit->ID) > 0)? IS_EDGE: IS_VERTEX;
	       IDTYPE target = r->find(connectit->ID, which);
	       if(preConnect.count(target) == 0)
	       {
		    vector<list<EXPRESSION> > vle;
		    preConnect[target] = vle;

		    if(coloured)
		    {
			vector<unsigned> lu;
			preWeight_Col[target] = lu;
			preWeight_Col[target].push_back(connectit->weight);
		    }
		    else
			preWeight[target] = connectit->weight;
	        }
	        else
		{ 
			if(coloured)
			 preWeight_Col[target].push_back(connectit->weight);
			else
			 preWeight[target] += connectit->weight;
		}

		preConnect[target].push_back(connectit->attributes);
	    }
	    for(connectit = tPostset.begin(); connectit != tPostset.end(); connectit++)
	    {
	       GRAPH_ELEMENT_TYPE which = (E.count(connectit->ID) > 0)? IS_EDGE: IS_VERTEX;
	       IDTYPE target = r->find(connectit->ID, which);
	       if(postConnect.count(target) == 0)
	       {
		     vector<list<EXPRESSION> > vle;
		     postConnect[target] = vle;

		     if(coloured)
		     {
			vector<unsigned> lu;
			postWeight_Col[target] = lu;
			postWeight_Col[target].push_back(connectit->weight);
		     }
		     else
			postWeight[target] = connectit->weight;
	        }
	        else
		{
		     if(coloured)
			postWeight_Col[target].push_back(connectit->weight);
		     else
		     	postWeight[target] += connectit->weight;
		}

		postConnect[target].push_back(connectit->attributes);
	    }

	       for(expressit = preConnect.begin(); expressit != preConnect.end(); expressit++)
	       {
		  if(coloured)
		  {
			for(int j=0; j<expressit->second.size(); j++)
			{
				TCONNECTION tc;
				tc.ID = expressit->first;
				tc.weight = preWeight_Col[tc.ID][j];
				tc.attributes = expressit->second[j];
				nPreset.push_back(tc);
		
				TCONNECTION dual;
				dual.ID = *nTit;
				dual.weight = tc.weight;
				dual.attributes = expressit->second[j];
				nPostsets[expressit->first].push_back(dual);
			}
		  }
		  else
		  {
			TCONNECTION tc;
			tc.ID = expressit->first;
			tc.weight = preWeight[tc.ID];
			tc.attributes = *(new list<EXPRESSION>());
			nPreset.push_back(tc);
	
			TCONNECTION dual;
			dual.ID = *nTit;
			dual.weight = tc.weight;
			dual.attributes = *(new list<EXPRESSION>());
			nPostsets[expressit->first].push_back(dual);
		  }
	       }
	       for(expressit = postConnect.begin(); expressit != postConnect.end(); expressit++)
	       {
		  if(coloured)
		  {
			for(int j=0; j<expressit->second.size(); j++)
			{
				TCONNECTION tc;
				tc.ID = expressit->first;
				tc.weight = postWeight_Col[tc.ID][j];
				tc.attributes = expressit->second[j];
				nPostset.push_back(tc);
			
				TCONNECTION dual;
				dual.ID = *nTit;
				dual.weight = tc.weight;
				dual.attributes = expressit->second[j];
				nPresets[expressit->first].push_back(dual);
			}
		  }
		  else
		  {
			TCONNECTION tc;
			tc.ID = expressit->first;
			tc.weight = postWeight[tc.ID];
			tc.attributes = *(new list<EXPRESSION>());
			nPostset.push_back(tc);
		
			TCONNECTION dual;
			dual.ID = *nTit;
			dual.weight = tc.weight;
			dual.attributes = *(new list<EXPRESSION>());
			nPresets[expressit->first].push_back(dual);
		  }
	       }

	    nPresets[*nTit] = nPreset;
	    nPostsets[*nTit] = nPostset;
	    
	    /*
	     * calculate mu mapping for the new transition
	     */
	    map<IDTYPE, IDTYPE> muT = this->petrigraph->getMu(repr).getMatch();
	    map<IDTYPE, IDTYPE>::iterator mutit;
	    MATCH munT;
	    for(mutit = muT.begin(); mutit != muT.end(); mutit++)
	    {
	       GRAPH_ELEMENT_TYPE which = (E.count(mutit->second) > 0)? IS_EDGE: IS_VERTEX;
	       IDTYPE muval = r->find(mutit->second, which);
	       munT.setMatch(mutit->first, muval);
	    }
	    nMu[*nTit] = munT;
	     
	    nDepth[*nTit] = r->getDepth(*nTit);
	    n_associated_rules[*nTit] = this->petrigraph->getAssociatedRuleID(repr);
	    if(*nTit > nLUI) nLUI = *nTit;
	 }
	 
	 /*
	  * calculate new iota mapping and initial marking
	  */
	 map<IDTYPE, IDTYPE> ioto = this->petrigraph->getIota().getMatch();
	 map<IDTYPE, IDTYPE>::iterator iotit;
	 list<IDTYPE> setInitial;
	 for(iotit = ioto.begin(); iotit != ioto.end(); iotit++)
	 {
	    GRAPH_ELEMENT_TYPE which = (E.count(iotit->second) > 0)? IS_EDGE: IS_VERTEX;
	    IDTYPE ioval = r->find(iotit->second, which);
	    nIota.setMatch(iotit->first, ioval);
	   
	    if(!nodes_as_places)
	      if(nE.count(ioval)<1)
	        continue;

	    TOKEN t;
	    t.setCount(1);
		
	    if(coloured)
	    {
		ALGORITHM * engine = storage.getEngine();
		list<EXPRESSION> le = this->petrigraph->getGrammar()->getStartgraph()->getEdge(iotit->first).getAttributes();
		for(list<EXPRESSION>::iterator ita = le.begin(); ita != le.end(); ita++)
			t.addAttributes(engine->run(*ita));
	    }

	    nInitialMarking.putToken(ioval, t);
	 }
 
	 /*
	  * now replace the old Petri graph by the new one
	  */
	 this->petrigraph->setAllValues(nIota, nMu,
	  n_associated_rules, ncV, ncE, nT, nPresets, nPostsets,
	  nInitialMarking, nCurrentMarking, nDepth, nV, nE, nLUI, this->petrigraph->getGrammar());
	 
	 // incremental
	 if(incremental_cov)
	 {
	  cout << "merge_petrigraph_uf is not implemented with incremental coverability\n";
	  exit(1);
	  /*
	  vector<EVENT> to_send;
	  
	  EVENT e1;
	  e1.type = MERGE_PLACES;
	  e1.ids = equivrel;
	  	  
	  EVENT e2;
	  e2.type = MERGE_TRANSITIONS;
	  e2.ids = equivrel;
	  
	  to_send.push_back(e1);
	  to_send.push_back(e2);
	  
	  VALIDATOR::validate(to_send);
	  */
	 } 
	 
	 // update IDs only if the Petrigraph is real
	 if(this->petrigraph->IsReal)
	 {
	 	map<IDTYPE, IDTYPE> equivrel;
		
		// to do - nodes
		
		for(map<IDTYPE, EDGE>::iterator e = E.begin(); e!=E.end(); e++)
		 equivrel[e->first]= r->find(e->first, IS_EDGE);
		 
		for(map<IDTYPE, TRANSITION>::iterator t = T.begin(); t!=T.end(); t++)
		 equivrel[t->first]= r->find(t->first, IS_TRANSITION);
				
		vector<EVENT> to_send;
		EVENT e;
		e.type = MERGE;
		e.ids = equivrel;
		to_send.push_back(e);
		VALIDATOR::validate(to_send);		
	 } 
      }
      
      void MERGE_PETRIGRAPH_UF::run( map<IDTYPE, IDTYPE> vunion,
                                  map<IDTYPE, IDTYPE> eunion, map<IDTYPE, IDTYPE> tunion )
      {
         this->run(new RELATION(this->petrigraph, vunion, eunion, tunion));
      }
      void MERGE_PETRIGRAPH_UF::run( set< pair<IDTYPE, IDTYPE> > vunion,
                                  set< pair<IDTYPE, IDTYPE> > eunion,
				  set< pair<IDTYPE, IDTYPE> > tunion )
      {
         this->run(new RELATION(this->petrigraph, vunion, eunion, tunion));
      }     
   
      MERGE_PETRIGRAPH_UF::RELATION::RELATION( PETRIGRAPH* pg, map<IDTYPE, IDTYPE> vunion,
                                            map<IDTYPE, IDTYPE> eunion,
					    map<IDTYPE, IDTYPE> tunion )
      {
         this->petrigraph = pg;
         map<IDTYPE, IDTYPE>::iterator mapit;
	 
	 set< pair<IDTYPE, IDTYPE> > vu;
	 for(mapit = vunion.begin(); mapit != vunion.end(); mapit++) vu.insert(*mapit);
	 set< pair<IDTYPE, IDTYPE> > eu;
	 for(mapit = eunion.begin(); mapit != eunion.end(); mapit++) eu.insert(*mapit);
	 set< pair<IDTYPE, IDTYPE> > tu;
	 for(mapit = tunion.begin(); mapit != tunion.end(); mapit++) tu.insert(*mapit);
	 
	 *this = RELATION(pg, vu, eu, tu);
      }
      
      MERGE_PETRIGRAPH_UF::RELATION::RELATION( PETRIGRAPH* pg,
                                            set< pair<IDTYPE, IDTYPE> > vunion,
                                            set< pair<IDTYPE, IDTYPE> > eunion,
				            set< pair<IDTYPE, IDTYPE> > tunion )
      {
         this->petrigraph = pg;
         set< pair<IDTYPE, IDTYPE> >::iterator unionit;
	 
         map<IDTYPE, VERTEX> V = pg->getVertices();
         map<IDTYPE, VERTEX>::iterator Vit;
         for(Vit = V.begin(); Vit != V.end(); Vit++)
	 {
	    this->vrel.add(Vit->first);
	    this->setDepth(Vit->first, pg->getDepth(Vit->first));
	 }
         for(unionit = vunion.begin(); unionit != vunion.end(); unionit++)
	 {
	    this->addVertexRel(unionit->first, unionit->second);
	 }
	 
         map<IDTYPE, EDGE> E = pg->getEdges();
         map<IDTYPE, EDGE>::iterator Eit;
         for(Eit = E.begin(); Eit != E.end(); Eit++)
	 {
	    this->erel.add(Eit->first);
	    this->setDepth(Eit->first, pg->getDepth(Eit->first));
	 }
         for(unionit = eunion.begin(); unionit != eunion.end(); unionit++)
	 {
	    this->addEdgeRel(unionit->first, unionit->second);
	 }
         
         map<IDTYPE, TRANSITION> T = pg->getTransitions();
         map<IDTYPE, TRANSITION>::iterator Tit;
         for(Tit = T.begin(); Tit != T.end(); Tit++)
	 {
	    this->trel.add(Tit->first);
	    this->setDepth(Tit->first, pg->getDepth(Tit->first));
	 }
         for(unionit = tunion.begin(); unionit != tunion.end(); unionit++)
	 {
	    this->addTransitionRel(unionit->first, unionit->second);
	 }
      }

      MERGE_PETRIGRAPH_UF::RELATION::RELATION(PETRIGRAPH* pg)
      {
         map<IDTYPE, IDTYPE> emptymap;
         RELATION(pg, emptymap, emptymap, emptymap);
      }
      
      IDTYPE MERGE_PETRIGRAPH_UF::RELATION::find(IDTYPE elem, GRAPH_ELEMENT_TYPE which)
      {
         switch(which)
	 {
	    case IS_VERTEX:
	       return this->vrel.find(elem); break;
	    case IS_EDGE:
	       return this->erel.find(elem); break;
	    case IS_TRANSITION:
	       return this->trel.find(elem); break;
	    default:
	       return (IDTYPE)(-1);
	 }
      }
      IDTYPE MERGE_PETRIGRAPH_UF::RELATION::getAnyElement(IDTYPE name, GRAPH_ELEMENT_TYPE which)
      {
         switch(which)
	 {
	    case IS_VERTEX:
	       return this->vrel.getAnyElement(name); break;
	    case IS_EDGE:
	       return this->erel.getAnyElement(name); break;
	    case IS_TRANSITION:
	       return this->trel.getAnyElement(name); break;
	    default:
	       return (IDTYPE)(-1);
	 }
      }
      set<IDTYPE> MERGE_PETRIGRAPH_UF::RELATION::getNames(GRAPH_ELEMENT_TYPE which)
      {
         switch(which)
	 {
	    case IS_VERTEX:
	       return this->vrel.getNames(); break;
	    case IS_EDGE:
	       return this->erel.getNames(); break;
	    case IS_TRANSITION:
	       return this->trel.getNames(); break;
	    default:
	       return set<IDTYPE>();
	 }
      }
      
      
      void MERGE_PETRIGRAPH_UF::RELATION::addRel(IDTYPE x, IDTYPE y, GRAPH_ELEMENT_TYPE which)
      {
         switch(which)
	 {
	    case IS_VERTEX:
	       this->addVertexRel(x, y); break;
	    case IS_EDGE:
	       this->addEdgeRel(x, y); break;
	    case IS_TRANSITION:
	       this->addTransitionRel(x, y); break;
	    default:
	       break;
	 }
      }
      void MERGE_PETRIGRAPH_UF::RELATION::addVertexRel(IDTYPE vertex1, IDTYPE vertex2)
      {
         unsigned mindepth = this->petrigraph->getDepth(vertex1);
	 if(mindepth > this->petrigraph->getDepth(vertex2))
	    mindepth = this->petrigraph->getDepth(vertex2);
         IDTYPE* k = new IDTYPE;
         this->vrel.mk_union(this->vrel.find(vertex1), this->vrel.find(vertex2), k);
	 this->setDepth(*k, mindepth);
	 if(debug)
	 {
	    cout << "vertices v" << vertex1 << ":d" << this->petrigraph->getDepth(vertex1);
	    cout << " and v" << vertex2 << ":d" << this->petrigraph->getDepth(vertex2);
	    cout << " were merged to v" << *k << ":d" << mindepth << "\n";
	 }
      }
      void MERGE_PETRIGRAPH_UF::RELATION::addEdgeRel(IDTYPE edge1, IDTYPE edge2)
      {
         unsigned mindepth = this->petrigraph->getDepth(edge1);
	 if(mindepth > this->petrigraph->getDepth(edge2))
	    mindepth = this->petrigraph->getDepth(edge2);
         IDTYPE* k = new IDTYPE;
         this->erel.mk_union(this->erel.find(edge1), this->erel.find(edge2), k);
	 this->setDepth(*k, mindepth);
	 if(debug)
	 {
	    cout << "edges e" << edge1 << ":d" << this->petrigraph->getDepth(edge1);
	    cout << " and e" << edge2 << ":d" << this->petrigraph->getDepth(edge2);
	    cout << " were merged to e" << *k << ":d" << mindepth << "\n";
	 }
      }
      void MERGE_PETRIGRAPH_UF::RELATION::addTransitionRel(IDTYPE t1, IDTYPE t2)
      {
         unsigned mindepth = this->petrigraph->getDepth(t1);
	 if(mindepth > this->petrigraph->getDepth(t2))
	    mindepth = this->petrigraph->getDepth(t2);
         IDTYPE* k = new IDTYPE;
         this->trel.mk_union(this->trel.find(t1), this->trel.find(t2), k);
	 this->setDepth(*k, mindepth);
	 if(debug)
	 {
	    cout << "transitions t" << t1 << ":d" << this->petrigraph->getDepth(t1);
	    cout << " and t" << t2 << ":d" << this->petrigraph->getDepth(t2);
	    cout << " were merged to t" << *k << ":d" << mindepth << "\n";
	 }
      }
      
      bool MERGE_PETRIGRAPH_UF::RELATION::areVerticesInRel(IDTYPE vertex1, IDTYPE vertex2)
      {
         return (this->vrel.find(vertex1) == this->vrel.find(vertex2));
      }
      bool MERGE_PETRIGRAPH_UF::RELATION::areEdgesInRel(IDTYPE edge1, IDTYPE edge2)
      {
         return (this->erel.find(edge1) == this->erel.find(edge2));
      }
      bool MERGE_PETRIGRAPH_UF::RELATION::areTransitionsInRel(IDTYPE t1, IDTYPE t2)
      {
         return (this->trel.find(t1) == this->trel.find(t2));
      }
      bool MERGE_PETRIGRAPH_UF::RELATION::areInRel(IDTYPE x, IDTYPE y, GRAPH_ELEMENT_TYPE which)
      {
         switch(which)
	 {
	    case IS_VERTEX:
	       return this->areVerticesInRel(x, y); break;
	    case IS_EDGE:
	       return this->areEdgesInRel(x, y); break;
	    case IS_TRANSITION:
	       return this->areTransitionsInRel(x, y); break;
	    default:
	       return false;
	 }
      }
      
      /*
       * extend to a closed equivalence relation
       * iterate until no extension could be performed
       */
      void MERGE_PETRIGRAPH_UF::RELATION::extend()
      {
         bool isClosedEquivRelation = false; 
         map<IDTYPE, VERTEX> V = this->petrigraph->getVertices();
         map<IDTYPE, EDGE> E = this->petrigraph->getEdges();
         map<IDTYPE, TRANSITION> T = this->petrigraph->getTransitions();
	 
	 while(!isClosedEquivRelation)
	 {
	    isClosedEquivRelation = true;
	    for(map<IDTYPE, TRANSITION>::iterator i = T.begin(); i != T.end(); i++)
	    {
	       map<IDTYPE, TRANSITION>::iterator j = i;
	       for(j++; j != T.end(); j++)
	       {
	          IDTYPE t1 = i->first;
		  IDTYPE t2 = j->first;
		  MATCH m1 = petrigraph->getMu(t1);
		  MATCH m2 = petrigraph->getMu(t2);
                  RULE* t1rule = this->petrigraph->getAssociatedRule(t1);
                  RULE* t2rule = this->petrigraph->getAssociatedRule(t2);
		  
	          if(this->areTransitionsInRel(t1, t2))
		  {
		     if(!quiet && (t1rule != t2rule))
		     {
		        cout << "RELATION::extend() has issued a warning:\n";
			cout << "transitions t" << t1 << " and t" << t2 << " will be merged, ";
			cout << "although they represent applications\nof different rules ";
			cout << "(" << t1rule->getID() << "[" << t1rule->getName() << "] and ";
			cout << t2rule->getID() << "[" << t2rule->getName() << "], respectively)\n";
		     }
		  
		     /*
		      * Def. 7, (4): for all e in (E_l U E_r): mu(t1)(e) ~ mu(t2)(e)
		      */
		     if(t1rule != NULL)
		     {
		        map<IDTYPE, EDGE> e_lhs = t1rule->getLeft()->getEdges();
			for( map<IDTYPE, EDGE>::iterator lhsit = e_lhs.begin(); 
			     lhsit != e_lhs.end();
			     lhsit++ )
		        {
			   if(m1.hasMatch(lhsit->first) && m2.hasMatch(lhsit->first))
			   {
			      IDTYPE e1 = m1.getMatch(lhsit->first);
			      IDTYPE e2 = m2.getMatch(lhsit->first);
			      if(!this->areEdgesInRel(e1, e2))
			      {
			         isClosedEquivRelation = false;
			         this->addEdgeRel(e1, e2);
			      }
			   }
			}
		        map<IDTYPE, EDGE> e_rhs = t1rule->getRight()->getEdges();
			for( map<IDTYPE, EDGE>::iterator rhsit = e_rhs.begin(); 
			     rhsit != e_rhs.end();
			     rhsit++ )
		        {
			   if(m1.hasMatch(rhsit->first) && m2.hasMatch(rhsit->first))
			   {
			      IDTYPE e1 = m1.getMatch(rhsit->first);
			      IDTYPE e2 = m2.getMatch(rhsit->first);
			      if(!this->areEdgesInRel(e1, e2))
			      {
			         isClosedEquivRelation = false;
			         this->addEdgeRel(e1, e2);
			      }
			   }
			}
		     }
		  }
		  else if (false)
		  {
		     /*
		      * Def. 7, (3):
		      * p_N(t1) == p_N(t2) and forall e in E_L: mu(t1)(e) ~ mu(t2)(e)
		      * implies t1 ~ t2
		      */
		     if(t1rule == NULL) { cout << "severe error.\n"; exit(-1); }
		     if((t1rule != NULL) && (t1rule->getID() == t2rule->getID()))
		     {
		        map<IDTYPE, EDGE> edges = t1rule->getLeft()->getEdges();
			bool ok = true;
			for( map<IDTYPE, EDGE>::iterator edgeit = edges.begin(); 
			     edgeit != edges.end();
			     edgeit++ )
			{
			   IDTYPE e1 = m1.getMatch(edgeit->first);
			   IDTYPE e2 = m2.getMatch(edgeit->first);
			   if(e1 != e2) ok = false;
			}
			
			if(ok)  // then t1 ~ t2
			{
			   isClosedEquivRelation = false;
			   this->addTransitionRel(t1, t2);
			}
		     }
		  }
	       }
	    }
	    
	    /*
	     * Def. 7, (5):
	     * e1 ~ e2 && c_g(e1) == v_1 .. v_m && c_g(e2) == w_1 .. w_m
	     * implies for all i: v_i ~ w_i
	     */
	    set<IDTYPE> eparts = this->erel.getNames();
	    set<IDTYPE>::iterator partit;
	    for(partit = eparts.begin(); partit != eparts.end(); partit++)
	    {
	       set<IDTYPE> similar = this->erel.getPartition(*partit);
	       if(false && (similar.size() > 1) && debug) cout << "loading edge partition " << *partit << " with " << similar.size() << " elements.\n";
	       for(set<IDTYPE>::iterator e1 = similar.begin(); e1 != similar.end(); e1++)
	       {
	          list<IDTYPE> cge1 = this->petrigraph->getConnectedVertices(*e1);
	          set<IDTYPE>::iterator e2 = e1;
		  for(e2++; e2 != similar.end(); e2++)
		  {
		     if(false) cout << "comparing vertices connected to " << *e1 << " and " << *e2 << "\n";
		     list<IDTYPE> cge2 = this->petrigraph->getConnectedVertices(*e2);
		     if(cge1.size() == cge2.size())  // same arity?
		     {
		        list<IDTYPE>::iterator cge1it = cge1.begin();
		        list<IDTYPE>::iterator cge2it = cge2.begin();
			while(cge1it != cge1.end())
			{
			   if(!this->areVerticesInRel(*cge1it, *cge2it))
			   {
			      isClosedEquivRelation = false;
			      this->addVertexRel(*cge1it, *cge2it);
			   }
			   cge1it++;
			   cge2it++;
			}
		     }
		     else cout << "warning: edges with different arity values will be merged.\n";
		  }
	       }
	    }
	 }
      }
      
      unsigned MERGE_PETRIGRAPH_UF::RELATION::getDepth(IDTYPE id)
      {
         if(this->depth.count(id) == 0) this->depth[id] = 0;
         return this->depth[id];
      }
      void MERGE_PETRIGRAPH_UF::RELATION::setDepth(IDTYPE id, unsigned d)
      {
         this->depth[id] = d;
      }
      
      ostream& operator<<(ostream& ost, MERGE_PETRIGRAPH_UF::PARTITION& th)
      {
         set<IDTYPE>::iterator elit;
	 ost << "partition ";
	 for(elit = th.el.begin(); elit != th.el.end(); elit++) ost << *elit << "->" << th.find(*elit) << " ";
	 return ost;
      }
      
      ostream& operator<<(ostream& ost, MERGE_PETRIGRAPH_UF::RELATION& th)
      {
         ost << "vertices: " << th.vrel << "\n";
	 ost << "edges: " << th.erel << "\n";
	 ost << "transitions: " << th.trel << "\n";
         return ost;
      }
   };
};

#endif
