/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#ifndef ALGORITHM_MERGE_PETRIGRAPH_UF_H
#define ALGORITHM_MERGE_PETRIGRAPH_UF_H

#include "algorithm.h"
#include "globals.h"
#include "petrigraph.h"

namespace algorithms
{
   namespace petrigraph
   {
      class MERGE_PETRIGRAPH_UF: public ALGORITHM
      {
       public:
       
         /*
          * a PARTITION is used to represent an equivalence relation
          * over a set of indexed elements.
          */
         class PARTITION
         {
          public:
            PARTITION();
            ~PARTITION();
            
            void add(IDTYPE elem, IDTYPE tname);
            IDTYPE add(IDTYPE elem);
      
            IDTYPE find(IDTYPE id);
            bool equiv(IDTYPE x, IDTYPE y);
      
            void mk_union(IDTYPE i, IDTYPE j, IDTYPE k);
            void mk_union(IDTYPE i, IDTYPE j, IDTYPE* k);
            void mk_union(IDTYPE i, IDTYPE j);
      
            IDTYPE getAnyElement(IDTYPE tname);
            set<IDTYPE> getNames();
            set<IDTYPE> getPartition(IDTYPE tname);
	    
	    friend ostream& operator<<(ostream& ost, PARTITION& th);
    
          protected:
            set<IDTYPE> el;  // lists all elements of the considered set
            map<IDTYPE, IDTYPE> pred;  // arborescence structure for union/find (predecessors)
            map< IDTYPE, set<IDTYPE> > succ;  // tree structure (postsets)
            map<IDTYPE, unsigned> count;  // number of elements with the same root
            map<IDTYPE, IDTYPE> name;
            map<IDTYPE, IDTYPE> root;
      
            IDTYPE unused_name;
            IDTYPE insert(IDTYPE elem, IDTYPE name);
            void exportSubtree(IDTYPE nd, IDTYPE rt);
            void exportSubtree(IDTYPE nd, IDTYPE rt, set<IDTYPE>* tpt);
         };
       
         class RELATION
	 {
	    public:
	       RELATION(PETRIGRAPH* pg);
	       RELATION( PETRIGRAPH* pg, map<IDTYPE, IDTYPE> vunion,
	                 map<IDTYPE, IDTYPE> eunion, map<IDTYPE, IDTYPE> tunion );
	       RELATION( PETRIGRAPH* pg, set< pair<IDTYPE, IDTYPE> > vunion,
	                 set< pair<IDTYPE, IDTYPE> > eunion,
			 set< pair<IDTYPE, IDTYPE> > tunion );
			 
	       IDTYPE find(IDTYPE elem, GRAPH_ELEMENT_TYPE which);
	       IDTYPE getAnyElement(IDTYPE name, GRAPH_ELEMENT_TYPE which);
	       set<IDTYPE> getNames(GRAPH_ELEMENT_TYPE which);
	       void addRel(IDTYPE x, IDTYPE y, GRAPH_ELEMENT_TYPE which);
			 
	       /* add vertex1 ~ vertex2 to the relation */
	       void addVertexRel(IDTYPE vertex1, IDTYPE vertex2);
	       
	       /*
	        * add edge1 ~ edge2 to the relation.
		* the caller must take care that labels of edge1 and edge2 are identical
		*/
	       void addEdgeRel(IDTYPE edge1, IDTYPE edge2);
	       
	       /*
	        * add t1 ~ t2 to the relation.
		* the caller must take care that the rules
		* associated with both transitions are identical
		*/
	       void addTransitionRel(IDTYPE t1, IDTYPE t2);
	       
	       bool areVerticesInRel(IDTYPE vertex1, IDTYPE vertex2);
	       bool areEdgesInRel(IDTYPE edge1, IDTYPE edge2);
	       bool areTransitionsInRel(IDTYPE t1, IDTYPE t2);
	       bool areInRel(IDTYPE x, IDTYPE y, GRAPH_ELEMENT_TYPE which);	           
	       
	       /*
	        * extend the relation to the least possible closed equivalence relation
		*/
	       void extend();
	       
	       unsigned getDepth(IDTYPE id);
	       void setDepth(IDTYPE id, unsigned d);
	       
	       friend ostream& operator<<(ostream& ost, RELATION& th);
		
	    private:
	       PETRIGRAPH* petrigraph;
	       PARTITION vrel;
	       PARTITION erel;
	       PARTITION trel;
	       map<IDTYPE, unsigned> depth;
	 };
              
         MERGE_PETRIGRAPH_UF();
	 virtual ~MERGE_PETRIGRAPH_UF();
       
         virtual void initialize(PETRIGRAPH* pg);
	 virtual void run(RELATION* r);
	 virtual void run( map<IDTYPE, IDTYPE> vunion,
	                   map<IDTYPE, IDTYPE> eunion, map<IDTYPE, IDTYPE> tunion );
	 virtual void run( set< pair<IDTYPE, IDTYPE> > vunion,
	                   set< pair<IDTYPE, IDTYPE> > eunion,
			   set< pair<IDTYPE, IDTYPE> > tunion );        
			   
       protected:
         PETRIGRAPH* petrigraph;
	 
	 // incremental coverability
	 bool incremental_cov;
      };
   };
};
using namespace algorithms;
using namespace algorithms::petrigraph;

#endif
