/***************************************************************************
 *   Copyright (C) 2006 by root   *
 *   root@Vitali   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "metrikff_call_scenario.h"
#include "petrigraph.h"
#include "gxlreader.h"
#include "pddlwriter.h"
#include "marking_reader.h"
#include "ffreader.h"

#include <string.h>

using namespace std;

namespace scenarios
{

METRIKFF_CALL_SCENARIO::METRIKFF_CALL_SCENARIO()
{
      this->ID = ALGORITHM_METRICFF_CALL;
      this->name = "metricff_scenario";
      
      this->inspec.push_back(pair<string, DATA_FORMAT>("petrinet", GXL));
      this->inspec.push_back(pair<string, DATA_FORMAT>("marking", PLAIN_TEXT));	
      this->outspec.push_back(pair<string, DATA_FORMAT>("trace", PLAIN_TEXT));
      
      this->call_descriptor = "Usage: augur -sc=coverability petrinet marking trace\n\n\
petrinet\tpetri net in GXL format\n\
marking\ttested marking\n\
trace\ttrace to coverable marking\n\n";
}


METRIKFF_CALL_SCENARIO::~METRIKFF_CALL_SCENARIO()
{
}

void METRIKFF_CALL_SCENARIO::run()
{
 if(coloured)
 {
   cout << "Error: Metric FF can not work with attributed Petri nets (not yet implemented)" << endl;
   exit(1);
 }

 //read petrinet
 GXLREADER gxlreader;
 gxlreader.initialize((char*)(this->filemap["petrinet"]).c_str());
 PETRIGRAPH pg;
 GTS gts;
 gxlreader.run(pg, gts);

 // marking
 MARKING_READER mr;
 mr.initialize((char*)(this->filemap["marking"].c_str()));
 list<MARKING> lm = mr.run(0);

 string pddlFile;
 char * pch;
 char filename_copy[strlen(this->filemap["petrinet"].c_str())+1];
 strcpy(filename_copy, this->filemap["petrinet"].c_str());
 pch = strtok (filename_copy,"/");

 while (1)
 {
	pddlFile += string("/");
	string tmp = (string)pch;
	pch = strtok (NULL, "/");

	if(pch==NULL)
	 break;	
	else
	 pddlFile += tmp;
 }

 string pddlFile_pn = pddlFile + string("pn");

 PDDLWriter pddl_writer;
 pddl_writer.initialize((char*)(pddlFile_pn.c_str()));
 pddl_writer.run((PTNET&)pg, lm);

 writePPID();

 init_time();

 // call metric ff
 string s = "./ff -f " + pddlFile + "pn.pddl" + " -o domain/petri1.pddl > " + pddlFile + "ffresult";
 system(s.c_str());

 double t = get_time();

 cout << "Result of Metric FF is written to " << pddlFile << "ffresult" << endl;

 // read the output of Metric FF
 FFREADER ffr;
 ffr.initialize((char*)string(pddlFile + "ffresult").c_str());
 bool cov;
 COUNTER_EXAMPLE * ce = ffr.run((PTNET*)(&pg), &cov);

 // write counter-example
 ofstream of((char*)(this->filemap["trace"].c_str()));

 int n = ce->trans_strings.size();
 for(int i=0; i<n; i++)
  of << ce->trans_strings[i] << endl << ce->trans_ids[i] << endl << endl;

 of.close();

 // output
 if(!cov)
  cout << endl << "The Final Marking(s) are not coverable.\n";
 else
  cout << endl << "The Final Marking(s) are coverable\n";

 cout << "time used: " << t << " sec." << endl;

}

}
