#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <stdexcept>
#include <iostream>
#include <string.h>
#include "newGTXLReader.h"
#include "rule.h"
// #include "gxlreader.h"
#include "match.h"
// #include "GTXLWriter.h"
// #include "GTXLReader.h"
#include <vector>

namespace algorithms {

namespace readers {

using namespace std;

newGTXLREADER::newGTXLREADER()
{
	lastID = 0;
}

newGTXLREADER::~newGTXLREADER()
{
	xmlFreeDoc(doc);
}
	
void newGTXLREADER::initialize(char* filename){
	doc = xmlParseFile(filename);
	if(doc == 0){
		cerr << "Parse error. File could not be read" << endl;
		xmlFreeDoc(doc);
		exit(1);
	}
    if (xmlStrcmp(xmlDocGetRootElement(doc)->name, (const xmlChar *) "gtxl")) 
      {
	cerr << "document of the wrong type, root node != gtxl" << endl;
	xmlFreeDoc(doc);
	exit(1);
      }
}
xmlNodePtr newGTXLREADER::GetRelend(xmlNodePtr rel, int order){
	xmlNodePtr relend = GetChildrenNode(rel, "relend");
	while(relend != 0){
		try{//Gucken, ob startorder nicht in xlink steht
			if(atoi(getProp(relend, "startorder").c_str()) == order)
				return relend;
		}
		catch(exception){//Gucken ob startorder in xlink steht
			try{
				xmlNodePtr type = GetChildrenNode(rel, "type");
				if(type != 0){
					xmlNodePtr xlink = get(getProp(type, "href"), xLinks);
					xmlNodePtr rele = GetChildrenNode(xlink, "relend");
					while(rele != 0){
						if(atoi(getProp(rele, "startorder").c_str()) == order){
							return relend;
						}
						rele = GetNextChild(rele, "relend");
					}
				}
				return 0;
			}
			catch(exception){
				return 0;
			}
		}	
		relend = GetNextChild(relend, "relend");
	}
	return 0;
}

xmlNodePtr newGTXLREADER::GetNextChild(xmlNodePtr child1, char* name_child)
{
	 xmlNodePtr child2 = child1->next;
	while((child2 != NULL) && (xmlStrcasecmp(child2->name, (const xmlChar*)name_child)))
		child2 = child2->next;
	return child2;
}

xmlNodePtr newGTXLREADER::GetAttributeNode(xmlNodePtr node, char* name)
{
	xmlNodePtr cur = GetChildrenNode(node, "attr");
	while(cur != NULL)
		{
			if (!(xmlStrcmp(xmlGetProp(cur, (const xmlChar*)"name"), (const xmlChar*)name)))
				return cur;
			else
				cur = GetNextChild(cur, "attr");
		}
	return cur;
}

xmlNodePtr newGTXLREADER::GetChildrenNode(xmlNodePtr parent, char* name_child)
{
	xmlNodePtr child = parent->xmlChildrenNode;
	while((child != NULL) && (xmlStrcasecmp(child->name, (const xmlChar*)name_child)))
		child = child->next;
	return child;	
}

TYPES newGTXLREADER::readTypes(string input){
	TYPES result;
	string type;
	for(int i = 0, k = 0; i < input.size();i++){
		if(input[i] == '-'){
			type = substr(input,k,i-1);
			k = i+1;
		}
		else if(i == (input.size()-1)){
			type = substr(input,k,i);
		}
		if(input[i] == '-' || i == (input.size()-1)){
			if(type == "int")
				result.push_back(INTEGER);
			else if(type == "str" || type == "string")
				result.push_back(STRING);
			else if(type == "bool")
				result.push_back(BOOL);
			else{
				cerr << "The type: " << type << " is not supported yet" << endl;
				exit(1);
			}
		}
	}
	return result;
}

/*
void newGTXLREADER::start(char * filename){
// Zum Testen von Reader. Ausgabe in file:///home/mayer/augur/optimized/src/madeGTXLfile.xml
	initialize(filename);
	GTS gts;
	run(gts);
	writers::GTXLWRITER testWriter;
	testWriter.initialize("madeGTXLfile.xml");
	testWriter.run(gts);
}
*/

list<EXPRESSION> newGTXLREADER::addAttributes(list<EXPRESSION> liste, xmlNodePtr firstAttr){
	xmlNodePtr pStr;
	list<EXPRESSION> result;
	for(list<EXPRESSION>::iterator iter = liste.begin(); iter != liste.end(); iter++){
		while(firstAttr != 0){
			if(getProp(firstAttr, "name") == iter->getName()){
				pStr = GetChildrenNode(firstAttr, "string");
				if(pStr != 0)
					pStr = pStr->xmlChildrenNode;
				else{
					cerr << "GTXLReader: There must be a <string> node in an <attr> node. Error in line: " << xmlGetLineNo(firstAttr) << endl;
					exit(1);
				} 
				if (pStr != NULL){
					iter->addExpr((string)(char*)pStr->content);
				}
				else{
					cerr << "GTXLReader: <string><\\string> found in line: " << xmlGetLineNo(firstAttr) << " please enter value" << endl;
					exit(1);
				}
				result.push_back(*iter);
			}
			else if(iter->getName() == "label")
				result.push_back(*iter);
		firstAttr = GetNextChild(firstAttr, "attr");
		}
	}
	return result;
}

string newGTXLREADER::getProp(xmlNodePtr ptr, char* name){
	xmlChar* xmlPointer = xmlGetProp(ptr, (const xmlChar *)name);
	if(xmlPointer == 0 && name != "startorder"){
		cerr << "There is no property with the name of: " << name << endl;
		cerr << "Error in xml File. Line: " << xmlGetLineNo(ptr) << endl;
		exit(1);
	}
	return (string)(char*)xmlPointer;
}

void newGTXLREADER::readXlinks(xmlNodePtr graph){
	string sid, kind, name, fromto;
	xmlNodePtr node = GetChildrenNode(graph,"node");
	xmlNodePtr attr, temp;
	while(node != 0){
		VERTEX v;
		sid = getProp(node, "id");
		nodes.push_back(sid);
		v.setLabel(sid);
		xLinks.insert(make_pair(sid,node));//Zum prüfen ob xlink vorhanden
		xmlNodePtr in_node_graph = GetChildrenNode(node, "graph");
		xmlNodePtr rel = GetChildrenNode(in_node_graph, "rel");
		xmlNodePtr edge = GetChildrenNode(in_node_graph,"edge");
		attr = GetChildrenNode(node, "attr");
		while(attr != 0){
			name = getProp(attr, "name");
			kind = getProp(attr, "kind");
			if(kind == "AttrType"){
				temp = GetChildrenNode(attr, "string");
				if(temp == 0){
					cerr << "There must be an node: <string> in an <attr> node." << endl;
					exit(1);
				}
				temp = temp->xmlChildrenNode;
				if(temp == 0){
					cerr << "There is no value in one attr node." << endl;
					exit(1);
				}
				TYPES typ = readTypes((string)(char*)temp->content);
				EXPRESSION ex;
				ex.addType(typ);
				ex.addName(name);
				v.setType(typ);
				v.addAttributes(ex);
			}
			attr = GetNextChild(attr, "attr");
		}
		xNodes.insert(make_pair(sid,v));
		while(rel != 0){
			EDGE e;
			attr = GetChildrenNode(rel, "attr");
			while(attr != 0){
				name = getProp(attr, "name");
				kind = getProp(attr, "kind");
				if(kind == "AttrType"){
					temp = GetChildrenNode(attr, "string");
					if(temp == 0){
						cerr << "There must be an node: <string> in an <attr> node." << endl;
						exit(1);
					}
					temp = temp->xmlChildrenNode;
					if(temp == 0){
						cerr << "There is no value in attr node." << endl;
						exit(1);
					}
					TYPES typ = readTypes((string)(char*)temp->content);
					EXPRESSION ex;
					ex.addType(typ);
					ex.addName(name);
					e.setType(typ);
					e.addAttributes(ex);
				}
				attr = GetNextChild(attr, "attr");
			}
			sid = getProp(rel, "id");
			e.setLabel(sid);
			xLinks.insert(make_pair(sid, rel));
			xEdges.insert(make_pair(sid,e));
			rel = GetNextChild(rel,"rel");
		}
		while(edge != 0){
			EDGE e;
			fromto = getProp(edge, "to");
			if(!contains(nodes,fromto)){
				cerr << "There is no property with the name of: " << fromto << endl;
				cerr << "Error in line: " << xmlGetLineNo(edge) << endl;
				exit(1);
			}
			fromto = getProp(edge, "from");
			if(!contains(nodes,fromto)){
				cerr << "There is no property with the name of: " << fromto << endl;
				cerr << "Error in line: " << xmlGetLineNo(edge) << endl;
				exit(1);
			}
			attr = GetChildrenNode(edge, "attr");
			while(attr != 0){
				name = getProp(attr, "name");
				kind = getProp(attr, "kind");
				if(kind == "AttrType"){
					temp = GetChildrenNode(attr, "string");
					if(temp == 0){
						cerr << "There must be an node: <string> in an <attr> node." << endl;
						exit(1);
					}
					temp = temp->xmlChildrenNode;
					if(temp == 0){
						cerr << "There is no value in attr node." << endl;
						exit(1);
					}
					TYPES typ = readTypes((string)(char*)temp->content);
					EXPRESSION ex;
					ex.addType(typ);
					ex.addName(name);
					e.setType(typ);
					e.addAttributes(ex);
				}
				attr = GetNextChild(attr, "attr");
			}
			sid = getProp(edge, "id");
			e.setLabel(sid);
			xEdges.insert(make_pair(sid, e));
			xLinks.insert(make_pair(sid,edge));
			edge = GetNextChild(edge,"edge");
		}
		node = GetNextChild(node,"node");
	}
}

void newGTXLREADER::readRules(xmlNodePtr firstRule, GTS& GraphGrammar){
	vector<string> rids, xids;
	string sid("");
	string xid("");
	string name("");
	xmlNodePtr preserved;
	xmlNodePtr dre;
	xmlNodePtr xlink;
	while(firstRule != 0){
		RULE rule;
		HYPERGRAPH lhs, rhs;
		rids.clear();//Sorgt für eindeutige IDs
		name = getProp(firstRule, "name");
		sid = getProp(firstRule,"id");
		if(contains(xids,sid)){
			cerr << IDERROR << xmlGetLineNo(firstRule) << endl;
			cerr << sid << " already exist in this file" << endl;
			exit(1);
		}
		xids.push_back(sid);
		rule.setName(name);
		rule.setID(++lastID);
		xmlNodePtr guard_node = GetChildrenNode(firstRule, "precondition");
		if(guard_node != NULL)
		{
			xmlNodePtr guard_node2 = GetChildrenNode(guard_node, "condition");
			if(guard_node2 != NULL)
			{
				xmlNodePtr guard_node3 = GetChildrenNode(guard_node2, "attrCondition");
				if(guard_node3 != NULL)
				{
					xmlNodePtr guards_node = GetChildrenNode(guard_node3, "string");
					if(guards_node != NULL) 
					{
						guards_node = guards_node->xmlChildrenNode;
						if (guards_node != NULL)
						{
							string guards = (string)((char*)guards_node->content);
	
							char ch[guards.size()+1];
							strcpy(ch, guards.c_str());
							char * pch;
							pch = strtok(ch,",");
							
							while (pch!=NULL)
							{
								EXPRESSION eg;
								eg.addName("guard");
								TYPES types;
								types.push_back(BOOL);
								eg.addType(types);
								eg.addExpr((string)pch);
								rule.addGuard(eg);
								pch = strtok (NULL, ",");
							}
						}
					}
				}
			}
		}
		preserved = GetChildrenNode(firstRule, "preserved");
		if(preserved == 0){
			cerr << "GTXLReader: Rule without Node: <preserved> found in line: " << xmlGetLineNo(firstRule) << endl;
			exit(1);
		}
		lhs = readHypergraph(preserved);
		rhs = readRHS(firstRule);
		rule.setInterface(createMapping(&lhs,&rhs));
		rule.setLeft(&lhs);
		rule.setRight(&rhs);
		GraphGrammar.addRule(&rule);
		firstRule = GetNextChild(firstRule,"rule");
	}
}

HYPERGRAPH newGTXLREADER::readRHS(xmlNodePtr rule){
	HYPERGRAPH result;
	ids.clear();
	vector<string> deletedIDs;
	string sid, xid;
	xmlNodePtr xlink, attr;
//DELETED
	xmlNodePtr ruleElement = GetChildrenNode(rule, "deleted");
	xmlNodePtr dre = GetChildrenNode(ruleElement, "node");
	if(dre != 0){
		cerr << "This software can not delete nodes" << endl;
		exit(1);
	}
	dre = GetChildrenNode(ruleElement, "edge");
	while(dre != 0){
		getProp(dre,"from");
		getProp(dre, "to");
		deletedIDs.push_back(getProp(dre, "id"));
		dre = GetNextChild(dre, "edge");
	}
	dre = GetChildrenNode(ruleElement, "rel");
	while(dre != 0){
		deletedIDs.push_back(getProp(dre, "id"));
		xmlNodePtr relend = GetChildrenNode(dre, "relend");
		while(relend != 0){
			getProp(relend, "target");
			relend = GetNextChild(relend, "relend");
		}
		dre = GetNextChild(dre, "rel");
	}
//PRESERVED
	ruleElement = GetChildrenNode(rule, "preserved");
	ruleElement = GetChildrenNode(ruleElement, "graph");
	dre = GetChildrenNode(ruleElement, "node");
	if(dre == 0){
		cerr << "In Rule: " << getProp(rule, "id") << " are in preserved no nodes" << endl;
	}
	while(dre != 0){
		sid = getProp(dre, "id");
		if(contains(ids,sid)){
			cerr << IDERROR << xmlGetLineNo(dre) << endl;
			exit(1);
		}
		ids.push_back(sid);
		xlink = GetChildrenNode(dre, "type");
		if(xlink == 0){
			cerr << XLINKERROR << xmlGetLineNo(dre) << endl;
			exit(1);
		}
		xid = getProp(xlink, "href");
		try{
			get(xid, xLinks);
		}
		catch(NoSuchElementException){
			cerr << XLINKERROR << xmlGetLineNo(xlink) << endl;
			exit(1);
		}
		attr = GetChildrenNode(dre, "attr");
		VERTEX v = get(xid, xNodes);
		if((attr == 0 xor v.getAttributes().empty())){
			cerr << "An error occured, based on a wrong attribute. Error in line: " << xmlGetLineNo(dre) << endl;
			exit(1);
		}
		v.setAttributes(addAttributes(v.getAttributes(), attr));
		v.setIdLabel(sid);
		v.setID(++lastID);
		result.addVertex(v,lastID);
		dre = GetNextChild(dre, "node");
	}
	list<IDTYPE> listOfVertices;
	dre = GetChildrenNode(ruleElement, "edge");
	while(dre != 0){
		sid = getProp(dre, "id");
		if(!contains(deletedIDs, sid)){
			if(contains(ids, sid)){
				cerr << IDERROR << xmlGetLineNo(dre) << endl;
				exit(1);
			}
			ids.push_back(sid);
			xlink = GetChildrenNode(dre, "type");
			if(xlink == 0){
				cerr << XLINKERROR << xmlGetLineNo(dre) << endl;
				exit(1);
			}
			xid = getProp(xlink, "href");
			try{
			get(xid, xLinks);
			}
			catch(NoSuchElementException){
				cerr << XLINKERROR << xmlGetLineNo(xlink) << endl;
				exit(1);
			}
			attr = GetChildrenNode(dre, "attr");
			EDGE edge = get(xid, xEdges);
			if((attr == 0 xor edge.getAttributes().empty())){
				cerr << "An error occured, based on a wrong attribute. Error in line: " << xmlGetLineNo(dre) << endl;
				exit(1);
			}
			edge.setAttributes(addAttributes(edge.getAttributes(), attr));
			edge.setIdLabel(sid);
			edge.setID(++lastID);
			string from = getProp(dre, "from");
			string to = getProp(dre, "to");
			if(!((result.hasVertex(from))&&(result.hasVertex(to)))){
				cerr << "The vertex used in edge's connection does not exist in this graph" << endl;
				cerr << "Error in line: " << xmlGetLineNo(dre) << endl;
				exit(1);
			}
			VERTEX ver1, ver2;
			listOfVertices.clear();
			ver1 = result.getVertex(from);
			ver2 = result.getVertex(to);
			listOfVertices.push_back(ver1.getID());
			listOfVertices.push_back(ver2.getID());
			
			result.addEdgeToCE(ver1.getID(), edge.getID());
			result.addEdgeToCE(ver2.getID(), edge.getID());
			result.addEdge(edge,lastID,listOfVertices);
		}
		dre = GetNextChild(dre, "edge");
	}
	dre = GetChildrenNode(ruleElement, "rel");
	while(dre != 0){
		sid = getProp(dre, "id");
		if(!contains(deletedIDs, sid)){
			if(contains(ids, sid)){
				cerr << IDERROR << xmlGetLineNo(dre) << endl;
				exit(1);
			}
			ids.push_back(sid);
			xlink = GetChildrenNode(dre, "type");
				if(xlink == 0){
					cerr << XLINKERROR << xmlGetLineNo(dre) << endl;
					exit(1);
				}
				xid = getProp(xlink, "href");
				try{
					get(xid,xLinks);
				}
				catch(NoSuchElementException){
					cerr << XLINKERROR << xmlGetLineNo(xlink) << endl;
					exit(1);
				}
				attr = GetChildrenNode(dre, "attr");
				EDGE edge = get(xid, xEdges);
				if((attr == 0 xor edge.getAttributes().empty())){
					cerr << "An error occured, based on a wrong attribute. Error in line: " << xmlGetLineNo(dre) << endl;
					exit(1);
				}
				edge.setAttributes(addAttributes(edge.getAttributes(), attr));
				edge.setIdLabel(sid);
				edge.setID(++lastID);
				int order = 0;
				xmlNodePtr relend = GetRelend(dre, order);
				listOfVertices.clear();
				string target;
				while(relend != 0){
					target = getProp(relend, "target");
					if(!result.hasVertex(target)){
						cerr << "The vertex used in edge's connection (relend) does not exist in this graph: " << xmlGetLineNo(dre) << endl;
						exit(1);
					}
					VERTEX ver = result.getVertex(target);
					listOfVertices.push_back(ver.getID());
					order++;
					relend = GetRelend(dre, order);
				}
			result.addEdge(edge,lastID, listOfVertices);
		}
		dre = GetNextChild(dre, "rel");	
	}
//CREATED
	ruleElement = GetChildrenNode(rule, "created");
	dre = GetChildrenNode(ruleElement, "node");
	while(dre != 0){
		sid = getProp(dre, "id");
		if(contains(ids,sid)){
			cerr << IDERROR << xmlGetLineNo(dre) << endl;
			exit(1);
		}
		ids.push_back(sid);
		xlink = GetChildrenNode(dre, "type");
		if(xlink == 0){
			cerr << XLINKERROR << xmlGetLineNo(dre) << endl;
			exit(1);
		}
		xid = getProp(xlink, "href");
		try{
			get(xid, xLinks);
		}
		catch(NoSuchElementException){
			cerr << XLINKERROR << xmlGetLineNo(xlink) << endl;
			exit(1);
		}
		attr = GetChildrenNode(dre, "attr");
		VERTEX v = get(xid, xNodes);
		if((attr == 0 xor v.getAttributes().empty())){
			cerr << "An error occured, based on a wrong attribute. Error in line: " << xmlGetLineNo(dre) << endl;
			exit(1);
		}
		v.setAttributes(addAttributes(v.getAttributes(), attr));
		v.setIdLabel(sid);
		v.setID(++lastID);
		result.addVertex(v,lastID);
		dre = GetNextChild(dre, "node");
	}
	dre = GetChildrenNode(ruleElement, "edge");
	while(dre != 0){
		sid = getProp(dre, "id");
		if(contains(ids, sid)){
			cerr << IDERROR << xmlGetLineNo(dre) << endl;
			exit(1);
		}
		ids.push_back(sid);
		xlink = GetChildrenNode(dre, "type");
		if(xlink == 0){
			cerr << XLINKERROR << xmlGetLineNo(dre) << endl;
			exit(1);
		}
		xid = getProp(xlink, "href");
		try{
			get(xid, xLinks);
		}
		catch(NoSuchElementException){
			cerr << XLINKERROR << xmlGetLineNo(xlink) << endl;
			exit(1);
		}
		attr = GetChildrenNode(dre, "attr");
		EDGE edge = get(xid, xEdges);
		if((attr == 0 xor edge.getAttributes().empty())){
			cerr << "An error occured, based on a wrong attribute. Error in line: " << xmlGetLineNo(dre) << endl;
			exit(1);
		}
		edge.setAttributes(addAttributes(edge.getAttributes(), attr));
		edge.setIdLabel(sid);
		edge.setID(++lastID);
		string from = getProp(dre, "from");
		string to = getProp(dre, "to");
		if(!((result.hasVertex(from))&&(result.hasVertex(to)))){
			cerr << "The vertex used in edge's connection does not exist in this graph " << endl;
			cerr << "Error in line: " << xmlGetLineNo(dre) << endl;
			exit(1);
		}
		VERTEX ver1, ver2;
		listOfVertices.clear();
		ver1 = result.getVertex(from);
		ver2 = result.getVertex(to);
		listOfVertices.push_back(ver1.getID());
		listOfVertices.push_back(ver2.getID());
		
		result.addEdgeToCE(ver1.getID(), edge.getID());
		result.addEdgeToCE(ver2.getID(), edge.getID());
		result.addEdge(edge,lastID,listOfVertices);
		dre = GetNextChild(dre, "edge");
	}
	dre = GetChildrenNode(ruleElement, "rel");
	while(dre != 0){
		sid = getProp(dre, "id");
		if(contains(ids, sid)){
			cerr << IDERROR << xmlGetLineNo(dre) << endl;
			exit(1);
		}
		ids.push_back(sid);
		xlink = GetChildrenNode(dre, "type");
			if(xlink == 0){
				cerr << XLINKERROR << xmlGetLineNo(dre) << endl;
				exit(1);
			}
			xid = getProp(xlink, "href");
			try{
				get(xid, xLinks);
			}
			catch(NoSuchElementException){
				cerr << XLINKERROR << xmlGetLineNo(xlink) << endl;
				exit(1);
			}
			attr = GetChildrenNode(dre, "attr");
			EDGE edge = get(xid, xEdges);
			if((attr == 0 xor edge.getAttributes().empty())){
				cerr << "An error occured, based on a wrong attribute. Error in line: " << xmlGetLineNo(dre) << endl;
				exit(1);
			}
			edge.setAttributes(addAttributes(edge.getAttributes(), attr));
			edge.setIdLabel(sid);
			edge.setID(++lastID);
			int order = 0;
			xmlNodePtr relend = GetRelend(dre, order);
			listOfVertices.clear();
			string target;
			while(relend != 0){
				target = getProp(relend, "target");
				if(!result.hasVertex(target)){
					cerr << "The vertex used in edge's connection (relend) does not exist in this graph" << endl;
					cerr << "Error in line: " << xmlGetLineNo(dre) << endl;
					exit(1);
				}
				VERTEX ver = result.getVertex(target);
				listOfVertices.push_back(ver.getID());
				order++;
				relend = GetRelend(dre, order);
			}
		result.addEdge(edge,lastID, listOfVertices);
	dre = GetNextChild(dre, "rel");	
	}
	return result;
}

MATCH newGTXLREADER::createMapping(HYPERGRAPH* lhs, HYPERGRAPH* rhs){
	map<IDTYPE, IDTYPE> match;
	MATCH result;
//Knoten
	map<IDTYPE, VERTEX>* lhsVertices = lhs->getVerticesPointer();
	map<IDTYPE, VERTEX>* rhsVertices = rhs->getVerticesPointer();
	map<IDTYPE, VERTEX>::iterator lhsVertIter = lhsVertices->begin();
	map<IDTYPE, VERTEX>::iterator rhsVertIter = rhsVertices->begin();
//Kanten
	map<IDTYPE, EDGE>* lhsEdges = lhs->getEdgesPointer();
	map<IDTYPE, EDGE>* rhsEdges = rhs->getEdgesPointer();
	map<IDTYPE, EDGE>::iterator lhsEdgeIter = lhsEdges->begin();
	map<IDTYPE, EDGE>::iterator rhsEdgeIter = rhsEdges->begin();
	while(lhsVertIter != lhsVertices->end()){
		rhsVertIter = rhsVertices->begin();
		while(rhsVertIter != rhsVertices->end()){
			if(lhsVertIter->second.getIdLabel() == rhsVertIter->second.getIdLabel()){
				lhsVertIter->second.setIdLabel(lhsVertIter->second.getIdLabel() + "_lhs");
				rhsVertIter->second.setIdLabel(rhsVertIter->second.getIdLabel() + "_rhs");
				match.insert(make_pair(lhsVertIter->second.getID(), rhsVertIter->second.getID()));
			}
			rhsVertIter++;
		}
		lhsVertIter++;
	}
	while(lhsEdgeIter != lhsEdges->end()){
		rhsEdgeIter = rhsEdges->begin();
		while(rhsEdgeIter != rhsEdges->end()){
			if(lhsEdgeIter->second.getIdLabel() == rhsEdgeIter->second.getIdLabel()){
				lhsEdgeIter->second.setIdLabel(lhsEdgeIter->second.getIdLabel() + "_lhs");
				rhsEdgeIter->second.setIdLabel(rhsEdgeIter->second.getIdLabel() + "_rhs");
				match.insert(make_pair(lhsEdgeIter->second.getID(), rhsEdgeIter->second.getID()));
			}
			rhsEdgeIter++;
		}
		lhsEdgeIter++;
	}
	result.setMatch(match);
	return result;
}

HYPERGRAPH newGTXLREADER::readHypergraph(xmlNodePtr root){
	ids.clear();
	HYPERGRAPH result;
	xmlNodePtr xlink;
	xmlNodePtr attr;
	xmlNodePtr graph = GetChildrenNode(root,"graph");
	string sid;//Für IDs
	string xid;//Für xlink IDs
	if(graph == 0){
		cerr << "Rule without `Graph` Node found in line: " << xmlGetLineNo(root) << endl;
		exit(1);
	}
	sid = getProp(graph,"id");
	result.setName(sid);
	xmlNodePtr graphElement = GetChildrenNode(graph,"node");
	while(graphElement != 0){
		VERTEX v;
		sid = getProp(graphElement,"id");
		if(contains(ids,sid)){
			cerr << IDERROR << xmlGetLineNo(graphElement) << endl;
			exit(1);
		}
		ids.push_back(sid);
		xlink = GetChildrenNode(graphElement, "type");
		if(xlink == 0){
			cerr << XLINKERROR << xmlGetLineNo(graphElement) << endl;
			exit(1);
		}
		xid = getProp(xlink,"href");
		try{
			get(xid, xLinks);
		}
		catch(NoSuchElementException){
			cerr << XLINKERROR << xmlGetLineNo(graphElement) << endl;
			exit(1);
		}
		v = get(xid, xNodes);
		attr = GetChildrenNode(graphElement, "attr");
		if((attr == 0 xor v.getAttributes().empty())){
			cerr << "An error occured, based on a wrong attribute. Error in line: " << xmlGetLineNo(graphElement) << endl;
			exit(1);
		}
		v.setAttributes(addAttributes(v.getAttributes(), attr));
		v.setIdLabel(sid);
		v.setID(++lastID);
		result.addVertex(v,lastID);
		graphElement = GetNextChild(graphElement,"node");
	}
	list<IDTYPE> listOfVertices;
	string from, to;
	graphElement = GetChildrenNode(graph,"edge");
	while(graphElement != 0){
		sid = getProp(graphElement, "id");
		if(contains(ids,sid)){
			cerr << IDERROR << xmlGetLineNo(graphElement) << endl;
			exit(1);
		}
		ids.push_back(sid);
		xlink = GetChildrenNode(graphElement, "type");
		if(xlink == 0){
			cerr << XLINKERROR << xmlGetLineNo(graphElement) << endl;
			exit(1);
		}
		xid = getProp(xlink, "href");
		try{
			get(xid, xLinks);
		}
		catch(NoSuchElementException){
			cerr << XLINKERROR << xmlGetLineNo(graphElement) << endl;
			exit(1);
		}
		EDGE edge = get(xid, xEdges);
		attr = GetChildrenNode(graphElement, "attr");
		if((attr == 0 xor edge.getAttributes().empty())){
			cerr << "An error occured, based on a wrong attribute. Error in line: " << xmlGetLineNo(graphElement) << endl;
			exit(1);
		}
		edge.setAttributes(addAttributes(edge.getAttributes(), attr));
		edge.setIdLabel(sid);
		edge.setID(++lastID);
		from = getProp(graphElement,"from");
		to = getProp(graphElement,"to");
		if(!((result.hasVertex(from))&&(result.hasVertex(to)))){
			cerr << "The vertexe used in edge's connection does not exist in this graph" << endl;
			cerr << "Error in line: " << xmlGetLineNo(graphElement) << endl;
			exit(1);
		}
		VERTEX ver1, ver2;
		listOfVertices.clear();
		ver1 = result.getVertex(from);
		ver2 = result.getVertex(to);
		listOfVertices.push_back(ver1.getID());
		listOfVertices.push_back(ver2.getID());
		result.addEdgeToCE(ver1.getID(), edge.getID());
		result.addEdgeToCE(ver2.getID(), edge.getID());

		result.addEdge(edge,lastID,listOfVertices);
		graphElement = GetNextChild(graphElement,"edge");
	}
	graphElement = GetChildrenNode(graph, "rel");
	while(graphElement != 0){
		sid = getProp(graphElement, "id");
		if(contains(ids,sid)){
			cerr << IDERROR << xmlGetLineNo(graphElement) << endl;
			exit(1);
		}
		ids.push_back(sid);
		xlink = GetChildrenNode(graphElement, "type");
		if(xlink == 0){
			cerr << XLINKERROR << xmlGetLineNo(graphElement) << endl;
			exit(1);
		}
		xid = getProp(xlink, "href");
		try{
			get(xid, xLinks);
		}
		catch(NoSuchElementException){
			cerr << XLINKERROR << xmlGetLineNo(graphElement) << endl;
			exit(1);
		}
		EDGE edge = get(xid, xEdges);
		attr = GetChildrenNode(graphElement, "attr");
		if((attr == 0 xor edge.getAttributes().empty())){
			cerr << "An error occured, based on a wrong attribute. Error in line: " << xmlGetLineNo(graphElement) << endl;
			exit(1);
		}
		edge.setAttributes(addAttributes(edge.getAttributes(), attr));
		edge.setIdLabel(sid);
		edge.setID(++lastID);
		int order = 0;
		xmlNodePtr relend = GetRelend(graphElement, order);
		listOfVertices.clear();
		while(relend != 0){
			string target = getProp(relend, "target");
			if(!result.hasVertex(target)){
				cerr << "The vertexe used in edge's connection (relend) does not exist in this graph." << endl;
				exit(1);
			}
			VERTEX ver;
			ver = result.getVertex(target);
			listOfVertices.push_back(ver.getID());
			relend = GetRelend(graphElement, ++order);
		}
		result.addEdge(edge,lastID,listOfVertices);
		graphElement = GetNextChild(graphElement, "rel");
	}
	result.setLastUsedID(lastID);
	return result;
}
	
void newGTXLREADER::run(GTS &GraphGrammar){
	try{
		xmlNodePtr root = xmlDocGetRootElement(doc);
		xmlNodePtr graph = GetChildrenNode(root,"graph");
		readXlinks(graph);
		gts = GetChildrenNode(root,"gts");
		xmlNodePtr init = GetChildrenNode(gts,"initial");
		initial = readHypergraph(init);
		xmlNodePtr firstRule = GetChildrenNode(gts, "rule");
		GraphGrammar.setStartgraph(&initial);
		readRules(firstRule, GraphGrammar);
	}
	catch(exception e){
		cerr << "An error occurd: " << e.what() << endl;
	}

}

}

}


