/***************************************************************************
 *   Copyright (C) 2006 by Michael Mayer   *
 *   mayer@lippe   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef NEWGTXLREADER_H
#define NEWGTXLREADER_H
#include "gts.h"
#include "petrigraph.h"

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <exception>
#include <string>
#include "algorithm.h"
#include "globals.h"

namespace algorithms {

namespace readers {

namespace new_gtxl_reader {

const char* const IDERROR = "GTXLReader: The IDs of the file are inconsistent. Error in line: ";
const char* const XLINKERROR = "GTXLReader: Graph-Element without xlink found in line: ";
// string getProp(xmlNodePtr, char*);

template<class T>
bool contains(vector<T>& vec, T& item){
	for(int i = 0; i < vec.size(); i++)
		if(vec[i] == item)
			return true;
	return false;
}

template<class T>
T get(string id, map<string, T>& m){
	for(typename map<string,T>::iterator iter = m.begin();
		iter != m.end(); iter++){
		if(iter->first == id)
			return iter->second;
	}
	throw NoSuchElementException();
}

};

using namespace new_gtxl_reader;

using namespace data_structures;

using namespace algorithm;

/**
	@author Michael Mayer <mayer@lippe>
*/

class newGTXLREADER : public ALGORITHM{
public:
    newGTXLREADER();
    virtual ~newGTXLREADER();
    virtual void initialize(char *filename);
    virtual void run(GTS &GraphGrammar);
//	virtual void start(char * filename);
private:
//Vorgegebene Typen
	xmlDocPtr doc;
	xmlNodePtr gts;
	HYPERGRAPH initial;
//Eigene Typen
	IDTYPE lastID;
	map<string, VERTEX> xNodes;
	map<string, EDGE> xEdges;
	map<string, xmlNodePtr> xLinks;
	vector<string> ids;//Vergebene IDs
	vector<string> nodes;
//Methoden für xmlParsen
	xmlNodePtr GetRelend(xmlNodePtr rel, int order);
	xmlNodePtr GetAttributeNode(xmlNodePtr node, char* name);
	xmlNodePtr GetNextChild(xmlNodePtr child1, char* child_name);
	xmlNodePtr GetChildrenNode(xmlNodePtr parent, char* child_name);
	string getProp(xmlNodePtr ptr, char* name);
//Eigene Methoden
	TYPES readTypes(string input);
	list<EXPRESSION> addAttributes(list<EXPRESSION> liste, xmlNodePtr graphElement);
	HYPERGRAPH readHypergraph(xmlNodePtr root);
	HYPERGRAPH readRHS(xmlNodePtr rule);
	MATCH createMapping(HYPERGRAPH* lhs, HYPERGRAPH* rhs);
	void readXlinks(xmlNodePtr graph);
	void readRules(xmlNodePtr firstRule, GTS& GraphGrammar);

	char * filename;
};

};

};


#endif
