/***************************************************************************
 *   Copyright (C) 2006 by Maxim Zaks   *
 *   mzaks@neckar   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>

#include "pddlwriter.h"

namespace algorithms {



namespace writers {


PDDLWriter::PDDLWriter()
{
}


PDDLWriter::~PDDLWriter()
{
}

void PDDLWriter::initialize(char *pfilename)
{
    int n = strlen(pfilename);
    
    filename = new char[n+1];
    strcpy(filename,pfilename);
}
    
void PDDLWriter::run(PTNET &pn, MARKING m)
{
	list<MARKING> m_list;
	m_list.push_back(m);

	run(pn, m_list);
// 	string pddlFile;
// 	char * pch;
// 	char filename_copy[strlen(filename)+1];
// 	strcpy(filename_copy, filename);
// 	pch = strtok (filename_copy,".");
// 		
// 	while (pch!=NULL && strcmp(pch, "ps")!=0)
//   	{
// 		if(pddlFile.size()!=0)
// 			pddlFile += string(".");
// 		pddlFile += (string)pch;
//     		pch = strtok (NULL, ".");		
//   	}
// 
// 	pddlFile += string(".pddl");
// 
// 	writePDDLFile(pn, (char*)pddlFile.c_str(), m);

// 	callDot((char*)pddlFile.c_str(), filename);
}

void PDDLWriter::run(PTNET &pn, list<MARKING> m_list)
{
	string pddlFile;
	char * pch;
	char filename_copy[strlen(filename)+1];
	strcpy(filename_copy, filename);
	pch = strtok (filename_copy,".");
		
	while (pch!=NULL && strcmp(pch, "ps")!=0)
  	{
		if(pddlFile.size()!=0)
			pddlFile += string(".");
		pddlFile += (string)pch;
    		pch = strtok (NULL, ".");
  	}

	pddlFile += string(".pddl");

	writePDDLFile(pn, (char*)pddlFile.c_str(), m_list);

// 	callDot((char*)pddlFile.c_str(), filename);
}

void PDDLWriter::writePDDLFile(PTNET & pn, char * pddlFile, list<MARKING> m_list)
{
	ofstream of(pddlFile);
	
	// Header
	of << "(define (problem augur)" << endl << "(:domain petri)" << endl << "(:objects" <<endl;

	map<IDTYPE, TRANSITION> transitions = pn.getTransitions();
	map<IDTYPE, EDGE> edges = pn.getEdges();

	IDTYPE lastUsedID = pn.getLastUsedID();

	// Transitions
	of << "    ";
	for(map<IDTYPE, TRANSITION>::iterator itt = transitions.begin(); itt!=transitions.end(); itt++)
	{

		of << "t" << itt->first << " ";
	}
	of << "- transition" << endl;


	// Places
	of << "    ";
	for(map<IDTYPE, EDGE>::iterator ite = edges.begin(); ite!=edges.end(); ite++)
	{

		of << "p" << ite->first << " ";
	}
	of << "- place)" << endl;

	// Init
	of << "(:init" << endl;

	for(map<IDTYPE, EDGE>::iterator ite = edges.begin(); ite!=edges.end(); ite++)
	{

		MARKING initial = pn.getInitialMarking();
		
		int summaTokens = 0;
		
		vector<TOKEN> vt = initial.getMarking(ite->first);
		for(vector<TOKEN>::iterator it = vt.begin(); it != vt.end(); it++)
		{
			
			summaTokens += it->getCount();
		}
		
		char buf[16];
		sprintf(buf, "%d", summaTokens);
		of << "     " << "(= (number-of-tokens p" << ite->first << ") " << buf << ")" << endl;
	}
	of << endl << endl;
	for(map<IDTYPE, TRANSITION>::iterator itt2 = transitions.begin(); itt2!=transitions.end(); itt2++)
	{
		// preset
		vector<TCONNECTION> c1 = pn.getPreset(itt2->first);
		for(vector<TCONNECTION>::iterator itc1=c1.begin(); itc1!=c1.end(); itc1++)
		{
			if(!coloured || itc1->attributes.size()==0)
			{
// 				of << "_" << itc1->ID << " -- " << "_" << itt2->first << "[dir=forward, label=\"";
//           			of << itc1->weight << "\"]\n";
				
				of << "     " << "(incoming p" << itc1->ID << " t" << itt2->first << ")" << endl;
				of << "     " << "(= (incoming-tokens t" << itt2->first << " p" << itc1->ID << ") " << itc1->weight << ")" << endl;
			}
			of << endl;
		}
		
		// postset
		vector<TCONNECTION> c2 = pn.getPostset(itt2->first);
		for(vector<TCONNECTION>::iterator itc2=c2.begin(); itc2!=c2.end(); itc2++)
		{
			if(!coloured || itc2->attributes.size()==0)
			{
// 				of << "_" << itt2->first << " -- " << "_" << itc2->ID << "[dir=forward, taillabel=\"";
//           			of << itc2->weight << "\"]\n";

				of << "     " << "(outgoing t" << itt2->first << " p" << itc2->ID << ")" << endl;
				of << "     " << "(= (outgoing-tokens t" << itt2->first << " p" << itc2->ID << ") " << itc2->weight << ")" << endl;
			}
			of << endl;
		}
	}



	of << ")" << endl;	// End Init

	of << "(:goal" << endl;
	of << "  (or"  << endl;
	for(list<MARKING>::iterator m = m_list.begin(); m!=m_list.end(); m++){

		of << "    (and" << endl;
		map< IDTYPE, vector<TOKEN> > markingsMap = m->getMarkings();
		for(map< IDTYPE, vector<TOKEN> >::iterator iter=markingsMap.begin(); iter != markingsMap.end(); iter++){

			if(iter->second.size() > 0)
			 of << "      (>= (number-of-tokens p" << iter->first << ") " << iter->second.front().getCount() << ")"  << endl;
		} 
		of << "    )" << endl;
	}
	of << "  )" << endl;
	of << ")" << endl;	//End Goal
	of << ")" << endl;	//End Header
	
}

}
}

