/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "petrigraph.h"

namespace data_structures
{
   PETRIGRAPH::PETRIGRAPH()
   {
	IsReal=true;
   }

   PETRIGRAPH::~PETRIGRAPH()
   {
   }

   PETRIGRAPH PETRIGRAPH::clone()
   {
      PETRIGRAPH pg = PETRIGRAPH();
      pg.iota = this->iota;
      pg.mu = this->mu;
      pg.grammar = this->grammar;
      return pg;
   }

   void PETRIGRAPH::addPlace(IDTYPE p, VERTEX v)
   {
      PTNET::addPlace(p, v);
      this->connectedEdges[p] = multiset<IDTYPE>();
   }
   
   void PETRIGRAPH::addPlace(IDTYPE p, EDGE e, list<IDTYPE> cV)
   {
      if(p > lastUsedID) lastUsedID = p;
      e.setDepth(maxDepth);
      HYPERGRAPH::addEdge(e, p, cV);
      presets[p] = vector<TCONNECTION>();
      postsets[p] = vector<TCONNECTION>();
      depth[p] = maxDepth;
   }
   
   void PETRIGRAPH::addPlace(IDTYPE p, EDGE e)
   {
      this->addPlace(p, e, list<IDTYPE>());
   }
   
   void PETRIGRAPH::deletePlace(IDTYPE p)
   {   
      if(this->vertices.count(p) > 0) this->deleteVertex(p);
      if(this->edges.count(p) > 0) this->deleteEdge(p);
   }
   
   void PETRIGRAPH::addVertex(VERTEX v, IDTYPE id)
   {
      this->addPlace(id, v);
   }
   
   void PETRIGRAPH::deleteVertex(IDTYPE id)
   {
      this->HYPERGRAPH::deleteVertex(id);
      this->PTNET::deletePlace(id);
   }
   
   void PETRIGRAPH::addEdge(EDGE e, IDTYPE id, list<IDTYPE> cV)
   {
      this->addPlace(id, e, cV);
   }
         
   void PETRIGRAPH::deleteEdge(IDTYPE id)
   {
      this->HYPERGRAPH::deleteEdge(id);
      this->PTNET::deletePlace(id);
   }

   MATCH PETRIGRAPH::getIota()
   {
      return this->iota;
   }
   
   void PETRIGRAPH::setIota(MATCH j)
   {
      this->iota = j;
   }
   
   map<IDTYPE, MATCH>* PETRIGRAPH::getMu()
   {
      return &(this->mu);
   }
   
   MATCH PETRIGRAPH::getMu(IDTYPE id)
   {
      assert(this->mu.count(id) > 0);
      return this->mu[id];
   }
   
   void PETRIGRAPH::setMu(map<IDTYPE, MATCH>* m)
   {
      this->mu = *m;
   }
   
   void PETRIGRAPH::setMu(IDTYPE id, MATCH m)
   {
      this->mu[id] = m;
   }
   
   GTS* PETRIGRAPH::getGrammar()
   {
      return this->grammar;
   }
   
   void PETRIGRAPH::setGrammar(GTS* gg)
   {
      this->grammar = gg;
   }
   
   void PETRIGRAPH::setAssociatedRule(IDTYPE t, IDTYPE r)
   {
      this->associated_rules[t] = r;
   }
   
   ostream& operator<<(ostream& ost, PETRIGRAPH& th)
   {
      ost << "===\npetrigraph with ID " << th.HYPERGRAPH::id << "\n\n1) hypergraph component\n"; 
      ost << (HYPERGRAPH&)th;
      ost << "\n2) Petri net component\n";
      ost << (PTNET&)th;
      ost << "\n===";
      return ost;
   }

   MARKING PETRIGRAPH::match2marking(MATCH m)
   {
	MARKING mark;
	
	map<IDTYPE, IDTYPE> match = m.getMatch();
	for(map<IDTYPE, IDTYPE>::iterator it = match.begin(); it != match.end(); it++)
	{
	        if(!nodes_as_places)
		 if(edges.count(it->second)==0)
		  continue;
		
		TOKEN t;
		if(mark.getMarking(it->second).size()==0)
			t.setCount(1);
		else
			t.setCount(mark.getMarking(it->second)[0].getCount()+1);
		
		vector<TOKEN> v_token;
		v_token.push_back(t);
		mark.setMarking(it->second, v_token);
		
	}
	
	return mark;
   }

   HYPERGRAPH PETRIGRAPH::getMarkedHypergraph(Marking m, map<IDTYPE, IDTYPE> * mergedEdges)
   {
    	 HYPERGRAPH hg;	 

	 list<IDTYPE> marked_edges;
	 
	 if(mergedEdges==NULL)
	 {
	  cout << "\npointer to NULL\n";
	  exit(1);
	 }
 
	 if(nodes_as_places)
	 {
	  cout << "\nPETRIGRAPH::getMarkedHypergraph is not implemented with nodes_as_places\n";
	  exit(1);
	 }

	 for(map< IDTYPE, unsigned >::iterator it=m.begin(); it!=m.end(); it++)
	 {
	 	  if(edges.count(it->first)==0)
		   continue;
		  assert(it->second!=0);
		  for(unsigned int i=0; i<it->second; i++)
			marked_edges.push_back(it->first);
	 }

	 for(list<IDTYPE>::iterator it_edges = marked_edges.begin(); it_edges!=marked_edges.end(); it_edges++)
	 {
		   list<IDTYPE> vertices = getConnectedVertices(*it_edges);
		   
		   for(list<IDTYPE>::iterator it_vertices = vertices.begin(); it_vertices != vertices.end(); it_vertices++)
		   {
		   	VERTEX v = GRAPH::vertices[*it_vertices];
			hg.addVertex(v, *it_vertices);
		   } 

		   EDGE e = GRAPH::edges[*it_edges];
		   IDTYPE edge = hg.addEdge2(e, *it_edges, vertices);
		   
                   (*mergedEdges)[edge] = (*it_edges);
	 }

	 return hg;
   }

   HYPERGRAPH PETRIGRAPH::getMarkedHypergraph(MARKING m, map<IDTYPE, IDTYPE> * mergedEdges)
   {
    	 HYPERGRAPH hg;	 

	 list<IDTYPE> marked_edges;
	 
	 if(mergedEdges==NULL)
	 {
	  cout << "\npointer to NULL\n";
	  exit(1);
	 }
 
	 if(nodes_as_places)
	 {
	  cout << "\nPETRIGRAPH::getMarkedHypergraph is not implemented with nodes_as_places\n";
	  exit(1);
	 }

	 map< IDTYPE, vector<TOKEN> > marking = m.getMarkings();

	 for(map< IDTYPE, vector<TOKEN> >::iterator it_edges = marking.begin(); it_edges!=marking.end(); it_edges++)
	 {
		   if(m.getTokenNumber(it_edges->first) <= 0)
			continue;

		   list<IDTYPE> vertices = getConnectedVertices(it_edges->first);
		   
		   for(list<IDTYPE>::iterator it_vertices = vertices.begin(); it_vertices != vertices.end(); it_vertices++)
		   {
		   	VERTEX v = GRAPH::vertices[*it_vertices];
			hg.addVertex(v, *it_vertices);
		   } 

		   for(vector<TOKEN>::iterator itt = it_edges->second.begin(); itt != it_edges->second.end(); itt++)
		   {
			for(unsigned int i = 0; i < itt->getCount(); i++)
			{
		   		EDGE e = GRAPH::edges[it_edges->first];
				e.setAttributes(itt->getAttributes());
		   		IDTYPE edge = hg.addEdge2(e, it_edges->first, vertices);

				(*mergedEdges)[edge] = it_edges->first;
			}
		   }
	 }

	 return hg;
   }

   void PETRIGRAPH::setAllValues(
		MATCH nIota,
		map<IDTYPE, MATCH> nMu,
		map<IDTYPE, IDTYPE> n_associated_rules,
		map< IDTYPE, list<IDTYPE> > ncV,
         	map< IDTYPE, multiset<IDTYPE> > ncE,
		map<IDTYPE, TRANSITION> nT,
		map< IDTYPE, vector<TCONNECTION> > nPresets,
         	map< IDTYPE, vector<TCONNECTION> > nPostsets,
		MARKING nInitialMarking,
		MARKING nCurrentMarking,
		map<IDTYPE, unsigned> nDepth,
		map<IDTYPE, VERTEX> nV,
         	map<IDTYPE, EDGE> nE,
		IDTYPE nLUI,
		GTS * grammar
   )
   {
	 iota = nIota;
	 mu = nMu;
	 associated_rules = n_associated_rules;
	 connectedVertices = ncV;
	 connectedEdges = ncE;
	 transitions = nT;
	 presets = nPresets;
	 postsets = nPostsets;
	 initial_marking = nInitialMarking;
	 current_marking = nCurrentMarking;
	 depth = nDepth;
	 vertices = nV;
	 edges = nE;
	 lastUsedID = nLUI;
	 this->grammar = grammar;	
   }

   IDTYPE PETRIGRAPH::getAssociatedRuleID(IDTYPE t)
   {
	return associated_rules[t];
   }

};


