
#ifndef HashTable_H
#define HashTable_H

#include <string>
#include <vector>

// QuadraticProbing Hash table class.

// ******************PUBLIC OPERATIONS*********************
// void insert( x )       --> Insert x
// void remove( x )       --> Remove x
// Object find( x )       --> Return item that matches x
// void makeEmpty( )      --> Remove all items
// ******************ERRORS********************************

template <class Object>
class HashTable
{
public:

  HashTable();

  int findPos(const Object &x);
  void makeEmpty();
  int insert(const Object &x);
  void remove(const Object & x);
  Object * find(const Object & x);
  int getTableSize() {return array.size();}
  int getActiveSize() const;
  enum EntryType {ACTIVE, EMPTY, DELETED};

 private:
  struct HashEntry
  {
    Object    element;
    EntryType info;

    HashEntry(const Object &e = Object(), EntryType i=EMPTY) : element(e), info(i) {}
  };

  int occupied;
  bool isActive(int currentPos) const;
  void rehash();
  bool isPrime(int n) const;
  int nextPrime(int n) const;


 protected:
  unsigned int hashFunction(const std::string & key) const;
  std::vector<HashEntry> array;
};



#include "HashTable.cpp"

#endif
