#ifndef _PLGTS_H
#define _PLGTS_H

#include "scenario.h"
#include "arithmetictree.h"
#include "booltree.h"
#include "symboltable.h"


class PLGTS {

public:

   GTS GraphGrammar;

private:
   int rule_id;
   int rule_counter;
   bool peqp_2rule;
   SymbolTable* symboltable;

   enum GRAPHHAND {LEFTHAND, RIGHTHAND};



   /*
     erstellt eine Reihe von Regeln die die Variable helpvar bis zur letzten Pfadkomponente fuehren
     es wird eine Error-Regel erstellt
     counter:  Programmcounter
     path:     Pfad
     halpvar:  Hilfsvariable, die schließlich auf die letzte Pfadkomponente zeigt
     rulename: Name der Regel
   */

   list<RULE*> gotoLastPathRule(int counter, list <char*> path, char *helpvar, char *rulename, char **returnrecordname);
   list <RULE*>gotoLastPathRule(int counter, int helpcounter, list <char*> path, char *helpvar, char *rulename, int mode, int &returnhelpcounter, char **returnrecordname);
   list <RULE*> gotoLastPathRule(int counter, int helpcounter, list <char*> path, char *helpvar, char *rulename, int mode, int &returnhelpcounter, int &returntype);

   list <RULE*> gotoLastPathRule(int counter, list <char*> path, char *helpvar, char *rulename);
   list <RULE*> gotoLastPathRule(int counter, int helpcounter, list <char*> path, char *helpvar, char *rulename, int mode, int &returnhelpcounter);
   /*
     erstellt eine ganze Reihe von Regeln die die Variable helpvar bis zur vorletzten Pfadkomponente führen
     es wird auch eine Error-Regel erstellt.
     counter:  Programmcounter
     path:     Pfad
     halpvar:  Hilfsvariable, die schlißlich auf die vorletzte Pfadkomponente zeigt
     rulename: Name der Regel
     returnrecordname:  Name von Record auf den die vorletzte Pfadkomponente zeigt 
   */
   list <RULE*> gotoSecondLastPathRule(int counter, list <char*> path, char *helpvar, char *rulename, char **returnrecordname);
   list <RULE*>gotoSecondLastPathRule(int counter, int helpcounter, list <char*> path_, char *helpvar, char *rulename, int mode, int &returnhelpcounter, char **returnrecordname);


   /* 
     erstellt eine neue Record-Hyperkante und gibt ihre id zurueck (return)

     hg:         Hypergraph in den die neue Record-Hyperkante eingefuegt werden soll
     graphhand:  linke oder rechte Seite der Regel
     id:         Startwert: Start id für neue graphische Elemente
                 Rückgabewert: letzte vergebene id + 1
     recordname: Name vom neuen Record
     vertexid:   id von Vertex an dem die neue Hyperkante angehaengt werden soll
     nextname:
     vertexnext: id von Vertex das mit der Recordkomponente nextname assoziert wird
   */
   int produceRecord(HYPERGRAPH *hg, GRAPHHAND graphhand, int &id, char *recordname, int vertexid, char *nextname, int &vertexnext);
   int produceRecord(HYPERGRAPH *hg, GRAPHHAND graphhand, int &id, char *recordname, int vertexid);


   /* 
     erstellt eine neue Record-Hyperkante und gibt ihre id zurueck (return)
     Dabei werden die Record-Komponente ensprechend initialisiert (null, integer, string)

     hg:         Hypergraph in den die neue Record-Hyperkante eingefuegt werden soll
     graphhand:  linke oder rechte Seite der Regel
     id:         Startwert: Start id für neue graphische Elemente
                 Rückgabewert: letzte vergebene id + 1
     recordname: Name vom neuen Record
     vertexid:   id von Vertex an dem die neue Hyperkante angehaengt werden soll
   */
   int produceRecordInit(HYPERGRAPH *hg, GRAPHHAND graphhand, int &id, char *recordname, int vertexid);


   /*
     counterleft -> counterright
     rulename: Name der Regel
   */
   void produceDummyRule(int counter, char *rulename);
   void produceDummyRule(int counterleft, int counterright, char *rulename);


   /*
     erstellt eine Error-Regel
     counter:  Programmzaehler
     varname:  Name der Variable mit der die Error-Hyperkante konstruiert werden soll 
     rulename: Name der Regel
   
     ---------           --------
     [varname] ----o----[ error ]
     ---------           --------
   */
   RULE *produceError(int counter, char *varname, char *rulename);
   RULE *produceError(char *strcounter, char *varname, char *rulename);


   /* 
     erstellt allgemeine Regel für die belibiegen Pointerzuweisungen 
     dabei werden die möglichen Zyklen in Pfaden nicht beruecksichtigt
     counter:   Programmzaehler
     lefthand:  linke Seite der Zuweisung
     righthand: rechte Seite der Zuweisung
  */
   RULE* producePointerToPointerRule(int counterleft, int counterright, list <char*> lefthand, list <char*> righthand, char *rulename);


   /*
     erstellt allgemeine Error-Regeln für die belibiegen Pfaden
     dabei werden die möglichen Zyklen in Pfaden nicht beruecksichtigt
     counter:   Programmzaehler
     path:      Pfad
   */
   list <RULE*> produceErrorRule(int counter, list <char*> path);


   /*
     erstellt eine Integer- oder String-Variable 
     hg:         Hypergraph in den die neue Record-Hyperkante eingefuegt werden soll
     graphhand:  linke oder rechte Seite der Regel
     id:         Startwert: Start id für neue graphische Elemente
                 Rückgabewert: letzte vergebene id + 1
     varname:    Name der Variable
     attrvalue:  Attribut (Wert der Variable)
     type:       Typ der Variable (integer, string)

                         --------
     ---------          [  type  ]
     [varname] ----o----[--------]
     ---------          [atrvalue]
                         --------
   */
   int produceVar(HYPERGRAPH *hg, GRAPHHAND graphhand, int &id, char *varname, char *attrvalue, char *type);


   /*
     erstellt Pointer.Recordcomp 
     hg:         Hypergraph in den die neue Record-Hyperkante eingefuegt werden soll
     graphhand:  linke oder rechte Seite der Regel
     id:         Startwert: Start id für neue graphische Elemente
                 Rückgabewert: letzte vergebene id + 1
     varname:    Name der Variable
     attrvalue:  Attribut (Wert der Variable)
     recordname: Name des Records
     type:       Typ der Variable (integer, string)
   
                         ----------  ---o         --------
     ---------          [          ]- recordcomp [  type  ]
     [varname] ----o----[recordname]----o--------[--------]
     ---------          [          ]-            [atrvalue]
                         ----------- ---o         --------
  */

   int producePVar(HYPERGRAPH *hg, GRAPHHAND graphhand, int &id, char *varname, char *recordcomp, char *recordname, char *attrvalue, char *type);



   /*
     strcounter_l: varnameleft = varnameright strcounter_r:

     strcounter_l: Programmcounter auf der linken Seite
     strcounter_r: Programmcounter auf der rechten Seite
     varnameleft,
     varnameright:   Pointer
     rulename:       Name der Regel
     deletevarleft, 
     deletevarright: legt fest, ob varnameleft bzw varnameright auf der rechten Seite der Regel erstellt wird
   */
   RULE *producePToPRule(char *strcounter_l, char *strcounter_r, char *varnameleft, char *varnameright, char *rulename, bool deletevarleft=false, bool deletevarright=false);


   /*
     strcounter_l: varname = varname.recordcomp; strcounter_r:

     varname:      Pointer auf Record recordname
     strcounter_l: Programmcounter auf der linken Seite
     strcounter_r: Programmcounter auf der rechten Seite
     rulename:     Name der Regel
   */
   RULE *producePToPcompRule(char *strcounter_l, char *strcounter_r, char *varname, char *recordcomp, char *recordname, char *rulename, bool deletevarright=false);


   /*
     strcounter_l: varleft = varright.recordcomp strcounter_r:

     varname:      Pointer auf Record recordname
     strcounter_l: Programmcounter auf der linken Seite
     strcounter_r: Programmcounter auf der rechten Seite
     rulename:     Name der Regel
   */
   RULE *producePiToPjcompRule(char *strcounter_l, char *strcounter_r, char *varleft, char *varright, char *recordcomp, char *recordname, char *rulename, bool deletepi=false, bool deletepj=false);


   /*
     strcounter_l: varname.recordcomp = varname strcounter_r:

     strcounter_l: Programmcounter auf der linken Seite
     strcounter_r: Programmcounter auf der rechten Seite
     varname:      Pointer auf Record recordname
     rulename:     Name der Regel
   */
   RULE *producePcompToPRule(char *strcounter_l, char *strcounter_r, char *varname, char *recordcomp, char *recordname, char *rulename, bool deletevarright=false);


   /*
     strcounter_l: varleft.recordcomp = varright strcounter_r:

     strcounter_l: Programmcounter auf der linken Seite
     strcounter_r: Programmcounter auf der rechten Seite
     varleft:      Pointer auf Record recordname
     rulename:     Name der Regel
     deletepi, 
     deletepj:     legt fest, ob varnameleft bzw. varnameright auf der rechten Seite der Regel erstellt wird
   */
   RULE *producePicompToPjRule(char *strcounter_l, char *strcounter_r, char *varnameleft, char *recordcomp, char *recordname, char *varnameright, char *rulename,bool deletepi=false, bool deletepj=false);


   /*
     strcounter_l: NEW(varname.recordcomp); strcounter_r:

     strcounter_l: Programmcounter auf der linken Seite
     strcounter_r: Programmcounter auf der rechten Seite
     varname:      Pointer auf Record recordname (erste Pfad-Komponente)
     recordcomp:   Zweite Pfad-Komponente
     rulename:     Name der Regel
     deletvar:     legt fest, ob varname auf der rechten Seite der Regel erstellt wird
   */
   RULE *produceNEWPcompRule(char *strcounter_l, char *strcounter_r, char *varname, char *recordcomp, char *recordname, char *rulename, bool deletevar=false);


   /*
     strcounter_l: DELETE(varname); strcounter_r:

     strcounter_l: Programmcounter auf der linken Seite
     strcounter_r: Programmcounter auf der rechten Seite
     varname:      Pointer auf Record recordname
     rulename:     Name der Regel
     deletvar:     legt fest, ob varname auf der rechten Seite der Regel erstellt wird
   */
   list<RULE*> produceDELETE_P_Rule(char *strcounter_l, char *strcounter_r, char *varname, char *recordname, char *rulename, bool deletevar=false);


   /*
     strcounter_l: DELETE(varname.recordcomp); strcounter_r:

     strcounter_l: Programmcounter auf der linken Seite
     strcounter_r: Programmcounter auf der rechten Seite
     varname:      Pointer auf Record recordname
     rulename:     Name der Regel
     deletvar:     legt fest, ob varname auf der rechten Seite der Regel erstellt wird
   */
   list<RULE*> produceDELETE_PComp_Rule(char *strcounter_l, char *strcounter_r, char *varname, char *recordcomp, char *recordname, char *rulename, bool deletevar=false);


   /*
     strcounter_l: varname.recordcomp:=NULL strcounter_r:;

     strcounter_l: Programmcounter auf der linken Seite
     strcounter_r: Programmcounter auf der rechten Seite
     varname:      Pointer auf Record recordname (erste Pfad-Komponente)
     recordcomp:   Zweite Pfad-Komponente
     rulename:     Name der Regel
     deletvar:     legt fest, ob varname auf der rechten Seite der Regel erstellt wird
   */
   RULE *produceNULLPcompRule(char *strcounter_l, char *strcounter_r, char *varname, char *recordcomp, char *recordname, char *rulename, bool deletvar=false);


   /*
     strcounter_IF: IF (left=right) strcounter_THEN: THEN S1 strcounter_ELSE: ELSE S2 FI;

     strcounter_IF, strcounter_THEN, strcounter_ELSE: Programmcounter
     left, right: Pointer
     rulename:    Name der Regel 
     deletevarleft:
     deletevarright: legt fest, ob left bzw right auf der rechten Seite der Regeln erstellt werden
   */
   //5 Regeln 
   list <RULE*> producePEqPRule(char *strcounter_IF, char* strcounter_THEN, char *strcounter_ELSE, char *left, char *right, char *recordname, char *rulename, bool deletevarleft=false, bool deletevarright=false);
   //2 Regeln
   list <RULE*> producePEqPRule_2Rules(char *strcounter_IF, char* strcounter_THEN, char *strcounter_ELSE, char *left, char *right, char *rulename, bool deletevarleft=false, bool deletevarright=false);


   /*
      strcounter_IF: IF (pointer=NULL) strcounter_THEN: THEN S1 strcounter_ELSE: ELSE S2 FI;

      strcounter_IF, strcounter_THEN, strcounter_ELSE: Programmcounter
      pointer:   Pointer
      rulename:  Name der Regel
   */
   list<RULE*>  producePEqNULL(char *strcounter_IF, char *strcounter_THEN, char *strcounter_ELSE, char *pointer, char *recordname, char *rulename, bool deletevar=false);

char *addRuleID(char *rulename);
void replaceSemicolon(char *str);

public:


   /* Konstruktor */
   /* erstellt auch einen Startgraph */
   /* counter: Label der Starthyperkante */
   /* Standard counter = 1 */
   PLGTS(SymbolTable* symboltable);
   PLGTS(SymbolTable* symboltable, int counter);
   PLGTS(SymbolTable* symboltable, char *strcounter); 

   ~PLGTS();

   //Startregel
   //void produceInitGraph(int &counter);

   /* -------- Regeln ------- */

   /* -- Deklarationen -- */

   /* Integer-Deklaration */
   void produceIntDeclRule(int counter, char *varname, char *rulename);
   void produceIntDeclRule(int counterleft, int counterright, char *varname, char *rulename);

   /* String-Deklaration */
   void produceStrDeclRule(int counter, char *varname, char *rulename);
   void produceStrDeclRule(int counterleft, int counterright, char *varname, char *rulename);

   /* Pointer-Deklaration */
   void producePtoRecDeclRule(int counter, char *varname, char *rulename);
   void producePtoRecDeclRule(int counterleft, int counterright, char *varname, char *rulename);

   /* Type-Deklaration */
   void produceTypeDeclRule(int counter, char *rulename);
   void produceTypeDeclRule(int counterleft, int counterright, char *rulename);


   /* -- Anweisungen -- */

   /* SKIP */
   void produceSKIPRule(int counterleft, int counterright, char *rulename);

   /* NEW */
   void produceNEWRule(int counterleft, int counterright, list<char*> path, char *rulename);

   /* DELETE */
   void produceDELETERule(int counterleft, int counterright, list<char*> path, char *rulename);

   /* Pointer-Zuweisung */
   void producePointerAssignerRule(int counterleft, int counterright, list <char*> left, list <char*> right, char *rulename);

   /* Pointer-NULL-Zuweisung */
   void producePointerAssignerNULLRule(int counterleft, int counterright, list<char*> path, char *rulename);

   /* String-Const-Zuweisung*/
   void produceStringAssignerRule(int counterleft, int counterright, list<char*> left, list <char*> right, char *rulename);
   /* String-Zuweisung*/
   void produceStringAssignerConstRule(int counterleft, int counterright, list<char*> left, char *strconst, char *rulename);

   /* Integer-Zuweisung */
   void produceIntegerAssignerRule(int counterleft, int counterright, list <char*> left, arithmetictree::Node *arithmetictree , char *rulename);

   /* IF-Anweisung */
      // ohne Pointervergleich
   void produceIFRule(int counterIF, int counterELSE, booltree::Node *booltree , char *rulename);
      // linke und rechte sind pointers
   void produceIFRule(int counterIF, int counterELSE, list<char*> left, list <char*> right, char *rulename, bool equal=true);
      // linke oder rechte Seite ist NULL
   void produceIFNULLRule(int counterIF, int counterELSE, list<char*> path, char *rulename, bool equal=true);

   /* WHILE-Schleife */
      // ohne Pointervergleich
   void produceWHILERule(int counterWHILE, int counterOD, booltree::Node *booltree , char *rulename);
      // linke und rechte Seite sind Pointers
   void produceWHILERule(int counterWHILE, int counterOD, list<char*> left, list<char*> right, char *rulename, bool equal=true);
      // linke oder rechte Seite ist NULL
   void produceWHILENULLRule(int counterWHILE, int counterOD, list<char*> path, char *rulename, bool equal=true);

   /* schreibt XML File   */
   /* input: GraphGrammar */
   /* output: xml file    */
   void WriteXML(char *filename="plgraph.xml");

   /* Gibt den Graph zurueck */
   GTS getGraph() {return GraphGrammar;}

   /*
     pep_2rule_=true:   producePEqPRule_2Rules
                        zwei Regeln für den Pointer-Vergleich
     pep_2rule_=false:  producePEqPRule
                        fünf Regeln für den Pointer-Vergleich
   */
   void setPeqPOption(bool peqp_2rule_);



};


#endif

