
#include "abstree.h"
#include "helpfunction.h"
#include "symboltable.h"

#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <iostream>



using namespace std;
int error;

void buildTypeDeclSymbolTable(SymbolTable*, NODE*);
void buildVarDeclSymbolTable(SymbolTable *, NODE*);
void buildTypeDeclListSymbolTable(SymbolTable *, NODE*, char *); 


void buildDeclSymbolTable(SymbolTable*, NODE*);
void buildBodySymbolTable(SymbolTable*, NODE*);

SymbolEntry* controlIdentifier(SymbolTable *, NODE*);
RecordEntry* controlPath (SymbolTable *, NODE*);
void controlType(SymbolTable *, NODE*);
int controlBEXP_(SymbolTable *symboltable, NODE *node, int counter);


RecordEntry* getNPathCompStart(SymbolTable*, NODE* , char*, int);
RecordEntry* getLastPathComp_(SymbolTable*, NODE*, char *recordname);

int getNPath_(SymbolTable* symboltable, NODE* node, int length, char **varname, char **recordname);

/* 
  void pretty_print(const NODE* node)
  gibt den Parserbaum abstree aus 
*/
void pretty_print(const NODE* node){

   if (node==NULL) {
      //printf("NULLPOINTER");
      return;
   }
   int type=node->type;

   switch(type) {

	case T_VARDECL: {
		VARDECL* vardecl=(VARDECL*)node->pointer;
		printf("VAR %s:",vardecl->name);
		pretty_print(vardecl->type);
		break;
	}
	case T_DECLLIST: {
		DECLLIST* decllist=(DECLLIST*)node->pointer;
		pretty_print(decllist->head);
		printf(";\n");
		pretty_print(decllist->tail);
		break;
	}
	case T_TYPEDECL: {
		TYPEDECL* typedecl=(TYPEDECL*)node->pointer;
		printf("TYPE %s IS RECORD(",typedecl->name);
		pretty_print(typedecl->head);
                if (typedecl->tail != NULL) {
		   printf(",");
		   pretty_print(typedecl->tail);
                }
		printf(")");
		break;
	}
        case T_TYPEDECLNULL: {
                TYPEDECLNULL* typedeclnull=(TYPEDECLNULL*)node->pointer;
                printf("TYPE %s", typedeclnull->name);
                break;
        }
	case T_TYPEDECLLIST: {
		TYPEDECLLIST* typedecllist=(TYPEDECLLIST*)node->pointer;
		pretty_print(typedecllist->head);
                if (typedecllist->tail != NULL) {
		   printf(",");
		   pretty_print(typedecllist->tail);
                }
		break;
	}
	case T_TVARDECL: {
		TVARDECL* tvardecl=(TVARDECL*)node->pointer;
		printf("%s:", tvardecl->name);
		pretty_print(tvardecl->type);
		break;
	}

	case T_INTEGER: {
		printf("INTEGER");
		break;
	}
	case T_STRING: {
		printf("STRING");
		break;
	}
	case T_PTOREC: {
		printf("POINTER TO RECORD ");
		PTOREC* ptorec=(PTOREC*)node->pointer;
		printf("%s",ptorec->name);
		break;
	}
	case T_PATH: {
		PATH* path=(PATH*)node->pointer;
		pretty_print(path->first);
		printf(".");
		pretty_print(path->last);
		break;
	}
	case T_PROGRAM: {
		PROGRAM* stmlist=(PROGRAM*)node->pointer;
		pretty_print(stmlist->declarations);
		pretty_print(stmlist->body);
		break;
	}
	case T_STMLIST: {
		STMLIST* stmlist=(STMLIST*)node->pointer;
		pretty_print(stmlist->head);
		printf(";\n");
		pretty_print(stmlist->tail);
		break;
	}
	case T_INTEGERCONST: {
	   	INTEGERCONST* intcons=(INTEGERCONST*)node->pointer;
		printf("%d",intcons->value);
	   	break;
	}
	case T_STRCONST: {
		STRCONST* strconst=(STRCONST*)node->pointer;
		printf("%s",strconst->value);
		break;
	}
	case T_ASSIGNER: {
		ASSIGNER* assign=(ASSIGNER*)node->pointer;
		pretty_print(assign->left);
		printf(STR_ASSIGNER);
		pretty_print(assign->right);
	   	break;
	}
	case T_PLUSOP: {
		PLUSOP* plusop=(PLUSOP*)node->pointer;
		//printf("(");
		pretty_print(plusop->left);
		printf(STR_ADD);
		pretty_print(plusop->right);
		//printf(")");
	   	break;
	}
	case T_MINUSOP: {
		MINUSOP* minusop=(MINUSOP*)node->pointer;
		//printf("(");
		pretty_print(minusop->left);
		printf(STR_MIN);
		pretty_print(minusop->right);
		//printf(")");
	   	break;
	}
	case T_MULTOP: {
		MULTOP* multop=(MULTOP*)node->pointer;
		//printf("(");
		pretty_print(multop->left);
		printf(STR_MULT);
		pretty_print(multop->right);
		//printf(")");
	   	break;
	}
	case T_DIVOP: {
		DIVOP* divop=(DIVOP*)node->pointer;
		pretty_print(divop->left);
		printf(STR_DIV);
		pretty_print(divop->right);
	   	break;
	}
	case T_MODOP: {
		MODOP* modop=(MODOP*)node->pointer;
		pretty_print(modop->left);
		printf(STR_MOD);
		pretty_print(modop->right);
	   	break;
	}
	case T_NEGATE: {
		//printf("(");
		NEGATE* negate=(NEGATE*)node->pointer;
		printf(STR_UMINUS);
		pretty_print(negate->expstm);
		//printf(")");
		break;
	}
	case T_IDENTIFIER: {
	   	IDENTIFIER* identi=(IDENTIFIER*)node->pointer;
		printf("%s",identi->name);
	   	break;
	}
	case T_NEW: {
		ALLOC* alloc=(ALLOC*)node->pointer;
		printf("NEW(");
		pretty_print(alloc->path);
		printf(")");
		break;
	}
	case T_DELETE: {
		DELETE_* del=(DELETE_*)node->pointer;
		printf("DELETE_(");
		pretty_print(del->path);
		printf(")");
		break;
	}
	case T_SKIP: {
		printf("SKIP");
		break;
	}
	case T_IFSTM: {
		IFSTM* ifstm=(IFSTM*)node->pointer;
		printf("IF (");
		pretty_print(ifstm->bexp);
		printf(") THEN ");
		pretty_print(ifstm->ifstm);
		printf("ELSE ");
		pretty_print(ifstm->elsestm);
		printf("FI");
		break;
	}
	case T_WHILE: {
		WHILESTM* whilestm=(WHILESTM*)node->pointer;
		printf("WHILE (");
		pretty_print(whilestm->bexp);
		printf(") DO ");
		pretty_print(whilestm->dostm);
		printf("OD");
		break;
	}
	case T_EQUAL: {
		EQUAL* equal=(EQUAL*)node->pointer;
		pretty_print(equal->left);
		printf(STR_EQUAL);
		pretty_print(equal->right);
	   	break;
	}
	case T_NOTEQUAL: {
		NOTEQUAL* notequal=(NOTEQUAL*)node->pointer;
		pretty_print(notequal->left);
		printf(STR_NOTEQUAL);
		pretty_print(notequal->right);
	   	break;
	}
	case T_SMALLER: {
		SMALLER* smaller=(SMALLER*)node->pointer;
		pretty_print(smaller->left);
		printf(STR_SMALLER);
		pretty_print(smaller->right);
	   	break;
	}
	case T_GREATER: {
		GREATER* greater=(GREATER*)node->pointer;
		pretty_print(greater->left);
		printf(STR_GREATER);
		pretty_print(greater->right);
	   	break;
	}
	case T_GREATEREQUAL: {
		GREATEREQUAL* greaterequal=(GREATEREQUAL*)node->pointer;
		pretty_print(greaterequal->left);
		printf(STR_GREATEREQUAL);
		pretty_print(greaterequal->right);
	   	break;
	}
	case T_SMALLEREQUAL: {
		SMALLEREQUAL* smallerequal=(SMALLEREQUAL*)node->pointer;
		pretty_print(smallerequal->left);
		printf(STR_SMALLEREQUAL);
		pretty_print(smallerequal->right);
	   	break;
	}
	case T_AND: {
		AND_* and_=(AND_*)node->pointer;
		//printf("(");
 		pretty_print(and_->left);
		printf(STR_AND);
		pretty_print(and_->right);
		//printf(")");
	   	break;
	}
	case T_OR: {
		OR_* or_=(OR_*)node->pointer;
		//printf("(");
		pretty_print(or_->left);
		printf(STR_OR);
		pretty_print(or_->right);

	   	break;
	}
	case T_NOT: {
		NOT_* not_=(NOT_*)node->pointer;
                printf(STR_NOT);
		printf("(");
		pretty_print(not_->left);
		printf(")");
	   	break;
	}
	case T_TRUE: {
		printf("TRUE");
		break;
	}
	case T_FALSE: {
		printf("FALSE");
		break;
	}
	case T_NULL: {
		printf("NULL");
		break;
	}
	}
}





/* 
  void printNodeType(NODE* node)
  gibt den Type des Knoten aus
*/
void printNodeType(const NODE* node)
{
   assert(node != NULL);
   switch (node->type) {

   	case T_PROGRAM: {
           printf("PROGRAM");
           break;
        }
        case T_DECLLIST: {
           printf("DECLLIST");
           break;
        }
        case T_VARDECL: {
           printf("VARDECL");
           break;
        }
        case T_TYPEDECL: {
           printf("TYPEDECL");
           break;
        }
        case T_TYPEDECLNULL: {
           printf("TYPEDECLNULL");
           break;
        }
        case T_TVARDECL: {
           printf("TVARDECL");
           break;
        }
        case T_TYPEDECLLIST: {
           printf("TYPEDECLLIST");
           break;
        }
        case T_STMLIST: {
           printf("STMLIST");
           break;
        }
        case T_VARINT: {
           printf("VARINT");
           break;
        }
        case T_VARSTR: {
           printf("VARSTR");
           break;
        }
        case T_VARPOINT: {
           printf("VARPOINT");
           break;
        }
        case T_INTEGER: {
	   printf("INTEGER");
           break;
        }
        case T_STRING: {
           printf("STRING");
           break;
        }
        case T_PTOREC: {
           printf("PTOREC");
           break;
        }
        case T_IDENTIFIER: {
           printf("IDENTIFIER");
           break;
        }
        case T_NEGATE: {
           printf("NEGATE");
           break;
        }
	case T_INTEGERCONST: {
           printf("INTEGERCONST");
           break;
        }
	case T_STRCONST: {
           printf("STRCONST");
           break;
        }
        case T_ASSIGNER: {
           printf("ASSIGNER");
           break;
        }
	case T_PLUSOP: {
           printf("PLUSOP");
           break;
        }
	case T_MINUSOP: {
           printf("MINUSOP");
           break;
        }
	case T_MULTOP: {
           printf("MULTOP");
           break;
        }
	case T_DIVOP: {
           printf("DIVOP");
           break;
        }
	case T_MODOP: {
           printf("MODOP");
           break;
        }
	case T_PATH: {
           printf("PATH");
           break;
        }
	case T_NEW: {
           printf("NEW");
           break;
        }
	case T_DELETE: {
           printf("DELETE_");
           break;
        }
	case T_SKIP: {
           printf("SKIP");
           break;
        }
	case T_EQUAL: {
           printf("EQUAL");
           break;
	}
	case T_SMALLER: {
           printf("SMALLER");
           break;
        }
	case T_GREATER: {
           printf("GREATER");
           break;
        }
	case T_SMALLEREQUAL: {
           printf("SMALLEREQUAL");
           break;
        }
	case T_GREATEREQUAL: {
           printf("GREATEREQUAL");
           break;
        }
	case T_NOTEQUAL: {
           printf("NOTEQUAL");
           break;
        }

	case T_AND: {
	   printf("AND");
	   break;
	}
	case T_OR: {
	   printf("OR");
	   break;
	}
	case T_NOT: {
	   printf("NOT");
	   break;
	}
	case T_IFSTM: {
	   printf("IFSTM");
	   break;
	}
	case T_WHILE: {
	   printf("WHILE");
	   break;
	}
	case T_TRUE: {
           printf("TRUE");
	   break;
	}
	case T_FALSE: {
           printf("FALSE");
	   break;
	}
	case T_NULL: {
           printf("NULL");
	   break;
	}
	default: {
	   printf("unexpected Node");
           assert(false);
	}
   }
}




/*
  int buildSymbolTable (SymbolTable *symboltable,  const NODE *node)
  Aufbau der Symboltabelle, Typueberpruefung
*/
int buildSymbolTable(SymbolTable *symboltable,  const NODE *node)
{

   if (symboltable==NULL)
      return -1;

   if (node == NULL)
      return -1;

   int type=node->type;
   int programcounter=0;
   error=0;

   if (type != T_PROGRAM)
      return -1;

   PROGRAM* program=(PROGRAM*)node->pointer;
   buildDeclSymbolTable(symboltable,program->declarations);
   if (error==0) {
      printf("Symboltable built\n");
      buildBodySymbolTable(symboltable,program->body);
   }
   if (error==0) {
      printf("checking type\n");
      controlType(symboltable,program->body);
   }
   return error;
}




/*
   void buildDeclSymbolTable(SymbolTable *symboltable,  NODE* node)
*/
void buildDeclSymbolTable(SymbolTable *symboltable,  NODE* node)
{
   if (symboltable == NULL)
      return;
   if (node==NULL)
      return;

   int type=node->type;

   switch(type) {

        case T_DECLLIST: {
		DECLLIST* decllist=(DECLLIST*)node->pointer;
		buildDeclSymbolTable(symboltable,decllist->head);
		buildDeclSymbolTable(symboltable,decllist->tail);
                break;
        }
	case T_VARDECL: {
                buildVarDeclSymbolTable(symboltable,node);
		break;
	}
        case T_TYPEDECL: {
		buildTypeDeclSymbolTable(symboltable,node);
		break;
	}
        case T_TYPEDECLNULL: {
		buildTypeDeclSymbolTable(symboltable,node);
		break;
	}
	default: return;
   }
}




/*
   void buildVarDeclSymbolTable(SymbolTable *symboltable,  NODE* node)
*/
void buildVarDeclSymbolTable(SymbolTable *symboltable,  NODE* node)
{

   if (symboltable==NULL)
       return;
   if (node == NULL)
       return;
   int type=node->type;

   VARDECL* vardecl=(VARDECL*)node->pointer;
   switch (vardecl->type->type) {
   	case T_INTEGER: {
		if (putIntegerSymbol(symboltable, vardecl->name)==0) {
                   printf ("error in line %d\n", vardecl->programcounter);
		   printf ("variable %s already exists\n", vardecl->name);
		   error=1;
		}
	        break;
	}
	case T_STRING: {
		if (putStringSymbol(symboltable, vardecl->name)==0) {
                   printf ("error in line %d\n", vardecl->programcounter);
		   printf ("variable %s already exists\n", vardecl->name);	
		   error=1;
		}
		break;
	}
	case T_PTOREC: {
                PTOREC* ptorec=(PTOREC*)vardecl->type->pointer;
		if (ptorec==NULL)
		   return;

                int msg=putPointerSymbol(symboltable, vardecl->name, ptorec->name);
                if (msg==0) {
                   printf ("error in line %d\n", vardecl->programcounter);
		   printf ("variable %s already exists\n", vardecl->name);	
		   error=1;
                }
		else if (msg==-2) {
                   printf ("error in line %d\n", ptorec->programcounter);
		   printf ("%s not defined\n" ,ptorec->name);
		   error=1;
		}
		break;	
		
	}
        default: return;
   }

}




/*
   void buildTypeDeclSymbolTable(SymbolTable *symboltable,  NODE* node)
*/
void buildTypeDeclSymbolTable(SymbolTable *symboltable,  NODE* node)
{

   if (symboltable==NULL)
       return;
   if (node == NULL)
       return;

   int type=node->type;
   assert(type == T_TYPEDECL || type==T_TYPEDECLNULL);
   if (type == T_TYPEDECLNULL) {
      TYPEDECLNULL* typedeclnull=(TYPEDECLNULL*)node->pointer;
      char *typename_=typedeclnull->name;
      int i=putRecordSymbol(symboltable, typename_);
      if (i==0) {
         printf ("error in line %d\n", typedeclnull->programcounter);
         printf ("TYPE %s already exists\n", typename_);
         error=1;
         return;	
      }
      return;
   }

   TYPEDECL* typedecl=(TYPEDECL*)node->pointer;
   char *typename_=typedecl->name;
   int i=putRecordSymbol(symboltable, typename_);

   if (i==0) {
      SymbolEntry *se=getSymbol(symboltable, typename_);
      if ((se->typref) != NULL) {
         printf ("error in line %d\n", typedecl->programcounter);
         printf ("TYPE %s already exists\n", typename_);
         error=1;
         return;
      }
   }   
   
   if (typedecl->head != NULL) {
      if (typedecl->head->type == T_TVARDECL) {
         TVARDECL* tvardecl=(TVARDECL*)typedecl->head->pointer;
         if (tvardecl->type !=NULL) {
            switch(tvardecl->type->type) {
            	case T_INTEGER: {
                        if (addIntRecordComp(symboltable,tvardecl->name ,typename_) ==0) {
                   	   printf ("error in line %d\n", tvardecl->programcounter);
     			   printf ("component %s of record %s already exists\n", tvardecl->name, typename_);
		           error=1;	
                        }
			break;
		}
		case T_STRING: {
                        if (addStrRecordComp(symboltable,tvardecl->name ,typename_) ==0) {
                   	   printf ("error in line %d\n", tvardecl->programcounter);
     			   printf ("component %s of record %s already exists\n", tvardecl->name, typename_);	
			   error=1;
                        }
			break;
		}
		case T_PTOREC: {
			PTOREC* ptorec=(PTOREC*)tvardecl->type->pointer;
			if (ptorec==NULL)
			   return;
			int i=addPointerRecordComp(symboltable,tvardecl->name,ptorec->name,typename_);
			if (i==0) {
                           printf ("error in line %d\n", tvardecl->programcounter);
     			   printf ("component %s of record %s already exists\n", tvardecl->name, typename_);
			   error=1;
			}
			else if (i==-3) {
                           printf ("error in line %d\n", tvardecl->programcounter);
     			   printf ("%s in record %s undefined\n", ptorec->name, typename_);
			   error=1;	
			}
			
			break;
		}
            }
         }
         
      }
   }
   if (typedecl->tail != NULL){
      buildTypeDeclListSymbolTable(symboltable,typedecl->tail,typename_);
   }
}





/*
   void buildTypeDeclListSymbolTable(SymbolTable *symboltable,  NODE* node, char *typename_) 
*/
void buildTypeDeclListSymbolTable(SymbolTable *symboltable,  NODE* node, char *typename_) 
{
      if (symboltable==NULL)
         return;
      if (node == NULL)
         return;
      int type=node->type;

      switch (type) {

   	 case T_TYPEDECLLIST: {
	     TYPEDECLLIST* typedecllist=(TYPEDECLLIST*)node->pointer;
	     if (typedecllist->head != NULL)
               buildTypeDeclListSymbolTable(symboltable,typedecllist->head,typename_);
             if (typedecllist->tail != NULL)
                buildTypeDeclListSymbolTable(symboltable,typedecllist->tail,typename_);
             break;
	 }
         case T_TVARDECL: {
	     TVARDECL* tvardecl=(TVARDECL*)node->pointer;
            if (tvardecl->type !=NULL) {
               switch(tvardecl->type->type) {
            	  case T_INTEGER: {
  		     int i=addIntRecordComp(symboltable,tvardecl->name ,typename_);
		     if (i==0) {
		        printf ("error in line %d\n", tvardecl->programcounter);
     			printf ("component %s of record %s already exists\n", tvardecl->name, typename_);
			error=1;
		     }
		     break;
		  }
		  case T_STRING: {
		     int i=addStrRecordComp(symboltable,tvardecl->name ,typename_);
		     if (i==0) {
			printf ("error in line %d\n", tvardecl->programcounter);
     			printf ("component %s of record %s already exists\n", tvardecl->name, typename_);
		        error=1;	
		     }
		     break;
		   }
		   case T_PTOREC: {
		       PTOREC* ptorec=(PTOREC*)tvardecl->type->pointer;
		      if (ptorec==NULL)
		         return;
		      int i= addPointerRecordComp(symboltable,tvardecl->name, ptorec->name, typename_);
		      if (i==0) {
		         printf ("error in line %d\n", tvardecl->programcounter);
     			 printf ("component %s of record %s already exists\n", tvardecl->name, typename_);	
		   	 error=1;				   
		      }
		      else if (i==-3) {
			 printf ("error in line %d\n", ptorec->programcounter);
     			 printf ("%s in record %s undefined\n", ptorec->name, typename_);
			 error=1;						
		      }
		      break;
		    }
            	}
             }
	     break;
	 }
      }
}





/*
   void buildBodySymbolTable(SymbolTable *symboltable,  NODE* node)
   Überprüft, ob alle Variablen deklariert sind und alle Pfade richtig sind
*/
void buildBodySymbolTable(SymbolTable *symboltable,  NODE* node)
{
	if (node==NULL) {
		return;
	}
	int type=node->type;

	switch(type) {

	case T_PATH: {
		PATH* path=(PATH*)node->pointer;
		if (controlPath(symboltable,node)==NULL)
		   error = 1;
		break;
	}
	case T_STMLIST: {
		STMLIST* stmlist=(STMLIST*)node->pointer;
		buildBodySymbolTable(symboltable,stmlist->head);
		buildBodySymbolTable(symboltable,stmlist->tail);
		break;
	}

	case T_ASSIGNER: {
		ASSIGNER* assign=(ASSIGNER*)node->pointer;
		buildBodySymbolTable(symboltable,assign->left);
		buildBodySymbolTable(symboltable,assign->right);
	   	break;
	}
	case T_PLUSOP: {
		PLUSOP* plusop=(PLUSOP*)node->pointer;
		buildBodySymbolTable(symboltable,plusop->left);
		buildBodySymbolTable(symboltable,plusop->right);
	   	break;
	}
	case T_MINUSOP: {
		MINUSOP* minusop=(MINUSOP*)node->pointer;
		buildBodySymbolTable(symboltable,minusop->left);
		buildBodySymbolTable(symboltable,minusop->right);
	   	break;
	}
	case T_MULTOP: {
		MULTOP* multop=(MULTOP*)node->pointer;
		buildBodySymbolTable(symboltable,multop->left);
		buildBodySymbolTable(symboltable,multop->right);
	   	break;
	}
	case T_DIVOP: {
		DIVOP* divop=(DIVOP*)node->pointer;
		buildBodySymbolTable(symboltable,divop->left);
		buildBodySymbolTable(symboltable,divop->right);
	   	break;
	}
	case T_MODOP: {
		MODOP* modop=(MODOP*)node->pointer;
		buildBodySymbolTable(symboltable,modop->left);
		buildBodySymbolTable(symboltable,modop->right);
	   	break;
	}
	case T_NEGATE: {
		NEGATE* negate=(NEGATE*)node->pointer;
		buildBodySymbolTable(symboltable,negate->expstm);
		break;
	}
	case T_IDENTIFIER: {
	   	IDENTIFIER* identi=(IDENTIFIER*)node->pointer;
                identi->symboltableentry=controlIdentifier(symboltable, node);
		if (identi->symboltableentry== NULL)
		   error = 1;
	   	break;
	}
	case T_NEW: {
		ALLOC* alloc=(ALLOC*)node->pointer;
		buildBodySymbolTable(symboltable,alloc->path);
		break;
	}
	case T_DELETE: {
		DELETE_* del=(DELETE_*)node->pointer;
		buildBodySymbolTable(symboltable,del->path);
		break;
	}
	case T_IFSTM: {
		IFSTM* ifstm=(IFSTM*)node->pointer;
		buildBodySymbolTable(symboltable,ifstm->bexp);
		buildBodySymbolTable(symboltable,ifstm->ifstm);
		buildBodySymbolTable(symboltable,ifstm->elsestm);
		break;
	}
	case T_WHILE: {
		WHILESTM* whilestm=(WHILESTM*)node->pointer;
		buildBodySymbolTable(symboltable,whilestm->bexp);
		buildBodySymbolTable(symboltable,whilestm->dostm);
		break;
	}
	case T_EQUAL: {
		EQUAL* equal=(EQUAL*)node->pointer;
		buildBodySymbolTable(symboltable,equal->left);
		buildBodySymbolTable(symboltable,equal->right);
	   	break;
	}
	case T_NOTEQUAL: {
		NOTEQUAL* notequal=(NOTEQUAL*)node->pointer;
		buildBodySymbolTable(symboltable,notequal->left);
		buildBodySymbolTable(symboltable,notequal->right);
	   	break;
	}
	case T_SMALLER: {
		SMALLER* smaller=(SMALLER*)node->pointer;
		buildBodySymbolTable(symboltable,smaller->left);
		buildBodySymbolTable(symboltable,smaller->right);
	   	break;
	}
	case T_GREATER: {
		GREATER* greater=(GREATER*)node->pointer;
		buildBodySymbolTable(symboltable,greater->left);
		buildBodySymbolTable(symboltable,greater->right);
	   	break;
	}
	case T_GREATEREQUAL: {
		GREATEREQUAL* greaterequal=(GREATEREQUAL*)node->pointer;
		buildBodySymbolTable(symboltable,greaterequal->left);
		buildBodySymbolTable(symboltable,greaterequal->right);
	   	break;
	}
	case T_SMALLEREQUAL: {
		SMALLEREQUAL* smallerequal=(SMALLEREQUAL*)node->pointer;
		buildBodySymbolTable(symboltable,smallerequal->left);
		buildBodySymbolTable(symboltable,smallerequal->right);
	   	break;
	}
	case T_AND: {
		AND_* and_=(AND_*)node->pointer;
		buildBodySymbolTable(symboltable,and_->left);
		buildBodySymbolTable(symboltable,and_->right);
	   	break;
	}
	case T_OR: {
		OR_* or_=(OR_*)node->pointer;
		buildBodySymbolTable(symboltable,or_->left);
		buildBodySymbolTable(symboltable,or_->right);

	   	break;
	}
	case T_NOT: {
		NOT_* not_=(NOT_*)node->pointer;
		buildBodySymbolTable(symboltable,not_->left);
	   	break;
	}
	default: return;
	}
}




/*
   RecordEntry* controlPath(SymbolTable *symboltable,  NODE *node)

   überprüft den Pfad auf die Richtigkeit
   Rückgabewert: falls alles ok - Recordentry der letzten Pfadkomponente
   sonst NULL 
*/
RecordEntry* controlPath(SymbolTable *symboltable,  NODE *node)
{

   if (symboltable == NULL || node==NULL)
      return NULL;

   //Typ der Knoten
   int type=node->type;
   if (type != T_PATH)
      return NULL;

   PATH* path=(PATH*)node->pointer;
   if (path->first == NULL || path->last == NULL)
      return NULL;
   if (path->first->type != T_IDENTIFIER)
      return NULL;

   IDENTIFIER* identifier = ((IDENTIFIER*)(path->first->pointer));
   SymbolEntry *symbolentry = NULL;
   symbolentry=getSymbol(symboltable,identifier->name);

   if (symbolentry == NULL) {
      printf("error in line %d\n", identifier->programcounter);
      printf("%s not defined\n", identifier->name);
      return NULL;
   }
   if (symbolentry->type != ptorecType) {
      printf("error in line %d\n", identifier->programcounter);
      printf("variable %s must be a pointer\n", identifier->name);
      return NULL;
   }
   RecordEntry *recordentry=NULL;
   PointerEntry *pointerentry=NULL;
   pointerentry=(PointerEntry*)(symbolentry->typref);
   char *record=pointerentry->recordname; 

   while (path->last->type != T_IDENTIFIER) {

        NODE *node=path->last;
       path= (PATH*)(node->pointer);

       if ((path->first==NULL) || (path->last==NULL))
          return NULL;

       if (path->first->type != T_IDENTIFIER)
          return NULL;

        IDENTIFIER* identifier = (IDENTIFIER *)(path->first->pointer);
       if (identifier == NULL)
	  return NULL;

        //überprüfen
       recordentry =getRecordComp(symboltable, identifier->name, record);
       if (recordentry==NULL) {
          printf("error in line %d\n", identifier->programcounter);
	  printf("component %s not in record %s\n", identifier->name,record);
	  return NULL;
       }
       if (recordentry->type != ptorecType) {
          printf("error in line %d\n", identifier->programcounter);
          printf("component %s of record %s not a pointer\n", identifier->name,record);
	  return NULL;
       }
       record = recordentry->recordname;

   }

   IDENTIFIER* identifierlast = (IDENTIFIER *)(path->last->pointer);
   if (identifier == NULL)
      return NULL;
   //RecordEntry *recordentry=NULL;
   recordentry=getRecordComp(symboltable, identifierlast->name, record);
   if (recordentry==NULL) {
      printf("error in line %d\n", identifier->programcounter);
      printf("component %s not in record %s\n", identifierlast->name,record);
      return NULL;
   }

   return recordentry;
}




/*
   SymbolEntry* controlIdentifier(SymbolTable *symboltable,  NODE *node)
   Überprüft ob der Bezeichner in der Symboltabelle vorhanden ist 
*/
SymbolEntry* controlIdentifier(SymbolTable *symboltable,  NODE *node)
{
   if (node==NULL)
      return NULL;
   int type=node->type;
   if (type != T_IDENTIFIER)
      return NULL;

   IDENTIFIER *identifier=(IDENTIFIER*)node->pointer; 
   SymbolEntry *symbolentry=NULL;
   symbolentry = getSymbol(symboltable,identifier->name);
   if (symbolentry == NULL ) {
      printf("error in line %d\n", identifier->programcounter);
      printf("variable %s not defined\n", identifier->name);
   }
   return symbolentry;
}




/*
int controlBEXP_(SymbolTable *symboltable, NODE *node)
Einschränkung der Bedingung der IF-Anweisung und der WHILE-Schleife
keine gemischte Ausdrücke
*/
int controlBEXP_(SymbolTable *symboltable, NODE *node, int line)
{

    int pointer=0;
    int null=0;
    int var=0;
    controlBEXP(symboltable, node, pointer, null, var);
    if (node->type==T_TRUE || node->type==T_FALSE)
       return -1;
    if (var>0) {
       if ((pointer>0) || (null>0)) {
          printf("error in line %d\n", line);
          printf("mixing types not allowed in Condition\n");
          printf("either boolean expression or pointer comparison allowed\n");
          error=1;
       }
    }
    else if (!((pointer==2 && null==0) || (pointer==1 && null==1))) {
       printf("error in line %d\n", line);
       printf("only one pointer<->pointer or pointer<->null comparison in IF-Condition allowed\n");
       error=1;
    }
    return 1;
}



/*
void controlBEXP(SymbolTable *symboltable, NODE *node, int &pointer, int &null, int &var)
*/
void controlBEXP(SymbolTable *symboltable, NODE *node, int &pointer, int &null, int &var)
{

    switch (node->type) {

        case T_IDENTIFIER: {
           SymbolEntry *se=getSymbol(symboltable, ((IDENTIFIER*)(node->pointer))->name);
           assert(se != NULL);
           if (se->type==ptorecType)
              pointer++;
           else if (se->type==intType || se->type==strType)
              var++;
           else
              assert(false);
           break;
        }
        case T_NEGATE: {
           controlBEXP(symboltable, ((NEGATE*)(node->pointer))->expstm, pointer, null, var);
           break;
        }
	case T_INTEGERCONST: {
           var++;
           break;
        }
	case T_STRCONST: {
           var++;
           break;
        }
	case T_PLUSOP: {
           controlBEXP(symboltable, ((PLUSOP*)(node->pointer))->left, pointer, null, var);
           controlBEXP(symboltable, ((PLUSOP*)(node->pointer))->right, pointer, null, var);
           break;
        }
	case T_MINUSOP: {
           controlBEXP(symboltable, ((MINUSOP*)(node->pointer))->left, pointer, null, var);
           controlBEXP(symboltable, ((MINUSOP*)(node->pointer))->right, pointer, null, var);
           break;
        }
	case T_MULTOP: {
           controlBEXP(symboltable, ((MULTOP*)(node->pointer))->left, pointer, null, var);
           controlBEXP(symboltable, ((MULTOP*)(node->pointer))->right, pointer, null, var);
           break;
        }
	case T_DIVOP: {
           controlBEXP(symboltable, ((DIVOP*)(node->pointer))->left, pointer, null, var);
           controlBEXP(symboltable, ((DIVOP*)(node->pointer))->right, pointer, null, var);
           break;
        }
	case T_MODOP: {
           controlBEXP(symboltable, ((MODOP*)(node->pointer))->left, pointer, null, var);
           controlBEXP(symboltable, ((MODOP*)(node->pointer))->right, pointer, null, var);
           break;
        }
	case T_PATH: {
           RecordEntry *recordentry=getLastPathComp(symboltable, node);
           if (recordentry->type==ptorecType)
              pointer++;
           else if (recordentry->type==intType || recordentry->type==strType)
              var++;
           else
              assert(false);
           break;
        }
	case T_EQUAL: {
           controlBEXP(symboltable, ((EQUAL*)(node->pointer))->left, pointer, null, var);
           controlBEXP(symboltable, ((EQUAL*)(node->pointer))->right, pointer, null, var);
           break;
	}
	case T_SMALLER: {
           controlBEXP(symboltable, ((SMALLER*)(node->pointer))->left, pointer, null, var);
           controlBEXP(symboltable, ((SMALLER*)(node->pointer))->right, pointer, null, var);
           break;
        }
	case T_GREATER: {
           controlBEXP(symboltable, ((GREATER*)(node->pointer))->left, pointer, null, var);
           controlBEXP(symboltable, ((GREATER*)(node->pointer))->right, pointer, null, var);
           break;
        }
	case T_SMALLEREQUAL: {
           controlBEXP(symboltable, ((SMALLEREQUAL*)(node->pointer))->left, pointer, null, var);
           controlBEXP(symboltable, ((SMALLEREQUAL*)(node->pointer))->right, pointer, null, var);
           break;
        }
	case T_GREATEREQUAL: {
           controlBEXP(symboltable, ((GREATEREQUAL*)(node->pointer))->left, pointer, null, var);
           controlBEXP(symboltable, ((GREATEREQUAL*)(node->pointer))->right, pointer, null, var);
           break;
        }
	case T_NOTEQUAL: {
           controlBEXP(symboltable, ((NOTEQUAL*)(node->pointer))->left, pointer, null, var);
           controlBEXP(symboltable, ((NOTEQUAL*)(node->pointer))->right, pointer, null, var);
           break;
        }
	case T_AND: {
           controlBEXP(symboltable, ((AND_*)(node->pointer))->left, pointer, null, var);
           controlBEXP(symboltable, ((AND_*)(node->pointer))->right, pointer, null, var);
	   break;
	}
	case T_OR: {
           controlBEXP(symboltable, ((OR_*)(node->pointer))->left, pointer, null, var);
           controlBEXP(symboltable, ((OR_*)(node->pointer))->right, pointer, null, var);
	   break;
	}
	case T_NOT: {
           controlBEXP(symboltable, ((NOT_*)(node->pointer))->left, pointer, null, var);
	   break;
	}
	case T_TRUE: {
	   break;
	}
	case T_FALSE: {
	   break;
	}
	case T_NULL: {
           null++;
	   break;
	}
	default: {
	   printf("unexpected Node");
           assert(false);
	}
   }

}




/*
   void controlType(SymbolTable *symboltable,  NODE* node)
*/
void controlType(SymbolTable *symboltable,  NODE* node)
{

	if (node==NULL) {
		return;
	}
	int type=node->type;

	switch(type) {

	case T_PATH: {
		PATH* path=(PATH*)node->pointer;
		break;
	}
	case T_STMLIST: {
		STMLIST* stmlist=(STMLIST*)node->pointer;
		controlType(symboltable,stmlist->head);
		controlType(symboltable,stmlist->tail);
		break;
	}

	case T_ASSIGNER: {
		ASSIGNER* assign=(ASSIGNER*)node->pointer;
		char *lefttype=getNodeType(symboltable, assign->left);
		char *righttype=getNodeType(symboltable, assign->right);

		if (righttype == "NULL" && (strstr(lefttype, "PTOREC ")!=NULL)) {
		   return;
		}	

		if (strcmp(lefttype, righttype) != 0 ) {
		   printf("error in line %d\n", assign->programcounter);
                   printf("expression type on the right side does not match left side\n");
		   error = 1;
		   return;
		}
		controlType(symboltable,assign->right);
	   	break;
	}
	case T_PLUSOP: {
		PLUSOP* plusop=(PLUSOP*)node->pointer;
		char *lefttype=getNodeType(symboltable, plusop->left);
		char *righttype=getNodeType(symboltable, plusop->right);
		//printf("lefttype: %s, righttype: %s\n", lefttype, righttype);
		if (lefttype != "INT" || righttype != "INT") {
		   printf("error in line %d\n", plusop->programcounter);
                   printf("'%s' accepts INTEGER types only", STR_ADD);
		   error = 1;
		   return;
		}
		controlType(symboltable,plusop->left);
		controlType(symboltable,plusop->right);
	   	break;
	}
	case T_MINUSOP: {
		MINUSOP* minusop=(MINUSOP*)node->pointer;
		char *lefttype=getNodeType(symboltable, minusop->left);
		char *righttype=getNodeType(symboltable, minusop->right);

		if (lefttype != "INT" || righttype != "INT") {
		   printf("error in line %d\n", minusop->programcounter);
                   printf("'%s' accepts INTEGER types only", STR_MIN);
		   error = 1;
		   return;
		}
		controlType(symboltable,minusop->left);
		controlType(symboltable,minusop->right);
	   	break;
	}
	case T_MULTOP: {
		MULTOP* multop=(MULTOP*)node->pointer;
		char *lefttype=getNodeType(symboltable, multop->left);
		char *righttype=getNodeType(symboltable, multop->right);

		if (lefttype != "INT" || righttype != "INT") {
		   printf("error in line %d\n", multop->programcounter);
                   printf("'%s' accepts INTEGER types only", STR_MULT);
		   error = 1;
		   return;
		}
		controlType(symboltable,multop->left);
		controlType(symboltable,multop->right);
	   	break;
	}
	case T_DIVOP: {
		DIVOP* divop=(DIVOP*)node->pointer;
		char *lefttype=getNodeType(symboltable, divop->left);
		char *righttype=getNodeType(symboltable, divop->right);

		if (lefttype != "INT" || righttype != "INT") {
		   printf("error in line %d\n", divop->programcounter);
                   printf("'%s' accepts INTEGER types only", STR_DIV);
		   error = 1;
		   return;
		}
		controlType(symboltable,divop->left);
		controlType(symboltable,divop->right);
	   	break;
	}
	case T_MODOP: {
		MODOP* modop=(MODOP*)node->pointer;
		char *lefttype=getNodeType(symboltable, modop->left);
		char *righttype=getNodeType(symboltable, modop->right);

		if (lefttype != "INT" || righttype != "INT") {
		   printf("error in line %d\n", modop->programcounter);
                   printf("'%s' accepts INTEGER types only", STR_MOD);
		   error = 1;
		   return;
		}
		controlType(symboltable,modop->left);
		controlType(symboltable,modop->right);
	   	break;
	}
	case T_NEGATE: {
		 NEGATE* negate=(NEGATE*)node->pointer;
	        char *type=getNodeType(symboltable, negate->expstm);
		if (type != "INT") {
		   printf("error in line %d\n", negate->programcounter);
                   printf("'%s' accepts INTEGER types only", STR_UMINUS);
		   error = 1;
		   return;		   
		}
		controlType(symboltable,negate->expstm);
		break;
	}
	case T_IDENTIFIER: {
	   	IDENTIFIER* identi=(IDENTIFIER*)node->pointer;
                identi->symboltableentry=controlIdentifier(symboltable, node);
		if (identi->symboltableentry== NULL)
		   error = 1;
	   	break;
	}
	case T_NEW: {
		 ALLOC* alloc=(ALLOC*)node->pointer;

                char *varname, *recordname;
                getLastPath(symboltable, alloc->path, &varname, &recordname);
                if ((strcmp(recordname, "#int#")==0) || (strcmp(recordname, "#str#")==0) || (strcmp(recordname,"#type#")==0)) {
                   printf("error in line %d\n", alloc->programcounter);
                   printf("pointer needed in NEW ");
		   error = 1;
                }
		break;

	}
	case T_DELETE: {
		 DELETE_* del=(DELETE_*)node->pointer;

                char *varname, *recordname;
                getLastPath(symboltable, del->path, &varname, &recordname);
                if ((strcmp(recordname, "#int#")==0) || (strcmp(recordname, "#str#")==0) || (strcmp(recordname,"#type#")==0)) {
                   printf("error in line %d\n", del->programcounter);
                   printf("pointer needed in DELETE ");
		   error = 1;
                }
		break;

	}
	case T_IFSTM: {
		IFSTM* ifstm=(IFSTM*)node->pointer;
                controlBEXP_(symboltable,ifstm->bexp, ifstm->programcounter); 
		controlType(symboltable,ifstm->bexp);
		controlType(symboltable,ifstm->ifstm);
		controlType(symboltable,ifstm->elsestm);
		break;
	}
	case T_WHILE: {
		WHILESTM* whilestm=(WHILESTM*)node->pointer;
                controlBEXP_(symboltable,whilestm->bexp, whilestm->programcounter); 
		controlType(symboltable,whilestm->bexp);
		controlType(symboltable,whilestm->dostm);
		break;
	}
	case T_EQUAL: {
		EQUAL* equal=(EQUAL*)node->pointer;
		char *lefttype=getNodeType(symboltable, equal->left);
		char *righttype=getNodeType(symboltable, equal->right);
		//alles ok
	        if (strcmp(lefttype,righttype)==0) 
		   return;
		//alles ok
		if (righttype == "NULL" && (strstr(lefttype, "PTOREC ")!=NULL))
		   return;	
		//alles ok
		if (lefttype == "NULL" && (strstr(righttype, "PTOREC ")!=NULL))  
		   return;
		if ((lefttype == "NULL") && (righttype == "NULL"))
		   return;
		printf("error in line %d\n", equal->programcounter);
                printf("types of left and right side of '=' don`t match\n");
		error = 1;
		break;
	}
	case T_NOTEQUAL: {
		 NOTEQUAL* notequal=(NOTEQUAL*)node->pointer;
		char *lefttype=getNodeType(symboltable, notequal->left);
		char *righttype=getNodeType(symboltable, notequal->right);
		//alles ok
	        if (strcmp(lefttype,righttype)==0) 
		   return;
		//alles ok
		if (righttype == "NULL" && (strstr(lefttype, "PTOREC ")!=NULL))
		   return;	
		//alles ok
		if (lefttype == "NULL" && (strstr(righttype, "PTOREC ")!=NULL))  
		   return;
		if ((lefttype == "NULL") && (righttype == "NULL"))
		   return;
		printf("error in line %d\n", notequal->programcounter);
                printf("types of left and right side of '%s' don't match\n", STR_NOTEQUAL);
		error = 1;
		break;

	}
	case T_SMALLER: {

		SMALLER* smaller=(SMALLER*)node->pointer;

		char *lefttype=getNodeType(symboltable, smaller->left);
		char *righttype=getNodeType(symboltable, smaller->right);

		if (lefttype != "INT" || righttype != "INT") {
		   printf("error in line %d\n", smaller->programcounter);
                   printf("'%s' accepts INTEGER types only", STR_SMALLER);
		   error = 1;
		   return;
		}
		controlType(symboltable,smaller->left);
		controlType(symboltable,smaller->right);
	   	break;
	}
	case T_GREATER: {

		GREATER* greater=(GREATER*)node->pointer;

		char *lefttype=getNodeType(symboltable, greater->left);
		char *righttype=getNodeType(symboltable, greater->right);

		if (lefttype != "INT" || righttype != "INT") {
		   printf("error in line %d\n", greater->programcounter);
                   printf("'%s' accepts INTEGER types only", STR_GREATER);
		   error = 1;
		   return;
		}
		controlType(symboltable,greater->left);
		controlType(symboltable,greater->right);
	   	break;
	}
	case T_GREATEREQUAL: {

		GREATEREQUAL* greaterequal=(GREATEREQUAL*)node->pointer;

		char *lefttype=getNodeType(symboltable, greaterequal->left);
		char *righttype=getNodeType(symboltable, greaterequal->right);

		if (lefttype != "INT" || righttype != "INT") {
		   printf("error in line %d\n", greaterequal->programcounter);
                   printf("'%s' accepts INTEGER types only", STR_GREATEREQUAL);
		   error = 1;
		   return;
		}
		controlType(symboltable,greaterequal->left);
		controlType(symboltable,greaterequal->right);
	   	break;

	}
	case T_SMALLEREQUAL: {

		SMALLEREQUAL* smallerequal=(SMALLEREQUAL*)node->pointer;

		char *lefttype=getNodeType(symboltable, smallerequal->left);
		char *righttype=getNodeType(symboltable, smallerequal->right);

		if (lefttype != "INT" || righttype != "INT") {
		   printf("error in line %d\n", smallerequal->programcounter);
                   printf("'%s' accepts INTEGER types only", STR_SMALLEREQUAL);
		   error = 1;
		   return;
		}
		controlType(symboltable,smallerequal->left);
		controlType(symboltable,smallerequal->right);
	   	break;

	}
	case T_AND: {
		AND_* and_=(AND_*)node->pointer;
		controlType(symboltable,and_->left);
		controlType(symboltable,and_->right);
	   	break;
	}
	case T_OR: {
		OR_* or_=(OR_*)node->pointer;
		controlType(symboltable,or_->left);
		controlType(symboltable,or_->right);
	   	break;
	}
	case T_NOT: {
		NOT_* not_=(NOT_*)node->pointer;
		controlType(symboltable,not_->left);
	   	break;
	}
	default: return;
	}

}




/*
char* getNodeType(SymbolTable *symboltable,  NODE *node)

T_PLUSOP, T_MINUSOP, T_MULTOP, T_DIVOP, T_MODOP, T_NEGATE, T_INTEGERCONST, T_IDENTIFIER (falls vom Typ INTEGER), 
T_PATH (falls letzte Pfadkomponente vom Typ INTEGER) ---> "INT"

T_STRCONST, T_IDENTIFIER (falls vom Typ STRING), T_PATH (falls letzte Pfadkomponente vom Typ STRING)
---> "STR"

T_NULL ---> "NULL" 

T_IDENTIFIER (falls vom Typ POINTER), T_PATH (falls vom Typ POINTER), T_NEW
---> "PTOREC [recordname]
*/
char* getNodeType(SymbolTable *symboltable,  NODE *node)
{

	int NodeKind=node->type;
	switch (NodeKind) {
   	   	case T_PLUSOP: {
			return "INT";
			break;
		}
      		case T_MINUSOP: {
			return "INT";
			break;
		}
      		case T_MULTOP:{
			return "INT";
			break;
		}
      		case T_DIVOP: {
			return "INT";
			break;
		}
      		case T_MODOP: {
			return "INT";
			break;
		}
		case T_NEGATE: {
			return "INT";
			break;
		}
		case T_INTEGERCONST: {
			return "INT";
			break;
		}
		case T_STRCONST: {
			return "STR";
			break;
		}
		case T_NULL: {
			return "NULL";
			break;
		}
		case T_IDENTIFIER: {
			IDENTIFIER* identi=(IDENTIFIER*)node->pointer;
			if (identi->symboltableentry==NULL) return NULL;
			if (identi->symboltableentry->type == intType)
				return "INT";
			if (identi->symboltableentry->type == strType)
				return "STR";
                        if (identi->symboltableentry->type == ptorecType) {
				char *str1="PTOREC ";
				char *str2=((PointerEntry*)(identi->symboltableentry->typref))->recordname;
				char *recordname=catstrstr(str1,str2);
				return recordname;
			}
			break;
		}
		case T_PATH: {
			
			PATH* path=(PATH*)node->pointer;
			RecordEntry *recordentry=getLastPathComp(symboltable, node);

			if (recordentry->type==intType)
				return "INT";
			if (recordentry->type==strType)
				return "STR";
			if (recordentry->type==ptorecType) {
				char *str1="PTOREC ";
				char *str2=(char*)(recordentry->recordname);
				char *recordname=catstrstr(str1,str2);

				return recordname;
  			}
			break;
		}
		case T_NEW: {
			ALLOC* alloc=(ALLOC*)node->pointer;
			RecordEntry *recordentry=getLastPathComp(symboltable, node);
                        assert(recordentry->type==ptorecType);
		        char *str1="PTOREC ";
			char *str2=(char*)(recordentry->recordname);
			char *recordname=catstrstr(str1,str2);
			return recordname;
			break;
		}
		case T_DELETE: {
			DELETE_* del=(DELETE_*)node->pointer;
			RecordEntry *recordentry=getLastPathComp(symboltable, node);
                        assert(recordentry->type==ptorecType);
		        char *str1="PTOREC ";
			char *str2=(char*)(recordentry->recordname);
			char *recordname=catstrstr(str1,str2);
			return recordname;
			break;
		}

		default: return "0";
   }
}



/* int getPathLength(NODE *node) */
/* gibt die Länge des Pfades zurück */
/* 0 bei Fehler */

int getPathLength(NODE *node)
{
   if (node==NULL)
      return 0;
   int type=node->type;
   if (type == T_IDENTIFIER)
      return 1;
   if (type != T_PATH)
      return 0;
    PATH* path=(PATH*)node->pointer;
   return (getPathLength(path->last)+1);
}



/*
int getNPath(SymbolTable* symboltable,  NODE *node, int length, char **varname, char **recordname)
*/
int getNPath(SymbolTable* symboltable,  NODE *node, int length, char **varname, char **recordname)
{
   //assert (node->type == T_PATH);
   if (symboltable == NULL || node==NULL)
      return 0;


   if (length==1) {
      if (node->type==T_IDENTIFIER) {
         SymbolEntry *symbolentry=getSymbol(symboltable, ((IDENTIFIER *)(node->pointer))->name);
         *varname=symbolentry->name;
         if (symbolentry->type != ptorecType) 
            if (symbolentry->type == intType)
               *recordname="#int#";
            else if (symbolentry->type == strType)
               *recordname="#str#";
            else if (symbolentry->type == recType)
               *recordname="#type#";
            else
               assert(false);
          else
            *recordname=((PointerEntry*)(symbolentry->typref))->recordname;
       return 1;
       }
   }

   PATH *path=(PATH*)node->pointer;
   IDENTIFIER *identifierfirst=(IDENTIFIER *)(path->first->pointer);

   SymbolEntry *symbolentry=getSymbol(symboltable,identifierfirst->name);
   *varname=identifierfirst->name;
   *recordname=((PointerEntry*)(symbolentry->typref))->recordname;
   if (length==1)
      return 1;

   if (path->last->type == T_IDENTIFIER) {

       IDENTIFIER *identifierlast=(IDENTIFIER *)(path->last->pointer);
      RecordEntry *recordentry = getRecordComp(symboltable,identifierlast->name,*recordname);

      *varname=recordentry->name;
      if (recordentry->type != ptorecType) 
         if (recordentry->type == intType)
            *recordname="#int#";
         else if (recordentry->type == strType)
            *recordname="#str#";
         else
            assert(false);
      else
         *recordname=recordentry->recordname;
      if (length == 2)
         return 1;
      return 0;
   }
   //cout <<"varname: " <<*varname <<", " <<"recordname: "<<*recordname <<endl;
   return getNPath_(symboltable, path->last, length-1, varname, recordname);

}



/*
   int getNPath_(SymbolTable* symboltable,  NODE *node, int length, char **varname, char **recordname)
*/
int getNPath_(SymbolTable* symboltable,  NODE *node, int length, char **varname, char **recordname)
{

   if (symboltable == NULL || node==NULL)
      return 0;
    PATH *path=(PATH*)node->pointer;
    IDENTIFIER *identifierfirst=(IDENTIFIER *)(path->first->pointer);
   RecordEntry *recordentry = getRecordComp(symboltable,identifierfirst->name,*recordname);

   int type=recordentry->type;
   *varname=recordentry->name;

   if (length==1) {
      if (type != ptorecType)
         if (recordentry->type == intType)
            *recordname="#int#";
         else if (recordentry->type == strType)
            *recordname="#str#";
         else
            assert(false);
      else
            *recordname=recordentry->recordname;
      return 1;
   }

   if (path->last->type == T_IDENTIFIER) {
      *recordname=recordentry->recordname;
       IDENTIFIER *identifierlast=(IDENTIFIER *)(path->last->pointer);
      RecordEntry *recordentry = getRecordComp(symboltable,identifierlast->name,*recordname);

      *varname=recordentry->name;
      if (recordentry->type != ptorecType)  {
         if (recordentry->type == intType)
            *recordname="#int#";
         else if (recordentry->type == strType)
            *recordname="#str#";
         else
            assert(false);
      }
      else 
         *recordname=recordentry->recordname;
      if (length == 2)
         return 1;
      return 0;
   }
   *recordname=recordentry->recordname;
   return getNPath_(symboltable,path->last,length-1,varname,recordname);

}



/*
   int getLastPath(SymbolTable* symboltable,  NODE *node, char **varname, char **recordname)
   gibt Eigenschaften (Record-Name, Komponente) der letzten Pfadkomponente
*/
int getLastPath(SymbolTable* symboltable,  NODE *node, char **varname, char **recordname)
{
   //assert (node->type == T_PATH || node->type == T_IDENTIFIER);
   if (symboltable == NULL || node==NULL)
      return 0;
   
   int pathlength=getPathLength(node);

   return getNPath(symboltable, node, pathlength, varname, recordname);
}



/*
   RecordEntry* getLastPathComp(SymbolTable* symboltable,  NODE *node)
   gibt Record-Eintrag der letzten Pfadkomponente
*/
RecordEntry* getLastPathComp(SymbolTable* symboltable,  NODE *node)
{

   if (symboltable == NULL || node==NULL)
      return NULL;


   PATH *path=(PATH*)node->pointer;
   IDENTIFIER *identifierfirst=(IDENTIFIER *)(path->first->pointer);
   SymbolEntry *symbolentry=getSymbol(symboltable,identifierfirst->name);
   char *recordname = ((PointerEntry*)(symbolentry->typref))->recordname;

   RecordEntry *recordentry=NULL;
   if (path->last->type == T_IDENTIFIER) {
      IDENTIFIER *identifierlast=(IDENTIFIER *)(path->last->pointer);
      recordentry = getRecordComp(symboltable,identifierlast->name,recordname);
   }
   else 
      recordentry=getLastPathComp_(symboltable, path->last, recordname);

   return recordentry;
}



/*
RecordEntry* getLastPathComp_(SymbolTable* symboltable,  NODE* node, char *recordname)
*/
RecordEntry* getLastPathComp_(SymbolTable* symboltable,  NODE* node, char *recordname)
{

   if (symboltable == NULL || node==NULL)
      return NULL;
    PATH *path=(PATH*)node->pointer;

   while (path->last->type != T_IDENTIFIER) {
      IDENTIFIER *identifierfirst=(IDENTIFIER *)(path->first->pointer);
      RecordEntry *recordentry = getRecordComp(symboltable,identifierfirst->name,recordname);

      if (recordentry != NULL)
         return getLastPathComp_(symboltable,path->last,recordentry->recordname);
      else
         return NULL;
   }
   if (path->last->type == T_IDENTIFIER) {

      IDENTIFIER *identifierfirst=(IDENTIFIER *)(path->first->pointer);
      IDENTIFIER *identifierlast=(IDENTIFIER *)(path->last->pointer);
      RecordEntry *recordentry = getRecordComp(symboltable,identifierfirst->name,recordname);
      recordname=recordentry->recordname;
      recordentry = getRecordComp(symboltable,identifierlast->name,recordname);
      if (recordentry != NULL)
         return recordentry;
      else
         return NULL;
   }
   return NULL;
}



/*
   void freeAbstree(NODE* node)
*/
void freeAbstree(NODE* node){

   if (node==NULL) {
      return;
   }
   int type=node->type;
   switch(type) {
	case T_VARDECL: {
		VARDECL* vardecl=(VARDECL*)node->pointer;
		freeAbstree(vardecl->type);
                free(vardecl);
                free(node);
		break;
	}
	case T_DECLLIST: {
		DECLLIST* decllist=(DECLLIST*)node->pointer;
		freeAbstree(decllist->head);
		freeAbstree(decllist->tail);
                free(decllist);
                free(node);
		break;
	}
	case T_TYPEDECL: {
		TYPEDECL* typedecl=(TYPEDECL*)node->pointer;
		freeAbstree(typedecl->head);
                freeAbstree(typedecl->tail);
		free(typedecl);
		free(node);
		break;
	}
        case T_TYPEDECLNULL: {
                TYPEDECLNULL* typedeclnull=(TYPEDECLNULL*)node->pointer;
		free(typedeclnull);
		free(node);
                break;
        }
	case T_TYPEDECLLIST: {
		TYPEDECLLIST* typedecllist=(TYPEDECLLIST*)node->pointer;
		freeAbstree(typedecllist->head);
		freeAbstree(typedecllist->tail);
                free(typedecllist);
		free(node);
		break;
	}
	case T_TVARDECL: {
		TVARDECL* tvardecl=(TVARDECL*)node->pointer;
		freeAbstree(tvardecl->type);
		free(tvardecl);
		free(node);
		break;
	}

	case T_INTEGER: {
		free(node);
		break;
	}
	case T_STRING: {
		free(node);
		break;
	}
	case T_PTOREC: {
		PTOREC* ptorec=(PTOREC*)node->pointer;
		free(ptorec);
		break;
	}
	case T_PATH: {
		PATH* path=(PATH*)node->pointer;
		freeAbstree(path->first);
		freeAbstree(path->last);
		free(path);
		free(node);
		break;
	}
	case T_PROGRAM: {
		PROGRAM* stmlist=(PROGRAM*)node->pointer;
		freeAbstree(stmlist->declarations);
		freeAbstree(stmlist->body);
		free(stmlist);
		free(node);
		break;
	}
	case T_STMLIST: {
		STMLIST* stmlist=(STMLIST*)node->pointer;
		freeAbstree(stmlist->head);
		freeAbstree(stmlist->tail);
		free(stmlist);
		free(node);
		break;
	}
	case T_INTEGERCONST: {
	   	INTEGERCONST* intconst=(INTEGERCONST*)node->pointer;
		free(intconst);
		free(node);
	   	break;
	}
	case T_STRCONST: {
		STRCONST* strconst=(STRCONST*)node->pointer;
		free(strconst->value);
		free(strconst);
		free(node);
		break;
	}
	case T_ASSIGNER: {
		ASSIGNER* assign=(ASSIGNER*)node->pointer;
		freeAbstree(assign->left);
		freeAbstree(assign->right);
		free(assign);
		free(node);
	   	break;
	}
	case T_PLUSOP: {
		PLUSOP* plusop=(PLUSOP*)node->pointer;
		freeAbstree(plusop->left);
		freeAbstree(plusop->right);
		free(plusop);
		free(node);
	   	break;
	}
	case T_MINUSOP: {
		MINUSOP* minusop=(MINUSOP*)node->pointer;
		freeAbstree(minusop->left);
		freeAbstree(minusop->right);
		free(minusop);
		free(node);
	   	break;
	}
	case T_MULTOP: {
		MULTOP* multop=(MULTOP*)node->pointer;
		freeAbstree(multop->left);
		freeAbstree(multop->right);
		free(multop);
		free(node);
	   	break;
	}
	case T_DIVOP: {
		DIVOP* divop=(DIVOP*)node->pointer;
		freeAbstree(divop->left);
		freeAbstree(divop->right);
		free(divop);
		free(node);
	   	break;
	}
	case T_MODOP: {
		MODOP* modop=(MODOP*)node->pointer;
		freeAbstree(modop->left);
		freeAbstree(modop->right);
		free(modop);
		free(node);
	   	break;
	}
	case T_NEGATE: {
		NEGATE* negate=(NEGATE*)node->pointer;
		freeAbstree(negate->expstm);
		free(negate);
		free(node);
		break;
	}
	case T_IDENTIFIER: {
	   	IDENTIFIER* identi=(IDENTIFIER*)node->pointer;
		free(identi);
		free(node);
	   	break;
	}
	case T_NEW: {
		ALLOC* alloc=(ALLOC*)node->pointer;
		freeAbstree(alloc->path);
		free(alloc);
		free(node);
		break;
	}
	case T_DELETE: {
		DELETE_* del=(DELETE_*)node->pointer;
		freeAbstree(del->path);
		free(del);
		free(node);
		break;
	}
	case T_SKIP: {
		free(node);
		break;
	}
	case T_IFSTM: {
		IFSTM* ifstm=(IFSTM*)node->pointer;
		freeAbstree(ifstm->bexp);
		freeAbstree(ifstm->ifstm);
		freeAbstree(ifstm->elsestm);
		free(ifstm);
		free(node);
		break;
	}
	case T_WHILE: {
		WHILESTM* whilestm=(WHILESTM*)node->pointer;
		freeAbstree(whilestm->bexp);
		freeAbstree(whilestm->dostm);
		free(whilestm);
		free(node);
		break;
	}
	case T_EQUAL: {
		EQUAL* equal=(EQUAL*)node->pointer;
		freeAbstree(equal->left);
		freeAbstree(equal->right);
		free(equal);
		free(node);
	   	break;
	}
	case T_NOTEQUAL: {
		NOTEQUAL* notequal=(NOTEQUAL*)node->pointer;
		freeAbstree(notequal->left);
		freeAbstree(notequal->right);
		free(notequal);
		free(node);
	   	break;
	}
	case T_SMALLER: {
		SMALLER* smaller=(SMALLER*)node->pointer;
		freeAbstree(smaller->left);
		freeAbstree(smaller->right);
		free(smaller);
		free(node);
	   	break;
	}
	case T_GREATER: {
		GREATER* greater=(GREATER*)node->pointer;
		freeAbstree(greater->left);
		freeAbstree(greater->right);
		free(greater);
		free(node);
	   	break;
	}
	case T_GREATEREQUAL: {
		GREATEREQUAL* greaterequal=(GREATEREQUAL*)node->pointer;
		freeAbstree(greaterequal->left);
		freeAbstree(greaterequal->right);
		free(greaterequal);
		free(node);
	   	break;
	}
	case T_SMALLEREQUAL: {
		SMALLEREQUAL* smallerequal=(SMALLEREQUAL*)node->pointer;
		freeAbstree(smallerequal->left);
		freeAbstree(smallerequal->right);
		free(smallerequal);
		free(node);
	   	break;
	}
	case T_AND: {
		AND_* and_=(AND_*)node->pointer;
 		freeAbstree(and_->left);
		freeAbstree(and_->right);
		free(and_);
		free(node);
	   	break;
	}
	case T_OR: {
		OR_* or_=(OR_*)node->pointer;
		freeAbstree(or_->left);
		freeAbstree(or_->right);
		free(or_);
		free(node);
	   	break;
	}
	case T_NOT: {
		NOT_* not_=(NOT_*)node->pointer;
		freeAbstree(not_->left);
		free(not_);
		free(node);
	   	break;
	}
	case T_TRUE: {
		free(node);
		break;
	}
	case T_FALSE: {
		free(node);
		break;
	}
	case T_NULL: {
		free(node);
		break;
	}
	}


}
