#ifndef _ABSTREE_H_
#define _ABSTREE_H_

#include "symboltable.h"

const int MAX_ID=25;

#define T_PROGRAM 10
#define T_DECLLIST 20
#define T_VARDECL 30
#define T_TYPEDECL 40
#define T_TYPEDECLNULL 45
#define T_TVARDECL 50
#define T_TYPEDECLLIST 60


#define T_STMLIST 65

#define T_VARINT 70
#define T_VARSTR 80
#define T_VARPOINT 90

#define T_INTEGER 100
#define T_STRING 110
#define T_PTOREC 120


#define T_IDENTIFIER 130
#define T_NEGATE 135
#define T_INTEGERCONST 140
#define T_STRCONST 145


#define T_ASSIGNER 150

#define T_PLUSOP 160
#define T_MINUSOP 170
#define T_MULTOP 180
#define T_DIVOP 190
#define T_MODOP 195

#define T_PATH 200
#define T_NEW 210
#define T_DELETE 215
#define T_SKIP 220

#define T_EQUAL 300
#define T_SMALLER 310
#define T_GREATER 320
#define T_SMALLEREQUAL 330
#define T_GREATEREQUAL 340
#define T_NOTEQUAL 350

#define T_AND 360
#define T_OR 370
#define T_NOT 380

#define T_IFSTM 400
#define T_WHILE 410

#define T_TRUE 500
#define T_FALSE 510
#define T_NULL 520



#define STR_ASSIGNER     ":="
#define STR_EQUAL        "="
#define STR_NOTEQUAL     "!="
#define STR_SMALLER      "<"
#define STR_GREATER      ">"
#define STR_SMALLEREQUAL "<="
#define STR_GREATEREQUAL ">="
#define STR_AND          " AND "
#define STR_OR           " OR "
#define STR_NOT          "NOT"
#define STR_ADD          "+"
#define STR_MIN          "-"
#define STR_MULT         "*"
#define STR_DIV          "/"
#define STR_MOD          "%"
#define STR_UMINUS       "-"

/* Hilfsstruktur */
typedef struct {
   int type;
   void* pointer;
} NODE;

/* Knoten des Parserbaumes */
typedef struct  {
   NODE* declarations;
   NODE* body;
}PROGRAM;

typedef struct {
   int programcounter;
   char name[MAX_ID];
   NODE* type;
}VARDECL;

typedef struct  {
   int programcounter;
   char name[MAX_ID];
   NODE* head;
   NODE* tail;
}TYPEDECL;

typedef struct {
   int programcounter;
   char name[MAX_ID];
}TYPEDECLNULL;

typedef struct {
   int programcounter;
   char name[MAX_ID];
   NODE* head;
   NODE* tail;
}TYPEDECLLIST;

typedef struct  {
   int programcounter;
   char name[MAX_ID];
   NODE* type;
}TVARDECL;

typedef struct  {
   NODE* head;
   NODE* tail;
}DECLLIST;

typedef struct {
   int programcounter;
   char name[MAX_ID];
}PTOREC;

typedef struct {
   NODE* head;
   NODE* tail;
}STMLIST;

typedef struct {
   int programcounter;
   NODE* expstm;
}NEGATE;

typedef struct  {
   int programcounter;
   NODE* left;
   NODE* right;
}ASSIGNER;


typedef struct {
   int programcounter;
   NODE* left;
   NODE* right;
}PLUSOP;

typedef struct {
   int programcounter;
   NODE* left;
   NODE* right;
}MINUSOP;

typedef struct  {
   int programcounter;
   NODE* left;
   NODE* right;
}MULTOP;

typedef struct {
   int programcounter;
   NODE* left;
   NODE* right;
}DIVOP;

typedef struct {
   int programcounter;
   NODE* left;
   NODE* right;
}MODOP;

typedef struct{
   int programcounter;
   NODE* left;
   NODE* right;
}EQUAL;

typedef struct {
   int programcounter;
   NODE* left;
   NODE* right;
}SMALLER;

typedef struct {
   int programcounter;
   NODE* left;
   NODE* right;
}GREATER;

typedef struct {
   int programcounter;
   NODE* left;
   NODE* right;
}SMALLEREQUAL;

typedef struct {
   int programcounter;
   NODE* left;
   NODE* right;
}GREATEREQUAL;

typedef struct {
   int programcounter;
   NODE* left;
   NODE* right;
}NOTEQUAL;

typedef struct {
   int programcounter;
   NODE* left;
   NODE* right;
}AND_;

typedef struct {
   int programcounter;
   NODE* left;
   NODE* right;
}OR_;

typedef struct {
   int programcounter;
   NODE* left;
}NOT_;

typedef struct  {
   int programcounter;
   NODE* bexp;
   NODE* ifstm;
   NODE* elsestm;
}IFSTM;

typedef struct {
   int programcounter;
   NODE* bexp;
   NODE* dostm;
}WHILESTM;

typedef struct  {
   int programcounter;
   int value;
}INTEGERCONST;

typedef struct  {
   int programcounter;
   char* value;
}STRCONST;

typedef struct  {
   int programcounter;
   SymbolEntry *symboltableentry;
   char name[MAX_ID];
}IDENTIFIER;

typedef struct {
   int programcounter;
   NODE* first;
   NODE* last;
}PATH;

typedef struct  {
  int programcounter;
  NODE* path;
}ALLOC;

typedef struct {
  int programcounter;
  NODE* path;
}DELETE_;



/* gibt den Parserbaum abstree aus */
void pretty_print(const NODE *abstree);


/* gibt den Typ des Knoten aus */
void printNodeType(const NODE *node);


/* Aufbau der Symboltabelle*/
/* Typueberpruefung */
int buildSymbolTable(SymbolTable *symboltable, const NODE *node);


/*
   int getLastPath(SymbolTable* symboltable,  NODE *node, char **varname, char **recordname);
   gibt Eigenschaften (Record-Name, Komponente) der letzten Pfadkomponente
*/
int getLastPath(SymbolTable* symboltable, NODE *node, char **varname, char **recordname);


/*
   RecordEntry* getLastPathComp(SymbolTable* symboltable,  NODE *node);
   gibt Record-Eintrag der letzten Pfadkomponente
*/
RecordEntry* getLastPathComp(SymbolTable* symboltable, NODE *path);


/*
  int getNPath(SymbolTable* symboltable, NODE *node, int length, char **varname, char **recordname);
  gibt Eigenschaften (Record-Name, Komponente) der N-te Pfadkomponente
*/
int getNPath(SymbolTable* symboltable, NODE *node, int length, char **varname, char **recordname);



/* int getPathLength(NODE *node) */
/* gibt die Länge des Pfades zurück */
/* 0 bei Fehler */
int getPathLength(NODE *node);





/*
char* getNodeType(SymbolTable *symboltable,  NODE *node);

T_PLUSOP, T_MINUSOP, T_MULTOP, T_DIVOP, T_MODOP, T_NEGATE, T_INTEGERCONST, T_IDENTIFIER (falls vom Typ INTEGER), 
T_PATH (falls letzte Pfadkomponente vom Typ INTEGER) ---> "INT"

T_STRCONST, T_IDENTIFIER (falls vom Typ STRING), T_PATH (falls letzte Pfadkomponente vom Typ STRING)
---> "STR"

T_NULL ---> "NULL" 

T_IDENTIFIER (falls vom Typ POINTER), T_PATH (falls vom Typ POINTER), T_NEW
---> "PTOREC [recordname]
*/
char* getNodeType(SymbolTable *symboltable, NODE *node);



/*
void controlBEXP(SymbolTable *symboltable, NODE *node, int &pointer, int &null, int &var);
zählt die Anzahl der pointer, null, und var in der Bedingung der IF-Anweisung
pointer: Anzahl der Pointer-Variablen
null: Anzahl von NULL
var: Anzahl der Integer-, String-Variablen 
*/
void controlBEXP(SymbolTable *symboltable, NODE *node, int &pointer, int &null, int &var);



/*
void freeAbstree(NODE* node)
löscht den abstrakten Parserbaum 
*/
void freeAbstree(NODE* node);

#endif
