#ifndef _AEXP_H
#define _AEXP_H


#include "scenario.h"
#include <list>

#define STR_ASGN         "="
#define STR_ADD          "+"
#define STR_MIN          "-"
#define STR_MULT         "*"
#define STR_DIV          "/"
#define STR_MOD          "%"

namespace arithmetictree {

typedef list <char*> CHARLIST;
enum MODE {PREORDER, POSTORDER, INORDER};


class Node
{
private:
   //char *arithexp;
public:
   virtual ~Node() {}
   virtual void getNumNode(list<CHARLIST> &list) const=0;
   virtual void getNumConstNode(list<int> &list) const=0;
   virtual void printTree(MODE mode=INORDER) const=0;
   virtual char* getArithmeticExpression()=0;
   virtual void changeNumNode(list<char*> changenode, list<char*> newnode) =0;

};


class NumConstNode: public Node
{
public:
   NumConstNode (int num) : value (num) {}
   void getNumNode(list<CHARLIST> &list) const {};
   void getNumConstNode(list<int> &list) const;
   void changeNumNode(list<char*> changenode, list<char*> newnode) {};
   void printTree(MODE mode=INORDER) const;
   char* getArithmeticExpression();
private:
   const int value;
};


class NumNode: public Node
{
public:
   NumNode(list <char*> path_) : path(path_) {}
   void getNumNode(list<CHARLIST> &list) const;
   void getNumConstNode(list<int> &list) const {};
   void printTree(MODE mode=INORDER) const;
   void changeNumNode(list<char*> changenode, list<char*> newnode);
   char* getArithmeticExpression();
private:
   list <char*> path;
};


class TwoOpNode: public Node
{
public:
   TwoOpNode(Node *_pLeft, Node *_pRight): pLeft (_pLeft), pRight(_pRight) {}
   ~TwoOpNode();
protected:
   Node* const pLeft;
   Node* const pRight;
};


class AddNode: public TwoOpNode
{
public:
   AddNode(Node *pLeft, Node *pRight) : TwoOpNode (pLeft, pRight) {}
   void getNumNode(list<CHARLIST> &list) const;
   void getNumConstNode(list<int> &list) const;
   void printTree(MODE mode=INORDER) const;
   void changeNumNode(list<char*> changenode, list<char*> newnode);
   char* getArithmeticExpression();
};


class MinNode: public TwoOpNode
{
public:
   MinNode(Node *pLeft, Node *pRight) : TwoOpNode (pLeft, pRight) {}
   void getNumNode(list<CHARLIST> &list) const;
   void getNumConstNode(list<int> &list) const;
   void printTree(MODE mode=INORDER) const;
   void changeNumNode(list<char*> changenode, list<char*> newnode);
   char* getArithmeticExpression();
};


class MultNode: public TwoOpNode
{
public:
   MultNode (Node *pLeft, Node *pRight) : TwoOpNode (pLeft, pRight) {}
   void getNumNode(list<CHARLIST> &list) const;
   void getNumConstNode(list<int> &list) const;
   void printTree(MODE mode=INORDER) const;
   void changeNumNode(list<char*> changenode, list<char*> newnode);
   char* getArithmeticExpression();
};


class DivNode: public TwoOpNode
{
public:
   DivNode (Node *pLeft, Node *pRight) : TwoOpNode (pLeft, pRight) {}
   void getNumNode(list<CHARLIST> &list) const;
   void getNumConstNode(list<int> &list) const;
   void printTree(MODE mode=INORDER) const;
   void changeNumNode(list<char*> changenode, list<char*> newnode);
   char* getArithmeticExpression();
};

class ModNode: public TwoOpNode
{
public:
   ModNode (Node *pLeft, Node *pRight) : TwoOpNode (pLeft, pRight) {}
   void getNumNode(list<CHARLIST> &list) const;
   void getNumConstNode(list<int> &list) const;
   void printTree(MODE mode=INORDER) const;
   void changeNumNode(list<char*> changenode, list<char*> newnode);
   char* getArithmeticExpression();
};

}



#endif
