#include <iostream>
#include <string.h>
#include "booltree.h"
#include "helpfunction.h"

using namespace std;
using namespace booltree;

/*                    */
/* ---NumConstNode--- */
/*                    */
void NumConstNode::printTree() const
{
   cout <<value;
}

char* NumConstNode::getExpression()
{
   char *strvalue=new char[15];
   sprintf(strvalue, "%d\0", value);
   return strvalue;
}

void NumConstNode::getNumConstNode(list<int> &list) const {
   list.push_back(value);
}

/*                    */
/* ---StrConstNode--- */
/*                    */
void StrConstNode::printTree() const
{
   cout <<value;
}

char *StrConstNode::getExpression()
{
   char *strvalue=new char[15];
   sprintf(strvalue, "%d\0", value);
   return value;
}

void StrConstNode::getStrConstNode(list<char*> &list) const 
{
   list.push_back(value);
}

/*                      */
/* ---IdentifierNode--- */
/*                      */
void IdentifierNode::printTree() const
{
   printcharlist(path);
}

void IdentifierNode::changeIdentifierNode(list<char*> changenode, list<char*> newnode)
{
   if (pathcmp(changenode, path) ==0) {
      list<char*>::iterator itr_begin = newnode.begin();
      list<char*>::iterator itr_end=newnode.end();
      path.clear();
      path.assign(newnode.begin(),newnode.end());
   }
}

char *IdentifierNode::getExpression()
{
   list<char*>::iterator itr=path.begin();
   char *returnstring=new char[strlen(*itr)+1];
   returnstring=*itr;
   itr++;
   while (itr != path.end())
   {
      returnstring=catstrstr(returnstring, *itr);
      itr++;
   }
   return returnstring;
}

void IdentifierNode::getIdentifierNode(list<CHARLIST> &list) const 
{
   list.push_back(path);
}

/*                 */
/* ---TwoOpNode--- */
/*                 */
TwoOpNode::~TwoOpNode()
{
   delete pLeft;
   delete pRight;
}

OneOpNode::~ OneOpNode()
{
  delete pExp;
}


/*                 */
/* ---EqualNode--- */
/*                 */
void EqualNode::printTree() const
{
   cout <<"(";
   pLeft->printTree();
   cout <<STR_EQUAL;
   pRight->printTree();
   cout <<")";
}

void EqualNode::changeIdentifierNode(list<char*> changenode, list<char*> newnode) {
   pLeft->changeIdentifierNode(changenode, newnode);
   pRight->changeIdentifierNode(changenode, newnode);
}

char *EqualNode::getExpression() 
{
  char *str1=catstrstr("(", pLeft->getExpression());
  char *str2=catstrstr(str1, STR_EQUAL);
  char *str3=catstrstr(str2, pRight->getExpression());
  return catstrstr(str3, ")");
}

void EqualNode::getIdentifierNode(list<CHARLIST> &list) const
{
   pLeft->getIdentifierNode(list);
   pRight->getIdentifierNode(list);
}

void EqualNode::getNumConstNode(list<int> &list) const
{
   pLeft->getNumConstNode(list);
   pRight->getNumConstNode(list);
}

void EqualNode::getStrConstNode(list<char*> &list) const
{
   pLeft->getStrConstNode(list);
   pRight->getStrConstNode(list);
}


/*                    */
/* ---NotEqualNode--- */
/*                    */
void NotEqualNode::printTree() const
{
   cout <<"(";
   pLeft->printTree();
   cout <<STR_NOTEQUAL;
   pRight->printTree();
   cout <<")";
}

void NotEqualNode::changeIdentifierNode(list<char*> changenode, list<char*> newnode) {
   pLeft->changeIdentifierNode(changenode, newnode);
   pRight->changeIdentifierNode(changenode, newnode);
}

char *NotEqualNode::getExpression() {
  char *str1=catstrstr("(", pLeft->getExpression());
  char *str2=catstrstr(str1, STR_NOTEQUAL);
  char *str3=catstrstr(str2, pRight->getExpression());
  return catstrstr(str3, ")");
}

void NotEqualNode::getIdentifierNode(list<CHARLIST> &list) const
{
   pLeft->getIdentifierNode(list);
   pRight->getIdentifierNode(list);
}

void NotEqualNode::getNumConstNode(list<int> &list) const
{
   pLeft->getNumConstNode(list);
   pRight->getNumConstNode(list);
}

void NotEqualNode::getStrConstNode(list<char*> &list) const
{
   pLeft->getStrConstNode(list);
   pRight->getStrConstNode(list);
}


/*                    */
/* ---SmallerNode---  */
/*                    */
void SmallerNode::printTree() const
{
   cout <<"(";
   pLeft->printTree();
   cout <<STR_SMALLER;
   pRight->printTree();
   cout <<")";
}

void SmallerNode::changeIdentifierNode(list<char*> changenode, list<char*> newnode) {
   pLeft->changeIdentifierNode(changenode, newnode);
   pRight->changeIdentifierNode(changenode, newnode);
}


char *SmallerNode::getExpression() {
  char *str1=catstrstr("(", pLeft->getExpression());
  char *str2=catstrstr(str1, STR_SMALLER);
  char *str3=catstrstr(str2, pRight->getExpression());
  return catstrstr(str3, ")");
}

void SmallerNode::getIdentifierNode(list<CHARLIST> &list) const
{
   pLeft->getIdentifierNode(list);
   pRight->getIdentifierNode(list);
}

void SmallerNode::getNumConstNode(list<int> &list) const
{
   pLeft->getNumConstNode(list);
   pRight->getNumConstNode(list);
}

void SmallerNode::getStrConstNode(list<char*> &list) const
{
   pLeft->getStrConstNode(list);
   pRight->getStrConstNode(list);
}

/*                   */
/* ---GreaterNode--- */
/*                   */
void GreaterNode::printTree() const
{
   cout <<"(";
   pLeft->printTree();
   cout <<STR_GREATER;
   pRight->printTree();
   cout <<")";
}

void GreaterNode::changeIdentifierNode(list<char*> changenode, list<char*> newnode) {
   pLeft->changeIdentifierNode(changenode, newnode);
   pRight->changeIdentifierNode(changenode, newnode);
}

char *GreaterNode::getExpression() {
  char *str1=catstrstr("(", pLeft->getExpression());
  char *str2=catstrstr(str1, STR_GREATER);
  char *str3=catstrstr(str2, pRight->getExpression());
  return catstrstr(str3, ")");
}

void GreaterNode::getIdentifierNode(list<CHARLIST> &list) const
{
   pLeft->getIdentifierNode(list);
   pRight->getIdentifierNode(list);
}

void GreaterNode::getNumConstNode(list<int> &list) const
{
   pLeft->getNumConstNode(list);
   pRight->getNumConstNode(list);
}

void GreaterNode::getStrConstNode(list<char*> &list) const
{
   pLeft->getStrConstNode(list);
   pRight->getStrConstNode(list);
}

/*                        */
/* ---SmallerEqualNode--- */
/*                        */
void SmallerEqualNode::printTree() const
{
   cout <<"(";
   pLeft->printTree();
   cout <<STR_SMALLEREQUAL;
   pRight->printTree();
   cout <<")";
}

void SmallerEqualNode::changeIdentifierNode(list<char*> changenode, list<char*> newnode) {
   pLeft->changeIdentifierNode(changenode, newnode);
   pRight->changeIdentifierNode(changenode, newnode);
}

char *SmallerEqualNode::getExpression() {
  char *str1=catstrstr("(", pLeft->getExpression());
  char *str2=catstrstr(str1, STR_SMALLEREQUAL);
  char *str3=catstrstr(str2, pRight->getExpression());
  return catstrstr(str3, ")");
}

void SmallerEqualNode::getIdentifierNode(list<CHARLIST> &list) const
{
   pLeft->getIdentifierNode(list);
   pRight->getIdentifierNode(list);
}

void SmallerEqualNode::getNumConstNode(list<int> &list) const
{
   pLeft->getNumConstNode(list);
   pRight->getNumConstNode(list);
}

void SmallerEqualNode::getStrConstNode(list<char*> &list) const
{
   pLeft->getStrConstNode(list);
   pRight->getStrConstNode(list);
}

/*                        */
/* ---GreaterEqualNode--- */
/*                        */
void GreaterEqualNode::printTree() const
{
   cout <<"(";
   pLeft->printTree();
   cout <<STR_GREATEREQUAL;
   pRight->printTree();
   cout <<")";
}

void GreaterEqualNode::changeIdentifierNode(list<char*> changenode, list<char*> newnode) {
   pLeft->changeIdentifierNode(changenode, newnode);
   pRight->changeIdentifierNode(changenode, newnode);
}

char *GreaterEqualNode::getExpression() {
  char *str1=catstrstr("(", pLeft->getExpression());
  char *str2=catstrstr(str1, STR_GREATEREQUAL);
  char *str3=catstrstr(str2, pRight->getExpression());
  return catstrstr(str3, ")");
}

void GreaterEqualNode::getIdentifierNode(list<CHARLIST> &list) const
{
   pLeft->getIdentifierNode(list);
   pRight->getIdentifierNode(list);
}

void GreaterEqualNode::getNumConstNode(list<int> &list) const
{
   pLeft->getNumConstNode(list);
   pRight->getNumConstNode(list);
}

void GreaterEqualNode::getStrConstNode(list<char*> &list) const
{
   pLeft->getStrConstNode(list);
   pRight->getStrConstNode(list);
}

/*               */
/* ---AndNode--- */
/*               */
void AndNode::printTree() const
{
   cout <<"(";
   pLeft->printTree();
   cout <<STR_AND;
   pRight->printTree();
   cout <<")";
}

void AndNode::changeIdentifierNode(list<char*> changenode, list<char*> newnode) {
   pLeft->changeIdentifierNode(changenode, newnode);
   pRight->changeIdentifierNode(changenode, newnode);
}

char *AndNode::getExpression() {
  char *str1=catstrstr("(", pLeft->getExpression());
  char *str2=catstrstr(str1, STR_AND);
  char *str3=catstrstr(str2, pRight->getExpression());
  return catstrstr(str3, ")");
}

void AndNode::getIdentifierNode(list<CHARLIST> &list) const
{
   pLeft->getIdentifierNode(list);
   pRight->getIdentifierNode(list);
}

void AndNode::getNumConstNode(list<int> &list) const
{
   pLeft->getNumConstNode(list);
   pRight->getNumConstNode(list);
}

void AndNode::getStrConstNode(list<char*> &list) const
{
   pLeft->getStrConstNode(list);
   pRight->getStrConstNode(list);
}

/*               */
/* ---OrNode---  */
/*               */
void OrNode::changeIdentifierNode(list<char*> changenode, list<char*> newnode) {
   pLeft->changeIdentifierNode(changenode, newnode);
   pRight->changeIdentifierNode(changenode, newnode);
}

void OrNode::printTree() const
{
   cout <<"(";
   pLeft->printTree();
   cout <<STR_OR;
   pRight->printTree();
   cout <<")";
}

char *OrNode::getExpression() {
  char *str1=catstrstr("(", pLeft->getExpression());
  char *str2=catstrstr(str1, STR_OR);
  char *str3=catstrstr(str2, pRight->getExpression());
  return catstrstr(str3, ")");
}

void OrNode::getIdentifierNode(list<CHARLIST> &list) const
{
   pLeft->getIdentifierNode(list);
   pRight->getIdentifierNode(list);
}

void OrNode::getNumConstNode(list<int> &list) const
{
   pLeft->getNumConstNode(list);
   pRight->getNumConstNode(list);
}

void OrNode::getStrConstNode(list<char*> &list) const
{
   pLeft->getStrConstNode(list);
   pRight->getStrConstNode(list);
}

/*               */
/* ---NotNode--- */
/*               */
void NotNode::printTree() const
{
   cout <<STR_NOT <<"("; 
   pExp->printTree();
   cout <<")";
}

void NotNode::changeIdentifierNode(list<char*> changenode, list<char*> newnode) {
   pExp->changeIdentifierNode(changenode, newnode);
}

char *NotNode::getExpression() {
  char *str1=catstrstr(STR_NOT, "(");
  char *str2=catstrstr(str1, pExp->getExpression());
  char *str3=catstrstr(str2, ")");
  return str3;
}

void NotNode::getIdentifierNode(list<CHARLIST> &list) const
{
   pExp->getIdentifierNode(list);
}

void NotNode::getNumConstNode(list<int> &list) const
{
   pExp->getNumConstNode(list);
}

void NotNode::getStrConstNode(list<char*> &list) const
{
   pExp->getStrConstNode(list);
}

/*               */
/* ---AddNode--- */
/*               */
void AddNode::changeIdentifierNode(list<char*> changenode, list<char*> newnode) {
   pLeft->changeIdentifierNode(changenode, newnode);
   pRight->changeIdentifierNode(changenode, newnode);
}

void AddNode::printTree() const
{
   cout <<"(";
   pLeft->printTree();
   cout <<STR_ADD;
   pRight->printTree();
   cout <<")";
}

char *AddNode::getExpression() {
  char *str1=catstrstr("(", pLeft->getExpression());
  char *str2=catstrstr(str1, STR_ADD);
  char *str3=catstrstr(str2, pRight->getExpression());
  return catstrstr(str3, ")");
}

void AddNode::getIdentifierNode(list<CHARLIST> &list) const
{
   pLeft->getIdentifierNode(list);
   pRight->getIdentifierNode(list);
}

void AddNode::getNumConstNode(list<int> &list) const
{
   pLeft->getNumConstNode(list);
   pRight->getNumConstNode(list);
}

void AddNode::getStrConstNode(list<char*> &list) const
{
   pLeft->getStrConstNode(list);
   pRight->getStrConstNode(list);
}

/*               */
/* ---MinNode--- */
/*               */
void MinNode::changeIdentifierNode(list<char*> changenode, list<char*> newnode) {
   pLeft->changeIdentifierNode(changenode, newnode);
   pRight->changeIdentifierNode(changenode, newnode);
}

void MinNode::printTree() const
{
   cout <<"(";
   pLeft->printTree();
   cout <<STR_MIN;
   pRight->printTree();
   cout <<")";
}

char *MinNode::getExpression() {
  char *str1=catstrstr("(", pLeft->getExpression());
  char *str2=catstrstr(str1, STR_MIN);
  char *str3=catstrstr(str2, pRight->getExpression());
  return catstrstr(str3, ")");
}

void MinNode::getIdentifierNode(list<CHARLIST> &list) const
{
   pLeft->getIdentifierNode(list);
   pRight->getIdentifierNode(list);
}

void MinNode::getNumConstNode(list<int> &list) const
{
   pLeft->getNumConstNode(list);
   pRight->getNumConstNode(list);
}

void MinNode::getStrConstNode(list<char*> &list) const
{
   pLeft->getStrConstNode(list);
   pRight->getStrConstNode(list);
}

/*                */
/* ---MultNode--- */
/*                */
void MultNode::changeIdentifierNode(list<char*> changenode, list<char*> newnode) {
   pLeft->changeIdentifierNode(changenode, newnode);
   pRight->changeIdentifierNode(changenode, newnode);
}

void MultNode::printTree() const
{
   cout <<"(";
   pLeft->printTree();
   cout <<STR_MULT;
   pRight->printTree();
   cout <<")";
}

char *MultNode::getExpression() {
  char *str1=catstrstr("(", pLeft->getExpression());
  char *str2=catstrstr(str1, STR_MULT);
  char *str3=catstrstr(str2, pRight->getExpression());
  return catstrstr(str3, ")");
}

void MultNode::getIdentifierNode(list<CHARLIST> &list) const
{
   pLeft->getIdentifierNode(list);
   pRight->getIdentifierNode(list);
}

void MultNode::getNumConstNode(list<int> &list) const
{
   pLeft->getNumConstNode(list);
   pRight->getNumConstNode(list);
}

void MultNode::getStrConstNode(list<char*> &list) const
{
   pLeft->getStrConstNode(list);
   pRight->getStrConstNode(list);
}

/*               */
/* ---DivNode--- */
/*               */
void DivNode::changeIdentifierNode(list<char*> changenode, list<char*> newnode) {
   pLeft->changeIdentifierNode(changenode, newnode);
   pRight->changeIdentifierNode(changenode, newnode);
}

void DivNode::printTree() const
{
   cout <<"(";
   pLeft->printTree();
   cout <<STR_DIV;
   pRight->printTree();
   cout <<")";
}

char *DivNode::getExpression() {
  char *str1=catstrstr("(", pLeft->getExpression());
  char *str2=catstrstr(str1, STR_DIV);
  char *str3=catstrstr(str2, pRight->getExpression());
  return catstrstr(str3, ")");
}

void DivNode::getIdentifierNode(list<CHARLIST> &list) const
{
   pLeft->getIdentifierNode(list);
   pRight->getIdentifierNode(list);
}

void DivNode::getNumConstNode(list<int> &list) const
{
   pLeft->getNumConstNode(list);
   pRight->getNumConstNode(list);
}

void DivNode::getStrConstNode(list<char*> &list) const
{
   pLeft->getStrConstNode(list);
   pRight->getStrConstNode(list);
}

/*               */
/* ---ModNode--- */
/*               */
void ModNode::changeIdentifierNode(list<char*> changenode, list<char*> newnode) {
   pLeft->changeIdentifierNode(changenode, newnode);
   pRight->changeIdentifierNode(changenode, newnode);
}

void ModNode::printTree() const
{
   cout <<"(";
   pLeft->printTree();
   cout <<STR_MOD;
   pRight->printTree();
   cout <<")";
}

char *ModNode::getExpression() {
  char *str1=catstrstr("(", pLeft->getExpression());
  char *str2=catstrstr(str1, STR_MOD);
  char *str3=catstrstr(str2, pRight->getExpression());
  return catstrstr(str3, ")");
}

void ModNode::getIdentifierNode(list<CHARLIST> &list) const
{
   pLeft->getIdentifierNode(list);
   pRight->getIdentifierNode(list);
}

void ModNode::getNumConstNode(list<int> &list) const
{
   pLeft->getNumConstNode(list);
   pRight->getNumConstNode(list);
}

void ModNode::getStrConstNode(list<char*> &list) const
{
   pLeft->getStrConstNode(list);
   pRight->getStrConstNode(list);
}
