#ifndef _BEXP_H
#define _BEXP_H

#include "scenario.h"
#include "scenario.h"
#include <list>

#define STR_EQUAL        "="
#define STR_NOTEQUAL     "!="
#define STR_SMALLER      "<"
#define STR_GREATER      ">"
#define STR_SMALLEREQUAL "<="
#define STR_GREATEREQUAL ">="
#define STR_AND          " AND "
#define STR_OR           " OR "
#define STR_NOT          "NOT"
#define STR_ADD          "+"
#define STR_MIN          "-"
#define STR_MULT         "*"
#define STR_DIV          "/"
#define STR_MOD          "%"

namespace booltree {

typedef list <char*> CHARLIST;


class Node
{
private:
   char *boolexp;
public:
   virtual ~Node() {}
   virtual void printTree() const=0;
   virtual void changeIdentifierNode(list<char*> changenode, list<char*> newnode)=0;
   virtual char *getExpression()=0;
   virtual void getIdentifierNode(list<CHARLIST> &list) const=0;
   virtual void getNumConstNode(list<int> &list) const=0;
   virtual void getStrConstNode(list<char*> &list) const=0;


};


class NumConstNode: public Node
{
public:
   NumConstNode (int num) : value (num) {}
   void printTree() const;
   void changeIdentifierNode(list<char*> changenode, list<char*> newnode) {};
   char *getExpression();
   void getIdentifierNode(list<CHARLIST> &list) const {};
   void getNumConstNode(list<int> &list) const;
   void getStrConstNode(list<char*> &list) const {};


private:
   const int value;
};


class StrConstNode: public Node
{
public:
   ~StrConstNode() {delete value;}
   StrConstNode (char *value_) : value(value_) {}
   void printTree() const;
   void changeIdentifierNode(list<char*> changenode, list<char*> newnode) {};
   char *getExpression();
   void getIdentifierNode(list<CHARLIST> &list) const {};
   void getNumConstNode(list<int> &list) const {};
   void getStrConstNode(list<char*> &list) const;
private:
   char *value;
};


class IdentifierNode: public Node
{
public:
   IdentifierNode(list <char*> path_) : path(path_) {}
   void printTree() const;
   void changeIdentifierNode(list<char*> changenode, list<char*> newnode);
   char *getExpression();
   void getIdentifierNode(list<CHARLIST> &list) const;
   void getNumConstNode(list<int> &list) const {};
   void getStrConstNode(list<char*> &list) const {};
private:
   list <char*> path;
};



class TwoOpNode: public Node
{
public:
   TwoOpNode(Node *_pLeft, Node *_pRight): pLeft (_pLeft), pRight(_pRight) {}
   ~TwoOpNode();
protected:
   Node* const pLeft;
   Node* const pRight;
};



class OneOpNode: public Node
{
public:
   OneOpNode(Node *_pExp): pExp(_pExp) {}
   ~OneOpNode();
protected:
   Node* const pExp;
};



class EqualNode: public TwoOpNode
{
public:
   EqualNode(Node *pLeft, Node *pRight) : TwoOpNode (pLeft, pRight) {};
   void printTree() const;
   void changeIdentifierNode(list<char*> changenode, list<char*> newnode);
   char *getExpression();
   void getIdentifierNode(list<CHARLIST> &list) const;
   void getNumConstNode(list<int> &list) const;
   void getStrConstNode(list<char*> &list) const;
};


class NotEqualNode: public TwoOpNode
{
public:
   NotEqualNode(Node *pLeft, Node *pRight) : TwoOpNode (pLeft, pRight) {};
   void printTree() const;
   void changeIdentifierNode(list<char*> changenode, list<char*> newnode);
   char *getExpression();
   void getIdentifierNode(list<CHARLIST> &list) const;
   void getNumConstNode(list<int> &list) const;
   void getStrConstNode(list<char*> &list) const;
};


class SmallerNode: public TwoOpNode
{
public:
   SmallerNode(Node *pLeft, Node *pRight) : TwoOpNode (pLeft, pRight) {};
   void printTree() const;
   void changeIdentifierNode(list<char*> changenode, list<char*> newnode);
   char *getExpression();
   void getIdentifierNode(list<CHARLIST> &list) const;
   void getNumConstNode(list<int> &list) const;
   void getStrConstNode(list<char*> &list) const;
};


class GreaterNode: public TwoOpNode
{
public:
   GreaterNode(Node *pLeft, Node *pRight) : TwoOpNode (pLeft, pRight) {};
   void printTree() const;
   void changeIdentifierNode(list<char*> changenode, list<char*> newnode);
   char *getExpression();
   void getIdentifierNode(list<CHARLIST> &list) const;
   void getNumConstNode(list<int> &list) const;
   void getStrConstNode(list<char*> &list) const;
};


class SmallerEqualNode: public TwoOpNode
{
public:
   SmallerEqualNode(Node *pLeft, Node *pRight) : TwoOpNode (pLeft, pRight) {};
   void printTree() const;
   void changeIdentifierNode(list<char*> changenode, list<char*> newnode);
   char *getExpression();
   void getIdentifierNode(list<CHARLIST> &list) const;
   void getNumConstNode(list<int> &list) const;
   void getStrConstNode(list<char*> &list) const;
};

class GreaterEqualNode: public TwoOpNode
{
public:
   GreaterEqualNode(Node *pLeft, Node *pRight) : TwoOpNode (pLeft, pRight) {};
   void printTree() const;
   void changeIdentifierNode(list<char*> changenode, list<char*> newnode);
   char *getExpression();
   void getIdentifierNode(list<CHARLIST> &list) const;
   void getNumConstNode(list<int> &list) const;
   void getStrConstNode(list<char*> &list) const;
};



class AndNode: public TwoOpNode
{

public:
   AndNode(Node *pLeft, Node *pRight) : TwoOpNode (pLeft, pRight) {};
   void printTree() const;
   void changeIdentifierNode(list<char*> changenode, list<char*> newnode);
   char *getExpression();
   void getIdentifierNode(list<CHARLIST> &list) const;
   void getNumConstNode(list<int> &list) const;
   void getStrConstNode(list<char*> &list) const;
};


class OrNode: public TwoOpNode
{
public:
   OrNode(Node *pLeft, Node *pRight) : TwoOpNode (pLeft, pRight) {};
   void printTree() const;
   void changeIdentifierNode(list<char*> changenode, list<char*> newnode);
   char *getExpression();
   void getIdentifierNode(list<CHARLIST> &list) const;
   void getNumConstNode(list<int> &list) const;
   void getStrConstNode(list<char*> &list) const;
};


class NotNode: public OneOpNode
{
public:
   NotNode(Node *pExp) : OneOpNode (pExp) {};
   void printTree() const;
   void changeIdentifierNode(list<char*> changenode, list<char*> newnode);
   char *getExpression();
   void getIdentifierNode(list<CHARLIST> &list) const;
   void getNumConstNode(list<int> &list) const;
   void getStrConstNode(list<char*> &list) const;
};


class AddNode: public TwoOpNode
{
public:
   AddNode(Node *pLeft, Node *pRight) : TwoOpNode (pLeft, pRight) {};
   void printTree() const;
   void changeIdentifierNode(list<char*> changenode, list<char*> newnode);
   char *getExpression();
   void getIdentifierNode(list<CHARLIST> &list) const;
   void getNumConstNode(list<int> &list) const;
   void getStrConstNode(list<char*> &list) const;
};

class MinNode: public TwoOpNode
{
public:
   MinNode(Node *pLeft, Node *pRight) : TwoOpNode (pLeft, pRight) {};
   void printTree() const;
   void changeIdentifierNode(list<char*> changenode, list<char*> newnode);
   char *getExpression();
   void getIdentifierNode(list<CHARLIST> &list) const;
   void getNumConstNode(list<int> &list) const;
   void getStrConstNode(list<char*> &list) const;
};

class MultNode: public TwoOpNode
{
public:
   MultNode(Node *pLeft, Node *pRight) : TwoOpNode (pLeft, pRight) {};
   void printTree() const;
   void changeIdentifierNode(list<char*> changenode, list<char*> newnode);
   char *getExpression();
   void getIdentifierNode(list<CHARLIST> &list) const;
   void getNumConstNode(list<int> &list) const;
   void getStrConstNode(list<char*> &list) const;
};

class DivNode: public TwoOpNode
{
public:
   DivNode(Node *pLeft, Node *pRight) : TwoOpNode (pLeft, pRight) {};
   void printTree() const;
   void changeIdentifierNode(list<char*> changenode, list<char*> newnode);
   char *getExpression();
   void getIdentifierNode(list<CHARLIST> &list) const;
   void getNumConstNode(list<int> &list) const;
   void getStrConstNode(list<char*> &list) const;
};

class ModNode: public TwoOpNode
{
public:
   ModNode(Node *pLeft, Node *pRight) : TwoOpNode (pLeft, pRight) {};
   void printTree() const;
   void changeIdentifierNode(list<char*> changenode, list<char*> newnode);
   char *getExpression();
   void getIdentifierNode(list<CHARLIST> &list) const;
   void getNumConstNode(list<int> &list) const;
   void getStrConstNode(list<char*> &list) const;
};


}

#endif
