#include <stdio.h>
#include <math.h>
#include <string.h>
#include <malloc.h>

#include "symboltable.h"


/*
   int hashFunction(const char* string) 
*/
int hashFunction(const char *string)
{
  const char *p;
  unsigned h=0,g;
  for(p=string;*p!='\0';p++){
    h=(h<<4)+(*p);
    if((g=h &(unsigned) 0xf0000000)!=0){
      h=h ^ (g>>24);
      h=h ^ g;
    }
  }
  return ((int)( h % HashSize));
}





/*
   Symboltable * createSymbolTable() 
   erstellt eine Symboltabelle und gibt pointer 
   auf diese Tabelle zurueck 
*/
SymbolTable* createSymbolTable() 
{
   SymbolTable* symboltable=NULL;
   symboltable=(SymbolTable*)malloc(sizeof(SymbolTable));
   int i=0;
   for (i=0; i<HashSize; i++)
      symboltable->table[i]=NULL;
   return symboltable;
}





/* 
   int putIntegerSymbol(SymbolTable *symboltable, char *name)
   traegt die Integer-Variable name in die Symboltabelle symboltable ein
   Rueckgabewerte:  
   -1 Symboltabelle symboltable ist null
    0 die Integer-Variable name in der Tabelle symboltable schon vorhanden
    1 die Integer-Variable name wurde in die Symboltabelle eingefuehrt
*/
int putIntegerSymbol(SymbolTable *symboltable, char* name)
{
   if (symboltable==NULL)
      return -1;

   SymbolEntry *help;

   int index=hashFunction(name);
   help = symboltable->table[index];

   while (symboltable->table[index]!=NULL) {
      if(strcmp(symboltable->table[index]->name,name)==0) {
         symboltable->table[index]=help;
         return 0;
      }
      symboltable->table[index] = symboltable->table[index]->next;

   }

   if (symboltable->table[index] == NULL) {
      symboltable->table[index] = (SymbolEntry*)malloc(sizeof(SymbolEntry));
      symboltable->table[index]->name=name;
      symboltable->table[index]->type=intType;
      symboltable->table[index]->next = NULL;
      if (help != NULL) {
         symboltable->table[index]->next = help;
      }
   }
   return 1;
}





/*
   int putStringSymbol(SymbolTable *symboltable, char *name)
   traegt die String-Variable name in die Symboltabelle symboltable ein
   Rueckgabewerte:
   -1 Symboltabelle symboltable ist null
    0 die String-Variable name in der Tabelle symboltable schon vorhanden
    1 die String-Variable name wurde in die Symboltabelle eingefuehrt
*/
int putStringSymbol(SymbolTable *symboltable, char* name)
{
   if (symboltable==NULL)
      return -1;

   SymbolEntry *help;
  
   int index=hashFunction(name);
   help = symboltable->table[index];

   while (symboltable->table[index]!=NULL) {
      if(strcmp(symboltable->table[index]->name,name)==0) {
         symboltable->table[index]=help;
         return 0;
      }
      symboltable->table[index] = symboltable->table[index]->next;

   }

   if (symboltable->table[index] == NULL) {
      symboltable->table[index] = (SymbolEntry*)malloc(sizeof(SymbolEntry));
      symboltable->table[index]->name=name;
      symboltable->table[index]->type=strType;
      symboltable->table[index]->next = NULL;
      if (help != NULL) {
         symboltable->table[index]->next = help;
      }
   }
   return 1;
}





/* int putPointerSymbol(SymbolTable *symboltable, char *name, char *recordname)
   traegt die Pointer-Variable name, die auf den record recordname zeigt in die 
   Symboltabelle symboltable ein
   Rueckgabewerte:
   -1 Symboltabelle symboltable ist null
   -2 Record recordname in der Symboltabelle symboltable nicht vorhanden
    0 die Pointer-Variable name in der Tabelle symboltable schon vorhanden
    1 die Pointer-Variable name wurde in die Symboltabelle symboltable eingefuehrt
*/
int putPointerSymbol(SymbolTable *symboltable, char *name, char *recordname)
{
   if (symboltable==NULL)
      return -1;

   SymbolEntry *help=NULL;;

   help = getSymbol(symboltable, recordname);
   if (help == NULL)
      return -2;
   if (help->type != recType)
      return -2;
   help = NULL;

   int index=hashFunction(name);
   help = symboltable->table[index];

   while (symboltable->table[index]!=NULL) {
      if(strcmp(symboltable->table[index]->name,name)==0) {
         symboltable->table[index]=help;
         return 0;
      }
      symboltable->table[index] = symboltable->table[index]->next;

   }

   if (symboltable->table[index] == NULL) {
      symboltable->table[index] = (SymbolEntry*)malloc(sizeof(SymbolEntry));
      symboltable->table[index]->name=name;
      symboltable->table[index]->type=ptorecType;
      symboltable->table[index]->next = NULL;
      symboltable->table[index]->typref=(PointerEntry*)malloc(sizeof(PointerEntry));
      ((PointerEntry*)(symboltable->table[index]->typref))->recordname=recordname;
      if (help != NULL) {
         symboltable->table[index]->next = help;
      }
   }
   return 1;
}



/* 
   int putRecordSymbol(SymbolTable *symboltable, char *name)
   traegt den Record recordname in die Symboltabelle symboltable ein
   Rueckgabewerte:
   -1 Symboltabelle symboltable ist null
    0 Record name in der Tabelle symboltable schon vorhanden
    1 Record name wurde in die Tabelle symboltable eingefuehrt
*/
int putRecordSymbol(SymbolTable *symboltable, char *recordname)
{
   if (symboltable==NULL)
      return -1;

   SymbolEntry *help;
  
   int index=hashFunction(recordname);
   help = symboltable->table[index];

   while (symboltable->table[index]!=NULL) {
      if(strcmp(symboltable->table[index]->name,recordname)==0) {
         symboltable->table[index]=help;
         return 0;
      }
      symboltable->table[index] = symboltable->table[index]->next;

   }
   if (symboltable->table[index] == NULL) {
      symboltable->table[index] = (SymbolEntry*)malloc(sizeof(SymbolEntry));
      symboltable->table[index]->name=recordname;
      symboltable->table[index]->type=recType;
      symboltable->table[index]->next = NULL;
      symboltable->table[index]->typref = NULL;
      if (help != NULL) {
         symboltable->table[index]->next = help;
      }
   }
   return 1;
}





/* 
   int addIntRecordComp(SymbolTable *symboltable, char *name, char *recordname)
   ergaenzt den Record recordname um die Integervariable name
   Rueckgabewerte:
   -1: Symboltabelle symboltable ist null
   -2: Record recordname in der Tabelle symboltable nicht vorhanden
    0: Variable name in dem Record recordname schon vorhanden
    1: Record recordname wurde um die Integervariable name ergaenzt
*/
int addIntRecordComp(SymbolTable *symboltable, char *name, char *recordname)
{

   if (symboltable==NULL)
      return -1;

   SymbolEntry *se=NULL;
   se=getSymbol(symboltable, recordname);

   if (se == NULL)
      return -2;
   if (se->type != recType)
      return -2;

   RecordEntry *help = (RecordEntry*)(se->typref);
   RecordEntry *help2=help;

   while (help != NULL){
      if (strcmp(help->name, name)==0) 
         return 0;
      help2=help;
      help =help->next;
   }

   help=(RecordEntry*)malloc(sizeof(RecordEntry));
   help->name=name;
   help->type=intType;
   help->next = NULL;
   if (se->typref==NULL)
      se->typref=help;
   else
      help2->next=help;
   return 1;
}





/* 
   int addStrRecordComp(SymbolTable *symboltable, char *name, char *recordname) 
   ergaenzt den Record recordname um die Stringvariable name
   Rueckgabewerte:
   -1: Symboltabelle symboltable ist null
   -2: Record recordname in der Tabelle symboltable nicht vorhanden
    0: Variable name in dem Record recordname schon vorhanden
    1: Record recordname wurde um die Stringvariable name ergaenzt
*/
int addStrRecordComp(SymbolTable *symboltable, char *name, char *recordname)
{

   if (symboltable==NULL)
      return -1;

   SymbolEntry *se=NULL;
   se=getSymbol(symboltable, recordname);

   if (se == NULL)
      return -2;
   if (se->type != recType)
      return -2;
  
   RecordEntry *help = (RecordEntry*)(se->typref);
   RecordEntry *help2=help;

   while (help != NULL){
      if (strcmp(help->name, name)==0) 
         return 0;
      help2=help;
      help =help->next;
   }

   help=(RecordEntry*)malloc(sizeof(RecordEntry));
   help->name=name;
   help->type=strType;
   help->next = NULL;
   if (se->typref==NULL)
      se->typref=help;
   else
      help2->next=help;
   return 1;
}





/* 
   int addPointerRecordComp(SymbolTable *symboltable, char *name, char *ptorecname,char *recordname) 
   ergaenzt den Record recordname um die Pointervariable name
   name ist pointer to ptorecname
   Rueckgabewerte:
   -1: Symboltabelle symboltable ist null
   -2: Record recordname in der Tabelle symboltable nicht vorhanden
   -3: Record ptorecname in der Tabelle symboltable nicht vorhanden
    0: Variable name in dem Record recordname schon vorhanden
    1: Record recordname wurde um die Pointervariable name ergaenzt
*/
int addPointerRecordComp(SymbolTable *symboltable, char *name, char *ptorecname, char *recordname)
{
   if (symboltable==NULL)
      return -1;

   SymbolEntry *se=NULL;
   se=getSymbol(symboltable, recordname);

   if (se == NULL)
      return -2;
   if (se->type != recType)
      return -2;

   SymbolEntry *ptorec=NULL;
   ptorec=getSymbol(symboltable, ptorecname);
   if (ptorec == NULL)
      return -3;
   if (ptorec->type != recType)
      return -3;
   
   RecordEntry *help = (RecordEntry*)(se->typref);
   RecordEntry *help2=help;

   while (help != NULL){
      if (strcmp(help->name, name)==0) 
         return 0;
      help2=help;
      help =help->next;
   }

   help=(RecordEntry*)malloc(sizeof(RecordEntry));
   help->name=name;
   help->recordname=ptorecname;
   help->type=ptorecType;
   help->next = NULL;
   if (se->typref==NULL)
      se->typref=help;
   else
      help2->next=help;
   return 1;
}





/* 
   SymbolEntry *getSymbol(const SymbolTable *symboltable, const char *name)
   sucht nach dem Eintrag name in der Symboltabelle symboltable
   beim Treffer gibt den Zeiger auf den entsprechenden Eintrag zurueck
   sonst NULL
*/
SymbolEntry* getSymbol(const SymbolTable *symboltable, const char *name)
{
   SymbolEntry *help;
   int index=hashFunction(name);
   help = symboltable->table[index];

   while (help != NULL) {
      if (strcmp(help->name,name)==0) {
         return help;
      }
      help = help->next;
   }
   return NULL;
}





/* 
   RecordEntry *getRecordComp(const SymbolTable *symboltable, char *name, char *recordname)
   sucht nach der Recordcomponente name des Records recordname in der Symboltabelle symboltable
   beim Treffer gibt den Zeiger auf den entsprechenden Eintrag zurueck
   sonst NULL
*/
RecordEntry* getRecordComp(const SymbolTable *symboltable, const char *name, const char *recordname)
{
   if (symboltable == NULL)
      return NULL;

   SymbolEntry *help;
   int index=hashFunction(recordname);
   help = symboltable->table[index];
   
   while (help != NULL) { 

      if (help->type == recType) {
         if (strcmp(help->name, recordname) ==0) {
            RecordEntry *recordentry=(RecordEntry *)help->typref;
            while(recordentry != NULL) {
               if (strcmp(recordentry->name, name)==0) {
                  return recordentry;
               }
               recordentry=recordentry->next;
            }
            return NULL;
         }
	 else
            help = help->next;
      }
      else {
	help = help->next;
      }
   }
   return NULL;
}





/* 
   RecordEntry* getNRecordComp(const SymbolTable *symboltable, const char *recordname, int i)
   sucht nach der i-te Komponente des Records recordname in der Symboltabelle symboltable
   beim Treffer gibt den Zeiger auf den entsprechenden Eintrag zurueck
   sonst NULL
*/
RecordEntry* getNRecordComp(const SymbolTable* symboltable, const char *recordname, int i)
{
   if (symboltable==NULL)
      return NULL;

   SymbolEntry *help;
   int index=hashFunction(recordname);
   help = symboltable->table[index];

   while (help != NULL) { 

      if (help->type == recType) {
         if (strcmp(help->name, recordname) ==0) {
            RecordEntry *recordentry=(RecordEntry *)help->typref;
            while (help != NULL && i >1) {
               i--;
               recordentry=recordentry->next;
            }
            return recordentry;;
         }
	 else
            help = help->next;
      }
      else {
	help = help->next;
      }
   }
   return NULL;
}





/* 
   char *getRecordName(const SymbolTable *symboltable, const char *varname)
   gibt den Namen von Record zurueck auf den die Pointer-Variable varname zeigt
   NULL falls varname kein Pointer
*/
char *getRecordName(const SymbolTable *symboltable, const char *varname)
{
   SymbolEntry *se=getSymbol(symboltable,varname);
   if (se !=NULL && se->type == ptorecType) {
      return ((PointerEntry*)(se->typref))->recordname;
   }
   return NULL;
}





/* 
   void printSymbolTable(const SymbolTable* symboltable)
   gibt Symboltabelle symboltable aus 
*/
void printSymbolTable(const SymbolTable *symboltable) {

   SymbolEntry* help=NULL;
   int i=0;
   for (i=0; i < HashSize; i++) {
      help=symboltable->table[i];
      printf("h[%d]: ", i);
      while (help != NULL) {
         switch (help->type) {
            case intType: {
               printf("%s (int); ", help->name);
               break;
	    }
            case strType: {
 	       printf("%s (str);  ", help->name);
               break;
            }
	    case  ptorecType: {
               printf("%s (pointer to record %s); ", help->name, 
               (( PointerEntry*)(help->typref))->recordname);
               break;
            }
            case recType: {
               printf("%s (record) Komponente: ", help->name);
	       RecordEntry* recordentry=(RecordEntry*)help->typref;
	       while (recordentry != NULL) {
                  switch (recordentry->type) {
                     case intType: {
			printf("%s (integer),", recordentry->name );
                        break;
		     }
                     case strType: {
			printf("%s (string),", recordentry->name);
                        break;
		     }
		     case ptorecType: {
		        printf("%s (pointer to record %s),", recordentry->name,recordentry->recordname);
			break;
                     }
                  }
                  recordentry=recordentry->next;
               }
               printf("\n");
               break;
            }
	 }
         help = help->next;
      }
      printf("\n");
   }
}





/* 
   void printRecordEntry(const RecordEntry* recordentry)
   gibt Recordeintrag recordentry aus 
*/
void printRecordEntry(const RecordEntry* recordentry)
{

   if (recordentry == NULL) {
      printf("NULL");
      return;
   }
   printf("name: %s, ", recordentry->name);
   switch (recordentry->type) {
      case intType: 
         printf("type: INTEGER, ");
         break;
      case strType:
         printf("type: STRING, ");
         break;
      case ptorecType:
         printf("type: POINTER TO RECORD %s, ", recordentry->recordname);
         break;
   }
}


