#ifndef _symbol_h
#define _symbol_h



/* Groesse der Symboltabelle*/
#define HashSize 127
//127
//317


/* Typ des Eintrages */
enum GroundType {intType,strType,ptorecType,recType};


/* Eintrag in der Symboltabelle */
typedef struct SymbolEntry_ {
   char *name;
   enum GroundType type;
   //fuer die Kollisionsbehandlung
   SymbolEntry_ *next;
   //Hilfszeiger
   void* typref;
} SymbolEntry;


/* Symboltabelle */
typedef struct SymbolTable_ {
   SymbolEntry *table[HashSize];
} SymbolTable;


/* Integereintrag */
typedef struct IntEntry_ {
   int value;
} IntEntry;


/* Stringeintrag */
typedef struct StrEntry_ {
   char* value;
} StrEntry;

/* pointer to record Eintrag */
typedef struct PointerEntry_ {
   char* recordname;
} PointerEntry;

/* Recordeintrag */
typedef struct RecordEntry_ {
   char *name;
   //intType,strType oder ptorecType
   enum GroundType type;
   //fals type=ptoRecType
   char *recordname;
   struct RecordEntry_ *next;
} RecordEntry;


/* SymbolTable* createSymbolTable(void)
   erstellt eine Symboltabelle und gibt pointer
   auf diese Tabelle zurueck
*/
SymbolTable* createSymbolTable(void);

/* int putIntegerSymbol(SymbolTable *symboltable, char *name)
   traegt die Integer-Variable name in die Symboltabelle symboltable ein
   Rueckgabewerte:  
   -1 Symboltabelle symboltable ist null
    0 die Integer-Variable name in der Tabelle symboltable schon vorhanden
    1 die Integer-Variable name wurde in die Symboltabelle eingefuehrt
*/
int putIntegerSymbol(SymbolTable *symboltable, char *name);

/* int putStringSymbol(SymbolTable *symboltable, char *name)
   traegt die String-Variable name in die Symboltabelle symboltable ein
   Rueckgabewerte:
   -1 Symboltabelle symboltable ist null
    0 die String-Variable name in der Tabelle symboltable schon vorhanden
    1 die String-Variable name wurde in die Symboltabelle eingefuehrt
*/
int putStringSymbol(SymbolTable *symboltable, char *name);

/* int putPointerSymbol(SymbolTable *symboltable, char *name, char *recordname)
   traegt die Pointer-Variable name, die auf den record recordname zeigt in die 
   Symboltabelle symboltable ein
   Rueckgabewerte:
   -1 Symboltabelle symboltable ist null
   -2 Record recordname in der Symboltabelle symboltable nicht vorhanden
    0 die Pointer-Variable name in der Tabelle symboltable schon vorhanden
    1 die Pointer-Variable name wurde in die Symboltabelle symboltable eingefuehrt
*/
int putPointerSymbol(SymbolTable *symboltable, char *name, char *recordname);

/* int putRecordSymbol(SymbolTable *symboltable, char *name)
   traegt den Record recordname in die Symboltabelle symboltable ein
   Rueckgabewerte:
   -1 Symboltabelle symboltable ist null
    0 Record name in der Tabelle symboltable schon vorhanden
    1 Record name wurde in die Tabelle symboltable eingefuehrt
*/
int putRecordSymbol(SymbolTable *symboltable, char *name);

/* int addIntRecordComp(SymbolTable *symboltable, char *name, char *recordname)
   ergaenzt den Record recordname um die Integervariable name
   Rueckgabewerte:
   -1: Symboltabelle symboltable ist null
   -2: Record recordname in der Tabelle symboltable nicht vorhanden
    0: Variable name in dem Record recordname schon vorhanden
    1: Record recordname wurde um die Integervariable name ergaenzt
*/
int addIntRecordComp(SymbolTable *symboltable, char *name, char *recordname);

/* int addStrRecordComp(SymbolTable *symboltable, char *name, char *recordname) 
   ergaenzt den Record recordname um die Stringvariable name
   Rueckgabewerte:
   -1: Symboltabelle symboltable ist null
   -2: Record recordname in der Tabelle symboltable nicht vorhanden
    0: Variable name in dem Record recordname schon vorhanden
    1: Record recordname wurde um die Stringvariable name ergaenzt
*/
int addStrRecordComp(SymbolTable *symboltable, char *name, char *recordname);

/* int addPointerRecordComp(SymbolTable *symboltable, char *name, char *ptorecname,char *recordname) 
   ergaenzt den Record recordname um die Pointervariable name
   name ist pointer to ptorecname
   Rueckgabewerte:
   -1: Symboltabelle symboltable ist null
   -2: Record recordname in der Tabelle symboltable nicht vorhanden
   -3: Record ptorecname in der Tabelle symboltable nicht vorhanden
    0: Variable name in dem Record recordname schon vorhanden
    1: Record recordname wurde um die Pointervariable name ergaenzt
*/
int addPointerRecordComp(SymbolTable *symboltable, char *name, char *ptorecname, char *recordname);

/* SymbolEntry *getSymbol(const SymbolTable *symboltable, const char *name)
   sucht nach dem Eintrag name in der Symboltabelle symboltable
   beim Treffer gibt den Zeiger auf den entsprechenden Eintrag zurueck
   sonst NULL
*/
SymbolEntry *getSymbol(const SymbolTable *symboltable, const char *name);

/* RecordEntry *getRecordComp(const SymbolTable*, const char *name, const char *recordname)
   sucht nach der Recordcomponente name des Records recordname in der Symboltabelle symboltable
   beim Treffer gibt den Zeiger auf den entsprechenden Eintrag zurueck
   sonst NULL
*/
RecordEntry *getRecordComp(const SymbolTable*, const char *name, const char *recordname);


/* RecordEntry* getNRecordComp(const SymbolTable* symboltable, const char *recordname, int i)
   sucht nach der i-te Komponente des Records recordname in der Symboltabelle symboltable
   beim Treffer gibt den Zeiger auf den entsprechenden Eintrag zurueck
   sonst NULL
*/
RecordEntry* getNRecordComp(const SymbolTable* symboltable, const char *recordname, int i);


/* char *getRecordName(const SymbolTable *symboltable, const char *varname)
   gibt den Namen von Record zurueck auf den die Pointer-Variable varname zeigt
   NULL falls varname kein Pointer
*/
char *getRecordName(const SymbolTable *symboltable, const char *varname);


/* printSymbolTable(const SymbolTable *symboltable)
   gibt Symboltabelle symboltable aus 
*/
void printSymbolTable(const SymbolTable *symboltable);

/* void printRecordEntry(const RecordEntry *recordentry)
   gibt Recordeintrag recordentry aus 
*/
void printRecordEntry(const RecordEntry *recordentry);

#endif
