/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdlib.h>
#include <stdio.h> 
#include <string.h>
#include "pointer_lang_reader.h"

#include "pointer_lang/symboltable.h"
#include "pointer_lang/abstree.h"

#include "id_checker.h"

#include "GTXLReader.h"

extern NODE *abstree;


extern "C" {
   int yyparse();
   extern int sera;
}

extern FILE *yyin;


namespace algorithms
{


namespace readers
{


POINTER_LANG_READER::POINTER_LANG_READER()
{
   symboltable=NULL;
   workDir = "";
}


POINTER_LANG_READER::~POINTER_LANG_READER()
{
}


/*
   void POINTER_LANG_READER::initialize(char *infilename)
*/
//SPL Datei einlesen
//syntaktische Analyse
//Aufbau eines Parserbaumes
//semantische Analyse
void POINTER_LANG_READER::initialize(char *infilename)
{
   cout << "initialize" <<endl;
   cout <<"open " <<infilename<<endl;

   if((yyin=fopen(infilename,"r"))==NULL){

      cerr<<"File "<<infilename <<" not found:\n";
      cerr<<"Program will be aborted\n";
      exit(1);
   }

   int length=strlen(infilename);
   if (length>4) {
      if (infilename[length-1]!='l' ||  infilename[length-2]!='p' || infilename[length-3]!='s' || infilename[length-4]!='.') {
        cerr <<"Invalid file extension\n";
        cerr<<"Program will be aborted\n";
        exit(1);
      }
   }
   else {
        cerr <<"Invalid file extension\n";
        exit(1);
   }


   cout<<"syntactic analysis\n";
   yyparse();
   cout<<"ok\n";
   cout<<"semantic analysis\n";  
   symboltable = createSymbolTable();

   if (!buildSymbolTable(symboltable,abstree))
      cout<<"ok\n";
   else {
      cerr<<"Program will be aborted\n";
      freeAbstree(abstree);
      exit(1);
   }

   char * pch;
   char filename_copy[strlen(infilename)+1];
   strcpy(filename_copy, infilename);
   pch = strtok (filename_copy,"/");

   while (1)
   {
	workDir += string("/");
	string tmp = (string)pch;
	pch = strtok (NULL, "/");

	if(pch==NULL)
	 break;	
	else
	 workDir += tmp;
   }
}


/*
   void POINTER_LANG_READER::run(char *outfilename)
*/
void POINTER_LANG_READER::run(GTS &GraphGrammar)
{
   assert(symboltable!=NULL);
   int programcounter=1;
   int counter = 1;
   int rule_id =1;

   PLGTS * PLGraphGrammar=new PLGTS(symboltable, 1);

   char *varname;
   char *recordname;
   int depth=1;
   int type=0;


   assert(abstree->type==T_PROGRAM);
   //Deklarationen
   NODE *declarations=((PROGRAM*)(abstree->pointer))->declarations;
   if (declarations!=NULL)
      produceDeclRules(PLGraphGrammar, declarations, programcounter);
   //Anweisungen
   NODE *body=((PROGRAM*)(abstree->pointer))->body;
   if (body!=NULL)
      produceBodyRules(PLGraphGrammar, body, programcounter, false, -1);

   // write down the gtxl file and read it with the usual reader
   
   string tmpFile = workDir + string("tmp.xml");	
   PLGraphGrammar->WriteXML((char*)tmpFile.c_str());
   GTXLREADER gtxlreader;
   gtxlreader.initialize((char*)tmpFile.c_str());
   gtxlreader.run(GraphGrammar);
   

   // delete 
   freeAbstree(abstree);
   // delete GraphGrammar;

   // This part of code works instead of writing of the gtxl file	
   /*
   GraphGrammar = PLGraphGrammar->GraphGrammar;
   ID_CHECKER ic;
   ic.initialize(GraphGrammar);
   ic.run();
   */
}




/*
   void TestTimur::produceDeclRules(PLGTS *GraphGrammar, NODE *abstree, int &counter)
*/
void POINTER_LANG_READER::produceDeclRules(PLGTS *GraphGrammar, NODE *abstree, int &counter)
{
   if (abstree==NULL)
      return;
   switch (abstree->type) {
      case T_DECLLIST: {
         NODE *decl_head=((DECLLIST*)(abstree->pointer))->head;
         NODE *decl_tail=((DECLLIST*)(abstree->pointer))->tail;
         produceDeclRules(GraphGrammar, decl_head, counter);
         counter++;
         produceDeclRules(GraphGrammar, decl_tail, counter);
         break;
      }
      case T_VARDECL: {
         NODE *vardecl = ((VARDECL*)(abstree->pointer))->type;
         int type = vardecl->type;
         char *varname = ((VARDECL*)(abstree->pointer))->name;
         if (type==T_INTEGER) {
            GraphGrammar->produceIntDeclRule(counter, varname, catintstr(counter, ": produce Integer Decl Rule"));
         }
         else if (type==T_STRING) {
            GraphGrammar->produceStrDeclRule(counter, varname, catintstr(counter, ": produce String Decl Rule"));
         }
         else if (type==T_PTOREC) {
            GraphGrammar->producePtoRecDeclRule(counter, varname,catintstr(counter, ": produce Pointer Decl Rule"));
         }
         else 
            assert(false);
         break;
      }
      case T_TYPEDECL: 
      case T_TYPEDECLNULL: {
         //GraphGrammar->produceTypeDeclRule(counter, catintstr(counter, ": produce Type Decl Rule"));
         counter--;
         break;
      }
      default: assert(false);

   }

}


/*
void TestTimur::produceBodyRules(PLGTS *GraphGrammar, NODE *abstree, int counter, bool laststm, int gotocounter) 
*/
void POINTER_LANG_READER::produceBodyRules(PLGTS *GraphGrammar, NODE *abstree, int counter, bool laststm, int gotocounter)
{

   switch (abstree->type) {

      //IFSTM:
      case T_IFSTM: {
         NODE *bexp=((IFSTM*)(abstree->pointer))->bexp;
         NODE *thenstm=((IFSTM*)(abstree->pointer))->ifstm;
         NODE *elsestm=((IFSTM*)(abstree->pointer))->elsestm;
         int elsecounter=computeSTMLIST(thenstm)+counter+1;
         int fi=computeSTMLIST(abstree)+counter+1;


         int pointer=0;
         int null=0;
         int var=0;
         controlBEXP(symboltable, bexp, pointer, null, var);
         if (var>0) {
            booltree::Node *booltree=BoolExpToBooltree(bexp);
            GraphGrammar->produceIFRule(counter, elsecounter, booltree, catintstr(counter, ": IF "));
	    delete booltree;
         }
         else if (bexp->type==T_EQUAL) {
            NODE *left=((EQUAL*)(bexp->pointer))->left;
            NODE *right=((EQUAL*)(bexp->pointer))->right;
            if (pointer==2) {
               list <char*> leftlist=PathToCharList(left);
               list <char*> rightlist=PathToCharList(right);
               GraphGrammar->produceIFRule(counter, elsecounter, leftlist, rightlist, catintstr(counter, ": IF p1=p2 "), true);
            }
            else if (pointer==1 && null==1) {
               if (left->type==T_NULL) {
                  list <char*> rightlist=PathToCharList(right);
                  GraphGrammar->produceIFNULLRule(counter, elsecounter, rightlist, catintstr(counter, ": IF NULL=p "), true);
               }
               else if (right->type==T_NULL) {
                  list <char*> leftlist=PathToCharList(left);
                  GraphGrammar->produceIFNULLRule(counter, elsecounter, leftlist, catintstr(counter, ": IF p=NULL "), true);
               }
               else {
                  assert(false);
               }
            }
            else 
               assert(false);
         }
         else if (bexp->type==T_NOTEQUAL) {
            NODE *left=((NOTEQUAL*)(bexp->pointer))->left;
            NODE *right=((NOTEQUAL*)(bexp->pointer))->right;
            if (pointer==2) {
               list <char*> leftlist=PathToCharList(left);
               list <char*> rightlist=PathToCharList(right);
               GraphGrammar->produceIFRule(counter, elsecounter, leftlist, rightlist, catintstr(counter, ": IF p1!=p2 "), false);
            }
            else if (pointer==1 && null==1) {
               if (left->type==T_NULL) {
                  list <char*> rightlist=PathToCharList(right);
                  GraphGrammar->produceIFNULLRule(counter, elsecounter, rightlist, catintstr(counter, ": IF NULL!=p "), false);
               }
               else if (right->type==T_NULL) {
                  list <char*> leftlist=PathToCharList(left);
                  GraphGrammar->produceIFNULLRule(counter, elsecounter, leftlist, catintstr(counter, ": IF p!=NULL "), false);
               }
               else {
                  assert(false);
               }
            }
            else 
               assert(false);
         }
         else if(bexp->type==T_NOT) {
            NODE *exp=((NOT_*)(bexp->pointer))->left;
            if (exp->type==T_EQUAL) {
               NODE *left=((EQUAL*)(exp->pointer))->left;
               NODE *right=((EQUAL*)(exp->pointer))->right;
               if (pointer==2) {
                  list <char*> leftlist=PathToCharList(left);
                  list <char*> rightlist=PathToCharList(right);
                  GraphGrammar->produceIFRule(counter, elsecounter, leftlist, rightlist, catintstr(counter, ": IF NOT(p1=p2) "), false);
               }
               else if (pointer==1 && null==1) {
                  if (left->type==T_NULL) {
                     list <char*> rightlist=PathToCharList(right);
                     GraphGrammar->produceIFNULLRule(counter, elsecounter, rightlist, catintstr(counter, ": IF NOT(NULL=p) "), false);
                  }
                  else if (right->type==T_NULL) {
                     list <char*> leftlist=PathToCharList(left);
                     GraphGrammar->produceIFNULLRule(counter, elsecounter, leftlist, catintstr(counter, ": IF NOT(p=NULL) "), false);
                  }
                  else {
                     assert(false);
                  }
               }
               else 
                  assert(false);
            }
            else if (exp->type==T_NOTEQUAL) {
               NODE *left=((NOTEQUAL*)(exp->pointer))->left;
               NODE *right=((NOTEQUAL*)(exp->pointer))->right;
               if (pointer==2) {
                  list <char*> leftlist=PathToCharList(left);
                  list <char*> rightlist=PathToCharList(right);
                  GraphGrammar->produceIFRule(counter, elsecounter, leftlist, rightlist, catintstr(counter, ": IF NOT(p1!=p2) "), true);
               }
               else if (pointer==1 && null==1) {
                  if (left->type==T_NULL) {
                     list <char*> rightlist=PathToCharList(right);
                     GraphGrammar->produceIFNULLRule(counter, elsecounter, rightlist, catintstr(counter, ": IF NOT(NULL!=p)"), true);
                  }
                  else if (right->type==T_NULL) {
                     list <char*> leftlist=PathToCharList(left);
                     GraphGrammar->produceIFNULLRule(counter, elsecounter, leftlist, catintstr(counter, ": IF NOT(p!=NULL)"), true);
                  }
                  else {
                     assert(false);
                  }
               }
               else 
                  assert(false);
            }
            else
               assert(false);
         }
         else {
            //
            //GraphGrammar->produceIFRule(counter, elsecounter, booltree, catintstr(counter, ": IF "));
         }


         if (laststm) {
            if (bexp->type==T_TRUE) {
	       GraphGrammar->produceSKIPRule(counter,counter+1, catintstr(counter,": IF (TRUE)"));
               produceBodyRules(GraphGrammar, thenstm, counter+1, false, gotocounter);
            }
            else if (bexp->type==T_FALSE) {
	       GraphGrammar->produceSKIPRule(counter,elsecounter, catintstr(counter,": IF (FALSE)"));
               produceBodyRules(GraphGrammar, elsestm, elsecounter, false, gotocounter);
            }
            else {
               produceBodyRules(GraphGrammar, thenstm, counter+1, false, gotocounter);
               produceBodyRules(GraphGrammar, elsestm, elsecounter, false, gotocounter);
            }
         }
         else {
            if (bexp->type==T_TRUE) {
	       GraphGrammar->produceSKIPRule(counter,counter+1, catintstr(counter,": IF (TRUE)"));
               produceBodyRules(GraphGrammar, thenstm, counter+1, false, fi);
            }
            else if (bexp->type==T_FALSE) {
	       GraphGrammar->produceSKIPRule(counter,elsecounter, catintstr(counter,": IF (FALSE)"));
               produceBodyRules(GraphGrammar, elsestm, elsecounter, false, fi);
	    }
            else {
               produceBodyRules(GraphGrammar, thenstm, counter+1, false, fi);
               produceBodyRules(GraphGrammar, elsestm, elsecounter, false, fi);
            }
         }
         break;
      }



      //WHILE
      case T_WHILE: {
         NODE *bexp=((WHILESTM*)(abstree->pointer))->bexp;
         NODE *dostm=((WHILESTM*)(abstree->pointer))->dostm;
         int odcounter=computeSTMLIST(dostm)+counter+1;

         int pointer=0;
         int null=0;
         int var=0;

         controlBEXP(symboltable, bexp, pointer, null, var);
         if (var>0) {
            booltree::Node *booltree=BoolExpToBooltree(bexp);
            if (laststm)
               GraphGrammar->produceWHILERule(counter, gotocounter, booltree, catintstr(counter, ": WHILE "));
            else
               GraphGrammar->produceWHILERule(counter, odcounter, booltree, catintstr(counter, ": WHILE "));
            delete booltree;
         }
         else if (bexp->type==T_EQUAL) {
            NODE *left=((EQUAL*)(bexp->pointer))->left;
            NODE *right=((EQUAL*)(bexp->pointer))->right;
            if (pointer==2) {
               list <char*> leftlist=PathToCharList(left);
               list <char*> rightlist=PathToCharList(right);
               if (laststm)
                  GraphGrammar->produceWHILERule(counter, gotocounter, leftlist, rightlist, catintstr(counter, ": WHILE p1=p2 "), true);
               else
                  GraphGrammar->produceWHILERule(counter, odcounter, leftlist, rightlist, catintstr(counter, ": WHILE p1=p2 "), true);
            }
            else if (pointer==1 && null==1) {
               if (left->type==T_NULL) {
                  list <char*> rightlist=PathToCharList(right);
                  if (laststm)
                     GraphGrammar->produceWHILENULLRule(counter, gotocounter, rightlist, catintstr(counter, ": WHILE NULL=p "), true);
                  else
                     GraphGrammar->produceWHILENULLRule(counter, odcounter, rightlist, catintstr(counter, ": WHILE NULL=p "), true);
               }
               else if (right->type==T_NULL) {
                  list <char*> leftlist=PathToCharList(left);
                  if (laststm)
                     GraphGrammar->produceWHILENULLRule(counter, gotocounter, leftlist, catintstr(counter, ": WHILE p=NULL "), true);
                  else
                     GraphGrammar->produceWHILENULLRule(counter, odcounter, leftlist, catintstr(counter, ": WHILE p=NULL "), true);
               }
               else {
                  assert(false);
               }
            }
            else 
               assert(false);
         }
         else if (bexp->type==T_NOTEQUAL) {
            NODE *left=((NOTEQUAL*)(bexp->pointer))->left;
            NODE *right=((NOTEQUAL*)(bexp->pointer))->right;
            if (pointer==2) {
               list <char*> leftlist=PathToCharList(left);
               list <char*> rightlist=PathToCharList(right);
               if (laststm)
                  GraphGrammar->produceWHILERule(counter, gotocounter, leftlist, rightlist, catintstr(counter, ": WHILE p1!=p2 "), false);
               else
                  GraphGrammar->produceWHILERule(counter, odcounter, leftlist, rightlist, catintstr(counter, ": WHILE p1!=p2 "), false);
            }
            else if (pointer==1 && null==1) {
               if (left->type==T_NULL) {
                  list <char*> rightlist=PathToCharList(right);
                  if (laststm)
                     GraphGrammar->produceWHILENULLRule(counter, gotocounter, rightlist, catintstr(counter, ": WHILE NULL!=p "), false);
                  else
                     GraphGrammar->produceWHILENULLRule(counter, odcounter, rightlist, catintstr(counter, ": WHILE NULL!=p "), false);
               }
               else if (right->type==T_NULL) {
                  list <char*> leftlist=PathToCharList(left);
                  if (laststm)
                     GraphGrammar->produceWHILENULLRule(counter, gotocounter, leftlist, catintstr(counter, ": WHILE p!=NULL "), false);
                  else
                     GraphGrammar->produceWHILENULLRule(counter, odcounter, leftlist, catintstr(counter, ": WHILE p!=NULL "), false);
               }
               else {
                  assert(false);
               }
            }
            else 
               assert(false);
         }

         else if(bexp->type==T_NOT) {
            NODE *exp=((NOT_*)(bexp->pointer))->left;
            if (exp->type==T_EQUAL) {
               NODE *left=((EQUAL*)(exp->pointer))->left;
               NODE *right=((EQUAL*)(exp->pointer))->right;
               if (pointer==2) {
                  list <char*> leftlist=PathToCharList(left);
                  list <char*> rightlist=PathToCharList(right);
                  if (laststm)
                     GraphGrammar->produceWHILERule(counter, gotocounter, leftlist, rightlist, catintstr(counter, ": WHILE NOT(p1=p2) "), false);
                  else
                     GraphGrammar->produceWHILERule(counter, odcounter, leftlist, rightlist, catintstr(counter, ": WHILE NOT(p1=p2) "), false);
               }
               else if (pointer==1 && null==1) {
                  if (left->type==T_NULL) {
                     list <char*> rightlist=PathToCharList(right);
                     if (laststm)
                        GraphGrammar->produceWHILENULLRule(counter, gotocounter, rightlist, catintstr(counter, ": WHILE NOT(NULL=p) "), false);
                     else
                        GraphGrammar->produceWHILENULLRule(counter, odcounter, rightlist, catintstr(counter, ": WHILE NOT(NULL=p) "), false);
                  }
                  else if (right->type==T_NULL) {
                     list <char*> leftlist=PathToCharList(left);
                     if (laststm)
                        GraphGrammar->produceWHILENULLRule(counter, gotocounter, leftlist, catintstr(counter, ": WHILE NOT(p=NULL) "), false);
                     else
                        GraphGrammar->produceWHILENULLRule(counter, odcounter, leftlist, catintstr(counter, ": WHILE NOT(p=NULL) "), false);
                  }
                  else {
                     assert(false);
                  }
               }
               else 
                  assert(false);
            }
            else if (exp->type==T_NOTEQUAL) {
               NODE *left=((NOTEQUAL*)(exp->pointer))->left;
               NODE *right=((NOTEQUAL*)(exp->pointer))->right;
               if (pointer==2) {
                  list <char*> leftlist=PathToCharList(left);
                  list <char*> rightlist=PathToCharList(right);
                  if (laststm)
                     GraphGrammar->produceIFRule(counter, gotocounter, leftlist, rightlist, catintstr(counter, ": WHILE NOT(p1!=p2) "), true);
                  else
                     GraphGrammar->produceIFRule(counter, odcounter, leftlist, rightlist, catintstr(counter, ": WHILE NOT(p1!=p2) "), true);
               }
               else if (pointer==1 && null==1) {
                  if (left->type==T_NULL) {
                     list <char*> rightlist=PathToCharList(right);
                     if (laststm)
                        GraphGrammar->produceWHILENULLRule(counter, gotocounter, rightlist, catintstr(counter, ": WHILE NOT(NULL!=p)"), true);
                     else
                        GraphGrammar->produceWHILENULLRule(counter, odcounter, rightlist, catintstr(counter, ": WHILE NOT(NULL!=p)"), true);
                  }
                  else if (right->type==T_NULL) {
                     list <char*> leftlist=PathToCharList(left);
                     if (laststm)
                        GraphGrammar->produceWHILENULLRule(counter, gotocounter, leftlist, catintstr(counter, ": WHILE NOT(p!=NULL)"), true);
                     else
                        GraphGrammar->produceWHILENULLRule(counter, odcounter, leftlist, catintstr(counter, ": WHILE NOT(p!=NULL)"), true);
                  }
                  else {
                     assert(false);
                  }
               }
               else 
                  assert(false);
            }
            else
               assert(false);
         }
         else {
            //
            //GraphGrammar->produceIFRule(counter, elsecounter, booltree, catintstr(counter, ": IF "));
         }

         if (bexp->type==T_TRUE) {
            GraphGrammar->produceSKIPRule(counter, counter+1, catintstr(counter, ": WHILE (TRUE)"));
            produceBodyRules(GraphGrammar, dostm, counter+1, laststm, counter);
         }
         else if (bexp->type==T_FALSE) {
            if (laststm)
               GraphGrammar->produceSKIPRule(counter, gotocounter, catintstr(counter, ": WHILE (FALSE)"));
            else
               GraphGrammar->produceSKIPRule(counter, odcounter, catintstr(counter, ": WHILE (FALSE)"));
         }
         else
            produceBodyRules(GraphGrammar, dostm, counter+1, laststm, counter);
         break;
      }

      //SKIP:
      case T_SKIP: {
         int leftcounter=counter;
         int rightcounter=0;

         if (laststm) {
            rightcounter=gotocounter;
         }
         else {
            rightcounter=counter+1;
         }
         GraphGrammar->produceSKIPRule(leftcounter, rightcounter, catintstr(counter,": produce Skip Rule"));
         break;
      }

      //NEW
      case T_NEW: {
         NODE *newpath=(( ALLOC*)(abstree->pointer))->path;

         int counterleft=counter;
         int counterright;

         if (laststm) {
            counterright=gotocounter;
         }
         else {
            counterright=counter+1;
         }

         list <char*> path = PathToCharList(newpath);
         GraphGrammar->produceNEWRule(counterleft, counterright, path, catintstr(counter,": produce NEW Pointer Rule"));
         break;
      }

      //DELETE
      case T_DELETE: {
         NODE *deletepath=((DELETE_*)(abstree->pointer))->path;

         int counterleft=counter;
         int counterright;

         if (laststm) {
            counterright=gotocounter;
         }
         else {
            counterright=counter+1;
         }

         list <char*> path = PathToCharList(deletepath);
         GraphGrammar->produceDELETERule(counterleft, counterright, path, catintstr(counter,": produce DELETE Pointer Rule"));
         break;
      }


      //ASSIGNER
      case T_ASSIGNER: {
         NODE* stmleft=(( ASSIGNER*)(abstree->pointer))->left;
         NODE* stmright=(( ASSIGNER*)(abstree->pointer))->right;

         int counterleft=counter;
         int counterright;

         if (laststm) {
            counterright=gotocounter;
         }
         else {
            counterright=counter+1;
         }

         //integer =
         if (strcmp(getNodeType(symboltable, stmleft), "INT")==0) {

             arithmetictree::Node *tree=ArithmeticExpToArithmetictree(stmright);
             GraphGrammar->produceIntegerAssignerRule(counterleft, counterright, PathToCharList(stmleft),tree,catintstr(counter, ": produce Integer Assigner Rule"));
             delete tree;

         }

         //string =
         else if (strcmp(getNodeType(symboltable, stmleft),"STR")==0) {
            char *rulename=catintstr(counter, ": produce String Assigner Rule");
             if (stmright->type == T_STRCONST)
                GraphGrammar->produceStringAssignerConstRule(counterleft, counterright, PathToCharList(stmleft),(( STRCONST*)(stmright->pointer))->value,rulename);
             else {
                assert(strcmp(getNodeType(symboltable, stmright),"STR")==0);
                GraphGrammar->produceStringAssignerRule(counterleft, counterright, PathToCharList(stmleft),PathToCharList(stmright), rulename);

             }
         }
         //pointer =
         else if(strstr(getNodeType(symboltable, stmleft),"PTOREC ") != NULL) {
             list <char*> lefthand = PathToCharList(stmleft);
             if (stmright->type == T_NULL) {;
                GraphGrammar->producePointerAssignerNULLRule(counterleft, counterright, lefthand, catintstr(counterleft, ": produce Pointer Null Assigner Rule"));
             }
             else {
                assert(strstr(getNodeType(symboltable, stmright),"PTOREC ") != NULL);
                list <char*> righthand = PathToCharList(stmright);
                GraphGrammar->producePointerAssignerRule(counterleft, counterright, lefthand, righthand, catintstr(counterleft,": produce Pointer Assigner Rule"));
             }
          }
          else {assert(false);}

         break;
      }


      //STMLIST
      case T_STMLIST: {
         NODE *stmlist_head=((STMLIST*)(abstree->pointer))->head;
         NODE *stmlist_tail=((STMLIST*)(abstree->pointer))->tail;
         if (stmlist_tail==NULL) {
            produceBodyRules(GraphGrammar, stmlist_head, counter, true, gotocounter);
         }
         else {
            produceBodyRules(GraphGrammar, stmlist_head, counter, false, gotocounter);
            produceBodyRules(GraphGrammar, stmlist_tail, computeSTMLIST(stmlist_head)+counter+1, false, gotocounter);
         }
         break;
      }

      default: assert(false);

   }
}




/*
   int POINTER_LANG_READER::computeSTMLIST(NODE *abstree)
   berechnet die Anzahl der Knoten vom Typ STMLIST
*/
int POINTER_LANG_READER::computeSTMLIST(NODE *abstree) {

   if (abstree==NULL)
      return 0;

   switch (abstree->type) {
      case T_STMLIST: {
         NODE *head=((STMLIST*)(abstree->pointer))->head;
         NODE *tail=((STMLIST*)(abstree->pointer))->tail;
         return 1+computeSTMLIST(head)+computeSTMLIST(tail);
      }
      case T_IFSTM: {
         NODE *ifstm=((IFSTM*)(abstree->pointer))->ifstm;
         NODE *elsestm=((IFSTM*)(abstree->pointer))->elsestm;
         return computeSTMLIST(ifstm)+computeSTMLIST(elsestm);
      }
      case T_WHILE: {
         NODE *dostm=((WHILESTM*)(abstree->pointer))->dostm;
         return computeSTMLIST(dostm);
      }
      case T_ASSIGNER: 
      case T_SKIP:
      case T_NEW: 
      case T_DELETE: return 0;
      
      default: assert(false);
   }
}



/* 
   list<char*> PathToCharList(NODE* tree)
   PATH oder IDENTIFIER -> list<char*>
*/
list <char*> POINTER_LANG_READER::PathToCharList(NODE* tree)
{
   assert(tree->type == T_PATH || tree->type == T_IDENTIFIER);
   list <char*> list;

   if (tree->type == T_PATH) {
      int l=1,i=0;
      char *recordname,*name; 
      while (i=getNPath(symboltable, tree, l++, &name, &recordname) == 1) 
         list.push_back(name);
   }
   else {
      SymbolEntry *se=getSymbol(symboltable,(( IDENTIFIER*)(tree->pointer))->name);
      //assert (se->type == ptorecType);
      list.push_back(se->name);
   }
   return list;
}



/*
   booltree::Node *POINTER_LANG_READER::BoolExpToBooltree(NODE* tree)
   konstruiert einen booleschen Baum
*/
booltree::Node *POINTER_LANG_READER::BoolExpToBooltree(NODE* tree)
{

   using namespace booltree;
   switch (tree->type) {

      case T_EQUAL: {
         NODE *left=((EQUAL*)(tree->pointer))->left;
         NODE *right=((EQUAL*)(tree->pointer))->right;
         Node *pNode=new EqualNode(BoolExpToBooltree(left),BoolExpToBooltree(right));
         return pNode;
      }

      case T_NOTEQUAL: {
         NODE *left=(( NOTEQUAL*)(tree->pointer))->left;
         NODE *right=(( NOTEQUAL*)(tree->pointer))->right;
         Node *pNode=new NotEqualNode(BoolExpToBooltree(left),BoolExpToBooltree(right));
         return pNode;
      }
      case T_SMALLER: {
         NODE *left=(( SMALLER*)(tree->pointer))->left;
         NODE *right=(( SMALLER*)(tree->pointer))->right;
         Node *pNode=new SmallerNode(BoolExpToBooltree(left),BoolExpToBooltree(right));
         return pNode;
      }
      case T_GREATER: {
         NODE *left=(( GREATER*)(tree->pointer))->left;
         NODE *right=(( GREATER*)(tree->pointer))->right;
         Node *pNode=new GreaterNode(BoolExpToBooltree(left),BoolExpToBooltree(right));
         return pNode;
      }
      case T_SMALLEREQUAL: {
         NODE *left=(( SMALLEREQUAL*)(tree->pointer))->left;
         NODE *right=(( SMALLEREQUAL*)(tree->pointer))->right;
         Node *pNode=new SmallerEqualNode(BoolExpToBooltree(left),BoolExpToBooltree(right));
         return pNode;
      }
      case T_GREATEREQUAL: {
         NODE *left=(( GREATEREQUAL*)(tree->pointer))->left;
         NODE *right=(( GREATEREQUAL*)(tree->pointer))->right;
         Node *pNode=new GreaterEqualNode(BoolExpToBooltree(left),BoolExpToBooltree(right));
         return pNode;
      }
      case T_AND: {
         NODE *left=(( AND_*)(tree->pointer))->left;
         NODE *right=(( AND_*)(tree->pointer))->right;
         Node *pNode=new AndNode(BoolExpToBooltree(left),BoolExpToBooltree(right));
         return pNode;
      }
      case T_OR: {
         NODE *left=(( OR_*)(tree->pointer))->left;
         NODE *right=(( OR_*)(tree->pointer))->right;
         Node *pNode=new OrNode(BoolExpToBooltree(left),BoolExpToBooltree(right));
         return pNode;
      }
      case T_NOT: {
         NODE *node=(( NOT_*)(tree->pointer))->left;
         Node *pNode=new NotNode(BoolExpToBooltree(node));
         return pNode;
      }
      case T_PLUSOP: {
         NODE *left=(( PLUSOP*)(tree->pointer))->left;
         NODE *right=(( PLUSOP*)(tree->pointer))->right;
         Node *pNode=new AddNode(BoolExpToBooltree(left),BoolExpToBooltree(right));
         return pNode;
      }
      case T_MINUSOP: {
         NODE *left=(( MINUSOP*)(tree->pointer))->left;
         NODE *right=(( MINUSOP*)(tree->pointer))->right;
         Node *pNode=new MinNode(BoolExpToBooltree(left),BoolExpToBooltree(right));
         return pNode;
      }
      case T_MULTOP: {
         NODE *left=(( MULTOP*)(tree->pointer))->left;
         NODE *right=(( MULTOP*)(tree->pointer))->right;
         Node *pNode=new MultNode(BoolExpToBooltree(left),BoolExpToBooltree(right));
         return pNode;
      }
      case T_DIVOP: {
         NODE *left=(( DIVOP*)(tree->pointer))->left;
         NODE *right=(( DIVOP*)(tree->pointer))->right;
         Node *pNode=new DivNode(BoolExpToBooltree(left),BoolExpToBooltree(right));
         return pNode;
      }
      case T_MODOP: {
         NODE *left=(( MODOP*)(tree->pointer))->left;
         NODE *right=(( MODOP*)(tree->pointer))->right;
         Node *pNode=new ModNode(BoolExpToBooltree(left),BoolExpToBooltree(right));
         return pNode;
      }
      case T_INTEGERCONST: {
         Node *pNode=new NumConstNode((( INTEGERCONST*)(tree->pointer))->value);
         return pNode;
      }
      case T_STRCONST: {
         Node *pNode=new StrConstNode((( STRCONST*)(tree->pointer))->value);
         return pNode;
      }
      case T_IDENTIFIER: {
         list <char*> list;
         list.push_back((( IDENTIFIER*)(tree->pointer))->name);
         Node *pNode=new IdentifierNode(list);
         return pNode;
      }
      case T_PATH:{
         list <char*> list=PathToCharList(tree);
         Node *pNode=new IdentifierNode(list);
         return pNode;
      }
      case T_NEGATE: {
         Node *pNodeNum=new NumConstNode(-1);
          NODE *expstm=(( NEGATE*)(tree->pointer))->expstm;
         Node *pNode=new MultNode(pNodeNum,BoolExpToBooltree(expstm));
         return pNode;
      }
      case T_TRUE: {
         list <char*> list;
         list.push_back("TRUE");
         Node *pNode=new IdentifierNode(list);
         return pNode;
      }
      case T_FALSE: {
         list <char*> list;
         list.push_back("FALSE");
         Node *pNode=new IdentifierNode(list);
         return pNode;
      }

   }
}



/*
   arithmetictree::Node *POINTER_LANG_READER::ArithmeticExpToArithmetictree(NODE* tree)
  konstruiert einen arithmetischen Baum
*/
arithmetictree::Node *POINTER_LANG_READER::ArithmeticExpToArithmetictree(NODE* tree)
{

   using namespace arithmetictree;
   switch  (tree->type) {

      case T_PLUSOP: {
         NODE *left=(( PLUSOP*)(tree->pointer))->left;
         NODE *right=(( PLUSOP*)(tree->pointer))->right;
         Node *pNode=new AddNode(ArithmeticExpToArithmetictree(left),ArithmeticExpToArithmetictree(right));
         return pNode;
      }
      case T_MINUSOP: {
         NODE *left=(( MINUSOP*)(tree->pointer))->left;
         NODE *right=(( MINUSOP*)(tree->pointer))->right;
         Node *pNode=new MinNode(ArithmeticExpToArithmetictree(left),ArithmeticExpToArithmetictree(right));
         return pNode;
      }
      case T_MULTOP: {
         NODE *left=(( MULTOP*)(tree->pointer))->left;
         NODE *right=(( MULTOP*)(tree->pointer))->right;
         Node *pNode=new MultNode(ArithmeticExpToArithmetictree(left),ArithmeticExpToArithmetictree(right));
         return pNode;
      }
      case T_DIVOP: {
         NODE *left=(( DIVOP*)(tree->pointer))->left;
         NODE *right=(( DIVOP*)(tree->pointer))->right;
         Node *pNode=new DivNode(ArithmeticExpToArithmetictree(left),ArithmeticExpToArithmetictree(right));
         return pNode;
      }
      case T_MODOP: {
         NODE *left=(( MODOP*)(tree->pointer))->left;
         NODE *right=(( MODOP*)(tree->pointer))->right;
         Node *pNode=new ModNode(ArithmeticExpToArithmetictree(left),ArithmeticExpToArithmetictree(right));
         return pNode;
      }
      case T_INTEGERCONST: {
         Node *pNode=new NumConstNode((( INTEGERCONST*)(tree->pointer))->value);
         return pNode;
      }
      case T_IDENTIFIER: {
         list <char*> list;
         list.push_back((( IDENTIFIER*)(tree->pointer))->name);
         Node *pNode=new NumNode(list);
         return pNode;
      }
      case T_PATH:{
         list <char*> list=PathToCharList(tree);
         Node *pNode=new NumNode(list);
         return pNode;
      }
      case T_NEGATE: {
         Node *pNodeNum=new NumConstNode(-1);
         NODE *expstm=(( NEGATE*)(tree->pointer))->expstm;
         Node *pNode=new MultNode(pNodeNum,ArithmeticExpToArithmetictree(expstm));
         return pNode;
      }
      default: assert(false);
   }
}

};
};
