/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "ptnet.h"

namespace data_structures
{
   TCONNECTION::TCONNECTION()
   {
      ID = 0;
      weight = 0;
   }

   TCONNECTION::~TCONNECTION()
   {
   }
   
   TCONNECTION TCONNECTION::clone()
   {
      TCONNECTION arc;
      arc.ID = this->ID;
      arc.weight = this->weight;
      arc.attributes = this->attributes;
      return arc;
   }    
   
   PTNET::PTNET()
   {
      initial_marking = MARKING();
      current_marking = MARKING();
      maxDepth = 0;
   }

   PTNET::~PTNET()
   {
   }
   
   PTNET PTNET::clone()
   {
      PTNET ptn = PTNET();
      ptn.transitions = this->transitions;
      ptn.presets = this->presets;
      ptn.postsets = this->postsets;
      ptn.initial_marking = this->initial_marking;
      ptn.current_marking = this->current_marking;
      ptn.maxDepth = this->maxDepth;
      ptn.depth = this->depth;
      ptn.vertices = this->vertices;
      ptn.edges = this->edges;
      ptn.lastUsedID = this->lastUsedID;
      return ptn;
   }

   void PTNET::addPlace(IDTYPE p, VERTEX v)
   {
      assert(this->edges.count(p) == 0);
      if(this->vertices.count(p) == 0)
      {
         this->vertices[p] = v;
         this->vertices[p].setID(p);
	 
	 this->presets[p] = vector<TCONNECTION>();
         this->postsets[p] = vector<TCONNECTION>();
	           
	 this->depth[p] = maxDepth;
      }
      if(p > lastUsedID) lastUsedID = p;
   }
   
   void PTNET::addPlace(IDTYPE p, EDGE e)
   {
      assert(this->vertices.count(p) == 0);
      if(this->edges.count(p) == 0)
      {
	 e.setDepth(maxDepth);
         this->edges[p] = e;
         this->edges[p].setID(p);
         this->presets[p] = vector<TCONNECTION>();
         this->postsets[p] = vector<TCONNECTION>();
  	 this->depth[p] = maxDepth;
      }
      if(p > lastUsedID) lastUsedID = p;
   }
   
   void PTNET::addPlace(IDTYPE p)
   {
      this->addPlace(p, EDGE());
   }
   
   void PTNET::deletePlace(IDTYPE p)
   {
      vector<TCONNECTION>::iterator arcit;
      
      /*
       * delete all transitions in p*
       */
      for( arcit = this->postsets[p].begin();
           arcit != this->postsets[p].end();
           arcit++ )
      {
         this->deleteTransition(arcit->ID);
      }
        
      /*
       * remove p from the postset of all transitions in *p
       */
      for( arcit = this->presets[p].begin();
           arcit != this->presets[p].end();
           arcit++ )
      {
         this->remove_y_fromPostsetOf_x(arcit->ID, p);
      }
       
      /*
       * delete p
       */
      this->vertices.erase(p);
      this->edges.erase(p);
      this->presets.erase(p);
      this->postsets.erase(p);
      this->initial_marking.deletePlace(p);
      this->current_marking.deletePlace(p);
      this->depth.erase(p);
   }
   
   set<IDTYPE> PTNET::getPlaces()
   {
      set<IDTYPE> places = set<IDTYPE>();
      
      if(nodes_as_places)
      {
	map<IDTYPE, VERTEX>::iterator vit;
	for(vit = this->vertices.begin(); vit != this->vertices.end(); vit++)
	{
		places.insert(vit->first);
	}
      }
      
      map<IDTYPE, EDGE>::iterator eit;
      for(eit = this->edges.begin(); eit != this->edges.end(); eit++)
      {
         places.insert(eit->first);
      }
      
      return places;
   }
   
   bool PTNET::hasPlace(IDTYPE p)
   {
      if(nodes_as_places)
       return ((this->vertices.count(p) > 0) || (this->edges.count(p) > 0));
      else
       return (this->edges.count(p) > 0); 
   }

   void PTNET::addTransition(IDTYPE t, TRANSITION tguard)
   {
      if(this->transitions.count(t) == 0)
      {
         this->transitions[t] = tguard;
	 this->transitions[t].setID(t);
         this->presets[t] = vector<TCONNECTION>();
         this->postsets[t] = vector<TCONNECTION>();
	 this->depth[t] = maxDepth;
	 if(t > lastUsedID) lastUsedID = t;	 
      }
   }
   
   void PTNET::deleteTransition(IDTYPE t)
   {
      vector<TCONNECTION>::iterator arcit;
      
      /*
       * delete t from the postset of all places in *t
       */
      for( arcit = this->presets[t].begin();
           arcit != this->presets[t].end();
           arcit++ )
      {
         this->remove_y_fromPostsetOf_x(arcit->ID, t);
      }
      
      /*
       * delete t from the preset of all places in t*
       */
      for( arcit = this->postsets[t].begin();
           arcit != this->postsets[t].end();
           arcit++ )
      {
         this->remove_y_fromPresetOf_x(arcit->ID, t);
      }
   
      /*
       * delete t
       */
      this->transitions.erase(t);
      this->presets.erase(t);
      this->postsets.erase(t);
      this->depth.erase(t);
   }
   
   bool PTNET::hasTransition(IDTYPE t)
   {
      return (this->transitions.count(t) > 0);
   }
   
   map<IDTYPE, TRANSITION> PTNET::getTransitions()
   {
      return this->transitions;
   }

   void PTNET::addTConnection(IDTYPE source, TCONNECTION arc)
   {
      assert(   (this->hasPlace(source) && this->hasTransition(arc.ID))
             || (this->hasPlace(arc.ID) && this->hasTransition(source)) );
   
      /*
       * add given arc
       */
      this->postsets[source].push_back(arc);

      /*
       * add dual arc
       */
      TCONNECTION dual = TCONNECTION();
      dual.ID = source;
      dual.weight = arc.weight;
      dual.attributes = arc.attributes;
      this->presets[arc.ID].push_back(dual);
   }
   
   /*
    * expression defaults to EXPRESSION()
    */
   void PTNET::addTConnection(IDTYPE source, IDTYPE target, long weight)
   {
      TCONNECTION arc = TCONNECTION();
      arc.ID = target;
      arc.weight = weight;
      this->addTConnection(source, arc);
   }
   
   /*
    * weight defaults to 1
    */
   void PTNET::addTConnection(IDTYPE source, IDTYPE target, list<EXPRESSION> attributes)
   {
      TCONNECTION arc = TCONNECTION();
      arc.ID = target;
      arc.weight = 1;
      arc.attributes = attributes;
      this->addTConnection(source, arc);
   }
   
   void PTNET::addTConnection(IDTYPE source, IDTYPE target)
   {
      this->addTConnection(source, target, 1);
   }
   
   void PTNET::remove_y_fromPresetOf_x(IDTYPE x, IDTYPE y)
   {
      assert(this->presets.count(x) > 0);
      vector<TCONNECTION>::iterator arcit;
      
      /*
       * replace *x with a set without occurances of y
       */
      vector<TCONNECTION> xpre = vector<TCONNECTION>();
      for( arcit = this->presets[x].begin();
           arcit != this->presets[x].end();
           arcit++ )
      {
         if(arcit->ID != y) xpre.push_back(*arcit);
      }
      this->presets[x] = xpre;
   }

   void PTNET::remove_y_fromPostsetOf_x(IDTYPE x, IDTYPE y)
   {
      assert(this->postsets.count(x) > 0);
      vector<TCONNECTION>::iterator arcit;
      
      /*
       * replace x* with a set without occurances of y
       */
      vector<TCONNECTION> xpost = vector<TCONNECTION>();
      for( arcit = this->postsets[x].begin();
           arcit != this->postsets[x].end();
           arcit++ )
      {
         if(arcit->ID != y) xpost.push_back(*arcit);
      }
      this->postsets[x] = xpost;
   }   
      
   void PTNET::deleteTConnection(IDTYPE source, IDTYPE target)
   {
      /*
       * replace source* with a set without occurances of target
       */
      this->remove_y_fromPostsetOf_x(source, target);
      
      /*
       * replace *target with a set without occurances of source
       */
      this->remove_y_fromPresetOf_x(target, source);
   }

   void PTNET::deletePreset(IDTYPE id)
   {
      assert(this->presets.count(id) > 0);
   
      /*
       * remove id from the postset of all elements in *id
       */
      vector<TCONNECTION>::iterator arcit;
      for(arcit = this->presets[id].begin(); arcit != this->presets[id].end(); arcit++)
      {
         this->remove_y_fromPostsetOf_x(arcit->ID, id);
      }

      /*
       * *id <- {}
       */
      this->presets[id] = vector<TCONNECTION>();
   }

   void PTNET::deletePostset(IDTYPE id)
   {
      assert(this->postsets.count(id) > 0);

      /*
       * remove id from the preset of all elements in id*
       */
      vector<TCONNECTION>::iterator arcit;
      for(arcit = this->postsets[id].begin(); arcit != this->postsets[id].end(); arcit++)
      {
         this->remove_y_fromPresetOf_x(arcit->ID, id);
      }

      /*
       * id* <- {}
       */
      this->postsets[id] = vector<TCONNECTION>();
   }

   vector<TCONNECTION> PTNET::getPreset(IDTYPE id)
   {
      if(this->presets.count(id) == 0)
	return vector<TCONNECTION>();

      return this->presets[id];
   }

   vector<TCONNECTION> PTNET::getPostset(IDTYPE id)
   {
      if(this->postsets.count(id) == 0)
	return vector<TCONNECTION>();

      return this->postsets[id];
   }

   void PTNET::setPreset(IDTYPE id, vector<TCONNECTION> preset)
   {
      this->deletePreset(id);
      this->presets[id] = preset;

      vector<TCONNECTION>::iterator arcit;
      for(arcit = preset.begin(); arcit != preset.end(); arcit++)
      {
         assert(this->postsets.count(arcit->ID) > 0);
         TCONNECTION dual = TCONNECTION();

         dual.ID = id;
         dual.weight = arcit->weight;
         dual.attributes = arcit->attributes;
         this->postsets[arcit->ID].push_back(dual);
      }
   }

   void PTNET::setPostset(IDTYPE id, vector<TCONNECTION> postset)
   {
      this->deletePostset(id);
      vector<TCONNECTION>::iterator arcit;
      for(arcit = postset.begin(); arcit != postset.end(); arcit++)
      {
         this->addTConnection(id, *arcit);
      }
   }

   MARKING & PTNET::getInitialMarking()
   {
      return this->initial_marking;
   }

   MARKING & PTNET::getCurrentMarking()
   {
      return this->current_marking;
   }

   void PTNET::setInitialMarking(MARKING m)
   {
      this->initial_marking = m;
   }

   void PTNET::setCurrentMarking(MARKING m)
   {
      this->current_marking = m;
   }

   unsigned PTNET::getDepth(IDTYPE id)
   {
      return this->depth[id];
   }
   void PTNET::setDepth(IDTYPE id, unsigned d)
   {
      this->depth[id] = d;
      if(this->edges.count(id) > 0) this->edges[id].setDepth(d);
      if(d > this->maxDepth) this->maxDepth = d;
   }
   unsigned PTNET::getMaximalDepth() {return this->maxDepth;}

   ostream& operator<<(ostream& ost, PTNET& th)
   {
      ost << "place/transition net \"" << th.name << "\" (ID " << th.id << ") with\nplaces";
      map<IDTYPE, VERTEX>::iterator vit;
      for(vit = th.vertices.begin(); vit != th.vertices.end(); vit++)
      {
         ost << " v[" << vit->first << ":d" << th.getDepth(vit->first) << ']';
      }
      map<IDTYPE, EDGE>::iterator eit;
      for(eit = th.edges.begin(); eit != th.edges.end(); eit++)
      {
         ost << " e[" << eit->first << ":d" << th.getDepth(eit->first) << ']';
      }

      ost << "\nand the following transitions:";
      map<IDTYPE, TRANSITION>::iterator transit;
      vector<TCONNECTION>::iterator connectit;
      for(transit = th.transitions.begin(); transit != th.transitions.end(); transit++)
      {
         ost << "\n";
	 for( connectit = th.presets[transit->first].begin();
	      connectit != th.presets[transit->first].end();
	      connectit++ )
         {
	    ost << connectit->ID;
	    if(connectit->weight != 1) ost << "x" << connectit->weight;
	    ost << " ";
	 }
	 ost << " ---|" << transit->first << ":d" << th.getDepth(transit->first) << "|--> ";
	 for( connectit = th.postsets[transit->first].begin();
	      connectit != th.postsets[transit->first].end();
	      connectit++ )
         {
	    ost << " " << connectit->ID;
	    if(connectit->weight != 1) ost << "x" << connectit->weight;
	 }
      }
      ost << "\ninitial: " << th.initial_marking;
      ost << "\ncurrent: " << th.current_marking;
      return ost;
   }

   unsigned PTNET::WeightInPreSet(IDTYPE trans_id, IDTYPE id)
   {
    return WeightInSet(presets[trans_id], id);
   }

   unsigned PTNET::WeightInPostSet(IDTYPE trans_id, IDTYPE id)
   {
    return WeightInSet(postsets[trans_id], id);
   }

   unsigned PTNET::WeightInSet(vector<TCONNECTION> & pset, IDTYPE id)
   {
    for(vector<TCONNECTION>::iterator it=pset.begin(); it!=pset.end(); it++)
    {
     if(it->ID==id)
      return it->weight;
    }
    return 0;
   }

   TRANSITION PTNET::getTransition(IDTYPE id)
   {
    return transitions[id];
   }

   set<IDTYPE> PTNET::getTransitions(string s)
   {
    set<IDTYPE> result;

    for(map<IDTYPE, TRANSITION>::iterator it = transitions.begin(); it != transitions.end(); it++)
    {
     string srule = getAssociatedRule(it->second.getID())->getName();
     if(s==srule)
      result.insert(it->second.getID());
    }

    return result;
   }

   RULE* PTNET::getAssociatedRule(IDTYPE t)
   {
      if((this->associated_rules.count(t) > 0) && (this->grammar != NULL))
      {
         return this->grammar->getRule(associated_rules[t]);
      }
      else 
      {
       cout << "\nwarning: no such rule (PTNET::getAssociatedRule)" << endl;
      } 
   }

   map<IDTYPE, IDTYPE> * PTNET::getAssociatedRules()
   {
       return &associated_rules;
   }

   IDTYPE PTNET::getAssociatedRuleID(IDTYPE t)
   {
      if(this->associated_rules.count(t) > 0)
      {
         return associated_rules[t];
      }
      else 
      {
       cout << "\nwarning: no such rule (PTNET::getAssociatedRuleID)" << endl;
      }
   }

};

