/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef DATA_STRUCTURESPTNET_H
#define DATA_STRUCTURESPTNET_H

#include "expression.h"
#include "graph.h"
#include "marking.h"
#include "transition.h"
#include "gts.h"

namespace data_structures
{
   class TCONNECTION
   {
      public:
         TCONNECTION();
         ~TCONNECTION();

         TCONNECTION clone();

         IDTYPE ID;
         long weight;
         list<EXPRESSION> attributes;
   };

   class PTNET: public virtual GRAPH
   {
      public:
         PTNET();
         virtual ~PTNET();

         PTNET clone();

         virtual void addPlace(IDTYPE p, VERTEX v);
         virtual void addPlace(IDTYPE p, EDGE e);
         virtual void addPlace(IDTYPE p);
         virtual void deletePlace(IDTYPE p);
         set<IDTYPE> getPlaces();
         bool hasPlace(IDTYPE p);
         map<IDTYPE, TRANSITION> getTransitions();
	 
         void addTransition(IDTYPE t, TRANSITION tguard);
         void deleteTransition(IDTYPE t);
         bool hasTransition(IDTYPE t);

         void addTConnection(IDTYPE source, TCONNECTION arc);
         void addTConnection(IDTYPE source, IDTYPE target, long weight);
         void addTConnection(IDTYPE source, IDTYPE target, list<EXPRESSION> attributes);
         void addTConnection(IDTYPE source, IDTYPE target);
         void deleteTConnection(IDTYPE source, IDTYPE target);

         void deletePreset(IDTYPE id);
         void deletePostset(IDTYPE id);
         vector<TCONNECTION> getPreset(IDTYPE id);
         vector<TCONNECTION> getPostset(IDTYPE id);
         void setPreset(IDTYPE id, vector<TCONNECTION> preset);
         void setPostset(IDTYPE id, vector<TCONNECTION> postset);

         MARKING & getInitialMarking();
         MARKING & getCurrentMarking();
         void setInitialMarking(MARKING m);
         void setCurrentMarking(MARKING m);
	 	 
	 
	 virtual unsigned getDepth(IDTYPE id);
	 virtual void setDepth(IDTYPE id, unsigned d);
	 unsigned getMaximalDepth();	 
	 
	 TRANSITION getTransition(IDTYPE id);
	 RULE* getAssociatedRule(IDTYPE t);
	 map<IDTYPE, IDTYPE> * getAssociatedRules();
	 IDTYPE getAssociatedRuleID(IDTYPE t);
	 set<IDTYPE> getTransitions(string s);
	 
	 unsigned WeightInPreSet(IDTYPE trans_id, IDTYPE id);
	 unsigned WeightInPostSet(IDTYPE trans_id, IDTYPE id);
	 
	 friend ostream& operator<<(ostream& ost, PTNET& th);

	//Für QuickUnfold
	void setAssociatedRules(map<IDTYPE, IDTYPE> assoRules){associated_rules=assoRules;}
	void setTransitionsQuick(map<IDTYPE, TRANSITION> transitions){this->transitions=transitions;}
	map<IDTYPE, TRANSITION> getTransitionsQuick(){return transitions;}
	void setPresets(map<IDTYPE, vector<TCONNECTION> > presets){this->presets=presets;}
	map<IDTYPE, vector<TCONNECTION> > getPresets(){return presets;}
	void setPostsets(map<IDTYPE, vector<TCONNECTION> > postsets){this->postsets=postsets;}
	map<IDTYPE, vector<TCONNECTION> > getPostsets(){return postsets;}
	void setDepth(map<IDTYPE, unsigned> depth){this->depth=depth;}
	map<IDTYPE, unsigned> getDepth(){return depth;}
	void setMaxDepth(unsigned maxDepth){this->maxDepth=maxDepth;}
	unsigned getMaxDepth(){return maxDepth;}
	void setGrammar(GTS* grammar){this->grammar=grammar;}
	GTS* getGrammar(){return grammar;}

      protected:

         /*
	  * maps transition IDs to rule IDs
	  */
	 map<IDTYPE, IDTYPE> associated_rules;

         map<IDTYPE, TRANSITION> transitions;
         map< IDTYPE, vector<TCONNECTION> > presets;
         map< IDTYPE, vector<TCONNECTION> > postsets;
         MARKING initial_marking;
         MARKING current_marking;
	 map<IDTYPE, unsigned> depth;
	 unsigned maxDepth;
	 	 
	 GTS* grammar;	

      private: 
         void remove_y_fromPresetOf_x(IDTYPE x, IDTYPE y);
         void remove_y_fromPostsetOf_x(IDTYPE x, IDTYPE y);
	 
	 unsigned WeightInSet(vector<TCONNECTION> & pset, IDTYPE id);

   };
};
using namespace data_structures;

#endif
