/***************************************************************************
 *   Copyright (C) 2006 by Michael Mayer   *
 *   mayer@lippe   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "ConflictCalculator.h"
namespace scenarios{
namespace quickUnfolding{
namespace conflictCheck{

set<IDTYPE> ConflictCalculator::getPresetFromTransition(IDTYPE id, PETRIGRAPH pg){
	set<IDTYPE> result;
	vector<TCONNECTION> cons = pg.getPreset(id);
	for(vector<TCONNECTION>::iterator iter = cons.begin(); iter != cons.end(); iter++)
		result.insert(iter->ID);
	return result;
}

set<IDTYPE> ConflictCalculator::getPostsetFromTransition(IDTYPE id, PETRIGRAPH pg){
	set<IDTYPE> result;
	vector<TCONNECTION> cons = pg.getPostset(id);
	for(vector<TCONNECTION>::iterator iter = cons.begin(); iter != cons.end(); iter++)
		result.insert(iter->ID);
	return result;
}

void ConflictCalculator::initialize(map<IDTYPE, set<IDTYPE> >& removingEdge, map<IDTYPE, IDTYPE>& createdByTransition){
	this->name = "ConflictCalculator";
	this->removingEdge = &removingEdge;
	this->createdByTransition = &createdByTransition;
}

void ConflictCalculator::insertConflictinBranch(PETRIGRAPH pg, IDTYPE branch, IDTYPE conflict)
  {
    set<IDTYPE> postEdges = getPostsetFromTransition(branch,pg);
    for(set<IDTYPE>::iterator postIt = postEdges.begin();
	postIt != postEdges.end();
	postIt++)
      {
	map<IDTYPE, set<IDTYPE> >::iterator removed = removingEdge->find(*postIt);
	if(removed != removingEdge->end())
	  {
	    for(set<IDTYPE>::iterator removedIt = (*removed).second.begin();
		removedIt != (*removed).second.end();
		removedIt++)
	      {
		map<IDTYPE, set<IDTYPE> >::iterator removTrans = transitionConflicts->find(*removedIt);
		if(removTrans != transitionConflicts->end())
		  {
		    (*removTrans).second.insert(conflict);
		  } else {
		    set<IDTYPE> newSet;
		    newSet.insert(conflict);
		    transitionConflicts->insert(pair<IDTYPE, set<IDTYPE> >(*removedIt, newSet));
		  }
		map<IDTYPE, set<IDTYPE> >::iterator newTrans = transitionConflicts->find(conflict);
		if(newTrans != transitionConflicts->end())
		  {
		    (*newTrans).second.insert(*removedIt);
		  } else {
		    set<IDTYPE> newSet;
		    newSet.insert(*removedIt);
		    transitionConflicts->insert(pair<IDTYPE, set<IDTYPE> >(conflict, newSet));
		  }
		if(*removedIt != branch)
		  insertConflictinBranch(pg, *removedIt, conflict);
	      }
	  }
      }
}

void ConflictCalculator::run(PETRIGRAPH& pg, list<IDTYPE>& transitions){
	for(list<IDTYPE>::iterator transition = transitions.begin();
	transition != transitions.end();
	transition++)
      {
	set<IDTYPE> conflicts;	
	set<IDTYPE> preset = getPresetFromTransition(*transition,pg);

	for(set<IDTYPE>::iterator presetIt = preset.begin();
	    presetIt != preset.end();
	    presetIt++)
	  {

		set<IDTYPE> edges = (*removingEdge->find(*presetIt)).second;
		conflicts.insert(edges.begin(), edges.end());
		IDTYPE created = (*createdByTransition->find(*presetIt)).second;
		if(created != 0)
		  { 
		    if(transitionConflicts->find(created) != transitionConflicts->end())
		      {
			set<IDTYPE> oldConflicts = (*transitionConflicts->find(created)).second;
			conflicts.insert(oldConflicts.begin(), oldConflicts.end());
		      }
		  }

	  }
	   
	if(conflicts.find(*transition) != conflicts.end())
	  {
	    conflicts.erase(*transition);
	  }
	if(transitionConflicts->find(*transition) == transitionConflicts->end())
	  {
	    transitionConflicts->insert(pair<IDTYPE, set<IDTYPE> >(*transition, conflicts));
	  } else {
	    (*transitionConflicts->find(*transition)).second.insert(conflicts.begin(), conflicts.end());
	  }

	//inserting these transitions in the conflicts, because of symmetric
	for(set<IDTYPE>::iterator conIt = conflicts.begin();
	    conIt != conflicts.end();
	    conIt++)
	  {
	    map<IDTYPE, set<IDTYPE> >::iterator confTrans = transitionConflicts->find(*conIt);
	    if(confTrans != transitionConflicts->end())
	      {
		(*confTrans).second.insert(*transition);
	      } else {
		set<IDTYPE> newConflictSet;
		newConflictSet.insert(*transition);
		transitionConflicts->insert(pair<IDTYPE, set<IDTYPE> >(*conIt, newConflictSet));
	      }
	    double inConTemp = double(clock());
	    insertConflictinBranch(pg, *conIt, *transition);
	    //insertConflictTime += (double(clock()) - inConTemp) / CLOCKS_PER_SEC ;
	  }
	
      }
}

}
}
}
