/***************************************************************************
 *   Copyright (C) 2006 by Michael Mayer   *
 *   mayer@lippe   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "InterfaceChecker.h"
#include "checkEdge.h"

namespace scenarios{
namespace quickUnfolding{

InterfaceChecker::InterfaceChecker(PETRIGRAPH pg){
	  map<IDTYPE, RULE>* rules = pg.getGrammar()->getRules();
    for(map<IDTYPE, RULE>::iterator rule = rules->begin();
	rule != rules->end();
	rule++)
      {
	set<IDTYPE> leftEdges = coverability::getKeys((*rule).second.getLeft()->getEdges());
	set<IDTYPE> rightEdges = coverability::getKeys((*rule).second.getRight()->getEdges());
	for(set<IDTYPE>::iterator leftEdge = leftEdges.begin();
	    leftEdge != leftEdges.end();
	    leftEdge++)
	  {
	    if(rightEdges.count(*leftEdge) == 1)
	      {
		list<IDTYPE> leftVertices = rule->second.getLeft()->getConnectedVertices(*leftEdge);
		list<IDTYPE> rightVertices = rule->second.getRight()->getConnectedVertices(*leftEdge);
		if(leftVertices.size() != rightVertices.size())
		  {
		    cerr << "Different number of vertices of Edge " << (*rule).second.getLeft()->getEdges()[*leftEdge].getLabel() << " in context of rule " << (*rule).second.getName() << ".\n";
		    validInterface = false;
			return;
		  }
		list<IDTYPE>::iterator leftVertice = leftVertices.begin();
		list<IDTYPE>::iterator rightVertice = rightVertices.begin();
		while(leftVertice != leftVertices.end())
		  {
		    if((*leftVertice) != (*rightVertice))
		      {
			cout << "Different vertices of Edge " << rule->second.getLeft()->getEdges()[*leftEdge].getLabel() << " in context of rule " << rule->second.getName() << ".\n";
			validInterface = false;
			return;
		      }
		    leftVertice++;
		    rightVertice++;
		  }
	      }
	  }
      }
    validInterface = true;
}

}
}
