/***************************************************************************
 *   Copyright (C) 2006 by Michael Mayer   *
 *   mayer@lippe   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "checkEdge.h"

namespace scenarios{
namespace quickUnfolding{
namespace coverability{

CheckEdge::CheckEdge(PETRIGRAPH &  pg, RULE rule, IDTYPE ruleedge, IDTYPE graphedge, map<IDTYPE, IDTYPE> phi_e, map<IDTYPE,IDTYPE> phi_v)
  {
    matching = true;
    bool initial = (phi_e.size() == 0);//function call is initial

    for(map<IDTYPE, IDTYPE>::iterator pe_it = phi_e.begin();
	pe_it != phi_e.end();
	pe_it++){
		if(graphedge == (*pe_it).second){
	    //graphedge may be only used once for mapping (Kanten dürfen nur einmal gemappt werden)
	    	matching = false;
	  	}
    }
    list<IDTYPE> graphV = pg.getConnectedVertices(graphedge);
    list<IDTYPE> ruleV = rule.getLeft()->getConnectedVertices(ruleedge);

    // ID in rule mapped to ID in graph
    map<IDTYPE, IDTYPE> newMappedVertices;

    if(graphV.size() == ruleV.size()) 
      {
	//number of connected Vertices is equal
       list<IDTYPE>::iterator graphVit = graphV.begin();
       list<IDTYPE>::iterator ruleVit = ruleV.begin();

	while((ruleVit != ruleV.end()) && (matching)) 
	{
	  map<IDTYPE, IDTYPE>::iterator phi_vIt = phi_v.find(*ruleVit);

	  //if *graphVit is not mapped, map it
	  if(phi_vIt == phi_v.end())
	    {
	      //mapped in newMappedVertices?
	      map<IDTYPE, IDTYPE>::iterator nMVit = newMappedVertices.find(*ruleVit);
	      if(nMVit == newMappedVertices.end()) {
		newMappedVertices.insert(pair<IDTYPE, IDTYPE>(*ruleVit, *graphVit));
	      } else {
		if((*nMVit).second != (*graphVit)) {
		  matching = false;
		}
	      }
	    } else {
	      if((*phi_vIt).second != (*graphVit)){
		// edge in phi_V already used
		matching = false;
	      }
	    }
	  graphVit++;
	  ruleVit++;
	}
	

      } else {

	matching = false;

      }

    //matching is true, if all connected vertices are mapped correctly
    if(matching) {

      map<IDTYPE, IDTYPE> edgeMapped;

      //adding of the new mapped edge
      edgeMapped.insert(pair<IDTYPE,IDTYPE>(ruleedge, graphedge));

      //adding new found mappings to resultList
      resultList.push_back(pair<map<IDTYPE, IDTYPE>, map<IDTYPE,IDTYPE> >(edgeMapped, newMappedVertices));


      //for all vertices of the rule starting the checkNode function, the answer of the
      //function has to be inserted in the mapping 
      map<IDTYPE,IDTYPE>::iterator newMappedVerticesIt = newMappedVertices.begin();
      while (matching && (newMappedVerticesIt != newMappedVertices.end())) {

	 matching = false;

	 //list of the result for this node
	 list<pair<map <IDTYPE,IDTYPE>, map<IDTYPE, IDTYPE> > > Vresult;

	 //diese Schleife versucht für alle Möglichkeiten der resultList ein Mapping zu finden
	 //wenn keine Mapping gefunden wurde, muss dieses Mapping entfernt werden
	 // die resultList wird für jeden Knoten neu erstellt und dann kopiert

	 //trying to comlete the mapping for the vertex of the graph with each mapping
	 //if answer is false, the mapping has to be removed
	 for(list<pair<map <IDTYPE,IDTYPE>, map<IDTYPE, IDTYPE> > >::iterator rsListIt = resultList.begin();
	     rsListIt != resultList.end();
	     rsListIt++)
	   {

	     map<IDTYPE, IDTYPE> phi_e_union(phi_e);
	     phi_e_union.insert((*rsListIt).first.begin(), (*rsListIt).first.end());

	     map<IDTYPE, IDTYPE> phi_v_union(phi_v);
	     phi_v_union.insert((*rsListIt).second.begin(), (*rsListIt).second.end());
	     //merging the mapping of phi and of the resultList

	     CheckNode result(pg, rule, (*newMappedVerticesIt).first , phi_e_union , phi_v_union);

	     if(result.isCoverable()) 
	       {
		 matching = true;

		 if(result.getMapping().size() > 0)
		   {
		     for(list<pair<map <IDTYPE,IDTYPE>, map<IDTYPE, IDTYPE> > >::iterator resultIt = result.getMapping().begin();
			 resultIt != result.getMapping().end();
			 resultIt++)
		       {
			 map<IDTYPE,IDTYPE> edgesUnion((*resultIt).first);
			 map<IDTYPE,IDTYPE> verticesUnion((*resultIt).second);
			 //adding all the edges and vertices of resultIt
			 for(map<IDTYPE, IDTYPE>::iterator it = (*rsListIt).first.begin();
			     it != (*rsListIt).first.end();
			     it++)
			   {
			     edgesUnion.insert((*it));
			   }
			 for(map<IDTYPE, IDTYPE>::iterator it = (*rsListIt).second.begin();
			     it != (*rsListIt).second.end();
			     it++)
			   {
			     verticesUnion.insert(*it);
			   }
			 //building the result
			 Vresult.push_back(pair<map<IDTYPE,IDTYPE>, map<IDTYPE,IDTYPE> >(edgesUnion, verticesUnion));
		       } 
		   } else {
		     // if the result is true, but no further mappings necessary 
		     Vresult.push_back(*rsListIt);
		   }
	       }
	     
	   }
	 //for this vertex the result in Vresult is complete and is the new resultList
	 resultList = Vresult;
	 newMappedVerticesIt++;

       }
    }
    //if the left side of the rule is not full connected, each part of the graph has to be mapped
    if(initial)
      {

	list<pair<map <IDTYPE,IDTYPE>, map<IDTYPE, IDTYPE> > > anotherResultList;

	bool complete = false;
	while(!complete)
	  {
	    bool missing = true;
	    complete = true;
	    for(list<pair<map <IDTYPE,IDTYPE>, map<IDTYPE, IDTYPE> > >::iterator rlIt = resultList.begin();
		rlIt != resultList.end();
		rlIt++)
	      {
		set<IDTYPE> ruleEdges = getKeys(rule.getLeft()->getEdges());
		//not all edges of the left side are mapped due to partial graphs
		if(ruleEdges.size() !=  (*rlIt).first.size())
		  {
		    missing = true;
		    complete = false;
		    for(set<IDTYPE>::iterator ruleIt = ruleEdges.begin();
			ruleIt != ruleEdges.end();
			ruleIt++)
		      {
			if((*rlIt).first.find(*ruleIt) == ((*rlIt).first.end()))
			  {
			    set<IDTYPE> graphEdges = getKeys(pg.getEdges());
			    //looking for all edges in the whole graph with the same label
			    for(set<IDTYPE>::iterator gEIt = graphEdges.begin();
				gEIt != graphEdges.end();
				gEIt++)
			      {
				if(rule.getLeft()->getEdge(*ruleIt).getLabel() == pg.getEdge(*gEIt).getLabel())
				  {	
					CheckEdge oneResult(pg, rule, *ruleIt, *gEIt, (*rlIt).first, (*rlIt).second);
				    if(oneResult.isCoverable()) 
				      {
					matching = true;
					
					if(oneResult.getMapping().size() > 0)
					  {
					    for(list<pair<map <IDTYPE,IDTYPE>, map<IDTYPE, IDTYPE> > >::iterator oneResultIt = oneResult.getMapping().begin();
						oneResultIt != oneResult.getMapping().end();
						oneResultIt++)
					      {
						
						map<IDTYPE,IDTYPE> edgesUnion((*oneResultIt).first);
						map<IDTYPE,IDTYPE> verticesUnion((*oneResultIt).second);
						//adding all vertices of rsListIt.first
						map<IDTYPE, IDTYPE> onePhie = (*rlIt).first;
						map<IDTYPE, IDTYPE> onePhiv = (*rlIt).second;
						for(map<IDTYPE, IDTYPE>::iterator it = onePhie.begin();
						    it != onePhie.end();
						    it++)
						  {
						    edgesUnion.insert((*it));
						  }
						for(map<IDTYPE, IDTYPE>::iterator it = onePhiv.begin();
						    it != onePhiv.end();
						    it++)
						  {
						    verticesUnion.insert(*it);
						  }
						// adding new results to the new resultlist
						anotherResultList.push_back(pair<map<IDTYPE,IDTYPE>, map<IDTYPE,IDTYPE> >(edgesUnion, verticesUnion));
					      } 
					  } else {
					  }
				      } else {
					continue;
				      }

				  }

			      }
			    break;
			  } 			    
			
		      }
		  } else {
		    //adding new result to the new resultlist
		    anotherResultList.push_back(*rlIt);
		  }
	      }
	    //new resultlist is now the resultlist
	    resultList = anotherResultList;
	    anotherResultList.clear();

	  }
   }
}

}
}
}
