/***************************************************************************
 *   Copyright (C) 2006 by Michael Mayer   *
 *   mayer@lippe   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "cutoffCheck.h"
#include "mapGraph.h"

namespace scenarios{
namespace quickUnfolding{
namespace newUnfold{

CutoffCheck::CutoffCheck(PETRIGRAPH pg,IDTYPE transition, set<IDTYPE> preSet, set<IDTYPE> postSet, map<IDTYPE, set<IDTYPE> >& necessaryTransitions, map<IDTYPE, TransitionCertificate>& transitionCertificates, map<IDTYPE, ExtendedTransitionCertificate>& extendedTransitionCertificates)
  {
    set<IDTYPE> necessaryTrans;
    //    set<IDTYPE> testnecessaryTrans;
    for(set<IDTYPE>::iterator edge = preSet.begin();
	edge != preSet.end();
	edge++)
      {
	//	getNecessaryTransitions(pg, necessaryTrans, (*edge));
	if(necessaryTransitions.find(*edge) != necessaryTransitions.end())
	  {
	    set<IDTYPE> edgeNecTrans = (*necessaryTransitions.find(*edge)).second;
	    necessaryTrans.insert(edgeNecTrans.begin(), edgeNecTrans.end());
	  }
      }


    //    if(testnecessaryTrans.size() != necessaryTrans.size())
    //      cout << "Unterschied alt: " << necessaryTrans << " \n versus neu: " << testnecessaryTrans << "\n";
    
    set<IDTYPE> newGraphEdges(coverability::getKeys(pg.getGrammar()->getStartgraph()->getEdges()));

    for(set<IDTYPE>::iterator transitionIt = necessaryTrans.begin();
	transitionIt != necessaryTrans.end();
	transitionIt++)
      {
	set<IDTYPE> postedges = convert(pg.getPostset(*transitionIt));
	newGraphEdges.insert(postedges.begin(), postedges.end());
      }
    for(set<IDTYPE>::iterator transitionIt = necessaryTrans.begin();
	transitionIt != necessaryTrans.end();
	transitionIt++)
      {
	set<IDTYPE> preedges = convert(pg.getPreset(*transitionIt));
	for(set<IDTYPE>::iterator preedge = preedges.begin();
	    preedge !=  preedges.end();
	    preedge++)
	  {
	    newGraphEdges.erase(*preedge);
	  }
	
      }

    //inserting the new transition
    necessaryTrans.insert(transition);
    newGraphEdges.insert(postSet.begin(), postSet.end());
    for(set<IDTYPE>::iterator preedge = preSet.begin();
	preedge != preSet.end();
	preedge++)
      {
	newGraphEdges.erase(*preedge);
      }


    map<LABELTYPE, set<IDTYPE> >newGraph;
    for(set<IDTYPE>::iterator newEdge = newGraphEdges.begin();
 	newEdge != newGraphEdges.end();
	newEdge++)
      {
	LABELTYPE label = pg.getEdges()[*newEdge].getLabel();//Label einer Kante mit der ID *newEdge
	if(newGraph.find(label) != newGraph.end())
	  {
	    (*newGraph.find(label)).second.insert(*newEdge);
	  } else {
	    set<IDTYPE> newSet;
	    newSet.insert(*newEdge);
	    newGraph.insert(pair<LABELTYPE, set<IDTYPE> >(label, newSet));
	  }
      }

    if(!exCerts)
      {
	string certificate = certificates::GenerateCertificate(newGraph).getCertificate();

	certificates::TransitionCertificate tg(certificate, newGraph, necessaryTrans.size());
	
	
	for(map<IDTYPE, certificates::TransitionCertificate>::iterator isocertificate = transitionCertificates.begin();
	    isocertificate != transitionCertificates.end();
	    isocertificate++)
	  {
	    
      	    if(necessaryTrans.size() > (*isocertificate).second.pretransitions)
	      {
		if(certificate == (*isocertificate).second.certificate)
		  {
		    
		    //cutChecks++; only for debuging
		    if(compareGraph(pg, tg, (*isocertificate).second))
		      {
			//cutSuccess++; only for debuging
			cutable = true;
		      }
		  }
	      }
	    
	  }
	transitionCertificates.insert(pair<IDTYPE, TransitionCertificate>(transition, tg));
      } else {
	
	list<pair<IDTYPE, list<pair<LABELTYPE, int> > > > newCertificate;
	
	newCertificate = certificates::GenerateExtendedCertificate(pg, newGraph).getCertificate();
	
	certificates::ExtendedTransitionCertificate etg(newCertificate, newGraph, necessaryTrans.size());
	
	for(map<IDTYPE, ExtendedTransitionCertificate>::iterator isocertificate = extendedTransitionCertificates.begin();
	    isocertificate != extendedTransitionCertificates.end();
	    isocertificate++)
	  {
	    if(exCertisEq(etg, (*isocertificate).second))
	      {
//		cutChecks++; only for debuging
		
		if(etg.pretransitions > (*isocertificate).second.pretransitions)
		  {
		    map<IDTYPE, IDTYPE> mapping;
		    //graphs can only be equal, if number of edges are equal, not checked by the certificates
		    if(etg.graph.size() == (*isocertificate).second.graph.size())
		      {
			if(certificates::CompareExtendedCertificate(pg, etg.certificate, (*isocertificate).second.certificate, etg.graph, (*isocertificate).second.graph, mapping).isEqual())
			  {
			    //cutSuccess++;only for debuging
			    
			    cutable = true;
			  }
		      }
		  }
	      }
	  }
	extendedTransitionCertificates.insert(pair<IDTYPE, ExtendedTransitionCertificate>(transition, etg));
      }

      

    cutable = false;
  }

set<IDTYPE>	CutoffCheck::convert(vector<TCONNECTION> tcons){
	set<IDTYPE> result;
	for(vector<TCONNECTION>::iterator iter = tcons.begin(); iter != tcons.end(); iter++)
		result.insert(iter->ID);
	return result;
}

bool CutoffCheck::compareGraph(HYPERGRAPH hg, TransitionCertificate cut, TransitionCertificate isoTransitionCertificate)
{
    if(cut.pretransitions <= isoTransitionCertificate.pretransitions)
      {
	return false;
      }

    map<LABELTYPE, set<IDTYPE> > cutGraph = cut.graph;
    map<LABELTYPE, set<IDTYPE> > isoGraph = isoTransitionCertificate.graph;

    map<IDTYPE, IDTYPE> mapping; //isoGraph Edge wird nach cutGraph gemapped

    return (newUnfold::MapGraph(hg, cutGraph, isoGraph, mapping).isEqual());
}

bool CutoffCheck::exCertisEq(certificates::ExtendedTransitionCertificate a, certificates::ExtendedTransitionCertificate b){
	list<pair<IDTYPE, list<pair<LABELTYPE, int> > > >  aCert = a.certificate;
    list<pair<IDTYPE, list<pair<LABELTYPE, int> > > >  bCert = b.certificate;

    if(bCert.size() != aCert.size())
      return false;

    list<pair<IDTYPE, list<pair<LABELTYPE, int> > > >::iterator aVertex = aCert.begin();
    list<pair<IDTYPE, list<pair<LABELTYPE, int> > > >::iterator bVertex = bCert.begin();

    while(aVertex != aCert.end())
      {

	if((*aVertex).second.size() != (*bVertex).second.size())
	  return false;

	list<pair<LABELTYPE, int> >::iterator aNode = (*aVertex).second.begin();
	list<pair<LABELTYPE, int> >::iterator bNode = (*bVertex).second.begin();

	while(aNode != (*aVertex).second.end())
	  {
	    if((*aNode).first != (*bNode).first)
	      return false;

	    if((*aNode).second != (*bNode).second)
	      return false;

	    aNode++;
	    bNode++;
	  }
	aVertex++;
	bVertex++;
      }

    return true;
}

}
}
}
