/***************************************************************************
 *   Copyright (C) 2006 by Michael Mayer   *
 *   mayer@lippe   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef DIRECTCONFLICTCHECKER_CPP
#define DIRECTCONFLICTCHECKER_CPP

#include "directConflictChecker.h"

namespace scenarios{
namespace quickUnfolding{
namespace conflictCheck{

void DirectConflictChecker::initialize(map<IDTYPE, set<IDTYPE> >& necessaryTransitions, map<IDTYPE, set<IDTYPE> >& directConflicts){
	this->necessaryTransitions = &necessaryTransitions;
	this->directConflicts = &directConflicts;
}

void DirectConflictChecker::run(PETRIGRAPH pg, map<IDTYPE, IDTYPE> preset){
	set<IDTYPE> necTransitionsUnion;
    for(map<IDTYPE, IDTYPE>::iterator edge = preset.begin();
	edge != preset.end();
	edge++)
      {
		if(necessaryTransitions->find((*edge).second) != necessaryTransitions->end())
	  	{
	   	 	set<IDTYPE> edgeNecTrans = (*necessaryTransitions->find((*edge).second)).second;
	    	necTransitionsUnion.insert(edgeNecTrans.begin(), edgeNecTrans.end());
	  	}
      }
    if(context)
      {
	set<IDTYPE> toCheck;
	toCheck.insert(necTransitionsUnion.begin(), necTransitionsUnion.end());
	set<IDTYPE> branch;
	while(toCheck.begin() != toCheck.end())
	  {
	    if(cycleChecker(toCheck, necTransitionsUnion, branch, (*toCheck.begin())) == false)
	      {
		coverable = false;
		return;
	      } 
	  }

      } else {
	set<IDTYPE> conflictUnion;
	conflictUnion = getDirectConflicts(necTransitionsUnion);

	for(set<IDTYPE>::iterator necTraTest = necTransitionsUnion.begin();
	    necTraTest != necTransitionsUnion.end();
	    necTraTest++)
	  {
	    if(conflictUnion.find(*necTraTest) != conflictUnion.end())
	      {
		coverable = false;	
		return;
	      }
	  }
      }
    coverable = true;
}

set<IDTYPE> DirectConflictChecker::getDirectConflicts(set<IDTYPE> transitions)
  {
    set<IDTYPE> conflicts;
    for(set<IDTYPE>::iterator transition = transitions.begin();
	transition != transitions.end();
	transition++)
      {
	if(directConflicts->find(*transition) != directConflicts->end())
	  {
	    set<IDTYPE> dirCon = (*directConflicts->find(*transition)).second;
	    if(dirCon.size() > 0)
	      conflicts.insert(dirCon.begin(), dirCon.end());
	  }
      }
    return conflicts;
  }

bool DirectConflictChecker::cycleChecker(set<IDTYPE> & toCheck, set<IDTYPE> necTransitions, set<IDTYPE> branch, IDTYPE transitionToCheck) {

    if(branch.count(transitionToCheck) != 0)
      {
	set<IDTYPE> testset(branch.find(transitionToCheck), branch.end());/*
	if(debug && testset.size() > 2)
	  {
	    cout << "Found Circle: " << testset << " following " << transitionToCheck << "\n";
	  }*/
	return false;
      }
    branch.insert(transitionToCheck);
    if(toCheck.count(transitionToCheck) != 0)
    {
      toCheck.erase(transitionToCheck);
      map<IDTYPE, set<IDTYPE> >::iterator dCIt = directConflicts->find(transitionToCheck);
      if(dCIt != directConflicts->end())
	{
	  set<IDTYPE> conflicts = (*dCIt).second;
	  for(set<IDTYPE>::iterator conflict = conflicts.begin();
	      conflict != conflicts.end();
	      conflict++)
	    {
	      if(necTransitions.count(*conflict) != 0)
		{
		  if(cycleChecker(toCheck, necTransitions, branch, (*conflict)) == false)
		    {
		      return false;
		    }
		}
	    }
	}
    }
    return true;
  }

}
}
}

#endif
