/***************************************************************************
 *   Copyright (C) 2006 by Michael Mayer   *
 *   mayer@lippe   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef FINDMATCHEDGE_CPP
#define FINDMATCHEDGE_CPP
#include "findMatchEdge.h"
#include "petrigraph.h"
#include "checkEdge.h"
namespace scenarios{
namespace quickUnfolding{

FindMatchEdge::FindMatchEdge(PETRIGRAPH & pg, RULE rule){
    //searching the most seldom edge of the left side in the graph
    pair<IDTYPE, int> minimum = findMinimumEdge(pg , rule);
    matching = false;

	if(minimum.second != 0){//one edge doesn't appear in the graph, no match possible
		map<IDTYPE,IDTYPE> phi_e; // mapping: first is the ID of the edge of the rule, second of the graph
		map<IDTYPE,IDTYPE> phi_v; // same with vertices
	
		pair<map<IDTYPE,IDTYPE>, map<IDTYPE,IDTYPE> >  phi =  pair<map<IDTYPE,IDTYPE>, map<IDTYPE,IDTYPE> >(phi_e, phi_v);
	
		std::string minimumedgelabel = rule.getLeft()->getEdges()[minimum.first].getLabel();
	
		set<IDTYPE> startGraEdges = coverability::getKeys(pg.getEdges());
	
		//mapping the minimum edge to all edges of the graph with the same labels
		for(set<IDTYPE>::iterator graphedge = startGraEdges.begin();
		graphedge != startGraEdges.end();
		graphedge++)
		{
	
		std::string graphedgelabel = pg.getEdges()[*graphedge].getLabel();
	
		if(graphedgelabel == minimumedgelabel) 
		{
			//if graphedge has the same label like the minimum edge, look for a mapping
			coverability::CheckEdge oneResult(pg, rule, minimum.first, (*graphedge), phi_e, phi_v);
			if(oneResult.isCoverable()) 
			{
			matching = true;
			resultList.merge(oneResult.getMapping());
			}
		}
		}
	}
  }

pair<IDTYPE, int> FindMatchEdge::findMinimumEdge(PETRIGRAPH & pg, RULE rule){
    
    pair<IDTYPE, int> minimal;

    set<IDTYPE> ruleLeftEdges = coverability::getKeys((rule.getLeft()->getEdges()));
    
    set<IDTYPE>::iterator anedge = ruleLeftEdges.begin();

    std::string edgelabel =  rule.getLeft()->getEdges()[*anedge].getLabel();

    int number = countAppearances(pg, edgelabel);
    minimal = pair<IDTYPE,int> ((*anedge), number);
    anedge++;
    while(minimal.second > 0 && anedge != ruleLeftEdges.end())
      {  
	edgelabel =  rule.getLeft()->getEdges()[*anedge].getLabel();
	number = countAppearances(pg, edgelabel);
	if(number < minimal.second) 
	  {  
	    minimal = pair<IDTYPE,int> ((*anedge), number);
	  }
	anedge++;
      }
    
    return minimal;

  }

  int FindMatchEdge::countAppearances(PETRIGRAPH & pg, std::string edgelabel) {
    int count = 0;
    set<IDTYPE> edges = coverability::getKeys((pg.getEdges()));
    for(set<IDTYPE>::iterator edgesIt = edges.begin();
	edgesIt != edges.end();
	edgesIt++)
      {
	
	if(pg.getEdges()[*edgesIt].getLabel() == edgelabel) {
	  count++;
	}
      }
    return count;  
  }

}
}
#endif
