/***************************************************************************
 *   Copyright (C) 2006 by Michael Mayer   *
 *   mayer@lippe   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef NEWUNFOLD_CPP
#define NEWUNFOLD_CPP

#include "newUnfold.h"

namespace scenarios{
namespace quickUnfolding{
namespace newUnfold{

void NewUnfold::initialize(map<IDTYPE, map<IDTYPE, IDTYPE> > &possibleRules, map<IDTYPE, ExtendedTransitionCertificate> &extendedTransitionCertificates, map<IDTYPE, map<IDTYPE, map<IDTYPE, IDTYPE> > > &dependencies, map<IDTYPE, IDTYPE> &createdByTransition, map<IDTYPE, TransitionCertificate> &transitionCertificates, map<IDTYPE, set<IDTYPE> >& concurrentEdges){
	this->possibleRules = &possibleRules;
	this->extendedTransitionCertificates = &extendedTransitionCertificates;
	this->dependencies = &dependencies;
	this->createdByTransition = &createdByTransition;
	this->transitionCertificates = &transitionCertificates;
	this->concurrentEdges = &concurrentEdges;
}

IDTYPE	NewUnfold::createNewID()
{
	counter++;
	return counter;
}

void NewUnfold::run(PETRIGRAPH & pg, RULE rule, pair<map<IDTYPE, IDTYPE>, map<IDTYPE, IDTYPE> > phi){

    transition = 0;

    counter  = pg.getLastUsedID();

    HYPERGRAPH rightHandSide = *(rule.getRight());
    HYPERGRAPH leftHandSide = *(rule.getLeft());

    set<IDTYPE> leftEdges = coverability::getKeys(leftHandSide.getEdges());
    set<IDTYPE> rightEdges = coverability::getKeys(rightHandSide.getEdges());

    set<IDTYPE> preSet;
    map<IDTYPE, unsigned> preweight;

    set<IDTYPE> postSet;
    map<IDTYPE, unsigned> postweight;

    set<IDTYPE> contextSet;
    map<IDTYPE, unsigned> contextweight;

    //necessary to remove the newVertices and newEdges in a cutoff-event
    set<IDTYPE> newVertices;
    set<IDTYPE> newEdges;

    map<IDTYPE, IDTYPE> mu_t;

    for(map<IDTYPE, IDTYPE>::iterator presetIt = phi.first.begin();
	presetIt != phi.first.end();
	presetIt++)
      {
	if(context && (rightEdges.find((*presetIt).first) != rightEdges.end()))
	  {
	    contextSet.insert((*presetIt).second);
	    contextweight.insert(pair<IDTYPE, unsigned>((*presetIt).second, 1));
	  } else {
	    preSet.insert((*presetIt).second);
	    preweight.insert(pair<IDTYPE, unsigned>((*presetIt).second, 1));
	  }
	mu_t.insert(*presetIt);
      }


    //looking for new vertices on the right hand side und adding these to graph and phi
    set<IDTYPE> rightVertices = coverability::getKeys(rightHandSide.getVertices());
    set<IDTYPE> leftVertices = coverability::getKeys(leftHandSide.getVertices());
    for(set<IDTYPE>::iterator rightVertex = rightVertices.begin();
	rightVertex != rightVertices.end();
	rightVertex++)
      {
	if(leftVertices.count(*rightVertex) == 0) 
	  {
	    //creating the vertex
	    IDTYPE newID = createNewID();
	    pg.addVertex(rightHandSide.getVertices()[*rightVertex],newID);//experimentel !!!
	    newVertices.insert(newID);
	    mu_t[*rightVertex] = newID;
	    phi.second.insert(pair<IDTYPE, IDTYPE>(*rightVertex, newID));
	  }
      }
			
    //Berechnen der nötigen Transitionen zur Ausführung dieser Entfaltung
    //Alle neu erzeugten Edges bekommen die Union der Listen der benötigten Listen
    //inklusive dieser auszuführenden Transition. Gespeichtert in necessaryTransitions.
    //Eintragen in removedEdges, dass diese Transitionen die Edges entfernen
    //Eintragen der Edges in necessaryEdges

    //nach oben verschoben
    //    set<IDTYPE> leftEdges = leftHandSide.getEdges();
    //    set<IDTYPE> rightEdges = rightHandSide.getEdges();

    //Alle Edges werden neu eingefügt, falls es keinen Kontext gibt
    //Die Vertices, die nicht in der linken Seite (also auch im Mapping sind), werden ebenfalls eingefügt
    list<IDTYPE> newConnectionList;
    for(set<IDTYPE>::iterator edgesIt = rightEdges.begin();
	edgesIt != rightEdges.end();
	edgesIt++)
      {
	list<IDTYPE> connectedV = rightHandSide.getConnectedVertices(*edgesIt);
	for(list<IDTYPE>::iterator verticesIt = connectedV.begin();
	    verticesIt != connectedV.end();
	    verticesIt++)
	  {
	    newConnectionList.push_back((*phi.second.find(*verticesIt)).second);
	    mu_t[*verticesIt] = (*phi.second.find(*verticesIt)).second;
	  }
	IDTYPE response;
	if(phi.first.count(*edgesIt) == 0)
	  {
		pg.addEdge(rightHandSide.getEdges()[*edgesIt],response,newConnectionList);//evtl. falsch ID
	    response++; 
	    newEdges.insert(response);
	    postSet.insert(response);
	    postweight.insert(pair<IDTYPE, unsigned>(response, 1));
	  } else {
	    //edge is in the context and won't be created
	    response = (*phi.first.find(*edgesIt)).second;
	    contextSet.insert(response);
	    contextweight.insert(pair<IDTYPE, unsigned>(response, 1));
	  }

	mu_t[*edgesIt] = response;

	phi.first.insert(pair<IDTYPE, IDTYPE>((*edgesIt), response));

	newConnectionList.clear();
      }

//    PTNetTransition newTransition = PTNetTransition(preSet, postSet, contextSet, preweight, postweight, contextweight, rule.getID());
    transition = createNewID();

    double tempCut = double(clock()) / CLOCKS_PER_SEC;
/*
    if(cutoff && CutoffCheck(pg, transition, preSet, postSet, necessaryTransitions, transitionCertificates, extendedTransitionCertificates).isCutable())
      {
	//cutoff event, deleting edges and vertices and returning a 0
	for(set<IDTYPE>::iterator newEdge = newEdges.begin();
	    newEdge != newEdges.end();
	    newEdge++)
	  {
	    pg.deleteEdge(*newEdge);
	  }
	for(set<IDTYPE>::iterator newVertex = newVertices.begin();
	    newVertex != newVertices.end();
	    newVertex++)
	  {
	    pg.deleteVertex(*newVertex);
	  }
       
//	timeCut += double(clock()) / CLOCKS_PER_SEC - tempCut;
	transition = 0;
      }*/
//    timeCut += double(clock()) / CLOCKS_PER_SEC - tempCut;


//vitali fragen! (transition, welche information ist wichtig?)
//    pg.addTransition(transition);



    map<IDTYPE, map<IDTYPE, map<IDTYPE, IDTYPE> > >::iterator depRuleIt = dependencies->find(rule.getID());
    if(depRuleIt != dependencies->end())
      {
	map<IDTYPE, map<IDTYPE, IDTYPE> > ruleDependencies = (*depRuleIt).second;

	//adding the suggestions for the next round
	for(map<IDTYPE, map<IDTYPE, IDTYPE> >::iterator newrulesIt = ruleDependencies.begin();
	    newrulesIt != ruleDependencies.end();
	    newrulesIt++)
	  {
	   
	    map<IDTYPE, IDTYPE> newmapping;
	    for(map<IDTYPE, IDTYPE>::iterator dependingEdges = (*newrulesIt).second.begin();
		dependingEdges != (*newrulesIt).second.end();
		dependingEdges++)
	      {
		map<IDTYPE, IDTYPE>::iterator edgeID = phi.first.find((*dependingEdges).first);
		newmapping.insert(pair<IDTYPE,IDTYPE>((*edgeID).second, (*dependingEdges).second));
	      }

	    map<IDTYPE, map<IDTYPE, IDTYPE> >::iterator possibleRuleIt = possibleRules->find((*newrulesIt).first);
	    
	    //checking if the new suggestion was already in the suggested
	    if(possibleRuleIt == possibleRules->end()) {
	      possibleRules->insert(pair<IDTYPE, map<IDTYPE,IDTYPE> >((*newrulesIt).first, newmapping));
	    } else {
	      (*possibleRuleIt).second.insert(newmapping.begin(), newmapping.end());
	    }

	  }
      }

    for(set<IDTYPE>::iterator preEdge = preSet.begin();
	preEdge != preSet.end();
	preEdge++)
      {
	map<IDTYPE, set<IDTYPE> >::iterator edgeIT = removingEdge->find(*preEdge);
	if(edgeIT == removingEdge->end())
	  {
	    set<IDTYPE> removingTransitions;
	    removingTransitions.insert(transition);
	    removingEdge->insert(pair<IDTYPE, set<IDTYPE> >((*preEdge), removingTransitions));
	  } else {
		(*edgeIT).second.insert(transition);
	  }
	  
	
      }
    if(context)
      {
	for(set<IDTYPE>::iterator contextEdge = contextSet.begin();
	    contextEdge != contextSet.end();
	    contextEdge++)
	  {
	    map<IDTYPE, set<IDTYPE> >::iterator contextIt = edgeContext->find(*contextEdge);
	    if(contextIt != edgeContext->end())
	      {
		(*contextIt).second.insert(transition);
	      } else {
		set<IDTYPE> newContext;
		newContext.insert(transition);
		edgeContext->insert(pair<IDTYPE, set<IDTYPE> >(*contextEdge, newContext));
	      }
	  }
      }

    if(directConflict)
      {
	set<IDTYPE> directConflictUnion;
	for(set<IDTYPE>::iterator preEdge = preSet.begin();
	    preEdge != preSet.end();
	    preEdge++)
	  {
	    if(removingEdge->find(*preEdge) != removingEdge->end())
	      {
		//edge in removedEdges
		set<IDTYPE> edgeConflict = (*removingEdge->find(*preEdge)).second;
		if(edgeConflict.find(transition) != edgeConflict.end())
		  {
		    edgeConflict.erase(transition);
		  }
		directConflictUnion.insert(edgeConflict.begin(), edgeConflict.end());
		for(set<IDTYPE>::iterator edgeIt = edgeConflict.begin();
			    edgeIt != edgeConflict.end();
			    edgeIt++)
		  {
		    //if edge is not a context, adding the conflict
		    if(directConflicts->find(*edgeIt) != directConflicts->end())
		      {
			(*directConflicts->find(*edgeIt)).second.insert(transition);
		      } else {
			set<IDTYPE> newSet;
			newSet.insert(transition);
			directConflicts->insert(pair<IDTYPE, set<IDTYPE> >((*edgeIt), newSet));
		      }
		  }
		  

	      }

	    //adding the transition as conflict of the transitions, reading the edge
	    if(context)
	      {
		map<IDTYPE, set<IDTYPE> >::iterator edgeContextIt = edgeContext->find(*preEdge);
		if(edgeContextIt != edgeContext->end())
		  {

		    for(set<IDTYPE>::iterator contextIt = (*edgeContextIt).second.begin();
			contextIt != (*edgeContextIt).second.end();
			contextIt++)
		      {	
			(*directConflicts->find(*contextIt)).second.insert(transition);
		      }
		  }
	      }
	      

	  }
	
	if(context)
	  {
	    for(set<IDTYPE>::iterator contextEdge = contextSet.begin();
		contextEdge != contextSet.end();
		contextEdge++)
	      {
		if(removingEdge->find(*contextEdge) != removingEdge->end())
		  {
		    //edge in removedEdges
		    set<IDTYPE> edgeConflict = (*removingEdge->find(*contextEdge)).second;
		    if(edgeConflict.find(transition) != edgeConflict.end())
		      {
			edgeConflict.erase(transition);
		      }
		    directConflictUnion.insert(edgeConflict.begin(), edgeConflict.end());
		  }
	      }
	  }
	directConflicts->insert(pair<IDTYPE, set<IDTYPE> >(transition, directConflictUnion));
      }




    for(set<IDTYPE>::iterator postEdge = postSet.begin();
	postEdge != postSet.end();
	postEdge++)
      {
	createdByTransition->insert(pair<IDTYPE, IDTYPE>(*postEdge, transition));
      }


    set<IDTYPE> conEdgesIntersection;
    set<IDTYPE> previousTransition;

    bool allOriginEdges = true;
    bool first = true;
    bool firstEdge = true;
    for(set<IDTYPE>::iterator preEdge = preSet.begin();
	preEdge != preSet.end();
	preEdge++)
      {
	if(concurrency)
	  {
	    if(firstEdge)
	      {
		firstEdge = false;
		map<IDTYPE, set<IDTYPE> >::iterator conEIt = concurrentEdges->find(*preEdge);
		if(conEIt != concurrentEdges->end())
		  {
		    conEdgesIntersection.insert((*conEIt).second.begin(), (*conEIt).second.end());
		  } 
	      } else {
		map<IDTYPE, set<IDTYPE> >::iterator conEIt = concurrentEdges->find(*preEdge);
		if(conEIt != concurrentEdges->end())
		  {
		    set<IDTYPE> newIntersection;
		    set<IDTYPE> conEdges = (*conEIt).second;
		    for(set<IDTYPE>::iterator edgesInterIt = conEdgesIntersection.begin();
			edgesInterIt != conEdgesIntersection.end();
			edgesInterIt++)
		      {
			if(conEdges.find(*edgesInterIt) != conEdges.end())
			  {
			    newIntersection.insert(*edgesInterIt);
			  }
		      }
		    conEdgesIntersection = newIntersection;
		  } else {
		    conEdgesIntersection.clear();
		  }
	      }
	  }
	
	if(directConflict || fullConflict || cutoff)
	  {
	    map<IDTYPE, set<IDTYPE> >::iterator edgePrev = necessaryTransitions->find(*preEdge);
	    if(edgePrev != necessaryTransitions->end())
	      {
		previousTransition.insert((*edgePrev).second.begin(), (*edgePrev).second.end());
	      }
	  }
	
      }

    if(fullConflict || directConflict || cutoff)
      { 
	if(context) 
	  {
	    for(set<IDTYPE>::iterator conEdge = contextSet.begin();
		conEdge != contextSet.end();
		conEdge++)
	      {
		map<IDTYPE, set<IDTYPE> >::iterator edgePrev = necessaryTransitions->find(*conEdge);
		if(edgePrev != necessaryTransitions->end())
		  {
		    previousTransition.insert((*edgePrev).second.begin(), (*edgePrev).second.end());
		  }
		
	      }
	  }
	previousTransition.insert(transition);
      }

    for(set<IDTYPE>::iterator postEdge = postSet.begin();
	postEdge != postSet.end();
	postEdge++)
      {
	if(concurrency) 
	  {
	    set<IDTYPE> postConEdges(postSet);
	    postConEdges.erase(*postEdge);
	    postConEdges.insert(conEdgesIntersection.begin(), conEdgesIntersection.end());
	    if(concurrentEdges->find(*postEdge) != concurrentEdges->end())
	      {
		(*concurrentEdges->find(*postEdge)).second.insert(postConEdges.begin(), postConEdges.end());
	      } else {
		concurrentEdges->insert(pair<IDTYPE, set<IDTYPE> >(*postEdge, postConEdges));
	      }
	    for(set<IDTYPE>::iterator cEIt = postConEdges.begin();
		cEIt != postConEdges.end();
		cEIt++)
	      {
		map<IDTYPE, set<IDTYPE> >::iterator reverse = concurrentEdges->find(*cEIt);
		if(reverse != concurrentEdges->end())
		  {
		    (*reverse).second.insert(*postEdge);
		  } else {
		    set<IDTYPE> newSet;
		    newSet.insert(*postEdge);
		    concurrentEdges->insert(pair<IDTYPE, set<IDTYPE> >(*cEIt, newSet));
		  }
	      }
	  }
	if(fullConflict || directConflict || cutoff)
	  {
	    if((!context) || (preSet.count(*postEdge) != 1))
	      necessaryTransitions->insert(pair<IDTYPE, set<IDTYPE> >(*postEdge, previousTransition));
	    //adding necessaryTransitions of the new transition
	  }
      }
    MATCH m;
    m.setMatch(mu_t);
    pg.setMu(transition, m);// zur datenstruktur hinzugefügt
}

}
}
}
#endif
