/***************************************************************************
 *   Copyright (C) 2006 by Michael Mayer   *
 *   mayer@lippe   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef QUICKUNFOLD_CPP
#define QUICKUNFOLD_CPP
#include "quickUnfold.h"
#include "checkEdge.h"
#include "DependenciesCalculator.h"
#include "checkEdge.h"
#include "RedundancyChecker.h"
#include "newUnfold.h"
#include "ConflictCalculator.h"
#include "RedundancyChecker.h"
#include "preConflictCheck.h"
#include "fullConflictChecker.h"
#include "concurrencyChecker.h"
#include "directConflictChecker.h"

namespace scenarios{
namespace quickUnfolding{
bool QuickUnfold::initialize(GTS* graphgrammar, ALGORITHM* writer){
	this->graphgrammar=graphgrammar;
	this->writer=writer;
	HYPERGRAPH startgraph(*(graphgrammar->getStartgraph()));
    // set initial depth 0
    startgraph.initDepth();
    
    // first, construct a Petrigraph with no transitions from startgraph,
    // including the identity mapping iota
    map<IDTYPE,unsigned> initialMarking;
    startgraph_edges = coverability::getKeys(startgraph.getEdges());
    startgraph_vertices = coverability::getKeys(startgraph.getVertices());

    for(set<IDTYPE>::iterator i = startgraph_edges.begin();
        i!=startgraph_edges.end();
        i++)
      {
	initialMarking[*i] = 1;

	if(concurrency)
	  {
	    set<IDTYPE> concurrent(startgraph_edges);
	    concurrent.erase(*i);
	    concurrentEdges.insert(pair<IDTYPE, set<IDTYPE> >(*i, concurrent));
	  }
      }



	PTNET ptnet;//  PTNET ptnet(startgraph_edges,list<PTNetTransition>(),initialMarking);
	for(set<IDTYPE>::iterator iter = startgraph_edges.begin(); iter != startgraph_edges.end(); iter++)
		ptnet.addPlace(*iter,startgraph.getEdges()[*iter]);


// evtl. unpassend!!
	MARKING m;
	for(map<IDTYPE,unsigned>::iterator iter = initialMarking.begin(); iter != initialMarking.end(); iter++){
		TOKEN t;
		vector<TOKEN> vec;
		t.setCount(iter->second);
		vec.push_back(t);
		m.setMarking(iter->first,vec);
	}
	ptnet.setInitialMarking(m);
// bis hierhin!


    map<IDTYPE,IDTYPE> identity_startgraph;
    for(set<IDTYPE>::iterator i = startgraph_edges.begin();
        i!=startgraph_edges.end();
        i++)
      identity_startgraph[*i] = *i;
    for(set<IDTYPE>::iterator i = startgraph_vertices.begin();
        i!=startgraph_vertices.end();
        i++)
      identity_startgraph[*i] = *i;
   /** 
     * Markierung welche Graph-Elemente aus dem Startgraphen stammen (alle Knoten + Kanten aus dem Stargraphen) 
     */
//  PETRIGRAPH P(startgraph,ptnet,graphgrammar,map<IDTYPE,map<IDTYPE,IDTYPE> >(),identity_startgraph);
//identity_startgraph nicht verarbeitet!!!
//GRAPH (HYPERGRAPH)
	result.setVertices(startgraph.getVertices());
	result.setEdges(startgraph.getEdges());
	result.setName(startgraph.getName());
	result.setID(startgraph.getID());
	result.setLastUsedID(startgraph.getLastUsedID());
//HYPERGRAPH
	result.setConnectedVertices(startgraph.getConnectedVertices());
	result.setConnectedEdges(startgraph.getConnectedEdges());
	result.setDepthmapEdges(startgraph.getDepthmapEdges());
	result.setDepthmapVertices(startgraph.getDepthmapVertices());
//GRAPH (PTNET)
	map<IDTYPE, VERTEX> vertices = ptnet.getVertices();
	map<IDTYPE, EDGE> edges = ptnet.getEdges();
	for(map<IDTYPE, VERTEX>::iterator iter = vertices.begin(); iter != vertices.end(); iter++){
		result.addVertexQuick(iter->first,iter->second);
	}
	for(map<IDTYPE, EDGE>::iterator iter = edges.begin(); iter != edges.end(); iter++){
		result.addEdgeQuick(iter->first,iter->second);
	}
	if(result.getLastUsedID() <= ptnet.getLastUsedID())//Auf den größeren wert setzen
		result.setLastUsedID(ptnet.getLastUsedID());
//PTNET
	result.setAssociatedRules(*ptnet.getAssociatedRules());
 	result.setTransitionsQuick(ptnet.getTransitionsQuick());
	result.setPresets(ptnet.getPresets());
	result.setPostsets(ptnet.getPostsets());
	result.setDepth(ptnet.getDepth());
	result.setMaxDepth(ptnet.getMaxDepth());
	result.setGrammar(ptnet.getGrammar());
    //checking the interface
    if(!InterfaceChecker(result).hasValidInterface())
      {
	cerr << "Error in Context!\n";
	return false;
      }
}

void QuickUnfold::CreatedByTransitionInit(PETRIGRAPH pg){
    set<IDTYPE> hgEdges = coverability::getKeys(pg.getEdges());
    for(set<IDTYPE>::iterator hgEdgesIt = hgEdges.begin();
	hgEdgesIt != hgEdges.end();
	hgEdgesIt++)
      {
	createdByTransition.insert(pair<IDTYPE, IDTYPE>((*hgEdgesIt), 0));
      }
}

int QuickUnfold::countAppearances(PETRIGRAPH & pg, string edgelabel){
int count = 0;
    set<IDTYPE> edges = coverability::getKeys(pg.getEdges());
    for(set<IDTYPE>::iterator edgesIt = edges.begin();
	edgesIt != edges.end();
	edgesIt++)
    {
		if(pg.getEdges()[*edgesIt].getLabel() == edgelabel) {
	  	count++;
		}
    }
    return count;
}

pair<IDTYPE, int> QuickUnfold::findMinimumEdge(PETRIGRAPH & pg, RULE rule){
	pair<IDTYPE, int> minimal;
    set<IDTYPE> ruleLeftEdges = coverability::getKeys(rule.getLeft()->getEdges());
    set<IDTYPE>::iterator anedge = ruleLeftEdges.begin();
    string edgelabel =  rule.getLeft()->getEdges()[*anedge].getLabel();
    int number = countAppearances(pg, edgelabel);
    minimal = pair<IDTYPE,int> ((*anedge), number);
    anedge++;
    while(minimal.second > 0 && anedge != ruleLeftEdges.end())
      {
	edgelabel = rule.getLeft()->getEdges()[*anedge].getLabel();
	number = countAppearances(pg, edgelabel);
	if(number < minimal.second) 
	  {  
	    minimal = pair<IDTYPE,int> ((*anedge), number);
	  }
	anedge++;
      }
    
    return minimal;
}

pair<bool, list<pair<map <IDTYPE,IDTYPE>, map<IDTYPE,IDTYPE> > > > QuickUnfold::findMatchingsForRule(PETRIGRAPH& pg, RULE rule){
	//searching the most seldom edge of the left side in the graph
    pair<IDTYPE, int> minimum = findMinimumEdge(pg , rule);

    pair<bool, list<pair<map <IDTYPE,IDTYPE>, map <IDTYPE,IDTYPE> > > > r;
    r.first = false;
    if(minimum.second == 0) {
      //one edge doesn't appear in the graph, no match possible
      return r;
    }

    map<IDTYPE,IDTYPE> phi_e; // mapping: first is the ID of the edge of the rule, second of the graph
    map<IDTYPE,IDTYPE> phi_v; // same with vertices

    pair<map<IDTYPE,IDTYPE>, map<IDTYPE,IDTYPE> >  phi =  pair<map<IDTYPE,IDTYPE>, map<IDTYPE,IDTYPE> >(phi_e, phi_v);

    string minimumedgelabel = rule.getLeft()->getEdges()[minimum.first].getLabel();

    set<IDTYPE> startGraEdges = coverability::getKeys(pg.getEdges());

    //mapping the minimum edge to all edges of the graph with the same labels
    for(set<IDTYPE>::iterator graphedge = startGraEdges.begin();
	graphedge != startGraEdges.end();
	graphedge++)
      {

	string graphedgelabel = pg.getEdges()[*graphedge].getLabel();

	if(graphedgelabel == minimumedgelabel) 
	  {
	    //if graphedge has the same label like the minimum edge, look for a mapping
		coverability::CheckEdge ce(pg, rule, minimum.first, (*graphedge), phi_e, phi_v);
	    if(ce.isCoverable()) 
	      {
		r.first = true;
		r.second.merge(ce.getMapping());
	      }
	  }
      }
    return r;
}

void QuickUnfold::run(PETRIGRAPH * pg){
	
    if(!quiet)
      cout << "Step: 1\n";

    //some time measurements
    double timeDep, timeCreatedInit, timeFirstRound, timeRedundancy = 0;
    double timeFind, timeFold, timeConflicts, timeCalc = 0;

    timeDep = double(clock()) / CLOCKS_PER_SEC;
    dependencies = DependenciesCalculator(graphgrammar).getResult();
    timeDep = double(clock()) / CLOCKS_PER_SEC - timeDep;

    timeCreatedInit = double(clock()) / CLOCKS_PER_SEC;
    CreatedByTransitionInit(result);
    timeCreatedInit = double(clock()) / CLOCKS_PER_SEC - timeCreatedInit;

    if(cutoff)
      {
	//creating the startgraph certificate
	map<string, set<IDTYPE> > newStartGraph;
	//hg = result!!
	for(set<IDTYPE>::iterator startEdge = startgraph_edges.begin();
	    startEdge != startgraph_edges.end();
	    startEdge++)
	  {
	    string label = result.getEdges()[*startEdge].getLabel();
	    if(newStartGraph.find(label) != newStartGraph.end())
	      {
		(*newStartGraph.find(label)).second.insert(*startEdge);
	      } else {
		set<IDTYPE> newSet;
		newSet.insert(*startEdge);
		newStartGraph.insert(pair<string, set<IDTYPE> >(label, newSet));
	      }
	  }
	if(!exCerts)
	  {
	    string certificate = GenerateCertificate(newStartGraph).getCertificate();
	    TransitionCertificate startTransitionCertificate(certificate, newStartGraph, 0);
	    transitionCertificates.insert(pair<IDTYPE, TransitionCertificate>(0, startTransitionCertificate));
	  } else {
	    ExtendedTransitionCertificate extendedStartTransitionCertificate(GenerateExtendedCertificate(result, newStartGraph).getCertificate(), newStartGraph, 0);
	    extendedTransitionCertificates.insert(pair<IDTYPE, ExtendedTransitionCertificate>(0, extendedStartTransitionCertificate));
	  }
      }

    multimap<IDTYPE, list<pair<map <IDTYPE,IDTYPE>, map<IDTYPE,IDTYPE> > > > resultMap;
    
    list<IDTYPE> unfolded;
    list<IDTYPE> transitions;


    timeFirstRound = double(clock()) / CLOCKS_PER_SEC;

    //searching possible matches without coverability check
    for(map<IDTYPE, RULE>::iterator rule = graphgrammar->getRules()->begin();
	rule != graphgrammar->getRules()->end();
	rule++)
      {
	pair<bool, list<pair<map <IDTYPE,IDTYPE>, map<IDTYPE,IDTYPE> > > > oneResult = findMatchingsForRule(result, (*rule).second);

	if(oneResult.first) 
	  {
	    if(resultMap.find((*rule).first) != resultMap.end())
	      {
		(*resultMap.find((*rule).first)).second.insert((*resultMap.find((*rule).first)).second.end(), oneResult.second.begin(), oneResult.second.end());
	      } else {
		resultMap.insert(pair <IDTYPE, list<pair<map <IDTYPE, IDTYPE>, map<IDTYPE, IDTYPE> > > >((*rule).first, oneResult.second));
	      }
	  }
      }
    timeFirstRound = double(clock()) / CLOCKS_PER_SEC - timeCreatedInit;
    timeFind += timeFirstRound;
    possibleRules.clear();

    double tempRed = double(clock()) / CLOCKS_PER_SEC ;
    if(red)
      resultMap = RedundancyChecker(resultMap).getNewResult();
    timeRedundancy += double(clock()) / CLOCKS_PER_SEC - tempRed;



    double timeFirstUnfolding = double(clock()) / CLOCKS_PER_SEC;
    //unfolding without coverability check in the first round
    for(multimap<IDTYPE, list<pair<map <IDTYPE,IDTYPE>, map<IDTYPE,IDTYPE> > > >::iterator resultIt =  resultMap.begin();
	resultIt != resultMap.end();
	resultIt++)
      {
	
	for(list<pair<map <IDTYPE,IDTYPE>, map<IDTYPE,IDTYPE> > >::iterator mappingIt = (*resultIt).second.begin();
	    mappingIt != (*resultIt).second.end();
	    mappingIt++)
	  {
		newUnfold::NewUnfold nuf;
		nuf.initialize(possibleRules,extendedTransitionCertificates,dependencies,createdByTransition,transitionCertificates, concurrentEdges);
		nuf.run(result, (*graphgrammar->getRules()->find((*resultIt).first)).second, (*mappingIt));
	    IDTYPE transID = nuf.getTransition();
	    //if transID equal 0, transition is a cutoff
	    if(transID != 0)
	      {
		//saving results
		transitions.push_back(transID);
		unfolded.push_back((*resultIt).first);
		  }
	  }
	
      }
    timeFirstUnfolding = double(clock()) / CLOCKS_PER_SEC - timeFirstUnfolding;
    timeFold += timeFirstUnfolding;
  
    double timeFirstConflicts = double(clock()) / CLOCKS_PER_SEC;
    //full conflict relation needs update
    if(fullConflict)
      {
	conflictCheck::ConflictCalculator confs;
	confs.initialize(removingEdge,createdByTransition);
	confs.run(result,transitions);
	transitionConflicts = confs.getTransitionConflicts();
//	calculateConflicts(result, transitions);
      }
    timeFirstConflicts = double(clock()) / CLOCKS_PER_SEC - timeFirstConflicts;
    timeConflicts += timeFirstConflicts;

    transitions.clear();

    if(!quiet)
      {
	//result output
	cout << "Unfolding: ";
	for(list<IDTYPE>::iterator unfoldedIt = unfolded.begin();
	    unfoldedIt != unfolded.end();
	    unfoldedIt++)
	  {
	    cout << " " << *unfoldedIt << " (" << (*result.getGrammar()->getRules()->find(*unfoldedIt)).second.getName() << ") ";
	  }
	cout << "\n";
      }
    unfolded.clear();
    //trace output
    if(trace)
      {
	char graphname[256];
	sprintf(graphname,"step1");
	writer->initialize(graphname);
	writer->run(result,graphname);
      }
    //unfolding loop

    for(int i = 2; i <= unfdepth; i++)
      {
	if(!quiet)
	  cout << "Step: " << i << "\n";

	resultMap.clear();

	double tempFind =  double(clock()) ;
	//possibleRules contains mapping of single edges with rules to check
	for(map<IDTYPE, map<IDTYPE, IDTYPE> >::iterator posRulesIt = possibleRules.begin();
	    posRulesIt != possibleRules.end();
	    posRulesIt++)
	  {
	    for(map<IDTYPE, IDTYPE>::iterator mappingsIT = (*posRulesIt).second.begin();
		mappingsIT != (*posRulesIt).second.end();
		mappingsIT++)
	      {
		map<IDTYPE, IDTYPE> phi_e;
		map<IDTYPE, IDTYPE> phi_v;

		//searching total matches
		coverability::CheckEdge ce(result, *(graphgrammar->getRule((*posRulesIt).first)), (*mappingsIT).second, (*mappingsIT).first, phi_e, phi_v);
//		pair<bool, list<pair<map <IDTYPE,IDTYPE>, map<IDTYPE,IDTYPE> > > > mappingResult = checkEdge(result, graphgrammar.getRule((*posRulesIt).first), (*mappingsIT).second, (*mappingsIT).first, phi_e, phi_v);
	
		if(ce.isCoverable()) {

		  if(resultMap.find((*posRulesIt).first) != resultMap.end())
		    {
		      (*resultMap.find((*posRulesIt).first)).second.insert((*resultMap.find((*posRulesIt).first)).second.end(), ce.getMapping().begin(), ce.getMapping().end());
		    } else {
		      resultMap.insert(pair <IDTYPE, list<pair<map <IDTYPE, IDTYPE>, map<IDTYPE, IDTYPE> > > >((*posRulesIt).first, ce.getMapping())); 
		    }
		    
		}
	      }
	  }
	timeFind += (double(clock()) - tempFind) / CLOCKS_PER_SEC;
	possibleRules.clear();

	double tempRed = double(clock());

	//eliminating of matches, which appear more than once
	if(red)
	  resultMap = RedundancyChecker(resultMap).getNewResult();
	timeRedundancy += (double(clock()) - tempRed) / CLOCKS_PER_SEC;

	//unfolding of the matches after the converability check
	for(multimap<IDTYPE, list<pair<map <IDTYPE,IDTYPE>, map<IDTYPE,IDTYPE> > > >::iterator resultIt =  resultMap.begin();
	    resultIt != resultMap.end();
	    resultIt++)
	  {
	    for(list<pair<map <IDTYPE,IDTYPE>, map<IDTYPE,IDTYPE> > >::iterator mappingIt = (*resultIt).second.begin();
		mappingIt != (*resultIt).second.end();
		mappingIt++)
	      {
		double tempChecker =  double(clock());
		//two level of coverability checks
		/**
			AN DIESER STELLE MUSS EINE DATENINTIGRTÄTSPRÜFUNG STATTFINDEN !!!
			evtl. müßen directConflict etc. upgedatet werden
		*/
		conflictCheck::PreConflictCheck pcc;
		pcc.initialize(necessaryTransitions,removingEdge);
		pcc.run((*mappingIt).first);
		if(concurrency || ncov || pcc.isReachable()) //check only necessary for conflict checks
		  {
		    //coverability check
			bool tempConflict = fullConflict;
			if(fullConflict){
				conflictCheck::FullConflictChecker fcc;
				fcc.initialize(necessaryTransitions,createdByTransition,transitionConflicts);
				fcc.run((*mappingIt).first);
				tempConflict = fcc.isCoverable();
			}
			bool tempConcurrency = concurrency;
			if(concurrency){
				ConcurrencyChecker ccc;
				ccc.initialize(concurrentEdges);
				ccc.run((*mappingIt).first);
				tempConcurrency = ccc.isCoverable();
			}
			bool tempDirectConflict = directConflict;
			if(context){
				conflictCheck::DirectConflictChecker dcc;
				dcc.initialize(necessaryTransitions,directConflicts);
				dcc.run(result,(*mappingIt).first);
				tempDirectConflict = dcc.isCoverable();
			}
		    if(ncov || tempConflict || tempConcurrency || tempDirectConflict)
		      {
			timeConflicts =  timeConflicts + (double(clock()) - tempChecker) / CLOCKS_PER_SEC;
			double tempUnfold = double(clock());
			
			//real unfolding
			newUnfold::NewUnfold nuf;
			nuf.initialize(possibleRules,extendedTransitionCertificates,dependencies,createdByTransition,transitionCertificates, concurrentEdges);
			nuf.run(result, (graphgrammar->getRules()->find((*resultIt).first)->second), (*mappingIt));
			IDTYPE transID = nuf.getTransition();
			if(transID != 0)
			  {
			    transitions.push_back(transID);
			    unfolded.push_back((*resultIt).first);
			  }
			timeFold +=  (double(clock()) - tempUnfold) / CLOCKS_PER_SEC;
		      } else {
		      }
		  } else {
		  }
	      }
	    
	  }
	//updating of full conflict relation if necessary
	if(fullConflict)
	  {
	    double tempCalc = double(clock()) / CLOCKS_PER_SEC;
		conflictCheck::ConflictCalculator cc;
		cc.initialize(removingEdge,createdByTransition);
		cc.run(result, transitions);
		transitionConflicts = cc.getTransitionConflicts();
	    timeCalc =  double(clock()) / CLOCKS_PER_SEC - tempCalc;
	  }
	transitions.clear();

	//output	
	if(!quiet)
	  {
	    cout << "Unfolding: ";
	    for(list<IDTYPE>::iterator unfoldedIt = unfolded.begin();
		unfoldedIt != unfolded.end();
		unfoldedIt++)
	      {
		cout << " " << *unfoldedIt << " (" << (*result.getGrammar()->getRules()->find(*unfoldedIt)).second.getName() << ") ";
	      }
	    cout << "\n";
	  }
	    
	unfolded.clear();

	if(trace && (i  < unfdepth))
	  {
	    char graphname[256];
	    sprintf(graphname,"step%i",i);
		writer->initialize(graphname);
		writer->run(result,graphname);
//	    writer.addGraph(P,graphname);
	  }
      }

    *pg = result;

    //some debug information
/*
    if(debug && concurrency)
      {
	for(map<IDTYPE, set<IDTYPE> >::iterator concIt = concurrentEdges.begin();
	    concIt != concurrentEdges.end();
	    concIt++)
	  {
	    cout << "Edge " << (*concIt).first << " hat Concurrent: " << (*concIt).second << "\n";
	  }
      }

    if(debug && (fullConflict || directConflict))
      {
	if(fullConflict)
	  {
	    for(map<IDTYPE, set<IDTYPE> >::iterator confIt = transitionConflicts.begin();
		confIt != transitionConflicts.end();
		confIt++)
	      {
		cout << "Transition " << (*confIt).first << " steht im Konflikt mit: " << (*confIt).second << "\n";
	      }
	  }
	if(directConflict)
	  {
	    for(map<IDTYPE, set<IDTYPE> >::iterator dcIt = directConflicts.begin();
		dcIt != directConflicts.end();
		dcIt++)
	      {
		cout << "Transition " << (*dcIt).first << " steht direkt im Konflikt mit: " << (*dcIt).second << "\n";
	      }
	  }
      }
    if(debug) {
      cout << "Times needed:\n" << "Dependencies: " << timeDep << "\n";
      cout << "Initial Created By: " << timeCreatedInit << "\n";
      cout << "First Round: " << timeFirstRound << "\n";
      cout << "Unfolding first Round: " << timeFirstUnfolding << "\n";
      cout << "First Conflicts: " << timeFirstConflicts << "\n";
      cout << "Find Rules: " << timeFind << "\n";
      cout << "Unfolding: " << timeFold << "\n";
      if(cutoff)
	{
	  cout << "including time of cut: " << timeCut << "\n";
	  cout << cutSuccess << " cuts of " << cutChecks << " checks\n";
	}
      cout << "Redundancy Checks: " << timeRedundancy << "\n";
      cout << "Checking Conflics: " << timeConflicts << "\n";
      cout << "Calculate Conflicts: " << timeCalc << "\n";
      cout << "Inserting Conflicts in Branches: " << insertConflictTime << "\n";
    }
*/
}


}
}
#endif
