/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef ALGORITHMS_REACHABILITY_ANALYSE_LPMODEL_H
#define ALGORITHMS_REACHABILITY_ANALYSE_LPMODEL_H

#include <vector>
#include <string>
#include <map>

#include <globals.h>
#include <lp_lib.h>

using namespace std;

class LPMODEL {
  private:
    lprec *lp;
    map<IDTYPE, int> variable_mapping;
    map<int, IDTYPE> reverse_mapping;

  public:
    enum CONSTRAINT_TYPE {
      CONSTRAINT_LE = ROWTYPE_LE,
      CONSTRAINT_GE = ROWTYPE_GE,
      CONSTRAINT_EQ = ROWTYPE_EQ
    };

    enum RESULT {
      RESULT_OPTIMAL = OPTIMAL,
      RESULT_INFEASIBLE = INFEASIBLE,
      RESULT_ERROR
    };

    double INFINITE;

    //LPMODEL (const vector<string> &variables);
    LPMODEL (const vector<IDTYPE> &variables, int num_rows);
    LPMODEL (const LPMODEL &model);
    ~LPMODEL ();

    bool add_row (const map<IDTYPE, double> &factors, CONSTRAINT_TYPE type, double rhs);
    bool add_row (const map<IDTYPE, int> &factors, CONSTRAINT_TYPE type, double rhs);
    bool set_constraint (const IDTYPE &row, CONSTRAINT_TYPE type);
    bool set_rhs (const IDTYPE &row, double rhs);
    bool set_column (const IDTYPE &transition, const map<IDTYPE, double> &factors);

    void set_int (const vector<IDTYPE> &variables);

    bool set_bounds (const IDTYPE &variable, double lower, double upper);

    void remove_last_row ();
    
    RESULT solve ();

    map<IDTYPE, double> get_variables ();
};

#endif
