/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef ALGORITHM_REDUCE_TRANSITIONS_CPP
#define ALGORITHM_REDUCE_TRANSITIONS_CPP

#include "reduce_transitions.h"

namespace algorithms
{
   namespace petrigraph
   {
      REDUCE_TRANSITIONS::REDUCE_TRANSITIONS()
      {
         this->ID = ALGORITHM_REDUCE_TRANSITIONS;
         this->name = "reduce_transitions";
	 this->petrigraph = NULL;
      }
      REDUCE_TRANSITIONS::~REDUCE_TRANSITIONS() {}
      
      void REDUCE_TRANSITIONS::initialize(PETRIGRAPH* pg)
      {
         this->petrigraph = pg;
      }
      
      void REDUCE_TRANSITIONS::run()
      {
         ALGORITHM* merge = this->load("call_merge");
	 merge->initialize(this->petrigraph);
	 
         bool merged = true;
	 while(merged)
	 {
	    merged = false;
	    map<IDTYPE, TRANSITION> transnet = this->petrigraph->getTransitions();
	    map<IDTYPE, TRANSITION>::iterator transit, forward;
	    for(transit = transnet.begin(); !merged && (transit != transnet.end()); transit++)
	    {
	       /*
	        * rule, preset and union relation for the MERGE algorithm
		*/
	       RULE* r = this->petrigraph->getAssociatedRule(transit->first);
	       vector<TCONNECTION> pre = this->petrigraph->getPreset(transit->first);
	       vector<TCONNECTION>::iterator connectit;
	       map<IDTYPE, unsigned long> preweight;
	       for(connectit = pre.begin(); connectit != pre.end(); connectit++)
	       {
	          if(preweight.count(connectit->ID) == 0)
		  {
		     preweight[connectit->ID] = connectit->weight;
		  }
	          else preweight[connectit->ID] += connectit->weight;;
	       }
	       set< pair<IDTYPE, IDTYPE> > rel;
	       set< pair<IDTYPE, IDTYPE> > emptyunion;
	       
	       forward = transit;
	       for(forward++; forward != transnet.end(); forward++)
	       {
	          /*
		   * rule and preset of the second transition
		   */
		  RULE* rfwd = this->petrigraph->getAssociatedRule(forward->first);
		  vector<TCONNECTION> prefwd = this->petrigraph->getPreset(forward->first);
		  
		  /*
		   * never merge transitions that represent different rules
		   */
		  if(r != rfwd) continue;
		  
		  /*
		   * check whether the presets are equal (i.e. iff balance == 0)
		   * continue with the next transition if they aren't
		   */
		  bool balanced = true;
		  map<IDTYPE, unsigned long> balance = map<IDTYPE, unsigned long>(preweight);
		  map<IDTYPE, unsigned long>::iterator balanceit;
		  for(connectit = prefwd.begin(); connectit != prefwd.end(); connectit++)
		  {
	             if(balance.count(connectit->ID) == 0) balance[connectit->ID] = 0;
		     if(connectit->weight > balance[connectit->ID])
		     {
		        balanced = false;
			break;
		     }
		     else balance[connectit->ID] -= connectit->weight;
		     if(balance[connectit->ID] == 0) balance.erase(connectit->ID);
		  }
		  for(balanceit = balance.begin(); balanced && (balanceit != balance.end()); balanceit++)
		  {
		     if(balanceit->second != 0) balanced = false;
		  }
		  if(!balanced) continue;

		  // check depth
		  for(vector<TCONNECTION>::iterator it1 = pre.begin(); it1 != pre.end(); it1++)
		  {
			if(this->petrigraph->getDepth(it1->ID) < depth)
			{
				balanced = false;
				break;
			}

			list<IDTYPE> lv = this->petrigraph->getConnectedVertices(it1->ID);

			for(list<IDTYPE>::iterator it = lv.begin(); it != lv.end(); it++)
			{
				if(this->petrigraph->getDepth(*it) < depth)
				{
					balanced = false;
					break;
				}
			}

			if(!balanced) break;
		  }
		  if(!balanced) continue;

		  /*
		   * insert the two transitions into the transition union relation
		   */
		  rel.insert(pair<IDTYPE, IDTYPE>(transit->first, forward->first));
		  merged = true;
	       }
	    
	       /*
	        * merge transitions
	        */
	       if(merged)
	       {
	          merge->run(emptyunion, emptyunion, rel);
		  break;
	       }
	    }
	 }
      }     

   };
};
using namespace algorithms;
using namespace algorithms::petrigraph;


#endif
