/***************************************************************************
 *   Copyright (C) 2006 by root   *
 *   root@Vitali   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "refinement_loop.h"
#include "approximated_unfolding.h"
#include "sponge_call.h"
#include "bwra.h"
#include "marking_reader.h"
#include "counter_example_reader.h"
#include "sample.h"
#include "gxlreader.h"
#include "example_checker.h"
#include "equivalence_constructor.h"
#include "abstraction_refinement.h"
#include "algorithm_storage.h"
#include "sample_writer.h"
#include "check_pg_structure.h"

namespace scenarios
{

REFINEMENT_LOOP::REFINEMENT_LOOP()
{
      this->ID = ALGORITHM_REFINEMENT_LOOP;
      this->name = "refinement_loop";

      this->inspec.push_back(pair<string, DATA_FORMAT>("grammar", GTXL));
      this->inspec.push_back(pair<string, DATA_FORMAT>("property", PLAIN_TEXT));

      this->call_descriptor = "Usage: augur -sc=refinement_loop [options] grammar property\n\n\
grammar\tgraph grammar in GTXL format\n\
property\tproperty to verify\n\n";
}


REFINEMENT_LOOP::~REFINEMENT_LOOP()
{
}

void REFINEMENT_LOOP::run()
{
 // gtxl reader
 ALGORITHM* gtxlReader = this->load("call_gtxl_reader");
 gtxlReader->initialize((char*)(this->filemap["grammar"]).c_str());
 GTS gts;
 gtxlReader->run(gts);

 // gxl writer
 ALGORITHM* gxlWriter = this->load("call_gxl_writer");
 gxlWriter->initialize("petrigraph");

 // unfolding
 APPROXIMATED_UNFOLDING au;
 au.initialize(&gts, gxlWriter);
 PETRIGRAPH pg;
 au.run(&pg);

// CHECK_PG_STRUCTURE chps;
// chps.initialize(&pg);
// if(!chps.run())
//  return;

 gxlWriter->initialize("petrigraph_0.xml");
 gxlWriter->run(pg, "petrigraph");
 gxlWriter->initialize("petrigraph");

 SAMPLE_STORAGE ss;
 // set sample storage into SECOND_MATCHHANDLER_REFINEMENT
 storage.setData("second_matchhandler_refinement", NULL);
 storage.setData("second_matchhandler_refinement", &ss);

 // refinement cycle
 bool checked = false;
 unsigned int count = 1;
 while(!checked)
 {
   if(count >= ref_loop_limit)
   {
     cout << "\nloop limit, interrupted\n";
     exit(0);
   }

   // writer
   gxlWriter->run(pg, "petrigraph");


   // encode property
   SPONGE_CALL_SCENARIO scs;
   list<string> ls;
   ls.push_back("petrigraph");
   ls.push_back(this->filemap["property"]);
   ls.push_back(string("marking"));
   scs.initialize(ls);
   scs.run();

   // sleep
   sleep(10);

   // read marking
   MARKING_READER mr;
   mr.initialize("marking");
   list<MARKING> lm = mr.run(0);

   // counter
   char buffer[16];
   sprintf(buffer, "%u", count);

   // check marking
   BWRA bwra;
   bwra.initialize((PTNET*)&pg);
   string t_str("transpath_");
   t_str += string(buffer);
   bool covered = bwra.run(lm, (char*)t_str.c_str());

   if(!covered)
   {
	checked = true;
   }
   else
   {
      COUNTER_EXAMPLE_READER cer;
      cer.initialize((char*)t_str.c_str());
      COUNTER_EXAMPLE ce = cer.run();

      // check the counter-example
      EXAMPLE_CHECKER ech;
      SAMPLE s;
      ech.initialize(&pg, ce.trans_ids);
      int trans_num;
      bool result;
      s = ech.run(&trans_num, &result);

      if(result)
      {
	cout << "The example is real \n";
	return;
      }	
      else
      {
	cout << "The example is spurious \n" << "Transition number " << trans_num << ", id: ";
	cout << ce.trans_ids[trans_num-1] << ", name: " << ce.trans_strings[trans_num-1];
	cout << ",\n" << "could not be fired \n";
      }

      // calculate the equivalence classes
      EQUIVALENCE_CONSTRUCTOR ec;
      ec.initialize(&pg, ce.trans_ids, &s);
      ec.run();

      // add to storage
      ss.m_list_storage.push_back(s);

      // write sample
      string s_str("sample_");
      s_str += string(buffer);
      s_str += string(".xml");
      SAMPLE_WRITER sw;
      sw.initialize((char*)s_str.c_str());
      sw.run(ss);

      PETRIGRAPH pg_ref;
      ABSTRACTION_REFINEMENT refiner;
      refiner.initialize(&gts, gxlWriter);
      refiner.run(&pg_ref);
      pg = pg_ref;

      string s_pg("petrigraph_");
      s_pg+=string(buffer);
      s_pg+=string(".xml");
      gxlWriter->initialize((char*)s_pg.c_str());
      gxlWriter->run(pg, "petrigraph");
      gxlWriter->initialize("petrigraph");
   }

   count++;
 }
}

}
