/***************************************************************************
 *   Copyright (C) 2006 by root   *
 *   root@Vitali   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>

#include "regexpr2marking.h"
#include "regexpreader.h"
#include "cross_product.h"
#include "gxlreader.h"
#include "hypergraph.h"
#include "semilin/automaton_2_grammar.h"
#include "semilin/commkleene_solve.h"
#include "semilin_caller.h"

namespace scenarios
{

REGEXPR_2_MARKING::REGEXPR_2_MARKING()
{
      this->ID = ALGORITHM_REGEXPR2MARKING;
      this->name = "regexpr2marking";

      this->inspec.push_back(pair<string, DATA_FORMAT>("hypergraph", GXL));
      this->inspec.push_back(pair<string, DATA_FORMAT>("regular_expression", PLAIN_TEXT));	
      this->outspec.push_back(pair<string, DATA_FORMAT>("marking", PLAIN_TEXT));

      this->call_descriptor = "Usage: augur -sc=reg2marking hypergraph regular_expression marking\n\n\
hypergraph\thypergraph in GXL format\n\
regular_expression\ttested property as regular expression\n\
marking\tmarking corresponding to regular expression\n\n";
}


REGEXPR_2_MARKING::~REGEXPR_2_MARKING()
{
}


void REGEXPR_2_MARKING::run()
{
	// read hypergraph
	GXLREADER gxlreader;
	gxlreader.initialize((char*)(this->filemap["hypergraph"]).c_str());
	PETRIGRAPH pg;
	GTS gts;
	gxlreader.run(pg, gts);

	// read regular expression
	REGEXPREADER regexpr_reader;
	regexpr_reader.initialize(this->filemap["regular_expression"].c_str());
	FINITE_AUTOMATON* fa;
	regexpr_reader.run(&fa);

	HYPERGRAPH hg;

	hg = (HYPERGRAPH)(pg);

	// write all temporary files in the work directory
	string workDir;
	char * pch;
	char filename_copy[strlen(this->filemap["hypergraph"].c_str())+1];
	strcpy(filename_copy, this->filemap["hypergraph"].c_str());
	pch = strtok (filename_copy,"/");
	
	while (1)
	{
		workDir += string("/");
		string tmp = (string)pch;
		pch = strtok (NULL, "/");
		
		if(pch==NULL)
		 break;	
		else
		 workDir += tmp;
	}

	string automFile = workDir + string("automation");

	fa->write((char*)automFile.c_str());
	cout << "finite automation is written to " << automFile.c_str() << endl;

	// cross product
	CROSS_PRODUCT cp;
	FINITE_AUTOMATON cross_pr;
	cp.initialize(fa, &hg);
	cp.run(&cross_pr);

	string crossFile = workDir + string("cross_product");

	cross_pr.write((char*)crossFile.c_str());
	cout << "cross product is written to " << crossFile.c_str() << endl;

	// grammar
	string grammarFile = workDir + string("grammar");
	AUTOMATON_2_GRAMMAR a2g;
	vector<CFEXPRESSION*> le = a2g.run(&cross_pr);
	
	for(vector<CFEXPRESSION*>::iterator itl = le.begin(); itl != le.end(); itl++)
	 (*itl)->write((char*)grammarFile.c_str());

	cout << "grammar is written to " << grammarFile.c_str() << endl;

	// variables and constants
	map<IDTYPE, string> variables = a2g.getVariables();
	map<string, char> constants = a2g.getConstants();

	string symbolsFile = workDir + string("symbols");

	ofstream ofs(symbolsFile.c_str());

	ofs << "Variables (State -> Var):" << endl << endl;

	for(map<IDTYPE, string>::iterator it1 = variables.begin(); it1 != variables.end(); it1++)
		ofs << it1->first << " -> " << it1->second << endl;

	ofs << endl << "Constants (Label -> Symbol):" << endl << endl;

	for(map<string, char>::iterator it2 = constants.begin(); it2 != constants.end(); it2++)
		ofs << it2->first << " -> " << it2->second << endl;

	cout << "Symbol maps are written to " << symbolsFile.c_str() << endl;

	// solver
	string solvedFile = workDir + string("solved");
	COMMKLEENE_SOLVE solver = COMMKLEENE_SOLVE();
	CFEXPRESSION exp = solver.run(le);
	exp.write((char*)solvedFile.c_str());

	cout << "Kleene solver result is written to " << solvedFile.c_str() << endl;

	// the last step is set to a separate class
	// because of some problems with template compilation
	SEMILIN_CALLER scall;
	scall.setConstants(constants);
	scall.initialize((char*)(this->filemap["marking"]).c_str());
	scall.run(exp);
}

}


