%{
/*****************************************************************************
 * reglex.ll: lex file for regular expressions
 *****************************************************************************
 * Authors: Salil Joshi <hellfeuer@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/


#include <stdio.h>
#include "../finite_automaton.h"
#include "regparse.tab.hh"

#undef YY_INPUT
char *re_strin;		/* current pointer on input string */

#define YY_INPUT(buf, result, max_size) \
	strncpy(buf, re_strin, max_size-1);\
	re_strin += result = strlen(buf);

%}

delim	[ \t\n]
white	{delim}+
char	[A-Za-z_]
digit	[0-9]
alphanum	{char}|{digit}
star	\*
plus	\+
lb	\(
rb	\)
eps	~|\"\"

%%
{white}	;/* eat up whitespace */
{alphanum}+	{
		re_lval.str = new string(re_text);
		return ALPHANUM;
		}
{star}	{return STAR;}
{plus}	{return PLUS;}
{lb}	{return LB;}
{rb}	{return RB;}
{eps}	{
		re_lval.str = new string('\0');
		return ALPHANUM;
	}

%%

