%{
/*****************************************************************************
 * regparse.yy: yacc grammar for regular expressions
 *****************************************************************************
 *
 * Authors: Salil Joshi <hellfeuer@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/
#include <stdio.h>
#include <algorithm>
#include <iostream>

#include "../finite_automaton.h"

extern "C"
{
	int re_parse(void);
	int re_wrap()
	{
		return 1;
	}
	void re_error(const char* str)
	{
		fprintf(stderr,"error: %s\n",str);
	}

	using namespace std;
}
int re_lex(void);
FINITE_AUTOMATON * finalFA;
%}

%token EPSILON ALPHANUM STAR PLUS LB RB

%union {
string* str;
FINITE_AUTOMATON * fa;
}

%type <fa> expr;
%type <fa> term;
%type <fa> factor;
%type <str> ALPHANUM;

%%
final	: expr
	{
		finalFA = $1;
	}
	;
expr	: expr PLUS term
	{
		$$ = new FINITE_AUTOMATON();
		//combine all states
		set_union($1->allStates.begin(), $1->allStates.end(), $3->allStates.begin(), $3->allStates.end(), inserter($$->allStates, $$->allStates.begin()));
		
		
		if ((*($1->initialStates.begin()))->id == (*($1->finalStates.begin()))->id) {
			//combine all transitions
			set_union($1->transitions.begin(), $1->transitions.end(), $3->transitions.begin(), $3->transitions.end(), inserter($$->transitions, $$->transitions.begin()));
			if ((*($3->initialStates.begin()))->id == (*($3->finalStates.begin()))->id) {
				STATE* ns1 = $$->addState();
				STATE* ns2 = $$->addState();
				$$->addTransition(ns1, '\0', *($1->initialStates.begin()));
				$$->addTransition(ns1, '\0', *($3->initialStates.begin()));
				$$->addTransition(*($1->finalStates.begin()), '\0', ns2);
				$$->addTransition(*($3->finalStates.begin()), '\0', ns2);
				$$->initialStates.insert(ns1);
				$$->finalStates.insert(ns2);
			} else {
				$$->addTransition(*($3->initialStates.begin()), '\0', *($1->initialStates.begin()));
				$$->addTransition(*($1->finalStates.begin()), '\0', *($3->finalStates.begin()));
				$$->initialStates.insert(*($3->initialStates.begin()));
				$$->finalStates.insert(*($3->finalStates.begin()));
			}
		} else if ((*($3->initialStates.begin()))->id == (*($3->finalStates.begin()))->id) {
			//combine all transitions
			set_union($1->transitions.begin(), $1->transitions.end(), $3->transitions.begin(), $3->transitions.end(), inserter($$->transitions, $$->transitions.begin()));
			
			$$->addTransition(*($1->initialStates.begin()), '\0', *($3->initialStates.begin()));
			$$->addTransition(*($3->finalStates.begin()), '\0', *($1->finalStates.begin()));
			$$->initialStates.insert(*($1->initialStates.begin()));
			$$->finalStates.insert(*($1->finalStates.begin()));
		} else {
			map<pair<STATE*, string>, set<STATE*> >::iterator j;
			for (j = $1->transitions.begin(); j != $1->transitions.end(); j++) {
				$$->transitions.insert(*j);
			}
			for (j = $3->transitions.begin(); j != $3->transitions.end(); j++) {
				pair<pair<STATE*, string>, set<STATE*> > tr;
				set<STATE*> to = (*j).second;
				
				set<STATE*>::iterator f = to.find(*($3->finalStates.begin()));
				if (f != to.end()) {
					to.erase(f);
					to.insert(*($1->finalStates.begin()));
					tr = pair<pair<STATE*, string>, set<STATE*> >((*j).first, to);
				} else {
					tr = *j;
				}
			
				if (((*j).first.first)->id == (*($3->initialStates.begin()))->id) {
					STATE* from = *($1->initialStates.begin());
					string a = (*j).first.second;
					
					$$->transitions[pair<STATE*, string>(from,a)].insert<set<STATE*>::iterator>(to.begin(), to.end());
				} else {
					
					$$->transitions.insert(tr);
				}
			}
			$$->allStates.erase(*($3->initialStates.begin()));
			$$->allStates.erase(*($3->finalStates.begin()));
			$$->initialStates.insert(*($1->initialStates.begin()));
			$$->finalStates.insert(*($1->finalStates.begin()));
			delete *($3->initialStates.begin());
			delete *($3->finalStates.begin());
		}
		delete $3;
		delete $1;
	}
	| term
	{
	}
	;
term	: term factor
	{
		$$ = new FINITE_AUTOMATON();
		
		//combine all states
		set_union($1->allStates.begin(), $1->allStates.end(), $2->allStates.begin(), $2->allStates.end(), inserter($$->allStates, $$->allStates.begin()));
		
		//combine transitions
		map<pair<STATE*, string>, set<STATE*> >::iterator j;
		for (j = $1->transitions.begin(); j != $1->transitions.end(); j++) {
			$$->transitions.insert(*j);
		}
		
		for (j = $2->transitions.begin(); j != $2->transitions.end(); j++) {
			if (((*j).first.first)->id == (*($2->initialStates.begin()))->id) {
				STATE* from = *($1->finalStates.begin());
				set<STATE*> to = (*j).second;
				string a = (*j).first.second;
				if ((*(to.begin()))->id == (*($2->initialStates.begin()))->id) {
					$$->transitions[pair<STATE*, string>(from,a)].insert<set<STATE*>::iterator>($1->finalStates.begin(), $1->finalStates.end());
				} else {
					$$->transitions[pair<STATE*, string>(from,a)].insert<set<STATE*>::iterator>(to.begin(), to.end());
				}
			} else if ((*((*j).second.begin()))->id == (*($2->initialStates.begin()))->id) {
				STATE* from = j->first.first;
				string a = j->first.second;
				$$->transitions[pair<STATE*, string>(from,a)].insert<set<STATE*>::iterator>($1->finalStates.begin(), $1->finalStates.end());
			} else {
				$$->transitions.insert(*j);
			}
		}
		$$->initialStates.insert(*($1->initialStates.begin()));
		
		//$$->transitions[pair<STATE*, string>(*($1->finalStates.begin()),'\0')].insert<set<STATE*>::iterator>($2->initialStates.begin(), $2->initialStates.end());
		
		STATE* tobedeleted = (*($2->initialStates.begin()));
		if (tobedeleted != *($2->finalStates.begin())) {
			$$->finalStates.insert(*($2->finalStates.begin()));
		} else {
			$$->finalStates.insert(*($1->finalStates.begin()));
		}
		$$->allStates.erase(tobedeleted);
		delete tobedeleted;
		
		delete $1;
		delete $2;
	}
	| factor
	{
	}
	;
factor	: LB expr RB
	{
		$$ = $2;
	}
	| factor STAR
	{
		$$ = new FINITE_AUTOMATON;
		$$->allStates = set<STATE*>($1->allStates);
		STATE* finalS = *($1->finalStates.begin());
		STATE* initialS = *($1->initialStates.begin());
		
		map<pair<STATE*, string>, set<STATE*> >::iterator j;
		
		for (j = $1->transitions.begin(); j != $1->transitions.end(); j++) {
			if (((*j).first.first)->id == initialS->id) {
				STATE* from = finalS;
				set<STATE*> to = (*j).second;
				string a = (*j).first.second;
				$$->addTransition(from, a, to);	
			} else {
				$$->transitions.insert(*j);
			}
		}
		$$->initialStates.insert(finalS);
		$$->finalStates.insert(finalS);
		$$->allStates.erase(initialS);
		delete initialS;
		delete $1;
		
	}
	| ALPHANUM
	{
		$$ = new FINITE_AUTOMATON();
		STATE* s1 = $$->addState();
		STATE* s2 = $$->addState();
		$$->addTransition(s1, string(*re_lval.str), s2);
		$$->finalStates.insert(s2);
		$$->initialStates.insert(s1);
		
	}
	| EPSILON
	{
	}
	;

%%

