/***************************************************************************                                                                       *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>

#include "rules_converter.h"
#include "GTXLReader.h"
#include "vis_hypergraph.h"
#include "rule.h"

namespace converters
{

RULES_CONVERTER::RULES_CONVERTER()
{
	this->inspec.push_back(pair<string, DATA_FORMAT>("gtxl-file", GTXL));
      this->outspec.push_back(pair<string, DATA_FORMAT>("ps_rules", PS));

      this->call_descriptor = "Usage: augur -sc=rules2ps [-size=n] gtxl-file ps_rules \n\n\
gtxl-file\tgraph grammar in GTXL format\n\
ps_rules\tograph grammar visualized in PS format\n\n";

this->name = "rules_converter";

}


RULES_CONVERTER::~RULES_CONVERTER()
{
}

void RULES_CONVERTER::run()
{	
	ALGORITHM* reader = this->load("call_gtxl_reader");
	reader->initialize((char*)this->filemap["gtxl-file"].c_str());
	GTS gts;
	reader->run(gts);

	string output;
	char * filename = new char(filemap["ps_rules"].size()+1);
	char * pch;
	strcpy(filename, (char*)this->filemap["ps_rules"].c_str());
	pch = strtok (filename,".");
			
	while (pch!=NULL && strcmp(pch, "ps")!=0)
  	{
		output += (string)pch;
    		pch = strtok (NULL, ".");		
  	}
	
	system((char*)string(string("mkdir ") + output + string("_files")).c_str());
	convert(gts, output+string("_files"));

	makeLaTex(output, gts, rules_size);

	callLatex((char*)string(output + string(".tex")).c_str(), (char*)string(output + string(".ps")).c_str());
}

void RULES_CONVERTER::makeLaTex(string output, GTS & ggr, float size)
{
	 string outputtex = output + string(".tex");
	 ofstream of(outputtex.c_str());

         char out[output.size()+1]; 
         strcpy(out, output.c_str());

         char * pch;
         pch = strtok(out,"/");
         while(pch!=NULL)
         {
             output = (string)pch;
             pch = strtok(NULL, "/");
         }

	// write head
	of << "\\documentclass{article}\n \\newcommand{\\epsff}[1]{\\scalebox{";
	of << size;
	of << "}{\\epsffile{#1}}} \n \\usepackage{epsf, graphicx} \n \\begin{document} \n ";

	// initial graph
        of << "Initial graph:\\qquad\n";
	of << "\\begin{tabular}[c]{r} \n \\epsff{" << output << "_files/ini.ps} \n \\end{tabular} \\\\ \n";

	string sfl = (string)output + "_files/";
	string sln = "l.ps";
	string srn = "r.ps";

	// rules
	map<IDTYPE, RULE>* rules = ggr.getRules();

	map<IDTYPE, RULE>::iterator it = rules->begin();

	// rules
	for(int i=1; i<=ggr.getRules()->size(); i++)
	{
		char buf[16];
		sprintf(buf, "%d", i);
		string sb(buf);

		string sl = sfl + sb + sln;
		string sr = sfl + sb + srn;

		// name of rule
		string s = it->second.getName();
		
		of << "\\bigskip\\hrule\\bigskip\n";

                of << s.c_str() << ": \\\\\n";

		// left
		of << "\\begin{tabular}{rcl} \n \\begin{tabular}[c]{r} \n \\epsff{" << sl.c_str() << "} \n \\end{tabular} & \n";

		// guard
		if(coloured && it->second.getGuard().size()>0)
		{
			of << "\\begin{tabular}[c]{c} \n";
			list<EXPRESSION> le = it->second.getGuard();
			for(list<EXPRESSION>::iterator itl = le.begin(); itl != le.end(); itl++)
			{
				if(itl->getExpr().size() <= 15)
					of << "$" << itl->getExpr() << "$\\\\\n";
				else
					of << "\\parbox{5cm}{\\raggedright $" << itl->getExpr() << "$}\\\\\n";
			}
			
			// arrow
			of << "$\\Longrightarrow$ \n";
			
			of << "\\end{tabular} & \n";
		}
		else
		{
			// arrow
			of << "$\\Longrightarrow$ & \n";
		}

		// right
		of << "\\begin{tabular}[c]{l} \n \\epsff{" << sr.c_str() << "} \n \\end{tabular} \n \\end{tabular} \\\\ \n";

		it++;
	}

	// write bottom
	of << "\\end{document} \n";
}

/** convert initial graph and rules to dot format */
void RULES_CONVERTER::convert(GTS & ggr, string output)
{
	// initial graph
	HYPERGRAPH * initial_graph = ggr.getStartgraph();

	string inis = output + (string)"/ini.ps";
	
	map<int, int> iniMap;
	
	VIS_HYPERGRAPH vh;
	vh.initialize((char*)inis.c_str());
	PRINT_TYPE pt = FULL;
	vh.run(*initial_graph, pt);

	// rules
	map<IDTYPE,RULE> *rules = ggr.getRules();

	int counter = 1;
	string sf;

	string sel = "l.ps";
	string ser = "r.ps";
	string sfl = output + "/";
		
	for(map<IDTYPE,RULE>::iterator itR = (*rules).begin(); itR!=(*rules).end(); itR++)
	{
		char buf[256];
		sprintf(buf, "%d", counter);
		string sb(buf);

		string sl = sfl + sb + sel;
		string sr = sfl + sb + ser;

		// mapping
		map<int, int> mapleft;
		map<int, int> mapright;

		HYPERGRAPH * hgl = itR->second.getLeft();
		HYPERGRAPH * hgr = itR->second.getRight();

		map<IDTYPE, VERTEX> verticesleft = hgl->getVertices();
		map<IDTYPE, VERTEX> verticesright = hgr->getVertices();

		MATCH interface = itR->second.getInterface();

		int count = 1;

		// calculate mapping
		for(map<IDTYPE, VERTEX>::iterator itV = verticesleft.begin(); itV!=verticesleft.end(); itV++)
		{
			if(interface.hasMatch(itV->first))
			{
				mapleft[itV->first] = count;
				mapright[interface.getMatch(itV->first)] = count;
			}
			else
			{
				mapleft[itV->first] = count;
				mapright[-interface.getMatch(itV->first)] = count;
			}

			count++;
		}

	// visualize
	PRINT_TYPE pt = FULL;

	vh.initialize((char*)sl.c_str());
	vh.setNodeLabel(mapleft);
	vh.run(*hgl, pt);
	
	vh.initialize((char*)sr.c_str());
	vh.setNodeLabel(mapright);
	vh.run(*hgr, pt);

	counter++;
	}

}

}
