/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef SAMPLE_H
#define SAMPLE_H

#include "globals.h"
#include "hypergraph.h"
#include "match.h"
#include "persistable.h"
#include "rule.h"

namespace algorithms {

namespace refinement {

namespace data_str_refinement {

// single real graph in sample
struct HYPERGRAPH_INFO
{
  HYPERGRAPH_INFO(HYPERGRAPH hg, unsigned int id, unsigned int link)
  {
	 this->hg = hg;
	 this->id = id;
	 this->link = link;
  }
	  
  HYPERGRAPH_INFO()
  {
  }
			  
  HYPERGRAPH hg;
  // id in sample
  unsigned int id;
  // linked id
  unsigned int link;
  // match from L to graph in Sample
  list<MATCH> matchesLG;
  // match from real graph to graph(m)
  // is initialized only for last not empty set [M_k]
  list<MATCH> matchesGgM;
  // labels for the equivalence classes
  // id_node -> list(labels)
  map<IDTYPE, list<unsigned> > eq_labels;
  
};

// sample consists of step-levels with s list of hypergraph in each level
struct SAMPLE{

  SAMPLE(vector<list<HYPERGRAPH_INFO> > & tree,  vector<string> & trans )
  {
	this->tree = tree;
	this->trans = trans;
  }
  
  SAMPLE()
  {
  }
  
   // one sample is a tree like structure
   vector<list<HYPERGRAPH_INFO> > tree;
   // counter-example, names of transitions
   vector<string> trans;

};

// all samples from the session
class SAMPLE_STORAGE: public DATA
{

public:

  list<SAMPLE> m_list_storage;

};

struct NODE_INFO
{
	IDTYPE id;
	unsigned int level;
	unsigned int graph_id;
	
	friend bool operator<(const NODE_INFO & n1, const NODE_INFO & n2);
};

struct CHECKED_FOLDINGS
{
	CHECKED_FOLDINGS(RULE * r, MATCH * match_first, MATCH * match_second)
	{
		this->rid = r->getID();
		this->match_first = *match_first;
		this->match_second = *match_second;
	}
		 
	IDTYPE rid;
	MATCH match_first;
	MATCH match_second;
   };
};

};

};

using namespace algorithms;
using namespace refinement;
using namespace data_str_refinement;

#endif
