/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>

#include "sample_reader.h"

namespace algorithms {

namespace readers {

SAMPLE_READER::SAMPLE_READER()
{
}


SAMPLE_READER::~SAMPLE_READER()
{
}

void SAMPLE_READER::initialize(char * filename)
{
 this->filename = (string)filename;
}

SAMPLE_STORAGE SAMPLE_READER::run()
{
try
{       
	SAMPLE_STORAGE sample_storage;
	 
	ifstream file;
	file.open(filename.c_str());
	if(file.fail())
	{
	 return sample_storage;
	} 
	
	xmlDocPtr doc = xmlParseFile(filename.c_str());
	if (doc == NULL )
	{
		fprintf(stderr,"Document not parsed successfully. \n");
		xmlFreeDoc(doc);
		exit(1);
	}	
	
	xmlNodePtr storage = xmlDocGetRootElement(doc);
	if (storage == NULL)
	{
		cout << "Empty xml document. \n";
		exit(1);
	}
	
		
	if (xmlStrcasecmp(storage->name, (const xmlChar*) "m_list_storage"))
	{
		cout << "Root node 'm_list_storage' expected. \n";
		exit(1);
	}
	

	xmlNodePtr sample_in_list = GetChildrenNode(storage,(char*)"sample");
	
	sample_storage.m_list_storage.clear();
	SAMPLE sample;
	sample.tree.clear();
	sample.trans.clear();
	
	while(sample_in_list != NULL)
	{	
		ReadSample(sample_in_list, sample);
		sample_storage.m_list_storage.push_back(sample);
		sample.tree.clear();
		sample.trans.clear();
		sample_in_list = GetNextChild(sample_in_list, (char*)"sample");
	}
	
	file.close();
	
	return sample_storage;
}
catch(exception e)
{
 cout << "Error while reading the sample file\n";
 exit(1);
}

}


void SAMPLE_READER::ReadSample(xmlNodePtr list_sample, SAMPLE &sample)
{
try
{
	xmlNodePtr sample_tree = GetChildrenNode(list_sample, (char*)"tree");
	
	if(sample_tree != NULL)
	{
		Read_m_list(sample_tree, sample.tree);
	}
	
	xmlNodePtr sample_trans = NULL;
	sample_trans = GetChildrenNode(list_sample, (char*)"trans");
	xmlNodePtr cur = NULL;
	
	if(sample_trans != NULL) 
	{
		string str;
		xmlNodePtr string_in_trans = GetChildrenNode(sample_trans, (char*)"string_in_vector");
	
		while(string_in_trans != NULL)
		{
			cur = string_in_trans->xmlChildrenNode;
			if(cur != NULL)
			{
				str = (char*)cur->content;
				sample.trans.push_back(str);
				cur = NULL;
			}
			
			string_in_trans = GetNextChild(string_in_trans, (char*) "string_in_vector");
		}
	}
	
}
catch(exception e)
{
 cout << "Error while reading the sample file\n";
 exit(1);
}

}

xmlNodePtr SAMPLE_READER::GetChildrenNode(xmlNodePtr parent, char* name_child)
{
	xmlNodePtr child = parent->xmlChildrenNode;
	while((child != NULL) && (xmlStrcasecmp(child->name, (const xmlChar*)name_child)))
	{
		child = child->next;
	}
	
	return child;
	
		
}

xmlNodePtr SAMPLE_READER::GetNextChild(xmlNodePtr child1, char* name_child)
{
	
	xmlNodePtr child2 = child1->next;
	while((child2 != NULL) && (xmlStrcasecmp(child2->name, (const xmlChar*)name_child)))
	{
		child2 = child2->next;
	}
	
	return child2;
	
	
}

void SAMPLE_READER::ReadHypergraphInfo(xmlNodePtr HyperInfo, HYPERGRAPH_INFO &hgInfo)
{
try
{
	string str;
	
	str = (char *)xmlGetProp(HyperInfo, (const xmlChar*) "id");
	hgInfo.id = atoi(str.c_str());

	str = (char *)xmlGetProp(HyperInfo, (const xmlChar*) "link");
	hgInfo.link = atoi(str.c_str());
	
	// xmlNodePtr HypGraph = NULL;
	// HypGraph = GetChildrenNode(HyperInfo, (char*)"Hypergraph");
	// if (HypGraph != NULL)
	{
		set<string> ids;
		IDTYPE last_ID=0;
		hgInfo.hg = ReadHypergraph(HyperInfo, ids, last_ID);
	}
	
	/*
	xmlNodePtr MatchesLG_XmlPointer = NULL;
	MatchesLG_XmlPointer = GetChildrenNode(HyperInfo, (char*)"matchesLG");
	
	
	//if (xmlStrcasecmp(MatchesLG_XmlPointer->name, (const xmlChar*) "matchesLG"))
	if(MatchesLG_XmlPointer == NULL)
	{
		cout << "Error while reading the sample file - matchesLG expected.\n";
		exit(1);
	}
	
	typedef pair<IDTYPE, IDTYPE> matches_type;
	map<IDTYPE, IDTYPE> map_el;
	map_el.clear();
	IDTYPE id1, id2;
	
	xmlNodePtr list_element = NULL;
	list_element = GetChildrenNode(MatchesLG_XmlPointer, (char*)"map");
	while(list_element != NULL)
	{
		xmlNodePtr map_element = NULL;
		map_element = GetChildrenNode(list_element,(char*)"map_element");
		while(map_element !=NULL)
		{
			str = (char *)xmlGetProp(map_element, (const xmlChar*) "id1");
			id1 = atoi(str.c_str());
			str = (char *)xmlGetProp(map_element, (const xmlChar*) "id2");
			id2 = atoi(str.c_str());
			map_el.insert(matches_type(id1,id2));
			map_element = GetNextChild(map_element, (char*)"map_element");
		}
		
		MATCH m_el;
		m_el.setMatch(map_el);
		hgInfo.matchesLG.push_back(m_el);
		map_el.clear();
		list_element = GetNextChild(list_element,(char*)"map");
	}
	*/
	
	/////////// labels //////////////
	
	xmlNodePtr MatchesLabelPointer = NULL;
	MatchesLabelPointer = GetChildrenNode(HyperInfo, (char*)"node_labels");
	
	
	if(MatchesLabelPointer == NULL)
	{
		cout << "Error while reading the sample file - node_labels expected\n";
      		exit(1);
	}
		
	typedef pair<IDTYPE, IDTYPE> matches_type;

	IDTYPE id1, id2;
	
	xmlNodePtr map_element = NULL;
	map_element = GetChildrenNode(MatchesLabelPointer,(char*)"map_element");
	while(map_element !=NULL)
	{
		str = (char *)xmlGetProp(map_element, (const xmlChar*) "id1");
		id1 = atoi(str.c_str());
		str = (char *)xmlGetProp(map_element, (const xmlChar*) "id2");
		
		list<unsigned> eq_list;
				
		char * ch;
		ch = strtok((char*)str.c_str(), ";");

		while(ch!=NULL)
		{
			eq_list.push_back(atoi(ch));
			ch = strtok(NULL, ";");
		}
		
		hgInfo.eq_labels.insert(pair<IDTYPE, list<unsigned> >(id1,eq_list));
		map_element = GetNextChild(map_element, (char*)"map_element");
	}	
	
	//////////////////////////////////////
	
	/*
	xmlNodePtr MatchesGgM_XmlPointer = NULL;
	MatchesGgM_XmlPointer = GetChildrenNode(HyperInfo, (char*)"matchesGgM");
	
	//if (xmlStrcasecmp(cur->name, (const xmlChar*) "matchesGgM"))
	if(MatchesGgM_XmlPointer == NULL)
	{
		cout << "Error while reading the sample file - matchesGgM expected\n";
      		exit(1);
	}
	
	list_element = NULL;	
	list_element = GetChildrenNode(MatchesGgM_XmlPointer,(char*)"map");
	map_el.clear();
	while(list_element != NULL)
	{
		xmlNodePtr map_element = NULL;
		map_element = GetChildrenNode(list_element,(char*)"map_element");
		while(map_element != NULL)
		{
			str = (char *)xmlGetProp(map_element, (const xmlChar*) "id1");
			id1 = atoi(str.c_str());
			str = (char *)xmlGetProp(map_element, (const xmlChar*) "id2");
			id2 = atoi(str.c_str());
			map_el.insert(matches_type(id1,id2));
			map_element = GetNextChild(map_element, (char*)"map_element");
		}
		MATCH m_el;
		m_el.setMatch(map_el);
		hgInfo.matchesGgM.push_back(m_el);
		map_el.clear();
		list_element = GetNextChild(list_element, (char*)"map");
	}
	*/
}
catch(exception e)
{ 
 cout << "Error while reading the sample file\n";
 exit(1);
}

}

void SAMPLE_READER::Read_m_list(xmlNodePtr m_list_xml, vector<list<HYPERGRAPH_INFO> > &m_list1)
{ 
try
{
	xmlNodePtr list_el = NULL;	
	xmlNodePtr vector_el = NULL;
	vector_el = GetChildrenNode(m_list_xml, (char*)"vector");
	
	if (vector_el != NULL)
	{
		vector_el = GetChildrenNode(vector_el, (char*)"list");

		if (xmlStrcasecmp(vector_el->name, (const xmlChar*) "list"))
		{
			cout << "Error while reading the sample file - Root node 'list' expected.\n";
			exit(1);
		}
	
		list<HYPERGRAPH_INFO> hgInfo_list;
		HYPERGRAPH_INFO hygInfo;
	
		while(vector_el != NULL)
		{	
			list_el = GetChildrenNode(vector_el,(char*)"HypergraphInfo");
			while(list_el != NULL)
			{
				ReadHypergraphInfo(list_el, hygInfo);
				hgInfo_list.push_back(hygInfo);
				list_el = GetNextChild(list_el, (char*)"HypergraphInfo");
			}	
			m_list1.push_back(hgInfo_list);
			vector_el = GetNextChild(vector_el, (char*)"list");
		}
	
	}
}
catch(exception e)
{	
	cout << "Error while reading HypergraphInfo from XML file " << endl; 
	exit(1);
}
}

HYPERGRAPH SAMPLE_READER::ReadHypergraph(xmlNodePtr root, set<string> &ids, IDTYPE lastID)
{
try{
    xmlNodePtr graph = GetChildrenNode(root, "Graph");
    if (graph == NULL)
    {
    	cout << "Error while reading Hypergraph from XML file - node 'Graph' expected.\n";
	exit(1);
    }
	
    // true if hypergraph=true is set
    bool isHypergraph(!xmlStrcmp(xmlGetProp(graph,(const xmlChar *)"hypergraph"),(const xmlChar *)"true"));
    const xmlChar *edgemodeName(xmlGetProp(graph,(const xmlChar *)"edgemode"));
    
    if( isHypergraph && (!xmlStrcmp(edgemodeName,(const xmlChar *)"directed")))
      {
	cerr << "This program does not support directed hypergraphs.\n";
	exit(1);
      }

    bool isDirected = false;
    if(!xmlStrcmp(edgemodeName,(const xmlChar *)"directed")) 
      isDirected = true;
    if(!xmlStrcmp(edgemodeName,(const xmlChar *)"defaultdirected"))
      isDirected = true;
      
    HYPERGRAPH result;
    result.setLastUsedID(lastID);
    //variables to check id's unique
   // set<string> ids;
    pair<set<string>::iterator, bool> checker;
    string id_str = "";
    
	xmlChar* tXMLID = xmlGetProp(graph, (const xmlChar *)"id");
        if(NULL != tXMLID) id_str = string((char *)tXMLID);
	checker = ids.insert(id_str);
		if (checker.second == false)
		{
			printf("Error: The id's in 'Graph' are not unique: id = %s\n", id_str.c_str());
			exit(1);
		} 

	result.setName(id_str);
	IDTYPE id;
	result.setID();
		
    // start parsing the node/edge list:
    
    //read Nodes
    
    xmlNodePtr graph_element = GetChildrenNode(graph, "node");
    while(graph_element != NULL)
       {
	  //check node's ids
	        id_str = "";
	        tXMLID = xmlGetProp(graph_element, (const xmlChar *)"id");
                if(NULL != tXMLID) id_str = string((char *)tXMLID);
	  	checker = ids.insert(id_str);
		if (checker.second == false)
		{
			printf("Error: The id's in 'node' are not unique: id = %s\n", id_str.c_str());
			exit(0);
		}
		
		VERTEX v;
		v.setIdLabel(id_str);

		/*
		id = result.getLastUsedID();
		v.setID(++id);
		result.setLastUsedID(id);
		*/

		id = atoi(id_str.c_str());
		v.setID(id);
		result.addVertex(v,id);
		
		graph_element = GetNextChild(graph_element, "node");
	} 
	  
	
   // read Edges
   
//   map<IDTYPE, list<IDTYPE> > conVertices;
//   map<IDTYPE, multiset<IDTYPE> > conEdges;
//   typedef pair<IDTYPE, list<IDTYPE> > type_conVertices;
//   typedef pair<IDTYPE, multiset<IDTYPE> >type_conEdges;
   list<IDTYPE> list_of_vertices;
//   multiset<IDTYPE> mset_of_edges;
   
	if (!isHypergraph)
	{
		// (biconnected) edge
		graph_element = GetChildrenNode(graph, "edge");
	  	while(graph_element != NULL)
       		{
	  		//check edge's IDs
	                id_str = "";
	                tXMLID = xmlGetProp(graph_element, (const xmlChar *)"id");
                        if(NULL != tXMLID) id_str = string((char *)tXMLID);
	  		checker = ids.insert(id_str);
			if (checker.second == false)
			{
				printf("Error: The id's in 'edge' are not unique: id = %s\n", id_str.c_str());
				exit(1);
			}
		
			EDGE edge;
			edge.setIdLabel(id_str);

			/*
			id = result.getLastUsedID();
			edge.setID(++id);
			result.setLastUsedID(id);
			*/

			id = atoi(id_str.c_str());
			edge.setID(id);
			
			string fromXMLID = (string)(char*)xmlGetProp(graph_element,(const xmlChar *)"from");
	    		string toXMLID = (string)(char*)xmlGetProp(graph_element,(const xmlChar *)"to");
			
			if (!((result.hasVertex(fromXMLID))&&(result.hasVertex(toXMLID))))
			{
				cout << "The vertex used in edge's connection does not exist in this graph.\n";
				exit(1);
			}

			VERTEX ver1;
			VERTEX ver2;
			list_of_vertices.clear();

			ver1 = result.getVertex(atoi(fromXMLID.c_str()));
			list_of_vertices.push_back(ver1.getID());
			ver2 = result.getVertex(atoi(toXMLID.c_str()));
			list_of_vertices.push_back(ver2.getID());

			// add Edge to ConnectedEdges
			result.addEdgeToCE(ver1.getID(), edge.getID());
			result.addEdgeToCE(ver2.getID(), edge.getID());
			
			//conVertices.insert(type_conVertices(id, list_of_vertices));
			
			xmlNodePtr attribut = GetChildrenNode(graph_element, "attr");
			if(attribut != NULL)
				if(!xmlStrcmp(xmlGetProp(attribut,(const xmlChar *)"name"),(const xmlChar *)"label")) 
				{
					attribut = GetChildrenNode(attribut, "string");
					if(attribut != NULL) 
						attribut = attribut->xmlChildrenNode;
						if (attribut != NULL)
						{
							edge.setLabel((char*)attribut->content);
						}
				}	
					
			result.addEdge(edge,id,list_of_vertices); 
			
			graph_element = GetNextChild(graph_element, "edge");
		}
	}
	
	graph_element = GetChildrenNode(graph, "rel");
	if((!isHypergraph)&&(graph_element != NULL)) // only hypergraphs have rel edges
	      {
		cerr << "Error in Sample: Hyperedge found in a non-hypergraph\n";
		exit(1);
	      }
	      
	while(graph_element != NULL)
	{
		//check edge's IDs
	        id_str = "";
	        tXMLID = xmlGetProp(graph_element, (const xmlChar *)"id");
                if(NULL != tXMLID) id_str = string((char *)tXMLID);
	  	checker = ids.insert(id_str);
		if (checker.second == false)
			{
				printf("Error in Sample: The id's in 'edge' are not unique: id = %s\n", id_str.c_str());
				exit(1);
			}
		
		EDGE edge;	
			
		edge.setIdLabel(id_str);
		id = atoi(id_str.c_str());
		edge.setID(id);
		
		xmlNodePtr attribut = GetChildrenNode(graph_element, "attr");
			if(attribut != NULL)
				if(!xmlStrcmp(xmlGetProp(attribut,(const xmlChar *)"name"),(const xmlChar *)"label")) 
				{
					attribut = GetChildrenNode(attribut, "string");
					if(attribut != NULL) 
						attribut = attribut->xmlChildrenNode;
						if (attribut != NULL)
						{
							edge.setLabel((char*)attribut->content);
						}
				}	
		
		int order = 0;
		xmlNodePtr relend = GetRelend(graph_element, order);
		list_of_vertices.clear();
		string s;
		while(relend != NULL)
		{
			s = (string)(char*)xmlGetProp(relend,(const xmlChar *)"target");
			if(!result.hasVertex(s))
			{
				cout << "Error in Sample: The vertex used in edge's connection (relend) does not exist in this graph.\n";
				exit(1);
			}
			VERTEX ver;
			ver = result.getVertex(atoi(s.c_str()));
			list_of_vertices.push_back(ver.getID());
			order++;
			relend = GetRelend(graph_element, order);
			
		}
		
		result.addEdge(edge, id, list_of_vertices);
		
		graph_element = GetNextChild(graph_element, "rel");
		
	}

	  return result;
       }
catch(exception e)
{
	cout << "Error in Sample: SAMPLE_READER::ReadHypergraph";
	exit(1);
}  
   
}

xmlNodePtr SAMPLE_READER:: GetRelend(xmlNodePtr rel, int order)
{
	string s;
	xmlNodePtr relend = GetChildrenNode(rel, "relend");
	while(relend != NULL)
	{
		s = (char*)xmlGetProp(relend,(const xmlChar *)"startorder");
		if(atoi(s.c_str()) == order)
			return relend;
		relend = GetNextChild(relend, "relend");
	}
	
	return NULL;
}



///////// namespaces ////////

}
}



