/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>

#include "sample_writer.h"

namespace algorithms {

namespace writers {

SAMPLE_WRITER::SAMPLE_WRITER()
{
}


SAMPLE_WRITER::~SAMPLE_WRITER()
{
}

void SAMPLE_WRITER::initialize(char *pfilename)
{
    int n = strlen(pfilename);
    
    filename = new char[n+1];
    strcpy(filename,pfilename);    
}

void SAMPLE_WRITER::run(SAMPLE_STORAGE & ss)
{
    xmlDocPtr doc;

    doc = xmlNewDoc((const xmlChar *)"1.0");
    xmlNodePtr root = xmlNewNode(NULL,(const xmlChar *)"m_list_storage");
    xmlDocSetRootElement(doc,root);

	xmlNodePtr storage_xml = xmlDocGetRootElement(doc);

	for(list<SAMPLE>::iterator it = ss.m_list_storage.begin(); it!=ss.m_list_storage.end(); it++)
	{ 
		WriteSample(storage_xml, *it);
	}

    if(-1 == xmlSaveFormatFile(filename,doc,true))
    {
      cout << "Error writing XML file\n";
      exit(1);
    }

}

void SAMPLE_WRITER::WriteSample(xmlNodePtr list_sample, SAMPLE &sample)
{
	xmlNodePtr sample_in_list = xmlNewChild(list_sample, NULL, (const xmlChar *)"sample", NULL);
	
	xmlNodePtr sample_tree = xmlNewChild(sample_in_list, NULL, (const xmlChar *)"tree", NULL);
	Write_m_list(sample_tree, sample.tree);
	
	xmlNodePtr sample_trans = xmlNewChild(sample_in_list, NULL, (const xmlChar *)"trans", NULL);
	for(vector<string>::iterator it = sample.trans.begin(); it != sample.trans.end(); it++)
	{
		xmlNewChild(sample_trans, NULL, (const xmlChar *)"string_in_vector", (const xmlChar *) (*it).c_str());
	}	
}

void SAMPLE_WRITER::Write_m_list(xmlNodePtr m_list_xml, vector<list<HYPERGRAPH_INFO> > &m_list1)
{
	xmlNodePtr vector_m_list = xmlNewChild(m_list_xml, NULL,(const xmlChar *)"vector", NULL);

	for(vector<list<HYPERGRAPH_INFO> >::iterator it = m_list1.begin(); it!=m_list1.end(); it++)
	{ 
		xmlNodePtr list_vector = xmlNewChild(vector_m_list, NULL,(const xmlChar *)"list", NULL);
		for(list<HYPERGRAPH_INFO>::iterator itl = it->begin(); itl!= it->end(); itl++)
		{
			WriteHypergraphInfo(list_vector, *itl);	
		}	
	
	}

}

// write HYPERGRAPH_INFO into xml-file
void SAMPLE_WRITER:: WriteHypergraphInfo(xmlNodePtr list_el, HYPERGRAPH_INFO &hgInfo)
{
	char ch[25];
	xmlNodePtr HyperInfo = xmlNewChild(list_el, NULL, (const xmlChar *)"HypergraphInfo", NULL);
	sprintf(ch,"%i",hgInfo.id);
	xmlSetProp(HyperInfo,(const xmlChar*)"id",(const xmlChar*)ch);
	sprintf(ch,"%i",hgInfo.link);
	xmlSetProp(HyperInfo,(const xmlChar*)"link",(const xmlChar*)ch);

	WriteHypergraph(HyperInfo, hgInfo.hg);

	/*
	xmlNodePtr matchesLG = xmlNewChild(HyperInfo, NULL,(const xmlChar*)"matchesLG", NULL);
	for(list<MATCH>::iterator it = hgInfo.matchesLG.begin(); it!= hgInfo.matchesLG.end(); it++)
	{
		xmlNodePtr map_matchesLG = xmlNewChild(matchesLG, NULL,(const xmlChar*)"map",NULL);	
		map<IDTYPE,IDTYPE> map_LG = it->getMatch();
		for(map<IDTYPE,IDTYPE>::iterator itm = map_LG.begin(); itm!= map_LG.end(); itm++)	
		{
			xmlNodePtr map_element = xmlNewChild(map_matchesLG,NULL,(const xmlChar*)"map_element",NULL);
			sprintf(ch,"%i",itm->first);
			xmlSetProp(map_element,(const xmlChar*)"id1",(const xmlChar*)ch);
			sprintf(ch,"%i",itm->second);
			xmlSetProp(map_element,(const xmlChar*)"id2",(const xmlChar*)ch);
		}
	}
	xmlNodePtr matchesGgM = xmlNewChild(HyperInfo, NULL,(const xmlChar*)"matchesGgM", NULL);
	for(list<MATCH>::iterator it = hgInfo.matchesGgM.begin(); it!= hgInfo.matchesGgM.end(); it++)
	{
		xmlNodePtr map_matchesGgM = xmlNewChild(matchesGgM, NULL,(const xmlChar*)"map",NULL);
		map<IDTYPE,IDTYPE> map_GgM = it->getMatch();
		for(map<IDTYPE,IDTYPE>::iterator itm = map_GgM.begin(); itm!= map_GgM.end(); itm++)	
		{
			xmlNodePtr map_element = xmlNewChild(map_matchesGgM,NULL,(const xmlChar*)"map_element",NULL);
			sprintf(ch,"%i",itm->first);
			xmlSetProp(map_element,(const xmlChar*)"id1",(const xmlChar*)ch);
			sprintf(ch,"%i",itm->second);
			xmlSetProp(map_element,(const xmlChar*)"id2",(const xmlChar*)ch);
		}
	}
	*/
	
	xmlNodePtr nodeLabels = xmlNewChild(HyperInfo, NULL,(const xmlChar*)"node_labels", NULL);
	for(map<IDTYPE,list<unsigned> >::iterator itm = hgInfo.eq_labels.begin(); itm!= hgInfo.eq_labels.end(); itm++)	
	{
		if(itm->second.size() != 0)
		{
			xmlNodePtr map_element = xmlNewChild(nodeLabels,NULL,(const xmlChar*)"map_element",NULL);
			sprintf(ch,"%i",itm->first);
			xmlSetProp(map_element,(const xmlChar*)"id1",(const xmlChar*)ch);
			string s = "";
			for(list<unsigned>::iterator itl=itm->second.begin(); itl!=itm->second.end(); itl++)
			{
			sprintf(ch,"%i",*itl);
			s += ch;
			s += ";";
			} 
			xmlSetProp(map_element,(const xmlChar*)"id2",(const xmlChar*)s.c_str());
		}
	}	
}

// write hypergraph
void SAMPLE_WRITER::WriteHypergraph(xmlNodePtr node, HYPERGRAPH &hg)
{
	xmlNodePtr graph = xmlNewChild(node, NULL, (const xmlChar *)"Graph", NULL);
	xmlSetProp(graph, (const xmlChar *)"id",(const xmlChar *)(hg.getName()).c_str());
	xmlSetProp(graph, (const xmlChar*)"edgeids", (const xmlChar*)"true");
	xmlSetProp(graph, (const xmlChar*)"hypergraph", (const xmlChar*)"true");
	xmlSetProp(graph, (const xmlChar*)"edgemode", (const xmlChar*)"undirected");

	//start to write nodes
	map<IDTYPE, VERTEX> map_tmp = hg.getVertices();
	for(map<IDTYPE, VERTEX>::iterator it = map_tmp.begin(); it != map_tmp.end(); it++)
	{
		xmlNodePtr node = xmlNewChild(graph, NULL, (const xmlChar*)"node", NULL);
		
		// xmlSetProp(node,(const xmlChar*)"id",(const xmlChar*)(it->second).getIdLabel().c_str());
		
		char ch[25];
		sprintf(ch,"%i",it->first);
		xmlSetProp(node,(const xmlChar*)"id",(const xmlChar*)ch);
	}
	
	//start to write edges (describe all edges as hyperedges)
	map<IDTYPE, EDGE> edges_tmp = hg.getEdges();
	for(map<IDTYPE, EDGE>::iterator it = edges_tmp.begin(); it != edges_tmp.end(); it++)
	{
		xmlNodePtr edge = xmlNewChild(graph, NULL, (const xmlChar*)"rel", NULL);
		
		// xmlSetProp(edge, (const xmlChar*)"id", (const xmlChar*)(it->second).getIdLabel().c_str());

		char ch[25];
		sprintf(ch,"%i",it->first);
		xmlSetProp(edge, (const xmlChar*)"id", (const xmlChar*)ch);

		xmlNodePtr attr = xmlNewChild(edge, NULL, (const xmlChar*)"attr", NULL);
		xmlSetProp(attr, (const xmlChar*)"name", (const xmlChar*)"label");
		xmlNewChild(attr, NULL, (const xmlChar*)"string",(const xmlChar*)(it->second).getLabel().c_str());
		
		list<IDTYPE> relends = hg.getConnectedVertices((it->second).getID());
		VERTEX v;
		int order = 0;
		string s;
		char buffer[25];
		for(list<IDTYPE>::iterator itr = relends.begin(); itr != relends.end(); itr++)
		{
			xmlNodePtr relend = xmlNewChild(edge, NULL, (const
			xmlChar*)"relend", NULL);

			// v = hg.getVertex(*itr);
			// xmlSetProp(relend, (const xmlChar*)"target", (const xmlChar*)v.getIdLabel().c_str());

			char ch[25];
			sprintf(ch,"%i",*itr);
			xmlSetProp(relend, (const xmlChar*)"target", (const xmlChar*)ch);

			sprintf(buffer, "%i", order);
  			xmlSetProp(relend, (const xmlChar*)"startorder", (const xmlChar*)buffer);
			order++;
			
			
			
		}
		
	}
	
}

}

}
