/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#ifndef ALGORITHM_SCENARIO_CPP
#define ALGORITHM_SCENARIO_CPP

#include <fstream>
#include "scenario.h"

namespace algorithm
{
   SCENARIO::SCENARIO()
   {
      this->ID = ALGORITHM_GENERIC_SCENARIO;
      this->name = "generic_scenario";
      /*
       * usage string of the scenario (should be overwritten by actual scenarios)
       */
      this->call_descriptor = "scenario usage: augur [options] -sc=<scenario> infiles outfiles";
   }
   SCENARIO::~SCENARIO() {}
   
   bool SCENARIO::initialize(list<string> files)
   {
      if(files.size() != this->filecount())
      {
         if(!quiet) cout << "severe error: a SCENARIO named \"" << this->name << "\" received "
	                 << files.size() << " instead of " << this->filecount() << ".\n";
	 cout << call_descriptor << "\n";
         exit(1);
      }
      list<string>::iterator fileit = files.begin();
      for(unsigned i=0; (fileit != files.end()) && (i < this->inspec.size()); i++)
      {
         this->filemap[this->inspec[i].first] = *fileit;
	 
	 /*
	  * try to access the file (read-only)
	  */
	 ifstream fin;
	 fin.open(fileit->c_str(), ios::in);
	 if(!fin)
	 {
	    if(!quiet) cout << "severe error: " << *fileit << " could not be opened for reading.\n";
	    cout << call_descriptor << "\n";
	    exit(1);
	 }
	 fin.close();
	 fileit++;
      }
      for(unsigned i=0; fileit != files.end(); i++)
      {
         this->filemap[this->outspec[i].first] = *fileit;
      
         /*
	  * try to access the file (write)
	  *
	 ofstream fout;
	 fout.open(fileit->c_str());
	 if(!fout)
	 {
	    if(!quiet) cout << "severe error: " << *fileit << " could not be opened for writing.\n";
	    cout << call_descriptor << "\n";
	    exit(1);
	 }
         fout.close();
	 */

	 fileit++;
      }
      return true;
   }
   
   unsigned SCENARIO::in_filecount() { return this->inspec.size(); }
   unsigned SCENARIO::out_filecount() { return this->outspec.size(); }
   unsigned SCENARIO::filecount()
   {
      return (this->inspec.size() + this->outspec.size());
   }
   
   pair<string, DATA_FORMAT> SCENARIO::in_descriptor(unsigned i) { return this->inspec[i]; }
   pair<string, DATA_FORMAT> SCENARIO::out_descriptor(unsigned i) { return this->outspec[i]; }
   pair<string, DATA_FORMAT> SCENARIO::descriptor(unsigned i)
   {
      return ((inspec.size() > i)? inspec[i]: outspec[i - inspec.size()]);
   }
};
using namespace algorithm;

#endif
