/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#ifndef ALGORITHM_SCENARIO_H
#define ALGORITHM_SCENARIO_H

#include "algorithm.h"

using namespace typedefs;

namespace algorithm
{
   class SCENARIO: public ALGORITHM
   {
    public:
      SCENARIO();
      virtual ~SCENARIO();
      virtual bool initialize(list<string> files);
      virtual void run() = 0;
    
      /*
       * these methods return the number of (in/out) arguments
       */
      unsigned in_filecount();
      unsigned out_filecount();
      unsigned filecount();
      
      /*
       * the first element of the pair is an informal description or name
       * of the i-th (in/out) argument to the scenario, the second element
       * is its data format
       */
      pair<string, DATA_FORMAT> in_descriptor(unsigned i);
      pair<string, DATA_FORMAT> out_descriptor(unsigned i);
      pair<string, DATA_FORMAT> descriptor(unsigned i);
      
    protected:
      /*
       * specification of in-/output files
       * these values must be set by the constructor of a specific scenario.
       */
      vector< pair<string, DATA_FORMAT> > inspec;
      vector< pair<string, DATA_FORMAT> > outspec;
      
      /*
       * initialize processes the file specification
       * and, on success, stores the filenames in a map, such
       * that the actual implementation need not remember the order
       * in which the arguments were provided.
       */
      map<string, string> filemap;
      
      /*
       * what should a call to this augur scenario look like?
       * this information is promulgated on parameter failure
       * (wrong arg. count, non-readable files, etc.)
       */
      string call_descriptor;
   };
};
using namespace algorithm;

#endif
