/***************************************************************************
 *   Copyright (C) 2006 by root   *
 *   root@Vitali   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "searchplan.h"
#include "algorithm_storage.h"

namespace algorithms {

namespace matchings
{
namespace data_str_searchplan
{
SEARCHPLAN::SEARCHPLAN()
{
	loaded = false;
}


SEARCHPLAN::~SEARCHPLAN()
{
}

bool SEARCHPLAN::initialize(COSTMATRIX* cm, HYPERGRAPH* hg, ALGORITHM* matchhandler, bool ptg){

	assert(cm != NULL);
	assert(hg != NULL);

	this->costMatrix = cm;
	this->hyperGraph = hg;
	this->matchhandler = matchhandler;
	
	this->ptg = ptg;
	
	if(!makeCostGraph())return false;

	makeCostTree();
// 	EDGE e;
// 	e.setID( 20 );
// 	e.setLabel( "Dummy" );
// 	costTree.addEdge( e , 0 , 6, 1 );
	list<IDTYPE> circle;
	circle = costTree.getFirstCircle();
	map<IDTYPE,int> removedEdges;
	while (!circle.empty()){
		
		float minC  = -1;
		float cost;
		IDTYPE minE = 0;
		IDTYPE v1,v2;
		// Find cheapest Edge in the Circle
		for(list<IDTYPE>::iterator iter = circle.begin(); iter != circle.end(); iter++){
			
			if( minC == -1 || costGraph.getCost( costTree.getIncoming( *iter ).front() ) < minC ) 
				minC = costGraph.getCost( costTree.getIncoming( *iter ).front() );
		}
		// Find incoming Edge to remove with		
		for(list<IDTYPE>::iterator iter = circle.begin(); iter != circle.end(); iter++){

			// Every Vertex in CostTree have only one incoming Edge
			IDTYPE eTree;
			eTree = costTree.getIncoming( *iter ).front();
			list<IDTYPE> edges = costGraph.getIncoming( *iter );

			for(list<IDTYPE>::iterator e_iter = edges.begin(); e_iter != edges.end(); e_iter++){

				if(eTree == *e_iter || removedEdges[*e_iter]==1) continue;

				float condC;
				condC = costGraph.getCost( *e_iter ) - costGraph.getCost( eTree ) + minC;
				if( minE == 0 ) minE = *e_iter;
				else if( condC<costGraph.getCost( minE ) ) minE = *e_iter;
			}
		}
		
		v1 = costGraph.getSource( minE );
		v2 = costGraph.getTarget( minE );
		cost = costGraph.getCost( minE );
		
		removedEdges[costTree.getIncoming( v2 ).front()] = 1;
		costTree.delEdge( costTree.getIncoming( v2 ).front() );
		
		costTree.addEdge( costGraph.getEdge( minE ), v1, v2, cost );
		circle = costTree.getFirstCircle();		
		
	}
	float cost=1;
	map<IDTYPE, EDGE> edges = costTree.getEdges();
	for(map<IDTYPE, EDGE>::iterator iter = edges.begin(); iter != edges.end(); iter ++){
//  		cout << "Kanten " << iter->first <<" von " <<costTree.getSource( iter->first ) << " nach " << costTree.getTarget( iter->first ) << " Kosten: " << costGraph.getCost( iter->first )  << endl;
// 		cout << costGraph.getCost( iter->first ) <<endl;
		cost = cost * costGraph.getCost( iter->first );
	}
	
	searchPlanList.clear();
	makeSearchPlanList( 0 );
	
	allMatches.clear();
	
// 	for(vector<IDTYPE>::iterator it = searchPlanList.begin(); it != searchPlanList.end(); it++){
// 
// 		cout << *it << ", ";
// 	}
// 	cout << endl;
// 	cout << "Es gibt " << cost << " Matches" << endl;
// 	MATCH result = startMatching();
// 	cout << "RESULT " << result << endl;
// 	
	/*list<MATCH> result_list = startMatchingForAll();
	cout << "FINDALLMATCHES!!!" << endl;
	for(list<MATCH>::iterator iter = result_list.begin(); iter!=result_list.end(); iter++){
		cout << *iter << endl;
	}*/	
	return true;
}

bool SEARCHPLAN::makeCostGraph(){
	
	//Insert Start Vertex
// 	COSTGRAPH* cg = new COSTGRAPH();
// 	this->costGraph = *cg;
	costGraph.clean();
	VERTEX v; v.setID( 0 ); v.setLabel( "*" );
	this->costGraph.addVertex( v );
	

	/*
	*  Make Vertexes from hyperGraph EDGES	
	*/
	map<IDTYPE, EDGE> edges = hyperGraph->getEdges();
         	for(map<IDTYPE, EDGE>::iterator i = edges.begin(); i != edges.end(); i++)
	 	{
// 			EDGE e = i->second;
			VERTEX v;
			v.setID(i->first);
			v.setLabel( i->second.getLabel() );
			costGraph.addVertex( v );
			
			
			
		}
	/*
	*  Make Edges for costGraph	
	*/
	map<IDTYPE, VERTEX> vertices = costGraph.getVertices();
         	for(map<IDTYPE, VERTEX>::iterator i = vertices.begin(); i != vertices.end(); i++)
	 	{
			if(i->first==0)continue;
// 			VERTEX v = i->second;
// 			VERTEX v1 = costGraph.getVertex( 0 );
			EDGE e;
			int id = (int)costGraph.getLastUsedID()+1;
			
			e.setID( id );
			e.setLabel( "*-" + i->second.getLabel() );
			if(costMatrix->getLookUp( i->second.getLabel() ).empty()) return false;
			costGraph.addEdge( e, 0, i->second.getID(), costMatrix->getDirectCost( i->second.getLabel() ) );
		}
	

	edges = hyperGraph->getEdges();
         	for(map<IDTYPE, EDGE>::iterator i = edges.begin(); i != edges.end(); i++)
	 	{

			list<IDTYPE> connVers = hyperGraph->getConnectedVertices(i->second.getID());
			int n1=0;
			for(list<IDTYPE>::iterator connVer_i = connVers.begin(); connVer_i != connVers.end(); connVer_i++){
				
				multiset<IDTYPE> connEs = hyperGraph->getConnectedEdges( *connVer_i );
				for(multiset<IDTYPE>::iterator connE_i = connEs.begin(); connE_i != connEs.end(); connE_i++){
					
					list<IDTYPE> connVers1 = hyperGraph->getConnectedVertices( *connE_i );
					int n2=0;
					for(list<IDTYPE>::iterator connVer1_i = connVers1.begin(); connVer1_i != connVers1.end(); connVer1_i++){
						if(*connVer1_i==*connVer_i){
							if( *connE_i==i->first && n1 == n2 )continue;
							char firstIndex[255+1]; char secondIndex[255+1];
	    						sprintf( firstIndex, "%i", n1 );
							sprintf( secondIndex, "%i", n2 );
							string label; 
							label = i->second.getLabel() + "-" + firstIndex + "-" + secondIndex + "-" + hyperGraph->getEdge( *connE_i ).getLabel();
							EDGE e;
							int id = (int)costGraph.getLastUsedID()+1;
			
							e.setID( id );
							e.setLabel( label );
							
// 							if(costMatrix->getCostOver( label ) == HUGE_VAL ) return false;
							costGraph.addEdge( e, i->first, *connE_i, costMatrix->getCostOver( label ));
						}
						n2++;
					}
				}
				n1++;
			}
		}	
// 	AUSGABE::::::::::
	map<IDTYPE, VERTEX> verts = costGraph.getVertices();
         	for(map<IDTYPE, VERTEX>::iterator i1 = verts.begin(); i1 != verts.end(); i1++)
	 	{
			IDTYPE id = i1->first;

// 			cout << id << " " << i1->second.getLabel() <<  endl;
		}

	map<IDTYPE, EDGE> edges1 = costGraph.getEdges();
		for(map<IDTYPE,	EDGE>::iterator i2 = edges1.begin(); i2 != edges1.end(); i2++){
			
			IDTYPE id1 = i2->first;

// 			cout << id1 << " " << i2->second.getLabel() << " " << costGraph.getCost( id1 ) << endl;
		}
	return true;
}

void SEARCHPLAN::makeCostTree(){

// 	this->costTree = *(new COSTTREE());
	costTree.clean();
	map<IDTYPE,VERTEX> vertexies = costGraph.getVertices();

	for(map<IDTYPE,VERTEX>::iterator ver_iter = vertexies.begin(); ver_iter != vertexies.end(); ver_iter++){
	
		costTree.addVertex( ver_iter->second );
	}

	for(map<IDTYPE,VERTEX>::iterator ver_iter = vertexies.begin(); ver_iter != vertexies.end(); ver_iter++){
		
		if(ver_iter->second.getLabel() == "*") continue;
		list<IDTYPE> incomings = costGraph.getIncoming(ver_iter->first);
		float cost=-1;
		IDTYPE best;
		for(list<IDTYPE>::iterator e_ID = incomings.begin(); e_ID != incomings.end(); e_ID++){

			if(costGraph.getCost( *e_ID )<cost || cost == -1){
				cost=costGraph.getCost( *e_ID );
				best = *e_ID;
			}
		} 
// 		cout<< "Best incoming " << best << endl;
		costTree.addEdge( costGraph.getEdge(best), costGraph.getSource(best), costGraph.getTarget(best), cost );

	}

}

void SEARCHPLAN::makeSearchPlanList(IDTYPE id){
	
	if(id!=0) searchPlanList.push_back(id);
	
	list<IDTYPE> edges = costTree.getOrderedOutgoing( id );
	for(list<IDTYPE>::iterator e_iter = edges.begin(); e_iter != edges.end(); e_iter++){
		makeSearchPlanList(costTree.getTarget(*e_iter));
	}
}

MATCH SEARCHPLAN::runOperation(MATCH* prevMatch, int operationPosition){

	MATCH result;
	if(searchPlanList.empty()) return result;
	IDTYPE opID = searchPlanList[operationPosition];
	list<IDTYPE> incoming = costTree.getIncoming( opID );
	list<IDTYPE> candidates;
	
	string label = costTree.getEdge( incoming.front() ).getLabel();

	//// 	Get candidate list
	
	if(label.find('*',0) != string::npos){
	//// 	For lookup Operation
		candidates = costMatrix->getLookUp( hyperGraph->getEdge( opID ).getLabel() );
// 		candidates.unique();
	}
	else{
	//// 	For extension Operation
		string::size_type loc = label.find( "-", 0 );

		label = label.substr(loc+1);
		int i = atoi(label.c_str());
		
		IDTYPE sourceID = costTree.getSource(costTree.getIncoming(opID).front());
		
		list<IDTYPE> ver = hyperGraph->getConnectedVertices( sourceID );
		int i1 = 0;
		IDTYPE v;
		for(list<IDTYPE>::iterator v_i = ver.begin(); v_i != ver.end(); v_i++){

			if(i==i1){
				v = prevMatch->getMatch( *v_i );
				break;
			}
			i1++;
		}
		
		loc = label.find( "-", 0 );
		label = label.substr(loc);
		char nodeID[255+1];
	    	sprintf( nodeID, "%i", v );
		string leer = "";
		label = leer + nodeID + label;
		
		candidates = costMatrix->getNodeConnection( label );
		if(candidates.empty()){
			
			map<IDTYPE, IDTYPE> r_match = result.getMatch();
			r_match.clear();
			result.setMatch( r_match );
			return result; //Fals keine möglichkeiten zum MATCH gibt, also vorheriger MATCH war falsch
		}
// 		candidates.unique();
		
	}

	for(list<IDTYPE>::iterator candidate = candidates.begin(); candidate != candidates.end(); candidate++){
	
		
		result.setMatch( opID, *candidate ); //Match for EDGE
		list<IDTYPE> sourceV, targetV; 
		sourceV = costMatrix->hyperGraph->getConnectedVertices( *candidate );
		targetV = hyperGraph->getConnectedVertices( opID );
		list<IDTYPE>::iterator sV, tV;
		
		bool badVertex = false;

		for(sV=sourceV.begin(), tV=targetV.begin(); sV!=sourceV.end(); sV++, tV++){
			if(( result.hasMatch( *tV ) && result.getMatch( *tV ) != *sV) || (prevMatch->hasMatch(*tV) && prevMatch->getMatch(*tV) != *sV)) 
				badVertex = true;
			
			else result.setMatch( *tV, *sV );
		}
			
		if(badVertex){ 
			
			//MATCH local_result;
// 			cout << "bad VERTEX!" << endl;
			map<IDTYPE, IDTYPE> r_match = result.getMatch();
			r_match.clear();
			result.setMatch( r_match );
			continue;
		}

		MATCH m1 = *prevMatch;
		map<IDTYPE,IDTYPE> matchMap = result.getMatch();
			
		for(map<IDTYPE,IDTYPE>::iterator match_iter = matchMap.begin(); match_iter != matchMap.end(); match_iter++){

			if(!m1.hasMatch( match_iter->first )) 
				m1.setMatch( match_iter->first, match_iter->second );
		}
				
		int n = operationPosition+1;
		if(searchPlanList.size()==n){
			bool check = true;

			
			
			if(matchhandler != NULL && ptg){	
	     			if(!loaded)
				{
					// set DATA to Algorithm Storage
					string to_alg = "findmatch_by_sp";
					storage.setData(to_alg, (DATA *)this);
				}

				check = matchhandler->run(&m1);
			}
			
			if(check){
// 				cout << m1 << endl;
				
				 return m1; //Am Ende Angekommen
			}
			else{
				matchMap.clear();
				result.setMatch( matchMap );
				continue;
			}
		}

		MATCH res = runOperation( &m1, n); //Iteration

		matchMap.clear();
						
		if(res.count()>m1.count()){
			return res;
		}
		else{
			result.setMatch( matchMap );
			continue;
		}
	}
	return result;	
}

MATCH SEARCHPLAN::runOperationForAll(MATCH* prevMatch, int operationPosition){

	MATCH result;
	if(searchPlanList.empty()) return result;
	IDTYPE opID = searchPlanList[operationPosition];
	list<IDTYPE> incoming = costTree.getIncoming( opID );
	list<IDTYPE> candidates;
	
	string label = costTree.getEdge( incoming.front() ).getLabel();

	//// 	Get candidate list
	
	if(label.find('*',0) != string::npos){
	//// 	For lookup Operation
		candidates = costMatrix->getLookUp( hyperGraph->getEdge( opID ).getLabel() );
// 		candidates.unique();
	}
	else{
	//// 	For extension Operation
		string::size_type loc = label.find( "-", 0 );

		label = label.substr(loc+1);
		int i = atoi(label.c_str());
		
		IDTYPE sourceID = costTree.getSource(costTree.getIncoming(opID).front());
		
		list<IDTYPE> ver = hyperGraph->getConnectedVertices( sourceID );
		int i1 = 0;
		IDTYPE v;
		for(list<IDTYPE>::iterator v_i = ver.begin(); v_i != ver.end(); v_i++){

			if(i==i1){
				v = prevMatch->getMatch( *v_i );
				break;
			}
			i1++;
		}
		
		loc = label.find( "-", 0 );
		label = label.substr(loc);
		char nodeID[255+1];
	    	sprintf( nodeID, "%i", v );
		string leer = "";
		label = leer + nodeID + label;
		
		candidates = costMatrix->getNodeConnection( label );
		if(candidates.empty()){
			
			map<IDTYPE, IDTYPE> r_match = result.getMatch();
			r_match.clear();
			result.setMatch( r_match );
			return result; //Fals keine möglichkeiten zum MATCH gibt, also vorheriger MATCH war falsch
		}
// 		candidates.unique();
		
	}

	for(list<IDTYPE>::iterator candidate = candidates.begin(); candidate != candidates.end(); candidate++){
	
		
		result.setMatch( opID, *candidate ); //Match for EDGE
		list<IDTYPE> sourceV, targetV; 
		sourceV = costMatrix->hyperGraph->getConnectedVertices( *candidate );
		targetV = hyperGraph->getConnectedVertices( opID );
		list<IDTYPE>::iterator sV, tV;
		
		bool badVertex = false;

		for(sV=sourceV.begin(), tV=targetV.begin(); sV!=sourceV.end(); sV++, tV++){
			if(( result.hasMatch( *tV ) && result.getMatch( *tV ) != *sV) || (prevMatch->hasMatch(*tV) && prevMatch->getMatch(*tV) != *sV)) 
				badVertex = true;
			
			else result.setMatch( *tV, *sV );
		}
			
		if(badVertex){ 
			
			//MATCH local_result;
// 			cout << "bad VERTEX!" << endl;
			map<IDTYPE, IDTYPE> r_match = result.getMatch();
			r_match.clear();
			result.setMatch( r_match );
			continue;
		}

		MATCH m1 = *prevMatch;
		map<IDTYPE,IDTYPE> matchMap = result.getMatch();
			
		for(map<IDTYPE,IDTYPE>::iterator match_iter = matchMap.begin(); match_iter != matchMap.end(); match_iter++){

			if(!m1.hasMatch( match_iter->first )) 
				m1.setMatch( match_iter->first, match_iter->second );
		}
				
		int n = operationPosition+1;
		if(searchPlanList.size()==n){
			bool check = true;

			allMatches.push_back( m1 );

			
			
			
			matchMap.clear();
			result.setMatch( matchMap );
			continue;
			
		}

		MATCH res = runOperationForAll( &m1, n); //Iteration

		matchMap.clear();
						
		if(res.count()>m1.count()){
			return res;
		}
		else{
			result.setMatch( matchMap );
			continue;
		}
	}
	return result;	
}


MATCH SEARCHPLAN::startMatching(){

	MATCH result;
	list<IDTYPE> edges = costTree.getOutgoing( 0 );
	for(list<IDTYPE>::iterator e_iter = edges.begin(); e_iter != edges.end(); e_iter++){

		MATCH res; 
		res = lookUp(costTree.getTarget(*e_iter), &result);

		map<IDTYPE,IDTYPE> matchMap = res.getMatch();
		
		if(matchMap.empty()) return res;
		for(map<IDTYPE,IDTYPE>::iterator match_iter = matchMap.begin(); match_iter != matchMap.end(); match_iter++){

			if(!result.hasMatch( match_iter->first )) result.setMatch( match_iter->first, match_iter->second );
			else{ 
				if(result.getMatch( match_iter->first ) != match_iter->second){

					/*MATCH badMatch;
					badMatch.setMatch( matchMap );
					setTestMatch( badMatch );
					return startMatching();*/
					problemNode = result.getMatch( match_iter->first );
					map<IDTYPE, IDTYPE> r_match = result.getMatch();
					r_match.clear();
					result.setMatch( r_match );
					e_iter = edges.begin();
				}
			}
		}
		
	}
	return result;
}

MATCH SEARCHPLAN::lookUp( IDTYPE eID, MATCH* resBefore ){
		
	MATCH result;	
	list<IDTYPE> posiblLookUps = costMatrix->getLookUp( hyperGraph->getEdge( eID ).getLabel() );
	if(posiblLookUps.empty()) return result; //Sollte garnicht vorkommen, da bei makeCostGraph vorgesorgt
	posiblLookUps.unique();

	for(list<IDTYPE>::iterator lu_iter = posiblLookUps.begin(); lu_iter != posiblLookUps.end(); lu_iter++){
	
		
		result.setMatch( eID, *lu_iter ); //Match for EDGE
		list<IDTYPE> sourceV, targetV; 
		sourceV = costMatrix->hyperGraph->getConnectedVertices( *lu_iter );
		targetV = hyperGraph->getConnectedVertices( eID );
		list<IDTYPE>::iterator sV, tV;
// 		for(sV=sourceV.begin(), tV=targetV.begin(); sV!=sourceV.end(); sV++, tV++){
// 
// 			result.setMatch( *tV, *sV ); //Match for VERTECIES
// 		}
// 		

		bool badVertex = false;

		for(sV=sourceV.begin(), tV=targetV.begin(); sV!=sourceV.end(); sV++, tV++){
			if(( result.hasMatch( *tV ) && result.getMatch( *tV ) != *sV) || (resBefore->hasMatch(*tV) && resBefore->getMatch(*tV) != *sV) || (*sV==problemNode) ) badVertex = true;
			else result.setMatch( *tV, *sV );
		}
			
		if(badVertex){ 
			
			//MATCH local_result;
// 			cout << "bad VERTEX!" << endl;
			map<IDTYPE, IDTYPE> r_match = result.getMatch();
			r_match.clear();
			result.setMatch( r_match );
			continue;
		}

		MATCH m1 = *resBefore;
		map<IDTYPE,IDTYPE> matchMap = result.getMatch();
			
			for(map<IDTYPE,IDTYPE>::iterator match_iter = matchMap.begin(); match_iter != matchMap.end(); match_iter++){

				if(!m1.hasMatch( match_iter->first )) 
						m1.setMatch( match_iter->first, match_iter->second );
// 					else cout << match_iter->first << " ist schon gematcht worden!!! gotoNext" << endl;
			}
// 		cout << m1 << endl;
		MATCH res = gotoNext( eID, &m1 );
// 		cout << res << endl;
		matchMap = res.getMatch();
		
		if(matchMap.empty()){ 
			result.setMatch( matchMap );
			continue; //Wenn gotoNext kein Match beinhaltet dann wa es eine falsche Wahl, also noch mal wählen
		}
		else if(res.IsEqual( result )) break;
		else{ 
			
			for(map<IDTYPE,IDTYPE>::iterator match_iter = matchMap.begin(); match_iter != matchMap.end(); match_iter++){

				if(!result.hasMatch( match_iter->first )) result.setMatch( match_iter->first, match_iter->second );
				else{

					if(result.getMatch( match_iter->first ) != match_iter->second){

						map<IDTYPE, IDTYPE> r_match = result.getMatch();
						r_match.clear();
						result.setMatch( r_match );
						return result;
					}
				}
			}

			break;
		}
		
	}
	return result;
}


MATCH SEARCHPLAN::gotoNext( IDTYPE eID, MATCH* m1 ){

	MATCH result;
	MATCH local_result;
	list<IDTYPE> outgoingEs = costTree.getOutgoing( eID );
	if(outgoingEs.empty()){ //Am Blatt des Spannbaums angelangt
		 
		if(testMatch.hasMatch( eID ) && !testIsDone ){
			
			if(m1->getMatch( eID ) == testMatch.getMatch( eID ) ){
// 				cout << eID << endl;
				testIsDone = true;
				return result;
			}
			if(testIsDone) return *m1;
		}	
		return *m1;
	}	

	for(list<IDTYPE>::iterator e_iter = outgoingEs.begin(); e_iter != outgoingEs.end(); e_iter++){
		
		
		string label = costTree.getEdge( *e_iter ).getLabel();
		string::size_type loc = label.find( "-", 0 );

		label = label.substr(loc+1);
		int i = atoi(label.c_str());
		list<IDTYPE> ver = hyperGraph->getConnectedVertices( eID );
		int i1 = 0;
		IDTYPE v;
		for(list<IDTYPE>::iterator v_i = ver.begin(); v_i != ver.end(); v_i++){

			if(i==i1){
				v = m1->getMatch( *v_i );
				break;
			}
			i1++;
		}
		
		loc = label.find( "-", 0 );
		label = label.substr(loc);
		char nodeID[255+1];
	    	sprintf( nodeID, "%i", v );
		string leer = "";
		label = leer + nodeID + label;
		
		list<IDTYPE> nodecon = costMatrix->getNodeConnection( label );
		if(nodecon.empty()){
			
			map<IDTYPE, IDTYPE> r_match = result.getMatch();
			r_match.clear();
			result.setMatch( r_match );
			return result; //Fals keine möglichkeiten zum MATCH gibt, also vorheriger MATCH war falsch
		}
		nodecon.unique();
		for(list<IDTYPE>::iterator node_it = nodecon.begin(); node_it != nodecon.end(); node_it++){

// 			if(*node_it == eID) continue;
			
			local_result.setMatch( costTree.getTarget( *e_iter ), *node_it );
			list<IDTYPE> sourceV, targetV; 
			sourceV = costMatrix->hyperGraph->getConnectedVertices( *node_it );
			targetV = hyperGraph->getConnectedVertices( costTree.getTarget( *e_iter ) );
			list<IDTYPE>::iterator sV, tV;
			bool badVertex = false;
// 			cout << "m1: " << *m1 << endl;
			for(sV=sourceV.begin(), tV=targetV.begin(); sV!=sourceV.end(); sV++, tV++){
				if(( local_result.hasMatch( *tV ) && local_result.getMatch( *tV ) != *sV) || (m1->hasMatch(*tV) && m1->getMatch(*tV) != *sV) || (*sV==problemNode) ) badVertex = true;
				else local_result.setMatch( *tV, *sV );
			}
// 			cout << local_result << endl;
			
			if(badVertex){ 
			
				//MATCH local_result;
// 				cout << "bad VERTEX!" << endl;
				map<IDTYPE, IDTYPE> r_match = local_result.getMatch();
				r_match.clear();
				local_result.setMatch( r_match );
				continue;
			}
			
			MATCH lookUpMatch;
			lookUpMatch.setMatch( m1->getMatch() );
			map<IDTYPE,IDTYPE> matchMap = local_result.getMatch();
			
			for(map<IDTYPE,IDTYPE>::iterator match_iter = matchMap.begin(); match_iter != matchMap.end(); match_iter++){

				if(!lookUpMatch.hasMatch( match_iter->first )) 
						lookUpMatch.setMatch( match_iter->first, match_iter->second );
// 					else cout << match_iter->first << " ist schon gematcht worden!!! gotoNext" << endl;
			}
			
// 			cout << lookUpMatch << endl;
			
			MATCH res = gotoNext( costTree.getTarget( *e_iter ), &lookUpMatch ); //REKURSION!!!!
			
			matchMap = res.getMatch();
			if(matchMap.empty()){ 
				
				local_result.setMatch( matchMap );
				continue;
			}
			else if( res.IsEqual( lookUpMatch ) ) {
// 				cout << "lookUpMatch == res" <<endl;
				local_result = lookUpMatch;
				break;
			}
			else{ 
			
				for(map<IDTYPE,IDTYPE>::iterator match_iter = matchMap.begin(); match_iter != matchMap.end(); match_iter++){

					if(!local_result.hasMatch( match_iter->first )) 
						local_result.setMatch( match_iter->first, match_iter->second );
					
					else if(local_result.getMatch( match_iter->first ) != match_iter->second){

// 					cout << "scheisse!!!" << endl;
					}
				}
// 				cout << "loca_result erweitern mit res" <<endl;
// 				cout << res << endl;
				
				break;
			}
		}
		if(local_result.getMatch().empty()) return local_result;
		
		map<IDTYPE,IDTYPE> matchMap = local_result.getMatch();
		for(map<IDTYPE,IDTYPE>::iterator match_iter = matchMap.begin(); match_iter != matchMap.end(); match_iter++){

			if(!result.hasMatch( match_iter->first )) result.setMatch( match_iter->first, match_iter->second );
			else if(result.getMatch( match_iter->first ) != match_iter->second){
				problemNode = result.getMatch( match_iter->first );
				map<IDTYPE, IDTYPE> r_match = result.getMatch();
				r_match.clear();
				result.setMatch( r_match );
				e_iter = outgoingEs.begin();
// 				cout << "scheisse!!!2" << problemNode << endl;
			}
		}
		//result.setMatch( local_result.getMatch() );
	}
// 	cout << "ende" << endl;
	return result;
}

void SEARCHPLAN::setTestMatch( MATCH test ){

	testMatch = test;
	testIsDone = false;
}


list<MATCH> SEARCHPLAN::startMatchingForAll(){
	 
	
	list<MATCH> result;
	
// 	MATCH test = startMatching();
// 	if(test.getMatch().empty()) return result;	

	list<IDTYPE> edges = costTree.getOutgoing( 0 );
	for(list<IDTYPE>::iterator e_iter = edges.begin(); e_iter != edges.end(); e_iter++){

		list<MATCH> res; 
		res = lookUpForAll(costTree.getTarget(*e_iter));
		if(res.empty())return res;
		
		if(result.empty()){ 

			result = res;
			continue;
		}

		list<MATCH> temp_res_list;
		for(list<MATCH>::iterator res_iter = res.begin(); res_iter!=res.end(); res_iter++){
			
			map<IDTYPE,IDTYPE> matchMap = (*res_iter).getMatch();
					
			if(result.empty()) temp_res_list = res;
			
			for(list<MATCH>::iterator result_iter = result.begin(); result_iter!=result.end(); result_iter++){
				bool good = true;
				MATCH temp_res = *result_iter;
				for(map<IDTYPE,IDTYPE>::iterator match_iter = matchMap.begin(); match_iter != matchMap.end(); match_iter++){

					if(!temp_res.hasMatch( match_iter->first )) temp_res.setMatch( match_iter->first, match_iter->second );
					else{ 
						if(temp_res.getMatch( match_iter->first ) != match_iter->second) good = false;
					}
				}
				
				
				if(good)temp_res_list.push_back(temp_res);
			}
		}
		result = temp_res_list;
		
	}
	//result.unique();
	return result;
}

list<MATCH> SEARCHPLAN::lookUpForAll(IDTYPE eID){

	
	list<MATCH> result_list;	
	list<IDTYPE> posiblLookUps = costMatrix->getLookUp( hyperGraph->getEdge( eID ).getLabel() );
	if(posiblLookUps.empty()) return result_list; //Sollte garnicht vorkommen, da bei makeCostGraph vorgesorgt
// 	posiblLookUps.unique();

	for(list<IDTYPE>::iterator lu_iter = posiblLookUps.begin(); lu_iter != posiblLookUps.end(); lu_iter++){
		MATCH result;
		result.setMatch( eID, *lu_iter ); //Match for EDGE
		list<IDTYPE> sourceV, targetV; 
		sourceV = costMatrix->hyperGraph->getConnectedVertices( *lu_iter );
		targetV = hyperGraph->getConnectedVertices( eID );
		list<IDTYPE>::iterator sV, tV;
		for(sV=sourceV.begin(), tV=targetV.begin(); sV!=sourceV.end(); sV++, tV++){

			result.setMatch( *tV, *sV ); //Match for VERTECIES
		}
		
		bool badVertex = false;

		for(sV=sourceV.begin(), tV=targetV.begin(); sV!=sourceV.end(); sV++, tV++){
			if(( result.hasMatch( *tV ) && result.getMatch( *tV ) != *sV) ) badVertex = true;
			else result.setMatch( *tV, *sV );
		}
			
		if(badVertex){ 
			
			//MATCH local_result;
// 			cout << "bad VERTEX!" << endl;
			map<IDTYPE, IDTYPE> r_match = result.getMatch();
			r_match.clear();
			result.setMatch( r_match );
			continue;
		}

		list<MATCH> res_list = gotoNextForAll( eID, &result );
		if(res_list.empty()) continue;

		for(list<MATCH>::iterator list_iter = res_list.begin(); list_iter!= res_list.end(); list_iter++){
			
			result_list.push_back(*list_iter);
		}
	}
	//result_list.unique();
	return result_list;
}

list<MATCH> SEARCHPLAN::gotoNextForAll( IDTYPE eID, MATCH* m1 ){

	MATCH result;
	list<MATCH> result_list;
	
	list<IDTYPE> outgoingEs = costTree.getOutgoing( eID );
	if(outgoingEs.empty()){ 
		
		result_list.push_back(*m1);
		return result_list; //Am Blatt des Spannbaums angelangt
	}
	
	for(list<IDTYPE>::iterator e_iter = outgoingEs.begin(); e_iter != outgoingEs.end(); e_iter++){
		
		
		string label = costTree.getEdge( *e_iter ).getLabel();
		string::size_type loc = label.find( "-", 0 );

		label = label.substr(loc+1);
		int i = atoi(label.c_str());
		list<IDTYPE> ver = hyperGraph->getConnectedVertices( eID );
		int i1 = 0;
		IDTYPE v;
		for(list<IDTYPE>::iterator v_i = ver.begin(); v_i != ver.end(); v_i++){

			if(i==i1){
				v = m1->getMatch( *v_i );
				break;
			}
			i1++;
		}
		
		loc = label.find( "-", 0 );
		label = label.substr(loc);
		char nodeID[255+1];
	    	sprintf( nodeID, "%i", v );
		string leer = "";
		label = leer + nodeID + label;
		
		list<IDTYPE> nodecon = costMatrix->getNodeConnection( label );
		if(nodecon.empty()){
			
			result_list.clear();
			return result_list; //Fals keine möglichkeiten zum MATCH gibt, also vorheriger MATCH war falsch
		}
		nodecon.unique();
		list<MATCH> local_res_list;
		for(list<IDTYPE>::iterator node_it = nodecon.begin(); node_it != nodecon.end(); node_it++){
			MATCH local_result;
// 			if(*node_it == eID) continue;
			
			local_result.setMatch( costTree.getTarget( *e_iter ), *node_it );
			list<IDTYPE> sourceV, targetV; 
			sourceV = costMatrix->hyperGraph->getConnectedVertices( *node_it );
			targetV = hyperGraph->getConnectedVertices( costTree.getTarget( *e_iter ) );
			list<IDTYPE>::iterator sV, tV;
			bool badVertex = false;
// 			cout << "m1: " << *m1 << endl;
			for(sV=sourceV.begin(), tV=targetV.begin(); sV!=sourceV.end(); sV++, tV++){
				if(( local_result.hasMatch( *tV ) && local_result.getMatch( *tV ) != *sV) || (m1->hasMatch(*tV) && m1->getMatch(*tV) != *sV) ) badVertex = true;
				else local_result.setMatch( *tV, *sV );
			}
// 			cout << local_result << endl;
			
			if(badVertex){ 
			
				//MATCH local_result;
// 				cout << "bad VERTEX!" << endl;
				map<IDTYPE, IDTYPE> r_match = local_result.getMatch();
				r_match.clear();
				local_result.setMatch( r_match );
				continue;
			}
			
			MATCH lookUpMatch;
			lookUpMatch.setMatch( m1->getMatch() );
			map<IDTYPE,IDTYPE> matchMap = local_result.getMatch();
			
			for(map<IDTYPE,IDTYPE>::iterator match_iter = matchMap.begin(); match_iter != matchMap.end(); match_iter++){

				if(!lookUpMatch.hasMatch( match_iter->first )) 
						lookUpMatch.setMatch( match_iter->first, match_iter->second );
// 					else cout << match_iter->first << " ist schon gematcht worden!!! gotoNext" << endl;
			}
// 			cout << lookUpMatch << endl;
			
			list<MATCH> res_list = gotoNextForAll( costTree.getTarget( *e_iter ), &lookUpMatch ); //REKURSION!!!!
			
			if(res_list.empty()) continue;
			
// 			local_res_list.merge(res_list);
			for(list<MATCH>::iterator list_iter = res_list.begin(); list_iter!= res_list.end(); list_iter++){
// 				cout<< "Add Match: "<< *list_iter << endl;
				local_res_list.push_back(*list_iter);
			}
		}
		if(local_res_list.empty()) return local_res_list;

		if(result_list.empty()){ 
			
			result_list = local_res_list;
			continue;
		}	
		list<MATCH> temp_res_list;
		for(list<MATCH>::iterator list_iter = local_res_list.begin(); list_iter!= local_res_list.end(); list_iter++){

			map<IDTYPE,IDTYPE> matchMap = (*list_iter).getMatch();
			
			for(list<MATCH>::iterator result_iter = result_list.begin(); result_iter!=result_list.end(); result_iter++){
				bool good = true;
				MATCH temp_res = *result_iter;
				for(map<IDTYPE,IDTYPE>::iterator match_iter = matchMap.begin(); match_iter != matchMap.end(); match_iter++){

					if(!temp_res.hasMatch( match_iter->first )) temp_res.setMatch( match_iter->first, match_iter->second );
					else{ 
						if(temp_res.getMatch( match_iter->first ) != match_iter->second) good = false;
					}
				}
				
				if(good)temp_res_list.push_back(temp_res);
			}
		}
		result_list = temp_res_list;
	}
// 	cout << "ende" << endl;
// 	result_list.unique();
	return result_list;
}

void SEARCHPLAN::setMatchHandler(ALGORITHM* matchhandler)
{
	this->matchhandler = matchhandler;
}

}

}

}
