/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "second_matchhandler_refinement.h"
#include "fold.h"
#include "find_all_matches.h"
#include "check_pg_structure.h"

namespace algorithms
{

namespace matchhandler
{

SECOND_MATCHHANDLER_REFINEMENT::SECOND_MATCHHANDLER_REFINEMENT()
{
	this->ID = ALGORITHM_SECOND_MATCHHANDLER_REFINEMENT;
	this->name = "second_matchhandler_refinement";
	petrigraph=NULL;
	prule_id=0;
	prule=NULL;
	phi_E_L=NULL;
	match_first=NULL;
	match_second=NULL;
}


SECOND_MATCHHANDLER_REFINEMENT::~SECOND_MATCHHANDLER_REFINEMENT()
{
}

void SECOND_MATCHHANDLER_REFINEMENT::initialize()
{
 checkedFoldings.clear(); 
}

void SECOND_MATCHHANDLER_REFINEMENT::initialize(PETRIGRAPH * i_petrigraph, IDTYPE i_prule_id,
					   RULE * i_prule, MARKING * i_phi_E_L,
					   MATCH * i_match_first)
{
	 assert(i_petrigraph != NULL);
	 assert(i_prule != NULL);
	 assert(i_phi_E_L != NULL);
	 assert(i_match_first != NULL);

	 this->petrigraph = i_petrigraph;
	 this->prule_id = i_prule_id;
	 this->prule = i_prule;
	 this->phi_E_L = i_phi_E_L;
	 this->match_first = i_match_first;
	 
	 if(coloured)
	  ft.initialize((PTNET*)i_petrigraph, storage.getEngine());
	 else
	  ft.initialize((PTNET*)i_petrigraph);
}
      
void SECOND_MATCHHANDLER_REFINEMENT::initialize( PETRIGRAPH * i_petrigraph,
				            IDTYPE i_prule_id, RULE * i_prule,
				            bool * i_canFold, bool * i_canUnfold,
				            MATCH* m1, MATCH* m2, MARKING* i_phi_E_L )
{
         this->initialize(i_petrigraph, i_prule_id, i_prule, i_phi_E_L, m1);
	 if(debug) cout << "PHI2 promulgation: match_second <- " << m2 << "\n";
	 this->match_second = m2;
}

bool SECOND_MATCHHANDLER_REFINEMENT::run(MATCH * match)
{  
	 assert(match != NULL);
	 if(debug)
	 {
	    cout << "\nrunning second matchhandler:\n";
	    cout << "match: " << *match << "\n" << "phi_E_L: " << *phi_E_L << "\n";
	 }

	 /*
	  * phi(E_L) is passed on initialization. Calculate phi2(E_L):
	  */
	 MARKING phi2_E_L;
	 set < IDTYPE > places = this->petrigraph->getPlaces();
	 set < IDTYPE >::iterator placeit;
	 // for(placeit = places.begin(); placeit != places.end(); placeit++)
	 // {
	 //   phi2_E_L.addPlace(*placeit);
	 // }
	 map < IDTYPE, EDGE > E_L = this->prule->getLeft()->getEdges();
	 map < IDTYPE, EDGE >::iterator elit;
	 for(elit = E_L.begin(); elit != E_L.end(); elit++)
	 {
	    TOKEN t;
	    t.setCount(1);
	    phi2_E_L.putToken(match->getMatch(elit->first), t);
	 }
	 if(debug) cout << "phi2_E_L: " << phi2_E_L << "\n";

	 /*
	  * AUGUR I says: "we need two different matchings. phi == phi = phi2?"
	  * Actually, two markings are compared. We consider all tokens as equal.
	  */
	 bool phi_equals_phi2 = true;
	 for(placeit = places.begin(); placeit != places.end(); placeit++)
	 {
	    if(!(phi2_E_L==*phi_E_L))
	    {
	       phi_equals_phi2 = false;
	    }
            
	 }
	 if(phi_equals_phi2)
	 {
	    if(debug) cout << "discarded because phi equals phi2.\n";
	    return false;
	 }

	 /*
	  * let us check the depth of edges
	  */
	 for(elit = E_L.begin(); elit != E_L.end(); elit++)
	 {
	    /*
	     * depth of gedge
	     */
	    IDTYPE gedge = match->getMatch(elit->first);
	    unsigned long dep = this->petrigraph->getDepth(gedge);

	    /*
	     * AUGUR I says: "if phi != phi2 and not depth then it doesn't pass"
	     * "not depth" means that the considered edge has lower than maximal depth
	     */
	    if((this->match_first->getMatch(elit->first) != gedge) && (depth > dep))
	    {
	       if(debug)
	       {
	          cout << "this->depth: " << depth << "\n";
	          cout << "edge in phi1: e" << this->match_first->getMatch(elit->first) << "\n";
	          cout << "edge in phi2: e" << gedge << ":d" << dep << "\n";
		  cout << "\"if phi != phi2 and not depth then it doesn't pass\" (AUGUR I)\n";
	       }
	       return false;
	    }
	 }
	 /*
	  * now consider the depth values of vertices
	  */
	 map < IDTYPE, VERTEX > V_L = this->prule->getLeft()->getVertices();
	 map < IDTYPE, VERTEX >::iterator vlit;
	 for(vlit = V_L.begin(); vlit != V_L.end(); vlit++)
	 {
	    IDTYPE gvertex = match->getMatch(vlit->first);
	    unsigned long dep = this->petrigraph->getDepth(gvertex);
	    if((this->match_first->getMatch(vlit->first) != gvertex) && (depth > dep))
	    {
	       if(debug)
	       {
	          cout << "this->depth: " << depth << "\n";
	          cout << "vertex in phi1: v" << this->match_first->getMatch(vlit->first) << "\n";
	          cout << "vertex in phi2: v" << gvertex << ":d" << dep << "\n";
		  cout << "\"if phi != phi2 and not depth then it doesn't pass\" (AUGUR I)\n";
	       }
	       return false;
	    }
	 }

	 /*
	  * Does a transition t exist such that ...
	  */
	 map < IDTYPE, TRANSITION > transitions = this->petrigraph->getTransitions();
	 map < IDTYPE, TRANSITION >::iterator transit;
	 ALGORITHM* causality_relation = this->load("call_causality");
	 causality_relation->initialize((PTNET *) (this->petrigraph));
	 for(transit = transitions.begin(); transit != transitions.end(); transit++)
	 {
	    /*
	     * ... p_N_i(t) == r and ...
	     */
	    unsigned transition_id;
	    /* istringstream ist;
	    ist.str(transit->second.getLabel());
	    ist >> transition_id; */
	    transition_id = this->petrigraph->getAssociatedRule(transit->first)->getID();
	    if(debug) cout << " {t" << transit->first << "/r" << transition_id << "?}";
	    if(this->prule_id == transition_id)
	    {
	       /*
	        * ... preset of t == phi2_E_L and ...
	        */
	       vector < TCONNECTION > tpreset = this->petrigraph->getPreset(transit->first);
	       vector < TCONNECTION >::iterator tpit;
	       bool equal = true;
	       for(placeit = places.begin(); placeit != places.end(); placeit++)
	       {		       
		  unsigned long weight = 0;
		  for(tpit = tpreset.begin(); tpit != tpreset.end(); tpit++)
		  {		     
		     if(tpit->ID == *placeit)
			weight += tpit->weight;
		  }
		  if(phi2_E_L.getTokenNumber(*placeit) != weight)
		     equal = false;
	       }
	       if(debug) cout << "[[*t" << transit->first << " == phi2_E_L?] " << equal << "]"; 

	       if(equal)
	       {
		  /* 
		   * for all e in phi_E_L: e in preset(t) or t <N e
		   * note that we just checked that *t == phi2_E_L, so the condition 
		   * of "e in *t" transforms to "e is marked in phi2_E_L".
		   */
		  bool e_condition_satisfied = true;
		  for(placeit = places.begin();
		      e_condition_satisfied && (placeit != places.end()); placeit++)
		  {
		     if(this->phi_E_L->IsMarked(*placeit) &&  // for all e in phi_E_L exclude:
			!phi2_E_L.IsMarked(*placeit) )  // e is not marked in phi2_E_L and ...
		     { 
		        /*
			 * ... does not depend on t.
			 */
			bool depends = causality_relation->run(transit->first, *placeit);
			if(debug) cout << "[[t" << transit->first << " <N p" << *placeit << "?] " << depends << "]";
			
			/*
			 * e_condition is satisfied only if *placeit is in tpreset
			 * as we checked before, this is the case if and only if
			 * the place *placeit is marked in phi2_E_L.
			 */
			if(!depends) e_condition_satisfied = false;
		     }
		  }
                  if(debug) cout << "[[\"e condition\" satisfied?] " << e_condition_satisfied << "] ";
		  if(e_condition_satisfied)
		  {
		     *(this->match_second) = *match;
		     if(debug) cout << "\nPHI2 promulgation: *(" << match_second << ") <- " << *match << "\n";
		     
		//////////////////// abstraction refinement part ////////////////////////////
		if(petrigraph->IsReal)
		{     
			// make a copy of petrigraph
			PETRIGRAPH pg_copy = *petrigraph;
			if(checked(prule, match_first, match_second))
			{
				// cout << "\n checked folding step forbiden \n";
				return false;
			}	

			pg_copy.IsReal=false;

			FOLD alg_fold;
			alg_fold.initialize(&pg_copy);
			alg_fold.run(prule, *match_second, *match_first);

			DATA * d = storage.getData(this->name);
			SAMPLE_STORAGE * ss = static_cast<SAMPLE_STORAGE *>(d);
			
			if(ss==NULL || d==NULL)
			{ 
			 cout << "\nSECOND_MATCHHANDLER_REFINEMENT: DATA error\n";
			 exit(1);
			}

			if(checkPetrigraph(&pg_copy, ss))
			{
				return true;
			}
			else
			{
				CHECKED_FOLDINGS cf(prule, match_first, match_second);
				checkedFoldings.push_back(cf);
				
				// cout << "\nfolding step forbiden\n";
				return false;
			}
		}    
		/////////////////////////////////////////////////////////////////////////////
		     
		     return true;
		  }
	       }
	       if(debug) cout << ' ';
	    }
	    else if (debug) cout << "[[wrong rule]] ";
	 }

	 if(debug) cout << "no corresponding instance of rule " << prule_id << " was found. discarding.\n";
	 return false;
}

/** check if this folding variant was already checked */
bool SECOND_MATCHHANDLER_REFINEMENT::checked(RULE * r, MATCH * match_first, MATCH * match_second)
{
	for(list<CHECKED_FOLDINGS>::iterator it = checkedFoldings.begin();
				 it != checkedFoldings.end(); it++)
	{
		if(r->getID()!=it->rid)
			continue;

		if(!(*match_first==it->match_first))
			continue;
			
		if(!(*match_second==it->match_second))  
			continue;	

		return true;	
	 }

	 return false;
}

void SECOND_MATCHHANDLER_REFINEMENT::update(vector<EVENT> events)
{
  // not initialized
  if(petrigraph==NULL)
   return;
  
  // empty event  
  if(events.size()==0)
   return;
        
  if(events.size()>1)
  {
   cout << "\ntoo many events fot SECOND_MATCHHANDLER_REFINEMENT\n";
   exit(1);
  }
  
  if(events[0].type!=MERGE)
  {
   cout << "\nwrong events fot SECOND_MATCHHANDLER_REFINEMENT\n";
   exit(1);
  }
   
  for(list<CHECKED_FOLDINGS>::iterator it = checkedFoldings.begin(); it!=checkedFoldings.end(); it++)
  {
  	map<IDTYPE, IDTYPE> map_first = it->match_first.getMatch();
	map<IDTYPE, IDTYPE> map_second = it->match_second.getMatch();
	
	for(map<IDTYPE, IDTYPE>::iterator itM=map_first.begin(); itM!=map_first.end(); itM++)
	{
		if(events[0].ids.find(itM->second)!=events[0].ids.end())
			map_first[itM->first] = events[0].ids[itM->second];
	}

	for(map<IDTYPE, IDTYPE>::iterator itM=map_second.begin(); itM!=map_second.end(); itM++)
	{
		if(events[0].ids.find(itM->second)!=events[0].ids.end())
			map_second[itM->first] = events[0].ids[itM->second];
	}
	
	it->match_first.setMatch(map_first);
	it->match_second.setMatch(map_second);
  }

}

/** check petrigraph */
bool SECOND_MATCHHANDLER_REFINEMENT::checkPetrigraph(PETRIGRAPH * pg, SAMPLE_STORAGE * ss)
{
  try
  {  
	for(list<SAMPLE>::iterator it = ss->m_list_storage.begin(); it!=ss->m_list_storage.end(); it++)
	{
	 // reachable markings
	 list<Marking> markings;
		
	 // start with initial
	 markings.push_back(pg->getInitialMarking().getUncoloured());
	
	 // transitions
	 vector<string>::iterator itTrans = (*it).trans.begin();
		
	 // tree
	 for(vector<list<HYPERGRAPH_INFO> >::iterator itVec = (*it).tree.begin(); itVec!=(*it).tree.end(); itVec++)
	 {
		 // hypergraphs
		 for(list<HYPERGRAPH_INFO>::iterator itH = (*itVec).begin(); itH!=(*itVec).end(); itH++)
		 {
		   	 // markings
			 for(list<Marking>::iterator itM = markings.begin(); itM!=markings.end(); itM++)
			 {	
				 // current marking
				 map<IDTYPE, IDTYPE> tmp;
				 HYPERGRAPH hgM = pg->getMarkedHypergraph(*itM, &tmp);

				 // matches
        			 TRIVIAL_MATCHHANDLER mh;
				 list<MATCH> matches;
				 
				 FIND_ALL_MATCHES fam;
				 fam.initialize(&hgM);
				 
				 /* Original from Augur I
				 if(automor)
					 matches = hgM.findAllMatchings(itH->hg, mh, true);
				 else
				 {
					 map<IDTYPE, IDTYPE> m = hgM.findMatching(itH->hg, mh);
					 matches.push_back(m);
				 }
				 */
				 
				 matches = fam.run(&(itH->hg), &mh, true);

				 for(list<MATCH>::iterator itMat = matches.begin(); itMat!=matches.end(); itMat++)
				 {
					 if(!checkLabels(&hgM, &(itH->hg), &(*itMat), &(*itH)))
						 return false;
				 }
			 }
		 } 

		 // end of transitions
		 if(itTrans == (*it).trans.end())
			 break;

		 // calculate new markings
		 list<Marking> markings_tmp;

		 // initialize fire transition
		 ft.initialize((PTNET*)pg);

		 // markings
		 for(list<Marking>::iterator itM = markings.begin(); itM!=markings.end(); itM++)
		 {
			  bool found;
 
			 // possible transitions
			 string rule_name = pg->getGrammar()->getRule(*itTrans)->getName();
			 set<IDTYPE> posSet = pg->getTransitions(*itTrans);

			 for(set<IDTYPE>::iterator its = posSet.begin(); its!=posSet.end(); its++)
			 {
				 Marking m;

				if(ft.isEnabledBy(*itM, *its)) 
				{
					m = ft.run(pg->getTransition(*its), *itM);

					markings_tmp.push_back(m);
				} 
			 }
		 }

		 if(markings_tmp.size()==0)
			 break;

		 markings.empty();
		 markings = markings_tmp; 

		 // transition list
		 itTrans++;
	 }

	}
	
	return true;
	
	}
	catch(exception e)
	{
		cout << "Error while checking the sample file" << endl;
		exit(1);
	}
}

/** check if some nodes having the same labels are matches in one */
bool SECOND_MATCHHANDLER_REFINEMENT::checkLabels(HYPERGRAPH * hg1, HYPERGRAPH * hg2, MATCH * match, HYPERGRAPH_INFO * hi)
{
	try
	{
		// h2 -> h1
		map<IDTYPE, VERTEX> nodes1 = hg1->getVertices();
		map<IDTYPE, VERTEX> nodes2 = hg2->getVertices();

		for(map<IDTYPE, VERTEX>::iterator it=nodes1.begin(); it!=nodes1.end(); it++)
		{
       			set<IDTYPE> selected_nodes;
       
			map<IDTYPE, IDTYPE> map_match = match->getMatch();
			for(map<IDTYPE, IDTYPE>::iterator itM = map_match.begin(); itM!=map_match.end(); itM++)
			{
			  if(itM->second == it->first)
            			selected_nodes.insert(itM->first);
			}
			
			if(selected_nodes.size()>1)
			{
			 multiset<unsigned int> mlabels;
			 
			 // check labels
         		 for(set<IDTYPE>::iterator itS = selected_nodes.begin(); itS!=selected_nodes.end(); itS++)
         		 {
				  // labels
				  list<unsigned int> labels = hi->eq_labels[*itS];

				  // add to multiset
				  for(list<unsigned int>::iterator itL = labels.begin(); itL!=labels.end(); itL++)
				  {
							mlabels.insert(*itL);

				  }
			 }

			// check
			for(multiset<unsigned int>::iterator itMultiset = mlabels.begin(); itMultiset != mlabels.end(); itMultiset++)
			if(mlabels.count(*itMultiset) > 1)
				return false;
		
			}
		}

		return true;
	}
	catch(exception e)
	{
		cout << "Error in SECOND_MATCHHANDLER_REFINEMENT\n";
		exit(1);
	}
}

///////// namespaces /////////////

}

}
