/***************************************************************************
 *   Copyright (C) 2006 by root   *
 *   root@Vitali   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "automaton_2_grammar.h"

namespace semilin {

AUTOMATON_2_GRAMMAR::AUTOMATON_2_GRAMMAR()
{
	variables.push_back("A");
	variables.push_back("B");
	variables.push_back("C");
	variables.push_back("D");
	variables.push_back("E");
	variables.push_back("F");
	variables.push_back("G");
	variables.push_back("H");
	variables.push_back("H");
	variables.push_back("I");
	variables.push_back("J");
	variables.push_back("K");
	variables.push_back("L");
	variables.push_back("M");
	variables.push_back("N");
	variables.push_back("O");
	variables.push_back("P");
	variables.push_back("Q");
	variables.push_back("R");
	variables.push_back("T");
	variables.push_back("U");
	variables.push_back("V");
	variables.push_back("W");
	variables.push_back("X");
	variables.push_back("Y");
	variables.push_back("Z");

	constants.push_back('a');
	constants.push_back('b');
	constants.push_back('c');
	constants.push_back('d');
	constants.push_back('e');
	constants.push_back('f');
	constants.push_back('g');
	constants.push_back('h');
	constants.push_back('i');
	constants.push_back('j');
	constants.push_back('k');
	constants.push_back('l');
	constants.push_back('m');
	constants.push_back('n');
	constants.push_back('o');
	constants.push_back('p');
	constants.push_back('q');
	constants.push_back('r');
	constants.push_back('s');
	constants.push_back('t');
	constants.push_back('u');
	constants.push_back('v');
	constants.push_back('w');
	constants.push_back('x');
	constants.push_back('w');
	constants.push_back('z');

	counter_var = 0;
	counter_con = 0;	
}


AUTOMATON_2_GRAMMAR::~AUTOMATON_2_GRAMMAR()
{
}

vector<CFEXPRESSION*> AUTOMATON_2_GRAMMAR::run(FINITE_AUTOMATON * a)
{
	vector<CFEXPRESSION*> result;

	// first check states having no output arcs
	list<IDTYPE> with_output;
	for(map<pair<STATE*, string>, set<STATE*> >::iterator it = a->transitions.begin(); it != a->transitions.end(); it++)
		with_output.push_back(it->first.first->id);

	map<IDTYPE, list<CFEXPRESSION *> > left_right;

	for(map<pair<STATE*, string>, set<STATE*> >::iterator it = a->transitions.begin(); it != a->transitions.end(); it++)
	{
		// set empty list
		if(left_right.find(it->first.first->id) == left_right.end())
		{
			list<CFEXPRESSION *> emptyList;
			left_right[it->first.first->id] = emptyList;
		}

		list<CFEXPRESSION *> right;

		for(set<STATE*>::iterator its = it->second.begin(); its != it->second.end(); its++)
		{
			// check liveness
			bool alive = false;
			for(list<IDTYPE>::iterator itl = with_output.begin(); itl != with_output.end(); itl++)
			{
				if((*its)->id == *itl)
				{
					alive = true;
					break;
				}
			}

			if(alive)
			{
				// right variable
				CFEXPRESSION * var_right = new CFEXPRESSION(VARIABLE, getVariable((*its)->id));
	
				// symbol
				string sch = "";
				sch.append(1, getConstant(it->first.second));
				CFEXPRESSION * symbol = new CFEXPRESSION(STRING, sch);
					
				// concatenation
				CFEXPRESSION * conncat = new CFEXPRESSION(OP_CONCAT, symbol, var_right);
	
				right.push_back(conncat);
			}
		
			// end states
			bool finiteState = false;
			for(set<STATE*>::iterator itf = a->finalStates.begin(); itf != a->finalStates.end(); itf++)
			{
				if((*itf)->id == (*its)->id)
				{
					finiteState = true;
					break;
				}
			}
	
			if(finiteState)
			{
				// symbol
				string sch = "";
				sch.append(1, getConstant(it->first.second));
				CFEXPRESSION * symbol = new CFEXPRESSION(STRING, sch);
					
				right.push_back(symbol);
			}

		}

		list<CFEXPRESSION *> current = left_right[it->first.first->id];
		for(list<CFEXPRESSION *>::iterator itr = right.begin(); itr != right.end(); itr++)
			current.push_back(*itr);

		left_right[it->first.first->id] = current;
	}

	// connect all with plus operation
	for(map<IDTYPE, list<CFEXPRESSION *> >::iterator itl = left_right.begin(); itl != left_right.end(); itl++)
	{
		// left variable
		CFEXPRESSION * var_left = new CFEXPRESSION(VARIABLE, getVariable(itl->first));

		// connect all with plus operation
		list<CFEXPRESSION *> right = itl->second;
		list<CFEXPRESSION *>::iterator itr =right.begin();
		CFEXPRESSION * r = NULL;
		
		if(itr != right.end())
		{
		 r = *itr;
		 itr++;
		}
		
		for(; itr != right.end(); itr++)
		{
			// plus
			CFEXPRESSION * op_plus = new CFEXPRESSION(OP_PLUS, *itr, r);

			// assign
			r = op_plus;
		}
		
		if(r != NULL)
		{
			// equal
			CFEXPRESSION * equal = new CFEXPRESSION(OP_EQUAL, var_left, r);
	
			result.push_back(equal);
		}
	}
	
	// left variable
	CFEXPRESSION * var_left = new CFEXPRESSION(VARIABLE, string("S"));

	list<CFEXPRESSION *> right;

	// initial states
	for(set<STATE*>::iterator iti = a->initialStates.begin(); iti != a->initialStates.end(); iti++)
	{
		// check liveness
		bool alive = false;
		for(list<IDTYPE>::iterator itl = with_output.begin(); itl != with_output.end(); itl++)
		{
			if((*iti)->id == *itl)
			{
				alive = true;
				break;
			}
		}

		if(!alive)
		 continue;
	
		// right variable
		CFEXPRESSION * var_right = new CFEXPRESSION(VARIABLE, getVariable((*iti)->id));

		right.push_back(var_right);
	}

	// check initial states
	if(right.size() == 0)
	{
		cout << "AUTOMATON_2_GRAMMAR Error: Empty initial states" << endl;
		exit(1);
	}

	// connect all with plus operation
	list<CFEXPRESSION *>::iterator itr =right.begin();
	CFEXPRESSION * r = *itr;
	itr++;
	for(; itr != right.end(); itr++)
	{
		// plus
		CFEXPRESSION * op_plus = new CFEXPRESSION(OP_PLUS, *itr, r);

		// assign
		r = op_plus;
	}

	// equal
	CFEXPRESSION * equal = new CFEXPRESSION(OP_EQUAL, var_left, r);

	result.push_back(equal);

	return result;
}


string AUTOMATON_2_GRAMMAR::getVariable(long index)
{
	if(indices_var.find(index) == indices_var.end())
	{
		if(counter_var+1 >= variables.size())
		{
			cout << "AUTOMATON_2_GRAMMAR::Too many variables ... terminated" << endl;
			exit(1);
		}

		indices_var[index] = variables[counter_var];
		counter_var++;
	}
	
	return indices_var[index];
}

const char AUTOMATON_2_GRAMMAR::getConstant(string s)
{
	if(indices_con.find(s) == indices_con.end())
	{
		if(counter_con+1 >= constants.size())
		{
			cout << "AUTOMATON_2_GRAMMAR::Too many constants ... terminated" << endl;
			exit(1);
		}

		indices_con[s] = constants[counter_con];
		counter_con++;
	}
	
	return indices_con[s];
}

map<IDTYPE, string> AUTOMATON_2_GRAMMAR::getVariables()
{
	return indices_var;
}

map<string, char> AUTOMATON_2_GRAMMAR::getConstants()
{
	return indices_con;
}

}

