/***************************************************************************
 *   Copyright (C) 2007 by Salil Joshi   *
 *   hellfeuer@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "cfexpression.h"
#include <sstream>

namespace semilin {

CFEXPRESSION::CFEXPRESSION()
{
	lchild = NULL;
	rchild = NULL;
}

CFEXPRESSION::CFEXPRESSION(CFEXPRESSION * exp)
{
	//recursively duplicate the entire tree;
	if (exp->lchild != NULL) {
		lchild = new CFEXPRESSION(exp->lchild);
	} else {
		lchild = NULL;
	}
	if (exp->rchild != NULL) {
		rchild = new CFEXPRESSION(exp->rchild);
	} else {
		rchild = NULL;
	}
	type = exp->type;
	str_val = exp->str_val;
	num_val = exp->num_val;
}

CFEXPRESSION::CFEXPRESSION(EXPRESSION_TYPE t, string val)
{
	type = t;
	str_val = val;
	lchild = NULL;
	rchild = NULL;
}

CFEXPRESSION::CFEXPRESSION(EXPRESSION_TYPE t, string* val)
{
	type = t;
	str_val = *val;
	lchild = NULL;
	rchild = NULL;
}

CFEXPRESSION::CFEXPRESSION(EXPRESSION_TYPE t, int val)
{
	type = t;
	num_val = val;
	lchild = NULL;
	rchild = NULL;
}

//Special case constructor for convenience
CFEXPRESSION::CFEXPRESSION(EXPRESSION_TYPE t, CFEXPRESSION* l, int n)
{
	//assert(t = OP_POWER);
	type = OP_POWER;
	lchild = l;
	num_val = n;
	rchild = NULL;
}

//Special case constructor for convenience
CFEXPRESSION::CFEXPRESSION(EXPRESSION_TYPE t, CFEXPRESSION* l)
{
	//assert(t = OP_STAR);
	type = OP_STAR;
	lchild = l;
	rchild = NULL;
}

CFEXPRESSION::CFEXPRESSION(EXPRESSION_TYPE t, CFEXPRESSION* l, CFEXPRESSION *r)
{
	type = t;
	lchild = l;
	rchild = r;
}

CFEXPRESSION::~CFEXPRESSION()
{
/*
	//recursively delete the entire tree;
	if (lchild != NULL) {
		delete lchild;
	}
	if (rchild != NULL) {
		delete rchild;
	}
*/
}
CFEXPRESSION* CFEXPRESSION::operator=(CFEXPRESSION * exp)
{
	type = exp->type;
	str_val = exp->str_val;
	num_val = exp->num_val;
	lchild = exp->lchild;
	rchild = exp->rchild;
}


bool CFEXPRESSION::operator==(CFEXPRESSION * exp)
{
	if (type == exp->type) {
		switch (type)
		{
			case STRING:
				return (str_val == exp->str_val);
				break;
			case VARIABLE:
				return (str_val == exp->str_val);
				break;
			case NUMBER:
				return (num_val == exp->num_val);
				break;
			case OP_CONCAT:
				return ((lchild == exp->lchild) && (rchild == exp->rchild));
				break;
			case OP_PLUS:
				return (((lchild == exp->lchild) && (rchild == exp->rchild)) || ((lchild == exp->rchild) && (rchild == exp->lchild)));
				break;
			case OP_STAR:
				return (lchild == exp->lchild);
				break;
			case OP_POWER:
				return ((num_val == exp->num_val) && (lchild == exp->lchild));
		}
	}
}

/*
Returns the expression as a string. basically traverses the tree
*/
string CFEXPRESSION::write_aux()
{
	switch (type)
	{
		case STRING:
			return (str_val);
			break;
		case VARIABLE:
			return (str_val);
			break;
		case NUMBER:
		{
			stringstream ss;
			string s;
			ss<<num_val;
			ss >> s;
			return s;
			break;
		}
		case OP_CONCAT:
			return (lchild->write_aux() + " " + rchild->write_aux());
			break;
		case OP_PLUS:
			return ("(" + lchild->write_aux() + "+" + rchild->write_aux() + ")");
			break;
		case OP_STAR:
			return ("(" + lchild->write_aux() + ")*");
			break;
		case OP_EQUAL:
			return lchild->write_aux() + "=" +rchild->write_aux();
			break;
		case OP_POWER:
		{
			stringstream ss;
			string s;
			ss<<num_val;
			ss >> s;
			string tmp = lchild->write_aux();
			return ("(" + tmp + ")" + s);
			break;
		}
	}
}

/*
Writes out the expression into 'filename'. This can be parsed back with expr_reader
*/
void CFEXPRESSION::write(char * filename)
{
	std::fstream file;
	file.open(filename, std::fstream::out|std::fstream::app);
	file <<write_aux();
	file<<"\n";
	file.close();
	
}

/*
Meant for debugging. Returns the Abstract Syntax Tree in a "pretty" format. Cannot be parsed back in
private function. Use write_ast
*/
string CFEXPRESSION::write_ast_aux(string spaces)
{
	switch (type)
	{
		case STRING:
			return (spaces + str_val+ "\n");
			break;
		case VARIABLE:
			return (spaces + str_val+ "\n");
			break;
		case NUMBER:
		{
			stringstream ss;
			string s;
			ss<<num_val;
			ss >> s;
			return spaces + s+ "\n";
			break;
		}
		case OP_CONCAT:
			return (spaces +"OP_CONCAT"+ "\n"+ lchild->write_ast_aux(spaces + "  ") + rchild->write_ast_aux(spaces + "  "));
			break;
		case OP_PLUS:
			return (spaces +"OP_PLUS"+ "\n"+ lchild->write_ast_aux(spaces + "  ") + rchild->write_ast_aux(spaces + "  "));
			break;
		case OP_STAR:
			return (spaces+"OP_STAR" + "\n" + lchild->write_ast_aux(spaces + "  "));
			break;
		case OP_EQUAL:
			return (spaces +"OP_EQUAL"+ "\n"+ lchild->write_ast_aux(spaces + "  ") + rchild->write_ast_aux(spaces + "  "));
			break;
		case OP_POWER:
		{
			stringstream ss;
			string s;
			ss<<num_val;
			ss >> s;
			string tmp = lchild->write_ast_aux(spaces + "  ");
			return (spaces + s + "\n" + tmp);
			break;
		}
	}
}

/*
Meant for debugging. Writes out the Abstract Syntax Tree in a "pretty" format. Cannot be parsed back in
*/
void CFEXPRESSION::write_ast(char * filename)
{
	std::fstream file;
	file.open(filename, std::fstream::out|std::fstream::trunc);
	file <<write_ast_aux("");
	file<<"\n";
	file.close();
}


}
