/***************************************************************************
 *   Copyright (C) 2007 by Salil Joshi   *
 *   hellfeuer@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef DATA_STRUCTURESEXPRESSION_H
#define DATA_STRUCTURESEXPRESSION_H
#include <string>
#include <iostream>
#include <fstream>
using std::string;
using std::stringstream;

namespace semilin {

/**
	@author hellfeuer <hellfeuer@gmail.com>
*/
/* Enum that represents the type of a node */
enum EXPRESSION_TYPE{
	OP_PLUS = 0,
	OP_CONCAT,
	OP_STAR,
	OP_POWER,	//in case a node is of type OP_POWER, then its numval stores the exponent, and left child has the expression to be exponentiated. i.e. in the case of (a)3, root is OP_POWER with numval = 3 and lchild be of type STRING with str_val = a
 	OP_EQUAL,
	NUMBER,
	STRING,
	VARIABLE,
};

/*Expression data type. It is basically a binary tree with each node having a type from EXPRESSION_TYPE.
In case of unary operator (eg STAR), only lchild is used 
Can be used for any expression, not only for use with commkleene_solve (i.e. it does not assume commutativity)
*/
class CFEXPRESSION{
public:
	CFEXPRESSION* lchild;
	CFEXPRESSION* rchild;
	
	EXPRESSION_TYPE type;
	string str_val;
	int num_val;
	
    CFEXPRESSION();
    CFEXPRESSION(EXPRESSION_TYPE t, CFEXPRESSION* l, CFEXPRESSION* r);
    CFEXPRESSION(CFEXPRESSION * exp);
    CFEXPRESSION(EXPRESSION_TYPE t, string val);
    CFEXPRESSION(EXPRESSION_TYPE t, string* val);
    CFEXPRESSION(EXPRESSION_TYPE t, int val);
    CFEXPRESSION(EXPRESSION_TYPE t, CFEXPRESSION* l, int n);
    CFEXPRESSION(EXPRESSION_TYPE t, CFEXPRESSION* l);
    
    bool operator==(CFEXPRESSION * exp);
    CFEXPRESSION* operator=(CFEXPRESSION * exp);
    void write(char * filename); //Writes out the expression into 'filename'. This can be parsed back with expr_reader
    void write_ast(char * filename); //Meant for debugging. Writes out the Abstract Syntax Tree in a "pretty" format. Cannot be parsed back in

    ~CFEXPRESSION();
	private:
		string write_aux();
		string write_ast_aux(string spaces);

};

}

using namespace semilin;

#endif
