/***************************************************************************
 *   Copyright (C) 2007 by Salil Joshi   *
 *   hellfeuer@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "commkleene_eval.h"

namespace semilin {

COMMKLEENE_EVAL::COMMKLEENE_EVAL()
{
}

void COMMKLEENE_EVAL::initialize(string var, CFEXPRESSION* val)
{
	variable = var;
	value = val;
}

/*basically traverses the tree and substitutes 'value' for 'variable'
It is long because it also does some simplifications along the way.
More simplification could be added
*/
void COMMKLEENE_EVAL::run(CFEXPRESSION* exp)
{
	if (exp ==NULL) {
		return;
	}
	
	run(exp->lchild);
	run(exp->rchild);
	
	switch (exp->type)
	{
		case OP_PLUS:
			if (exp->lchild == exp->rchild) {// this comparison is expensive
				delete exp->lchild;
				CFEXPRESSION* tmp = exp->rchild;
				*exp=tmp;
				tmp->lchild = NULL;
				tmp->rchild = NULL;
				delete tmp;
			} else if ((exp->lchild->type == STRING) && (exp->lchild->str_val == "/0")) {
				delete exp->lchild;
				CFEXPRESSION* tmp = exp->rchild;
				*exp =tmp;
				tmp->lchild = NULL;
				tmp->rchild = NULL;
				delete tmp;
			} else if ((exp->rchild->type == STRING) && (exp->rchild->str_val == "/0")) {
				delete exp->rchild;
				CFEXPRESSION* tmp = exp->lchild;
				*exp=tmp;
				tmp->lchild = NULL;
				tmp->rchild = NULL;
				delete tmp;
			}
			break;
		case OP_CONCAT:
			if ((exp->lchild->type == STRING) && (exp->rchild->type == STRING)) {
				if ((exp->lchild->str_val == "/0") || (exp->rchild->str_val == "~")) {
					delete exp->rchild;
					CFEXPRESSION* tmp = exp->lchild;
					*exp = tmp;
					tmp->lchild = NULL;
					tmp->rchild = NULL;
					delete tmp;
				} else if ((exp->rchild->str_val == "/0") || (exp->lchild->str_val == "~")) {
					delete exp->lchild;
					CFEXPRESSION* tmp = exp->rchild;
					*exp= exp->rchild;
					tmp->lchild = NULL;
					tmp->rchild = NULL;
					delete tmp;
				} else {
					exp->type = STRING;
					exp->str_val = exp->lchild->str_val + exp->rchild->str_val;
					delete exp->lchild;
					delete exp->rchild;
					exp->lchild = NULL;
					exp->rchild = NULL;
				}
			} else if (exp->lchild->type == STRING) {
				if (exp->lchild->str_val == "~") {
					delete exp->lchild;
					CFEXPRESSION* tmp = exp->rchild;
					*exp = tmp;
					tmp->lchild = NULL;
					tmp->rchild = NULL;
					delete tmp;
				} else if (exp->lchild->str_val == "/0") {
					delete exp->rchild;
					CFEXPRESSION* tmp = exp->lchild;
					*exp = tmp;
					tmp->lchild = NULL;
					tmp->rchild = NULL;
					delete tmp;
				}
			} else if (exp->rchild->type == STRING) {
				if (exp->rchild->str_val == "~") {
					delete exp->rchild;
					CFEXPRESSION* tmp = exp->lchild;
					*exp = tmp;
					tmp->lchild = NULL;
					tmp->rchild = NULL;
					delete tmp;
				} else if (exp->rchild->str_val == "/0") {
					delete exp->lchild;
					CFEXPRESSION* tmp = exp->rchild;
					*exp = tmp;
					tmp->lchild = NULL;
					tmp->rchild = NULL;
					delete tmp;
				}
			}
			break;
		case OP_POWER:
			if ((exp->lchild->type == STRING) && ((exp->lchild->str_val == "/0") ||(exp->lchild->str_val == "~"))){
				CFEXPRESSION* tmp = exp->lchild;
				*exp = tmp;
				delete tmp;
			} else if (exp->lchild->type == OP_STAR) {
				CFEXPRESSION* tmp = exp->lchild;
				*exp = tmp;
				tmp->lchild = NULL;
				tmp->rchild = NULL;
				delete tmp;
			} else if (exp->lchild->type == OP_POWER) {
				CFEXPRESSION* tmp = exp->lchild;
				exp->num_val *= exp->lchild->num_val;
				exp->lchild = exp->lchild->lchild;
				tmp->lchild = NULL;
				tmp->rchild = NULL;
				delete tmp;
			}
			break;
		case OP_STAR:
			if ((exp->lchild->type == STRING) && ((exp->lchild->str_val == "/0") ||(exp->lchild->str_val == "~"))){
				CFEXPRESSION* tmp = exp->lchild;
				*exp = tmp;
				exp->str_val = "~";
				delete tmp;
			}
			break;
		case VARIABLE:
			if (exp->str_val == variable) {
				*exp = new CFEXPRESSION(value); //this copy is essential
			}
			break;
			
		default:
			break;
	}
}

COMMKLEENE_EVAL::~COMMKLEENE_EVAL()
{
}


}
