/***************************************************************************
 *   Copyright (C) 2007 by Salil Joshi   *
 *   hellfeuer@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "commkleene_solve.h"

namespace semilin {

COMMKLEENE_SOLVE::COMMKLEENE_SOLVE()
{
}

CFEXPRESSION* COMMKLEENE_SOLVE::run(vector<CFEXPRESSION*> &equations)
{
	vector<string> variables = vector<string>(equations.size());
	vector<CFEXPRESSION*> functions = vector<CFEXPRESSION*>(equations.size());
	for (int i = 0; i != equations.size(); i++)
	{
		variables[i] = equations[i]->lchild->str_val;	//equation must be of the form X = something
		functions[i] = equations[i]->rchild;
	}
	return run_aux(variables, functions);
}

/*Direct implementation of method in Hopkins-Kozen paper*/
CFEXPRESSION* COMMKLEENE_SOLVE::run_aux(vector<string> &variables, vector<CFEXPRESSION*> &functions)
{
	COMMKLEENE_DIFFERENTIAL d;
	COMMKLEENE_EVAL eval;
	CFEXPRESSION* fdash;
	CFEXPRESSION* result;
	CFEXPRESSION * zero = new CFEXPRESSION(STRING, "/0");
	
	for (int i = 0; i <variables.size(); i++)
	{
		if (i>0) {
			eval.initialize(variables[i-1], result);
			for (int j = i; j <variables.size(); j++)
			{
				eval.run(functions[j]);
			}
			delete result;
		}
		d.initialize(variables[i]);
		fdash = d.run(functions[i]);
		
		eval.initialize(variables[i], zero);
		eval.run(functions[i]);    //functions[i] now contains functions[i](0)
		
		eval.initialize(variables[i], functions[i]);
		eval.run(fdash);    //fdash now contains fdash(functions[i](0))
		
		result = new CFEXPRESSION(OP_CONCAT, new CFEXPRESSION(OP_STAR, fdash), functions[i]);
	}
	delete zero;
	
	return result;
}

COMMKLEENE_SOLVE::~COMMKLEENE_SOLVE()
{
}


}

