/***************************************************************************
 *   Copyright (C) 2007 by Salil Joshi   *
 *   hellfeuer@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef ALGORITHMS_EXPRESSIONCOMMKLEENE_SOLVE_H
#define ALGORITHMS_EXPRESSIONCOMMKLEENE_SOLVE_H

#include <vector>

#include "cfexpression.h"
#include "commkleene_differential.h"
#include "commkleene_eval.h"

using std::vector;

namespace semilin {

/**
	@author hellfeuer <hellfeuer@gmail.com>
	Solves a set of equations in a Commutative Kleene Algebra and returns the result
	Method used: ref "Parikh's Theorem in Commutative Kleen Algebra" by Mark Hopkins and Dexter Kozen (Jan 1999)
	Using Gaussian Elimination (see example on page 9)
*/
class COMMKLEENE_SOLVE{
private:
	CFEXPRESSION* run_aux(vector<string> &variables, vector<CFEXPRESSION*> &functions);
public:
	COMMKLEENE_SOLVE();
	
	void initialize();
	/*
	Note: Equations must have only a single variable on LHS
	If you have equations in the following form:
	X=exp1
	Y=exp2
	...
	and you wish to solve for X then the vector you pass to function 'run' must have the "X = exp1" as the LAST element
	*/
	CFEXPRESSION* run(vector<CFEXPRESSION*> &equations);

	~COMMKLEENE_SOLVE();

};

}

#endif
