/***************************************************************************
 *   Copyright (C) 2007 by Salil Joshi   *
 *   hellfeuer@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "re_semilin_converter.h"

namespace semilin {

RE_SEMILIN_CONVERTER::RE_SEMILIN_CONVERTER()
{
}

void RE_SEMILIN_CONVERTER::initialize(unsigned int alphabetsize, map<char, unsigned int>* chartoindex)
{
	n = alphabetsize;
	charmap = chartoindex;
}

SEMILINEAR<slmarking>* RE_SEMILIN_CONVERTER::run(CFEXPRESSION* re)
{
	SEMILINEAR<slmarking>* s1;
	SEMILINEAR<slmarking>* s2;
	SEMILINEAR<slmarking>* result;
	switch (re->type)
	{
		case OP_PLUS:
			/* Result is the union of semilinear sets of lchild and rchild */
			s1 = run(re->lchild);
			s2 = run(re->rchild);
			result= SEMILINEAR<slmarking>::set_union(s1, s2);
			delete s1;
			delete s2;
			break;
		case OP_CONCAT:
			/* Result is the sum of semilinear sets of lchild and rchild */
			s1 = run(re->lchild);
			s2 = run(re->rchild);
			result= *s1 + *s2;
			delete s1;
			delete s2;
			break;
		case OP_STAR:
			/* Result is the STAR of semilinear set*/
			result = run(re->lchild);
			result->closure();
			break;
		case OP_POWER:
		{
			s1 =run(re->lchild);
			result = s1;
			for (int i = 1; i<re->num_val; i++)
			{
				result = *result + *s1;
			}
			break;
		}
		case STRING:
			if (re->str_val == "/0") {
				//represents the empty set
				return new SEMILINEAR<slmarking>(n);
			} else if (re->str_val == "~") {
				slmarking a = slmarking(n);	//the empty marking
				return new SEMILINEAR<slmarking>(a);
			} else {
				/*count the number of times each character occurs in the string and generate the correct marking
				i.e. "aaba" and "abaa" will both give [3, 1] if a is mapped to 0 and b to 1
				*/
				slmarking m = slmarking(n);
				for (int i = 0; i <re->str_val.size(); i++)
				{
					unsigned int t = (*charmap)[re->str_val[i]];
					m[t]++;
				}
				return new SEMILINEAR<slmarking>(m);
			}
			break;
	//default: ERROR. The above EXPRESSION_TYPES are the only ones that are supported
	}
 	return result;
}

RE_SEMILIN_CONVERTER::~RE_SEMILIN_CONVERTER()
{
}


}


