/***************************************************************************
 *   Copyright (C) 2007 by Salil Joshi   *
 *   hellfeuer@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <iostream>
using std::cout;

namespace data_structures {

template <typename T> SEMILINEAR<T>::SEMILINEAR(unsigned int N)
{
	n = N;
}

template <typename T> SEMILINEAR<T>::SEMILINEAR(T &element)
{
	n = element.getN();
	lsets.insert(new LINEARSET<T>(element));
}



/*******DO NOT USE*******
For the sake of completeness, this function should exist.
However, at present LINEARSET does not implement the hasElement function, since this functionality is not needed for calculating the Parikh image.
Implementing LINEARSET::hasElement would require proving the existence of a solution in the naturals, of a certain set of linear equations
If LINEARSET were to implement the hasElement correctly, the following would work fine
*/
template <typename T> bool SEMILINEAR<T>::hasElement(T* element)
{
	typename set<LINEARSET<T>*, Compare>::iterator i;
	
	for (i=lsets.begin(); i != lsets.end(); i++)
	{
		if (i->hasElement(element)) {
			return true;
		}
	}
	return false;
}



template <typename T> SEMILINEAR<T>* SEMILINEAR<T>::operator+(SEMILINEAR<T> & a)
{
	SEMILINEAR<T>* s = new SEMILINEAR<T>(n);
	typename set<LINEARSET<T>*, Compare>::iterator i;
	typename set<LINEARSET<T>*, Compare>::iterator j;
	for (i = lsets.begin(); i != lsets.end(); i++)
	{
		for (j = a.lsets.begin(); j != a.lsets.end(); j++)
		{
			s->lsets.insert(*(*i)+*(*j));
		}
	}
	return s;
}



template <typename T> void SEMILINEAR<T>::closure()
{
	bool zero = false;
	for (typename set<LINEARSET<T>*, Compare>::iterator i = lsets.begin(); i != lsets.end(); i++)
	{
		if ((*i)->isZero())
		{
			zero = true;
		} else {
			(*i)->closure();
		}
	}
	if (zero == false) {
		T tmp = T(n);
		lsets.insert(new LINEARSET<T>(tmp));
	}
	set<LINEARSET<T> *, Compare> oldlsets = set<LINEARSET<T>*, Compare>(lsets);
	
	for (typename set<LINEARSET<T>*, Compare>::iterator i = oldlsets.begin(); i != oldlsets.end(); i++)
	{
		if ((*i)->isZero()) {
			continue;
		}
		
		typename set<LINEARSET<T>*, Compare>::iterator j=i;
		for (++j;j != oldlsets.end(); j++)
		{
			
			LINEARSET<T>* l = *i;
			for (typename set<LINEARSET<T>*, Compare>::iterator k=j; k != oldlsets.end(); k++)
			{
				if ((*k)->isZero()) {
					continue;
				}
				l = *l + *(*k);			
				lsets.insert(l);

			}
		}
	}
}

    
template <typename T> SEMILINEAR<T>* SEMILINEAR<T>::set_union(SEMILINEAR* a, SEMILINEAR* b)
{
	SEMILINEAR<T>* t = new SEMILINEAR<T>(a->n);
	t->lsets = set<LINEARSET<T> *, Compare>(a->lsets);
	typename set<LINEARSET<T>*, Compare>::iterator i;
	for (i = b->lsets.begin(); i != b->lsets.end(); i++)
	{
		t->lsets.insert(*i);
	}
	return t;
}

template <typename T> void SEMILINEAR<T>::print()
{
	cout<<"{ ";
	for (typename set<LINEARSET<T>*, Compare>::iterator i = lsets.begin(); i != lsets.end(); i++)
	{
		cout<<" {";
		(*i)->print();
		cout<<"}\n ";
	}
	cout<<"} ";
}
	
template <typename T> SEMILINEAR<T>::~SEMILINEAR()
{
}


}
