/***************************************************************************
 *   Copyright (C) 2007 by Salil Joshi   *
 *   hellfeuer@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef DATA_STRUCTURESSEMILINEAR_H
#define DATA_STRUCTURESSEMILINEAR_H
#include <list>
#include "linearset.h"

namespace data_structures {

/**
	@author hellfeuer,,, <hellfeuer@hellfeuer-ubuntu>
	Semilinear set contains a set of linear set. Both these classes are templated
	The template parameter class must support the following operations:
	1) +=
	2) <
	3) isZero() method
	4) copy constructor
	5) constructor that takes one unsigned int argument representing size of alphabet, and creates a "zero element"
	6) print() method
*/
template<typename T>
class SEMILINEAR{
private:
	
	class Compare
	{
		public:
		bool operator()(LINEARSET<T>* a, LINEARSET<T>* b)
		{
			return (*a < *b);
		}
	};
public:
	unsigned int n; //no. of symbols
    SEMILINEAR(unsigned int N);
    SEMILINEAR(T &element);
    bool hasElement(T* element);
    set<LINEARSET<T> *, Compare> lsets;
    void closure();
    void print();
    static SEMILINEAR<T>* set_union(SEMILINEAR<T>* a, SEMILINEAR<T>* b);
    SEMILINEAR<T>* operator+(SEMILINEAR<T> & a);
    ~SEMILINEAR();
    

};

}
#include "semilinear.cpp"

#endif
