
%{
   #include "../pointer_lang/abstree.h"
   #include "parser.tab.h"
   #include <stdio.h>
   #include <string.h>
   int row = 1;
   char *charconst;
%}

letter			[A-Za-z]
digit			[0-9]
id			{letter}({letter}|{digit}|_)*
number			{digit}+
anythingstr		[^\"\n\t]
string	         	"\""{anythingstr}*"\""

%%
\n	{row++;}
[ \t]+	{;/**/}
\-\-[^\n]*	{;/* ignore comments and whitespaces */}

"+"			{return PLUS;}
"-"			{return MINUS;}
"*"			{return MULT;}
"/"			{return DIV;}
"%"			{return MOD;}

":="			{return ASGN;}
"("			{return LEFT;}
")"			{return RIGHT;}
";"			{return SEPARATOR;}
","			{return COMMA;}
":"			{return COLON;}
[.\^]			{return yytext[0];}
'^'			{return yytext[0];}

"="			{return EQ;}
"<"			{return LT;}
">"			{return GT;}
"<="			{return LE;}
">="			{return GE;}
"!="			{return NE;}


TRUE		 	{return TRUE;}
FALSE		 	{return FALSE;}

NOT			{return NOT;}
AND			{return AND;}
OR			{return OR;}
IS-NULL			{return ISNULL;}

IF			{return IF;}
THEN			{return THEN;}
ELSE			{return ELSE;}
FI			{return FI;}
WHILE			{return WHILE;}
DO			{return DO;}
OD			{return OD;}
TO			{return TO;}

NEW			{return NEW;}
DELETE			{return DELETE;}

NULL			{return NULLPOINTER;}
SKIP			{return SKIP;}

VAR			{return VAR;}
TYPE			{return TYPE;}
IS			{return IS;}

INTEGER			{return INTEGER;}
STRING			{return STRING;}
POINTER			{return POINTER;}
RECORD			{return RECORD;}

{id}			{
                                if (strlen(yytext) > 25 ) 
                            	{ 
   					printf("length of identifier > 25\n"); 
					printf("Program will be aborted\n");
					exit(1); 
                            	}
                         	strcpy(yylval.name,yytext);return ID;
			}

{number}		{sscanf(yytext,"%d",&yylval.integer); return INT_CONST;}
{string}		{
				charconst=(char*)(malloc(sizeof(char)*(strlen(yytext)+1)));
				strncpy(charconst,yytext,strlen(yytext)+1);
				yylval.string = charconst;
			 	return STR_CONST;
			}
.			{ printf("lexical error in line %d\n", row); exit(1);}

%%

