%{
#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "../pointer_lang/abstree.h"
#include "../pointer_lang/symboltable.h"

extern int yylex();
extern int yydebug;


extern int row;

extern "C"
{
   int yywrap(){
       return 1; 
   }
   extern int yyparse();
   extern void cleanup() {
      fprintf(stderr, "syntax error in line %d\n",row);	
      fprintf(stderr, "Program will be aborted\n");
      exit(1);
   }
   extern void yyerror(const char* s)
   {
      //fprintf(stderr, "%s\n", s);
   }
}
NODE *abstree;

%}

%token PLUS MINUS MULT DIV
%token ASGN LEFT RIGHT SEPARATOR COMMA COLON
%token '^' '.'

%token EQ LT GT LE GE NE
%token TRUE FALSE
%token NOT AND OR ISNULL
%token IF THEN ELSE FI WHILE DO OD TO

%token NEW DELETE NULLPOINTER
%token VAR TYPE IS
%token STR_CONST SKIP

%token <name>ID
%token <integer>INT_CONST
%token <string>STR_CONST

%token  INTEGER STRING POINTER RECORD 



%left OR
%left AND
%nonassoc NOT

%left PLUS MINUS
%left MOD MULT DIV
%left UMINUS

%start program

%union  {
   int  integer ;
   char*  string;
   char name[MAX_ID];
   NODE* Node;
}

%type <Node> program statements const exp ACTION statementsA stm decl vdecl first last path type tvdecl tdecllist bexp


%%

program		:  decl statements {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type = T_PROGRAM;
			$$->pointer=(PROGRAM*)malloc(sizeof(PROGRAM));
			((PROGRAM*)$$->pointer)->declarations=$1;
			((PROGRAM*)$$->pointer)->body=$2;
			abstree=$$;
		}
		|  error {cleanup();}
;

decl		:  vdecl SEPARATOR decl {
			
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_DECLLIST;
			$$->pointer=(DECLLIST*)malloc(sizeof(DECLLIST));
			((DECLLIST*)$$->pointer)->head=$1;
			((DECLLIST*)$$->pointer)->tail=$3;
		}
		|  {$$=NULL;}
;

vdecl		:  VAR ID COLON type {

			$$=(NODE*)malloc(sizeof(NODE));
			$$->type=T_VARDECL;
			$$->pointer=(VARDECL*)malloc(sizeof(VARDECL));
			strcpy(((VARDECL*)$$->pointer)->name,$2);	
			((VARDECL*)$$->pointer)->type=$4;
			((VARDECL*)$$->pointer)->programcounter=row;
			
		}
		|  TYPE ID IS RECORD LEFT tvdecl tdecllist RIGHT {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type=T_TYPEDECL;
			$$->pointer=(TYPEDECL*)malloc(sizeof(TYPEDECL));
			((TYPEDECL*)$$->pointer)->head=$6;
			((TYPEDECL*)$$->pointer)->tail=$7;
			strcpy(((TYPEDECL*)$$->pointer)->name,$2);
			((TYPEDECL*)$$->pointer)->programcounter=row;
		}
		|  TYPE ID  {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type=T_TYPEDECLNULL;
			$$->pointer=(TYPEDECLNULL*)malloc(sizeof(TYPEDECLNULL));
		        strcpy(((TYPEDECL*)$$->pointer)->name,$2);
			((TYPEDECL*)$$->pointer)->programcounter=row;	
		}
;

tvdecl		:  ID COLON type {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type=T_TVARDECL;
			$$->pointer=NULL;
			$$->pointer=(TVARDECL*)malloc(sizeof(TVARDECL));
			strcpy(((TVARDECL*)$$->pointer)->name,$1);	
			((TVARDECL*)$$->pointer)->type=$3;
			((TVARDECL*)$$->pointer)->programcounter=row;
		}
;

tdecllist	:  COMMA tvdecl tdecllist  {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type=T_TYPEDECLLIST;
			$$->pointer=(TYPEDECLLIST*)malloc(sizeof(TYPEDECLLIST));
			((TYPEDECLLIST*)$$->pointer)->head=$2;
			((TYPEDECLLIST*)$$->pointer)->tail=$3;
			((TYPEDECLLIST*)$$->pointer)->programcounter=row;
		}
		|  {$$=NULL;}
;

type		:  INTEGER {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_INTEGER;
			$$->pointer = NULL;	
		}
		|  STRING {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_STRING;
			$$->pointer = NULL;
		}

		|  POINTER TO RECORD ID {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_PTOREC;
			$$->pointer = (PTOREC*)malloc(sizeof(PTOREC));
			strcpy(((PTOREC*)$$->pointer)->name,$4);
			((PTOREC*)$$->pointer)->programcounter=row;
		}
;

statements	:  stm SEPARATOR statementsA {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_STMLIST;
			$$->pointer=(STMLIST*)malloc(sizeof(STMLIST));
			((STMLIST*)$$->pointer)->head=$1;
			((STMLIST*)$$->pointer)->tail=$3;
		}  
                |  {$$=NULL;}
;

statementsA	:  stm SEPARATOR statementsA {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_STMLIST;
			$$->pointer=(STMLIST*)malloc(sizeof(STMLIST));
			((STMLIST*)$$->pointer)->head=$1;
			((STMLIST*)$$->pointer)->tail=$3;
		}
		|  {$$=NULL;}
;

ACTION	:	ID {	
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type = T_IDENTIFIER;
			$$->pointer=(IDENTIFIER*)malloc(sizeof(IDENTIFIER));
			strcpy(((IDENTIFIER*)$$->pointer)->name,$1);	
			((IDENTIFIER*)$$->pointer)->symboltableentry=NULL;
			((IDENTIFIER*)$$->pointer)->programcounter=row;

		} 
;

stm	:	ACTION ASGN exp {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_ASSIGNER;
			$$->pointer=(ASSIGNER*)malloc(sizeof(ASSIGNER));
			((ASSIGNER*)$$->pointer)->right=$3;
			((ASSIGNER*)$$->pointer)->left=$1;
			((ASSIGNER*)$$->pointer)->programcounter=row;
		}
	|	path ASGN exp {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_ASSIGNER;
			$$->pointer=(ASSIGNER*)malloc(sizeof(ASSIGNER));
			((ASSIGNER*)$$->pointer)->right=$3;
			((ASSIGNER*)$$->pointer)->left=$1;
			((ASSIGNER*)$$->pointer)->programcounter=row;
		}
	|	NEW LEFT ACTION RIGHT {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_NEW;
			$$->pointer=(ALLOC*)malloc(sizeof(ALLOC));
			((ALLOC*)$$->pointer)->path=$3;
			((ALLOC*)$$->pointer)->programcounter=row;
		}
	|	NEW LEFT path RIGHT {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_NEW;
			$$->pointer=(ALLOC*)malloc(sizeof(ALLOC));
			((ALLOC*)$$->pointer)->path=$3;
			((ALLOC*)$$->pointer)->programcounter=row;
		}
	|	DELETE LEFT ACTION RIGHT {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_DELETE;
			$$->pointer=(DELETE_*)malloc(sizeof(DELETE_));
			((DELETE_*)$$->pointer)->path=$3;
			((DELETE_*)$$->pointer)->programcounter=row;
		}
	|	DELETE LEFT path RIGHT {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_DELETE;
			$$->pointer=(DELETE_*)malloc(sizeof(DELETE_));
			((DELETE_*)$$->pointer)->path=$3;
			((DELETE_*)$$->pointer)->programcounter=row;
		}
	|	IF bexp THEN statements ELSE statements FI {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_IFSTM;
			$$->pointer=(IFSTM*)malloc(sizeof(IFSTM));
			((IFSTM*)$$->pointer)->bexp=$2;
			((IFSTM*)$$->pointer)->ifstm=$4;
			((IFSTM*)$$->pointer)->elsestm=$6;
			((IFSTM*)$$->pointer)->programcounter=row;
		}
	|	WHILE bexp DO statements OD {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_WHILE;
			$$->pointer=(WHILESTM*)malloc(sizeof(WHILESTM));
			((WHILESTM*)$$->pointer)->bexp=$2;
			((WHILESTM*)$$->pointer)->dostm=$4;
		        ((WHILESTM*)$$->pointer)->programcounter=row;
		}
	|	SKIP {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_SKIP;
			$$->pointer=NULL;
		}
;

bexp	:	TRUE {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_TRUE;
			$$->pointer=NULL;
		}
	|	FALSE {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_FALSE;
			$$->pointer=NULL;
		}
	|	exp EQ exp {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_EQUAL;
			$$->pointer=(EQUAL*)malloc(sizeof(EQUAL));
			((EQUAL*)$$->pointer)->right=$3;
			((EQUAL*)$$->pointer)->left=$1;
		        ((EQUAL*)$$->pointer)->programcounter=row;
		}
	|	exp LT exp {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_SMALLER;
			$$->pointer=(SMALLER*)malloc(sizeof(SMALLER));
			((SMALLER*)$$->pointer)->right=$3;
			((SMALLER*)$$->pointer)->left=$1;
			((SMALLER*)$$->pointer)->programcounter=row;
		}
	|	exp GT exp {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_GREATER;
			$$->pointer=(GREATER*)malloc(sizeof(GREATER));
			((GREATER*)$$->pointer)->right=$3;
			((GREATER*)$$->pointer)->left=$1;
			((GREATER*)$$->pointer)->programcounter=row;
		}
	|	exp LE exp {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_SMALLEREQUAL;
			$$->pointer=(SMALLEREQUAL*)malloc(sizeof(SMALLEREQUAL));
			((SMALLEREQUAL*)$$->pointer)->right=$3;
			((SMALLEREQUAL*)$$->pointer)->left=$1;
			((SMALLEREQUAL*)$$->pointer)->programcounter=row;
		}
	| 	exp GE exp {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_GREATEREQUAL;
			$$->pointer=(GREATEREQUAL*)malloc(sizeof(GREATEREQUAL));
			((GREATEREQUAL*)$$->pointer)->right=$3;
			((GREATEREQUAL*)$$->pointer)->left=$1;
			((GREATEREQUAL*)$$->pointer)->programcounter=row;
		}
	|	exp NE exp {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_NOTEQUAL;
			$$->pointer=(NOTEQUAL*)malloc(sizeof(NOTEQUAL));
			((NOTEQUAL*)$$->pointer)->right=$3;
			((NOTEQUAL*)$$->pointer)->left=$1;
			((NOTEQUAL*)$$->pointer)->programcounter=row;	
		}
	|	bexp AND bexp {
                        if ($1->type==T_FALSE || $3->type==T_FALSE) {
			   $$=(NODE*)malloc(sizeof(NODE));
			   $$->type=T_FALSE;
               		   $$->pointer=NULL;
		           free($1);free($3);
                        }
                        else if ($1->type==T_TRUE && $3->type==T_TRUE) {
			   $$=(NODE*)malloc(sizeof(NODE));
			   $$->type=T_TRUE;
			   $$->pointer=NULL;
		           free($1);free($3);
                        }
                        else if ($1->type==T_TRUE) {
                           $$=$3;
                           free($1);
                        }
                        else if ($3->type==T_TRUE) {
                           $$=$1;
                           free($3);
                        }
                        else {
			   $$=(NODE*)malloc(sizeof(NODE));
			   $$->type =T_AND;
			   $$->pointer=(AND_*)malloc(sizeof(AND_));
			   ((AND_*)$$->pointer)->right=$3;
			   ((AND_*)$$->pointer)->left=$1;
			   ((AND_*)$$->pointer)->programcounter=row;
                        }
		}
	|	bexp OR bexp {
                        if ($1->type==T_TRUE || $3->type==T_TRUE) {
			   $$=(NODE*)malloc(sizeof(NODE));
			   $$->type=T_TRUE;
			   $$->pointer=NULL;
		           free($1);free($3);
                        }
                        else if ($1->type==T_FALSE && $3->type==T_FALSE) {
			   $$=(NODE*)malloc(sizeof(NODE));
			   $$->type=T_FALSE;
			   $$->pointer=NULL;
		           free($1);free($3);
                        }
                        else if ($1->type==T_FALSE) {
                           $$=$3;
                           free($1);
                        }
                        else if ($3->type==T_FALSE) {
                           $$=$1;
                           free($3);
                        }
                        else {
			   $$=(NODE*)malloc(sizeof(NODE));
			   $$->type =T_OR;
			   $$->pointer=(OR_*)malloc(sizeof(OR_));
			   ((OR_*)$$->pointer)->right=$3;
			   ((OR_*)$$->pointer)->left=$1;
			   ((OR_*)$$->pointer)->programcounter=row;
                        }
		}
	|	NOT bexp {
                        if ($2->type==T_TRUE) {
                           $2->type=T_FALSE;
			   $$=$2;
                        }
                        else if ($2->type==T_FALSE) {
                           $2->type=T_TRUE;
			   $$=$2;
                        }
                        else {
			   $$=(NODE*)malloc(sizeof(NODE));
			   $$->type =T_NOT;
			   $$->pointer=(NOT_*)malloc(sizeof(NOT_));
			   ((NOT_*)$$->pointer)->left=$2;
			   ((NOT_*)$$->pointer)->programcounter=row;
                        }

		}
	|	LEFT bexp RIGHT {$$=$2;}
;

exp	:	const {$$=$1;}
	|	ID {	
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type = T_IDENTIFIER;
			$$->pointer=(IDENTIFIER*)malloc(sizeof(IDENTIFIER));
			strcpy(((IDENTIFIER*)$$->pointer)->name,$1);
			((IDENTIFIER*)$$->pointer)->programcounter=row;
		}
	|	path { $$=$1;}
	| 	exp PLUS exp {
			$$=(NODE*)malloc(sizeof(NODE));
			if ($1->type == T_INTEGERCONST && $3->type == T_INTEGERCONST) {	
			    $$->type = T_INTEGERCONST;
			    $$->pointer=(INTEGERCONST*)malloc(sizeof(INTEGERCONST));
			    ((INTEGERCONST*)$$->pointer)->value = ((INTEGERCONST*)$1->pointer)->value + ((INTEGERCONST*)$3->pointer)->value;
		            free($1);free($3);
                        }
			else {
			    $$->type =T_PLUSOP;
			    $$->pointer=(PLUSOP*)malloc(sizeof(PLUSOP));
			    ((PLUSOP*)$$->pointer)->right=$3;
			    ((PLUSOP*)$$->pointer)->left=$1;
			    ((PLUSOP*)$$->pointer)->programcounter=row;
			}
		}
	|	exp MINUS exp {
			$$=(NODE*)malloc(sizeof(NODE));
			if ($1->type == T_INTEGERCONST && $3->type == T_INTEGERCONST) {
			    $$->type = T_INTEGERCONST;
			    $$->pointer=(INTEGERCONST*)malloc(sizeof(INTEGERCONST));
			    ((INTEGERCONST*)$$->pointer)->value = ((INTEGERCONST*)$1->pointer)->value - ((INTEGERCONST*)$3->pointer)->value;
		            free($1);free($3);
			}
			else {
			   $$->type =T_MINUSOP;
			   $$->pointer=(MINUSOP*)malloc(sizeof(MINUSOP));
			   ((MINUSOP*)$$->pointer)->right=$3;
			   ((MINUSOP*)$$->pointer)->left=$1;
			   ((MINUSOP*)$$->pointer)->programcounter=row;
			}	           
			
		
		}
	|	exp MULT exp {
			$$=(NODE*)malloc(sizeof(NODE));
			if ($1->type == T_INTEGERCONST && $3->type == T_INTEGERCONST) {	
			    $$->type = T_INTEGERCONST;
			    $$->pointer=(INTEGERCONST*)malloc(sizeof(INTEGERCONST));
			    ((INTEGERCONST*)$$->pointer)->value = ((INTEGERCONST*)$1->pointer)->value * ((INTEGERCONST*)$3->pointer)->value;
		            free($1);free($3);			   
			}
			else {
			   $$->type =T_MULTOP;
			   $$->pointer=(MULTOP*)malloc(sizeof(MULTOP));
			   ((MULTOP*)$$->pointer)->right=$3;
			   ((MULTOP*)$$->pointer)->left=$1;
			   ((MULTOP*)$$->pointer)->programcounter=row;
			}
			
		}
	|	exp DIV exp {
			$$=(NODE*)malloc(sizeof(NODE));
			if ($1->type == T_INTEGERCONST && $3->type == T_INTEGERCONST) {
			    $$->type = T_INTEGERCONST;
			    $$->pointer=(INTEGERCONST*)malloc(sizeof(INTEGERCONST));
			    ((INTEGERCONST*)$$->pointer)->value = ((INTEGERCONST*)$1->pointer)->value / ((INTEGERCONST*)$3->pointer)->value;
		            free($1);free($3);			
			}
			else {
			   $$->type =T_DIVOP;
			   $$->pointer=(DIVOP*)malloc(sizeof(DIVOP));
			   ((DIVOP*)$$->pointer)->right=$3;
			   ((DIVOP*)$$->pointer)->left=$1;
			   ((DIVOP*)$$->pointer)->programcounter=row;
			}	
		}
	|	exp MOD exp {
			$$=(NODE*)malloc(sizeof(NODE));
			if ($1->type == T_INTEGERCONST && $3->type == T_INTEGERCONST) {
			    $$->type = T_INTEGERCONST;
			    $$->pointer=(INTEGERCONST*)malloc(sizeof(INTEGERCONST));
			    ((INTEGERCONST*)$$->pointer)->value = ((INTEGERCONST*)$1->pointer)->value % ((INTEGERCONST*)$3->pointer)->value;
		            free($1);free($3);			
			}
			else {
			   $$->type =T_MODOP;
			   $$->pointer=(MODOP*)malloc(sizeof(MODOP));
			   ((MODOP*)$$->pointer)->right=$3;
			   ((MODOP*)$$->pointer)->left=$1;
			   ((MODOP*)$$->pointer)->programcounter=row;
			}	
		}
	|	MINUS exp %prec UMINUS {
			  $$=(NODE*)malloc(sizeof(NODE));
			  $$=(NODE*)malloc(sizeof(NODE));
			  $$->type =T_NEGATE;
			  $$->pointer=(NEGATE*)malloc(sizeof(NEGATE));
			  ((NEGATE*)$$->pointer)->expstm=$2;
 			  ((NEGATE*)$$->pointer)->programcounter=row;
			

		}
	|	NULLPOINTER {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_NULL;
			$$->pointer=NULL;
		}
	|	LEFT exp RIGHT {$$=$2;}
;

path	:	first '.' last {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_PATH;
			$$->pointer=(PATH*)malloc(sizeof(PATH));
			((PATH*)$$->pointer)->last=$3;
			((PATH*)$$->pointer)->first=$1;
			((PATH*)$$->pointer)->programcounter=row;	
		}
;

last	:	first '.' last {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type =T_PATH;
			$$->pointer=(PATH*)malloc(sizeof(PATH));
			((PATH*)$$->pointer)->last=$3;
			((PATH*)$$->pointer)->first=$1;
			((PATH*)$$->pointer)->programcounter=row;
		}
;

first	:	ID {	
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type = T_IDENTIFIER;
			$$->pointer=(IDENTIFIER*)malloc(sizeof(IDENTIFIER));
			strcpy(((IDENTIFIER*)$$->pointer)->name,$1);	
			((IDENTIFIER*)$$->pointer)->symboltableentry=NULL;
			((IDENTIFIER*)$$->pointer)->programcounter=row;
		}
;

first	:	ID '^' {	
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type = T_IDENTIFIER;
			$$->pointer=(IDENTIFIER*)malloc(sizeof(IDENTIFIER));
			strcpy(((IDENTIFIER*)$$->pointer)->name,$1);	
			((IDENTIFIER*)$$->pointer)->symboltableentry=NULL;
			((IDENTIFIER*)$$->pointer)->programcounter=row;
		}
;

last	:	ID {	
			$$=(NODE*)malloc(sizeof(NODE));
			$$->type = T_IDENTIFIER;
			$$->pointer=(IDENTIFIER*)malloc(sizeof(IDENTIFIER));
			strcpy(((IDENTIFIER*)$$->pointer)->name,$1);	
			((IDENTIFIER*)$$->pointer)->symboltableentry=NULL;
			((IDENTIFIER*)$$->pointer)->programcounter=row;
		}
;

const	:	INT_CONST { 
			$$=(NODE*)malloc(sizeof(NODE));
			$$->pointer=(INTEGERCONST*)malloc(sizeof(INTEGERCONST));
			$$->type = T_INTEGERCONST;
			((INTEGERCONST*)$$->pointer)->value = $1;
		}
	|	STR_CONST {
			$$=(NODE*)malloc(sizeof(NODE));
			$$->pointer=(STRCONST*)malloc(sizeof(STRCONST));
			$$->type = T_STRCONST;		
			((STRCONST*)$$->pointer)->value = $1;
		}
;


%%

