/*****************************************************************************
 * error.h: error header
 *****************************************************************************
 * Copyright (C) 2004 Universitaet Stuttgart
 * $Id: error.h,v 1.1.2.1 2004/08/30 15:18:22 koziouvi Exp $
 *
 * Authors: Nicolas Relange <relangns@studi.informatik.uni-stuttgart.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#ifndef ERROR_H
#define ERROR_H

#include <iostream>

//! General class that manages all the error / information messages
/*!
 *  
 */
class Error
{
 public:
  //! Constructor
  Error();

  //! Destructor
  virtual ~Error();

  //! Display an information (standard output) if cond <= verbosity
  /*! \param a Type of the displayed information
   *  \param b Information to display
   *  \param cond Condition on which the information is displayed
   */
  template<typename A,typename B>
    static void info(const A a, const B& b, unsigned int cond = 0);
  
  //! Display an error (error output) if cond <= verbosity
  /*! This method works like Error::info
   *  \param a Type of the displayed information
   *  \param b Information to display
   *  \param cond Condition on which the informaiton is displayed
   */
  template<typename A,typename B>
    static void error(const A a, const B b, unsigned int cond = 0);

  //! Level of verbosity
  static int verbosity;

 protected:

 private:
};



/////////////////////////////////////////////
// implementation of the template functions
/////////////////////////////////////////////

template<typename A,typename B>
void
Error::info(const A a, const B& b, unsigned int cond)
{
  if(cond <= verbosity)
    std::cerr << "info: " << a << ": " << b << "\n";
}

template<typename A,typename B>
void
Error::error(const A a, const B b, unsigned int cond)
{
  if(cond <= verbosity)
    std::cerr << "error: " << a << ": " << b << "\n";
}

#endif // ERROR_H
