/*****************************************************************************
 * gxlreader.h: GXL (Graph eXchange Language) reader (header)
 *****************************************************************************
 * Copyright (C) 2004 Universitaet Stuttgart
 * $Id: gxlreader.h,v 1.1.2.1 2004/08/30 15:18:22 koziouvi Exp $
 *
 * Authors: Nicolas Relange <relangns@studi.informatik.uni-stuttgart.de>
 *          Ingo Walther
 *          Lars Heinemann
 *          Barbara Knig <koenigba@fmi.uni-stuttgart.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#ifndef GXLREADER_H
#define GXLREADER_H

#include <libxml/xmlmemory.h>

template<typename T,typename L>
class Hypergraph;

//! Reading of a GXL (Graph eXchange Language) file
/*! reads a file in XML GXL (Graph eXchange Language)
 *  format and generates a Hypergraph object from it.
 */
template<typename T,typename L>
class GXLReader
{
 public:
  //! Constructor
  /*! \param filename Name of the XML file to parse
   */
  GXLReader(char *filename);

  //! Destructor
  ~GXLReader();

  //! Parse the XML file
  /*! \return The hypergraph read by the parser
   */
  Hypergraph<T,L>* read();

 private:
  Hypergraph<T,L>* graph; //!< hypergraph
  xmlDocPtr doc; //!< libxml2 document structure

  //! Parse a vertex
  /*! \param node Initial node of the vertex in the XML file
   *  The result is directly stored in the private variable graph
   */
  void parseVertex(xmlNodePtr node);

  //! Parse an edge / a relation
  /*! \param node Initial node of the relation in the XML file
   *  The result is directly stored in the private variable graph
   */
  void parseEdge(xmlNodePtr node);

  //! Convert a const xmlChar* to an unsigned int
  /*! "_123" or "123" gives 123 for example
   */
  T xmlCharToT(const xmlChar*);
};

#endif // GXLREADER_H
