%{
/*****************************************************************************
 * reglex.ll: lex file for regular expressions
 *****************************************************************************
 * Copyright (C) 2004 Universitaet Stuttgart
 * $Id: reglex.ll,v 1.1.2.1 2004/08/30 15:18:24 koziouvi Exp $
 *
 * Authors: Nicolas Relange <relangns@studi.informatik.uni-stuttgart.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#include "common.h"

#include <stdio.h>
#include <libxml/parser.h>

#include "regparse.h"
#include "mincover.cc"

Hypergraph<idType,labelType> *spHg;
MinCover<idType,labelType> *spMc;
unsigned mode;

%}

delim	[ \t\n]
white	{delim}+
char	[A-Za-z_]
digit	[0-9]
motif	\'{char}*\'
number	{digit}*
dcd	{digit}{char}{digit}
dmd	{digit}{motif}{digit}
nmn	\'{number}{char}*{number}\'
eps	~|\"\"

%%
{white}	/* eat up whitespace */
{char}	{
	spMc = new MinCover<idType,labelType>((labelType) yytext, spHg, mode);
	return MOTIF;
	}
{motif}	{
	spMc = new MinCover<idType,labelType>((labelType) xmlStrsub((xmlChar*) yytext,1,xmlStrlen((xmlChar*) yytext)-2), spHg, mode);
	return MOTIF;
	}
{dcd}	{
	spMc = new MinCover<idType,labelType>((labelType) xmlStrsub((xmlChar*) yytext,1,xmlStrlen((xmlChar*) yytext)-2), yytext[0]-48, yytext[xmlStrlen((xmlChar*) yytext)-1]-48, spHg);
	return MOTIF;
	}
{dmd}	{
	spMc = new MinCover<idType,labelType>((labelType) xmlStrsub((xmlChar*) yytext,2,xmlStrlen((xmlChar*) yytext)-4), yytext[0]-48, yytext[xmlStrlen((xmlChar*) yytext)-1]-48, spHg);
	return MOTIF;
	}
{nmn}	{
	unsigned begin, length, i, source, target, part;
	begin = length = source = target = part = 0;
	for(i=1; yytext[i] != '\''; ++i)
	   {
	   if( part == 0 && yytext[i] > 57 )
		{
		++part;
		begin = i;
		}
	   if( part == 1 && yytext[i] <= 57 )
	 	++part;
	   switch(part)
	   	{
		case 0:
		  {
		  source *= 10;
		  source += yytext[i]-48;
		  }
		break;
		case 1:
		  ++length;
		break;
		case 2:
		  {
		  target *= 10;
		  target += yytext[i]-48;
		  }
		break;
	   	}
	   }
	spMc = new MinCover<idType,labelType>((labelType) xmlStrsub((xmlChar*) yytext,begin,length), source, target, spHg);
	return MOTIF;
	}
{eps}	return EPSILON;
.	return (int) yytext[0];

%%

