%{
/*****************************************************************************
 * regparse.yy: yacc grammar for regular expressions
 *****************************************************************************
 * Copyright (C) 2004 Universitaet Stuttgart
 * $Id: regparse.yy,v 1.1.2.1 2004/08/30 15:18:24 koziouvi Exp $
 *
 * Authors: Nicolas Relange <relangns@studi.informatik.uni-stuttgart.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#include "mincover.cc"
#include "common.h"

#include <stdio.h>
#include "error.h"

#define YYSTYPE MinCover<idType,labelType>*

extern "C"
{
	int yyparse(void);
	int yywrap()
	{
		return 1;
	}
	void yyerror(const char* str)
	{
		fprintf(stderr,"error: %s\n",str);
	}
}

int yylex(void);

extern Hypergraph<idType,labelType> *spHg;
extern MinCover<idType,labelType> *spMc;
bool lolaOut;

%}

%token MOTIF EPSILON

%%
final	: expr
	{
		Error::info("result",*$1,1);
		if(lolaOut)
			$1->lola();
		else
		{
			if(Error::verbosity >= 1)
				$1->lola(std::cerr);
			$1->markings();
		}
	}
	;
expr	: expr '+' term
	{
		Error::info("regparse","PLUS",4);
		$$ = $1;
		(*$$) += (*$3);
		delete $3;
	}
	| term
	{
	}
	;
term	: term factor
	{
		Error::info("regparse","CONCAT",4);
		$$ = $1;
		(*$$) *= (*$2);
		delete $2;
	}
	| factor
	{
	}
	;
factor	: '(' expr ')'
	{
		Error::info("regparse","()",4);
		$$ = $2;
	}
	| factor '*'
	{
		Error::info("regparse","STAR",4);
		$$ = $1;
		$$->starAss();
	}
	| MOTIF
	{
		Error::info("regparse","MOTIF",4);
		$$ = new MinCover<idType,labelType>(*spMc);
		delete spMc;
	}
	| EPSILON
	{
		Error::info("regparse","EPSILON",4);
		$$ = new MinCover<idType,labelType>( spHg );
	}
	;

%%

