/*****************************************************************************
 * sponge.cpp: main source code
 *****************************************************************************
 * Copyright (C) 2004 Universitaet Stuttgart
 * $Id: sponge.cc,v 1.1.2.1 2004/08/30 15:18:24 koziouvi Exp $
 *
 * Authors: Nicolas Relange <relangns@studi.informatik.uni-stuttgart.de>
 *          Ingo Walther
 *          Lars Heinemann
 *          Barbara Knig <koenigba@fmi.uni-stuttgart.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"
#include "error.h"
#include "hypergraph.cc"
#include "gxlreader.cc"
#include "mincover.cc"

#include <iostream>
#include <cstdlib>
#include <cstring>

extern "C"
{
  int yyparse(void);
}

extern Hypergraph<idType,labelType>* spHg;
extern unsigned mode;
extern bool lolaOut;

int Error::verbosity = 0;

int
main(int argc, char**argv)
{
  // version
  const char* version = "1.0.0";

  // usage
  const char* usage = "Usage: sponge [-q|-d] [-m] [-l] graph_file.xml\n       sponge --help\n       sponge --version\n\nOptions:\n\n  -q quiet   quiet mode, disable verbose output\n  -d debug   debug mode, show many debugging information\n  -m mode1   mode 1 for hypergraphs: from all i to j>i (default: mode 0, from 0 to j > 0)\n  -f formula  formula output (default: markings output)\n  --help     display this small help\n  --version  display the name and the version of this software  \n\n";
  
  // parse args:
  int i;

  // default mode: 0: from vertice 0 to all vertices >= 1
  mode = 0;

  // default output: markings
  lolaOut = false;

  Error::verbosity = 1; // Default: 1

  for(i=1; i<argc; i++)
    {
      if(!strcmp(argv[i],"-q")) // quiet mode
	Error::verbosity = 0;
      else if(!strcmp(argv[i],"-d")) // debug output
	Error::verbosity = 4;
      else if(!strcmp(argv[i],"-m")) // mode 1
	mode = 1;
      else if(!strcmp(argv[i],"-f")) // lola/formula output
	lolaOut = true;
      else if(!strcmp(argv[i],"--version")) // version
	{
	  std::cout << "sponge " << version << "\n";
	  return 1;
	}
      else if(!strcmp(argv[i],"--help")) //help
	{
	  std::cout << usage;
	  return 1;
	}
      else if(*argv[i] == '-')
	{
	  std::cout << "Unknown option: " << argv[i] << "\n" << usage;
	  return 1;
	}
      else
	break;
    }
  if(i > argc-1)
    {
      std::cout << usage;
      return 1;
    }
  
  char* infile = argv[i];

  GXLReader<idType,labelType> reader(infile);
  spHg = reader.read();

  Error::info("Hypergraph",*spHg,1);

  yyparse();

  Error::info("sponge","end",3);

  std::cout << std::endl;

  return 0;
}

