#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "globals.h"
#include "petrigraph.h"
#include "findmatch.h"
#include "find_all_matches.h"

int main(int argc, char** argv)
{
   PETRIGRAPH petrigraph;
   petrigraph.HYPERGRAPH::setID(1);
   petrigraph.HYPERGRAPH::setName("petrigraph");
   petrigraph.addVertex(VERTEX(), 1);
   petrigraph.addVertex(VERTEX(), 2);
   petrigraph.addVertex(VERTEX(), 3);
   petrigraph.addVertex(VERTEX(), 4);
   petrigraph.addVertex(VERTEX(), 5);
   
   EDGE eVI;
   list<IDTYPE> eVI_cv;
   eVI.setLabel("E");
   eVI_cv.push_back(1);
   eVI_cv.push_back(3);
   petrigraph.addPlace(6, eVI, eVI_cv);
   EDGE eVII;
   list<IDTYPE> eVII_cv;
   eVII.setLabel("G");
   eVII_cv.push_back(2);
   eVII_cv.push_back(3);
   eVII_cv.push_back(1);
   petrigraph.addPlace(7, eVII, eVII_cv);
   EDGE eVIII;
   list<IDTYPE> eVIII_cv;
   eVIII.setLabel("F");
   eVIII_cv.push_back(3);
   eVIII_cv.push_back(2);
   petrigraph.addPlace(8, eVIII, eVIII_cv);
   EDGE eIX;
   list<IDTYPE> eIX_cv;
   eIX.setLabel("E");
   eIX_cv.push_back(4);
   eIX_cv.push_back(3);
   petrigraph.addPlace(9, eIX, eIX_cv);
   EDGE eX;
   list<IDTYPE> eX_cv;
   eX.setLabel("F");
   eX_cv.push_back(4);
   eX_cv.push_back(5);
   petrigraph.addPlace(10, eX, eX_cv);
   EDGE eXI;
   list<IDTYPE> eXI_cv;
   eXI.setLabel("E");
   eXI_cv.push_back(5);
   eXI_cv.push_back(1);
   petrigraph.addPlace(11, eXI, eXI_cv);
   
   petrigraph.addTransition(1023, TRANSITION());
   petrigraph.addTConnection( 1, 1023);
   petrigraph.addTConnection(1023,  1);
   petrigraph.addTConnection( 3, 1023);
   petrigraph.addTConnection(1023,  3);
   petrigraph.addTConnection(1023,  6);
   petrigraph.addTConnection(1023,  2);
   petrigraph.addTConnection(1023,  8);
   petrigraph.addTConnection( 9, 1023);
   petrigraph.addTConnection( 4, 1023);
   petrigraph.addTConnection(10, 1023);
   petrigraph.addTConnection( 5, 1023);
   petrigraph.addTConnection(11, 1023);
   
   MARKING initial_marking;
   initial_marking.putToken( 4, TOKEN());
   initial_marking.putToken( 5, TOKEN());
   initial_marking.putToken( 9, TOKEN());
   initial_marking.putToken(10, TOKEN());
   initial_marking.putToken(11, TOKEN());
   petrigraph.setInitialMarking(initial_marking);
   
   HYPERGRAPH pattern;
   pattern.addVertex(VERTEX(), 1);
   pattern.addVertex(VERTEX(), 2);
   pattern.addVertex(VERTEX(), 3);
   EDGE p_eIV;
   list<IDTYPE> p_eIV_cv;
   p_eIV.setLabel("E");
   p_eIV_cv.push_back(1);
   p_eIV_cv.push_back(2);
   pattern.addEdge(p_eIV, 4, p_eIV_cv);
   EDGE p_eV;
   list<IDTYPE> p_eV_cv;
   p_eV.setLabel("F");
   p_eV_cv.push_back(2);
   p_eV_cv.push_back(3);
   pattern.addEdge(p_eV, 5, p_eV_cv);
   
   /*
    * output test objects
    */
   cout << "pattern:\n" << pattern << "\n\n";
   cout << "petrigraph:\n" << petrigraph << "\n\n";
   
   /*
    * test FINDMATCH with a TRIVIAL_MATCHHANDLER
    */
   cout << "applying FINDMATCH with a TRIVIAL_MATCHHANDLER.\n";
   ALGORITHM* findmatch = new FINDMATCH();
   findmatch->initialize((HYPERGRAPH*) &petrigraph);
   ALGORITHM* trivial = new TRIVIAL_MATCHHANDLER();
   bool found = false;
   MATCH* match = findmatch->run(&pattern, trivial, &found);
   if(found) cout << "match found:\n" << *match << "\n";
   else cout << "no match found.\n";
   
   /*
    * test FIND_ALL_MATCHES with a TRIVIAL_MATCHHANDLER
    */
   cout << "\napplying FIND_ALL_MATCHES with a TRIVIAL_MATCHHANDLER.\n";
   ALGORITHM* findall = new FIND_ALL_MATCHES();
   findall->initialize((HYPERGRAPH*) &petrigraph);
   list<MATCH> matches = findall->run(&pattern, trivial, false);
   cout << matches.size() << " match(es) found:\n";
   for(list<MATCH>::iterator matchit = matches.begin(); matchit != matches.end(); matchit++)
   {
      cout << *matchit << "\n";
   }
   
   return 0;
}
