/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#ifndef ALGORITHM_UNFOLD_CPP
#define ALGORITHM_UNFOLD_CPP

#include "unfold.h"
#include "cov_graph.h"
#include "algorithm_storage.h"

namespace algorithms
{
   namespace petrigraph
   {
      UNFOLD::UNFOLD()
      {
         this->ID = ALGORITHM_UNFOLD;
         this->name = "unfold";
	 this->petrigraph = NULL;
      }
      UNFOLD::~UNFOLD() {}

      void UNFOLD::initialize(PETRIGRAPH* pg)
      {
         this->petrigraph = pg;
      }

      void UNFOLD::run(RULE* r, MATCH phi)
      {
         if(debug)
	 {
	    cout << "\n\n---\nunfolding:\n" << *r << "\n---\n\n";
	 }
 
         /*
	  * first, add P(t, r) to this petrigraph as defined between Defs. 9 and 10
	  */
	 HYPERGRAPH* left = r->getLeft();
	 HYPERGRAPH* right = r->getRight();
	 map<IDTYPE, VERTEX>::iterator vertit;
	 map<IDTYPE, VERTEX> vertices_left = left->getVertices();
	 map<IDTYPE, VERTEX> vertices_right = right->getVertices();
	 MATCH mu_t;
	 unsigned long lastUsedID = this->petrigraph->getLastUsedID();
	 if(debug) cout << "last used ID: " << lastUsedID << "\n";
	 
	 /* 
	  * add vertices from r to this petrigraph
	  */
	 for(vertit = vertices_left.begin(); vertit != vertices_left.end(); vertit++)
	 {
 	    IDTYPE vertexid = vertit->first;
	    if(vertexid > lastUsedID)
	    {
	       lastUsedID = vertexid;
               this->petrigraph->setLastUsedID(lastUsedID);
	    }
	    else if(this->petrigraph->hasVertex(vertexid) ||
		 this->petrigraph->hasEdge(vertexid) || this->petrigraph->hasTransition(vertexid))
	    {
	       lastUsedID++;
	       this->petrigraph->setLastUsedID(lastUsedID);
	       vertexid = lastUsedID;
	    }
	    VERTEX nvertex = vertit->second;
	    nvertex.setID(vertexid);

	    // depth, do not make this for isolated nodes
	    // they are not in the MATCH -> by getMatch() the severe error will be obtained
	    unsigned td = 0;
	    if(left->getConnectedEdges(vertit->first).size()!=0)
	     td = this->petrigraph->getDepth(phi.getMatch(vertit->first));
	    if(debug) cout << "inserting (left) vertex " << vertexid << " at depth " << td << "\n";

	    this->petrigraph->addVertex(nvertex, vertexid);
	    mu_t.setMatch(vertit->first, vertexid);
	    this->petrigraph->setDepth(vertexid, td);
	 }
 
	 map<IDTYPE, EDGE> edges_left = left->getEdges();
	 map<IDTYPE, EDGE> edges_right = right->getEdges();
	 map<IDTYPE, EDGE>::iterator edgeit;
	 vector<TCONNECTION> preset;
	 vector<TCONNECTION> postset;
	 unsigned long max_depth = 0;
 
	 for(edgeit = edges_left.begin(); edgeit != edges_left.end(); edgeit++)
	 {
	    IDTYPE edgeid = edgeit->first;
	    if(edgeid > lastUsedID)
	    {
	       lastUsedID = edgeid;
               this->petrigraph->setLastUsedID(lastUsedID);
	    }
	    else if(this->petrigraph->hasVertex(edgeid) ||
		 this->petrigraph->hasEdge(edgeid) ||
		 this->petrigraph->hasTransition(edgeid))
	    {
	       lastUsedID++;
	       this->petrigraph->setLastUsedID(lastUsedID);
	       edgeid = lastUsedID;
	    }
	    list<IDTYPE> ocV = left->getConnectedVertices(edgeit->first);
	    list<IDTYPE>::iterator connectit;
	    list<IDTYPE> ncV;
	    for(connectit = ocV.begin(); connectit != ocV.end(); connectit++)
	    {
	       ncV.push_back(mu_t.getMatch(*connectit));
	    }
	    EDGE nedge = edgeit->second;
	    nedge.setID(edgeid);
	    this->petrigraph->addEdge(nedge, edgeid, ncV);
	    mu_t.setMatch(edgeit->first, edgeid);
 
	    unsigned long dep = petrigraph->getDepth(phi.getMatch(edgeit->first));
	    if(debug) cout << "inserting (left) edge " << edgeid << " at depth " << dep << "\n";
	    this->petrigraph->setDepth(edgeid, dep);
	    if(dep > max_depth) max_depth = dep;
 
	    TCONNECTION ttc;
	    ttc.ID = edgeid;
	    ttc.weight = 1;
	    ttc.attributes = edgeit->second.getAttributes();
	    preset.push_back(ttc);
	 } 
 
	 for(vertit = vertices_right.begin(); vertit != vertices_right.end(); vertit++)
	 {
	    if(!mu_t.hasMatch(vertit->first))
	    {
	       IDTYPE vertexid = vertit->first;
	       if(vertexid > lastUsedID)
	       {
   	          lastUsedID = vertexid;
                  this->petrigraph->setLastUsedID(lastUsedID);
	       }
	       else if(this->petrigraph->hasVertex(vertexid) ||
		 this->petrigraph->hasEdge(vertexid) || this->petrigraph->hasTransition(vertexid))
	       {
 	          lastUsedID++;
	          this->petrigraph->setLastUsedID(lastUsedID);
	          vertexid = lastUsedID;
	       }
	       VERTEX nvertex = vertit->second;
	       nvertex.setID(vertexid);
	       if(debug) cout << "inserting (right) vertex " << vertexid << " at depth " << max_depth + 1 << "\n";
	       this->petrigraph->addVertex(nvertex, vertexid);
	       mu_t.setMatch(vertit->first, vertexid);
	       this->petrigraph->setDepth(vertexid, max_depth + 1);
	    }
 	 }
	 for(edgeit = edges_right.begin(); edgeit != edges_right.end(); edgeit++)
	 {
	    IDTYPE edgeid = edgeit->first;
	    if(edgeid > lastUsedID)
	    {
	       lastUsedID = edgeid;
               this->petrigraph->setLastUsedID(lastUsedID);
	    }
	    else if(this->petrigraph->hasVertex(edgeid) ||
		 this->petrigraph->hasEdge(edgeid) || this->petrigraph->hasTransition(edgeid))
	    {
 	       lastUsedID++;
	       this->petrigraph->setLastUsedID(lastUsedID);
	       edgeid = lastUsedID;
	    }
	    list<IDTYPE> ocV = right->getConnectedVertices(edgeit->first);
	    list<IDTYPE>::iterator connectit;
	    list<IDTYPE> ncV;
	    for(connectit = ocV.begin(); connectit != ocV.end(); connectit++)
	    {
	       ncV.push_back(mu_t.getMatch(*connectit));
	    }
	    EDGE nedge = edgeit->second;
	    nedge.setID(edgeid);
	    if(debug) cout << "inserting (right) edge " << edgeid << " at depth " << max_depth+1 << "\n";
	    this->petrigraph->addEdge(nedge, edgeid, ncV);
	    mu_t.setMatch(edgeit->first, edgeid);
	    this->petrigraph->setDepth(edgeid, max_depth + 1);
 
	    TCONNECTION ttc;
	    ttc.ID = edgeid;
	    ttc.weight = 1;
	    ttc.attributes = edgeit->second.getAttributes();
	    postset.push_back(ttc);
         }
	 
	 /*
	  * create the new transition
	  */
	 lastUsedID++;
	 this->petrigraph->setLastUsedID(lastUsedID);
	 IDTYPE tid = lastUsedID;
	 TRANSITION tt = TRANSITION();
	 tt.setID(tid);
	 tt.setLabel(r->getName());
	 tt.setGuard(r->getGuard());
	 this->petrigraph->addTransition(tid, tt);
	 this->petrigraph->setPreset(tid, preset);
	 this->petrigraph->setPostset(tid, postset);
	 this->petrigraph->setAssociatedRule(tid, r->getID());
	 this->petrigraph->setMu(tid, mu_t);
	 
	 // set data for coverability graph (for incremental coverability)
	 TRANSITION_INFO * ti = new TRANSITION_INFO();
	 ti->transition = tid;
	 storage.setData("coverability", ti);

	 /*
	  * merge the left hand side of the rule with its match in the (old) Petri graph
	  */
	 set<pair<IDTYPE, IDTYPE> > emptymap;
	 set<pair<IDTYPE, IDTYPE> > eunion;
	 set<pair<IDTYPE, IDTYPE> > vunion;
	 
	 for(edgeit = edges_left.begin(); edgeit != edges_left.end(); edgeit++)
	 {
	    eunion.insert(pair<IDTYPE, IDTYPE>(phi.getMatch(edgeit->first),mu_t.getMatch(edgeit->first)));
	 }

	 // interface
	 MATCH interface = r->getInterface();
	 map<IDTYPE, IDTYPE> interface_map = interface.getMatch();
	 for(map<IDTYPE, IDTYPE>::iterator itm = interface_map.begin(); itm != interface_map.end(); itm++)
	 {
		if(left->hasEdge(itm->first))
		 eunion.insert(pair<IDTYPE, IDTYPE>(itm->first,itm->second));
		else
		 vunion.insert(pair<IDTYPE, IDTYPE>(itm->first,itm->second));
	 }

	 ALGORITHM* merge = this->load("call_merge");
	 merge->initialize(this->petrigraph);
	 merge->run(vunion, eunion, emptymap);
      }

   };
};
using namespace algorithms;
using namespace algorithms::petrigraph;

#endif
