/***************************************************************************                                                                       *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "vis_hypergraph.h"

#include <string.h>

namespace algorithms {

namespace writers {

VIS_HYPERGRAPH::VIS_HYPERGRAPH()
{
}


VIS_HYPERGRAPH::~VIS_HYPERGRAPH()
{
}

void VIS_HYPERGRAPH::initialize(char * filename)
{
    int n = strlen(filename);

    this->filename = new char[n+1];
    strcpy(this->filename,filename);

    nodeLabels.clear();
}

void VIS_HYPERGRAPH::initialize(char * filename, ALGORITHM * p_engine)
{
    int n = strlen(filename);

    this->filename = new char[n+1];
    strcpy(this->filename,filename);

    nodeLabels.clear();

    this->engine = p_engine;
}

void VIS_HYPERGRAPH::setNodeLabel(map<int, int> nL)
{
	nodeLabels = nL;
}

void VIS_HYPERGRAPH::run(HYPERGRAPH & h, PRINT_TYPE pt)
{
	if(pt != FULL && engine==NULL)
	{
		cout << "\nerror: VIS_HYPERGRAPH - no attribute engine\n";
		exit(1);
	}
		
	string dotFile;
	char * pch;
	char filename_copy[strlen(filename)+1];
	strcpy(filename_copy, filename);
	pch = strtok (filename_copy,".");
		
	while (pch!=NULL && strcmp(pch, "ps")!=0)
  	{
		if(dotFile.size()!=0)
			dotFile += string(".");
		dotFile += (string)pch;
    		pch = strtok (NULL, ".");		
  	}

	dotFile += string(".dot");

	writeDotFile(h, (char*)dotFile.c_str(), pt);

	callNeato((char*)dotFile.c_str(), filename);
}

/** write a dot file for this hypergraph */
void VIS_HYPERGRAPH::writeDotFile(HYPERGRAPH & hg, char * dotFile, PRINT_TYPE pt)
{
	ofstream of;
	of.open(dotFile);

	// vertices
	map<IDTYPE, VERTEX> vertices = hg.getVertices();

	IDTYPE last_used_id = hg.getLastUsedID();

	of << "graph \"out\" {" << endl;

	for(map<IDTYPE, VERTEX>::iterator it = vertices.begin(); it!=vertices.end(); it++)
	{
		of << "_" << it->first << " [shape=circle, width=.15, label=\"";

		if(nodeLabels[it->first]!=0)
	    		of << nodeLabels[it->first];
		
		of << "\"] \n";
	}

	// isolated nodes
	for(map<int, int>::iterator it = nodeLabels.begin(); it!=nodeLabels.end(); it++)
	{
		if(it->first < 0)
		{
			last_used_id++;
			of << "_" << last_used_id << " [shape=circle, width=.15, label=\"";
			of << it->second;
			of << "\"] \n";
		}
			
	}

	// edges
	map<IDTYPE, EDGE> edges = hg.getEdges();

	for(map<IDTYPE, EDGE>::iterator it = edges.begin(); it!=edges.end(); it++)
	{
	
	string label = it->second.getLabel();
	list<EXPRESSION> attributes = it->second.getAttributes();

	if (attributes.size() == 0 || !coloured)
		of <<  "_" << it->first <<" [label=\"" <<label << " (" << it->first << ") " << "\", shape=box] \n";
	else 
	{
		int maxsize= 0;
		for(list<EXPRESSION>::iterator ite = attributes.begin();
			ite != attributes.end(); ite++)
		{
			if(pt == FULL && ((ite->getExpr()).size() > maxsize) )
				maxsize = (ite->getExpr()).size();
		}

		string line="-";
		maxsize = max(maxsize, (int)label.size());
		if (maxsize<=3)
		 line="----";
		else
		{
		 for (int i=1; i<=maxsize; i++)
			line=line+"-";
		}
		
		of << "_" << it->first << " [label=\"" << label << " (" << it->first << ") " <<"\\n" <<line <<"\\n"; 
		
		for(list<EXPRESSION>::iterator ite = attributes.begin();
			ite != attributes.end(); ite++)
		{
			if(pt == FULL)
			 of << ite->getName() << "=" << ite->getExpr() <<"\\n";
			else
			{
			 // evaluate
			 EXPRESSION e = engine->run(*ite);
			 // print
			 of << ite->getName() << "=" << engine->run(e, pt) <<"\\n";
			
			}
		}

		of << "\", shape=box] \n";
	}
	
	list<IDTYPE> connectedV =  hg.getConnectedVertices(it->first);
	int counter = 0;
	map<int, int> edgeLabelPos;
	const float PI = 3.1415;
	float angle = 0;
	int distance = 0;
	
	for(list<IDTYPE>::iterator itV = connectedV.begin(); itV!=connectedV.end(); itV++)
	{
	int pos = edgeLabelPos[*itV]+1;
	
	
	if ((pos % 2)  == 0)
	{
		angle = atan(tan(35.0*PI/180)/(pos/2))/PI*180;
		distance = pos / 2;
	}
	else
	{
		angle = - atan(tan(35.0*PI/180)/((pos+1)/2))/PI*180;
		distance = (pos+1)/2;
	}
	
	of << "_" << it->first << " -- _" << *itV << "[taillabel=\"" << counter << "\", labelangle=";
				of << angle << ", labeldistance=" << distance << "]\n";
	
				counter++;
				edgeLabelPos[*itV]++;
			}
		}
	
	of << "}\n";

}

// namespaces

}

}
