/***************************************************************************                                                                       *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "vis_petrinet.h"

#include <string.h>

namespace algorithms {

namespace writers {

VIS_PETRINET::VIS_PETRINET()
{
}


VIS_PETRINET::~VIS_PETRINET()
{
}

void VIS_PETRINET::initialize(char * filename)
{
    int n = strlen(filename);

    this->filename = new char[n+1];
    strcpy(this->filename,filename);
}

void VIS_PETRINET::run(PTNET & pn)
{
	string dotFile;
	char * pch;
	char filename_copy[strlen(filename)+1];
	strcpy(filename_copy, filename);
	pch = strtok (filename_copy,".");
		
	while (pch!=NULL && strcmp(pch, "ps")!=0)
  	{
		if(dotFile.size()!=0)
			dotFile += string(".");
		dotFile += (string)pch;
    		pch = strtok (NULL, ".");		
  	}

	dotFile += string(".dot");

	writeDotFile(pn, (char*)dotFile.c_str());

	callDot((char*)dotFile.c_str(), filename);
}

/** write a dot file for this hypergraph */
void VIS_PETRINET::writeDotFile(PTNET & pn, char * dotFile)
{
	ofstream of(dotFile);

	of << "graph \"out\" {" << endl;

	map<IDTYPE, TRANSITION> transitions = pn.getTransitions();
	map<IDTYPE, EDGE> edges = pn.getEdges();

	IDTYPE lastUsedID = pn.getLastUsedID();

	// transitions
	for(map<IDTYPE, TRANSITION>::iterator itt = transitions.begin(); itt!=transitions.end(); itt++)
	{
		of << "_" << itt->first << " [shape=box, label=\"";
		of << itt->second.getLabel();
		of << " (" << itt->first << ") \\n";

		// line and tokens
		if(coloured)
		{
			int maxsize= 0;
			list<EXPRESSION> guards = itt->second.getGuard();
			
			vector<string> str_guards;
			for(list<EXPRESSION>::iterator ite = guards.begin(); ite != guards.end(); ite++)
			{
				if((ite->getExpr()).size() > maxsize)
					maxsize = (ite->getExpr()).size();
	
				str_guards.push_back(ite->getExpr());			
			}
	
			// line
			string line="-";
			maxsize = max(maxsize, (int)itt->second.getLabel().size());
			if (maxsize<=3)
			line="----";
			else
			{
			for (int i=1; i<=maxsize; i++)
				line=line+"-";
			}
	
			if(guards.size() != 0)
			of << line << "\\n";
	
			// guards
			for(vector<string>::iterator its = str_guards.begin(); its != str_guards.end(); its++)
				of << *its << "\\n";
		}

		of << "\"] \n";
	}

	MARKING initial = pn.getInitialMarking();

	// edges
	for(map<IDTYPE, EDGE>::iterator ite = edges.begin(); ite!=edges.end(); ite++)
	{
		of << "_" << ite->first << " [shape=ellipse, label=\"";
		of << ite->second.getLabel();
		of << " (" << ite->first << ") \\n";

		// line and tokens
		int maxsize= 0;
		int summaTokens = 0;
		vector<string> tokens;
		vector<TOKEN> vt = initial.getMarking(ite->first);
		for(vector<TOKEN>::iterator it = vt.begin(); it != vt.end(); it++)
		{
			string token = "*";

			char buf[16];
			sprintf(buf, "%d", it->getCount());
			token += string(buf);

			summaTokens += it->getCount();

			list<EXPRESSION> attributes = it->getAttributes();
			if(attributes.size() != 0)
				token += "\\n";
			for(list<EXPRESSION>::iterator ite = attributes.begin(); ite != attributes.end(); ite++)
			{
				if((ite->getExpr()).size() > maxsize && coloured)
					maxsize = (ite->getExpr()).size();

				token += ite->getName();
				token += ":";
				token += ite->getExpr();
				token += "\\n";
			}

			tokens.push_back(token);
		}

		// line
		string line="-";
		maxsize = max(maxsize, (int)ite->second.getLabel().size());
		if (maxsize<=3)
		 line="----";
		else
		{
		 for (int i=1; i<=maxsize; i++)
			line=line+"-";
		}

		// tokens
		if(coloured)
		{
			for(vector<string>::iterator its = tokens.begin(); its != tokens.end(); its++)
				of << line << "\\n" << *its << "\\n";
		}
		else
		{
			if(summaTokens > 0)
			{
				char buf[16];
				sprintf(buf, "%d", summaTokens);
				of << line << "\\n" << "*" << buf;
			}
		}

		of << "\"] \n";
	}

	// connections
	for(map<IDTYPE, TRANSITION>::iterator itt2 = transitions.begin(); itt2!=transitions.end(); itt2++)
	{
		// preset
		vector<TCONNECTION> c1 = pn.getPreset(itt2->first);
		for(vector<TCONNECTION>::iterator itc1=c1.begin(); itc1!=c1.end(); itc1++)
		{
			if(!coloured || itc1->attributes.size()==0)
			{
				of << "_" << itc1->ID << " -- " << "_" << itt2->first << "[dir=forward, label=\"";
          			of << itc1->weight << "\"]\n";
			}
			else
			{
				lastUsedID++;
				of << "_" << lastUsedID << " [shape=plaintext, label=\"";
				of << itc1->weight << "\\n";
				for(list<EXPRESSION>::iterator ite = itc1->attributes.begin(); ite != itc1->attributes.end(); ite++)
				{
					of << ite->getName() << ":" << ite->getExpr() << "\\n";
				}
				of << "\"]\n";

				// tentacles
				of << "_" << itc1->ID << " -- " << "_" << lastUsedID << "[]\n";
				of << "_" << lastUsedID << " -- " << "_" << itt2->first << "[dir=forward]\n";
			}
		}
		
		// postset
		vector<TCONNECTION> c2 = pn.getPostset(itt2->first);
		for(vector<TCONNECTION>::iterator itc2=c2.begin(); itc2!=c2.end(); itc2++)
		{
			if(!coloured || itc2->attributes.size()==0)
			{
				of << "_" << itt2->first << " -- " << "_" << itc2->ID << "[dir=forward, taillabel=\"";
          			of << itc2->weight << "\"]\n";
			}

			else
			{
				lastUsedID++;
				of << "_" << lastUsedID << " [shape=plaintext, label=\"";
				of << itc2->weight << "\\n";
				for(list<EXPRESSION>::iterator ite = itc2->attributes.begin(); ite != itc2->attributes.end(); ite++)
				{
					of << ite->getName() << ":" << ite->getExpr() << "\\n";
				}
				of << "\"]\n";

				// tentacles
				of << "_" << itt2->first << " -- " << "_" << lastUsedID << "[]\n";
				of << "_" << lastUsedID << " -- " << "_" << itc2->ID << "[dir=forward]\n";
			}

			
		}
	}

	of << "}\n";
}

}

}
