/***************************************************************************
 *   Copyright (C) 2005 by SZS                                             *
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#ifndef ALGORITHM_WRITE_LLNET_CPP
#define ALGORITHM_WRITE_LLNET_CPP

#include "write_llnet.h"
#include <stdio.h>

namespace algorithms
{
   namespace writers
   {
      WRITE_LLNET::WRITE_LLNET()
      {
         this->ID = ALGORITHM_WRITE_LLNET;
         this->name = "write_llnet";
	 this->ptnet = NULL;
      }
      WRITE_LLNET::~WRITE_LLNET() {}
      
      void WRITE_LLNET::initialize( PTNET* i_ptnet,
                                    map<IDTYPE, unsigned>* pp2l, map<unsigned, IDTYPE>* pl2p,
				    map<IDTYPE, unsigned>* tp2l, map<unsigned, IDTYPE>* tl2p  )
      {
         this->ptnet = i_ptnet;
	 this->p_ptnet_to_llnet = pp2l;
	 this->p_llnet_to_ptnet = pl2p;
	 this->t_ptnet_to_llnet = tp2l;
	 this->t_llnet_to_ptnet = tl2p;
      }
      void WRITE_LLNET::initialize(PTNET* i_ptnet)
      {
         this->initialize( i_ptnet,
	                   new map<IDTYPE, unsigned>, new map<unsigned, IDTYPE>,
			   new map<IDTYPE, unsigned>, new map<unsigned, IDTYPE>  );
      }
      
      void WRITE_LLNET::run(char* filename)
      {
         FILE* llout;
	 llout = fopen(filename, "w");
	 if(llout == NULL)
	 {
	    cout << "error writing low level net file \"" << filename << "\".\n";
	    return;
	 }

	 set<IDTYPE> places = this->ptnet->getPlaces();
         set<IDTYPE>::iterator placeit;
	 map<IDTYPE, TRANSITION> transitions = this->ptnet->getTransitions();
         map<IDTYPE, TRANSITION>::iterator transit;
	 MARKING initial_marking = this->ptnet->getInitialMarking();
	 MARKING current_marking = this->ptnet->getCurrentMarking();
	 
         unsigned i, m;
         fprintf(llout, "%s", "PEP\nPTNet\nFORMAT_N\nPL\n");
         i = 0;
         for(placeit = places.begin(); placeit != places.end(); placeit++)
         {
	    i++;
	    this->p_ptnet_to_llnet->insert(pair<IDTYPE,unsigned>(*placeit, i));
	    this->p_llnet_to_ptnet->insert(pair<unsigned,IDTYPE>(i, *placeit));
	    fprintf(llout, "%u%s%u%s", i, "\"p", *placeit, "\"1@1");
	    m = initial_marking.getMarking(*placeit).size();
	    if(m>0) fprintf(llout, "%s%u", "M", m);
	    m = current_marking.getMarking(*placeit).size();
	    if(m>0) fprintf(llout, "%s%u", "n", m);
            fprintf(llout, "%s%u%s", "b\"p", *placeit, "\"\n");
         }
         fprintf(llout, "TR\n");
         i = 0;
         for(transit = transitions.begin(); transit != transitions.end(); transit++)
         {
	    i++;
	    this->t_ptnet_to_llnet->insert(pair<IDTYPE,unsigned>(transit->first, i));
	    this->t_llnet_to_ptnet->insert(pair<unsigned,IDTYPE>(i, transit->first));
	    fprintf(llout, "%u%s%u%s", i, "\"t", transit->first, "\"1@1\n");
         }

         vector<TCONNECTION> connectedP;
	 vector<TCONNECTION>::iterator cPit;
         fprintf(llout, "TP\n");
         for(transit = transitions.begin(); transit != transitions.end(); transit++)
         {
	    i = (*(this->t_ptnet_to_llnet))[transit->first];
	    connectedP = this->ptnet->getPostset(transit->first);
	    for(cPit = connectedP.begin(); cPit != connectedP.end(); cPit++)
	    {
	       unsigned w = cPit->weight;
	       if(w == 1)
	       {
	          fprintf(llout, "%u%s%u\n", i, "<", (*(this->p_ptnet_to_llnet))[cPit->ID]);
	       }
	       else if (w > 1)
	       {
	          fprintf( llout, "%u%s%u%s%u\n", i, "<",
		           (*(this->p_ptnet_to_llnet))[cPit->ID], "w", w );
	       }
	    }
         }
         fprintf(llout, "PT\n");
         for(transit = transitions.begin(); transit != transitions.end(); transit++)
         {
	    i = (*(this->t_ptnet_to_llnet))[transit->first];
	    connectedP = this->ptnet->getPreset(transit->first);
	    for(cPit = connectedP.begin(); cPit != connectedP.end(); cPit++)
	    {
	       unsigned w = cPit->weight;
	       if(w == 1)
	       {
	          fprintf(llout, "%u%s%u\n", (*(this->p_ptnet_to_llnet))[cPit->ID], ">", i);
	       }
	       else if(w > 1)
	       {
	          fprintf( llout, "%u%s%u%s%u\n",
		           (*(this->p_ptnet_to_llnet))[cPit->ID], ">", i, 'w', w );
	       }
	    }
         }

         fclose(llout);
         if(!quiet) cout << "low level net written to " << filename << "\n";
      }     

   };
};
using namespace algorithms;
using namespace algorithms::writers;

#endif
